/*
    Copyright 1991-2015 Amebis

    This file is part of Setup.

    Setup is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Setup is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Setup. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "..\..\include\Setup.h"
#ifdef USE_WINXP_THEMES
#pragma comment(lib, "Comctl32.lib")
#ifdef _WIN64
#pragma comment(linker, "/MANIFESTDEPENDENCY:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='amd64' publicKeyToken='6595b64144ccf1df' language='*'\"")
#else
#pragma comment(linker, "/MANIFESTDEPENDENCY:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='x86' publicKeyToken='6595b64144ccf1df' language='*'\"")
#endif
#endif
#pragma comment(lib, "Urlmon.lib")


////////////////////////////////////////////////////////////////////////////
// Local constants
////////////////////////////////////////////////////////////////////////////

#define IDC_LANGUAGE    8


////////////////////////////////////////////////////////////////////////////
// Local declarations
////////////////////////////////////////////////////////////////////////////

typedef INSTALLUILEVEL (WINAPI* MsiSetInternalUIF)(INSTALLUILEVEL dwUILevel, HWND *phWnd);
typedef UINT (WINAPI* MsiViewExecuteF)(MSIHANDLE hView, MSIHANDLE hRecord);
typedef UINT (WINAPI* MsiViewFetchF)(MSIHANDLE hView, MSIHANDLE *phRecord);
typedef UINT (WINAPI* MsiCloseHandleF)(MSIHANDLE h);
#ifdef _UNICODE
typedef UINT (WINAPI* MsiInstallProductF)(LPCTSTR szPackagePath, LPCTSTR szCommandLine);
typedef UINT (WINAPI* MsiApplyPatchF)(LPCTSTR szPatchPackage, LPCTSTR szInstallPackage, INSTALLTYPE eInstallType, LPCTSTR szCommandLine);
typedef UINT (WINAPI* MsiReinstallProductF)(LPCTSTR szProduct, DWORD dwReinstallMode);
typedef INSTALLSTATE (WINAPI* MsiQueryProductStateF)(LPCTSTR szProduct);
typedef UINT (WINAPI* MsiOpenDatabaseF)(LPCTSTR szDatabasePath, LPCTSTR szPersist, MSIHANDLE *phDatabase);
typedef UINT (WINAPI* MsiDatabaseOpenViewF)(MSIHANDLE hDatabase, LPCTSTR szQuery, MSIHANDLE *phView);
typedef UINT (WINAPI* MsiRecordGetStringF)(MSIHANDLE hRecord, unsigned int uiField, LPTSTR szValue, DWORD *pcchValueBuf);
#else
typedef UINT (WINAPI* MsiInstallProductF)(LPCSTR szPackagePath, LPCSTR szCommandLine);
typedef UINT (WINAPI* MsiApplyPatchF)(LPCSTR szPatchPackage, LPCSTR szInstallPackage, INSTALLTYPE eInstallType, LPCSTR szCommandLine);
typedef UINT (WINAPI* MsiReinstallProductF)(LPCSTR szProduct, DWORD dwReinstallMode);
typedef INSTALLSTATE (WINAPI* MsiQueryProductStateF)(LPCSTR szProduct);
typedef UINT (WINAPI* MsiOpenDatabaseF)(LPCSTR szDatabasePath, LPCSTR szPersist, MSIHANDLE *phDatabase);
typedef UINT (WINAPI* MsiDatabaseOpenViewF)(MSIHANDLE hDatabase, LPCSTR szQuery, MSIHANDLE *phView);
typedef UINT (WINAPI* MsiRecordGetStringF)(MSIHANDLE hRecord, unsigned int uiField, LPSTR szValue, DWORD *pcchValueBuf);
#endif


////////////////////////////////////////////////////////////////////////////
// CDownloadProgress
////////////////////////////////////////////////////////////////////////////

class CDownloadProgress : public IBindStatusCallback
{
public:
    CDownloadProgress();

    void Reset(_In_z_ LPCWSTR pszwMainInstruction);

    DWORD Start(_In_opt_ HWND hWndParent = NULL);
    DWORD End();
    BOOL IsCancelled();

protected:
    // IUnknown
    STDMETHODIMP QueryInterface(_In_ REFIID riid, _Out_ void **ppvObject);
    STDMETHODIMP_(ULONG) AddRef();
    STDMETHODIMP_(ULONG) Release();

    // IBindStatusCallback
    STDMETHOD(OnStartBinding)(_In_ DWORD dwReserved, _In_ IBinding *pib);
    STDMETHOD(GetPriority)(_Out_ LONG *pnPriority);
    STDMETHOD(OnLowResource)(_In_ DWORD reserved);
    STDMETHOD(OnProgress)(_In_ ULONG ulProgress, _In_ ULONG ulProgressMax, _In_ ULONG ulStatusCode, _In_z_ LPCWSTR szStatusText);
    STDMETHOD(OnStopBinding)(_In_ HRESULT hResult, _In_z_ LPCWSTR szError);
    STDMETHOD(GetBindInfo)(_Out_ DWORD *grfBINDF, _Inout_ BINDINFO *pbindinfo);
    STDMETHOD(OnDataAvailable)(_In_ DWORD grfBSCF, _In_ DWORD dwSize, _In_ FORMATETC *pformatetc, _In_ STGMEDIUM *pstgmed);
    STDMETHOD(OnObjectAvailable)(_In_ REFIID riid, _In_ IUnknown *punk);

    // Callback function when used with task dialogs
    static HRESULT CALLBACK TaskDialogCallbackProc(_In_ HWND hWnd, _In_ UINT uNotification, _In_ WPARAM wParam, _In_ LPARAM lParam, _In_ LONG_PTR dwRefData);
    virtual HRESULT TaskDialogCallbackProc(_In_ HWND hWnd, _In_ UINT uNotification, _In_ WPARAM wParam, _In_ LPARAM lParam);

    // UI thread function
    static DWORD WINAPI ThreadProc(_In_ LPVOID lpParameter);

public:
    HWND m_hWnd;    // Handle to task dialog window

protected:
    ATL::CComAutoCriticalSection m_csLock;

    HANDLE m_hThread;
    HWND m_hWndParent;

    ATL::CAtlStringW m_sMainInstruction;

    ULONG m_ulProgress;
    ULONG m_ulProgressMax;
    ULONG m_ulStatusCode;
    //ATL::CAtlStringW m_sContent;
    volatile BOOL m_fCancelled;

    HRESULT m_hResult;
    ATL::CAtlStringW m_sError;

    CComPtr<IBinding> m_pBinding;

    // IUnknown
    ULONG m_uRef;

private:
    BOOL m_fMarqee;
    HANDLE m_hEventReady;
};


////////////////////////////////////////////////////////////////////////////
// Local data
////////////////////////////////////////////////////////////////////////////

static HINSTANCE         g_hInstance       = NULL;                // Module instance handle
static ATL::CAtlString   g_sModulePath;                           // Full hMsiMsg path
static ATL::CAtlString   g_sPath;                                 // Module path (incl. trailing \, excl: filename)
static ATL::CAtlString   g_sFilename;                             // Module file name
static int               g_iArgCount       = 0;                   // Command line parameter count
static LPCWSTR*          g_ppszwArgs       = NULL;                // Command line parameters
static LANGID            g_langIDDefault   = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT); // User default language
static BOOL              g_fWin9X          = FALSE;               // Is Windows 9x?
static int               m_iMajorVersion   = 0;                   // Windows major version
static BOOL              g_fWinx64         = FALSE;               // Is this a 64-bit Windows?
static BOOL              g_fAdmin          = FALSE;               // Is user administrator?
static BOOL              g_fDelayRebootReq = FALSE;               // Is reboot requested?
static ExecutionModeEnum g_emExecMode      = ExecutionModePreset; // Execution mode

static HMODULE               g_hMSI               = NULL;
static MsiSetInternalUIF     g_pfnMsiSetInternalUI     = NULL;
static MsiInstallProductF    g_pfnMsiInstallProduct    = NULL;
static MsiApplyPatchF        g_pfnMsiApplyPatch        = NULL;
static MsiReinstallProductF  g_pfnMsiReinstallProduct  = NULL;
static MsiQueryProductStateF g_pfnMsiQueryProductState = NULL;
static MsiOpenDatabaseF      g_pfnMsiOpenDatabase      = NULL;
static MsiDatabaseOpenViewF  g_pfnMsiDatabaseOpenView  = NULL;
static MsiViewExecuteF       g_pfnMsiViewExecute       = NULL;
static MsiViewFetchF         g_pfnMsiViewFetch         = NULL;
static MsiRecordGetStringF   g_pfnMsiRecordGetString   = NULL;
static MsiCloseHandleF       g_pfnMsiCloseHandle       = NULL;


////////////////////////////////////////////////////////////////////////////
// Local function declarations
////////////////////////////////////////////////////////////////////////////

static DWORD _SetupPackage(_In_ LPCMSIPACKAGE pPackage);
static LPCMSIPACKAGELANGUAGE _GetLanguage(_In_ LPCMSIPACKAGE pPackage, _In_ LANGID langIDDefault);


////////////////////////////////////////////////////////////////////////////
// Main function
////////////////////////////////////////////////////////////////////////////

int CALLBACK WinMain(_In_ HINSTANCE hInstance, _In_ HINSTANCE hPrevInstance, _In_ LPSTR lpCmdLine, _In_ int nCmdShow)
{
    DWORD dwResult = NO_ERROR;

    g_hInstance = hInstance;

    {
        // Get instance filename and path.
        ATL::CAtlString sBufRaw;
        if (::GetModuleFileName(hInstance, sBufRaw)) {
            ATL::CAtlString sBuf;
            if (!::PathCanonicalize(sBuf, sBufRaw))
                sBuf = sBufRaw;
            g_sModulePath = sBuf;
            LPCTSTR pszOffset = _tcsrchr(sBuf, _T('\\'));
            if (pszOffset) {
                pszOffset++;
                g_sFilename.SetString(pszOffset);
                g_sPath.SetString(sBuf, pszOffset - sBuf);
            } else {
                g_sFilename.SetString(g_sModulePath);
                g_sPath.SetString(_T(".\\"));
            }
        } else
            ATLTRACE(atlTraceGeneral, 0, _T("GetModuleFileName failed (%ld).\r\n"), dwResult = ::GetLastError());
    }

    // Get user default language.
    g_langIDDefault = ::GetUserDefaultLangID();

    // Other initializations
    g_fWin9X  = ::IsOSWin9X(&m_iMajorVersion);
    g_fWinx64 = ::IsOSx64();
    g_fAdmin  = ::IsAdmin(g_fWin9X, m_iMajorVersion);

#ifdef USE_WINXP_THEMES
    {
        // Initialize Windows XP visual styles.
        INITCOMMONCONTROLSEX icc;
        icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
        icc.dwICC = ICC_WIN95_CLASSES | ICC_STANDARD_CLASSES;
        ::InitCommonControlsEx(&icc);
    }
#endif

    // Obtain minimum required MSI version.
    // Make sure required MSI version is a valid value -- must be >= 150.
    if (g_sProduct.uiInstallerVersionMin < MINIMUM_SUPPORTED_MSI_VERSION) {
        // Invalid minimum version string.
        ::PostError(::GetDesktopWindow(), IDS_ERR_MSI_INVALID_VER, g_sProduct.uiInstallerVersionMin, MINIMUM_SUPPORTED_MSI_VERSION);
        dwResult = ERROR_INVALID_PARAMETER;
        goto CleanUp;
    }

    g_fDelayRebootReq = FALSE;

    // Figure out what we want to do.
    if ((g_ppszwArgs = (LPCWSTR*)CommandLineToArgvW(GetCommandLineW(), &g_iArgCount)) == NULL) {
        ::PostError(::GetDesktopWindow(), IDS_ERR_PARSING_ARGUMENTS);
        dwResult = ERROR_INVALID_PARAMETER;
        goto CleanUp;
    }

    g_emExecMode = ::GetExecutionMode(g_iArgCount, g_ppszwArgs);
    if (g_emExecMode == ExecutionModeVerify) {
        //
        // We don't want any UI to be displayed in this case. The return value
        // from the EXE is the result of the verification. Therefore, this
        // should be done before initializing the UI.
        //
        switch (::IsPackageTrusted(g_sModulePath, g_ppszwArgs[2], NULL)) {
        case FileTrustWintrustNotInstalled:
            dwResult = (DWORD)TRUST_E_PROVIDER_UNKNOWN;
            goto CleanUp;
        case FileTrustTrusted:
            dwResult = NO_ERROR;
            goto CleanUp;
        case FileTrustUntrusted:
        default:
            dwResult = (DWORD)TRUST_E_SUBJECT_NOT_TRUSTED;
            goto CleanUp;
        }
    }

    if (dwResult == ERROR_BAD_ARGUMENTS || g_emExecMode == ExecutionModeHelp) {
        ::ShowMessageBox(::GetDesktopWindow(), IDS_USAGE, MB_OK | MB_ICONINFORMATION);
        goto CleanUp;
    }

    // Only run one instance at a time.
    if (::AlreadyInProgress(g_fWin9X, m_iMajorVersion)) {
        // Silently return.
        dwResult = ERROR_INSTALL_ALREADY_RUNNING;
        goto CleanUp;
    }

    {
        ATL::CAtlString sMsiDllLocation;

        {
            // Find MSI.DLL path in registry.
            ATL::CRegKey regInst;
            if (regInst.Open(HKEY_LOCAL_MACHINE, _T("Software\\Microsoft\\Windows\\CurrentVersion\\Installer"), KEY_READ) == NO_ERROR) {
                if (::RegQueryStringValue(regInst, _T("InstallerLocation"), sMsiDllLocation) == NO_ERROR) {
                    sMsiDllLocation += sMsiDllLocation.IsEmpty() || ::CharPrev(sMsiDllLocation, (LPCTSTR)sMsiDllLocation + sMsiDllLocation.GetLength())[0] == _T('\\') ? _T("msi.dll") : _T("\\msi.dll");
                    g_hMSI = ::LoadLibrary(sMsiDllLocation);
                }

                regInst.Close();
            }
        }

        if (!g_hMSI) {
            // If reading registry failed, load MSI.DLL using default path.
            g_hMSI = ::LoadLibrary(_T("msi.dll"));
            if (g_hMSI)
                sMsiDllLocation.ReleaseBuffer(::GetModuleFileName(g_hMSI, sMsiDllLocation.GetBuffer(MAX_PATH * 2), MAX_PATH * 2) ? -1 : 0);
        }
        if (!g_hMSI) {
            dwResult = ::GetLastError();
            ::PostError(::GetDesktopWindow(), IDS_ERR_MSI_LOAD, dwResult);
            goto CleanUp;
        }
        if (sMsiDllLocation.IsEmpty()) {
            // Can't obtain version information.
            ATLTRACE(atlTraceGeneral, 0, _T("Can't obtain MSI.DLL location.\r\n"));
            dwResult = ERROR_NOT_FOUND;
            goto CleanUp;
        }

        DWORD dwInstallerVersion;
        if (::GetFileVersionNumber(sMsiDllLocation, &dwInstallerVersion, NULL) != NO_ERROR) {
            // Can't obtain version information.
            ATLTRACE(atlTraceGeneral, 0, _T("Can't obtain version information.\r\n"));
            dwResult = ERROR_INTERNAL_ERROR;
            goto CleanUp;
        }

        UINT uiInstallerVersion = HIWORD(dwInstallerVersion) * 100 + LOWORD(dwInstallerVersion);
        if (uiInstallerVersion < g_sProduct.uiInstallerVersionMin) {
            ATLTRACE(atlTraceGeneral, 2, _T("Windows Installer too old.\r\n"));
            ATL::CAtlString sMain;
            sMain.FormatMessage(IDS_ERR_MSI_OUT_OF_DATE, uiInstallerVersion / 100, uiInstallerVersion % 100, g_sProduct.uiInstallerVersionMin / 100, g_sProduct.uiInstallerVersionMin % 100);
            ::TaskDialog(NULL, g_hInstance, MAKEINTRESOURCEW(IDS_APP_TITLE), sMain, MAKEINTRESOURCEW(IDS_ERR_MSI_OUT_OF_DATE2), TDCBF_CLOSE_BUTTON, TD_ERROR_ICON, NULL);
            dwResult = ERROR_INSTALL_USEREXIT;
            goto CleanUp;
        }
    }

    if (
#ifdef _UNICODE
        (g_pfnMsiInstallProduct    =    (MsiInstallProductF)::GetProcAddress(g_hMSI, "MsiInstallProductW"   )) == NULL ||
        (g_pfnMsiApplyPatch        =        (MsiApplyPatchF)::GetProcAddress(g_hMSI, "MsiApplyPatchW"       )) == NULL ||
        (g_pfnMsiReinstallProduct  =  (MsiReinstallProductF)::GetProcAddress(g_hMSI, "MsiReinstallProductW" )) == NULL ||
        (g_pfnMsiQueryProductState = (MsiQueryProductStateF)::GetProcAddress(g_hMSI, "MsiQueryProductStateW")) == NULL ||
        (g_pfnMsiOpenDatabase      =      (MsiOpenDatabaseF)::GetProcAddress(g_hMSI, "MsiOpenDatabaseW"     )) == NULL ||
        (g_pfnMsiDatabaseOpenView  =  (MsiDatabaseOpenViewF)::GetProcAddress(g_hMSI, "MsiDatabaseOpenViewW" )) == NULL ||
        (g_pfnMsiRecordGetString   =   (MsiRecordGetStringF)::GetProcAddress(g_hMSI, "MsiRecordGetStringW"  )) == NULL ||
#else
        (g_pfnMsiInstallProduct    =    (MsiInstallProductF)::GetProcAddress(g_hMSI, "MsiInstallProductA"   )) == NULL ||
        (g_pfnMsiApplyPatch        =        (MsiApplyPatchF)::GetProcAddress(g_hMSI, "MsiApplyPatchA"       )) == NULL ||
        (g_pfnMsiReinstallProduct  =  (MsiReinstallProductF)::GetProcAddress(g_hMSI, "MsiReinstallProductA" )) == NULL ||
        (g_pfnMsiQueryProductState = (MsiQueryProductStateF)::GetProcAddress(g_hMSI, "MsiQueryProductStateA")) == NULL ||
        (g_pfnMsiOpenDatabase      =      (MsiOpenDatabaseF)::GetProcAddress(g_hMSI, "MsiOpenDatabaseA"     )) == NULL ||
        (g_pfnMsiDatabaseOpenView  =  (MsiDatabaseOpenViewF)::GetProcAddress(g_hMSI, "MsiDatabaseOpenViewA" )) == NULL ||
        (g_pfnMsiRecordGetString   =   (MsiRecordGetStringF)::GetProcAddress(g_hMSI, "MsiRecordGetStringA"  )) == NULL ||
#endif
        (g_pfnMsiSetInternalUI     =     (MsiSetInternalUIF)::GetProcAddress(g_hMSI, "MsiSetInternalUI"     )) == NULL ||
        (g_pfnMsiViewExecute       =       (MsiViewExecuteF)::GetProcAddress(g_hMSI, "MsiViewExecute"       )) == NULL ||
        (g_pfnMsiViewFetch         =         (MsiViewFetchF)::GetProcAddress(g_hMSI, "MsiViewFetch"         )) == NULL ||
        (g_pfnMsiCloseHandle       =       (MsiCloseHandleF)::GetProcAddress(g_hMSI, "MsiCloseHandle"       )) == NULL)
    {
        ::PostError(::GetDesktopWindow(), IDS_ERR_MSI_INVALID);
        dwResult = ERROR_INSTALL_FAILURE;
        goto CleanUp;
    }


    if (dwResult != NO_ERROR)
        goto CleanUp;

    for (LPCMSIPACKAGE pPackage = g_sProduct.pPackages; pPackage && pPackage->pszProductName; pPackage++) {
        if ((dwResult = _SetupPackage(pPackage)) != NO_ERROR)
            break;
    }

CleanUp:
    if (g_ppszwArgs) ::LocalFree(g_ppszwArgs);
    if (g_hMSI)      ::FreeLibrary(g_hMSI);

    return dwResult;
}


////////////////////////////////////////////////////////////////////////////
// CDownloadProgress
////////////////////////////////////////////////////////////////////////////

CDownloadProgress::CDownloadProgress() :
    m_hWnd(NULL),
    m_hThread(NULL),
    m_hWndParent(NULL),
    m_ulProgress(0),
    m_ulProgressMax(0),
    m_ulStatusCode(NO_ERROR),
    m_fCancelled(FALSE),
    m_hResult(S_OK),
    m_uRef(0),
    m_fMarqee(FALSE),
    m_hEventReady(NULL)
{
}


void CDownloadProgress::Reset(_In_z_ LPCWSTR pszwMainInstruction)
{
    m_csLock.Lock();
    m_sMainInstruction = pszwMainInstruction;
    m_ulProgress = 0;
    m_ulProgressMax = 0;
    m_ulStatusCode = NO_ERROR;
    //m_sContent.Empty();
    m_hResult = S_OK;
    m_sError.Empty();
    m_csLock.Unlock();
}


DWORD CDownloadProgress::Start(_In_opt_ HWND hWndParent)
{
    // Prepare and launch the thread.
    m_hWndParent = hWndParent;
    m_hEventReady = ::CreateEvent(NULL, TRUE, FALSE, NULL);
    m_hThread = ::CreateThread(NULL, 0, CDownloadProgress::ThreadProc, this, 0, NULL);

    // Wait for the task dialog to get ready.
    ::WaitForSingleObject(m_hEventReady, INFINITE);
    ::CloseHandle(m_hEventReady);
    m_hEventReady = NULL;

    return m_hThread ? NO_ERROR : ::GetLastError();
}


DWORD CDownloadProgress::End()
{
    if (!m_hThread) {
        // Thread is already finished.
        return NO_ERROR;
    }

    // Instruct task dialog to close.
    if (::IsWindow(m_hWnd))
        SendMessage(m_hWnd, TDM_CLICK_BUTTON, IDCLOSE, 0);

    // Wait for thread to make a clean exit.
    ::WaitForSingleObject(m_hThread, INFINITE);

    // Task dialog is finished; reset it's handle.
    m_hWnd = NULL;

    // Destroy thread remainings.
    ::CloseHandle(m_hThread);
    m_hThread = NULL;

    return NO_ERROR;
}


BOOL CDownloadProgress::IsCancelled()
{
    m_csLock.Lock();
    BOOL fResult = m_fCancelled;
    m_csLock.Unlock();

    return fResult;
}


STDMETHODIMP CDownloadProgress::QueryInterface(_In_ REFIID riid, _Out_ void **ppvObject)
{
    if (riid == IID_IBindStatusCallback)
        *ppvObject = static_cast<IBindStatusCallback*>(this);
    else if (riid == IID_IUnknown)
        *ppvObject = static_cast<IUnknown*>(this);
    else {
        *ppvObject = 0;
        return E_NOINTERFACE;
    }

    AddRef();
    return S_OK;
}


STDMETHODIMP_(ULONG) CDownloadProgress::AddRef()
{
    return ++m_uRef;
}


STDMETHODIMP_(ULONG) CDownloadProgress::Release()
{
    return --m_uRef;
}


STDMETHODIMP CDownloadProgress::OnStartBinding(_In_ DWORD dwReserved, _In_ IBinding *pib)
{
    m_csLock.Lock();
    m_pBinding = pib;
    m_csLock.Unlock();

    return S_OK;
}


STDMETHODIMP CDownloadProgress::GetPriority(_Out_ LONG *pnPriority)
{
    if (pnPriority)
        *pnPriority = THREAD_PRIORITY_NORMAL;

    return S_OK;
}


STDMETHODIMP CDownloadProgress::OnLowResource(_In_ DWORD reserved)
{
    return S_OK;
}


STDMETHODIMP CDownloadProgress::OnProgress(_In_ ULONG ulProgress, _In_ ULONG ulProgressMax, _In_ ULONG ulStatusCode, _In_z_ LPCWSTR szStatusText)
{
    HRESULT hResult;

    // Save data and prepare return value while locked.
    m_csLock.Lock();
    m_ulProgress    = ulProgress;
    m_ulProgressMax = ulProgressMax;
    //m_ulStatusCode  = ulStatusCode;
    //m_sContent      = szStatusText;
    hResult = m_fCancelled ? E_ABORT : S_OK;
    m_csLock.Unlock();

    return hResult;
}


STDMETHODIMP CDownloadProgress::OnStopBinding(_In_ HRESULT hResult, _In_z_ LPCWSTR szError)
{
    // Save data.
    m_csLock.Lock();
    m_hResult = hResult;
    m_sError  = szError;
    m_pBinding = NULL;
    m_csLock.Unlock();

    return S_OK;
}


STDMETHODIMP CDownloadProgress::GetBindInfo(_Out_ DWORD *grfBINDF, _Inout_ BINDINFO *pbindinfo)
{
    *grfBINDF = BINDF_GETNEWESTVERSION | BINDF_NEEDFILE;
    return S_OK;
}


STDMETHODIMP CDownloadProgress::OnDataAvailable(_In_ DWORD grfBSCF, _In_ DWORD dwSize, _In_ FORMATETC *pformatetc, _In_ STGMEDIUM *pstgmed)
{
    return S_OK;
}


STDMETHODIMP CDownloadProgress::OnObjectAvailable(_In_ REFIID riid, _In_ IUnknown *punk)
{
    return S_OK;
}


HRESULT CALLBACK CDownloadProgress::TaskDialogCallbackProc(_In_ HWND hWnd, _In_ UINT uNotification, _In_ WPARAM wParam, _In_ LPARAM lParam, _In_ LONG_PTR dwRefData)
{
    if (!dwRefData) {
        // dwRefData should be a pointer to CDownloadProgress instance class.
        ATLTRACE(atlTraceGeneral, 1, _T(__FUNCTION__) _T(" was called with NULL dwRefData.\r\n"));
        return S_OK;
    }

    // Call instance's TaskDialogCallbackProc() virtual method.
    CDownloadProgress* pProgress = (CDownloadProgress*)dwRefData;
    return pProgress->TaskDialogCallbackProc(hWnd, uNotification, wParam, lParam);
}


HRESULT CDownloadProgress::TaskDialogCallbackProc(_In_ HWND hWnd, _In_ UINT uNotification, _In_ WPARAM wParam, _In_ LPARAM lParam)
{
    if (uNotification == TDN_DIALOG_CONSTRUCTED) {
        // Save hWnd for later reference.
        m_hWnd = hWnd;

        // Signal CDownloadProgress::Start() method task dialog is ready.
        ::SetEvent(m_hEventReady);
    }

    if (uNotification == TDN_DIALOG_CONSTRUCTED || uNotification == TDN_TIMER) {
        // Prepare a local copy of data, and unlock the class ASAP.
        m_csLock.Lock();
        // Show progress only when m_ulProgressMax is not null (avoiding division-by-zero), and m_ulProgress != m_ulProgressMax (downloading does that when file size is unknown).
        int iProgress = m_ulProgressMax && m_ulProgress != m_ulProgressMax ? (int)(pow((double)m_ulProgress / (double)m_ulProgressMax, 1.5) * 100.0) : -1;
        WPARAM wpProgressState = SUCCEEDED(m_hResult) ? PBST_NORMAL : PBST_ERROR;
        ATL::CAtlStringW sMainInstruction(m_sMainInstruction);
        //ATL::CAtlStringW sContent(m_sContent);
        ATL::CAtlStringW sError(m_sError);
        LPCWSTR pszwFooterText = SUCCEEDED(m_hResult) ? NULL : sError.GetString();
        m_csLock.Unlock();

        // Update progress bar.
        if (iProgress >= 0) {
            if (m_fMarqee) {
                // Don't rely on defaults. Set progress bar range to [0,100].
                ::SendMessage(hWnd, TDM_SET_MARQUEE_PROGRESS_BAR, FALSE, 0);
                ::SendMessage(hWnd, TDM_SET_PROGRESS_BAR_MARQUEE, FALSE, 0);
                ::SendMessage(hWnd, TDM_SET_PROGRESS_BAR_RANGE, 0, MAKELPARAM(0, 100));
                m_fMarqee = FALSE;
            }
            ::SendMessage(hWnd, TDM_SET_PROGRESS_BAR_POS, iProgress, 0);
        } else {
            if (!m_fMarqee) {
                ::SendMessage(hWnd, TDM_SET_MARQUEE_PROGRESS_BAR, TRUE, 0);
                ::SendMessage(hWnd, TDM_SET_PROGRESS_BAR_MARQUEE, TRUE, 0);
                m_fMarqee = TRUE;
            }
        }
        ::SendMessage(hWnd, TDM_SET_PROGRESS_BAR_STATE, wpProgressState, 0);

        // Update text and icons.
        ::SendMessage(hWnd, TDM_SET_ELEMENT_TEXT, (WPARAM)TDE_MAIN_INSTRUCTION, (LPARAM)sMainInstruction.GetString());
        //::SendMessage(hWnd, TDM_SET_ELEMENT_TEXT, (WPARAM)TDE_CONTENT,          (LPARAM)sContent.GetString());
        ::SendMessage(hWnd, TDM_SET_ELEMENT_TEXT, (WPARAM)TDE_FOOTER,           (LPARAM)pszwFooterText);
        if (pszwFooterText)
            ::SendMessage(hWnd, TDM_UPDATE_ICON, (WPARAM)TDIE_ICON_FOOTER, (LPARAM)TD_ERROR_ICON);
    }

    if (uNotification == TDN_BUTTON_CLICKED && wParam == IDCANCEL) {
        // User pushed "Cancel" button.
        m_csLock.Lock();
        m_fCancelled = TRUE;
        m_csLock.Unlock();

        // Disable Cancel button, to visually notify user, we acknowledged his cancellation attempt.
        ::SendMessage(hWnd, TDM_ENABLE_BUTTON, IDCANCEL, FALSE);

        return S_FALSE;
    }

    return S_OK;
}


DWORD WINAPI CDownloadProgress::ThreadProc(_In_ LPVOID lpParameter)
{
    // Run task dialog.
    CDownloadProgress* pProgress = (CDownloadProgress*)lpParameter;
    TASKDIALOGCONFIG sProgressDialogCfg = {
        sizeof(sProgressDialogCfg),                                                                        // cbSize
        pProgress->m_hWndParent,                                                                           // hwndParent
        g_hInstance,                                                                                       // hInstance
        TDF_SHOW_PROGRESS_BAR | TDF_SHOW_MARQUEE_PROGRESS_BAR | TDF_CALLBACK_TIMER | TDF_CAN_BE_MINIMIZED, // dwFlags
        TDCBF_CANCEL_BUTTON,                                                                               // dwCommonButtons
        MAKEINTRESOURCEW(IDS_APP_TITLE),                                                                   // pszWindowTitle
        {(HICON)MAKEINTRESOURCEW(IDI_SETUP)},                                                              // hMainIcon/pszMainIcon
        NULL,                                                                                              // pszMainInstruction
        MAKEINTRESOURCEW(IDS_PROGRESS2),                                                                   // pszContent
        0,                                                                                                 // cButtons
        NULL,                                                                                              // pButtons
        IDCANCEL,                                                                                          // nDefaultButton
        0,                                                                                                 // cRadioButtons
        NULL,                                                                                              // pRadioButtons
        0,                                                                                                 // nDefaultRadioButton
        NULL,                                                                                              // pszVerificationText
        NULL,                                                                                              // pszExpandedInformation
        NULL,                                                                                              // pszExpandedControlText
        NULL,                                                                                              // pszCollapsedControlText
        {NULL},                                                                                            // hFooterIcon/pszFooterIcon
        NULL,                                                                                              // pszFooter
        CDownloadProgress::TaskDialogCallbackProc,                                                         // pfCallback
        (LONG_PTR)pProgress,                                                                               // lpCallbackData
        250,                                                                                               // cxWidth
    };
    TaskDialogIndirect(&sProgressDialogCfg, NULL, NULL, NULL);

    return 0;
}


////////////////////////////////////////////////////////////////////////////
// _SetupPackage
////////////////////////////////////////////////////////////////////////////

static DWORD _SetupPackage(_In_ LPCMSIPACKAGE pPackage)
{
    const static TCHAR szDefaultOperation[]    = _T("DEFAULT");
    const static TCHAR szInstallOperation[]    = _T("INSTALL");
    const static TCHAR szMinPatchOperation[]   = _T("MINPATCH");
    const static TCHAR szMajPatchOperation[]   = _T("MAJPATCH");
    const static TCHAR szInstallUpdOperation[] = _T("INSTALLUPD");

    LPCMSIPACKAGELANGUAGE pLanguage = NULL;  // Selected package language

    LPCTSTR pszSetupFile      = NULL;
    LPCTSTR pszOperation      = NULL;
    LPCTSTR pszProperties     = NULL;
    LPCTSTR pszAdminImagePath = NULL;

    BOOL bPatch    = FALSE;
    BOOL bQFE      = FALSE;
    BOOL fDownload = FALSE;

    ATL::CAtlString sText;
    ATL::CAtlString sInstallPath;

    DWORD dwResult = NO_ERROR;

    if (pPackage->fPrompt) {
        // Prompt user to install?
        sText.FormatMessage(IDS_ADDITIONAL_PACKAGE, pPackage->pszProductName);
        if (::ShowMessageBox(::GetDesktopWindow(), sText, MB_YESNO | MB_ICONQUESTION) != IDYES)
            return NO_ERROR;
    }

    if (!pPackage->pLanguages || pPackage->pLanguages->langID == MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL)) {
        // Package has no language available.
        return ERROR_BAD_ARGUMENTS;
    } else if ((pPackage->pLanguages + 1)->langID) {
        // There's more than one language.
        pLanguage = _GetLanguage(pPackage, g_langIDDefault);
        if (!pLanguage)
            return ERROR_INSTALL_USEREXIT;
    } else
        pLanguage = pPackage->pLanguages;

    g_langIDDefault = pLanguage->langID;

    //
    // NOTE:
    // Delay handling admin. installs until we have determined if we are
    // patching an existing install or if we are doing a default install.
    //

    // Determine operation, default (if not present) is INSTALL.
    if ((pszOperation = pPackage->pszOperation) == NULL) {
        // set operation to default which is install
        pszOperation = szDefaultOperation;
    }

    // Determine if this is a patch or a normal install.
    if ((pszSetupFile =
         g_fWinx64 && pLanguage->Packages.Setup.pszWinx64 ? pLanguage->Packages.Setup.pszWinx64 :
        !g_fWin9X                                         ? pLanguage->Packages.Setup.pszWinNT  :
                                                            pLanguage->Packages.Setup.pszWin9X) == NULL)
    {
        // Look for a patch.
        if ((pszSetupFile =
             g_fWinx64 && pLanguage->Packages.Patch.pszWinx64 ? pLanguage->Packages.Patch.pszWinx64 :
            !g_fWin9X                                         ? pLanguage->Packages.Patch.pszWinNT  :
                                                                pLanguage->Packages.Patch.pszWin9X) == NULL)
        {
            ::PostError(::GetDesktopWindow(), IDS_ERR_NO_DATABASE);
            return ERROR_INSTALL_USEREXIT;
        }

        bPatch = TRUE;
    }

    //
    // If we are here, this is either an admin. install or a default install.
    // File signature verification, help and other invalid parameters have
    // already been taken care of above.
    //
    if (g_emExecMode == ExecutionModeAdminInstall)
        pszAdminImagePath = g_ppszwArgs[2];

    //
    // At this point, the validation of the commandline arguments is complete
    // and we have all the information we need.
    //

    // Perform some extra authoring validation.
    if (bPatch &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szMinPatchOperation, -1) != CSTR_EQUAL &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szMajPatchOperation, -1) != CSTR_EQUAL &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szDefaultOperation,  -1) != CSTR_EQUAL)
    {
        // Wrong operation
        ATLTRACE(atlTraceGeneral, 0, _T("Operation \"%s\" is not valid for a patch.\r\n"), pszOperation);
        ::PostError(::GetDesktopWindow(), IDS_ERR_INVALID_OPERATION, pszOperation);
        return ERROR_INVALID_PARAMETER;
    }
    else if (!bPatch &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szInstallOperation,    -1) != CSTR_EQUAL &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szInstallUpdOperation, -1) != CSTR_EQUAL &&
        CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szDefaultOperation,    -1) != CSTR_EQUAL)
    {
        // Wrong operation
        ATLTRACE(atlTraceGeneral, 0, _T("Operation \"%s\" is not valid for a package.\r\n"), pszOperation);
        ::PostError(::GetDesktopWindow(), IDS_ERR_INVALID_OPERATION, pszOperation);
        return ERROR_INVALID_PARAMETER;
    }

    // By now we either have a MSI or a MSP.
    if (           CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szMinPatchOperation,   -1) == CSTR_EQUAL ||
                   CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szInstallUpdOperation, -1) == CSTR_EQUAL ||
        (bPatch && CompareString(LOCALE_INVARIANT , NORM_IGNORECASE, pszOperation, -1, szDefaultOperation,    -1) == CSTR_EQUAL))
        bQFE = TRUE;

    //
    // Start task dialog.
    //
    CDownloadProgress dlgProgress;
    sText.FormatMessage(IDS_PREPARING_SETUP, pPackage->pszProductName);
    dlgProgress.Reset(sText);
    dwResult = dlgProgress.Start(::GetDesktopWindow());
    if (dwResult != NO_ERROR)
        return dwResult;

    //
    // Check if the setup file is present in the same folder as Setup.exe.
    // If setup file is OK, don't download it.
    //

    // Generate the path to the Msi file = g_sModulePath + msiFile
    //   Note: msiFile is a relative path.
    sInstallPath  = g_sPath;
    sInstallPath += pszSetupFile;

    if (::GetFileAttributes(sInstallPath) != INVALID_FILE_ATTRIBUTES) {
        // Set validating signature description.
        sText.FormatMessage(IDS_VALIDATING_SIGNATURE, pszSetupFile);
        dlgProgress.Reset(sText);

        // Validate signature.
        FileTrustEnum itv = ::IsPackageTrusted(g_sModulePath, sInstallPath, dlgProgress.m_hWnd);
        if (dlgProgress.IsCancelled()) {
            dlgProgress.End();
            return ERROR_INSTALL_USEREXIT;
        } else if (itv == FileTrustWintrustNotInstalled) {
            dlgProgress.End();
            ::PostError(::GetDesktopWindow(), IDS_ERR_NO_WINTRUST);
            return ERROR_CALL_NOT_IMPLEMENTED;
        } else if (itv == FileTrustUntrusted) {
            fDownload = TRUE;
        } else if (pPackage->pszVersion) {
            // Setup file exists & signature valid; check if correct version.
            MSIHANDLE
                hDatabase = NULL,
                hView = NULL,
                hRec = NULL;
            TCHAR szVersion[16] = {0};
            DWORD dwVersionLen = _countof(szVersion);

            // Get setup file version.
            if (g_pfnMsiOpenDatabase(sInstallPath, MSIDBOPEN_READONLY, &hDatabase) == NO_ERROR &&
                g_pfnMsiDatabaseOpenView(hDatabase, _T("SELECT `Value` FROM `Property` WHERE `Property`='ProductVersion'"), &hView) == NO_ERROR &&
                g_pfnMsiViewExecute(hView, 0) == NO_ERROR &&
                g_pfnMsiViewFetch(hView, &hRec) == NO_ERROR &&
                g_pfnMsiRecordGetString(hRec, 1, szVersion, &dwVersionLen) == NO_ERROR)
            {
                // Does version match?
                if (CompareString(LOCALE_INVARIANT, NORM_IGNORECASE, pPackage->pszVersion, -1, szVersion, dwVersionLen) != CSTR_EQUAL) {
                    // The setup file is wrong version.
                    fDownload = TRUE;
                }
            } else {
                // Getting setup file version failed; download a new one.
                fDownload = TRUE;
            }

            if (hDatabase) g_pfnMsiCloseHandle(hDatabase);
            if (hView)     g_pfnMsiCloseHandle(hView);
            if (hRec)      g_pfnMsiCloseHandle(hRec);
        }
    } else {
        // Setup file missing.
        fDownload = TRUE;
    }

    if (fDownload) {
        // Download from web.
        if (!pPackage->pszURLAddress) {
            // Download URL not available.
            dlgProgress.End();
            ::PostError(::GetDesktopWindow(), IDS_ERR_NO_URL, pszSetupFile);
            return ERROR_FILE_NOT_FOUND;
        }

        // Download setup file.
        ATL::CAtlString sInstMsiURL;
        HRESULT hResult;

        sInstMsiURL  = pPackage->pszURLAddress;
        sInstMsiURL += pszSetupFile;

        // Set download text.
        if (dlgProgress.IsCancelled()) {
            dlgProgress.End();
            return ERROR_INSTALL_USEREXIT;
        }
        sText.FormatMessage(IDS_DOWNLOADING_PACKAGE, pszSetupFile);
        dlgProgress.Reset(sText);

        // Download file.
        ATLTRACE(atlTraceGeneral, 2, _T("Downloading MSI file \"%s\" for WinVerifyTrust check...\r\n"), sInstMsiURL.GetString());
        hResult = ::URLDownloadToCacheFile(NULL, sInstMsiURL, sInstallPath.GetBuffer(MAX_PATH * 2), MAX_PATH * 2, 0, &dlgProgress);
        if (SUCCEEDED(hResult))
            sInstallPath.ReleaseBuffer();
        else
            sInstallPath.ReleaseBuffer(0);

        // Has user cancelled the download?
        if (dlgProgress.IsCancelled()) {
            dlgProgress.End();
            return ERROR_INSTALL_USEREXIT;
        } else if (FAILED(hResult) || sInstallPath.IsEmpty()) {
            // Error downloading.
            dlgProgress.End();
            ::PostError(::GetDesktopWindow(), IDS_ERR_DOWNLOAD, sInstMsiURL.GetString());
            return ERROR_FILE_NOT_FOUND;
        }

        // Set validating signature description.
        sText.FormatMessage(IDS_VALIDATING_SIGNATURE, pszSetupFile);

        // Validate signature.
        FileTrustEnum itv = ::IsPackageTrusted(g_sModulePath, sInstallPath, dlgProgress.m_hWnd);
        if (dlgProgress.IsCancelled()) {
            dlgProgress.End();
            return ERROR_INSTALL_USEREXIT;
        } else if (itv == FileTrustWintrustNotInstalled) {
            dlgProgress.End();
            ::PostError(::GetDesktopWindow(), IDS_ERR_NO_WINTRUST);
            return ERROR_CALL_NOT_IMPLEMENTED;
        } else if (itv == FileTrustUntrusted) {
            dlgProgress.End();
            ::PostError(::GetDesktopWindow(), IDS_ERR_UNTRUSTED, pszSetupFile);
            return HRESULT_CODE(TRUST_E_SUBJECT_NOT_TRUSTED);
        }
    }

    //
    // Good to go -- terminate our UI and let the Windows Installer take over.
    //
    dlgProgress.End();

    // Retrieve the optional command line PROPERTY = VALUE strings if available.
    pszProperties = pPackage->pszAdditionalProperties;

    //
    // Perform install -- use registered location of msi.dll so that we can make use of the "delay-reboot"
    //                    upgrade MSI feature.
    //
    ATLTRACE(atlTraceGeneral, 2, _T("Setting Internal UI level to FULL...\r\n"));
    g_pfnMsiSetInternalUI(pPackage->UILevel, 0);

    if (!bPatch) {
        // Performing install or reinstall/recache.
        ATLTRACE(atlTraceGeneral, 2, _T("Calling MsiInstallProduct with sInstallPath=\"%s\" and szCommandLine=\"%s\"...\r\n"), sInstallPath.GetString(), pszProperties ? pszProperties : _T("{null}"));

        // Default operation for a package is INSTALL.
        if (bQFE) {
            MSIHANDLE
                hDatabase = NULL,
                hView = NULL,
                hRec = NULL;
            TCHAR szProductCode[40] = {0};
            DWORD dwProductCodeSize = _countof(szProductCode);

            // Check to see if this product is already installed.
            if (g_pfnMsiOpenDatabase(sInstallPath, MSIDBOPEN_READONLY, &hDatabase) == NO_ERROR &&
                g_pfnMsiDatabaseOpenView(hDatabase, _T("SELECT `Value` FROM `Property` WHERE `Property`='ProductCode'"), &hView) == NO_ERROR &&
                g_pfnMsiViewExecute(hView, 0) == NO_ERROR &&
                g_pfnMsiViewFetch(hView, &hRec) == NO_ERROR &&
                g_pfnMsiRecordGetString(hRec, 1, szProductCode, &dwProductCodeSize) == NO_ERROR)
            {
                INSTALLSTATE isProduct = INSTALLSTATE_UNKNOWN;

                isProduct = g_pfnMsiQueryProductState(szProductCode);
                ATLTRACE(atlTraceGeneral, 2, _T("MsiQueryProductState returned %d.\r\n"), isProduct);
                if (isProduct != INSTALLSTATE_ADVERTISED && isProduct != INSTALLSTATE_DEFAULT) {
                    // Product is unknown, so this will be a first time install.
                    ATLTRACE(atlTraceGeneral, 2, _T("The product code \"%s\" is unknown. Will use first time install logic...\r\n"), szProductCode);
                    bQFE = FALSE;
                } else {
                    // Product is known, use QFE syntax.
                    ATLTRACE(atlTraceGeneral, 2, _T("The product code \"%s\" is known. Will use QFE recache and reinstall upgrade logic...\r\n"), szProductCode);
                }
            } else {
                // Some failure occurred when processing the product code, so treat as non-QFE.
                ATLTRACE(atlTraceGeneral, 2, _T("Unable to process product code. Will treat as first time install...\r\n"));
                bQFE = FALSE;
            }

            if (hDatabase) g_pfnMsiCloseHandle(hDatabase);
            if (hView)     g_pfnMsiCloseHandle(hView);
            if (hRec)      g_pfnMsiCloseHandle(hRec);
        }

        //
        // Set up the properties to be passed into MSIInstallProduct.
        //
        ATL::CAtlString sInstProperties;
        if (bQFE && !pszProperties)
            sInstProperties = _T("REINSTALL=ALL REINSTALLMODE=vomus");
        else if (pszProperties)
            sInstProperties = pszProperties;
        if (g_emExecMode == ExecutionModeAdminInstall) {
            sInstProperties += _T(" ACTION=ADMIN");
            if (pszAdminImagePath) {
                sInstProperties += _T(" TARGETDIR=\"");
                sInstProperties += pszAdminImagePath;
                sInstProperties += _T("\"");
            }
        }

        dwResult = g_pfnMsiInstallProduct(sInstallPath, sInstProperties);
        ATLTRACE(atlTraceGeneral, 2, _T("MsiInstallProduct returned %d.\r\n"), dwResult);
    } else {
        // Default Operation for a patch is MINPATCH.

        // If pszProperties is NULL, use our default value for QFE patches.
        if (!pszProperties && bQFE) {
            const static TCHAR szDefaultMinPatchCommandLine[] = _T("REINSTALL=ALL REINSTALLMODE=omus");
            ATLTRACE(atlTraceGeneral, 2, _T("Patch is a MINPATCH (small or minor update patch) so using default command line \"%s\".\r\n"), szDefaultMinPatchCommandLine);
            pszProperties = szDefaultMinPatchCommandLine;
        }

        if (g_emExecMode == ExecutionModeAdminInstall) {
            // Performing a patch.
            ATLTRACE(atlTraceGeneral, 2, _T("Calling MsiApplyPatch with szPatchPackage=\"%s\" and szInstallPackage=\"%s\" and eInstallType=INSTALLTYPE_NETWORK_IMAGE and szCommandLine=\"%s\"...\r\n"), sInstallPath, pszAdminImagePath, pszProperties ? pszProperties : _T("{null}"));
            dwResult = g_pfnMsiApplyPatch(sInstallPath, pszAdminImagePath, INSTALLTYPE_NETWORK_IMAGE, pszProperties);
        } else {
            // Performing a patch.
            ATLTRACE(atlTraceGeneral, 2, _T("Calling MsiApplyPatch with szPatchPackage=\"%s\" and szInstallPackage=\"{null}\" and eInstallType=INSTALLTYPE_DEFAULT and szCommandLine=\"%s\"...\r\n"), (LPCTSTR)sInstallPath, pszProperties ? pszProperties : _T("{null}"));
            dwResult = g_pfnMsiApplyPatch(sInstallPath, NULL, INSTALLTYPE_DEFAULT, pszProperties);
        }
        ATLTRACE(atlTraceGeneral, 2, _T("MsiApplyPatch returned %d.\r\n"), dwResult);
    }

    if (dwResult != NO_ERROR)
        ::PostMSIError(::GetDesktopWindow(), dwResult);

    return dwResult;
}


////////////////////////////////////////////////////////////////////////////
// _GetLanguage
////////////////////////////////////////////////////////////////////////////

static LPCMSIPACKAGELANGUAGE _GetLanguage(_In_ LPCMSIPACKAGE pPackage, _In_ LANGID langIDDefault)
{
    ATL::CAtlArray<TASKDIALOG_BUTTON> tButtons;
    LPCMSIPACKAGELANGUAGE p;
    int iButtonID, iDefaultButton = IDCANCEL;

    // Prepare language buttons.
    for (p = pPackage->pLanguages, iButtonID = IDC_LANGUAGE; p && p->langID != MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL); p++, iButtonID++) {
        TASKDIALOG_BUTTON sButton = {
            iButtonID,
            p->pszLanguageName
        };
        tButtons.Add(sButton);

        if (p->langID == g_langIDDefault) {
            // Set button corresponding to default language to default.
            iDefaultButton = iButtonID;
        }
    }

    ATL::CAtlString sMain;
    sMain.FormatMessage(IDS_LANGUAGE, pPackage->pszProductName);

    TASKDIALOGCONFIG sLanguageDialogCfg = {
        sizeof(sLanguageDialogCfg),           // cbSize
        ::GetDesktopWindow(),                 // hwndParent
        g_hInstance,                          // hInstance
        TDF_USE_COMMAND_LINKS,                // dwFlags
        TDCBF_CANCEL_BUTTON,                  // dwCommonButtons
        MAKEINTRESOURCEW(IDS_APP_TITLE),      // pszWindowTitle
        {(HICON)MAKEINTRESOURCEW(IDI_SETUP)}, // hMainIcon/pszMainIcon
        sMain,                                // pszMainInstruction
        MAKEINTRESOURCEW(IDS_LANGUAGE2),  // pszContent
        (int)tButtons.GetCount(),             // cButtons
        tButtons.GetData(),                   // pButtons
        iDefaultButton,                       // nDefaultButton
        //0,                                    // cRadioButtons
        //NULL,                                 // pRadioButtons
        //0,                                    // nDefaultRadioButton
        //NULL,                                 // pszVerificationText
        //NULL,                                 // pszExpandedInformation
        //NULL,                                 // pszExpandedControlText
        //NULL,                                 // pszCollapsedControlText
        //{NULL},                               // hFooterIcon/pszFooterIcon
        //NULL,                                 // pszFooter
        //NULL,                                 // pfCallback
        //NULL,                                 // lpCallbackData
        //0,                                    // cxWidth
    };
    if (SUCCEEDED(::TaskDialogIndirect(&sLanguageDialogCfg, &iButtonID, NULL, NULL))) {
        if (iButtonID == IDCANCEL) {
            // User cancelled.
            return NULL;
        } else if (IDC_LANGUAGE <= iButtonID && iButtonID < IDC_LANGUAGE + (int)tButtons.GetCount()) {
            // User selected a language.
            return pPackage->pLanguages + (iButtonID - IDC_LANGUAGE);
        } else {
            // Unknown command ID?!
            ATLTRACE(atlTraceGeneral, 0, _T("TaskDialogIndirect returned unknown command (%d).\r\n"), iButtonID);
            return NULL;
        }
    }

    return NULL;
}
