/*
    Copyright 1991-2015 Amebis

    This file is part of Setup.

    Setup is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Setup is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Setup. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SETUP_H__
#define __SETUP_H__


////////////////////////////////////////////////////////////////////////////
// Constants
////////////////////////////////////////////////////////////////////////////

#define SETUP_VERSION           0x03ff0300

#define SETUP_VERSION_MAJ       3
#define SETUP_VERSION_MIN       255
#define SETUP_VERSION_REV       3
#define SETUP_VERSION_BUILD     0

#define SETUP_VERSION_STR       "4.0-pre3"
#define SETUP_BUILD_YEAR_STR    "2015"


////////////////////////////////////////////////////////////////////
// Resource codes
////////////////////////////////////////////////////////////////////

#define IDI_SETUP                       100

#define IDS_APP_TITLE                   100
#define IDS_USAGE                       101
#define IDS_ADDITIONAL_PACKAGE          102
#define IDS_LANGUAGE                    103
#define IDS_LANGUAGE2                   104
#define IDS_PROGRESS2                   105
#define IDS_PREPARING_SETUP             106
#define IDS_VALIDATING_SIGNATURE        107
#define IDS_DOWNLOADING_PACKAGE         108
#define IDS_ERR_PARSING_ARGUMENTS       109
#define IDS_ERR_MSI_INVALID_VER         110
#define IDS_ERR_MSI_LOAD                111
#define IDS_ERR_MSI_INVALID             112
#define IDS_ERR_MSI_OUT_OF_DATE         113
#define IDS_ERR_MSI_OUT_OF_DATE2        114
#define IDS_ERR_NO_DATABASE             115
#define IDS_ERR_INVALID_OPERATION       116
#define IDS_ERR_NO_WINTRUST             117
#define IDS_ERR_UNTRUSTED               118
#define IDS_ERR_NO_URL                  119
#define IDS_ERR_DOWNLOAD                120
#define IDS_ERR_NO_RESOURCE             121
#define IDS_ERR_INSTALL                 122

#ifndef RC_INVOKED

#include <Windows.h>

#include <atlstr.h>

#include <Msi.h>


////////////////////////////////////////////////////////////////////
// Constants
////////////////////////////////////////////////////////////////////

#define MINIMUM_SUPPORTED_MSI_VERSION   150


////////////////////////////////////////////////////////////////////
// FileTrustEnum
////////////////////////////////////////////////////////////////////

enum FileTrustEnum {
    FileTrustWintrustNotInstalled = 0,
    FileTrustTrusted              = 1,
    FileTrustUntrusted            = 2,
};


////////////////////////////////////////////////////////////////////
// ExecutionModeEnum
////////////////////////////////////////////////////////////////////

enum ExecutionModeEnum {
    ExecutionModePreset       = 0,
    ExecutionModeHelp         = 1,
    ExecutionModeVerify       = 2,
    ExecutionModeAdminInstall = 3,
};


////////////////////////////////////////////////////////////////////
// MSIPACKAGELANGUAGE
////////////////////////////////////////////////////////////////////

typedef struct _MSIPACKAGELANGUAGE {
    LANGID langID;                  // MS language code of the package
    LPCTSTR pszLanguageName;        // Language name (in original language)
    struct {
        struct {
            LPCTSTR pszWin9X;       // Setup MSI package filename (for Windows 95/98/ME)
            LPCTSTR pszWinNT;       // Setup MSI package filename (for Windows NT/2000/XP ...)
            LPCTSTR pszWinx64;      // Setup MSI package filename (for Windows XP/Vista 64-bit ...)
        } Setup;
        struct {
            LPCTSTR pszWin9X;       // Patch MSI package filename (for Windows 95/98/ME)
            LPCTSTR pszWinNT;       // Patch MSI package filename (for Windows NT/2000/XP ...)
            LPCTSTR pszWinx64;      // Patch MSI package filename (for Windows XP/Vista 64-bit ...)
        } Patch;
    } Packages;
} MSIPACKAGELANGUAGE, *PMSIPACKAGELANGUAGE, FAR* LPMSIPACKAGELANGUAGE;
typedef const MSIPACKAGELANGUAGE FAR* LPCMSIPACKAGELANGUAGE;


////////////////////////////////////////////////////////////////////
// MSIPACKAGE
////////////////////////////////////////////////////////////////////

typedef struct _MSIPACKAGE {
    LPCTSTR pszURLAddress;              // URL root address of MSI packages
    LPCTSTR pszOperation;               // Operation for setup to perform ("DEFAULT", "INSTALL", "MINPATCH", "MAJPATCH", "INSTALLUPD")
    LPCTSTR pszProductName;             // Product name (language independant if possible)
    LPCTSTR pszAdditionalProperties;    // Additional MSI properties
    LPCTSTR pszVersion;                 // Product version
    BOOL fPrompt;                       // Prompt user to install?
    INSTALLUILEVEL UILevel;             // Setup verbosity
    LPCMSIPACKAGELANGUAGE pLanguages;   // Pointer to language-specific package table (terminated with 0 language)
} MSIPACKAGE, *PMSIPACKAGE, FAR *LPMSIPACKAGE;
typedef const MSIPACKAGE FAR* LPCMSIPACKAGE;


////////////////////////////////////////////////////////////////////
// MSIPRODUCT
////////////////////////////////////////////////////////////////////

typedef struct _MSIPRODUCT {
    UINT uiInstallerVersionMin; // Minimum Microsoft Installer version required to install this product
    LPCMSIPACKAGE pPackages;    // Pointer to install packages (NULL terminated)
} MSIPRODUCT, *PMSIPRODUCT, FAR* LPMSIPRODUCT;
typedef const MSIPRODUCT FAR* LPCMSIPRODUCT;

extern const MSIPRODUCT g_sProduct;


////////////////////////////////////////////////////////////////////
// Utility functions
////////////////////////////////////////////////////////////////////

ExecutionModeEnum GetExecutionMode(_In_ int iArgCount, _In_count_(iArgCount) LPCWSTR* g_ppszwArgs);
BOOL AlreadyInProgress(_In_ BOOL fWin9X, _In_ int iMajorVersion);
BOOL IsTerminalServerInstalled(_In_ BOOL fWin9X, _In_ int iMajorVersion);
DWORD GetFileVersionNumber(_In_z_ LPCTSTR pszFilename, _Out_opt_ DWORD *pdwMSVer, _Out_opt_ DWORD *pdwLSVer);
BOOL IsOSWin9X(_Out_opt_ int *piMajorVersion);
BOOL IsOSx64();
BOOL IsAdmin(_In_ BOOL fWin9X, _In_ int iMajorVersion);
BOOL AcquireShutdownPrivilege();
void PostError(_In_ HWND hWndParent, _In_ UINT uiErrorID, ...);
void PostMSIError(_In_ HWND hWndParent, _In_ UINT uiErrorID);
int ShowMessageBox(_In_ HWND hWndParent, _In_z_ LPCTSTR lpszText, _In_opt_ UINT nType = MB_OK, _In_opt_ UINT nIDHelp = 0);
int ShowMessageBox(_In_ HWND hWndParent, _In_ UINT nIDPrompt, _In_opt_ UINT nType = MB_OK, _In_opt_ UINT nIDHelp = 0);


////////////////////////////////////////////////////////////////////
// WinVerifyTrust functions
////////////////////////////////////////////////////////////////////

FileTrustEnum IsFileTrusted(_In_z_ LPCWSTR pszwFile, _In_ HWND hwndParent, _In_ DWORD dwUIChoice, _Out_opt_ BOOL *pfIsSigned, _Out_opt_ PCCERT_CONTEXT *ppcSigner);
FileTrustEnum IsPackageTrusted(_In_z_ LPCTSTR pszSetupExe, _In_z_ LPCTSTR pszPackage, _In_ HWND hwndParent);

#endif // RC_INVOKED

#endif // __SETUP_H__
