/*
    Copyright 1991-2015 Amebis

    This file is part of Setup.

    Setup is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Setup is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Setup. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#pragma comment(lib, "Version.lib")


////////////////////////////////////////////////////////////////////////////
// GetExecutionMode
////////////////////////////////////////////////////////////////////////////

ExecutionModeEnum GetExecutionMode(_In_ int iArgCount, _In_count_(iArgCount) LPCWSTR* g_ppszwArgs)
{
    if (iArgCount < 2)
        return ExecutionModePreset;

    if (g_ppszwArgs[1][0] == _T('/') || g_ppszwArgs[1][0] == _T('-')) {
        switch (::CharNext(g_ppszwArgs[1])[0]) {
        case _T('a'):
        case _T('A'):
            return iArgCount >= 3 ? ExecutionModeAdminInstall : ExecutionModeHelp;
        case _T('v'):
        case _T('V'):
            return iArgCount >= 3 ? ExecutionModeVerify : ExecutionModeHelp;
        default:
            return ExecutionModeHelp;
        }
    } else
        return ExecutionModeHelp;
}


/////////////////////////////////////////////////////////////////////////////
// AlreadyInProgress
//
//  Attempts to create the MSISETUP mutex. Returns TRUE
//  if mutex already exists or failed to create mutex
////////////////////////////////////////////////////////////////////////////

BOOL AlreadyInProgress(_In_ BOOL fWin9X, _In_ int iMajorVersion)
{
    // If Windows 2000 or greater or Terminal Server installed, must use Global prefix.
    return !::CreateMutex(NULL, FALSE,
        !fWin9X && iMajorVersion >= 5 || ::IsTerminalServerInstalled(fWin9X, iMajorVersion) ?
            _T("Global\\_MSISETUP_{2956EBA1-9B5A-4679-8618-357136DA66CA}") :
            _T("_MSISETUP_{2956EBA1-9B5A-4679-8618-357136DA66CA}")) || ::GetLastError() == ERROR_ALREADY_EXISTS ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////////////////
// IsTerminalServerInstalled
//
//  Determines whether terminal services are installed
////////////////////////////////////////////////////////////////////////////

BOOL IsTerminalServerInstalled(_In_ BOOL fWin9X, _In_ int iMajorVersion)
{
    DWORD dwSize = 0;
    HKEY  hKey   = NULL;
    DWORD dwType = 0;

    // Win9X is not terminal server.
    if (fWin9X)
        return FALSE;

    BOOL fIsTerminalServer = FALSE;

    // On Windows 2000 and greater, the ProductSuite "Terminal Server"
    // value will always be present. Use GetVersionEx to get the right
    // answer.
    if (iMajorVersion > 4) {
        OSVERSIONINFOEX osVersionInfo = { sizeof(osVersionInfo) };
        if (::GetVersionEx((OSVERSIONINFO*)&osVersionInfo) &&
             (osVersionInfo.wSuiteMask & VER_SUITE_TERMINAL    ) &&
            !(osVersionInfo.wSuiteMask & VER_SUITE_SINGLEUSERTS))
            fIsTerminalServer = TRUE;
    } else {
        // Other NT versions, check the registry key
        // If the value we want exists and has a non-zero size...
        if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("System\\CurrentControlSet\\Control\\ProductOptions"), 0, KEY_READ, &hKey) == NO_ERROR &&
            RegQueryValueEx(hKey, _T("ProductSuite"), NULL, &dwType, NULL, &dwSize) == NO_ERROR &&
            dwSize > 0 &&
            dwType == REG_MULTI_SZ)
        {
            TCHAR *szSuiteList = new TCHAR[dwSize / sizeof(TCHAR)];
            if (szSuiteList) {
                ::ZeroMemory(szSuiteList, dwSize);
                if (RegQueryValueEx(hKey, _T("ProductSuite"), NULL, &dwType, (LPBYTE)szSuiteList, &dwSize) == NO_ERROR) {
                    static const TCHAR s_szTSSearchStr[] = _T("Terminal Server"); // Not localized
                    LPCTSTR szSubString = szSuiteList;              // pointer to current substring

                    while (*szSubString) { // Break on consecutive zero bytes
                        SIZE_T nSubStringLen = _tcslen(szSubString);
                        if (nSubStringLen == _countof(s_szTSSearchStr) - 1 && wmemcmp(szSubString, s_szTSSearchStr, _countof(s_szTSSearchStr) - 1) == 0) {
                            fIsTerminalServer = TRUE;
                            break;
                        }

                        // Substring does not match, skip forward the length of the substring
                        // plus 1 for the terminating null.
                        szSubString += nSubStringLen + 1;
                    }
                }

                delete [] szSuiteList;
            }
        }

        if (hKey)
            RegCloseKey(hKey);
    }

    return fIsTerminalServer;
}


////////////////////////////////////////////////////////////////////////////
// GetFileVersionNumber
////////////////////////////////////////////////////////////////////////////

DWORD GetFileVersionNumber(_In_z_ LPCTSTR pszFilename, _Out_opt_ DWORD *pdwMSVer, _Out_opt_ DWORD *pdwLSVer)
{
    DWORD dwResult = NO_ERROR;

    if (pdwMSVer)
        *pdwMSVer = MAXDWORD;
    if (pdwLSVer)
        *pdwLSVer = MAXDWORD;

    // Read version info.
    ATL::CAtlArray<BYTE> aVersionInfo;
    if (::GetFileVersionInfo(pszFilename, 0, aVersionInfo)) {
        // Get the value for translation.
        VS_FIXEDFILEINFO *lpVSFixedFileInfo = NULL;
        UINT uiLen;
        if (::VerQueryValue(aVersionInfo.GetData(), _T("\\"), (LPVOID*)&lpVSFixedFileInfo, &uiLen) && uiLen != 0) {
            // Save version numbers.
            if (pdwMSVer)
                *pdwMSVer = lpVSFixedFileInfo->dwFileVersionMS;
            if (pdwLSVer)
                *pdwLSVer = lpVSFixedFileInfo->dwFileVersionLS;

            ATLTRACE(atlTraceGeneral, 2, _T("%s is version %d.%d.%d.%d\r\n"), pszFilename,
                HIWORD(lpVSFixedFileInfo->dwFileVersionMS), LOWORD(lpVSFixedFileInfo->dwFileVersionMS),
                HIWORD(lpVSFixedFileInfo->dwFileVersionLS), LOWORD(lpVSFixedFileInfo->dwFileVersionLS));
        } else
            ATLTRACE(atlTraceGeneral, 0, _T("VerQueryValue failed (%ld).\r\n"), dwResult = ERROR_NOT_FOUND);
    } else
        ATLTRACE(atlTraceGeneral, 0, _T("GetFileVersionInfo failed (%ld).\r\n"), dwResult = ::GetLastError());

    return dwResult;
}


////////////////////////////////////////////////////////////////////////////
// IsOSWin9X
//
//  Returns TRUE if running on a Win9X platform
//  Returns FALSE if running on a WinNT platform
////////////////////////////////////////////////////////////////////////////

BOOL IsOSWin9X(_Out_opt_ int *piMajorVersion)
{
    OSVERSIONINFO sInfoOS = { sizeof(sInfoOS) };
    ::GetVersionEx(&sInfoOS);

    if (piMajorVersion)
        *piMajorVersion = sInfoOS.dwMajorVersion;

    if (VER_PLATFORM_WIN32_NT == sInfoOS.dwPlatformId)
        return FALSE;
    else
        return TRUE;
}


////////////////////////////////////////////////////////////////////////////
// IsOSx64
//
//  Returns TRUE if running on a 64-bit platform
//  Returns FALSE if running on a 32-bit platform
////////////////////////////////////////////////////////////////////////////

BOOL IsOSx64()
{
#if defined(_WIN64)
    // 64-bit versions can only run on 64-bit OS.
    return TRUE;
#else
    typedef BOOL (WINAPI *PFnIsWow64Process)(HANDLE, PBOOL);
    PFnIsWow64Process pfnIsWow64Process = (PFnIsWow64Process)::GetProcAddress(::GetModuleHandle(_T("kernel32.dll")), "IsWow64Process");
    if (pfnIsWow64Process != NULL) {
        BOOL fIsWow64 = FALSE;
        if (pfnIsWow64Process(::GetCurrentProcess(), &fIsWow64)) {
            // Return the result.
            return fIsWow64;
        } else {
            // The IsWow64Process() function failed. Assume, it's not a 64-bit platform.
            return FALSE;
        }
    } else {
        // The IsWow64Process() was not found. Old OS, definitely not 64-bit one.
        return FALSE;
    }
#endif
}


////////////////////////////////////////////////////////////////////////////
// IsAdmin
//
//  Returns TRUE if current user is an administrator (or if on Win9X)
//  Returns FALSE if current user is not an adminstrator
//
//  implemented as per KB Q118626
////////////////////////////////////////////////////////////////////////////

BOOL IsAdmin(_In_ BOOL fWin9X, _In_ int iMajorVersion)
{
    if (fWin9X) {
        // Always admin on Win9X!
        return TRUE;
    }

    // Get the administrator SID.
    PSID psidAdministrators;
    SID_IDENTIFIER_AUTHORITY siaNtAuthority = SECURITY_NT_AUTHORITY;
    if (!::AllocateAndInitializeSid(&siaNtAuthority, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, &psidAdministrators))
        return FALSE;

    // On NT5, use the CheckTokenMembershipAPI to correctly handle cases where
    // the Adiminstrators group might be disabled. fIsAdmin is BOOL for...
    BOOL fIsAdmin = FALSE;
    if (iMajorVersion >= 5) {
        // CheckTokenMembership checks if the SID is enabled in the token. NULL for
        // the token means the token of the current thread. Disabled groups, restricted
        // SIDS, and SE_GROUP_USE_FOR_DENY_ONLY are all considered. If the function
        // returns FALSE, ignore the result.
        HMODULE hAdvapi32 = ::LoadLibrary(_T("advapi32.dll"));
        if (hAdvapi32) {
            typedef BOOL (WINAPI* PFnCheckTokenMembership)(HANDLE TokenHandle, PSID SidToCheck, PBOOL IsMember);
            PFnCheckTokenMembership pfnCheckTokenMembership = (PFnCheckTokenMembership)::GetProcAddress(hAdvapi32, "CheckTokenMembership");
            if (!pfnCheckTokenMembership || !pfnCheckTokenMembership(NULL, psidAdministrators, &fIsAdmin))
                fIsAdmin = FALSE;

            ::FreeLibrary(hAdvapi32);
        } else
            fIsAdmin = FALSE;
    } else {
        // NT4, check groups of user.
        HANDLE hAccessToken = 0;
        BYTE *pbInfoBuffer = new BYTE[1024]; // may need to resize if TokenInfo too big
        DWORD dwInfoBufferSize = 1024;

        if (pbInfoBuffer && ::OpenProcessToken(::GetCurrentProcess(), TOKEN_READ, &hAccessToken)) {
            BOOL fSuccess = FALSE;
            DWORD dwRetInfoBufferSize = 0;
            fSuccess = ::GetTokenInformation(hAccessToken, TokenGroups, pbInfoBuffer, dwInfoBufferSize, &dwRetInfoBufferSize) == TRUE;

            if(dwRetInfoBufferSize > dwInfoBufferSize) {
                delete [] pbInfoBuffer;
                pbInfoBuffer = new BYTE[dwRetInfoBufferSize];
                if (pbInfoBuffer) {
                    dwInfoBufferSize = dwRetInfoBufferSize;
                    fSuccess = ::GetTokenInformation(hAccessToken, TokenGroups, pbInfoBuffer, dwInfoBufferSize, &dwRetInfoBufferSize) == TRUE;
                }
            }

            ::CloseHandle(hAccessToken);

            if (fSuccess) {
                PTOKEN_GROUPS ptgGroups = (PTOKEN_GROUPS)pbInfoBuffer;
                for (DWORD x = 0; x < ptgGroups->GroupCount; x++) {
                    if (::EqualSid(psidAdministrators, ptgGroups->Groups[x].Sid)) {
                        fIsAdmin = TRUE;
                        break;
                    }
                }
            }

            delete [] pbInfoBuffer;
        }
    }

    ::FreeSid(psidAdministrators);
    return fIsAdmin ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////////////////
// AcquireShutdownPrivilege
//
//  Attempts to enable the SE_SHUTDOWN_NAME privilege in the process token
////////////////////////////////////////////////////////////////////////////

BOOL AcquireShutdownPrivilege()
{
    HANDLE hToken = 0;
    TOKEN_PRIVILEGES tkp;

    HMODULE hAdvapi32 = ::LoadLibrary(_T("advapi32.dll"));
    if (hAdvapi32)
        return FALSE;

    typedef BOOL (WINAPI* PFnOpenProcessToken)(HANDLE ProcessHandle, DWORD DesiredAccess, PHANDLE TokenHandle);
    PFnOpenProcessToken pfnOpenProcessToken = (PFnOpenProcessToken)::GetProcAddress(hAdvapi32, "OpenProcessToken");

#ifdef _UNICODE
    typedef BOOL (WINAPI* PFnLookupPrivilegeValue)(LPCTSTR lpSystemName, LPCTSTR lpName, PLUID lpLuid);
    PFnLookupPrivilegeValue pfnLookupPrivilegeValue = (PFnLookupPrivilegeValue)::GetProcAddress(hAdvapi32, "LookupPrivilegeValueW");
#else
    typedef BOOL (WINAPI* PFnLookupPrivilegeValue)(LPCSTR lpSystemName, LPCSTR lpName, PLUID lpLuid);
    PFnLookupPrivilegeValue pfnLookupPrivilegeValue = (PFnLookupPrivilegeValue)::GetProcAddress(hAdvapi32, "LookupPrivilegeValueA");
#endif

    typedef BOOL (WINAPI* PFnAdjustTokenPrivileges)(HANDLE TokenHandle, BOOL DisableAllPrivileges, PTOKEN_PRIVILEGES NewState, DWORD BufferLength, PTOKEN_PRIVILEGES PreviousState, PDWORD ReturnLength);
    PFnAdjustTokenPrivileges pfnAdjustTokenPrivileges = (PFnAdjustTokenPrivileges)::GetProcAddress(hAdvapi32, "AdjustTokenPrivileges");

    if (!pfnOpenProcessToken || !pfnLookupPrivilegeValue || !pfnAdjustTokenPrivileges) {
        ::FreeLibrary(hAdvapi32);
        return FALSE;
    }

    // Grab this process's token.
    if (!pfnOpenProcessToken(::GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) {
        ::FreeLibrary(hAdvapi32);
        return FALSE;
    }

    // Get the LUID for the shutdown privilege.
    pfnLookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);

    tkp.PrivilegeCount = 1; // One privilege to set
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    // Get the shutdown privilege for this process.
    pfnAdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

    // Cannot test return value of AdjustTokenPrivileges.
    if (::GetLastError() != NO_ERROR) {
        ::FreeLibrary(hAdvapi32);
        return FALSE;
    }

    ::FreeLibrary(hAdvapi32);
    return TRUE;
}


////////////////////////////////////////////////////////////////////////////
// PostError
////////////////////////////////////////////////////////////////////////////

void PostError(_In_ HWND hWndParent, _In_ UINT uiErrorID, ...)
{
    ATL::CAtlString sFormat;

    if (sFormat.LoadString(uiErrorID)) {
        va_list pArgs;
        ATL::CAtlString sMessage;

        va_start(pArgs, uiErrorID);
#if _ATL_VER >= 0x0700
        sMessage.FormatMessageV(sFormat, &pArgs);
#else
        LPTSTR lpszTemp;

        if (::FormatMessage(FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ALLOCATE_BUFFER, (LPCTSTR)sFormat, 0, 0, (LPTSTR)&lpszTemp, 0, &pArgs) == 0 || lpszTemp == NULL)
            AtlThrow(E_OUTOFMEMORY);

        // assign lpszTemp into the resulting string and free the temporary
        sMessage = lpszTemp;
        ::LocalFree(lpszTemp);
#endif
        va_end(pArgs);

        ::ShowMessageBox(hWndParent, sMessage, MB_OK | MB_ICONERROR);
    } else {
        sFormat.FormatMessage(IDS_ERR_NO_RESOURCE, uiErrorID);
        ::ShowMessageBox(hWndParent, sFormat, MB_OK | MB_ICONERROR);
    }
}


////////////////////////////////////////////////////////////////////////////
// PostMSIError
////////////////////////////////////////////////////////////////////////////

void PostMSIError(_In_ HWND hWndParent, _In_ UINT uiErrorID)
{
    switch (uiErrorID) {
    case ERROR_INSTALL_SUSPEND:
    case ERROR_INSTALL_USEREXIT:
    case ERROR_INSTALL_FAILURE:
    case ERROR_SUCCESS_REBOOT_REQUIRED:
    case ERROR_SUCCESS_REBOOT_INITIATED:
    case ERROR_APPHELP_BLOCK:
        return;

    case ERROR_FILE_NOT_FOUND:
    case ERROR_INVALID_NAME:
    case ERROR_PATH_NOT_FOUND:
        uiErrorID = ERROR_INSTALL_PACKAGE_OPEN_FAILED;
    }

    HMODULE hMsiMsg = ::LoadLibraryEx(_T("msimsg.dll"), NULL, LOAD_LIBRARY_AS_DATAFILE);
    if (hMsiMsg) {
        ATL::CAtlString sError;
        if (sError.LoadString(hMsiMsg, uiErrorID)) {
            ::ShowMessageBox(hWndParent, sError, MB_OK | MB_ICONERROR);
            ::FreeLibrary(hMsiMsg);
            return;
        }
        ::FreeLibrary(hMsiMsg);
    }

    // Error string does not exist, use default.
    ::PostError(hWndParent, IDS_ERR_INSTALL, uiErrorID);
}


////////////////////////////////////////////////////////////////////////////
// ShowMessageBox
////////////////////////////////////////////////////////////////////////////

int ShowMessageBox(_In_ HWND hWndParent, _In_z_ LPCTSTR lpszText, _In_opt_ UINT nType, _In_opt_ UINT nIDHelp)
{
    ATL::CAtlString sCaption;
    sCaption.LoadString(IDS_APP_TITLE);
    return ::MessageBox(hWndParent, lpszText, sCaption, nType);
}


int ShowMessageBox(_In_ HWND hWndParent, _In_ UINT nIDPrompt, _In_opt_ UINT nType, _In_opt_ UINT nIDHelp)
{
    ATL::CAtlString sText;
    sText.LoadString(nIDPrompt);
    return ::ShowMessageBox(hWndParent, sText, nType, nIDHelp);
}
