/*
    Copyright 1991-2015 Amebis

    This file is part of Setup.

    Setup is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Setup is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Setup. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


////////////////////////////////////////////////////////////////////////////
// IsFileTrusted
////////////////////////////////////////////////////////////////////////////

FileTrustEnum IsFileTrusted(_In_z_ LPCWSTR pszwFile, _In_ HWND hwndParent, _In_ DWORD dwUIChoice, _Out_opt_ BOOL *pfIsSigned, _Out_opt_ PCCERT_CONTEXT *ppcSigner)
{
    FileTrustEnum itv = FileTrustUntrusted;

    if (pfIsSigned)
        *pfIsSigned = FALSE;
    if (ppcSigner)
        *ppcSigner  = 0;

    HMODULE hWinTrust = ::LoadLibrary(_T("wintrust.dll"));
    if (hWinTrust) {
        typedef HRESULT (WINAPI *PFnWinVerifyTrust)(HWND hwnd, GUID *pgActionID, WINTRUST_DATA *pWinTrustData);
        PFnWinVerifyTrust pfnWinVerifyTrust = (PFnWinVerifyTrust)::GetProcAddress(hWinTrust, "WinVerifyTrust");

        typedef PCRYPT_PROVIDER_DATA (WINAPI *PFnWTHelperProvDataFromStateData)(HANDLE hStateData);
        PFnWTHelperProvDataFromStateData pfnWTHelperProvDataFromStateData= (PFnWTHelperProvDataFromStateData)::GetProcAddress(hWinTrust, "WTHelperProvDataFromStateData");

        typedef PCRYPT_PROVIDER_SGNR (WINAPI *PFnWTHelperGetProvSignerFromChain)(PCRYPT_PROVIDER_DATA pProvData, DWORD idxSigner, BOOL fCounterSigner, DWORD idxCounterSigner);
        PFnWTHelperGetProvSignerFromChain pfnWTHelperGetProvSignerFromChain = (PFnWTHelperGetProvSignerFromChain)::GetProcAddress(hWinTrust, "WTHelperGetProvSignerFromChain");

        typedef PCRYPT_PROVIDER_CERT (WINAPI* PFnWTHelperGetProvCertFromChain)(PCRYPT_PROVIDER_SGNR pSgnr, DWORD idxCert);
        PFnWTHelperGetProvCertFromChain pfnWTHelperGetProvCertFromChain = (PFnWTHelperGetProvCertFromChain)::GetProcAddress(hWinTrust, "WTHelperGetProvCertFromChain");

        if (pfnWinVerifyTrust && pfnWTHelperProvDataFromStateData && pfnWTHelperGetProvSignerFromChain && pfnWTHelperGetProvCertFromChain) {
            HRESULT hr;
            GUID guidAction = WINTRUST_ACTION_GENERIC_VERIFY_V2;
            WINTRUST_FILE_INFO sWintrustFileInfo = {
                sizeof(sWintrustFileInfo),
                pszwFile
            };
            WINTRUST_DATA sWintrustData = {
                sizeof(sWintrustData),
                NULL,
                NULL,
                dwUIChoice,
                WTD_REVOKE_NONE,
                WTD_CHOICE_FILE,
                { &sWintrustFileInfo },
                ppcSigner ? WTD_STATEACTION_VERIFY : 0
            };

            hr = pfnWinVerifyTrust(dwUIChoice == WTD_UI_NONE ? (HWND)INVALID_HANDLE_VALUE : hwndParent, &guidAction, &sWintrustData);
            ATLTRACE(atlTraceGeneral, 2, _T("WVT returned %#010x\r\n"), hr);
            itv = hr == TRUST_E_PROVIDER_UNKNOWN ? FileTrustWintrustNotInstalled : SUCCEEDED(hr) ? FileTrustTrusted : FileTrustUntrusted;

            if (itv != FileTrustWintrustNotInstalled) {
                if (pfIsSigned)
                    *pfIsSigned = hr == TRUST_E_NOSIGNATURE ? FALSE : TRUE;

                if (hr != TRUST_E_NOSIGNATURE) {
                    if (ppcSigner) {
                        CRYPT_PROVIDER_CERT *psProvCert = NULL;

                        // Grab the provider data.
                        const CRYPT_PROVIDER_DATA *psProvData = pfnWTHelperProvDataFromStateData(sWintrustData.hWVTStateData);
                        if (psProvData) {
                            // Grab the signer data from the CRYPT_PROV_DATA.
                            CRYPT_PROVIDER_SGNR *psProvSigner = pfnWTHelperGetProvSignerFromChain((PCRYPT_PROVIDER_DATA)psProvData, 0 /*first signer*/, FALSE /* not a counter signer */, 0);
                            if (psProvSigner) {
                                // Grab the signer cert from CRYPT_PROV_SGNR (pos 0 = signer cert; pos csCertChain-1 = root cert).
                                psProvCert = pfnWTHelperGetProvCertFromChain(psProvSigner, 0);
                            }
                        }

                        *ppcSigner = 0;
                        if (psProvCert) {
                            // Duplicate the cert.
                            HMODULE hCrypt32 = ::LoadLibrary(_T("crypt32.dll"));
                            if (hCrypt32) {
                                typedef PCCERT_CONTEXT (WINAPI* PFnCertDuplicateCertificateContext)(PCCERT_CONTEXT pCertContext);
                                PFnCertDuplicateCertificateContext pfnCertDuplicateCertificateContext = (PFnCertDuplicateCertificateContext)::GetProcAddress(hCrypt32, "CertDuplicateCertificateContext");
                                if (pfnCertDuplicateCertificateContext)
                                    *ppcSigner = pfnCertDuplicateCertificateContext(psProvCert->pCert);
                                ::FreeLibrary(hCrypt32);
                            } else {
                                // Crypt32 is unavailable on the machine.
                                ATLTRACE(atlTraceGeneral, 0, _T("LoadLibrary failed (%ld).\r\n"), ::GetLastError());
                            }
                        }
                    }
                }
            }

            // Release state data.
            sWintrustData.dwUIChoice = WTD_UI_NONE;
            sWintrustData.dwStateAction = WTD_STATEACTION_CLOSE;
            pfnWinVerifyTrust((HWND)INVALID_HANDLE_VALUE, &guidAction, &sWintrustData);
        } else {
            // WinTrust is unavailable on the machine.
            ATLTRACE(atlTraceGeneral, 0, _T("WinTrust functions not found.\r\n"));
            itv = FileTrustWintrustNotInstalled;
        }

        ::FreeLibrary(hWinTrust);
    } else {
        // WinTrust is unavailable on the machine.
        ATLTRACE(atlTraceGeneral, 0, _T("LoadLibrary failed (%ld).\r\n"), ::GetLastError());
        itv = FileTrustWintrustNotInstalled;
    }

    return itv;
}


////////////////////////////////////////////////////////////////////////////
// IsPackageTrusted
////////////////////////////////////////////////////////////////////////////

FileTrustEnum IsPackageTrusted(_In_z_ LPCTSTR pszSetupExe, _In_z_ LPCTSTR pszPackage, _In_ HWND hwndParent)
{
    FileTrustEnum itv = FileTrustUntrusted;

    HMODULE hCrypt32 = ::LoadLibrary(_T("crypt32.dll"));
    if (hCrypt32) {
        typedef BOOL (WINAPI* PFnCertCompareCertificate)(DWORD dwCertEncodingType, PCERT_INFO pCertId1, PCERT_INFO pCertId2);
        PFnCertCompareCertificate pfnCertCompareCertificate = (PFnCertCompareCertificate)::GetProcAddress(hCrypt32, "CertCompareCertificate");

        typedef BOOL (WINAPI* PFnCertFreeCertificateContext)(PCCERT_CONTEXT pCertContext);
        PFnCertFreeCertificateContext pfnCertFreeCertificateContext = (PFnCertFreeCertificateContext)::GetProcAddress(hCrypt32, "CertFreeCertificateContext");

        if (pfnCertCompareCertificate && pfnCertFreeCertificateContext) {
            LPCWSTR pszwSetup = NULL;
#ifdef _UNICODE
            pszwSetup = pszSetupExe;
#else
            {
                // Convert pszSetupExe to Unicode.
                int cchWide = ::MultiByteToWideChar(CP_ACP, 0, pszSetupExe, -1, 0, 0);
                pszwSetup = new WCHAR[cchWide];
                if (!pszwSetup) AtlThrow(E_OUTOFMEMORY);
                ::MultiByteToWideChar(CP_ACP, 0, pszSetupExe, -1, (LPWSTR)pszwSetup, cchWide);
            }
#endif

            PCCERT_CONTEXT pcExeSigner = NULL;
            PCCERT_CONTEXT pcMsiSigner = NULL;
            BOOL fSetupExeIsSigned = FALSE;
            BOOL fPackageIsSigned  = FALSE;

            //
            // Step 1: Silently call WinVerifyTrust on pszSetupExe, ignore return value - except for ivtWintrustNotOnMachine.
            //
            ATLTRACE(atlTraceGeneral, 2, _T("Step 1: Silently call WinVerifyTrust on \"%ls\", ignoring return value.\r\n"), pszwSetup);
            if ((itv = ::IsFileTrusted(pszwSetup, hwndParent, WTD_UI_NONE, &fSetupExeIsSigned, &pcExeSigner)) != FileTrustWintrustNotInstalled) {
                ATLTRACE(atlTraceGeneral, 2, _T("fSetupExeIsSigned = %s\r\n"), fSetupExeIsSigned ? "TRUE" : "FALSE");

                LPCWSTR pszwPackage = NULL;
#ifdef _UNICODE
                pszwPackage = pszPackage;
#else
                {
                    // Convert pszPackage to Unicode.
                    int cchWide = ::MultiByteToWideChar(CP_ACP, 0, pszPackage, -1, 0, 0);
                    pszwPackage = new WCHAR[cchWide];
                    if (!pszwPackage) AtlThrow(E_OUTOFMEMORY);
                    ::MultiByteToWideChar(CP_ACP, 0, pszPackage, -1, (LPWSTR)pszwPackage, cchWide));
                }
#endif
                //
                // Step 2: Silently call WinVerifyTrust on pszPackage, ignore return value - except for ivtWintrustNotOnMachine.
                //
                if (fSetupExeIsSigned) {
                    ATLTRACE(atlTraceGeneral, 2, _T("Step 2: Silently call WinVerifyTrust on \"%ls\", ignoring return value.\r\n"), pszwPackage);
                    if ((itv = IsFileTrusted(pszwPackage, hwndParent, WTD_UI_NONE, &fPackageIsSigned, &pcMsiSigner)) != FileTrustWintrustNotInstalled)
                        ATLTRACE(atlTraceGeneral, 2, _T("fPackageIsSigned = %s\r\n"), fPackageIsSigned ? "TRUE" : "FALSE");
                }

                if (itv != FileTrustWintrustNotInstalled) {
                    //
                    // Step 3: Call WinVerifyTrust on pszPackage, return value matters; use proper UI-level.
                    //
                    DWORD dwUILevel = 0;
                    if (!fSetupExeIsSigned || // EXE is not signed.
                        !fPackageIsSigned  || // Package is not signed.
                        !pcExeSigner       || // EXE signer cert is missing.
                        !pcMsiSigner       || // Package signer cert is missing.
                        !pfnCertCompareCertificate(X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, pcExeSigner->pCertInfo, pcMsiSigner->pCertInfo)) // Signed by different certs.
                    {
                        // Always show UI.
                        ATLTRACE(atlTraceGeneral, 2, _T("Step 3: Last call to WinVerifyTrust using full UI.\r\n"));
                        dwUILevel = WTD_UI_ALL;
                    } else {
                        // Show UI only if bad.
                        ATLTRACE(atlTraceGeneral, 2, _T("Step 3: Last call to WinVerifyTrust showing UI only if something is wrong.\r\n"));
                        dwUILevel = WTD_UI_NOGOOD;
                    }
                    itv = ::IsFileTrusted(pszwPackage, hwndParent, dwUILevel, NULL, NULL);
                }

#ifndef _UNICODE
                delete [] pszwPackage;
#endif
            }

            if (pcMsiSigner) pfnCertFreeCertificateContext(pcMsiSigner);
            if (pcExeSigner) pfnCertFreeCertificateContext(pcExeSigner);
#ifndef _UNICODE
            delete [] pszwSetup;
#endif
        } else {
            // Crypt32 is unavailable on the machine.
            ATLTRACE(atlTraceGeneral, 0, _T("Crypt32 functions not found.\r\n"));
            itv = FileTrustWintrustNotInstalled;
        }

        ::FreeLibrary(hCrypt32);
    } else {
        // Crypt32 is unavailable on the machine.
        ATLTRACE(atlTraceGeneral, 0, _T("LoadLibrary failed (%ld).\r\n"), ::GetLastError());
        itv = FileTrustWintrustNotInstalled;
    }

    return itv;
}
