/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __AOSNATL_H__
#define __AOSNATL_H__

#include "AOsn.h"
#include <atlbase.h>
#include <atlcom.h>
#include <atldef.h>
#include <basetyps.h>


#ifdef AOSN_ODTISNIK

//////////////////////////////////////////////////////////////////////
// IOdtisnikImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik> class ATL_NO_VTABLE IOdtisnikImpl :
    public CComObjectRootEx<ThreadModel>,
    public CComCoClass<T_cc, pclsid_cc>,
    public IDispatchImpl<T_base, &__uuidof(T_base)>
{
public:
    typedef IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik> T_this;

    BEGIN_COM_MAP(T_this)
        COM_INTERFACE_ENTRY(IOdtisnik)
        COM_INTERFACE_ENTRY(IDispatch)
    END_COM_MAP()

public:
    IOdtisnikImpl();

    STDMETHOD(Resetiraj)();
    STDMETHOD(IzracunajBlok)(const VARIANT *bin);
    STDMETHOD(Zakljuci)(VARIANT *odtis);

    STDMETHOD(Izracunaj)(const VARIANT *bin, VARIANT *odtis);

protected:
    STDMETHOD(IzracunajBlok)(const T_VOID *bin, T_SIZE dolzina);
    STDMETHOD(Izracunaj)(const T_VOID *bin, T_SIZE dolzina, VARIANT *odtis);

protected:
    T_odtisnik odtisnik;
};

#endif


//////////////////////////////////////////////////////////////////////
// ITokPodatkovImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
class ATL_NO_VTABLE ITokPodatkovImpl :
    public CComObjectRootEx<ThreadModel>,
    public CComCoClass<T_cc, pclsid_cc>,
    public IDispatchImpl<T_base, &__uuidof(T_base)>,
    public IStream
{
public:
    typedef ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_this;

    BEGIN_COM_MAP(T_this)
        COM_INTERFACE_ENTRY(ITokPodatkov)
        COM_INTERFACE_ENTRY(ISequentialStream)
        COM_INTERFACE_ENTRY(IStream)
        COM_INTERFACE_ENTRY(IDispatch)
    END_COM_MAP()

public:
    ITokPodatkovImpl();

    // metode ISequentialStream
    STDMETHOD(Read)(void *pv, ULONG cb, ULONG *pcbRead);
    STDMETHOD(Write)(const void *pv, ULONG cb, ULONG *pcbWritten);

    // metode IStream
    STDMETHOD(Seek)(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER *plibNewPosition);
    STDMETHOD(SetSize)(ULARGE_INTEGER libNewSize);
    STDMETHOD(CopyTo)(IStream *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER *pcbRead, ULARGE_INTEGER *pcbWritten);
    STDMETHOD(Commit)(DWORD grfCommitFlags);
    STDMETHOD(Revert)();
    STDMETHOD(LockRegion)(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);
    STDMETHOD(UnlockRegion)(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);
    STDMETHOD(Stat)(__RPC__out STATSTG *pstatstg, DWORD grfStatFlag);
    STDMETHOD(Clone)(__RPC__deref_out_opt IStream **ppstm);

    // metode ITokPodatkov
    STDMETHOD(Preberi)(UINT_PTR dolzina, VARIANT *podatki);
    STDMETHOD(PreberiBajt)(VARIANT *bajt);
    STDMETHOD(Zapisi)(const VARIANT *podatki, UINT_PTR *zapisano);
    STDMETHOD(ZapisiBajt)(BYTE bajt, UINT_PTR koliko = 1);
    STDMETHOD(Zapri)();
    STDMETHOD(Izplakni)();

    STDMETHOD(VrniStanje)(LONG *stanje);
    STDMETHOD(ResetirajStanje)();
    STDMETHOD(JeKonec)(VARIANT_BOOL *je_konec);

protected:
    T_tok tok;

    ULARGE_INTEGER polozaj;     // navidezni poloaj v toku podatkov
    ULARGE_INTEGER dolzina;     // navidezna dolina toka podatkov
};


//////////////////////////////////////////////////////////////////////
// IDatotekaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
class ATL_NO_VTABLE IDatotekaImpl :
    public ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>
{
public:
    typedef IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_this;
    typedef ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_base2;

    BEGIN_COM_MAP(T_this)
        COM_INTERFACE_ENTRY(IDatoteka)
        COM_INTERFACE_ENTRY_CHAIN(T_base2)
    END_COM_MAP()

public:
    IDatotekaImpl();

    // metode IStream
    STDMETHOD(Seek)(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER *plibNewPosition);
    STDMETHOD(SetSize)(ULARGE_INTEGER libNewSize);
    STDMETHOD(LockRegion)(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);
    STDMETHOD(UnlockRegion)(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);
    STDMETHOD(Stat)(__RPC__out STATSTG *pstatstg, DWORD grfStatFlag);

    // metode IDatoteka
    STDMETHOD(Prestavi)(INT_PTR odmik, UINT_PTR kako, UINT_PTR *koncni);
    STDMETHOD(VrniOdmik)(UINT_PTR *koncni);

    STDMETHOD(ZakleniObmocje)(UINT_PTR odmik, UINT_PTR dolzina);
    STDMETHOD(OdkleniObmocje)(UINT_PTR odmik, UINT_PTR dolzina);

    STDMETHOD(VrniDolzino)(UINT_PTR *dolzina);
    STDMETHOD(PostaviKonec)();

    STDMETHOD(DolociKodnoStran)(WORD privzeta_kodna_stran, WORD* kodna_stran);

    STDMETHOD(get_datum_stvaritve)(VARIANT *datum);
    STDMETHOD(put_datum_stvaritve)(const VARIANT *datum);
    STDMETHOD(get_datum_dostopa)(VARIANT *datum);
    STDMETHOD(put_datum_dostopa)(const VARIANT *datum);
    STDMETHOD(get_datum_pisanja)(VARIANT *datum);
    STDMETHOD(put_datum_pisanja)(const VARIANT *datum);
};


//////////////////////////////////////////////////////////////////////
// IPodvojevalnikTokaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
class ATL_NO_VTABLE IPodvojevalnikTokaImpl :
    public ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>
{
public:
    typedef IPodvojevalnikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_this;
    typedef ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_base2;

    BEGIN_COM_MAP(T_this)
        COM_INTERFACE_ENTRY(IPodvojevalnikToka)
        COM_INTERFACE_ENTRY_CHAIN(T_base2)
    END_COM_MAP()

public:
    IPodvojevalnikTokaImpl();
    void FinalRelease();

public:
    // metode IPodvojevalnikToka
    STDMETHOD(Dodaj)(VARIANT cilj);
    STDMETHOD(Odstrani)(VARIANT cilj);

protected:
    RSeznam<RISequentialStream> cilji;
};


//////////////////////////////////////////////////////////////////////
// IPretvornikTokaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
class ATL_NO_VTABLE IPretvornikTokaImpl :
    public ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>
{
public:
    typedef IPretvornikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_this;
    typedef ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok> T_base2;

    BEGIN_COM_MAP(T_this)
        COM_INTERFACE_ENTRY(IPretvornikToka)
        COM_INTERFACE_ENTRY_CHAIN(T_base2)
    END_COM_MAP()

public:
    IPretvornikTokaImpl();

    STDMETHOD(Uporabi)(VARIANT izvor);
    STDMETHOD(Odpusti)();

protected:
    RISequentialStream izvor;
};


//////////////////////////////////////////////////////////////////////
// CAOsnAtlDllModuleT
//////////////////////////////////////////////////////////////////////

template <class T>
class CAOsnAtlDllModuleT :
    public CAtlDllModuleT<T>
{
public:
    CAOsnAtlDllModuleT();
    virtual ~CAOsnAtlDllModuleT();

    HRESULT DllCanUnloadNow();
    HRESULT DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv);
    HRESULT DllRegisterServer(BOOL bRegTypeLib = TRUE);
    HRESULT DllUnregisterServer(BOOL bUnRegTypeLib = TRUE);

    virtual T_STANJE Vzpostavi();
    virtual T_STANJE Zakljuci();

protected:
    RNevarenOdsek kljucavnica;  // kljuavnica za sinhronizacijo dostopa do lanov razreda
    T_BOOL je_vzpostavljen;     // logina spremenljivka, ki doloa, ali je modul vzpostavljen
};


//////////////////////////////////////////////////////////////////////
// Metode
//////////////////////////////////////////////////////////////////////

#ifdef AOSN_ODTISNIK

//////////////////////////////////////////////////////////////////////
// IOdtisnikImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::IOdtisnikImpl() : odtisnik()
{
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::Resetiraj()
{
    AOSN_TRY {
        odtisnik.Resetiraj();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::IzracunajBlok(const VARIANT *bin)
{
    AOSN_TRY {
        if (!bin) return E_INVALIDARG;

        if (V_VT(bin) == VT_BSTR) {
            return IzracunajBlok(V_BSTR(bin), ::SysStringByteLen(V_BSTR(bin)));
        } else if ((V_VT(bin) & VT_ARRAY) && V_ARRAY(bin)) {
            T_VOID *p;
            T_SIZE n;
            HRESULT hr;

            hr = AOsn::ZakleniVsebinoSAFEARRAY(V_ARRAY(bin), &p, &n);
            if (SUCCEEDED(hr)) {
                hr = IzracunajBlok(p, n);
                verify(SUCCEEDED(AOsn::OdkleniVsebinoSAFEARRAY(V_ARRAY(bin))));
            }
            return hr;
        } else
             return E_INVALIDARG;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::IzracunajBlok(const T_VOID *bin, T_SIZE dolzina)
{
    AOSN_TRY {
        if (!bin && dolzina) return E_INVALIDARG;
        odtisnik.Izracunaj(bin, dolzina);
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::Zakljuci(VARIANT *odtis)
{
    AOSN_TRY {
        if (!odtis) return E_INVALIDARG;

        odtisnik.Zakljuci();
        ::VariantClear(odtis);
        V_ARRAY(odtis) = AOsn::VrniSAFEARRAY(VT_UI1, 1, odtisnik.VrniDolzinoOdtisa(), odtisnik.VrniOdtis());
        if (V_ARRAY(odtis)) {
            V_VT(odtis) = VT_ARRAY | VT_UI1;
            return S_OK;
        } else {
            V_VT(odtis) = VT_NULL;
            return E_OUTOFMEMORY;
        }
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::Izracunaj(const VARIANT *bin, VARIANT *odtis)
{
    AOSN_TRY {
        if (!bin || !odtis) return E_INVALIDARG;

        if (V_VT(bin) == VT_BSTR) {
            return Izracunaj(V_BSTR(bin), ::SysStringByteLen(V_BSTR(bin)), odtis);
        } else if ((V_VT(bin) & VT_ARRAY) && V_ARRAY(bin)) {
            T_VOID *p;
            T_SIZE n;
            HRESULT hr;

            hr = AOsn::ZakleniVsebinoSAFEARRAY(V_ARRAY(bin), &p, &n);
            if (SUCCEEDED(hr)) {
                hr = Izracunaj(p, n, odtis);
                verify(SUCCEEDED(AOsn::OdkleniVsebinoSAFEARRAY(V_ARRAY(bin))));
            }
            return hr;
        } else
             return E_INVALIDARG;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_odtisnik>
STDMETHODIMP IOdtisnikImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_odtisnik>::Izracunaj(const T_VOID *bin, T_SIZE dolzina, VARIANT *odtis)
{
    AOSN_TRY {
        if (!bin && dolzina || !odtis) return E_INVALIDARG;

        odtisnik.Resetiraj();
        odtisnik.Izracunaj(bin, dolzina);
        odtisnik.Zakljuci();
        ::VariantClear(odtis);
        V_ARRAY(odtis) = AOsn::VrniSAFEARRAY(VT_UI1, 1, odtisnik.VrniDolzinoOdtisa(), odtisnik.VrniOdtis());
        if (V_ARRAY(odtis)) {
            V_VT(odtis) = VT_ARRAY | VT_UI1;
            return S_OK;
        } else {
            V_VT(odtis) = VT_NULL;
            return E_OUTOFMEMORY;
        }
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}

#endif

//////////////////////////////////////////////////////////////////////
// ITokPodatkovImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::ITokPodatkovImpl() : tok()
{
    polozaj.QuadPart = 0;
    dolzina.QuadPart = 0;
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Read(void *pv, ULONG cb, ULONG *pcbRead)
{
    AOSN_TRY {
        ULONG n;
        T_STANJE stanje;

        if (!pv && cb) return E_INVALIDARG;

        n = (ULONG)tok.Preberi(pv, cb);
        stanje = tok.VrniStanje();
        if (stanje == AOSN_NAPAKA_KONEC_TOKA)
            return S_FALSE;
        else if (NEUSPESNO(stanje))
            return E_FAIL;

        polozaj.QuadPart += n;
        if (polozaj.QuadPart > dolzina.QuadPart) dolzina.QuadPart = n;
        if (pcbRead) *pcbRead = n;
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Write(const void *pv, ULONG cb, ULONG *pcbWritten)
{
    AOSN_TRY {
        ULONG n;

        if (!pv && cb) return E_INVALIDARG;

        if (NEUSPESNO(tok.VrniStanje())) return E_FAIL;
        n = (ULONG)tok.Zapisi(pv, cb);

        polozaj.QuadPart += n;
        if (polozaj.QuadPart > dolzina.QuadPart) dolzina.QuadPart = n;
        if (pcbWritten) *pcbWritten = n;
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Seek(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER *plibNewPosition)
{
    AOSN_TRY {
        // Skakanja po toku podatkov ne podpiramo. Kljub temu preraunamo navidezni poloaj, ker doloeni odjemalci kliejo Seek(0).
        switch (dwOrigin) {
            case STREAM_SEEK_SET: polozaj.QuadPart =                    dlibMove.QuadPart; break;
            case STREAM_SEEK_CUR: polozaj.QuadPart = polozaj.QuadPart + dlibMove.QuadPart; break;
            case STREAM_SEEK_END: polozaj.QuadPart = dolzina.QuadPart + dlibMove.QuadPart; break;
            default: return E_INVALIDARG;
        }

        if (plibNewPosition) plibNewPosition->QuadPart = polozaj.QuadPart;
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::SetSize(ULARGE_INTEGER libNewSize)
{
    AOSN_TRY {
        UNREFERENCED_PARAMETER(libNewSize);

        // Nastavljanja doline toka ne podpiramo.

        return E_NOTIMPL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::CopyTo(IStream *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER *pcbRead, ULARGE_INTEGER *pcbWritten)
{
    AOSN_TRY {
        assert(0); // TODO: Preizkusi ta del kode.

        HRESULT hr;
        T_SIZE korak;
        T_DOLZINA ostanek;
        ULARGE_INTEGER prebrano_skupno, zapisano_skupno;
        void *pom;

        if (!pstm) return E_INVALIDARG;

        korak = (T_SIZE)amin<T_DOLZINA>(cb.QuadPart, 0x10000);
        pom = aalloc(korak);
        if (!pom) return E_OUTOFMEMORY;

        prebrano_skupno.QuadPart = 0;
        zapisano_skupno.QuadPart = 0;

        for (ostanek = cb.QuadPart, hr = S_OK; ostanek; ) {
            ULONG trenutno = (ULONG)amin<T_DOLZINA>(ostanek, korak), prebrano, zapisano;

            hr = Read(pom, trenutno, &prebrano);
            if (FAILED(hr)) break;
            prebrano_skupno.QuadPart += prebrano;

            hr = pstm->Write(pom, prebrano, &zapisano);
            if (FAILED(hr)) break;
            zapisano_skupno.QuadPart += zapisano;

            ostanek -= zapisano;
            if (trenutno != prebrano || prebrano != zapisano) break;
        }

        afree(pom);

        if (pcbRead)       *pcbRead = prebrano_skupno;
        if (pcbWritten) *pcbWritten = zapisano_skupno;

        return hr;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Commit(DWORD grfCommitFlags)
{
    AOSN_TRY {
        UNREFERENCED_PARAMETER(grfCommitFlags);

        // Transakcij ne podpiramo. Vseeno izplaknemo tok podatkov.
        return USPESNO(tok.Izplakni()) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Revert()
{
    AOSN_TRY {
        // Transakcij ne podpiramo.

        return E_NOTIMPL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::LockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
    AOSN_TRY {
        UNREFERENCED_PARAMETER(libOffset);
        UNREFERENCED_PARAMETER(cb);
        UNREFERENCED_PARAMETER(dwLockType);

        // Zaklepanja toka ne podpiramo.

        return E_NOTIMPL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::UnlockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
    AOSN_TRY {
        UNREFERENCED_PARAMETER(libOffset);
        UNREFERENCED_PARAMETER(cb);
        UNREFERENCED_PARAMETER(dwLockType);

        // Odklepanja toka ne podpiramo.

        return E_NOTIMPL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Stat(__RPC__out STATSTG *pstatstg, DWORD grfStatFlag)
{
    AOSN_TRY {
        if (!pstatstg) return E_INVALIDARG;

        ZeroMemory(pstatstg, sizeof(STATSTG));

        pstatstg->type  = STGTY_STREAM;
        pstatstg->clsid = CLSID_NULL;

        if ((grfStatFlag & STATFLAG_NONAME) == 0) {
            static OLECHAR ime[] = L"ITokPodatkovImpl";
            pstatstg->pwcsName = (LPOLESTR)CoTaskMemAlloc(_countof(ime));
            if (!pstatstg->pwcsName) return STG_E_INSUFFICIENTMEMORY;
            memcpy(pstatstg->pwcsName, ime, _countof(ime));
        }

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Clone(__RPC__deref_out_opt IStream **ppstm)
{
    AOSN_TRY {
        UNREFERENCED_PARAMETER(ppstm);

        // Kloniranja toka ne podpiramo.

        return E_NOTIMPL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Preberi(UINT_PTR dolzina, VARIANT *podatki)
{
    AOSN_TRY {
        T_SIZE prebrano;
        T_VOID *p;

        if (!podatki) return E_INVALIDARG;

        ::VariantClear(podatki);
        V_ARRAY(podatki) = ::SafeArrayCreateVector(VT_UI1, 0, (ULONG)dolzina);
        if (V_ARRAY(podatki)) {
            V_VT(podatki) = VT_ARRAY | VT_UI1;
        } else {
            V_VT(podatki) = VT_NULL;
            return E_OUTOFMEMORY;
        }

        verify(SUCCEEDED(::SafeArrayAccessData(V_ARRAY(podatki), &p)));
        prebrano = tok.Preberi(p, dolzina);
        verify(SUCCEEDED(::SafeArrayUnaccessData(V_ARRAY(podatki))));
        if (prebrano < dolzina) {
            // Sprostimo odveen pomnilnik.
            SAFEARRAYBOUND bounds = { (ULONG)prebrano, 1 };
            verify(SUCCEEDED(::SafeArrayRedim(V_ARRAY(podatki), &bounds)));
        }

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::PreberiBajt(VARIANT *bajt)
{
    AOSN_TRY {
        if (!bajt) return E_INVALIDARG;

        ::VariantClear(bajt);
        V_VT(bajt)  = VT_UI1;
        V_UI1(bajt) = tok.PreberiBajt();

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Zapisi(const VARIANT *podatki, UINT_PTR *zapisano)
{
    AOSN_TRY {
        T_SIZE n;

        if (!podatki) return E_INVALIDARG;

        if (V_VT(podatki) == VT_BSTR) {
            n = tok.Zapisi(V_BSTR(podatki), ::SysStringByteLen(V_BSTR(podatki)));
            if (zapisano) *zapisano = n;
            return S_OK;
        } else if ((V_VT(podatki) & VT_ARRAY) && V_ARRAY(podatki)) {
            T_VOID *p;
            T_SIZE dolzina;
            HRESULT hr;

            hr = AOsn::ZakleniVsebinoSAFEARRAY(V_ARRAY(podatki), &p, &dolzina);
            if (SUCCEEDED(hr)) {
                n = tok.Zapisi(p, dolzina);
                if (zapisano) *zapisano = n;
                verify(SUCCEEDED(AOsn::OdkleniVsebinoSAFEARRAY(V_ARRAY(podatki))));
                return S_OK;
            } else
                return hr;
        } else
            return E_INVALIDARG;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::ZapisiBajt(BYTE bajt, UINT_PTR koliko)
{
    AOSN_TRY {
        tok.ZapisiBajt(bajt, koliko);
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Zapri()
{
    AOSN_TRY {
        return USPESNO(tok.Zapri()) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Izplakni()
{
    AOSN_TRY {
        return USPESNO(tok.Izplakni()) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::VrniStanje(LONG *stanje)
{
    AOSN_TRY {
        if (!stanje) return E_INVALIDARG;
        *stanje = tok.VrniStanje();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::ResetirajStanje()
{
    AOSN_TRY {
        tok.ResetirajStanje();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::JeKonec(VARIANT_BOOL *je_konec)
{
    AOSN_TRY {
        if (!je_konec) return E_INVALIDARG;
        *je_konec = tok.JeKonec() ? VARIANT_TRUE : VARIANT_FALSE;
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


//////////////////////////////////////////////////////////////////////
// IDatotekaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::IDatotekaImpl() : ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>()
{
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Seek(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER *plibNewPosition)
{
    AOSN_TRY {
        ULARGE_INTEGER odmik;

        switch (dwOrigin) {
            case STREAM_SEEK_SET: odmik.QuadPart = tok.Prestavi(dlibMove.QuadPart, RDAT_OD_ZACETKA);    break;
            case STREAM_SEEK_CUR: odmik.QuadPart = tok.Prestavi(dlibMove.QuadPart, RDAT_OD_TRENUTNEGA); break;
            case STREAM_SEEK_END: odmik.QuadPart = tok.Prestavi(dlibMove.QuadPart, RDAT_OD_KONCA);      break;
            default: return E_INVALIDARG;
        }
        if (plibNewPosition) *plibNewPosition = odmik;

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::SetSize(ULARGE_INTEGER libNewSize)
{
    AOSN_TRY {
        T_ODMIK odmik = tok.VrniOdmik();

        // RDatoteka ne pozna ustrezne metode, jo pa lahko ustvarimo s kombinacijo drugih.
        // Skoimo na mesto elenega konca datoteke, postavimo konec in skoimo nazaj.
        if (tok.Prestavi(libNewSize.QuadPart, RDAT_OD_ZACETKA) != libNewSize.QuadPart) return E_FAIL;
        if (NEUSPESNO(tok.PostaviKonec())) return E_FAIL;
        if (tok.Prestavi(odmik, RDAT_OD_ZACETKA) != libNewSize.QuadPart) return E_FAIL;

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::LockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
    AOSN_TRY {
        if (dwLockType != LOCK_EXCLUSIVE) return STG_E_INVALIDFUNCTION;

        return USPESNO(tok.ZakleniObmocje(libOffset.QuadPart, cb.QuadPart)) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::UnlockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
    AOSN_TRY {
        if (dwLockType != LOCK_EXCLUSIVE) return STG_E_INVALIDFUNCTION;

        return USPESNO(tok.OdkleniObmocje(libOffset.QuadPart, cb.QuadPart)) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Stat(__RPC__out STATSTG *pstatstg, DWORD grfStatFlag)
{
    AOSN_TRY {
        if (!pstatstg) return E_INVALIDARG;

        ZeroMemory(pstatstg, sizeof(STATSTG));

        pstatstg->type  = STGTY_STREAM;
        pstatstg->clsid = CLSID_NULL;

        if ((grfStatFlag & STATFLAG_NONAME) == 0) {
            static OLECHAR ime[] = L"IDatotekaImpl";
            pstatstg->pwcsName = (LPOLESTR)CoTaskMemAlloc(_countof(ime));
            if (!pstatstg->pwcsName) return STG_E_INSUFFICIENTMEMORY;
            memcpy(pstatstg->pwcsName, ime, _countof(ime));
        }

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Prestavi(INT_PTR odmik, UINT_PTR kako, UINT_PTR *koncni)
{
    AOSN_TRY {
        if (koncni)
            *koncni = (UINT_PTR)tok.Prestavi(odmik, (T_DATNACINODMIKA)kako);
        else
            tok.Prestavi(odmik, (T_DATNACINODMIKA)kako);
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::VrniOdmik(UINT_PTR *koncni)
{
    AOSN_TRY {
        if (!koncni) return E_INVALIDARG;
        *koncni = (UINT_PTR)tok.VrniOdmik();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::ZakleniObmocje(UINT_PTR odmik, UINT_PTR dolzina)
{
    AOSN_TRY {
        return USPESNO(tok.ZakleniObmocje(odmik, dolzina)) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::OdkleniObmocje(UINT_PTR odmik, UINT_PTR dolzina)
{
    AOSN_TRY {
        return USPESNO(tok.OdkleniObmocje(odmik, dolzina)) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::VrniDolzino(UINT_PTR *dolzina)
{
    AOSN_TRY {
        if (!dolzina) return E_INVALIDARG;
        *dolzina = (UINT_PTR)tok.VrniDolzino();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::PostaviKonec()
{
    AOSN_TRY {
        return USPESNO(tok.PostaviKonec()) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::DolociKodnoStran(WORD privzeta_kodna_stran, WORD* kodna_stran)
{
    AOSN_TRY {
        if (!kodna_stran) return E_INVALIDARG;
        *kodna_stran = (WORD)tok.DolociKodnoStran((T_KODNASTRAN)privzeta_kodna_stran);
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::get_datum_stvaritve(VARIANT *datum)
{
    AOSN_TRY {
        T_TRENUTEK t;
        if (!datum) return E_INVALIDARG;

        ::VariantClear(datum);
        t = tok.VrniDatumStvaritve();
        if (t != T_TRENUTEK_MIN) {
            V_VT(datum) = VT_DATE;
            RDatum::Pretvori(t, &V_DATE(datum));
        } else
            V_VT(datum) = VT_EMPTY;

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::put_datum_stvaritve(const VARIANT *datum)
{
    AOSN_TRY {
        if (!datum || V_VT(datum) != VT_DATE) return E_INVALIDARG;
        return USPESNO(tok.NastaviDatumStvaritve(RDatum::Pretvori(V_DATE(datum)))) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::get_datum_dostopa(VARIANT *datum)
{
    AOSN_TRY {
        T_TRENUTEK t;
        if (!datum) return E_INVALIDARG;

        ::VariantClear(datum);
        t = tok.VrniDatumDostopa();
        if (t != T_TRENUTEK_MIN) {
            V_VT(datum) = VT_DATE;
            RDatum::Pretvori(t, &V_DATE(datum));
        } else
            V_VT(datum) = VT_EMPTY;

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::put_datum_dostopa(const VARIANT *datum)
{
    AOSN_TRY {
        if (!datum || V_VT(datum) != VT_DATE) return E_INVALIDARG;
        return USPESNO(tok.NastaviDatumDostopa(RDatum::Pretvori(V_DATE(datum)))) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::get_datum_pisanja(VARIANT *datum)
{
    AOSN_TRY {
        T_TRENUTEK t;
        if (!datum) return E_INVALIDARG;

        ::VariantClear(datum);
        t = tok.VrniDatumPisanja();
        if (t != T_TRENUTEK_MIN) {
            V_VT(datum) = VT_DATE;
            RDatum::Pretvori(t, &V_DATE(datum));
        } else
            V_VT(datum) = VT_EMPTY;

        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IDatotekaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::put_datum_pisanja(const VARIANT *datum)
{
    AOSN_TRY {
        if (!datum || V_VT(datum) != VT_DATE) return E_INVALIDARG;
        return USPESNO(tok.NastaviDatumPisanja(RDatum::Pretvori(V_DATE(datum)))) ? S_OK : E_FAIL;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


//////////////////////////////////////////////////////////////////////
// IPodvojevalnikTokaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
IPodvojevalnikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::IPodvojevalnikTokaImpl() :
    ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>()
{
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
void IPodvojevalnikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::FinalRelease()
{
    AOSN_TRY {
        T_ROCICA rocica;

        while ((rocica = cilji.VrniPrvega()) != NULL) {
            RISequentialStream &c = cilji[rocica];
            verify(USPESNO(tok.Odstrani(&c)));
            c.Odpusti()->Release();
            cilji.Brisi(rocica);
        }

        T_base2::FinalRelease();
    } AOSN_CATCH
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IPodvojevalnikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Dodaj(VARIANT cilj)
{
    AOSN_TRY {
        HRESULT hr;
        ISequentialStream *_cilj = NULL;
        T_ROCICA rocica;

        if (V_VT(&cilj) == VT_DISPATCH)
            hr = V_DISPATCH(&cilj)->QueryInterface(IID_ISequentialStream, (void**)&_cilj);
        else if (V_VT(&cilj) == VT_UNKNOWN)
            hr = V_UNKNOWN(&cilj)->QueryInterface(IID_ISequentialStream, (void**)&_cilj);
        else
            hr = E_INVALIDARG;

        if (FAILED(hr) || !_cilj)
            return E_INVALIDARG;

        // Ustvarimo zapis v paralelnem seznamu ciljev.
        rocica = cilji.UstvariVozel();
        if (rocica) {
            RISequentialStream &c = cilji[rocica];
            // Pripnemo ciljni tok podatkov zapisu.
            if (USPESNO(c.Uporabi(_cilj))) {
                // Dodamo zapis na konec seznama ciljev.
                cilji.DodajVozel(rocica);

                // Dodamo zapis toku podatkov (ki mora biti vrste RPodvojevalnikToka).
                if (USPESNO(tok.Dodaj(&c))) {
                    _cilj->Release();
                    return S_OK;
                } else
                    hr = E_FAIL;

                cilji.IzlociVozel(rocica);
                c.Odpusti();
            } else
                hr = E_FAIL;

            cilji.UniciVozel(rocica);
        } else
            hr = E_OUTOFMEMORY;

        _cilj->Release();
        return hr;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IPodvojevalnikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Odstrani(VARIANT cilj)
{
    AOSN_TRY {
        HRESULT hr;
        ISequentialStream *_cilj = NULL;
        T_ROCICA rocica;

        if (V_VT(&cilj) == VT_DISPATCH)
            hr = V_DISPATCH(&cilj)->QueryInterface(IID_ISequentialStream, (void**)&_cilj);
        else if (V_VT(&cilj) == VT_UNKNOWN)
            hr = V_UNKNOWN(&cilj)->QueryInterface(IID_ISequentialStream, (void**)&_cilj);
        else
            hr = E_INVALIDARG;

        if (FAILED(hr) || !_cilj)
            return E_INVALIDARG;

        // Poiemo cilj v svojem seznamu.
        for (rocica = cilji.VrniPrvega(); rocica; rocica = cilji.VrniNaslednjega(rocica)) {
            RISequentialStream &c = cilji[rocica];
            if (c.VrniIzvor() == _cilj) {
                _cilj->Release();
                if (USPESNO(tok.Odstrani(&c))) {
                    c.Odpusti();
                    cilji.Brisi(rocica);
                    return S_OK;
                } else
                    return E_FAIL;
            }
        }
        _cilj->Release();

        return E_INVALIDARG;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


//////////////////////////////////////////////////////////////////////
// IPretvornikTokaImpl
//////////////////////////////////////////////////////////////////////

template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
IPretvornikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::IPretvornikTokaImpl() :
    ITokPodatkovImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>(),
    izvor()
{
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IPretvornikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Uporabi(VARIANT _izvor)
{
    AOSN_TRY {
        HRESULT hr;
        ISequentialStream *i = NULL;

        if (V_VT(&_izvor) == VT_DISPATCH)
            hr = V_DISPATCH(&_izvor)->QueryInterface(IID_ISequentialStream, (void**)&i);
        else if (V_VT(&_izvor) == VT_UNKNOWN)
            hr = V_UNKNOWN(&_izvor)->QueryInterface(IID_ISequentialStream, (void**)&i);
        else
            hr = E_INVALIDARG;

        if (FAILED(hr) || !i)
            return E_INVALIDARG;

        polozaj.QuadPart = 0;
        dolzina.QuadPart = 0;

        if (USPESNO(izvor.Uporabi(i))) {
            if (USPESNO(tok.Uporabi(&izvor))) {
                i->Release();
                return S_OK;
            } else
                hr = E_FAIL;

            izvor.Odpusti();
        } else
            hr = E_INVALIDARG;

        i->Release();
        return hr;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


template <typename ThreadModel, class T_base, class T_cc, const CLSID *pclsid_cc, class T_tok>
STDMETHODIMP IPretvornikTokaImpl<ThreadModel, T_base, T_cc, pclsid_cc, T_tok>::Odpusti()
{
    AOSN_TRY {
        tok.Odpusti();
        izvor.Odpusti();
        return S_OK;
    } AOSN_CATCH_RETURN(E_UNEXPECTED)
}


//////////////////////////////////////////////////////////////////////
// CAOsnAtlDllModuleT
//////////////////////////////////////////////////////////////////////

template <class T>
CAOsnAtlDllModuleT<T>::CAOsnAtlDllModuleT() :
    CAtlDllModuleT<T>(),
    je_vzpostavljen(FALSE)
{
}


template <class T>
CAOsnAtlDllModuleT<T>::~CAOsnAtlDllModuleT()
{
    assert(!je_vzpostavljen);
}


template <class T>
HRESULT CAOsnAtlDllModuleT<T>::DllCanUnloadNow()
{
    HRESULT hr = CAtlDllModuleT<T>::DllCanUnloadNow();

    if (SUCCEEDED(hr)) {
        // Zakljuimo modul.
        verify(USPESNO(Zakljuci()));
    }

    return hr;
}


template <class T>
HRESULT CAOsnAtlDllModuleT<T>::DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
    // Vzpostavimo modul.
    if (NEUSPESNO(Vzpostavi())) return E_FAIL;

    return CAtlDllModuleT<T>::DllGetClassObject(rclsid, riid, ppv);
}


template <class T>
HRESULT CAOsnAtlDllModuleT<T>::DllRegisterServer(BOOL bRegTypeLib)
{
    HRESULT hr;

    // Vzpostavimo modul.
    if (NEUSPESNO(Vzpostavi())) return E_FAIL;

    hr = CAtlDllModuleT<T>::DllRegisterServer(bRegTypeLib);

    // Zakljuimo modul.
    verify(USPESNO(Zakljuci()));

    return hr;
}


template <class T>
HRESULT CAOsnAtlDllModuleT<T>::DllUnregisterServer(BOOL bUnRegTypeLib)
{
    HRESULT hr;

    // Vzpostavimo modul.
    if (NEUSPESNO(Vzpostavi())) return E_FAIL;

    hr = CAtlDllModuleT<T>::DllUnregisterServer(bUnRegTypeLib);

    // Zakljuimo modul.
    verify(USPESNO(Zakljuci()));

    return hr;
}


template <class T>
T_STANJE CAOsnAtlDllModuleT<T>::Vzpostavi()
{
    RCakalnik cakalnik(kljucavnica);
    T_STANJE stanje;

    if (je_vzpostavljen)
        return A_OK;

#ifndef AOSN_DLLIMP
    stanje = AOsn::Vzpostavi();
    if (USPESNO(stanje))
#endif
    {
        je_vzpostavljen = TRUE;
        return A_OK;
    }

    return stanje;
}


template <class T>
T_STANJE CAOsnAtlDllModuleT<T>::Zakljuci()
{
    RCakalnik cakalnik(kljucavnica);

    if (!je_vzpostavljen)
        return A_OK;

    verify(USPESNO(AOsn::PocakajVzpostavitev()));
#ifndef AOSN_DLLIMP
    verify(USPESNO(AOsn::Zakljuci()));
#endif

    je_vzpostavljen = FALSE;

    return A_OK;
}

#endif // __AOSNATL_H__
