/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_BITMAP

#pragma comment(lib, "gdi32.lib")


T_VOID AOSN_API AOsn::ZapisiBitmap(HDC dc, HBITMAP slika, RTokPodatkov &datoteka)
{
    assert(0); // TODO: Preizkusi ta del kode.

    BITMAP bmp;
    BITMAPINFOHEADER bmh;
    T_UINT dolzina_palete;
    BITMAPFILEHEADER glava;
    LPBITMAPINFO info;
    LPVOID raster;

    // Zahtevamo parametre slike.
    verify(GetObject(slika, sizeof(BITMAP), (LPSTR)&bmp));

    // Doloimo parametre datoteke.
    memset(&bmh, 0, sizeof(BITMAPINFOHEADER));
    bmh.biSize = sizeof(BITMAPINFOHEADER);
    verify(GetDIBits(dc, slika, 0, (WORD)bmp.bmHeight, NULL, (LPBITMAPINFO)&bmh, DIB_RGB_COLORS));
    dolzina_palete = sizeof(RGBQUAD) * bmh.biClrUsed;

    // Pripravimo in zapiemo strukturo BITMAPFILEHEADER.
    memset(&glava, 0, sizeof(BITMAPFILEHEADER));
    glava.bfType    = *(WORD*)"BM";
    glava.bfSize    = (DWORD)(sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete + bmh.biSizeImage);
    glava.bfOffBits = (DWORD)(sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete);
    verify(datoteka.ZapisiTab(&glava, sizeof(BITMAPFILEHEADER), 1) == 1);

    // Pripravimo in zapiemo strukturo BITMAPINFO, barvno paleto ter raster.
    info = (BITMAPINFO*)aalloc(sizeof(BITMAPINFOHEADER) + dolzina_palete);
    assert(info);
    memcpy(&(info->bmiHeader), &bmh, sizeof(BITMAPINFOHEADER));
    memset(&(info->bmiColors), 0, dolzina_palete);
    raster = aalloc(bmh.biSizeImage);
    assert(raster);
    verify(::GetDIBits(dc, slika, 0, (WORD)bmp.bmHeight, raster, info, DIB_RGB_COLORS));
    verify(datoteka.ZapisiTab(info, sizeof(BITMAPINFOHEADER) + dolzina_palete, 1) == 1);
    verify(datoteka.ZapisiTab(raster, sizeof(T_U1), bmh.biSizeImage) == bmh.biSizeImage);

    // Sprostimo pomnilnik.
    afree(raster);
    afree(info);
}


LPSAFEARRAY AOSN_API AOsn::VrniBitmapSAFEARRAY(HDC dc, HBITMAP slika)
{
    BITMAP bmp;
    BITMAPINFOHEADER bmh;
    T_UINT dolzina_palete;
    LPSAFEARRAY sa;
    T_U1 *podatki;
    LPBITMAPFILEHEADER glava;
    LPBITMAPINFO info;
    LPVOID raster;

    // Zahtevamo parametre slike.
    verify(GetObject(slika, sizeof(BITMAP), (LPSTR)&bmp));

    // Doloimo parametre datoteke.
    memset(&bmh, 0, sizeof(BITMAPINFOHEADER));
    bmh.biSize = sizeof(BITMAPINFOHEADER);
    verify(GetDIBits(dc, slika, 0, (WORD)bmp.bmHeight, NULL, (LPBITMAPINFO)&bmh, DIB_RGB_COLORS));

    // Rezerviramo prostor v pomnilniku.
    dolzina_palete = sizeof(RGBQUAD) * bmh.biClrUsed;
    sa = ::SafeArrayCreateVector(VT_UI1, 0, (ULONG)(sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete + bmh.biSizeImage));
    if (!sa)
        return NULL;

    // Pripravimo kazalce na posamezne dele podatkov slike.
    verify(SUCCEEDED(::SafeArrayAccessData(sa, (T_VOID**)&podatki)));
    glava  =(LPBITMAPFILEHEADER)(podatki + 0);
    info   =      (LPBITMAPINFO)(podatki + sizeof(BITMAPFILEHEADER));
    raster =            (LPVOID)(podatki + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete);

    // Nastavimo strukturo BITMAPFILEHEADER.
    memset(glava, 0, sizeof(BITMAPFILEHEADER));
    glava->bfType    = *(WORD*)"BM";
    glava->bfSize    = (DWORD)(sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete + bmh.biSizeImage);
    glava->bfOffBits = (DWORD)(sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dolzina_palete);

    // Nastavimo strukturo BITMAPINFO.
    memcpy(&(info->bmiHeader), &bmh, sizeof(BITMAPINFOHEADER));
    memset(&(info->bmiColors), 0, dolzina_palete);

    // Nastavimo paleto in raster slike.
    verify(::GetDIBits(dc, slika, 0, (WORD)bmp.bmHeight, raster, info, DIB_RGB_COLORS));

    verify(SUCCEEDED(::SafeArrayUnaccessData(sa)));
    return sa;
}

#endif // AOSN_BITMAP
