/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"
#ifdef _DEBUG
#include <crtdbg.h>
#endif
#ifdef AOSN_DIAGNOSTIKA_AMALLOC
#include <stdlib.h>
#endif

#ifdef _WIN64
#define AMALLOC_KONTROLA    ((T_UINT)0xdcba98766789abcd)
#else
#define AMALLOC_KONTROLA    ((T_UINT)0xdcbaabcd)
#endif

#ifndef HeapEnableTerminationOnCorruption
#define HeapEnableTerminationOnCorruption   (HEAP_INFORMATION_CLASS)1
#endif


//////////////////////////////////////////////////////////////////////
// Inicializacija
//   Spodnji klic zagotavlja, da bo kopica vzpostavljena enkrat med
//   vzpostavitvijo standardne knjinice, ko e sploh ni ve niti. S
//   tem se izognemo nevarnosti, ker funkcija agetheap() ni
//   thread-safe.
//////////////////////////////////////////////////////////////////////

T_ROCICAOS _amalloc_rocica = agetheap();
#ifdef AOSN_DIAGNOSTIKA_AMALLOC
RDatoteka *_amalloc_dnevnik = NULL; // Ne vzpostavljamo ga tukaj, ker ga vzpostavi funkcija agetheap().
RNevarenOdsek *_amalloc_kljucavnica = NULL;
#endif

//////////////////////////////////////////////////////////////////////
// Globalne funkcije
//////////////////////////////////////////////////////////////////////

T_ROCICAOS AOSN_API agetheap()
{
    // Vzpostavitev ni thread safe!!!
    static T_ROCICAOS kopica = NULL;
    if (kopica) return kopica;

#if _MSC_VER >= 1600 && defined(_DEBUG)
    // Vklopimo diagnostiko pomnilnika v knjinici CRT.
    _CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_ALLOC_MEM_DF);
#endif

    // Pripravimo kopico.
    // Ne uporabljaj zastavice HEAP_GENERATE_EXCEPTIONS, saj vpliva na vse kopice procesa!
    // e pa je na modul vtinik DLL, pa ne smemo posegati v nastavitve, ki so globalne za proces.
    // Poleg tega HEAP_GENERATE_EXCEPTIONS tudi precej zakomplicira preverjanje (ne)uspene rezervacije, kar vendarle velikokrat preverimo.
    kopica = ::HeapCreate(0, 0, 0);
    assert(kopica);

#if defined(_DEBUG) || defined(AOSN_DIAGNOSTIKA)
    // Vklopimo preverjanje veljavnosti kopice.
    verify(::HeapSetInformation(kopica, HeapEnableTerminationOnCorruption, NULL, 0));
#endif

#ifdef AOSN_DIAGNOSTIKA_AMALLOC
    // Kopica je vzpostavljena. Pripravimo e datoteko za zapisovanje dnevnika.
    static RFizDatoteka dnevnik;
    static RNevarenOdsek kljucavnica;
    T_ZNAK pot_datoteke[MAX_PATH], pot_modula[MAX_PATH];
    T_NIZ ime_datoteke, ime_modula;
    RDatum datum;
    T_DATUM _datum;

    // Doloimo pot zaasnega imenika.
    verify(::GetTempPath(MAX_PATH, pot_datoteke));
    ime_datoteke = pot_datoteke + astrnlen(pot_datoteke, MAX_PATH);
    if (ime_datoteke - pot_datoteke > 0 && ime_datoteke - pot_datoteke < MAX_PATH && ime_datoteke[-1] != _T(LOCILO_IMENIKOV)) {
        // Prilepimo loilo imenikov na konec.
        ime_datoteke[0] = _T(LOCILO_IMENIKOV); ime_datoteke++;
    }

    // Doloimo ime modula.
    verify(::GetModuleFileName(NULL, pot_modula, MAX_PATH));
    ime_modula = astrrnchr(pot_modula, _T(LOCILO_IMENIKOV), MAX_PATH);
    if (ime_modula)
        ime_modula++;
    else
        ime_modula = pot_modula;

    // Prilepimo ime modula k poti datoteke dnevnika.
    ime_datoteke += astrncpy2(ime_datoteke, MAX_PATH - (ime_datoteke - pot_datoteke), ime_modula, MAX_PATH - (ime_modula - pot_modula));

    // Dodamo e "-", asovno oznako in konnico k imenu datoteke.
    datum.NastaviTrenutniCas();
    datum.VrniDatum(&_datum);
#ifdef _sntprintf_s
    _sntprintf_s(ime_datoteke, MAX_PATH - (ime_datoteke - pot_datoteke), _TRUNCATE, _T("-%04u%02u%02u%02u%02u%02u.amb"), (int)_datum.leto, (int)_datum.mesec, (int)_datum.dan, (int)_datum.ura, (int)_datum.minuta, (int)_datum.sekunda);
#else
    _sntprintf  (ime_datoteke, MAX_PATH - (ime_datoteke - pot_datoteke),            _T("-%04u%02u%02u%02u%02u%02u.amb"), (int)_datum.leto, (int)_datum.mesec, (int)_datum.dan, (int)_datum.ura, (int)_datum.minuta, (int)_datum.sekunda);
#endif

    // Odpremo datoteko.
    verify(USPESNO(dnevnik.Odpri(pot_datoteke, RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_BINARNO | RDAT_DOSTOP_NAKLJUCEN)));

#ifdef AOSN_DIAGNOSTIKA_AMALLOC_ZAPISOV_MAX
    // Skoimo na zadnji bajt v datoteki in zapiemo vanj eno nilo. Da sistem naredi datoteko elene doline.
    verify(dnevnik.PrestaviZac(sizeof(T_U8) + AOSN_DIAGNOSTIKA_AMALLOC_ZAPISOV_MAX * sizeof(T_AMALLOC_ZAPIS_OPERACIJE) - 1) == sizeof(T_U8) + AOSN_DIAGNOSTIKA_AMALLOC_ZAPISOV_MAX * sizeof(T_AMALLOC_ZAPIS_OPERACIJE) - 1);
    verify(USPESNO(dnevnik.ZapisiBajt(0)));

    // Prestavimo se nazaj na zaetek datoteke.
    verify(dnevnik.PrestaviZac(0) == 0);
#endif

    // Zapiemo glavo.
    dnevnik << (T_U8)sizeof(T_U8);
    assert(USPESNO(dnevnik.VrniStanje()));

    verify(USPESNO(dnevnik.Izplakni()));

    // Nastavimo globalna kazalca na datoteko z dnevnikom in njen semafor.
    _amalloc_dnevnik     = &dnevnik;
    _amalloc_kljucavnica = &kljucavnica;
#endif // AOSN_DIAGNOSTIKA_AMALLOC

    return kopica;
}


#ifdef AOSN_DIAGNOSTIKA_AMALLOC

inline T_STANJE _amalloc_zapisi(const T_AMALLOC_ZAPIS_OPERACIJE *z)
{
    assert(z);
    RCakalnik cakalnik(*_amalloc_kljucavnica);
    T_ODMIK odmik;
    T_STANJE stanje;

    odmik = _amalloc_dnevnik->VrniOdmik();
#ifdef AOSN_DIAGNOSTIKA_AMALLOC_ZAPISOV_MAX
    if (odmik >= sizeof(T_U8) + AOSN_DIAGNOSTIKA_AMALLOC_ZAPISOV_MAX * sizeof(T_AMALLOC_ZAPIS_OPERACIJE)) {
        // Presegli smo konec datoteke. Prestavimo se nazaj na zaetek.
        odmik = _amalloc_dnevnik->PrestaviZac(sizeof(T_U8));
        stanje = _amalloc_dnevnik->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
    }
#endif

    // Zapiemo zapis.
    odmik += _amalloc_dnevnik->ZapisiTab(z, sizeof(T_AMALLOC_ZAPIS_OPERACIJE), 1) * sizeof(T_AMALLOC_ZAPIS_OPERACIJE);
    stanje = _amalloc_dnevnik->VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;

    // Prestavimo se nazaj na zaetek datoteke in popravimo glavo.
    _amalloc_dnevnik->PrestaviZac(0);
    stanje = _amalloc_dnevnik->VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    (*_amalloc_dnevnik) << (T_U8)odmik;
    stanje = _amalloc_dnevnik->VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;

#ifdef AOSN_DIAGNOSTIKA_AMALLOC_VELIKOST_MAX
    if (z->dolzina > AOSN_DIAGNOSTIKA_AMALLOC_VELIKOST_MAX || z->dolzina2 > AOSN_DIAGNOSTIKA_AMALLOC_VELIKOST_MAX) {
        // Izplaknemo sistemske medpomnilnike.
        stanje = _amalloc_dnevnik->Izplakni();
        if (NEUSPESNO(stanje)) return stanje;

        // Zakljuimo program.
        exit(1);
    }
#endif

    // Vrnemo se na vsebino.
    _amalloc_dnevnik->PrestaviZac(odmik);
    stanje = _amalloc_dnevnik->VrniStanje();
    return stanje;
}


inline T_STANJE _amalloc_zapisi(T_AMALLOC_OPERACIJA operacija, T_KAZALEC pom, TK_NIZ ime_izvorne_datoteke_rezervacije = NULL, T_UINT vrstica_rezervacije = 0, TK_NIZ ime_izvorne_datoteke = NULL, T_UINT vrstica = 0, T_SIZE dolzina = 0, T_SIZE dolzina2 = 0)
{
    T_AMALLOC_ZAPIS_OPERACIJE z;

    memset(&z, 0, sizeof(T_AMALLOC_ZAPIS_OPERACIJE));

    z.operacija = (T_U1)operacija;
    z.pom       = (T_U8)pom;

    if (ime_izvorne_datoteke_rezervacije)
        astrncpy(z.ime_izvorne_datoteke_rezervacije, MAX_PATH, ime_izvorne_datoteke_rezervacije, T_UINT_MAX);
    z.vrstica_rezervacije = (T_U4)vrstica_rezervacije;

    if (ime_izvorne_datoteke)
        astrncpy(z.ime_izvorne_datoteke, MAX_PATH, ime_izvorne_datoteke, T_UINT_MAX);
    z.vrstica  = (T_U4)vrstica;

    z.dolzina  = dolzina;
    z.dolzina2 = dolzina2;

    return _amalloc_zapisi(&z);
}

#endif // AOSN_DIAGNOSTIKA_AMALLOC


#if defined(_DEBUG) || defined(AOSN_DIAGNOSTIKA)

T_KAZALEC AOSN_API _aalloc_diag(T_SIZE dolzina, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    T_ROCICAOS kopica = agetheap();
#ifdef AOSN_DIAGNOSTIKA_AMALLOC
    _amalloc_zapisi(AMALLOC_OPERACIJA_AALLOC, NULL, NULL, 0, ime_izvorne_datoteke, vrstica, dolzina);
#endif
    T_SIZE dolzina_imena = (astrlen(ime_izvorne_datoteke) + 1) * sizeof(T_ZNAK);
    T_KAZALEC pom, p;

    // Podaljamo prostor za ime izvorne datoteke, da bo poravnan na celo tevilo T_UINT-ov.
    dolzina_imena = (dolzina_imena + sizeof(T_UINT) - 1) / sizeof(T_UINT) * sizeof(T_UINT);

    // Rezerviramo veji kos pomnilnika, da vanj spravimo e ime izvorne datoteke, tevilko vrstice ter kontrolni kodi.
    pom = ::HeapAlloc(kopica, 0, 4*sizeof(T_UINT) + dolzina_imena + dolzina);
    if (!pom) {
        // Rezervacija ni uspela, vrnemo NULL.
        return NULL;
    }

    p = pom;

    // Zapiemo tevilko vrstice.
    *(T_UINT*)p = vrstica;
    p = (T_U1*)p + sizeof(T_UINT);

    // Zapiemo ime izvorne datoteke.
    astrncpy((T_NIZ)p, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke, T_UINT_MAX);
    p = (T_U1*)p + dolzina_imena;

    // Zapiemo dolino imena izvorne datoteke.
    *(T_UINT*)p = dolzina_imena;
    p = (T_U1*)p + sizeof(T_UINT);

    // Zapiemo kontrolo na zaetek podatkov.
    *(T_UINT*)p = AMALLOC_KONTROLA;
    p = (T_U1*)p + sizeof(T_UINT);

    // Dodamo kontrolo na konec podatkov.
    *(T_UINT*)((T_U1*)p + dolzina) = AMALLOC_KONTROLA;

    return p;
}


T_KAZALEC AOSN_API _arealloc_diag(T_KAZALEC pom, T_SIZE dolzina2, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    if (dolzina2) {
        // Ali je pomnilnik e rezerviran?
        if (pom) {
            T_ROCICAOS kopica = agetheap();
            T_KAZALEC p = pom;
            TK_NIZ ime_izvorne_datoteke_rezervacije;
            T_UINT vrstica_rezervacije;
            T_SIZE dolzina_imena, dolzina;

            // Preverimo zaetno kontrolo.
            assert((T_SIZE)p > sizeof(T_UINT));
            p = (T_U1*)p - sizeof(T_UINT);
            if (*(T_UINT*)p != AMALLOC_KONTROLA) {
                assert(0);
                adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo skuali uporabiti v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, ime_izvorne_datoteke, vrstica);
                return pom;
            }

            // Preberemo dolzino imena izvorne datoteke.
            assert((T_SIZE)p > sizeof(T_UINT));
            p = (T_U1*)p - sizeof(T_UINT);
            dolzina_imena = *(T_UINT*)p;

            // Preberemo ime izvorne datoteke.
            assert((T_SIZE)p > dolzina_imena);
            p = (T_U1*)p - dolzina_imena;
            ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;

            // Preberemo vrstico izvorne datoteke.
            assert((T_SIZE)p > sizeof(T_UINT));
            p = (T_U1*)p - sizeof(T_UINT);
            vrstica_rezervacije = *(T_UINT*)p;

            // Nali smo zaetek bloka.
            dolzina = ::HeapSize(kopica, 0, p) - 4*sizeof(T_UINT) - dolzina_imena;

#ifdef AOSN_DIAGNOSTIKA_AMALLOC
            _amalloc_zapisi(AMALLOC_OPERACIJA_AREALLOC, pom, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica, dolzina, dolzina2);
#endif

            // Zdaj imamo dolino in lahko preverimo e konno kontrolo.
            if (*(T_UINT*)((T_U1*)pom + dolzina) != AMALLOC_KONTROLA) {
                assert(0);
                adebug(ADNE_NAPAKA, _T("Konec pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo rezervirali v datoteki \"%.*s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije);
                return pom;
            }

            // Skrajamo/podaljamo pomnilnik.
            pom = ::HeapReAlloc(kopica, 0, p, 4*sizeof(T_UINT) + dolzina_imena + dolzina2);
            if (!pom) {
                // Rezervacija ni uspela, vrnemo NULL.
                return NULL;
            }

            // Ponovno zapiemo kontrolo na konec podatkov.
            p = (T_U1*)pom + 3*sizeof(T_UINT) + dolzina_imena;
            *(T_UINT*)((T_U1*)p + dolzina2) = AMALLOC_KONTROLA;
            return p;
        } else {
            // Pomnilnik e ni rezerviran.
            return _aalloc_diag(dolzina2, ime_izvorne_datoteke, vrstica);
        }
    } else if (pom) {
        // Kadar uporabnik nastavi dolino veljavnega bloka na 0, sprostimo pomnilnik.
        _afree_diag(pom, ime_izvorne_datoteke, vrstica);
    }

    return NULL;
}


T_KAZALEC AOSN_API _aqrealloc_diag(T_KAZALEC pom, T_SIZE dolzina2, T_BOOL ne_krajsaj, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    // Ali je pomnilnik e rezerviran?
    if (pom) {
        T_ROCICAOS kopica = agetheap();
        T_KAZALEC p = pom;
        TK_NIZ ime_izvorne_datoteke_rezervacije;
        T_UINT vrstica_rezervacije;
        T_SIZE dolzina_imena, dolzina;

        // Preverimo zaetno kontrolo.
        assert((T_SIZE)p > sizeof(T_UINT));
        p = (T_U1*)p - sizeof(T_UINT);
        if (*(T_UINT*)p != AMALLOC_KONTROLA) {
            assert(0);
            adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo skuali uporabiti v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, ime_izvorne_datoteke, vrstica);
            return pom;
        }

        // Preberemo dolino imena izvorne datoteke.
        assert((T_SIZE)p > sizeof(T_UINT));
        p = (T_U1*)p - sizeof(T_UINT);
        dolzina_imena = *(T_UINT*)p;

        // Preberemo ime izvorne datoteke.
        assert((T_SIZE)p > dolzina_imena);
        p = (T_U1*)p - dolzina_imena;
        ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;

        // Preberemo vrstico izvorne datoteke.
        assert((T_SIZE)p > sizeof(T_UINT));
        p = (T_U1*)p - sizeof(T_UINT);
        vrstica_rezervacije = *(T_UINT*)p;

        // Nali smo zaetek bloka.
        dolzina = ::HeapSize(kopica, 0, p) - 4*sizeof(T_UINT) - dolzina_imena;

#ifdef AOSN_DIAGNOSTIKA_AMALLOC
        _amalloc_zapisi(AMALLOC_OPERACIJA_AREALLOC, pom, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica, dolzina, dolzina2);
#endif

        // Zdaj imamo dolino in lahko preverimo e konno kontrolo.
        if (*(T_UINT*)((T_U1*)pom + dolzina) != AMALLOC_KONTROLA) {
            assert(0);
            adebug(ADNE_NAPAKA, _T("Konec pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo rezervirali v datoteki \"%.*s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije);
            return pom;
        }

        if (dolzina >= dolzina2) {
            // Uporabnik rabi manj pomnilnika.
            if (ne_krajsaj) {
                // Uporabnik ni elel krajati pomnilnika.
                return pom;
            } else {
                // Skrajamo pomnilnik.
                pom = ::HeapReAlloc(kopica, 0, p, 4*sizeof(T_UINT) + dolzina_imena + dolzina2);
                if (!pom)
                    return NULL;

                // Ponovno zapiemo kontrolo na konec podatkov.
                p = (T_U1*)pom + 3*sizeof(T_UINT) + dolzina_imena;
                *(T_UINT*)((T_U1*)p + dolzina2) = AMALLOC_KONTROLA;
                return p;
            }
        }

        // e moramo podaljati, pomnilnik, ga najprej sprostimo.
        ::HeapFree(kopica, 0, p);
    }

    return _aalloc_diag(dolzina2, ime_izvorne_datoteke, vrstica);
}


T_VOID AOSN_API _afree_diag(T_KAZALEC pom, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    T_ROCICAOS kopica = agetheap();
    T_KAZALEC p = pom;
    TK_NIZ ime_izvorne_datoteke_rezervacije;
    T_UINT vrstica_rezervacije;
    T_SIZE dolzina_imena, dolzina;

    // Preverimo zaetno kontrolo.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    if (*(T_UINT*)p != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo skuali uporabiti v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, ime_izvorne_datoteke, vrstica);
        return;
    }

    // Preberemo dolino imena izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    dolzina_imena = *(T_UINT*)p;

    // Preberemo ime izvorne datoteke.
    assert((T_SIZE)p > dolzina_imena);
    p = (T_U1*)p - dolzina_imena;
    ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;

    // Preberemo vrstico izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    vrstica_rezervacije = *(T_UINT*)p;

    // Nali smo zaetek bloka.
    dolzina = ::HeapSize(kopica, 0, p) - 4*sizeof(T_UINT) - dolzina_imena;

#ifdef AOSN_DIAGNOSTIKA_AMALLOC
    _amalloc_zapisi(AMALLOC_OPERACIJA_AFREE, pom, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica, dolzina);
#endif

    // Zdaj imamo dolino in lahko preverimo e konno kontrolo.
    if (*(T_UINT*)((T_U1*)pom + dolzina) != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Konec pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo rezervirali v datoteki \"%.*s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije);
        return;
    }

    ::HeapFree(kopica, 0, p);
    return;
}


T_SIZE AOSN_API _amsize_diag(T_KAZALEC pom, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    T_ROCICAOS kopica = agetheap();
    T_KAZALEC p = pom;
    TK_NIZ ime_izvorne_datoteke_rezervacije;
    T_UINT vrstica_rezervacije;
    T_SIZE dolzina_imena, dolzina;

    // Preverimo zaetno kontrolo.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    if (*(T_UINT*)p != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo skuali uporabiti v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, ime_izvorne_datoteke, vrstica);
        return 0;
    }

    // Preberemo dolino imena izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    dolzina_imena = *(T_UINT*)p;

    // Preberemo ime izvorne datoteke.
    assert((T_SIZE)p > dolzina_imena);
    p = (T_U1*)p - dolzina_imena;
    ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;

    // Preberemo vrstico izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    vrstica_rezervacije = *(T_UINT*)p;

    // Nali smo zaetek bloka.
    dolzina = ::HeapSize(kopica, 0, p) - 4*sizeof(T_UINT) - dolzina_imena;

#ifdef AOSN_DIAGNOSTIKA_AMALLOC
    _amalloc_zapisi(AMALLOC_OPERACIJA_AMSIZE, pom, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica, dolzina);
#endif

    // Zdaj imamo dolino in lahko preverimo e konno kontrolo.
    if (*(T_UINT*)((T_U1*)pom + dolzina) != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Konec pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo rezervirali v datoteki \"%.*s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije);
        return 0;
    }

    return dolzina;
}


T_BOOL AOSN_API _aheapvalid_diag(T_KAZALEC pom, T_SIZE dolzina2, TK_NIZ ime_izvorne_datoteke, T_UINT vrstica)
{
    T_ROCICAOS kopica = agetheap();
    T_KAZALEC p = pom;
    TK_NIZ ime_izvorne_datoteke_rezervacije;
    T_UINT vrstica_rezervacije;
    T_SIZE dolzina_imena, dolzina;

    // Preverimo zaetno kontrolo.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    if (*(T_UINT*)p != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo skuali preveriti v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, ime_izvorne_datoteke, vrstica);
        return FALSE;
    }

    // Preberemo dolino imena izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    dolzina_imena = *(T_UINT*)p;

    // Preberemo ime izvorne datoteke.
    assert((T_SIZE)p > dolzina_imena);
    p = (T_U1*)p - dolzina_imena;
    ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;

    // Preberemo vrstico izvorne datoteke.
    assert((T_SIZE)p > sizeof(T_UINT));
    p = (T_U1*)p - sizeof(T_UINT);
    vrstica_rezervacije = *(T_UINT*)p;

    // Nali smo zaetek bloka.
    dolzina = ::HeapSize(kopica, 0, p) - 4*sizeof(T_UINT) - dolzina_imena;
#ifdef AOSN_DIAGNOSTIKA_AMALLOC
    _amalloc_zapisi(AMALLOC_OPERACIJA_AHEAPVALID, pom, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica, dolzina, dolzina2);
#endif
    if (dolzina < dolzina2) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Dolina pomnilnikega bloka na naslovu %p je kraja od priakovane (%") _T(T_SIZE_PRINTF_SIZE_PREFIX) _T("u namesto %") _T(T_SIZE_PRINTF_SIZE_PREFIX) _T("u B). Blok smo rezervirali v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, preverjali pa v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina, dolzina2, ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica);
        return FALSE;
    }

    // Zdaj imamo dolino in lahko preverimo e konno kontrolo.
    if (*(T_UINT*)((T_U1*)pom + dolzina) != AMALLOC_KONTROLA) {
        assert(0);
        adebug(ADNE_NAPAKA, _T("Konec pomnilnikega bloka na naslovu %p je pokvarjen. Blok smo rezervirali v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, preverjali pa v datoteki \"%s\", vrstica %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.\n"), pom, dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, ime_izvorne_datoteke, vrstica);
        return FALSE;
    }

    // Preizkusimo, ali je pomnilniki blok v kopici veljaven.
    return ::HeapValidate(kopica, 0, p) ? TRUE: FALSE;
}


T_BOOL AOSN_API _adumpheap()
{
    T_ROCICAOS kopica = agetheap();
    T_BOOL vsebuje_bloke = FALSE;

#ifdef _DEBUG
#if _MSC_VER >= 1600
    // Vklopimo sporoanje o puanju pomnilnika v knjinici CRT.
    _CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);
#else
    if (_CrtDumpMemoryLeaks()) {
        // Nali smo nesproene pomnilnike bloke.

        // Sproimo past in obvestimo uporabnika, da lahko pred nadaljevanjem zaene orodje DbgView (e razhroevalnik ni na voljo).
        assert(0);

        // Zdaj pa e enkrat kliemo funkcijo _CrtDumpMemoryLeaks(), da bo orodje DbgView videlo, kaj je bilo narobe.
        _CrtDumpMemoryLeaks();
        vsebuje_bloke = TRUE;
    }
#endif
#endif

    if (::HeapLock(kopica)) {
        PROCESS_HEAP_ENTRY he;
        T_KAZALEC prvi_blok = NULL;   // kazalec na prvi blok v pomnilniki regiji (prvi blok v regiji je obiajno za reijske namene, zato ga ne obravnavamo kot pozabljenega)
        he.lpData = NULL;

        while (::HeapWalk(kopica, &he)) {
            if (he.wFlags & PROCESS_HEAP_REGION) {
                prvi_blok = he.Region.lpFirstBlock;
            } else if ((he.wFlags & PROCESS_HEAP_ENTRY_BUSY) && he.lpData != prvi_blok) {
                // Nali smo nesproeni pomnilniki blok.

                // Past sproimo kar takoj, da lahko pred nadaljevanjem zaenemo orodje DbgView (e razhroevalnik ni na voljo). V nadaljevanju pa bomo z adebug() izpisali ve informacij.
                assert(0);

                T_UINT i;
                T_KAZALEC p = he.lpData;
                TK_NIZ ime_izvorne_datoteke_rezervacije;
                T_UINT vrstica_rezervacije;
                T_SIZE dolzina_imena, dolzina;

                // Preberemo vrstico izvorne datoteke.
                vrstica_rezervacije = *(T_UINT*)p;
                p = (T_U1*)p + sizeof(T_UINT);

                // Preberemo ime izvorne datoteke.
                ime_izvorne_datoteke_rezervacije = (TK_NIZ)p;
                dolzina_imena = (astrlen(ime_izvorne_datoteke_rezervacije) + 1) * sizeof(T_ZNAK);
                dolzina_imena = (dolzina_imena + sizeof(T_UINT) - 1) / sizeof(T_UINT) * sizeof(T_UINT);
                p = (T_U1*)p + dolzina_imena;

                // Preverimo dolino imena izvorne datoteke.
                if (*(T_UINT*)p == dolzina_imena) {
                    p = (T_U1*)p + sizeof(T_UINT);

                    // Preverimo zaetno kontrolo.
                    if (*(T_UINT*)p == AMALLOC_KONTROLA) {
                        p = (T_U1*)p + sizeof(T_UINT);

                        // Nali smo zaetek bloka.
                        dolzina = he.cbData - 4*sizeof(T_UINT) - dolzina_imena;

                        adebug(ADNE_OPOZORILO, _T("%.*s(%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u) : Nesproen pomnilniki blok na naslovu %p (dolina: %") _T(T_SIZE_PRINTF_SIZE_PREFIX) _T("u B, regija: %") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u)\n\t"), dolzina_imena / sizeof(T_ZNAK), ime_izvorne_datoteke_rezervacije, vrstica_rezervacije, p, dolzina, (T_U4)he.iRegionIndex);
                        for (i = 0; i < 16 && i < dolzina; i++)
                            adebug(ADNE_OPOZORILO, _T("%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("x "), (T_U4)((LPBYTE)p)[i]);
                        adebug(ADNE_OPOZORILO, _T("\t"));
                        for (i = 0; i < 16 && i < dolzina; i++)
                            adebug(ADNE_OPOZORILO, _T("%hc"), ((LPBYTE)p)[i] >= ' ' ? ((LPBYTE)p)[i] : '.');
                        adebug(ADNE_OPOZORILO, i < dolzina ? _T("...\n") : _T("\n"));

                        if (*(T_UINT*)((T_U1*)p + dolzina) != AMALLOC_KONTROLA) {
                            adebug(ADNE_NAPAKA, _T("\tPa e konec pomnilnikega bloka je pokvarjen!\n"));
                            return 0;
                        }

                        if (!::HeapValidate(kopica, 0, he.lpData))
                            adebug(ADNE_NAPAKA, _T("\t^^^^^ !!! Pokvarjen !!! ^^^^^\n"));
                    } else
                        adebug(ADNE_NAPAKA, _T("Zaetek pomnilnikega bloka na naslovu %p je pokvarjen, ali pa ga nismo rezervirali z uporabo diagnostinih funkcij AMalloc.\n"), he.lpData);
                } else
                    adebug(ADNE_NAPAKA, _T("Dolina imena izvorne datoteke rezervacije pomnilnikega bloka na naslovu %p se ne ujema z imenom.\n"), he.lpData);

                vsebuje_bloke = TRUE;
            }
        }

        ::HeapUnlock(kopica);
    }

    return vsebuje_bloke;
}

#endif // _DEBUG || AOSN_DIAGNOSTIKA
