/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"
#include "AOsnP.h"
#include <winhttp.h>
#ifdef AOSN_VTIC
#include <winsock2.h>
#endif
#include <windows.h>

#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "version.lib")


//////////////////////////////////////////////////////////////////////
// Interni podatkovni tipi
//////////////////////////////////////////////////////////////////////

struct T_NAZIV_KODNE_STRANI {
    T_KODNASTRAN kodna_stran;
    TK_ASCII naziv;
};


struct T_NAZIV_JEZIKA {
    T_KODAJEZIKA2 koda_jezika;          // koda jezika
    TK_ASCII naziv;                     // naziv jezika
    T_UINT st_podjezikov;               // tevilo podjezikov
    const T_NAZIV_JEZIKA *podjeziki;    // kazalec na tabelo podjezikov
};


//////////////////////////////////////////////////////////////////////
// Deklaracije internih funkcij
//////////////////////////////////////////////////////////////////////

static T_KODAJEZIKA2 _VrniKodoJezika(const T_NAZIV_JEZIKA *tabela, T_UINT st_elementov, TK_ASCII naziv_rfc1766, T_UINT dolzina);
#ifdef AOSN_IZJEMA
static LONG WINAPI _RokovalnikNeujetihIzjem(PEXCEPTION_POINTERS ep);
#endif


//////////////////////////////////////////////////////////////////////
// Globalne spremenljivke in statini podatki
//////////////////////////////////////////////////////////////////////

T_LOCALE AOSN_API AOsn::locale_C         = NULL;                              // kazalec na jezikovno neodvisni locale
T_LOCALE AOSN_API AOsn::locale_ACP       = NULL;                              // kazalec na locale z uporabnikovo kodno stranjo
T_LOCALE AOSN_API AOsn::locale_OCP       = NULL;                              // kazalec na locale z uporabnikovo kodno stranjo (za konzolne aplikacije)
T_VERZIJAOS       AOsn::verzija_os       = AOSN_OS_NEZNAN;                    // verzija operacijskega sistema
#ifndef _WIN64
T_BOOL            AOsn::je_proces_wow64  = FALSE;                             // ali na 32-bitni proces tee znotraj 64-bitnega okolja
#endif
T_UINT            AOsn::st_vzpostavitev  = 0;                                 // kolikokrat smo vzpostavili knjinico AOsn
T_ROCICAOS        AOsn::kljucavnica      = ::CreateMutex(NULL, FALSE, NULL);  // sinhronizacija vzpostavitev in zakljuitev
static HMODULE    rocica_kernel32 = NULL;
#ifndef _WIN64
BOOL    (WINAPI *AOsn::IsWow64Process)(__in HANDLE hProcess, __out PBOOL Wow64Process) = NULL;
#endif
#ifdef RDATUM_NALOZI_JEDRO
VOID    (WINAPI *AOsn::GetSystemTimeAsFileTime)(__out LPFILETIME lpSystemTimeAsFileTime) = NULL;
#endif
static HMODULE   rocica_user32 = NULL;
HWINSTA (WINAPI *AOsn::GetProcessWindowStation)() = NULL;
BOOL    (WINAPI *AOsn::GetUserObjectInformation)(__in HANDLE hObj, __in int nIndex, __out_opt PVOID pvInfo, __in DWORD nLength, __out_opt LPDWORD lpnLengthNeeded) = NULL;
#ifdef AOSN_IZJEMA
RNevarenOdsek AOsn::kljucavnica_izjem;                             // sinhronizacija obdelave napak
LPTOP_LEVEL_EXCEPTION_FILTER AOsn::rokovalnik_izjem_prej = NULL;   // kazalec prejnjega rokovalnika izjem
RIzjemeGlobalnaNit* AOsn::izjeme_nit                     = NULL;   // nit, ki v ozadju vzpostavi rokovanje izjem in preveri, ali aka kakna datoteka s poroilom o napaki na poiljanje in jo polje streniku
#endif
HANDLE AOsn::rocica_std[3] = { NULL,  NULL,  NULL  };  // roice standardnega vhoda, izhoda in napake konzole
T_BOOL AOsn::je_konzola[3] = { FALSE, FALSE, FALSE };  // logine spremenljivke, ki doloajo, ali je standardni vhod/izhod/napaka konzola ali datoteka


const T_NAZIV_KODNE_STRANI nazivi_kodnih_strani[] = {
    { KODNASTRAN_WIN1250,    "windows-1250" },
    { KODNASTRAN_WIN1251,    "windows-1251" },
    { KODNASTRAN_WIN1252,    "windows-1252" },
    { KODNASTRAN_WIN1253,    "windows-1253" },
    { KODNASTRAN_WIN1254,    "windows-1254" },
    { KODNASTRAN_WIN1255,    "windows-1255" },
    { KODNASTRAN_WIN1256,    "windows-1256" },
    { KODNASTRAN_WIN1257,    "windows-1257" },
    { KODNASTRAN_WIN1258,    "windows-1258" },
    { KODNASTRAN_WIN874,     "windows-874"  },
    { KODNASTRAN_WIN949,     "euc-kr"       },
    { KODNASTRAN_SHIFTJIS,   "shift_jis"    },
    { KODNASTRAN_GBK,        "gb2312"       },
    { KODNASTRAN_BIG5,       "big5"         },
    { KODNASTRAN_IBM437,     "us-ascii"     },
    { KODNASTRAN_IBM852,     "cp852"        },
    { KODNASTRAN_ISO8859_1,  "iso-8859-1"   },
    { KODNASTRAN_ISO8859_2,  "iso-8859-2"   },
    { KODNASTRAN_ISO8859_3,  "iso-8859-3"   },
    { KODNASTRAN_ISO8859_4,  "iso-8859-4"   },
    { KODNASTRAN_ISO8859_5,  "iso-8859-5"   },
    { KODNASTRAN_ISO8859_6,  "iso-8859-6"   },
    { KODNASTRAN_ISO8859_8,  "iso-8859-8"   },
    { KODNASTRAN_ISO8859_9,  "iso-8859-9"   },
    { KODNASTRAN_ISO8859_15, "iso-8859-15"  },
    { KODNASTRAN_UTF7,       "utf-7"        },
    { KODNASTRAN_UTF8,       "utf-8"        },
    { KODNASTRAN_UTF16,      "utf-16"       },
};


const T_NAZIV_JEZIKA nazivi_jezikov_arabscina[] = {
    { (T_KODAJEZIKA2)14337, "-ae" },    // Arabic(U.A.E.)
    { (T_KODAJEZIKA2)15361, "-bh" },    // Arabic(Bahrain)
    { (T_KODAJEZIKA2)5121,  "-dz" },    // Arabic(Algeria)
    { (T_KODAJEZIKA2)3073,  "-eg" },    // Arabic(Egypt)
    { (T_KODAJEZIKA2)2049,  "-iq" },    // Arabic(Iraq)
    { (T_KODAJEZIKA2)11265, "-jo" },    // Arabic(Jordan)
    { (T_KODAJEZIKA2)13313, "-kw" },    // Arabic(Kuwait)
    { (T_KODAJEZIKA2)12289, "-lb" },    // Arabic(Lebanon)
    { (T_KODAJEZIKA2)4097,  "-ly" },    // Arabic(Libya)
    { (T_KODAJEZIKA2)6145,  "-ma" },    // Arabic(Morocco)
    { (T_KODAJEZIKA2)8193,  "-om" },    // Arabic(Oman)
    { (T_KODAJEZIKA2)16385, "-qa" },    // Arabic(Qatar)
    { (T_KODAJEZIKA2)1025,  "-sa" },    // Arabic(Saudi Arabia)
    { (T_KODAJEZIKA2)10241, "-sy" },    // Arabic(Syria)
    { (T_KODAJEZIKA2)7169,  "-tn" },    // Arabic(Tunisia)
    { (T_KODAJEZIKA2)9217,  "-ye" },    // Arabic(Yemen)
};


const T_NAZIV_JEZIKA nazivi_jezikov_nemscina[] = {
    { (T_KODAJEZIKA2)3079,  "-at" },    // German(Austrian)
    { (T_KODAJEZIKA2)2055,  "-ch" },    // German(Swiss)
    { (T_KODAJEZIKA2)5127,  "-li" },    // German(Liechtenstein)
    { (T_KODAJEZIKA2)4103,  "-lu" },    // German(Luxembourg)
};


const T_NAZIV_JEZIKA nazivi_jezikov_anglescina[] = {
    { (T_KODAJEZIKA2)3081,  "-au" },    // English(Australian)
    { (T_KODAJEZIKA2)10249, "-bz" },    // English(Belize)
    { (T_KODAJEZIKA2)4105,  "-ca" },    // English(Canadian)
    { (T_KODAJEZIKA2)9225,  "-ca" },    // English(Caribbean)
    { (T_KODAJEZIKA2)2057,  "-gb" },    // English(British)
    { (T_KODAJEZIKA2)6153,  "-ie" },    // English(Ireland)
    { (T_KODAJEZIKA2)8201,  "-jm" },    // English(Jamaica)
    { (T_KODAJEZIKA2)5129,  "-nz" },    // English(New Zealand)
    { (T_KODAJEZIKA2)11273, "-tt" },    // English(Trinidad)
    { (T_KODAJEZIKA2)1033,  "-us" },    // English(United States)
    { (T_KODAJEZIKA2)7177,  "-za" },    // English(South Africa)
};


const T_NAZIV_JEZIKA nazivi_jezikov_spanscina[] = {
    { (T_KODAJEZIKA2)11274, "-ar" },    // Spanish(Argentina)
    { (T_KODAJEZIKA2)16394, "-bo" },    // Spanish(Bolivia)
    { (T_KODAJEZIKA2)13322, "-c"  },    // Spanish(Chile)
    { (T_KODAJEZIKA2)9226,  "-co" },    // Spanish(Colombia)
    { (T_KODAJEZIKA2)5130,  "-cr" },    // Spanish(Costa Rica)
    { (T_KODAJEZIKA2)7178,  "-do" },    // Spanish(Dominican Republic)
    { (T_KODAJEZIKA2)12298, "-ec" },    // Spanish(Ecuador)
    { (T_KODAJEZIKA2)4106,  "-gt" },    // Spanish(Guatemala)
    { (T_KODAJEZIKA2)18442, "-hn" },    // Spanish(Honduras)
    { (T_KODAJEZIKA2)2058,  "-mx" },    // Spanish(Mexican)
    { (T_KODAJEZIKA2)19466, "-ni" },    // Spanish(Nicaragua)
    { (T_KODAJEZIKA2)6154,  "-pa" },    // Spanish(Panama)
    { (T_KODAJEZIKA2)10250, "-pe" },    // Spanish(Peru)
    { (T_KODAJEZIKA2)20490, "-pr" },    // Spanish(Puerto Rico)
    { (T_KODAJEZIKA2)15370, "-py" },    // Spanish(Paraguay)
    { (T_KODAJEZIKA2)17418, "-sv" },    // Spanish(El Salvador)
    { (T_KODAJEZIKA2)14346, "-uy" },    // Spanish(Uruguay)
    { (T_KODAJEZIKA2)8202,  "-ve" },    // Spanish(Venezuela)
};


const T_NAZIV_JEZIKA nazivi_jezikov_francoscina[] = {
    { (T_KODAJEZIKA2)2060,  "-be" },    // French(Belgian)
    { (T_KODAJEZIKA2)3084,  "-ca" },    // French(Canadian)
    { (T_KODAJEZIKA2)4108,  "-ch" },    // French(Swiss)
    { (T_KODAJEZIKA2)5132,  "-lu" },    // French(Luxembourg)
};


const T_NAZIV_JEZIKA nazivi_jezikov_italijanscina[] = {
    { (T_KODAJEZIKA2)2064,  "-ch" },    // Italian(Swiss)
};


const T_NAZIV_JEZIKA nazivi_jezikov_korejscina[] = {
    { (T_KODAJEZIKA2)2066,  "-johab" }, // Korean(Johab)
};


const T_NAZIV_JEZIKA nazivi_jezikov_nizozemscina[] = {
    { (T_KODAJEZIKA2)2067,  "-be" },    // Dutch(Belgian)
};


const T_NAZIV_JEZIKA nazivi_jezikov_norvescina[] = {
    { (T_KODAJEZIKA2)1044,  "-bokmaal" },   // Norwegian(Bokmaal)
    { (T_KODAJEZIKA2)2068,  "-nynorsk" },   // Norwegian(Nynorsk)
};


const T_NAZIV_JEZIKA nazivi_jezikov_portugalscina[] = {
    { (T_KODAJEZIKA2)1046,  "-br" },    // Portuguese(Brazil)
};


const T_NAZIV_JEZIKA nazivi_jezikov_romunscina[] = {
    { (T_KODAJEZIKA2)2072,  "-mo" },    // Romanian(Moldavia)
};


const T_NAZIV_JEZIKA nazivi_jezikov_ruscina[] = {
    { (T_KODAJEZIKA2)2073,  "-mo" },    // Russian(Moldavia)
};


const T_NAZIV_JEZIKA nazivi_jezikov_srbscina[] = {
    { (T_KODAJEZIKA2)3098,  "-cyrillic" },  // Serbian(Cyrillic)
};


const T_NAZIV_JEZIKA nazivi_jezikov_svedscina[] = {
    { (T_KODAJEZIKA2)2077,  "-fi" },    // Swedish(Finland)
};


const T_NAZIV_JEZIKA nazivi_jezikov_kitajscina[] = {
    { (T_KODAJEZIKA2)2052,  "-cn" },    // Chinese(PRC)
    { (T_KODAJEZIKA2)3076,  "-hk" },    // Chinese(Hong Kong)
    { (T_KODAJEZIKA2)4100,  "-sg" },    // Chinese(Singapore)
    { (T_KODAJEZIKA2)1028,  "-tw" },    // Chinese(Taiwan)
};


const T_NAZIV_JEZIKA nazivi_jezikov[] = {
    { (T_KODAJEZIKA2)1078,  "af" }, // Afrikaans
    { (T_KODAJEZIKA2)1,     "ar",   // Arabic
        _countof(nazivi_jezikov_arabscina), nazivi_jezikov_arabscina },
    { (T_KODAJEZIKA2)1059,  "be" }, // Belarusian
    { (T_KODAJEZIKA2)1026,  "bg" }, // Bulgarian
    { (T_KODAJEZIKA2)1027,  "ca" }, // Catalan
    { (T_KODAJEZIKA2)1029,  "cs" }, // Czech
    { (T_KODAJEZIKA2)1030,  "da" }, // Danish
    { (T_KODAJEZIKA2)1031,  "de",   // German(Standard)
        _countof(nazivi_jezikov_nemscina), nazivi_jezikov_nemscina },
    { (T_KODAJEZIKA2)1032,  "el" }, // Greek
    { (T_KODAJEZIKA2)9,     "en",   // English
        _countof(nazivi_jezikov_anglescina), nazivi_jezikov_anglescina },
    { (T_KODAJEZIKA2)1034,  "es",   // Spanish(Spain - Traditional Sort)
        _countof(nazivi_jezikov_spanscina), nazivi_jezikov_spanscina },
    { (T_KODAJEZIKA2)1061,  "et" }, // Estonian
    { (T_KODAJEZIKA2)1069,  "eu" }, // Basque
    { (T_KODAJEZIKA2)1065,  "fa" }, // Farsi
    { (T_KODAJEZIKA2)1035,  "fi" }, // Finnish
    { (T_KODAJEZIKA2)1080,  "fo" }, // Faeroese
    { (T_KODAJEZIKA2)1036,  "fr",   // French(Standard)
        _countof(nazivi_jezikov_francoscina), nazivi_jezikov_francoscina },
    { (T_KODAJEZIKA2)1084,  "gd" }, // Gaelic(Scots)
    { (T_KODAJEZIKA2)1037,  "he" }, // Hebrew
    { (T_KODAJEZIKA2)1081,  "hi" }, // Hindi
    { (T_KODAJEZIKA2)1050,  "hr" }, // Croatian
    { (T_KODAJEZIKA2)1038,  "hu" }, // Hungarian
    { (T_KODAJEZIKA2)1057,  "in" }, // Indonesian
    { (T_KODAJEZIKA2)1039,  "is" }, // Icelandic
    { (T_KODAJEZIKA2)1040,  "it",   // Italian(Standard)
        _countof(nazivi_jezikov_italijanscina), nazivi_jezikov_italijanscina },
    { (T_KODAJEZIKA2)1041,  "ja" }, // Japanese
    { (T_KODAJEZIKA2)1085,  "ji" }, // Yiddish
    { (T_KODAJEZIKA2)1042,  "ko",   // Korean
        _countof(nazivi_jezikov_korejscina), nazivi_jezikov_korejscina },
    { (T_KODAJEZIKA2)1063,  "lt" }, // Lithuanian
    { (T_KODAJEZIKA2)1062,  "lv" }, // Latvian
    { (T_KODAJEZIKA2)1071,  "mk" }, // Macedonian (FYROM)
    { (T_KODAJEZIKA2)1086,  "ms" }, // Malaysian
    { (T_KODAJEZIKA2)1082,  "mt" }, // Maltese
    { (T_KODAJEZIKA2)1043,  "nl",   // Dutch(Standard)
        _countof(nazivi_jezikov_nizozemscina), nazivi_jezikov_nizozemscina },
    { (T_KODAJEZIKA2)1044,  "no",    // Norwegian(Bokmaal)
        _countof(nazivi_jezikov_norvescina), nazivi_jezikov_norvescina },
    { (T_KODAJEZIKA2)1045,  "pl" }, // Polish
    { (T_KODAJEZIKA2)2070,  "pt",   // Portuguese(Portugal)
        _countof(nazivi_jezikov_portugalscina), nazivi_jezikov_portugalscina },
    { (T_KODAJEZIKA2)1047,  "rm" }, // Rhaeto-Romanic
    { (T_KODAJEZIKA2)1048,  "ro",   // Romanian
        _countof(nazivi_jezikov_romunscina), nazivi_jezikov_romunscina },
    { (T_KODAJEZIKA2)1049,  "ru",   // Russian
        _countof(nazivi_jezikov_ruscina), nazivi_jezikov_ruscina },
    { (T_KODAJEZIKA2)1070,  "sb" }, // Sorbian
    { (T_KODAJEZIKA2)1051,  "sk" }, // Slovak
    { (T_KODAJEZIKA2)1060,  "sl" }, // Slovenian
    { (T_KODAJEZIKA2)1052,  "sq" }, // Albanian
    { (T_KODAJEZIKA2)2074,  "sr",   // Serbian(Latin)
        _countof(nazivi_jezikov_srbscina), nazivi_jezikov_srbscina},
    { (T_KODAJEZIKA2)1053,  "sv",   // Swedish
        _countof(nazivi_jezikov_svedscina), nazivi_jezikov_svedscina },
    { (T_KODAJEZIKA2)1072,  "sx" }, // Sutu
    { (T_KODAJEZIKA2)1083,  "sz" }, // Sami(Lappish)
    { (T_KODAJEZIKA2)1054,  "th" }, // Thai
    { (T_KODAJEZIKA2)1074,  "tn" }, // Tswana
    { (T_KODAJEZIKA2)1055,  "tr" }, // Turkish
    { (T_KODAJEZIKA2)1073,  "ts" }, // Tsonga
    { (T_KODAJEZIKA2)1058,  "uk" }, // Ukrainian
    { (T_KODAJEZIKA2)1056,  "ur" }, // Urdu
    { (T_KODAJEZIKA2)1075,  "ve" }, // Venda
    { (T_KODAJEZIKA2)1066,  "vi" }, // Vietnamese
    { (T_KODAJEZIKA2)1076,  "xh" }, // Xhosa
    { (T_KODAJEZIKA2)4,     "zh",   // Chinese
        _countof(nazivi_jezikov_kitajscina), nazivi_jezikov_kitajscina },
    { (T_KODAJEZIKA2)1077,  "zu" }, // Zulu
};


//////////////////////////////////////////////////////////////////////
// Interne funkcije
//////////////////////////////////////////////////////////////////////

T_KODAJEZIKA2 _VrniKodoJezika(const T_NAZIV_JEZIKA *tabela, T_UINT st_elementov, TK_ASCII naziv_rfc1766, T_UINT dolzina)
{
    assert(tabela || !st_elementov);
    assert(naziv_rfc1766 || !dolzina);

    register T_UINT l, d, s, n;
    register T_INT r;

    for (l = 0, d = st_elementov; l < d; ) {
        s = (l + d) / 2;
        n = astrlen(tabela[s].naziv);
        r = astrnicmp(tabela[s].naziv, n, naziv_rfc1766, amin(n, dolzina));
        if (r < 0) {
            // Iskani element bo desno, e bo.
            l = s + 1;
        } else if (r > 0) {
            // Iskani element bo levo, e bo.
            d = s;
        } else {
            // Nali smo element.
            if (tabela[s].podjeziki) {
                // Preverimo e podjezik.
                T_KODAJEZIKA2 koda_jezika = _VrniKodoJezika(tabela[s].podjeziki, tabela[s].st_podjezikov, naziv_rfc1766 + n, dolzina - n);
                return koda_jezika != KODAJEZIKA2_NEZNANA ? koda_jezika : tabela[s].koda_jezika;
            } else
                return tabela[s].koda_jezika;
        }
    }

    return KODAJEZIKA2_NEZNANA;
}


#ifdef AOSN_IZJEMA

LONG WINAPI _RokovalnikNeujetihIzjem(PEXCEPTION_POINTERS ep)
{
    assert(ep);

    LONG rezultat = EXCEPTION_CONTINUE_SEARCH;

#if defined(_DEBUG) && defined(AOSN_HROSCAR_ASSERT)
    if (AOsn::JeProcesInteraktiven()) {
        // Prilo je do izjeme. Pripni razhroevalnik!
        assert(0);
    }
#endif

    adebug(ADNE_DIAG, ADNE_AOSN _T("Na naslovu %p je prilo do izjeme %x.\n"), ep->ExceptionRecord->ExceptionAddress, ep->ExceptionRecord->ExceptionCode);

    AOsn::kljucavnica_izjem.Zakleni();
    __try {
        if (AOsn::izjeme_nit) {
            if (USPESNO(AOsn::izjeme_nit->Pocakaj(AHROSCAR_KONCANJE_NITI))) {
                // Izjemo obdelamo s svojim rokovalnikom izjem.
                verify(USPESNO(AOsn::izjeme_nit->rokovalnik_izjem.Odpri(ep)));
            }
        }

        if (rezultat == EXCEPTION_CONTINUE_SEARCH && AOsn::rokovalnik_izjem_prej) {
            // Kliemo e prejnji rokovalnik izjem, da si e on ogleda izjemo.
            // Vendar smo previdni, ker se lahko sesuje tudi ta (primer wwlib.dll).
            __try {
                rezultat = AOsn::rokovalnik_izjem_prej(ep);
            } __except(EXCEPTION_EXECUTE_HANDLER) {}
        }
    } __finally {
        AOsn::kljucavnica_izjem.Odkleni();
    }

    return rezultat;
}

#endif // AOSN_IZJEMA


//////////////////////////////////////////////////////////////////////
// Funkcije AOsn
//////////////////////////////////////////////////////////////////////

T_STANJE AOSN_API AOsn::VrniVerzijo(T_UINT *verzija, T_UINT *maj, T_UINT *min, T_UINT *rev, T_UINT *build, TK_ASCII *niz)
{
    if (verzija) *verzija = AOSN_VERZIJA;
    if (maj)     *maj     = AOSN_VERZIJA_MAJ;
    if (min)     *min     = AOSN_VERZIJA_MIN;
    if (rev)     *rev     = AOSN_VERZIJA_REV;
#ifdef BUILD_NUM
    if (build)   *build   = BUILD_NUM;
#else
    if (build)   *build   = 0;
#endif
    if (niz)     *niz     = AOSN_VERZIJA_STR;

    return A_OK;
}


T_STANJE AOSN_API AOsn::Vzpostavi()
{
    T_STANJE stanje = A_OK;

    verify(::WaitForSingleObject(kljucavnica, INFINITE) == WAIT_OBJECT_0);

    if (!(st_vzpostavitev++)) {
        adebug(ADNE_DIAG, ADNE_AOSN _T("Vzpostavljam modul AOsn ...\n"));

        // Pripravimo pogosto uporabljene locale.
#if _MSC_VER >= 1400
        AOsn::locale_C   = _create_locale(LC_ALL, "C");
        AOsn::locale_ACP = _create_locale(LC_ALL, AOSN_PRIVZETI_LOCALE);
        AOsn::locale_OCP = _create_locale(LC_ALL, ".OCP");
#else
        static const struct _T_LOCALE locale_C   = { LC_ALL, "C" };
        static const struct _T_LOCALE locale_ACP = { LC_ALL, AOSN_PRIVZETI_LOCALE };
        static const struct _T_LOCALE locale_OCP = { LC_ALL, ".OCP" };

        AOsn::locale_C   = (T_LOCALE)&locale_C;
        AOsn::locale_ACP = (T_LOCALE)&locale_ACP;
        AOsn::locale_OCP = (T_LOCALE)&locale_OCP;
#endif

#ifdef AOSN_VTIC
        {
            WSAData wsa_data;
            int rezultat;

            // Vzpostavimo Windows Sockets 1.1.
            adebug(ADNE_DIAG, ADNE_AOSN _T("Vzpostavljam vtinice ...\n"));
            rezultat = ::WSAStartup(MAKEWORD(1, 1), &wsa_data);
            if (rezultat != 0) {
                adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri vzpostavljanju vtinic (koda MS %x).\n"), rezultat);
                stanje =  AOSN_NAPAKA_WSASTARTUP;
                goto napaka1;
            }
        }
#endif

#if defined(_WIN32)
        {
            OSVERSIONINFO verzija;

            // Preberemo verzijo Windowsov.
            verzija.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
            adebug(ADNE_DIAG, ADNE_AOSN _T("Doloam verzijo Windows ...\n"));
            if (!::GetVersionEx(&verzija)) {
                DWORD rezultat = ::GetLastError();
                adebug(ADNE_NAPAKA, ADNE_AOSN _T("Pri doloanju verzije Windows je prilo do napake (koda MS %x).\n"), rezultat);
                stanje = astanje(rezultat, AOSN_NAPAKA_GETVERSIONEX);
                goto napaka2;
            }

            // Doloimo platformo in verzijo.
            switch (verzija.dwPlatformId) {
            case VER_PLATFORM_WIN32s:
                verzija_os = AOSN_OS_WIN31;
                break;
            case VER_PLATFORM_WIN32_WINDOWS:
                if (verzija.dwMinorVersion < 10)
                    verzija_os = AOSN_OS_WIN95;
                else if (verzija.dwMinorVersion < 90)
                    verzija_os = AOSN_OS_WIN98;
                else
                    verzija_os = AOSN_OS_WINME;
                break;
            case VER_PLATFORM_WIN32_NT:
                if (verzija.dwMajorVersion < 4)
                    verzija_os = AOSN_OS_WINNT3;
                else if (verzija.dwMajorVersion < 5)
                    verzija_os = AOSN_OS_WINNT4;
                else if (verzija.dwMinorVersion < 1)
                    verzija_os = AOSN_OS_WIN2000;
                else if (verzija.dwMinorVersion < 2)
                    verzija_os = AOSN_OS_WINXP;
                else
                    verzija_os = AOSN_OS_WIN2003;
                break;
#ifdef _WIN32_WCE
            case VER_PLATFORM_WIN32_CE:
                if (verzija.dwMajorVersion < 2)
                    verzija_os = AOSN_OS_WINCE1;
                else if (verzija.dwMajorVersion < 3)
                    verzija_os = AOSN_OS_WINCE2;
                else
                    verzija_os = AOSN_OS_WINCE3;
#endif
            default:
                // e ne prepoznamo platforme, prevzamemo najnovejo mono verzijo OS.
                verzija_os = AOSN_OS_WIN2003;
            }
        }
#elif defined(__GNUC__)
        verzija_os = AOSN_OS_UNIX;
#else
#error Neznana platforma!
#endif

        {
            // Naloimo knjinico KERNEL32.DLL. Pomembno je, da naloimo tisto pravo, da nas kdo ne pretenta z lano funkcijo GetSystemTimeAsFileTime().
            T_ZNAK pot_kernel32[2*MAX_PATH];
            T_UINT znakov;

            adebug(ADNE_DIAG, ADNE_AOSN _T("Nalagam knjinjico KERNEL32 ...\n"));

            // Ugotovimo pot do C:\WINDOWS\system32.
            znakov = ::GetSystemDirectory(pot_kernel32, 2*MAX_PATH);
            if (znakov && znakov < 2*MAX_PATH)
                pot_kernel32[znakov++] = _T('\\');
            else
                znakov = 0;

            // Dodamo "KERNEL32.DLL".
            verify(astrncpy2(pot_kernel32 + znakov, _T("KERNEL32.DLL"), 2*MAX_PATH - znakov) < 2*MAX_PATH - znakov);
            ::rocica_kernel32 = ::LoadLibraryEx(pot_kernel32, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
            assert(::rocica_kernel32);

            // Poiemo naslove funkcij.
#ifndef _WIN64
            AOsn::IsWow64Process          = (BOOL(WINAPI*)(__in HANDLE, __out PBOOL))::GetProcAddress(::rocica_kernel32, "IsWow64Process");
#endif
#ifdef RDATUM_NALOZI_JEDRO
            AOsn::GetSystemTimeAsFileTime = (VOID(WINAPI*)(__out LPFILETIME))::GetProcAddress(::rocica_kernel32, "GetSystemTimeAsFileTime");
#endif
        }

#ifndef _WIN64
        // Doloimo, ali na 32-bitni proces tee znotraj 64-bitnega okolja.
        if (AOsn::IsWow64Process) {
            BOOL rezultat;

            // Preverimo, kaj pravi funkcija IsWow64Process().
            if (AOsn::IsWow64Process(::GetCurrentProcess(), &rezultat)) {
                // Nastavimo rezultat.
                je_proces_wow64 = rezultat ? TRUE : FALSE;
            } else {
                // Funkcija IsWow64Process()je vrnila napako, zato privzamemo, da proces ni WOW64.
                je_proces_wow64 = FALSE;
            }
        } else {
            // Ta platforma sploh ne pozna funkcije IsWow64Process(), zato proces zagotovo ni WOW64.
            je_proces_wow64 = FALSE;
        }
#endif

        {
            // Naloimo knjinico USER32.DLL.
            adebug(ADNE_DIAG, ADNE_AOSN _T("Nalagam knjinjico USER32 ...\n"));
            ::rocica_user32 = ::LoadLibrary(_T("USER32.DLL"));
            assert(::rocica_user32);

            // Poiemo naslove funkcij.
            AOsn::GetProcessWindowStation  = (HWINSTA(WINAPI*)())::GetProcAddress(::rocica_user32, "GetProcessWindowStation");
            AOsn::GetUserObjectInformation = (BOOL   (WINAPI*)(__in HANDLE, __in int, __out_opt PVOID, __in DWORD, __out_opt LPDWORD))::GetProcAddress(::rocica_user32,
#ifdef _UNICODE
                    "GetUserObjectInformationW"
#else
                    "GetUserObjectInformationA"
#endif
                );
        }

#ifdef AOSN_IZJEMA
        {
            MEMORY_BASIC_INFORMATION mbi;
            RCakalnik cakalnik(kljucavnica_izjem);

            adebug(ADNE_DIAG, ADNE_AOSN _T("Zaganjam vzpostavitev prestrezanja in poiljanja izjem ...\n"));

            izjeme_nit = new RIzjemeGlobalnaNit();
            if (izjeme_nit) {
                // Poenemo nit, ki v ozadju vzpostavi rokovanje izjem in preveri, ali aka kakna datoteka s poroilom o napaki na poiljanje in jo polje streniku.
                stanje = izjeme_nit->Odpri();
            }

            // Ugotovimo roico svojega modula tako, da pogledamo podatke pomnilnika, kjer se nahaja ta funkcija.
            if (VirtualQuery(AOsn::Vzpostavi, &mbi, sizeof(MEMORY_BASIC_INFORMATION)) >= 2*sizeof(PVOID) &&
                (HMODULE)(mbi.AllocationBase) == ::GetModuleHandle(NULL))
            {
                // Nismo DLL. Lahko si priredimo nastavitve procesa in knjinice CRT po svoje.

                // Registriramo rokovalnik izjem.
                rokovalnik_izjem_prej = ::SetUnhandledExceptionFilter(_RokovalnikNeujetihIzjem);
            }
        }
#endif

        {
            DWORD  nacin_konzole;

            adebug(ADNE_DIAG, ADNE_AOSN _T("Doloam konzolne roice ...\n"));

            rocica_std[AOSN_STD_VHOD]   = ::GetStdHandle(STD_INPUT_HANDLE );
            rocica_std[AOSN_STD_IZHOD]  = ::GetStdHandle(STD_OUTPUT_HANDLE);
            rocica_std[AOSN_STD_NAPAKA] = ::GetStdHandle(STD_ERROR_HANDLE );
            je_konzola[AOSN_STD_IZHOD]  = (::GetFileType(rocica_std[AOSN_STD_IZHOD] ) & FILE_TYPE_CHAR) && ::GetConsoleMode(rocica_std[AOSN_STD_IZHOD],  &nacin_konzole);
            je_konzola[AOSN_STD_NAPAKA] = (::GetFileType(rocica_std[AOSN_STD_NAPAKA]) & FILE_TYPE_CHAR) && ::GetConsoleMode(rocica_std[AOSN_STD_NAPAKA], &nacin_konzole);
        }

        adebug(ADNE_DIAG, ADNE_AOSN _T("Modul AOsn vzpostavljen.\n"));
    }

    stanje = A_OK;
    goto konec;

napaka2:
#ifdef AOSN_VTIC
    verify(!::WSACleanup());
napaka1:
#endif
konec:
    verify(::ReleaseMutex(kljucavnica));
    return stanje;
}


T_STANJE AOSN_API AOsn::PocakajVzpostavitev(T_UINT naj_cas)
{
#ifdef AOSN_IZJEMA
    RCakalnik cakalnik(kljucavnica_izjem);
    return izjeme_nit ? izjeme_nit->Pocakaj(naj_cas) : AOSN_OPOZORILO_NI_VZPOSTAVLJEN;
#else
    UNREFERENCED_PARAMETER(naj_cas);
    return A_OK;
#endif
}


T_STANJE AOSN_API AOsn::Zakljuci()
{
    T_STANJE stanje;

    verify(::WaitForSingleObject(kljucavnica, INFINITE) == WAIT_OBJECT_0);

    if (!st_vzpostavitev) {
        stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
        goto napaka1;
    }

    if (!(--st_vzpostavitev)) {
        adebug(ADNE_DIAG, ADNE_AOSN _T("Zakljuujem modul AOsn ...\n"));

#ifdef AOSN_IZJEMA
        verify(USPESNO(ZakljuciRokovalnikIzjem()));
#endif

        if (::rocica_user32) {
            adebug(ADNE_DIAG, ADNE_AOSN _T("Sproam knjinjico USER32 ...\n"));

            AOsn::GetProcessWindowStation  = NULL;
            AOsn::GetUserObjectInformation = NULL;
            ::FreeLibrary(::rocica_user32);
            ::rocica_user32 = NULL;
        }

        if (::rocica_kernel32) {
            adebug(ADNE_DIAG, ADNE_AOSN _T("Sproam knjinjico KERNEL32 ...\n"));

#ifndef _WIN64
            AOsn::IsWow64Process          = NULL;
#endif
#ifdef RDATUM_NALOZI_JEDRO
            AOsn::GetSystemTimeAsFileTime = NULL;
#endif
            ::FreeLibrary(::rocica_kernel32);
            ::rocica_kernel32 = NULL;
        }

#ifdef AOSN_VTIC
        adebug(ADNE_DIAG, ADNE_AOSN _T("Zakljuujem vtinice ...\n"));
        verify(!::WSACleanup());
#endif

#if _MSC_VER >= 1400
        _free_locale(AOsn::locale_OCP);
        _free_locale(AOsn::locale_ACP);
        _free_locale(AOsn::locale_C);
#endif
        AOsn::locale_OCP = NULL;
        AOsn::locale_ACP = NULL;
        AOsn::locale_C   = NULL;

        adebug(ADNE_DIAG, ADNE_AOSN _T("Modul AOsn zakljuen.\n"));
    }

    stanje = A_OK;

napaka1:
    verify(::ReleaseMutex(kljucavnica));
    return stanje;
}


#ifdef AOSN_IZJEMA

T_STANJE AOSN_API AOsn::ZakljuciRokovalnikIzjem()
{
    RCakalnik cakalnik(kljucavnica_izjem);

    adebug(ADNE_DIAG, ADNE_AOSN _T("Zakljuujem prestrezanje in poiljanje izjem ...\n"));

    if (rokovalnik_izjem_prej) {
        // Nastavimo prejnji rokovalnik izjem.
        verify(::SetUnhandledExceptionFilter(rokovalnik_izjem_prej));
        rokovalnik_izjem_prej = NULL;
    }

    if (izjeme_nit) {
        // Uniimo nit za vzpostavitev prestrezanja in poiljanja izjem.
        if (NEUSPESNO(izjeme_nit->Pocakaj(0))) {
            adebug(ADNE_DIAG, ADNE_AOSN _T("Vzpostavitev prestrezanja in poiljanja izjem se e ni zakljuila! Prekinjam nit ...\n"));
            assert(0);
            izjeme_nit->Prekini(1);
        }
        verify(USPESNO(izjeme_nit->Zapri()));

        // Uniimo oddajnik.
        delete izjeme_nit;
        izjeme_nit = NULL;
    }

    return A_OK;
}


T_STANJE AOSN_API AOsn::UstvariDodatniOpisZaIzjemo(TK_ASCII niz, T_UINT znakov, T_KODNASTRAN kodna_stran)
{
    RCakalnik cakalnik(kljucavnica_izjem);
    return izjeme_nit ? izjeme_nit->rokovalnik_izjem.UstvariDodatniOpis(niz, znakov, kodna_stran) : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


T_STANJE AOSN_API AOsn::UstvariDodatniOpisZaIzjemo(TK_UNICODE niz, T_UINT znakov)
{
    RCakalnik cakalnik(kljucavnica_izjem);
    return izjeme_nit ? izjeme_nit->rokovalnik_izjem.UstvariDodatniOpis(niz, znakov) : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


T_STANJE AOSN_API AOsn::UniciDodatniOpisZaIzjemo()
{
    RCakalnik cakalnik(kljucavnica_izjem);
    return izjeme_nit ? izjeme_nit->rokovalnik_izjem.UniciDodatniOpis() : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


LONG AOsn::RokovalnikIzjem(PEXCEPTION_POINTERS ep)
{
    assert(ep);

    LONG rezultat = EXCEPTION_CONTINUE_SEARCH;

#if defined(_DEBUG) && defined(AOSN_HROSCAR_ASSERT)
    if (AOsn::JeProcesInteraktiven()) {
        // Prilo je do izjeme. Pripni razhroevalnik!
        assert(0);
    }
#endif

    kljucavnica_izjem.Zakleni();
    __try {
        if (izjeme_nit) {
            if (USPESNO(izjeme_nit->Pocakaj(AHROSCAR_KONCANJE_NITI))) {
                // Izjemo obdelamo s svojim rokovalnikom izjem.
                verify(USPESNO(izjeme_nit->rokovalnik_izjem.Odpri(ep)));
            }
        }
    } __finally {
        kljucavnica_izjem.Odkleni();
    }

    return rezultat;
}

#endif


T_VERZIJAOS AOSN_API AOsn::VrniVerzijoOS()
{
    T_VERZIJAOS verzija_os;

    verify(::WaitForSingleObject(kljucavnica, INFINITE) == WAIT_OBJECT_0);
    assert(st_vzpostavitev);
    verzija_os = AOsn::verzija_os;
    verify(::ReleaseMutex(kljucavnica));

    return verzija_os;
}


T_STANJE AOSN_API AOsn::UstvariImenik(TK_NIZ imenik, TK_NIZ predloga)
{
    assert(imenik);
    T_STANJE stanje = A_OK;
    T_NIZ niz, konec;

    niz = konec = astrdup<T_ZNAK>(imenik);
    if (astrncmp(konec, 2, _T("\\\\"), 2) == 0) {
        // Pot je podana v zapisu UNC (zane se z "\\"). Preskoimo do naslednjega "\".
        konec = astrchr(konec + 2, _T(LOCILO_IMENIKOV));
        if (konec) konec++;
    }

    while (konec && *konec) {
        // Poiemo loilo imenikov.
        konec = astrchr(konec, _T(LOCILO_IMENIKOV));
        // e smo nali loilo imenikov, na tem mestu zakljuimo niz.
        if (konec)
            *konec = 0;

#ifdef _WIN32_WCE
        assert(!predloga);
#else
        if (predloga)
            stanje = ::CreateDirectoryEx(predloga, niz, NULL) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_USTVARI_IMENIK);
        else
#endif
            stanje = ::CreateDirectory(niz, NULL) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_USTVARI_IMENIK);

        // e ustvarjanje ni uspelo, prekinemo.
        // Izjemi sta napaki AOSN_NAPAKA_ZE_OBSTAJA in AOSN_NAPAKA_NI_DOSTOPA, ki pomenita, da je podimenik e ustvarjen, ali pa ga nimamo pravice ustvariti.
        if (NEUSPESNO(stanje) && stanje != AOSN_NAPAKA_ZE_OBSTAJA && stanje != AOSN_NAPAKA_NI_DOSTOPA)
            break;

        // Restavriramo loilo imenikov.
        if (konec) {
            *konec = _T(LOCILO_IMENIKOV);
            konec++;
        }
    }

    afree(niz);
    return stanje;
}


#if (defined(_DEBUG) || defined(AOSN_DIAGNOSTIKA))

T_UINT AOSN_API AOsn::IzpisiDiagnosticnoSpP(TK_ASCII format, T_LOCALE locale, va_list parametri)
{
    T_ASCII buf = NULL;
    T_UINT znakov = 0;
    T_INT zapisano;

    // Oblikujemo niz.
    do {
        znakov += 1000;
        buf = (T_ASCII)aqrealloc(buf, znakov * sizeof(T_AZNAK), FALSE);
        assert(buf);

        zapisano = avsnprintf(buf, znakov, format, locale, parametri);
    } while (zapisano < 0 || (T_UINT)zapisano == znakov);

#ifdef _WIN32_WCE
    // WinCE ne pozna funkcije OutputDebugStringA(), zato niz pretvorimo v Unicode in kliemo OutputDebugStringU().
    T_UNICODE buf_u;
    verify(astrndup2(&buf_u, buf, zapisano + 1) == zapisano);
    OutputDebugStringW(buf_u);
    afree(buf_u);
#else
    OutputDebugStringA(buf);
#endif
    afree(buf);

    return zapisano;
}


T_UINT AOSN_API AOsn::IzpisiDiagnosticnoSpP(TK_UNICODE format, T_LOCALE locale, va_list parametri)
{
    T_UNICODE buf = NULL;
    T_UINT znakov = 0;
    T_INT zapisano;

    // Oblikujemo niz.
    do {
        znakov += 1000;
        buf = (T_UNICODE)aqrealloc(buf, znakov * sizeof(T_UZNAK), FALSE);
        assert(buf);

        zapisano = avsnprintf(buf, znakov, format, locale, parametri);
    } while (zapisano < 0 || (T_UINT)zapisano == znakov);

    OutputDebugStringW(buf);
    afree(buf);

    return zapisano;
}

#endif // _DEBUG || AOSN_DIAGNOSTIKA


T_UINT AOSN_API AOsn::Izpisi(T_STD_ROCICA izhod, TK_ASCII niz, T_UINT znakov)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(0 <= izhod && izhod < 3);
    assert(niz || !znakov);

    DWORD zapisano;
    znakov = astrnlen(niz, znakov);

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleA(rocica_std[izhod], niz, (DWORD)znakov, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        ::WriteFile(rocica_std[izhod], niz, (DWORD)znakov, &zapisano, 0);
    }

    return zapisano;
}


T_UINT AOSN_API AOsn::Izpisi(T_STD_ROCICA izhod, TK_UNICODE niz, T_UINT znakov)
{
    assert(0 <= izhod && izhod < 3);
    assert(niz || !znakov);

    DWORD zapisano;
    znakov = astrnlen(niz, znakov);

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleW(rocica_std[izhod], niz, (DWORD)znakov, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        UINT kodna_stran_konzole = ::GetConsoleOutputCP();
        int znakov_izh2 = ::WideCharToMultiByte(kodna_stran_konzole, 0, niz, (int)znakov, 0, 0, 0, 0);
        LPSTR buf2 = (LPSTR)aalloc((znakov_izh2 + 1) * sizeof(CHAR));

        ::WideCharToMultiByte(kodna_stran_konzole, 0, niz, (int)znakov, buf2, znakov_izh2, 0, 0);
        ::WriteFile(rocica_std[izhod], buf2, znakov_izh2, &zapisano, 0);
        afree(buf2);
    }

    return zapisano;
}


T_UINT AOSN_API AOsn::Izpisi(T_STD_ROCICA izhod, HMODULE modul, T_UINT koda, T_KODAJEZIKA2 jezik)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(0 <= izhod && izhod < 3);

    TK_UNICODE niz;
    T_UINT znakov = aloadstring(modul, koda, &niz, jezik);
    DWORD zapisano;

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleW(rocica_std[izhod], niz, (DWORD)znakov, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        UINT kodna_stran_konzole = ::GetConsoleOutputCP();
        int znakov_izh2 = ::WideCharToMultiByte(kodna_stran_konzole, 0, niz, (int)znakov, 0, 0, 0, 0);
        LPSTR buf2 = (LPSTR)aalloc(znakov_izh2 * sizeof(CHAR));

        ::WideCharToMultiByte(kodna_stran_konzole, 0, niz, (int)znakov, buf2, znakov_izh2, 0, 0);
        ::WriteFile(rocica_std[izhod], buf2, znakov_izh2, &zapisano, 0);
        afree(buf2);
    }

    return zapisano;
}


T_UINT AOSN_API AOsn::IzpisiOP(T_STD_ROCICA izhod, TK_ASCII format, T_LOCALE locale, va_list parametri)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(0 <= izhod && izhod < 3);
    assert(format);

    T_ASCII buf = NULL;
    T_UINT znakov = 0;
    T_INT znakov_izh;
    DWORD zapisano;

    // Oblikujemo niz.
    do {
        znakov += 1000;
        buf = (T_ASCII)aqrealloc(buf, znakov * sizeof(T_AZNAK), FALSE);
        assert(buf);

        znakov_izh = avsnprintf(buf, znakov, format, locale, parametri);
    } while (znakov_izh < 0 || (T_UINT)znakov_izh == znakov);

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleA(rocica_std[izhod], buf, (DWORD)znakov_izh, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        ::WriteFile(rocica_std[izhod], buf, (DWORD)znakov_izh, &zapisano, 0);
    }
    afree(buf);

    return zapisano;
}


T_UINT AOSN_API AOsn::IzpisiOP(T_STD_ROCICA izhod, TK_UNICODE format, T_LOCALE locale, va_list parametri)
{
    assert(0 <= izhod && izhod < 3);
    assert(format);

    T_UNICODE buf = NULL;
    T_UINT znakov = 0;
    T_INT znakov_izh;
    DWORD zapisano;

    // Oblikujemo niz.
    do {
        znakov += 1000;
        buf = (T_UNICODE)aqrealloc(buf, znakov * sizeof(T_UZNAK), FALSE);
        assert(buf);

        znakov_izh = avsnprintf(buf, znakov, format, locale, parametri);
    } while (znakov_izh < 0 || (T_UINT)znakov_izh == znakov);

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleW(rocica_std[izhod], buf, (DWORD)znakov_izh, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        UINT kodna_stran_konzole = ::GetConsoleOutputCP();
        int znakov_izh2 = ::WideCharToMultiByte(kodna_stran_konzole, 0, buf, -1, 0, 0, 0, 0);
        LPSTR buf2 = (LPSTR)aalloc(znakov_izh2 * sizeof(CHAR));

        ::WideCharToMultiByte(kodna_stran_konzole, 0, buf, -1, buf2, znakov_izh2, 0, 0);
        ::WriteFile(rocica_std[izhod], buf2, znakov_izh2 - 1, &zapisano, 0);
        afree(buf2);
    }
    afree(buf);

    return zapisano;
}


T_UINT AOSN_API AOsn::IzpisiOP(T_STD_ROCICA izhod, HMODULE modul, T_UINT koda, T_KODAJEZIKA2 jezik, T_LOCALE locale, va_list parametri)
{
    assert(0 <= izhod && izhod < 3);

    TK_UNICODE format;
    T_UINT znakov = 0;
    T_UNICODE buf = NULL;
    T_INT znakov_izh;
    DWORD zapisano;

    aloadstring(modul, koda, &format, jezik);

    // Oblikujemo niz.
    do {
        znakov += 1000;
        buf = (T_UNICODE)aqrealloc(buf, znakov * sizeof(T_UZNAK), FALSE);
        assert(buf);

        znakov_izh = avsnprintf(buf, znakov, format, locale, parametri);
    } while (znakov_izh < 0 || (T_UINT)znakov_izh == znakov);

    if (je_konzola[izhod]) {
        // Zapiemo v konzolo.
        ::WriteConsoleW(rocica_std[izhod], buf, (DWORD)znakov_izh, &zapisano, 0);
    } else {
        // Zapiemo v datoteko.
        UINT kodna_stran_konzole = ::GetConsoleOutputCP();
        int znakov_izh2 = ::WideCharToMultiByte(kodna_stran_konzole, 0, buf, -1, 0, 0, 0, 0);
        LPSTR buf2 = (LPSTR)aalloc(znakov_izh2 * sizeof(CHAR));

        ::WideCharToMultiByte(kodna_stran_konzole, 0, buf, -1, buf2, znakov_izh2, 0, 0);
        ::WriteFile(rocica_std[izhod], buf2, znakov_izh2 - 1, &zapisano, 0);
        afree(buf2);
    }
    afree(buf);

    return zapisano;
}


T_STANJE AOSN_API AOsn::VrniPotModula(T_ROCICAOS rocica, RNizP<T_AZNAK> &pot)
{
    DWORD dolzina = MAX_PATH;
    T_ASCII _pot = pot.Odpusti(dolzina - 1, FALSE);

    for (;;) {
        DWORD rezultat = ::GetModuleFileNameA((HMODULE)rocica, _pot, dolzina);
        if (rezultat == 0) {
            pot.Uporabi(_pot, 0, FALSE);
#ifndef AOSN_HITROST_POMEMBNEJSA_OD_POMNILNIKA_VER
            pot.SprostiOdvecne();
#endif
            return astanje(::GetLastError(), AOSN_NAPAKA_NEZNANA_POT_MODULA);
        } else if (rezultat < dolzina) {
            pot.Uporabi(_pot, rezultat, FALSE);
#ifndef AOSN_HITROST_POMEMBNEJSA_OD_POMNILNIKA_VER
            pot.SprostiOdvecne();
#endif
            return A_OK;
        } else {
            dolzina += RNIZ_KORAK_POVECEVANJA;
            _pot = (T_ASCII)aqrealloc(_pot, dolzina * sizeof(T_AZNAK), FALSE);
        }
    }
}


T_STANJE AOSN_API AOsn::VrniPotModula(T_ROCICAOS rocica, RNizP<T_UZNAK> &pot)
{
    DWORD dolzina = MAX_PATH;
    T_UNICODE _pot = pot.Odpusti(dolzina - 1, FALSE);

    for (;;) {
        DWORD rezultat = ::GetModuleFileNameW((HMODULE)rocica, _pot, dolzina);
        if (rezultat == 0) {
            pot.Uporabi(_pot, 0, FALSE);
#ifndef AOSN_HITROST_POMEMBNEJSA_OD_POMNILNIKA_VER
            pot.SprostiOdvecne();
#endif
            return astanje(::GetLastError(), AOSN_NAPAKA_NEZNANA_POT_MODULA);
        } else if (rezultat < dolzina) {
            pot.Uporabi(_pot, rezultat, FALSE);
#ifndef AOSN_HITROST_POMEMBNEJSA_OD_POMNILNIKA_VER
            pot.SprostiOdvecne();
#endif
            return A_OK;
        } else {
            dolzina += RNIZ_KORAK_POVECEVANJA;
            _pot = (T_UNICODE)aqrealloc(_pot, dolzina * sizeof(T_UZNAK), FALSE);
        }
    }
}


T_STANJE AOSN_API AOsn::VrniTrenutniImenik(RNizP<T_AZNAK> &pot)
{
    assert(0); // TODO: Preizkusi ta del kode!

    DWORD dolzina = ::GetCurrentDirectoryA(0, NULL);
    if (dolzina) {
        T_ASCII _pot = pot.Odpusti(dolzina - 1, FALSE);
        dolzina = ::GetCurrentDirectoryA(dolzina, _pot);
        pot.Uporabi(_pot, dolzina, FALSE);
        return A_OK;
    } else
        return astanje(::GetLastError(), AOSN_NAPAKA_NEZNAN_TRENUTNI_IMENIK);
}


T_STANJE AOSN_API AOsn::VrniTrenutniImenik(RNizP<T_UZNAK> &pot)
{
    assert(0); // TODO: Preizkusi ta del kode!

    DWORD dolzina = ::GetCurrentDirectoryW(0, NULL);
    if (dolzina) {
        T_UNICODE _pot = pot.Odpusti(dolzina - 1, FALSE);
        dolzina = ::GetCurrentDirectoryW(dolzina, _pot);
        pot.Uporabi(_pot, dolzina, FALSE);
        return A_OK;
    } else
        return astanje(::GetLastError(), AOSN_NAPAKA_NEZNAN_TRENUTNI_IMENIK);
}


T_STANJE AOSN_API AOsn::NastaviTrenutniImenik(TK_ASCII pot)
{
    return ::SetCurrentDirectoryA(pot) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_SPREMEMBA_TRENUTNEGA_IMENIKA);
}


T_STANJE AOSN_API AOsn::NastaviTrenutniImenik(TK_UNICODE pot)
{
    return ::SetCurrentDirectoryW(pot) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_SPREMEMBA_TRENUTNEGA_IMENIKA);
}


T_ROCICA AOSN_API AOsn::RezervirajPomnilnik(T_ROCICA stari, T_SIZE nova_dolzina)
{
    return arealloc(stari, nova_dolzina);
}


T_VOID AOSN_API AOsn::SprostiPomnilnik(T_ROCICA blok)
{
    assert(blok);
    afree(blok);
}


T_BOOL AOSN_API AOsn::DatotekaObstaja(TK_ASCII ime_datoteke, T_UINT znakov)
{
    assert(ime_datoteke || !znakov);
    if (znakov == T_UINT_MAX)
        return ::GetFileAttributesA(ime_datoteke) != INVALID_FILE_ATTRIBUTES;
    else {
        RNizA niz(ime_datoteke, znakov);
        return ::GetFileAttributesA(niz.Vrni()) != INVALID_FILE_ATTRIBUTES;
    }
}


T_BOOL AOSN_API AOsn::DatotekaObstaja(TK_UNICODE ime_datoteke, T_UINT znakov)
{
    assert(ime_datoteke || !znakov);
    if (znakov == T_UINT_MAX)
        return ::GetFileAttributesW(ime_datoteke) != INVALID_FILE_ATTRIBUTES;
    else {
        RNizU niz(ime_datoteke, znakov);
        return ::GetFileAttributesW(niz.Vrni()) != INVALID_FILE_ATTRIBUTES;
    }
}


T_UINT AOSN_API AOsn::UrediImeDatoteke(T_ASCII ime_datoteke, T_UINT znakov)
{
    T_UINT i_zac, i_kon, j_zac, j_kon;
    T_ASCII locilo_l, locilo_d;
    WIN32_FIND_DATA dat;
    HANDLE rocica;

    i_zac = j_zac = 0;
    znakov = astrnlen(ime_datoteke, znakov);

    while (i_zac < znakov) {
        locilo_d = astrnchr(ime_datoteke + i_zac, LOCILO_IMENIKOV, znakov - i_zac);
        i_kon = locilo_d ? locilo_d - ime_datoteke : znakov;

        if (i_kon - i_zac == 1 && ime_datoteke[i_zac] == '.') {
            // Nali smo ".". Preskoimo ga.
            i_zac = locilo_d ? i_kon + 1 : i_kon;
        } else if (i_kon - i_zac == 2 && ime_datoteke[i_zac] == '.' && ime_datoteke[i_zac + 1] == '.') {
            // Nali smo "..". Poiemo, e obstaja podimenik levo.
            locilo_l = astrrnchr(ime_datoteke, LOCILO_IMENIKOV, j_zac ? j_zac - 1 : 0);
            if (locilo_l) {
                // Odreemo prejnji podimenik.
                j_zac = locilo_l - ime_datoteke + 1;
            } else {
                // Prejnjega podimenika ni. Prepiemo ".." in po potrebi zakljuimo z "\".
                i_zac = i_kon;
                ime_datoteke[j_zac++] = '.';
                ime_datoteke[j_zac++] = '.';
                if (locilo_d)
                    ime_datoteke[j_zac++] = LOCILO_IMENIKOV;
            }
            i_zac = locilo_d ? i_kon + 1 : i_kon;
        } else if (i_kon - i_zac == 2 && ('A' <= ime_datoteke[i_zac] && ime_datoteke[i_zac] <= 'Z' || 'a' <= ime_datoteke[i_zac] && ime_datoteke[i_zac] <= 'z') && ime_datoteke[i_zac + 1] == L':') {
            // Nali smo rko pogona.
            ime_datoteke[j_zac++] = atoupper(ime_datoteke[i_zac]);
            ime_datoteke[j_zac++] = L':';
            if (locilo_d) {
                ime_datoteke[j_zac++] = LOCILO_IMENIKOV;
                i_zac = i_kon + 1;
            } else
                i_zac = i_kon;
        } else {
            // Prepiemo podimenik.
            astrncpy2(ime_datoteke + j_zac, znakov - j_zac, ime_datoteke + i_zac, i_kon - i_zac);
            j_kon = j_zac + i_kon - i_zac;

            if (DatotekaObstaja(ime_datoteke, j_kon)) {
                // Ta pot obstaja. Preverimo, kako se v resnici imenuje.
                ime_datoteke[j_kon] = 0;
                rocica = ::FindFirstFileExA(ime_datoteke, (FINDEX_INFO_LEVELS)1 /*FindExInfoBasic*/, &dat, FindExSearchNameMatch, NULL, 0);
                if (rocica != INVALID_HANDLE_VALUE) {
                    // Prepiemo ime datoteke.
                    astrncpy2(ime_datoteke + j_zac, j_kon - j_zac, dat.cFileName, MAX_PATH);
                    ::FindClose(rocica);
                }
            }

            if (locilo_d) {
                i_zac = i_kon + 1;
                ime_datoteke[j_kon] = LOCILO_IMENIKOV;
                j_zac = j_kon + 1;
            } else {
                i_zac = i_kon;
                j_zac = j_kon;
            }
        }
    }

    if (j_zac < znakov) ime_datoteke[j_zac] = 0;

    return j_zac;
}


T_UINT AOSN_API AOsn::UrediImeDatoteke(T_UNICODE ime_datoteke, T_UINT znakov)
{
    T_UINT i_zac, i_kon, j_zac, j_kon;
    T_UNICODE locilo_l, locilo_d;
    WIN32_FIND_DATA dat;
    HANDLE rocica;

    i_zac = j_zac = 0;
    znakov = astrnlen(ime_datoteke, znakov);

    while (i_zac < znakov) {
        locilo_d = astrnchr(ime_datoteke + i_zac, _L(LOCILO_IMENIKOV), znakov - i_zac);
        i_kon = locilo_d ? locilo_d - ime_datoteke : znakov;

        if (i_kon - i_zac == 1 && ime_datoteke[i_zac] == L'.') {
            // Nali smo ".". Preskoimo ga.
            i_zac = locilo_d ? i_kon + 1 : i_kon;
        } else if (i_kon - i_zac == 2 && ime_datoteke[i_zac] == L'.' && ime_datoteke[i_zac + 1] == L'.') {
            // Nali smo "..". Poiemo, e obstaja podimenik levo.
            locilo_l = astrrnchr(ime_datoteke, _L(LOCILO_IMENIKOV), j_zac ? j_zac - 1 : 0);
            if (locilo_l) {
                // Odreemo prejnji podimenik.
                j_zac = locilo_l - ime_datoteke + 1;
            } else {
                // Prejnjega podimenika ni. Prepiemo ".." in po potrebi zakljuimo z "\".
                i_zac = i_kon;
                ime_datoteke[j_zac++] = L'.';
                ime_datoteke[j_zac++] = L'.';
                if (locilo_d)
                    ime_datoteke[j_zac++] = _L(LOCILO_IMENIKOV);
            }
            i_zac = locilo_d ? i_kon + 1 : i_kon;
        } else if (i_kon - i_zac == 2 && (L'A' <= ime_datoteke[i_zac] && ime_datoteke[i_zac] <= L'Z' || L'a' <= ime_datoteke[i_zac] && ime_datoteke[i_zac] <= L'z') && ime_datoteke[i_zac + 1] == L':') {
            // Nali smo rko pogona.
            ime_datoteke[j_zac++] = atoupper(ime_datoteke[i_zac]);
            ime_datoteke[j_zac++] = L':';
            if (locilo_d) {
                ime_datoteke[j_zac++] = _L(LOCILO_IMENIKOV);
                i_zac = i_kon + 1;
            } else
                i_zac = i_kon;
        } else {
            // Prepiemo podimenik.
            astrncpy2(ime_datoteke + j_zac, znakov - j_zac, ime_datoteke + i_zac, i_kon - i_zac);
            j_kon = j_zac + i_kon - i_zac;

            if (DatotekaObstaja(ime_datoteke, j_kon)) {
                // Ta pot obstaja. Preverimo, kako se v resnici imenuje.
                ime_datoteke[j_kon] = 0;
                rocica = ::FindFirstFileExW(ime_datoteke, (FINDEX_INFO_LEVELS)1 /*FindExInfoBasic*/, &dat, FindExSearchNameMatch, NULL, 0);
                if (rocica != INVALID_HANDLE_VALUE) {
                    // Prepiemo ime datoteke.
                    astrncpy2(ime_datoteke + j_zac, j_kon - j_zac, dat.cFileName, MAX_PATH);
                    ::FindClose(rocica);
                }
            }

            if (locilo_d) {
                i_zac = i_kon + 1;
                ime_datoteke[j_kon] = _L(LOCILO_IMENIKOV);
                j_zac = j_kon + 1;
            } else {
                i_zac = i_kon;
                j_zac = j_kon;
            }
        }
    }

    if (j_zac < znakov) ime_datoteke[j_zac] = 0;

    return j_zac;
}


T_STANJE AOSN_API AOsn::ZbrisiDatoteko(TK_ASCII ime_datoteke, T_UINT znakov)
{
    assert(ime_datoteke || !znakov);
    if (znakov == T_UINT_MAX)
        return ::DeleteFileA(ime_datoteke) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_BRISANJE);
    else {
        RNizA niz(ime_datoteke, znakov);
        return ::DeleteFileA(niz.Vrni()) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_BRISANJE);
    }
}


T_STANJE AOSN_API AOsn::ZbrisiDatoteko(TK_UNICODE ime_datoteke, T_UINT znakov)
{
    assert(ime_datoteke || !znakov);
    if (znakov == T_UINT_MAX)
        return ::DeleteFileW(ime_datoteke) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_BRISANJE);
    else {
        RNizU niz(ime_datoteke, znakov);
        return ::DeleteFileW(niz.Vrni()) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_BRISANJE);
    }
}


T_STANJE AOSN_API AOsn::PreimenujDatoteko(TK_ASCII ime_datoteke, TK_ASCII novo_ime_datoteke)
{
    assert(ime_datoteke);
    assert(novo_ime_datoteke);
    return ::MoveFileA(ime_datoteke, novo_ime_datoteke) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREIMENOVANJE);
}


T_STANJE AOSN_API AOsn::PreimenujDatoteko(TK_UNICODE ime_datoteke, TK_UNICODE novo_ime_datoteke)
{
    assert(ime_datoteke);
    assert(novo_ime_datoteke);
    return ::MoveFileW(ime_datoteke, novo_ime_datoteke) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREIMENOVANJE);
}


T_STANJE AOSN_API AOsn::PrepisiDatoteko(TK_ASCII ime_datoteke, TK_ASCII novo_ime_datoteke)
{
    assert(ime_datoteke);
    assert(novo_ime_datoteke);
    return ::CopyFileA(ime_datoteke, novo_ime_datoteke, FALSE) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREPISOVANJE);
}


T_STANJE AOSN_API AOsn::PrepisiDatoteko(TK_UNICODE ime_datoteke, TK_UNICODE novo_ime_datoteke)
{
    assert(ime_datoteke);
    assert(novo_ime_datoteke);
    return ::CopyFileW(ime_datoteke, novo_ime_datoteke, FALSE) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREPISOVANJE);
}


T_STANJE AOSN_API AOsn::VrniVerzijoDatoteke(TK_ASCII ime_datoteke, T_UINT *maj, T_UINT *min, T_UINT *rev, T_UINT *build)
{
    T_STANJE stanje;
    UINT uiSize;
    DWORD dwVerInfoSize;
    DWORD dwHandle;
    BYTE *prgbVersionInfo;
    VS_FIXEDFILEINFO *lpVSFixedFileInfo;

    dwVerInfoSize = ::GetFileVersionInfoSizeA((LPSTR)ime_datoteke, &dwHandle);
    if (dwVerInfoSize == 0) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka1;
    }

    prgbVersionInfo = (LPBYTE)::GlobalAlloc(GPTR, dwVerInfoSize);
    if (NULL == prgbVersionInfo) {
        stanje = AOSN_NAPAKA_POMNILNIK;
        goto napaka1;
    }

    if (!::GetFileVersionInfoA((LPSTR)ime_datoteke, dwHandle, dwVerInfoSize, prgbVersionInfo)) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka2;
    }

    if (!::VerQueryValueA(prgbVersionInfo, "\\", (LPVOID*)&lpVSFixedFileInfo, &uiSize) && (uiSize != 0)) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka2;
    }

    if (maj) *maj = HIWORD(lpVSFixedFileInfo->dwFileVersionMS);
    if (min) *min = LOWORD(lpVSFixedFileInfo->dwFileVersionMS);
    if (rev) *rev = HIWORD(lpVSFixedFileInfo->dwFileVersionLS);
    if (build) *build = LOWORD(lpVSFixedFileInfo->dwFileVersionLS);
    stanje = A_OK;

napaka2:
    ::GlobalFree(prgbVersionInfo);
napaka1:
    return stanje;
}


T_STANJE AOSN_API AOsn::VrniVerzijoDatoteke(TK_UNICODE ime_datoteke, T_UINT *maj, T_UINT *min, T_UINT *rev, T_UINT *build)
{
    T_STANJE stanje;
    UINT uiSize;
    DWORD dwVerInfoSize;
    DWORD dwHandle;
    BYTE *prgbVersionInfo;
    VS_FIXEDFILEINFO *lpVSFixedFileInfo;

    dwVerInfoSize = ::GetFileVersionInfoSizeW((LPWSTR)ime_datoteke, &dwHandle);
    if (dwVerInfoSize == 0) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka1;
    }

    prgbVersionInfo = (LPBYTE)::GlobalAlloc(GPTR, dwVerInfoSize);
    if (NULL == prgbVersionInfo) {
        stanje = AOSN_NAPAKA_POMNILNIK;
        goto napaka1;
    }

    if (!::GetFileVersionInfoW((LPWSTR)ime_datoteke, dwHandle, dwVerInfoSize, prgbVersionInfo)) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka2;
    }

    if (!::VerQueryValueW(prgbVersionInfo, L"\\", (LPVOID*)&lpVSFixedFileInfo, &uiSize) && (uiSize != 0)) {
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VERZIJA_DATOTEKE);
        goto napaka2;
    }

    if (maj) *maj = HIWORD(lpVSFixedFileInfo->dwFileVersionMS);
    if (min) *min = LOWORD(lpVSFixedFileInfo->dwFileVersionMS);
    if (rev) *rev = HIWORD(lpVSFixedFileInfo->dwFileVersionLS);
    if (build) *build = LOWORD(lpVSFixedFileInfo->dwFileVersionLS);
    stanje = A_OK;

napaka2:
    ::GlobalFree(prgbVersionInfo);
napaka1:
    return stanje;
}


TK_ASCII AOSN_API AOsn::VrniNazivKodneStrani(T_KODNASTRAN kodna_stran)
{
    T_UINT i;

    for (i = 0; i < _countof(nazivi_kodnih_strani); i++) {
        if (nazivi_kodnih_strani[i].kodna_stran == kodna_stran)
            return nazivi_kodnih_strani[i].naziv;
    }

    return NULL;
}


T_KODNASTRAN AOSN_API AOsn::VrniKodoKodneStrani(TK_ASCII naziv_kodne_strani, T_UINT dolzina)
{
    T_UINT i;

    for (i = 0; i < _countof(nazivi_kodnih_strani); i++) {
        if (!astrnicmp(nazivi_kodnih_strani[i].naziv, naziv_kodne_strani, dolzina))
            return nazivi_kodnih_strani[i].kodna_stran;
    }

    return KODNASTRAN_NEZNANA;
}


T_KODAJEZIKA2 AOSN_API AOsn::VrniKodoJezika(TK_ASCII naziv_rfc1766, T_UINT dolzina)
{
    return _VrniKodoJezika(::nazivi_jezikov, _countof(nazivi_jezikov), naziv_rfc1766, dolzina);
}


T_BOOL AOSN_API AOsn::JeProcesInteraktiven()
{
    HWINSTA rocica;
    DWORD dolzina, dolzina2;
    T_BOOL je_interaktiven;
    RNiz ime;
    T_NIZ niz;

    verify(::WaitForSingleObject(kljucavnica, INFINITE) == WAIT_OBJECT_0);
    assert(st_vzpostavitev);

    if (!AOsn::GetProcessWindowStation || !AOsn::GetUserObjectInformation) {
        // Operacijski sistem ne podpira funkcij GetProcessWindowStation() in GetUserObjectInformation(), zato sklepamo, da gre za stareje Windowse in so vsi procesi interaktivni.
        je_interaktiven = TRUE;
        goto konec;
    }

    // Ugotovimo roico "window station".
    rocica = AOsn::GetProcessWindowStation();
    if (!rocica) {
        // e je pri ugotavljanju interaktivnosti prilo do napake, privzamemo, da je proces interaktiven.
        je_interaktiven = TRUE;
        goto konec;
    }

    // Najprej ugotovimo dolino imena "window station".
    AOsn::GetUserObjectInformation(rocica, UOI_NAME, NULL, 0, &dolzina);
    dolzina /= sizeof(T_ZNAK);

    niz = ime.Odpusti(dolzina, FALSE);

    // Preberemo ime "window station".
    if (AOsn::GetUserObjectInformation(rocica, UOI_NAME, niz, dolzina * sizeof(T_ZNAK), &dolzina2)) {
        ime.Uporabi(niz, dolzina2 / sizeof(T_ZNAK), FALSE);

        // Samo "window station" z imenom "WinSta0" je interaktiven (vir KB171890).
        je_interaktiven = ime.PrimerjajI(_T("WinSta0")) ? FALSE : TRUE;
    } else {
        // Pri ugotavljanju imena je prilo do napake. Obupamo in javimo, da je proces interaktiven.
        ime.Uporabi(niz, 0, FALSE);
        je_interaktiven = TRUE;
    }

konec:
    verify(::ReleaseMutex(kljucavnica));
    return je_interaktiven;
}


T_BOOL AOSN_API AOsn::JeUporabnikAdministrator()
{
    T_BOOL je_admin = FALSE;
    HANDLE zeton = NULL;

    // Ugotovimo dostopni eton procesa.
    if (::OpenProcessToken(::GetCurrentProcess(), TOKEN_QUERY, &zeton)) {
        PTOKEN_GROUPS skupine;
        DWORD dolzina;

        // Ugotovimo dolino zapisa TokenGroups.
        verify(!::GetTokenInformation(zeton, TokenGroups, NULL, 0, &dolzina));
        assert(::GetLastError() == ERROR_INSUFFICIENT_BUFFER);

        // Pripravimo pomnilnik za zapis TokenGroups.
        if ((skupine = (PTOKEN_GROUPS)aalloc(dolzina)) != NULL) {
            // Preberemo zapis TokenGroups.
            if (::GetTokenInformation(zeton, TokenGroups, (LPVOID)skupine, dolzina, &dolzina)) {
                SID_IDENTIFIER_AUTHORITY avtoriteta = SECURITY_NT_AUTHORITY;
                PSID sid_admins = NULL;

                // Doloimo SID skupine "Administrators".
                if (::AllocateAndInitializeSid(&avtoriteta, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, &sid_admins)) {
                    // Preverimo, ali je skupina "Administrators" med skupinami v TokenGroups.
                    for (DWORD i = 0; i < skupine->GroupCount; i++) {
                        if (::EqualSid(sid_admins, skupine->Groups[i].Sid)) {
                            je_admin = TRUE;
                            break;
                        }
                    }
                    ::FreeSid(sid_admins);
                }
            }
            afree(skupine);
        }
        ::CloseHandle(zeton);
    }

    return je_admin;
}


#ifndef _WIN64

T_BOOL AOSN_API AOsn::JeProcesWOW64()
{
    T_BOOL je_proces_wow64;

    verify(::WaitForSingleObject(kljucavnica, INFINITE) == WAIT_OBJECT_0);
    assert(st_vzpostavitev);
    je_proces_wow64 = AOsn::je_proces_wow64;
    verify(::ReleaseMutex(kljucavnica));

    return je_proces_wow64;
}

#endif
