/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_WAV


//////////////////////////////////////////////////////////////////////
// T_WAV_GLAVA
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_GLAVA_ID::id = *(const T_WAV_ID*)"RIFF";


//////////////////////////////////////////////////////////////////////
// T_WAV_WAVE
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_WAVE_ID::id = *(const T_WAV_ID*)"WAVE";


//////////////////////////////////////////////////////////////////////
// T_WAV_FORMAT
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_FORMAT_ID::id = *(const T_WAV_ID*)"fmt ";


//////////////////////////////////////////////////////////////////////
// T_WAV_PODATKI
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_PODATKI_ID::id = *(const T_WAV_ID*)"data";


//////////////////////////////////////////////////////////////////////
// T_WAV_TISINA
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_TISINA_ID::id = *(const T_WAV_ID*)"slnt";


//////////////////////////////////////////////////////////////////////
// T_WAV_TABELA_OZNAK
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_OZNAKE_ID::id = *(const T_WAV_ID*)"cue ";


//////////////////////////////////////////////////////////////////////
// T_WAV_LTXT
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_LTXT_ID::id = *(const T_WAV_ID*)"ltxt";


//////////////////////////////////////////////////////////////////////
// T_WAV_NASLOV
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_NASLOV_ID::id = *(const T_WAV_ID*)"labl";


//////////////////////////////////////////////////////////////////////
// T_WAV_OPOMBA
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_OPOMBA_ID::id = *(const T_WAV_ID*)"note";


//////////////////////////////////////////////////////////////////////
// T_WAV_SEZNAM
//////////////////////////////////////////////////////////////////////

const T_WAV_ID T_WAV_SEZNAM_ID::id = *(const T_WAV_ID*)"LIST";


//////////////////////////////////////////////////////////////////////
// Funkcije za delo z datotekami WAV
//////////////////////////////////////////////////////////////////////

T_STANJE AOSN_API AOsn::WavPreberiOznake(RDatoteka &dat, T_WAV_TABELA_OZNAK_EX &oznake, T_ODMIK konec_bloka)
{
    T_ODMIK zacetek;
    T_STANJE stanje;

    zacetek = dat.VrniOdmik();
    while (dat.VrniOdmik() < konec_bloka) {
        T_WAV_DOLZINA dolzina;
        T_U4 st_oznak;
        T_WAV_OZNAKA_EX *p;
        T_UINT i;
        T_BOOL so_urejene;

        stanje = T_IDPODATKI<T_WAV_TABELA_OZNAK, T_WAV_ID, T_WAV_DOLZINA, WAV_PORAVNAVA>::NajdiGlavo(dat, konec_bloka);
        if (USPESNO(stanje)) {
            // Preberemo dolino.
            dat >> dolzina;
            stanje = dat.VrniStanje();
            if (NEUSPESNO(stanje)) return stanje;

            ROmejenTokPodatkov _dat(&dat, dolzina);

            // Preberemo tevilo oznak.
            _dat >> st_oznak;
            stanje = _dat.VrniStanje();
            if (NEUSPESNO(stanje)) return stanje;

            p = oznake.oznake.Odpusti(st_oznak, FALSE);
            if (!p) {
                assert(0); // TODO: Preizkusi ta del kode.

                // Zmanjkalo je pomnilnika.
                oznake.oznake.Uporabi(p, 0, FALSE);
                return AOSN_NAPAKA_POMNILNIK;
            }

            // Preberemo oznake eno po eno.
            for (i = 0, so_urejene = TRUE; i < st_oznak; i++) {
                (p + i)->T_WAV_OZNAKA_EX::T_WAV_OZNAKA_EX();
                _dat.PreberiTab((T_WAV_OZNAKA*)(p + i), sizeof(T_WAV_OZNAKA), 1);
                if (i && p[i - 1].id > p[i].id)
                    so_urejene = FALSE;
                stanje = _dat.VrniStanje();
                if (NEUSPESNO(stanje)) {
                    (p + i)->T_WAV_OZNAKA_EX::~T_WAV_OZNAKA_EX();
                    break;
                }
            }
            st_oznak = (T_U4)i;
            oznake.oznake.Uporabi(p, st_oznak, FALSE);

            if (!so_urejene) {
                // Uredimo oznake po naraajoem ID-ju.
#if _MSC_VER >= 1300
                oznake.oznake.Uredi<WavPrimerjajOznakiExPoID>();
#else
                T_UINT n = oznake.oznake.VrniStevilo();
                T_WAV_OZNAKA_EX *tab = oznake.oznake.Odpusti(n, FALSE);

                aqsort<T_WAV_OZNAKA_EX, WavPrimerjajOznakiExPoID>(tab, n);
                oznake.oznake.Uporabi(tab, n, FALSE);
#endif
            }
        }
    }

    // Tako, imamo oznake. Sedaj pa jih dopolnimo e z dodatnimi informacijami.
    dat.Prestavi(zacetek);
    while (dat.VrniOdmik() < konec_bloka) {
        T_ODMIK konec_vsebine;

        stanje = WavNajdiPrviBlok<T_WAV_SEZNAM>(dat, *(const T_WAV_ID*)"adtl", konec_bloka, &konec_vsebine);
        if (USPESNO(stanje)) {
            // Preberemo vse zapise seznama.
            while (dat.VrniOdmik() < konec_vsebine) {
                T_WAV_ID id;

                dat >> id;
                if (dat.JeKonec()) break;

                if (id == T_WAV_LTXT_ID::id) {
                    // Nali smo podatek LTXT.
                    T_WAV_LTXT ltxt;
                    T_WAV_OZNAKA_EX pom;
                    T_UINT i;

                    dat >> T_WAV_LTXT_ID(ltxt);
                    pom.id = ltxt.id;
#if _MSC_VER >= 1300
                    i = oznake.oznake.PoisciUrejeno<WavPrimerjajOznakiExPoID>(pom);
#else
                    i = abinsearch<T_WAV_OZNAKA_EX, WavPrimerjajOznakiExPoID>(oznake.oznake.Vrni(), oznake.oznake.VrniStevilo(), pom);
#endif
                    if (i != T_UINT_MAX) {
                        T_WAV_OZNAKA_EX &oznaka = oznake.oznake[i];
                        oznaka.dolzina     = ltxt.dolzina;
                        oznaka.id_namena   = ltxt.id_namena;
                        oznaka.drzava      = ltxt.drzava;
                        oznaka.jezik       = ltxt.jezik;
                        oznaka.narecje     = ltxt.narecje;
                        oznaka.kodna_stran = ltxt.kodna_stran;
                        oznaka.opis        = ltxt.opis;
                    }
                } else if (id == T_WAV_NASLOV_ID::id) {
                    // Nali smo naslov.
                    T_WAV_NASLOV naslov;
                    T_WAV_OZNAKA_EX pom;
                    T_UINT i;

                    dat >> T_WAV_NASLOV_ID(naslov);
                    pom.id = naslov.id;
#if _MSC_VER >= 1300
                    i = oznake.oznake.PoisciUrejeno<WavPrimerjajOznakiExPoID>(pom);
#else
                    i = abinsearch<T_WAV_OZNAKA_EX, WavPrimerjajOznakiExPoID>(oznake.oznake.Vrni(), oznake.oznake.VrniStevilo(), pom);
#endif
                    if (i != T_UINT_MAX) {
                        T_WAV_OZNAKA_EX &oznaka = oznake.oznake[i];
                        oznaka.naslov = naslov.naslov;
                    }
                } else if (id == T_WAV_OPOMBA_ID::id) {
                    // Nali smo opombo.
                    T_WAV_OPOMBA opomba;
                    T_WAV_OZNAKA_EX pom;
                    T_UINT i;

                    dat >> T_WAV_OPOMBA_ID(opomba);
                    pom.id = opomba.id;
#if _MSC_VER >= 1300
                    i = oznake.oznake.PoisciUrejeno<WavPrimerjajOznakiExPoID>(pom);
#else
                    i = abinsearch<T_WAV_OZNAKA_EX, WavPrimerjajOznakiExPoID>(oznake.oznake.Vrni(), oznake.oznake.VrniStevilo(), pom);
#endif
                    if (i != T_UINT_MAX) {
                        T_WAV_OZNAKA_EX &oznaka = oznake.oznake[i];
                        oznaka.opomba = opomba.opomba;
                    }
                } else {
                    _PreskociIdPodatke<T_WAV_DOLZINA, WAV_PORAVNAVA>(dat);
                }
            }
        }
    }

    dat.Prestavi(zacetek);
    return A_OK;
}


T_STANJE AOSN_API AOsn::WavIzlociOznake(RDatoteka &vhod, RDatoteka &izhod, T_ODMIK konec_bloka)
{
    T_STANJE stanje;
    T_WAV_ID id;
    T_UINT i;
    T_BOOL izloci;
    T_WAV_DOLZINA dolzina;
    static T_WAV_ID izlocljivi[] = {
        T_WAV_OZNAKE_ID::id,
        T_WAV_LTXT_ID::id,
        T_WAV_NASLOV_ID::id,
        T_WAV_OPOMBA_ID::id,
    };

    while (vhod.VrniOdmik() < konec_bloka) {
        vhod >> id;
        if (vhod.JeKonec()) return A_OK;

        for (i = 0, izloci = FALSE; i < _countof(izlocljivi); i++) {
            if (id == izlocljivi[i]) {
                // Nali smo element, ki ga elimo izloiti.
                izloci = TRUE;
                break;
            }
        }

        if (izloci) {
            // Preskoimo element, ki ga elimo izloiti
            _PreskociIdPodatke<T_WAV_DOLZINA, WAV_PORAVNAVA>(vhod);
        } else {
            // Element prepiemo.

            T_WAV_ID id2 = id;
            T_ODMIK zacetek_bloka;

            vhod >> dolzina;
            if (vhod.JeKonec()) return AOSN_NAPAKA_KONEC_TOKA;

            // Odpremo glavo elementa v izhodni datoteki.
            zacetek_bloka = _OdpriGlavo<T_WAV_ID, T_WAV_DOLZINA>(izhod, id);

            astrnupr((T_ASCII)&id2, sizeof(T_WAV_ID)/sizeof(T_AZNAK));
            if (id != id2 ||
                id == *(const T_WAV_ID*)"ICRD")
            {
                // ID ni sestavljen iz samih velikih rk (ali pa gre za izjemo), zato je element trivialen in ga prepiemo.
                stanje = izhod.VrniStanje();
                if (NEUSPESNO(stanje)) return stanje;
                izhod.ZapisiTok(vhod, dolzina);
                if (vhod.JeKonec()) return AOSN_NAPAKA_KONEC_TOKA;

                // Zakljuimo glavo elementa v izhodni datoteki.
                _ZapriGlavo<T_WAV_ID, T_WAV_DOLZINA, WAV_PORAVNAVA>(izhod, zacetek_bloka);

                dolzina = (T_WAV_DOLZINA)(WAV_PORAVNAVA - dolzina) % WAV_PORAVNAVA; // tevilo bajtov, ki so dodani, da je zapis dolg celo tevilo enot doline WAV_PORAVNAVA
                if (dolzina)
                    verify(USPESNO(vhod.Preskoci(dolzina)));
            } else {
                // ID je sestavljen iz samih velikih rk, zato ga obdelamo rekurzivno.
                T_WAV_ID id;
                T_ODMIK konec_bloka;

                // Preberemo ID elementa.
                vhod >> id;
                if (vhod.JeKonec()) return AOSN_NAPAKA_KONEC_TOKA;

                // Zapiemo glavo elementa.
                stanje = izhod.VrniStanje();
                if (NEUSPESNO(stanje)) return stanje;
                izhod << id;

                konec_bloka = izhod.VrniOdmik();

                // Sedaj obdelamo element rekurzivno.
                stanje = WavIzlociOznake(vhod, izhod, vhod.VrniOdmik() + dolzina - sizeof(T_WAV_ID));

                if (konec_bloka != izhod.VrniOdmik()) {
                    // Zakljuimo glavo elementa v izhodni datoteki.
                    _ZapriGlavo<T_WAV_ID, T_WAV_DOLZINA, WAV_PORAVNAVA>(izhod, zacetek_bloka);
                } else {
                    // Element smo spraznili, zato ga ukinemo.
                    izhod.Prestavi(zacetek_bloka);
                }
            }
        }
    }

    verify(USPESNO(izhod.PostaviKonec()));

    return A_OK;
}

#endif // AOSN_WAV
