/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_BASE64

#define IZRAVNALNI_ZNAK  ((T_AZNAK)'=')


//////////////////////////////////////////////////////////////////////
// Lokalne preslikovalne tabele
//////////////////////////////////////////////////////////////////////

TK_ASCII _koda = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
const T_BYTE _dekoda[] = {
      /*  0    1    2    3    4    5    6    7    8    9    A    B    C    D    E    F  */
/* 0 */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* 1 */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* 2 */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,  62, 255, 255, 255,  63,
/* 3 */     52,  53,  54,  55,  56,  57,  58,  59,  60,  61, 255, 255, 255,  64, 255, 255,
/* 4 */    255,   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,
/* 5 */     15,  16,  17,  18,  19,  20,  21,  22,  23,  24,  25, 255, 255, 255, 255, 255,
/* 6 */    255,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,  36,  37,  38,  39,  40,
/* 7 */     41,  42,  43,  44,  45,  46,  47,  48,  49,  50,  51, 255, 255, 255, 255, 255,
/* 8 */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* 9 */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* A */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* B */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* C */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* D */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* E */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
/* F */    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255
};


//////////////////////////////////////////////////////////////////////
// Definicije lokalnih funkcij
//////////////////////////////////////////////////////////////////////

inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], T_AZNAK izhod[4]);
inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RNizP<T_AZNAK> &izhod);
inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RTokPodatkov &izhod);
inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], T_AZNAK izhod[4], T_UINT dolzina);
inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RNizP<T_AZNAK> &izhod, T_UINT dolzina);
inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RTokPodatkov &izhod, T_UINT dolzina);
inline static T_UINT _DeKodiraj(const T_BYTE d_pom[4], T_BYTE izhod[3]);
inline static T_UINT _DeKodiraj(const T_BYTE d_pom[4], RTokPodatkov &izhod);


//////////////////////////////////////////////////////////////////////
// RBase64
//////////////////////////////////////////////////////////////////////

RBase64::RBase64()
{
    k_stevilo = 0;
    d_stevilo = 0;
}


T_UINT RBase64::Kodiraj(TK_KAZALEC podatki, T_SIZE dolzina, RNizP<T_AZNAK> &izhod, T_BOOL zadnji)
{
    assert(podatki || !dolzina);

    T_UINT i = 0, j = 0;

    // pretvorimo podatke znak po znak
    for (;;) {
        if (k_stevilo >= 3) {
            _Kodiraj(k_pom, izhod);
            k_stevilo = 0;
            j += 4;
        }

        if (i >= dolzina)
            break;

        k_pom[k_stevilo++] = ((T_BYTE*)podatki)[i++];
    }

    // ce je zadnji blok, izpraznimo buffer
    if (zadnji && k_stevilo) {
        _Kodiraj(k_pom, izhod, k_stevilo);
        k_stevilo = 0;
        j += 4;
    }

    return j;
}


T_UINT RBase64::Kodiraj(TK_KAZALEC podatki, T_SIZE dolzina, T_ASCII izhod, T_BOOL zadnji)
{
    assert(podatki || !dolzina);

    T_UINT i = 0, j = 0;

    for (;;) {
        if (k_stevilo >= 3) {
            _Kodiraj(k_pom, izhod + j);
            k_stevilo = 0;
            j += 4;
        }

        if (i >= dolzina)
            break;

        k_pom[k_stevilo++] = ((T_BYTE*)podatki)[i++];
    }

    // ce je zadnji blok, izpraznimo buffer
    if (zadnji && k_stevilo) {
        _Kodiraj(k_pom, izhod + j, k_stevilo);
        k_stevilo = 0;
        j += 4;
    }

    izhod[j] = 0;

    return j;
}


T_UINT RBase64::Kodiraj(TK_KAZALEC podatki, T_SIZE dolzina, RTokPodatkov &izhod, T_BOOL zadnji)
{
    assert(podatki || !dolzina);

    T_UINT i = 0, j = 0;

    // pretvorimo podatke znak po znak
    for (;;) {
        if (k_stevilo >= 3) {
            _Kodiraj(k_pom, izhod);
            k_stevilo = 0;
            j += 4;
        }

        if (i >= dolzina)
            break;

        k_pom[k_stevilo++] = ((T_BYTE*)podatki)[i++];
    }

    // ce je zadnji blok, izpraznimo buffer
    if (zadnji && k_stevilo) {
        _Kodiraj(k_pom, izhod, k_stevilo);
        k_stevilo = 0;
        j += 4;
    }

    return j;
}


T_SIZE RBase64::DeKodiraj(TK_ASCII podatki, RTokPodatkov &izhod, T_BOOL &zadnji)
{
    T_UINT i = 0, j = 0;
    T_UINT zlogov;

    zadnji = FALSE;

    for (;;) {
        if (d_stevilo >= 4) {
            zlogov = _DeKodiraj(d_pom, izhod);
            j += zlogov;
            d_stevilo = 0;
            if (zlogov < 3) {
                zadnji = TRUE;
                break;
            }
        }

        if (!podatki[i])
            break;

        if ((d_pom[d_stevilo] = _dekoda[(T_BYTE)podatki[i++]]) != 255)
            d_stevilo++;

    }

    return j;
}


T_SIZE RBase64::DeKodiraj(TK_ASCII podatki, T_KAZALEC izhod, T_BOOL &zadnji)
{
    T_UINT i = 0, j = 0;
    T_UINT zlogov;

    zadnji = FALSE;

    for (;;) {
        if (d_stevilo >= 4) {
            zlogov = _DeKodiraj(d_pom, (T_BYTE*)izhod + j);
            j += zlogov;
            d_stevilo = 0;
            if (zlogov < 3) {
                zadnji = TRUE;
                break;
            }
        }

        if (!podatki[i])
            break;

        if ((d_pom[d_stevilo] = _dekoda[(T_BYTE)podatki[i++]]) != 255)
            d_stevilo++;
    }

    return j;
}


//////////////////////////////////////////////////////////////////////
// RBase64Tok
//////////////////////////////////////////////////////////////////////

RBase64Tok::RBase64Tok(T_UINT _max_blokov, RTokPodatkov *izvor) :
    RPretvornikToka(izvor), RBase64(),
    max_blokov(_max_blokov)
{
}


T_STANJE RBase64Tok::Uporabi(RTokPodatkov *izvor)
{
    pom_odmik = pom_dolzina = st_blokov = 0;
    return RPretvornikToka::Uporabi(izvor);
}


RTokPodatkov* RBase64Tok::Odpusti()
{
    assert(izvor);

    // izpraznimo buffer
    if (k_stevilo) {
        _Kodiraj(k_pom, *izvor, k_stevilo);
        k_stevilo = 0;
    }

    return RPretvornikToka::Odpusti();
}


T_SIZE RBase64Tok::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    T_UINT j = 0;
    T_BYTE podatek = 0;
    T_SIZE ostanek = dolzina;

    while (ostanek) {
        // Ce nam je zmanjkalo podatkov v pomoznem pomnilniku, jih gremo iskat.
        if (!pom_dolzina) {
            // Preberemo iz vhoda stiri bajte (za Base64).
            while (d_stevilo < 4) {
                *izvor >> podatek;
                stanje = izvor->VrniStanje();
                if (NEUSPESNO(stanje))
                    goto konec;

                if ((d_pom[d_stevilo] = _dekoda[podatek]) != 255)
                    d_stevilo++;
            }

            // Napolnimo pomozni pomnilnik
            assert(d_stevilo == 4);
            pom_odmik = 0;
            pom_dolzina = _DeKodiraj(d_pom, pom);
            assert(pom_dolzina);
            d_stevilo = 0;
            if (pom_dolzina < 3) // Ce zaznamo konec v zapisu Base64, potvorimo ostanek, da preberemo le se repek.
                ostanek = amin(ostanek, pom_dolzina);
        }

        assert(j < dolzina);
        assert(ostanek);
        assert(pom_odmik < 3);
        assert(pom_dolzina);

        ((T_BYTE*)podatki)[j++] = pom[pom_odmik++];
        ostanek--; pom_dolzina--;
    }

konec:
    return j;
}


T_SIZE RBase64Tok::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    T_UINT i = 0, j = 0;

    // pretvorimo podatke znak po znak
    for (;;) {
        if (k_stevilo >= 3) {
            if (++st_blokov > max_blokov) {
                *izvor << (T_AZNAK)'\n';
                st_blokov = 1;
            }

            _Kodiraj(k_pom, *izvor);
            k_stevilo = 0;
            j += 4;
        }

        if (i >= dolzina)
            break;

        k_pom[k_stevilo++] = ((T_BYTE*)podatki)[i++];
    }

    return i;
}


//////////////////////////////////////////////////////////////////////
// Lokalne funkcije
//////////////////////////////////////////////////////////////////////

inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], T_AZNAK izhod[4])
{
    izhod[0] = _koda[k_pom[0] >> 2];
    izhod[1] = _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
    izhod[2] = _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
    izhod[3] = _koda[k_pom[2] & 0x3f];
}


inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RNizP<T_AZNAK> &izhod)
{
    izhod += _koda[k_pom[0] >> 2];
    izhod += _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
    izhod += _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
    izhod += _koda[k_pom[2] & 0x3f];
}


inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RTokPodatkov &izhod)
{
    izhod << _koda[k_pom[0] >> 2];
    izhod << _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
    izhod << _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
    izhod << _koda[k_pom[2] & 0x3f];
}


inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], T_AZNAK izhod[4], T_UINT dolzina)
{
    if (dolzina>0) {
        izhod[0] = _koda[k_pom[0] >> 2];
        if (dolzina>1) {
            izhod[1] = _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
            if (dolzina>2) {
                izhod[2] = _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
                izhod[3] = _koda[k_pom[2] & 0x3f];
            } else {
                izhod[2] = _koda[(k_pom[1] << 2) & 0x3f];
                izhod[3] = IZRAVNALNI_ZNAK;
            }
        } else {
            izhod[1] = _koda[(k_pom[0] << 4) & 0x3f];
            izhod[2] = IZRAVNALNI_ZNAK;
            izhod[3] = IZRAVNALNI_ZNAK;
        }
    } else {
        izhod[0] = IZRAVNALNI_ZNAK;
        izhod[1] = IZRAVNALNI_ZNAK;
        izhod[2] = IZRAVNALNI_ZNAK;
        izhod[3] = IZRAVNALNI_ZNAK;
    }
}


inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RNizP<T_AZNAK> &izhod, T_UINT dolzina)
{
    if (dolzina>0) {
        izhod += _koda[k_pom[0] >> 2];
        if (dolzina>1) {
            izhod += _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
            if (dolzina>2) {
                izhod += _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
                izhod += _koda[k_pom[2] & 0x3f];
            } else {
                izhod += _koda[(k_pom[1] << 2) & 0x3f];
                izhod += IZRAVNALNI_ZNAK;
            }
        } else {
            izhod += _koda[(k_pom[0] << 4) & 0x3f];
            izhod += IZRAVNALNI_ZNAK;
            izhod += IZRAVNALNI_ZNAK;
        }
    } else {
        izhod += IZRAVNALNI_ZNAK;
        izhod += IZRAVNALNI_ZNAK;
        izhod += IZRAVNALNI_ZNAK;
        izhod += IZRAVNALNI_ZNAK;
    }
}


inline static T_VOID _Kodiraj(const T_BYTE k_pom[3], RTokPodatkov &izhod, T_UINT dolzina)
{
    if (dolzina>0) {
        izhod << _koda[k_pom[0] >> 2];
        if (dolzina>1) {
            izhod << _koda[((k_pom[0] << 4) | (k_pom[1] >> 4)) & 0x3f];
            if (dolzina>2) {
                izhod << _koda[((k_pom[1] << 2) | (k_pom[2] >> 6)) & 0x3f];
                izhod << _koda[k_pom[2] & 0x3f];
            } else {
                izhod << _koda[(k_pom[1] << 2) & 0x3f];
                izhod << IZRAVNALNI_ZNAK;
            }
        } else {
            izhod << _koda[(k_pom[0] << 4) & 0x3f];
            izhod << IZRAVNALNI_ZNAK;
            izhod << IZRAVNALNI_ZNAK;
        }
    } else {
        izhod << IZRAVNALNI_ZNAK;
        izhod << IZRAVNALNI_ZNAK;
        izhod << IZRAVNALNI_ZNAK;
        izhod << IZRAVNALNI_ZNAK;
    }
}


inline static T_UINT _DeKodiraj(const T_BYTE d_pom[4], T_BYTE izhod[3])
{
    izhod[0] = (T_BYTE)(((d_pom[0] << 2) | (d_pom[1] >> 4)) & 0xff);
    if (d_pom[2]<64) {
        izhod[1]  = (T_BYTE)(((d_pom[1] << 4) | (d_pom[2] >> 2)) & 0xff);
        if (d_pom[3]<64) {
            izhod[2]  = (T_BYTE)(((d_pom[2] << 6) | d_pom[3]) & 0xff);
            return 3;
        } else
            return 2;
    } else
        return 1;
}


inline static T_UINT _DeKodiraj(const T_BYTE d_pom[4], RTokPodatkov &izhod)
{
    izhod << (T_BYTE)(((d_pom[0] << 2) | (d_pom[1] >> 4)) & 0xff);
    if (d_pom[2]<64) {
        izhod << (T_BYTE)(((d_pom[1] << 4) | (d_pom[2] >> 2)) & 0xff);
        if (d_pom[3]<64) {
            izhod << (T_BYTE)(((d_pom[2] << 6) | d_pom[3]) & 0xff);
            return 3;
        } else
            return 2;
    } else
        return 1;
}

#endif // AOSN_BASE64
