/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"
#include <basetyps.h>
#include <limits.h>
#include <winioctl.h>


//////////////////////////////////////////////////////////////////////
// RTokPodatkov
//////////////////////////////////////////////////////////////////////

RTokPodatkov::RTokPodatkov() : RIzpElement()
{
    stanje = A_OK;
}


T_SIZE RTokPodatkov::PreberiOstanek(T_KAZALEC &p, T_SIZE dolzina_max)
{
    T_SIZE odmik, dolzina;

    // Preberemo ostanek toka v pomnilnik po blokih.
    p       = NULL;
    odmik   = 0;
    dolzina = RDAT_VELIKOST_BLOKA;
    while (odmik < dolzina_max) {
        T_KAZALEC p2;

        dolzina = amin(dolzina, dolzina_max);
        p2 = arealloc(p, dolzina);
        if (!p2) {
            afree(p);
            p = NULL;
            stanje = AOSN_NAPAKA_POMNILNIK;
            return 0;
        }
        p = p2;

        odmik += PreberiTab((T_U1*)p + odmik, 1, dolzina - odmik);
        if (NEUSPESNO(stanje)) break;

        dolzina += RDAT_VELIKOST_BLOKA;
    }

    return odmik;
}


T_UINT RTokPodatkov::ZapisiObl(TK_ASCII format, T_LOCALE locale, ...)
{
    va_list parametri;
    T_UINT znakov;

    va_start(parametri, locale);
    znakov = ZapisiOblP(format, locale, parametri);
    va_end(parametri);

    return znakov;
}


T_UINT RTokPodatkov::ZapisiObl(TK_UNICODE format, T_LOCALE locale, ...)
{
    va_list parametri;
    T_UINT znakov;

    va_start(parametri, locale);
    znakov = ZapisiOblP(format, locale, parametri);
    va_end(parametri);

    return znakov;
}


T_UINT RTokPodatkov::ZapisiOblP(TK_ASCII format, T_LOCALE locale, va_list parametri)
{
    T_UINT znakov = 0;
    T_ASCII podatki = NULL;
    T_INT zapisano;

    do {
        // Po potrebi poveujemo blok v pomnilniku s korakom RDAT_VELIKOST_BLOKA.
        znakov += RDAT_VELIKOST_BLOKA;

        // Rezerviramo pomnilniki blok za prepisovanje.
        podatki = (T_ASCII)arealloc(podatki, znakov * sizeof(T_AZNAK));
        assert(podatki || !znakov);

        // Zapiemo v pomnilniki blok, in upamo, da je bil dovolj velik.
        zapisano = avsnprintf(podatki, znakov, format, locale, parametri);
    } while (zapisano < 0);

    // Zapiemo podatke v datoteko.
    znakov = ZapisiTab(podatki, sizeof(T_AZNAK), zapisano);
    afree(podatki);

    return znakov;
}


T_UINT RTokPodatkov::ZapisiOblP(TK_UNICODE format, T_LOCALE locale, va_list parametri)
{
    T_SIZE znakov = 0;
    T_UNICODE podatki = NULL;
    T_INT zapisano;

    do {
        // Po potrebi poveujemo blok v pomnilniku s korakom RDAT_VELIKOST_BLOKA.
        znakov += RDAT_VELIKOST_BLOKA;

        // Rezerviramo buffer za prepisovanje.
        podatki = (T_UNICODE)arealloc(podatki, znakov * sizeof(T_UZNAK));
        assert(podatki || !znakov);

        // Zapiemo v buffer, in upamo, da je bil dovolj velik.
        zapisano = avsnprintf(podatki, znakov, format, locale, parametri);
    } while (zapisano<0);

    // Zapiemo podatke v datoteko.
    znakov = ZapisiTab(podatki, sizeof(T_UZNAK), zapisano);
    afree(podatki);

    return znakov;
}


T_STANJE RTokPodatkov::Preskoci(T_DOLZINA koliko)
{
    if (koliko == 1) {
        // Preberemo le en bajt in ne kompliciramo.
        PreberiBajt();
    } else if (koliko < RDAT_VELIKOST_ITERIRANJA) {
        // eleno tevilo bajtov je dovolj kratko, da jih preberemo v zanki.
        for (T_UINT i = 0; i < (T_UINT)koliko; i++) {
            PreberiBajt();
#ifdef RDAT_NE_PREVERJAJ_STANJA_VER
            assert(!JeKonec());
#else
            if (JeKonec()) break;
#endif
        }
    } else {
        // Bajtov je ve. Rezerviramo blok v pomnilniku in beremo po blokih.
        T_SIZE blok = (T_SIZE)amin<T_DOLZINA>(koliko, RDAT_VELIKOST_BLOKA);
        T_KAZALEC dummy = aalloc(blok);
        assert(dummy);
        while (koliko) {
            koliko -= PreberiTab(dummy, 1, (T_SIZE)amin<T_DOLZINA>(koliko, blok));
            // Obvezno preverjamo stanje, sicer se lahko ujamemo v neskonno zanko, e se branje kona zaradi napake.
            if (JeKonec()) break;
        }
        afree(dummy);
    }

    return stanje;
}


//////////////////////////////////////////////////////////////////////
// RDatoteka
//////////////////////////////////////////////////////////////////////

T_SIZE RDatoteka::PreberiOstanek(T_KAZALEC &p, T_SIZE dolzina_max)
{
    T_SIZE dolzina;

    // Ker vemo, koliko je e do konca datoteke, lahko izraunamo tono dolino.
    dolzina = amin<T_SIZE>(dolzina_max, (T_SIZE)(VrniDolzino() - VrniOdmik()));

    // Rezerviramo pomnilnik.
    p = aalloc(dolzina);
    if (!p) {
        stanje = AOSN_NAPAKA_POMNILNIK;
        return 0;
    }

    // Preberemo podatke.
    return PreberiTab(p, 1, dolzina);
}


T_STANJE RDatoteka::Preskoci(T_DOLZINA koliko)
{
    Prestavi(koliko, RDAT_OD_TRENUTNEGA);
    return stanje;
}


T_STANJE RDatoteka::ZakleniObmocje(T_ODMIK, T_DOLZINA)
{
    return stanje = AOSN_NAPAKA_ZAKLEP;
}


T_STANJE RDatoteka::OdkleniObmocje(T_ODMIK, T_DOLZINA)
{
    return stanje = AOSN_NAPAKA_ODKLEP;
}


T_TRENUTEK RDatoteka::VrniDatumStvaritve() const
{
    return T_TRENUTEK_MIN;
}


T_TRENUTEK RDatoteka::VrniDatumDostopa() const
{
    return T_TRENUTEK_MIN;
}


T_TRENUTEK RDatoteka::VrniDatumPisanja() const
{
    return T_TRENUTEK_MIN;
}


T_STANJE RDatoteka::NastaviDatumStvaritve(T_TRENUTEK datum)
{
    UNREFERENCED_PARAMETER(datum);
    return AOSN_NAPAKA_NI_PODPRTO;
}


T_STANJE RDatoteka::NastaviDatumDostopa(T_TRENUTEK datum)
{
    UNREFERENCED_PARAMETER(datum);
    return AOSN_NAPAKA_NI_PODPRTO;
}


T_STANJE RDatoteka::NastaviDatumPisanja(T_TRENUTEK datum)
{
    UNREFERENCED_PARAMETER(datum);
    return AOSN_NAPAKA_NI_PODPRTO;
}


//////////////////////////////////////////////////////////////////////
// RPodvojevalnikToka
//////////////////////////////////////////////////////////////////////

RPodvojevalnikToka::RPodvojevalnikToka() : RTokPodatkov()
{
}


RPodvojevalnikToka::~RPodvojevalnikToka()
{
#ifdef AOSN_NIT
    T_UINT i, n = cilji.VrniStevilo();

    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);

        // Naroimo niti, naj kona, jo poakamo in poistimo.
        nit->ukaz = RCiljnaNit::KONCAJ;
        verify(USPESNO(nit->zacni.Postavi()));
        verify(USPESNO(nit->Pocakaj()));
        verify(USPESNO(nit->Zapri()));
        delete nit;
    }
#else
    cilji.Unici();
#endif
}


T_STANJE RPodvojevalnikToka::Dodaj(RTokPodatkov *cilj)
{
#ifdef AOSN_NIT
    assert(cilj);

    // Ustvarimo nit.
    RCiljnaNit *nit = new RCiljnaNit(cilj);
    if (!nit)
        return AOSN_NAPAKA_POMNILNIK;

    // Dodamo jo na spisek ciljev.
    cilji    += nit;
    koncanje += nit->koncano.VrniRocico();

    // Poenemo nit.
    return nit->Odpri();
#else
    cilji.Dodaj(cilj);
    return A_OK;
#endif
}


T_STANJE RPodvojevalnikToka::Odstrani(RTokPodatkov *cilj)
{
#ifdef AOSN_NIT
    assert(cilj);

    T_UINT i, n = cilji.VrniStevilo();

    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);
        if (nit->cilj == cilj) {
            // Nali smo eleni cilj. Naroimo njegovi niti, naj kona, jo poakamo in poistimo.
            nit->ukaz = RCiljnaNit::KONCAJ;
            verify(USPESNO(nit->zacni.Postavi()));
            verify(USPESNO(nit->Pocakaj()));
            verify(USPESNO(nit->Zapri()));
            delete nit;

            // Izloimo izbrisani element iz tabel cilji in koncanje.
            RCiljnaNit **_cilji = cilji.Odpusti(n, FALSE);
            memcpy(_cilji + i, _cilji + i + 1, (n - i - 1) * sizeof(RCiljnaNit*));
            cilji.Uporabi(_cilji, n - 1, FALSE);

            T_ROCICAOS *_koncanje = koncanje.Odpusti(n, FALSE);
            memcpy(_koncanje + i, _koncanje + i + 1, (n - i - 1) * sizeof(T_ROCICAOS));
            koncanje.Uporabi(_koncanje, n - 1, FALSE);

            return A_OK;
        }
    }
#else
    T_ROCICA rocica;

    for (rocica = cilji.VrniPrvega(); rocica; rocica = cilji.VrniNaslednjega(rocica)) {
        if (cilji[rocica] == cilj) {
            // Nali smo eleni cilj.
            cilji.Brisi(rocica);
            return A_OK;
        }
    }
#endif

    return AOSN_NAPAKA_NI_NAJDENO;
}


T_SIZE RPodvojevalnikToka::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    UNREFERENCED_PARAMETER(podatki);
    UNREFERENCED_PARAMETER(dolzina);

    stanje = AOSN_NAPAKA_NI_PODPRTO;
    return 0;
}


T_SIZE RPodvojevalnikToka::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    T_SIZE zapisano = dolzina;

#ifdef AOSN_NIT
    T_UINT i, n = cilji.VrniStevilo();

    // Naroimo pisanje vsem ciljem.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);
        nit->ukaz    = RCiljnaNit::ZAPISI;
        nit->podatki = podatki;
        nit->dolzina = dolzina;
        nit->koncano.Resetiraj();
        verify(USPESNO(nit->zacni.Postavi()));
    }

    // Poakamo, da vsi konajo.
    verify(::WaitForMultipleObjects((DWORD)n, (const HANDLE*)koncanje.Vrni(), TRUE, INFINITE) == WAIT_OBJECT_0);

    // Analiziramo njihove rezultate.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);

        stanje = nit->cilj->VrniStanje();
        if (NEUSPESNO(stanje)) return 0;

        if (nit->zapisano < zapisano)
            zapisano = nit->zapisano;
    }
#else
    T_ROCICA rocica;

    for (rocica = cilji.VrniPrvega(); rocica; rocica = cilji.VrniNaslednjega(rocica)) {
        RTokPodatkov *&cilj = cilji[rocica];
        assert(cilj);

        zapisano = amin(cilj->Zapisi(podatki, dolzina), zapisano);
        stanje = cilj->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
    }
#endif

    stanje = zapisano == dolzina ? A_OK : AOSN_NAPAKA_PISANJE;
    return zapisano;
}


T_STANJE RPodvojevalnikToka::Zapri()
{
#ifdef AOSN_NIT
    T_UINT i, n = cilji.VrniStevilo();

    // Naroimo zapiranje vsem ciljem.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);
        nit->ukaz = RCiljnaNit::ZAPRI;
        nit->koncano.Resetiraj();
        verify(USPESNO(nit->zacni.Postavi()));
    }

    // Poakamo, da vsi konajo.
    verify(::WaitForMultipleObjects((DWORD)n, (const HANDLE*)koncanje.Vrni(), TRUE, INFINITE) == WAIT_OBJECT_0);

    // Analiziramo njihove rezultate.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);

        stanje = nit->cilj->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
    }
#else
    T_ROCICA rocica;

    for (rocica = cilji.VrniPrvega(); rocica; rocica = cilji.VrniNaslednjega(rocica)) {
        RTokPodatkov *&cilj = cilji[rocica];
        assert(cilj);

        stanje = cilj->Zapri();
        if (NEUSPESNO(stanje)) return stanje;
    }
#endif

    return stanje = A_OK;
}


T_STANJE RPodvojevalnikToka::Izplakni()
{
#ifdef AOSN_NIT
    T_UINT i, n = cilji.VrniStevilo();

    // Naroimo izplakovanje vsem ciljem.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);
        nit->ukaz = RCiljnaNit::IZPLAKNI;
        nit->koncano.Resetiraj();
        verify(USPESNO(nit->zacni.Postavi()));
    }

    // Poakamo, da vsi konajo.
    verify(::WaitForMultipleObjects((DWORD)n, (const HANDLE*)koncanje.Vrni(), TRUE, INFINITE) == WAIT_OBJECT_0);

    // Analiziramo njihove rezultate.
    for (i = 0; i < n; i++) {
        RCiljnaNit *&nit = cilji[i];
        assert(nit);

        stanje = nit->cilj->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
    }
#else
    T_ROCICA rocica;

    for (rocica = cilji.VrniPrvega(); rocica; rocica = cilji.VrniNaslednjega(rocica)) {
        RTokPodatkov *&cilj = cilji[rocica];
        assert(cilj);

        stanje = cilj->Izplakni();
        if (NEUSPESNO(stanje)) return stanje;
    }
#endif

    return stanje = A_OK;
}


#ifdef AOSN_NIT

RPodvojevalnikToka::RCiljnaNit::RCiljnaNit(RTokPodatkov *_cilj) :
    cilj(_cilj),
    koncano(RSEM_POSTAVLJEN | RSEM_ROCNI_RESET),
    RNit()
{
}


T_REZULTAT RPodvojevalnikToka::RCiljnaNit::Izvedi(T_KAZALEC param)
{
    UNREFERENCED_PARAMETER(param);
    assert(cilj);

    for (;;) {
        verify(USPESNO(zacni.Pocakaj()));
        switch (ukaz) {
        case KONCAJ:
            // Zakljuimo delo.
            return 0;

        case ZAPISI:
            // Zapiemo blok podatkov.
            zapisano = cilj->Zapisi(podatki, dolzina);
            break;

        case ZAPRI:
            // Zapremo tok podatkov.
            cilj->Zapri();
            break;

        case IZPLAKNI:
            // Izplaknemo tok podatkov.
            cilj->Izplakni();
            break;

        default:
            assert(0); // Neznan ukaz
        }

        verify(USPESNO(koncano.Postavi()));
    }
}

#endif


//////////////////////////////////////////////////////////////////////
// RPretvornikToka
//////////////////////////////////////////////////////////////////////

RPretvornikToka::RPretvornikToka(RTokPodatkov *i) : RTokPodatkov(), izvor(NULL)
{
    if (i)
        verify(USPESNO(Uporabi(i)));
}


RPretvornikToka::~RPretvornikToka()
{
    // Pretvornik razpustimo, e ga se nismo e prej.
    if (izvor)
        Odpusti();

    assert(!izvor);
}


T_STANJE RPretvornikToka::Uporabi(RTokPodatkov *_izvor)
{
    assert(_izvor);
    assert(!izvor);

    izvor = _izvor;
    return stanje = _izvor->stanje;
}


RTokPodatkov* RPretvornikToka::Odpusti()
{
    assert(izvor);
    RTokPodatkov *stari_izvor = izvor;

    stanje = A_OK;
    izvor = NULL;
    return stari_izvor;
}


T_SIZE RPretvornikToka::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    T_SIZE prebrano = izvor->Preberi(podatki, dolzina);
    stanje = izvor->stanje;
    return prebrano;
}


T_SIZE RPretvornikToka::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    T_SIZE zapisano = izvor->Zapisi(podatki, dolzina);
    stanje = izvor->stanje;
    return zapisano;
}


T_STANJE RPretvornikToka::Zapri()
{
    return izvor ? Odpusti()->Zapri() : stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RPretvornikToka::Izplakni()
{
    assert(izvor);
    return stanje = izvor->Izplakni();
}


//////////////////////////////////////////////////////////////////////
// RIzravnalnik
//////////////////////////////////////////////////////////////////////

RIzravnalnik::RIzravnalnik(RTokPodatkov *izvor, T_SIZE dolzina_branja, T_SIZE dolzina_pisanja) : RPretvornikToka(izvor)
{
    izravnalnik[0].dolzina = dolzina_branja;
    izravnalnik[0].podatki = NULL;

    izravnalnik[1].dolzina = dolzina_pisanja;
    izravnalnik[1].podatki = NULL;
}


RIzravnalnik::~RIzravnalnik()
{
    assert(!izravnalnik[0].podatki);
    assert(!izravnalnik[1].podatki);
}


T_STANJE RIzravnalnik::Uporabi(RTokPodatkov *_izvor)
{
    assert(!izvor);
    assert(!izravnalnik[0].podatki);
    assert(!izravnalnik[1].podatki);

    // Pripravimo pomnilnik za branje.
    izravnalnik[0].odmik = 0;
    izravnalnik[0].zasedeno = 0;
    if (izravnalnik[0].dolzina != 0) {
        izravnalnik[0].podatki = aalloc(izravnalnik[0].dolzina);
        if (!izravnalnik[0].podatki) {
            stanje = AOSN_NAPAKA_POMNILNIK;
            goto napaka1;
        }
    }

    // Pripravimo pomnilnik za pisanje.
    izravnalnik[1].odmik = 0;
    izravnalnik[1].zasedeno = 0;
    if (izravnalnik[1].dolzina != 0) {
        izravnalnik[1].podatki = aalloc(izravnalnik[1].dolzina);
        if (!izravnalnik[1].podatki) {
            stanje = AOSN_NAPAKA_POMNILNIK;
            goto napaka2;
        }
    }

    stanje = RPretvornikToka::Uporabi(_izvor);
    if (NEUSPESNO(stanje)) goto napaka3;
    return stanje;

napaka3:
    if (izravnalnik[1].podatki) {
        afree(izravnalnik[1].podatki);
        izravnalnik[1].podatki = NULL;
    }
napaka2:
    if (izravnalnik[0].podatki) {
        afree(izravnalnik[0].podatki);
        izravnalnik[0].podatki = NULL;
    }
napaka1:
    return stanje;
}


RTokPodatkov* RIzravnalnik::Odpusti()
{
    assert(izvor);

    RTokPodatkov *_izvor;

    // Izplaknemo izravnalnik in odpustimo tok iz uporabe.
    verify(USPESNO(IzplakniIzravnalnik()));
    _izvor = RPretvornikToka::Odpusti();

    // Sprostimo pomnilnika.
    if (izravnalnik[1].podatki) {
        afree(izravnalnik[1].podatki);
        izravnalnik[1].podatki = NULL;
    }

    if (izravnalnik[0].podatki) {
        afree(izravnalnik[0].podatki);
        izravnalnik[0].podatki = NULL;
    }

    return _izvor;
}


T_SIZE RIzravnalnik::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    if (!dolzina) {
        // Ni potrebe po branju, javimo, da je vse v redu.
        stanje = A_OK;
        return 0;
    } else if (izravnalnik[0].podatki) {
        T_SIZE
            prebrano = 0,
            ostanek = dolzina,
            ostanek_v_izravnalniku;

        stanje = A_OK;

        for (;;) {
            assert(izravnalnik[0].odmik <= izravnalnik[0].zasedeno);
            assert(izravnalnik[0].zasedeno <= izravnalnik[0].dolzina);

            ostanek_v_izravnalniku = izravnalnik[0].zasedeno - izravnalnik[0].odmik;

            // Ali imamo dovolj podatkov v izravnalniku?
            if (ostanek <= ostanek_v_izravnalniku) {
                // Imamo dovolj podatkov. Prepiemo, kar potrebujemo in lahko konamo.
                assert(prebrano <= dolzina && prebrano + ostanek <= dolzina);
                assert(izravnalnik[0].odmik <= izravnalnik[0].zasedeno && izravnalnik[0].odmik + ostanek <= izravnalnik[0].zasedeno);

                amemcpy((T_U1*)podatki + prebrano, (T_U1*)(izravnalnik[0].podatki) + izravnalnik[0].odmik, ostanek);
                prebrano             += ostanek;
                izravnalnik[0].odmik += ostanek;
                stanje = prebrano ? A_OK : izvor->stanje;
                return prebrano;
            } else {
                // Nimamo dovolj podatkov. Prepiemo, kar imamo v izravnalniku.
                if (ostanek_v_izravnalniku) {
                    assert(prebrano <= dolzina && prebrano + ostanek_v_izravnalniku <= dolzina);
                    assert(izravnalnik[0].odmik <= izravnalnik[0].zasedeno && izravnalnik[0].odmik + ostanek_v_izravnalniku <= izravnalnik[0].zasedeno);

                    amemcpy((T_U1*)podatki + prebrano, (T_U1*)(izravnalnik[0].podatki) + izravnalnik[0].odmik, ostanek_v_izravnalniku);
                    prebrano += ostanek_v_izravnalniku;
                    ostanek  -= ostanek_v_izravnalniku;
                }

                izravnalnik[0].odmik = 0;

                // e je uporabnik zahteval ve podatkov, kot pa jih sprejme izravnalnik, beremo mimo izravnalnika.
                if (ostanek > izravnalnik[0].dolzina) {
                    izravnalnik[0].zasedeno = 0; // Razveljavimo izravnalnik.
                    assert(prebrano <= dolzina && prebrano + ostanek == dolzina);
                    prebrano += izvor->Preberi((T_U1*)podatki + prebrano, ostanek);
                    stanje = prebrano ? A_OK : izvor->stanje;
                    return prebrano;
                } else {
                    // Ponovno napolnimo izravnalnik.
                    izravnalnik[0].zasedeno = izvor->Preberi(izravnalnik[0].podatki, izravnalnik[0].dolzina);

                    // e je prilo premalo podatkov, bomo predhodno konali branje.
                    if (izravnalnik[0].zasedeno < izravnalnik[0].dolzina && izravnalnik[0].zasedeno < ostanek)
                        ostanek = izravnalnik[0].zasedeno;
                }
            }
        }
    } else
        return RPretvornikToka::Preberi(podatki, dolzina);
}


T_SIZE RIzravnalnik::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);

    if (izravnalnik[1].podatki) {
        if (dolzina) {
            T_SIZE
                zapisano = 0,
                ostanek = dolzina,
                ostanek_v_izravnalniku;

            for (;;) {
                assert(izravnalnik[1].odmik <= izravnalnik[1].zasedeno);
                assert(izravnalnik[1].zasedeno <= izravnalnik[1].dolzina);

                ostanek_v_izravnalniku = izravnalnik[1].dolzina - izravnalnik[1].zasedeno;

                // Ali je e dovolj prostora v izravnalniku?
                if (ostanek <= ostanek_v_izravnalniku) {
                    // V izdarvanlniku je e dovolj prostora. Prepiemo podatke v izravnalnik in lahko konamo.
                    assert(izravnalnik[1].zasedeno <= izravnalnik[1].dolzina && izravnalnik[1].zasedeno + ostanek <= izravnalnik[1].dolzina);
                    assert(zapisano <= dolzina && zapisano + ostanek <= dolzina);

                    amemcpy((T_U1*)(izravnalnik[1].podatki) + izravnalnik[1].zasedeno, (T_U1*)podatki + zapisano, ostanek);
                    zapisano                += ostanek;
                    izravnalnik[1].zasedeno += ostanek;
                    return zapisano;
                } else {
                    // V izravnalniku ni ve dovolj prostora. Prepiemo toliko podatkov v izravnalnik, kolikor jih e gre.
                    if (ostanek_v_izravnalniku) {
                        assert(izravnalnik[1].zasedeno <= izravnalnik[1].dolzina && izravnalnik[1].zasedeno + ostanek_v_izravnalniku <= izravnalnik[1].dolzina);
                        assert(zapisano <= dolzina && zapisano + ostanek_v_izravnalniku <= dolzina);

                        amemcpy((T_U1*)(izravnalnik[1].podatki) + izravnalnik[1].zasedeno, (T_U1*)podatki + zapisano, ostanek_v_izravnalniku);
                        zapisano                += ostanek_v_izravnalniku;
                        ostanek                 -= ostanek_v_izravnalniku;
                        izravnalnik[1].zasedeno += ostanek_v_izravnalniku;

                        assert(izravnalnik[1].zasedeno == izravnalnik[1].dolzina);
                    }

                    // Izraunamo, koliken del izravnalnika je e nezapisan.
                    assert(izravnalnik[1].zasedeno >= izravnalnik[1].odmik);
                    ostanek_v_izravnalniku = izravnalnik[1].zasedeno - izravnalnik[1].odmik;

                    // Spraznimo izravnalnik, tako da zapiemo, kar je v izravnalniku od prej (e je kaj zapisati).
                    if (ostanek_v_izravnalniku) {
                        assert(izravnalnik[1].odmik <= izravnalnik[1].zasedeno && izravnalnik[1].odmik + ostanek_v_izravnalniku <= izravnalnik[1].zasedeno);
                        izravnalnik[1].odmik += izvor->Zapisi(((T_U1*)izravnalnik[1].podatki) + izravnalnik[1].odmik, ostanek_v_izravnalniku);
                        stanje = izvor->stanje;

                        // e se je zapisalo premalo podatkov, bomo predhodno konali pisanje.
                        if (izravnalnik[1].odmik == izravnalnik[1].zasedeno)
                            izravnalnik[1].odmik = izravnalnik[1].zasedeno = 0;
                        else
                            return zapisano;
                    }

                    // e je ostanek prepredolg, da bi ga sploh lahko v celoti sprejel izravnalnik, ga zapiemo v datoteko kar mimo izravnalnika.
                    if (ostanek > izravnalnik[1].dolzina) {
                        assert(zapisano <= dolzina && zapisano + ostanek <= dolzina);
                        zapisano += izvor->Zapisi(((T_U1*)podatki) + zapisano, ostanek);
                        stanje = izvor->stanje;
                        return zapisano;
                    }
                }
            }
        } else {
            assert(izravnalnik[1].odmik <= izravnalnik[1].zasedeno);
            assert(izravnalnik[1].zasedeno <= izravnalnik[1].dolzina);

            T_SIZE ostanek_v_izravnalniku = izravnalnik[1].zasedeno - izravnalnik[1].odmik;

            // e je e kaj v izravnalniku, zapiemo.
            if (ostanek_v_izravnalniku) {
                assert(izravnalnik[1].odmik <= izravnalnik[1].zasedeno && izravnalnik[1].odmik + ostanek_v_izravnalniku <= izravnalnik[1].zasedeno);
                izravnalnik[1].odmik += izvor->Zapisi(((T_U1*)izravnalnik[1].podatki) + izravnalnik[1].odmik, ostanek_v_izravnalniku);
                stanje = izvor->stanje;

                // e se je zapisalo premalo podatkov, bomo predhodno konali pisanje.
                if (izravnalnik[1].odmik == izravnalnik[1].zasedeno)
                    izravnalnik[1].odmik = izravnalnik[1].zasedeno = 0;
                else
                    return 0;
            }

            izvor->Zapisi(NULL, 0);
            stanje = izvor->stanje;
            return 0;
        }
    } else
        return RPretvornikToka::Zapisi(podatki, dolzina);
}


T_STANJE RIzravnalnik::Izplakni()
{
    return USPESNO(IzplakniIzravnalnik()) ? stanje = RPretvornikToka::Izplakni() : stanje;
}


T_STANJE RIzravnalnik::IzplakniIzravnalnik()
{
    assert(izvor);

    if (izravnalnik[1].podatki) {
        T_SIZE ostanek_v_izravnalniku = izravnalnik[1].zasedeno - izravnalnik[1].odmik;

        // e je e kaj v izravnalniku, zapiemo.
        if (ostanek_v_izravnalniku) {
            izravnalnik[1].odmik += izvor->Zapisi(((T_U1*)izravnalnik[1].podatki) + izravnalnik[1].odmik, ostanek_v_izravnalniku);

            // Preverimo, e se je zapisalo vse.
            if (izravnalnik[1].odmik == izravnalnik[1].zasedeno) {
                izravnalnik[1].odmik = 0;
                izravnalnik[1].zasedeno = 0;
            } else
                return stanje = izvor->stanje;
        }
    }

    return stanje = A_OK;
}


//////////////////////////////////////////////////////////////////////
// ROmejenTokPodatkov
//////////////////////////////////////////////////////////////////////

ROmejenTokPodatkov::ROmejenTokPodatkov(RTokPodatkov *izvor, T_DOLZINA _dolzina_branja, T_DOLZINA _dolzina_pisanja) :
    RPretvornikToka(izvor),
    dolzina_branja(_dolzina_branja),
    dolzina_pisanja(_dolzina_pisanja)
{
}


T_SIZE ROmejenTokPodatkov::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);

    T_SIZE prebrano;

    if (dolzina <= dolzina_branja) {
        prebrano = izvor->Preberi(podatki, dolzina);
        stanje   = izvor->VrniStanje();
    } else if (dolzina && !dolzina_branja) {
        prebrano = 0;
        stanje   = AOSN_NAPAKA_KONEC_TOKA;
    } else {
        prebrano = izvor->Preberi(podatki, (T_SIZE)dolzina_branja);
        stanje   = izvor->VrniStanje();
    }

    dolzina_branja -= prebrano;
    return prebrano;
}


T_SIZE ROmejenTokPodatkov::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);

    T_SIZE zapisano;

    if (dolzina < dolzina_pisanja) {
        zapisano = izvor->Zapisi(podatki, dolzina);
        stanje   = izvor->VrniStanje();
    } else if (dolzina && !dolzina_pisanja) {
        zapisano = 0;
        stanje   = AOSN_NAPAKA_PISANJE;
    } else {
        zapisano = izvor->Zapisi(podatki, (T_SIZE)dolzina_pisanja);
        stanje   = izvor->VrniStanje();
    }

    dolzina_pisanja -= zapisano;
    return zapisano;
}


//////////////////////////////////////////////////////////////////////
// RMedpomnilnik
//////////////////////////////////////////////////////////////////////

RMedpomnilnik::RMedpomnilnik() : RDatoteka()
{
    izvor = NULL;
    medpomnilnik = NULL;
    dolzina = dolzina_bloka = 0;
    spremenjen = FALSE;
    zacetek = odmik = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif
}


RMedpomnilnik::RMedpomnilnik(RDatoteka *_izvor, T_SIZE _dolzina_bloka) : RDatoteka()
{
    izvor = NULL;
    medpomnilnik = NULL;
    dolzina = dolzina_bloka = 0;
    spremenjen = FALSE;
    zacetek = odmik = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif
    verify(USPESNO(Uporabi(_izvor, _dolzina_bloka)));
}


RMedpomnilnik::~RMedpomnilnik()
{
    if (izvor)
        Odpusti();

    assert(!izvor);
    assert(!medpomnilnik);
}


T_STANJE RMedpomnilnik::Uporabi(RDatoteka *_izvor, T_SIZE _dolzina_bloka)
{
    assert(_izvor);
    assert(!izvor);
    assert(!medpomnilnik);

    if (_dolzina_bloka) {
        medpomnilnik = aalloc(_dolzina_bloka);
        if (!medpomnilnik)
            return stanje = AOSN_NAPAKA_POMNILNIK;
    }

    izvor = _izvor;
    dolzina = 0;
    dolzina_bloka = _dolzina_bloka;
    spremenjen = FALSE;
    zacetek = 0;
    stanje = _izvor->stanje;
    odmik = USPESNO(stanje) ? _izvor->VrniOdmik() : 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    return stanje;
}


RDatoteka* RMedpomnilnik::Odpusti()
{
    assert(izvor);
    RDatoteka *stari_izvor = izvor;

    if (medpomnilnik) {
        // Najprej izplaknemo medpomnilnik, e je to treba.
        if (spremenjen && dolzina)
            verify(USPESNO(IzplakniMedpomnilnik()));

        // Sprostimo medpomnilnik.
        afree(medpomnilnik);
        medpomnilnik = NULL;

        // Preden odpustimo datoteko, uskladimo fizini in logini odmik v datoteki.
        izvor->Prestavi(odmik);
    }

    stanje = A_OK;
    izvor = NULL;
    dolzina = dolzina_bloka = 0;
    spremenjen = FALSE;
    zacetek = odmik = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    return stari_izvor;
}


T_SIZE RMedpomnilnik::Preberi(T_KAZALEC podatki, T_SIZE _dolzina)
{
    assert(izvor);
    assert(podatki || !_dolzina);

    if (medpomnilnik) {
        T_SIZE
            prebrano = 0,
            ostanek = _dolzina,
            ostanek_v_medpomnilniku;
        T_ODMIK
            konec;

        stanje = A_OK;

        for (;;) {
            konec = zacetek + dolzina;

            // Ali sta spodnja pogoja res ekvivalentna v vseh primerih? e se bo kdaj sproil spodnji assert, oitno nista in je zato Petru nagajala RMedFizDatoteka.
            assert((odmik - zacetek < dolzina) == (zacetek <= odmik && odmik < konec));

            // Ali je zadetek v medpomnilniku?
            if (zacetek <= odmik && odmik < konec) {
                ostanek_v_medpomnilniku = (T_SIZE)(konec - odmik);

                // Ali imamo dovolj podatkov v medpomnilniku?
                if (ostanek <= ostanek_v_medpomnilniku) {
                    amemcpy((T_U1*)podatki + prebrano, (T_U1*)medpomnilnik + (T_SIZE)(odmik - zacetek), ostanek);
                    prebrano += ostanek;
                    odmik    += ostanek;
                    stanje    = A_OK;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
                    datum_dostopa = RDatum::VrniTrenutniCas();
#endif
                    return prebrano;
                } else {
                    // Prepiemo, kar imamo v medpomnilniku.
                    amemcpy((T_U1*)podatki + prebrano, (T_U1*)medpomnilnik + (T_SIZE)(odmik - zacetek), ostanek_v_medpomnilniku);
                    prebrano += ostanek_v_medpomnilniku;
                    odmik    += ostanek_v_medpomnilniku;
                    ostanek  -= ostanek_v_medpomnilniku;
                }
            }

            // e je zahtevana koliina podatkov veja od kapacitete medpomnilnika, beremo podatke mimo medpomnilnika.
            if (ostanek > dolzina_bloka) {
                // Razveljavimo medpomnilnik.
                if (NEUSPESNO(RazveljaviMedpomnilnik()))
                    return prebrano;

                verify(izvor->Prestavi(odmik) == odmik);
                T_SIZE _prebrano = izvor->Preberi((T_U1*)podatki + prebrano, ostanek);
                prebrano += _prebrano;
                odmik    += _prebrano;
                stanje    = prebrano ? A_OK : izvor->stanje;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
                datum_dostopa = T_TRENUTEK_MIN;
#endif
                return prebrano;
            }

            // Napolnimo medpomnilnik.
            if (NEUSPESNO(ZamenjajMedpomnilnik(odmik)))
                return prebrano;

            konec = zacetek + dolzina;

            // e je prilo premalo podatkov, bomo predhodno konali branje.
            ostanek_v_medpomnilniku = (T_SIZE)(konec - odmik);
            if (dolzina < dolzina_bloka && ostanek_v_medpomnilniku < ostanek) {
                amemcpy((T_U1*)podatki + prebrano, (T_U1*)medpomnilnik + (T_SIZE)(odmik - zacetek), ostanek_v_medpomnilniku);
                prebrano += ostanek_v_medpomnilniku;
                odmik    += ostanek_v_medpomnilniku;
                stanje    = prebrano ? A_OK : AOSN_NAPAKA_KONEC_TOKA;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
                datum_dostopa = RDatum::VrniTrenutniCas();
#endif
                return prebrano;
            }
        }
    } else {
        T_SIZE prebrano = izvor->Preberi(podatki, _dolzina);
        odmik += prebrano;
        stanje = izvor->stanje;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
        datum_dostopa = T_TRENUTEK_MIN;
#endif
        return prebrano;
    }
}


T_SIZE RMedpomnilnik::Zapisi(TK_KAZALEC podatki, T_SIZE _dolzina)
{
    assert(izvor);
    assert(podatki || !_dolzina);

    if (medpomnilnik) {
        T_SIZE
            zapisano = 0,
            ostanek = _dolzina,
            ostanek_v_medpomnilniku;

        stanje = A_OK;

        if (!dolzina) zacetek = odmik;

        for (;;) {
            // Ali je zadetek v medpomnilniku? (Dovoljujemo dodajanje na konec medpomnilnika)
            if (zacetek <= odmik && odmik <= zacetek + dolzina) {
                ostanek_v_medpomnilniku = (T_SIZE)(zacetek + dolzina_bloka - odmik);

                // Ali imamo dovolj prostora v medpomnilniku?
                if (ostanek <= ostanek_v_medpomnilniku) {
                    amemcpy((T_U1*)medpomnilnik + (T_SIZE)(odmik - zacetek), (T_U1*)podatki + zapisano, ostanek);
                    zapisano  += ostanek;
                    odmik     += ostanek;
                    spremenjen = TRUE;
                    dolzina    = amax(dolzina, (T_SIZE)(odmik - zacetek));
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
                    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif
                    return zapisano;
                } else {
                    // Prepiemo, v medpomnilnik, kolikor lahko.
                    amemcpy((T_U1*)medpomnilnik + (T_SIZE)(odmik - zacetek), (T_U1*)podatki + zapisano, ostanek_v_medpomnilniku);
                    zapisano  += ostanek_v_medpomnilniku;
                    odmik     += ostanek_v_medpomnilniku;
                    ostanek   -= ostanek_v_medpomnilniku;
                    spremenjen = TRUE;
                    dolzina    = dolzina_bloka;
                }
            }

            // e je zahtevana koliina podatkov veja od kapacitete medpomnilnika, piemo podatke mimo medpomnilnika.
            if (ostanek > dolzina_bloka) {
                // Razveljavimo medpomnilnik.
                if (NEUSPESNO(RazveljaviMedpomnilnik()))
                    return zapisano;

                verify(izvor->Prestavi(odmik) == odmik);
                T_SIZE _zapisano = izvor->Zapisi((T_U1*)podatki + zapisano, ostanek);
                zapisano += _zapisano;
                odmik    += _zapisano;
                stanje    = izvor->stanje;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
                datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif
                return zapisano;
            }

            // Napolnimo medpomnilnik.
            if (NEUSPESNO(ZamenjajMedpomnilnik(odmik)))
                return zapisano;
        }
    } else {
        T_SIZE zapisano = izvor->Zapisi(podatki, _dolzina);
        odmik += zapisano;
        stanje = izvor->stanje;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
        datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif
        return zapisano;
    }
}


T_STANJE RMedpomnilnik::Zapri()
{
    return izvor ? Odpusti()->Zapri() : stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RMedpomnilnik::Izplakni()
{
    assert(izvor);

    // Najprej izplaknemo medpomnilnik, e je to treba.
    if (spremenjen && dolzina)
        if (NEUSPESNO(IzplakniMedpomnilnik()))
            return stanje;

#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    // Sedaj e fizino izplaknemo.
    return USPESNO(stanje) ? stanje = izvor->Izplakni() : stanje;
}


T_ODMIK RMedpomnilnik::Prestavi(T_RODMIK _odmik, T_DATNACINODMIKA kako)
{
    if (medpomnilnik) {
        stanje = A_OK;
        switch (kako) {
        case RDAT_OD_ZACETKA:
            return odmik  = _odmik;
        case RDAT_OD_TRENUTNEGA:
            return odmik += _odmik;
        case RDAT_OD_KONCA:
            return odmik  = VrniDolzino() + _odmik;
        default:
            assert(0);
            return T_ODMIK_MAX;
        }
    } else {
        assert(izvor);

        odmik  = izvor->Prestavi(_odmik, kako);
        stanje = izvor->VrniStanje();
        return odmik;
    }
}


T_ODMIK RMedpomnilnik::VrniOdmik()
{
    stanje = A_OK;
    return odmik;
}


T_STANJE RMedpomnilnik::ZakleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    assert(izvor);

    return stanje = izvor->ZakleniObmocje(odmik, dolzina);
}


T_STANJE RMedpomnilnik::OdkleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    assert(izvor);

    return stanje = izvor->OdkleniObmocje(odmik, dolzina);
}


T_DOLZINA RMedpomnilnik::VrniDolzino() const
{
    assert(izvor);

    return amax(izvor->VrniDolzino(), zacetek + dolzina);
}


T_STANJE RMedpomnilnik::PostaviKonec()
{
    assert(izvor);

    // e uporabljamo medpomnilnik, se logini in fizini odmik lahko razlikujeta, zato ju sedaj uskladimo.
    if (medpomnilnik)
        izvor->Prestavi(odmik);

#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    if (zacetek + dolzina <= odmik) {
        // Rezanje ne vpliva na medpomnilnik.
    } else if (zacetek <= odmik) {
        // Rezanje skraja medpomnilnik.
        dolzina = (T_SIZE)(odmik - zacetek);
    } else {
        // Rezanje odpravi medpomnilnik.
        zacetek = 0;
        dolzina = 0;
        spremenjen = FALSE;
    }

    return stanje = izvor->PostaviKonec();
}


T_TRENUTEK RMedpomnilnik::VrniDatumStvaritve() const
{
    return izvor ? izvor->VrniDatumStvaritve() : T_TRENUTEK_MIN;
}


T_TRENUTEK RMedpomnilnik::VrniDatumDostopa() const
{
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    return izvor ? amax(datum_dostopa, izvor->VrniDatumDostopa()) : T_TRENUTEK_MIN;
#else
    return izvor ? izvor->VrniDatumDostopa() : T_TRENUTEK_MIN;
#endif
}


T_TRENUTEK RMedpomnilnik::VrniDatumPisanja() const
{
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    return izvor ? amax(datum_pisanja, izvor->VrniDatumPisanja()) : T_TRENUTEK_MIN;
#else
    return izvor ? izvor->VrniDatumPisanja() : T_TRENUTEK_MIN;
#endif
}


T_STANJE RMedpomnilnik::NastaviDatumStvaritve(T_TRENUTEK datum)
{
    return stanje = izvor ? izvor->NastaviDatumStvaritve(datum) : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


T_STANJE RMedpomnilnik::NastaviDatumDostopa(T_TRENUTEK datum)
{
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum;
#endif
    return stanje = izvor ? izvor->NastaviDatumDostopa(datum) : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


T_STANJE RMedpomnilnik::NastaviDatumPisanja(T_TRENUTEK datum)
{
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_pisanja = datum;
#endif
    return stanje = izvor ? izvor->NastaviDatumPisanja(datum) : AOSN_NAPAKA_NI_VZPOSTAVLJEN;
}


inline T_STANJE RMedpomnilnik::IzplakniMedpomnilnik()
{
    assert(izvor);
    assert(spremenjen && dolzina);

    // Zapiemo vsebino medpomnilnika.
    izvor->Prestavi(zacetek);
    izvor->Zapisi(medpomnilnik, dolzina);
    spremenjen = FALSE;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    return stanje = izvor->stanje;
}


inline T_STANJE RMedpomnilnik::RazveljaviMedpomnilnik()
{
    // Najprej izplaknemo medpomnilnik, e je to treba.
    if (spremenjen && dolzina)
        if (NEUSPESNO(IzplakniMedpomnilnik()))
            return stanje;

    zacetek = 0;
    dolzina = 0;

    return stanje = A_OK;
}


inline T_STANJE RMedpomnilnik::ZamenjajMedpomnilnik(T_ODMIK _zacetek)
{
    assert(izvor);

    // Najprej izplaknemo medpomnilnik, e je to treba.
    if (spremenjen && dolzina)
        if (NEUSPESNO(IzplakniMedpomnilnik()))
            return stanje;

    // Preberemo novi blok.
    verify(izvor->Prestavi(zacetek = _zacetek) == _zacetek);
    dolzina = izvor->Preberi(medpomnilnik, dolzina_bloka);
    spremenjen = FALSE;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = T_TRENUTEK_MIN;
#endif

    return stanje = A_OK; // e branje ni uspelo vseeno ponaredimo stanje OK, saj bo dolina 0.
}


//////////////////////////////////////////////////////////////////////
// RSisTokPodatkov
//////////////////////////////////////////////////////////////////////

RSisTokPodatkov::RSisTokPodatkov(T_ROCICAOS rocica) :
    RTokPodatkov(),
    RSisElement(rocica)
{
}


RSisTokPodatkov::~RSisTokPodatkov()
{
    // Sistemski tok odpustimo (e ga se nismo prej).
    if (JE_VELJAVNA_ROCICA(rocica))
        Odpusti();
}


T_STANJE RSisTokPodatkov::Uporabi(T_ROCICAOS rocica)
{
    return stanje = RSisElement::Uporabi(rocica);
}


T_ROCICAOS RSisTokPodatkov::Odpusti()
{
    stanje = A_OK;
    return RSisElement::Odpusti();
}


T_SIZE RSisTokPodatkov::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    // e imamo kaj brati, gremo brati.
    if (dolzina) {
        DWORD koliko;
        BOOL je_uspesno;

        // Funkcija ReadFile() lahko sproi izjemo (npr. STATUS_FILE_BAD_FORMAT/0xE0000002).
        // Ker ne elimo, da nam to zaustavi izvajanje programa, izjemo prestreemo in pretvorimo v napaen status.
        __try {
            je_uspesno = ::ReadFile(rocica, podatki, (DWORD)dolzina, &koliko, NULL);
        } __except (EXCEPTION_EXECUTE_HANDLER) {
            // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko prebranih.
            // Zato nastavimo stanje napake in vrnemo, da nismo uspeli prebrati niesar, saj podatki v pomnilniku niso zanesljivi.
            stanje = AOSN_NAPAKA_BRANJE_IZJEMA;
            return 0;
        }

        if (je_uspesno) {
            stanje = koliko ? A_OK : AOSN_NAPAKA_KONEC_TOKA;
        } else {
            DWORD rezultat = ::GetLastError();
            if (rezultat == ERROR_NO_SYSTEM_RESOURCES) {
                // Prilo je do napake 1450 "Insufficient system resources exist to complete the requested service.".
                // Ta napaka se pojavi, kadar elimo naenkrat prebrati prevelik blok pomnilnika (npr. preko \\TSClient), zato preberemo po blokih.
                T_SIZE prebrano = 0;

                for (stanje = A_OK; dolzina;) {
                    __try {
                        je_uspesno = ::ReadFile(rocica, podatki, amin<DWORD>((DWORD)dolzina, RDAT_VELIKOST_BLOKA), &koliko, NULL);
                    } __except (EXCEPTION_EXECUTE_HANDLER) {
                        // Prilo je do izjeme in nimamo pojma, koliko novih podatkov je bilo dejansko prebranih.
                        // Zato nastavimo stanje napake in vrnemo le, kolikor smo do zdaj uspeli prebrati, saj podatki v pomnilniku za tem niso zanesljivi.
                        stanje = AOSN_NAPAKA_BRANJE_IZJEMA;
                        return prebrano;
                    }

                    if (je_uspesno) {
                        if (koliko) {
                            podatki   = ((T_U1*)podatki) + koliko;
                            prebrano += koliko;
                            dolzina  -= koliko;
                        } else {
                            stanje = AOSN_NAPAKA_KONEC_TOKA;
                            break;
                        }
                    } else {
                        stanje = astanje(::GetLastError(), AOSN_NAPAKA_BRANJE);
                        break;
                    }
                }

                return prebrano;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        }

        return koliko;
    } else {
        // Ni potrebe po branju, javimo, da je vse v redu.
        stanje = A_OK;
        return 0;
    }
}


T_SIZE RSisTokPodatkov::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD koliko;
    BOOL je_uspesno;

    // Funkcija ReadFile() lahko sproi izjemo (npr. STATUS_FILE_BAD_FORMAT/0xE0000002). Zato smo previdni tudi pri funkciji WriteFile().
    // Ker ne elimo, da nam pisanje zaustavi izvajanje programa, izjemo prestreemo in pretvorimo v napaen status.
    __try {
        je_uspesno = ::WriteFile(rocica, podatki, (DWORD)dolzina, &koliko, NULL);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko zapisanih.
        // Zato nastavimo stanje napake in vrnemo, da nismo uspeli zapisati niesar.
        // eprav smo mogoe kaj podatkov vendarle zapisali, nimamo o koliini nobene oprijemljive ocene.
        stanje = AOSN_NAPAKA_PISANJE_IZJEMA;
        return 0;
    }

    stanje = je_uspesno ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PISANJE);
    return koliko;
}


T_STANJE RSisTokPodatkov::Zapri()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    stanje = RSisElement::Zapri();
    return stanje;
}


T_STANJE RSisTokPodatkov::Izplakni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    return stanje = FlushFileBuffers(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_IZPLAKNI);
}


T_UINT RSisTokPodatkov::VrniAtribute()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    BY_HANDLE_FILE_INFORMATION a;

    return ::GetFileInformationByHandle(rocica, &a) ? a.dwFileAttributes : 0;
}


T_STANJE RSisTokPodatkov::VrniAtribute(T_UINT *atributi, T_TRENUTEK *cas_stvaritve, T_TRENUTEK *cas_dostopa, T_TRENUTEK *cas_pisanja, T_DOLZINA *dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    BY_HANDLE_FILE_INFORMATION a;

    if (::GetFileInformationByHandle(rocica, &a)) {
        if (atributi)       *atributi       = a.dwFileAttributes;
        if (cas_stvaritve)  *cas_stvaritve  = RDatum::Pretvori(&(a.ftCreationTime));
        if (cas_dostopa)    *cas_dostopa    = RDatum::Pretvori(&(a.ftLastAccessTime));
        if (cas_pisanja)    *cas_pisanja    = RDatum::Pretvori(&(a.ftLastWriteTime));
        if (dolzina)        *dolzina        = ((T_DOLZINA)a.nFileSizeHigh << 32) + a.nFileSizeLow;
        return A_OK;
    } else {
        if (atributi)       *atributi       = 0;
        if (cas_stvaritve)  *cas_stvaritve  = 0;
        if (cas_dostopa)    *cas_dostopa    = 0;
        if (cas_pisanja)    *cas_pisanja    = 0;
        if (dolzina)        *dolzina        = 0;
        return astanje(::GetLastError(), AOSN_NAPAKA_MANJKAJO_ATRIBUTI);
    }
}


T_UINT RSisTokPodatkov::VrniAtribute(TK_NIZ ime)
{
    assert(ime);

    WIN32_FILE_ATTRIBUTE_DATA a;

    return ::GetFileAttributesEx(ime, GetFileExInfoStandard, &a) ? a.dwFileAttributes : 0;
}


T_STANJE RSisTokPodatkov::VrniAtribute(TK_NIZ ime, T_UINT *atributi, T_TRENUTEK *cas_stvaritve, T_TRENUTEK *cas_dostopa, T_TRENUTEK *cas_pisanja, T_DOLZINA *dolzina)
{
    assert(ime);

    WIN32_FILE_ATTRIBUTE_DATA a;

    if (::GetFileAttributesEx(ime, GetFileExInfoStandard, &a)) {
        if (atributi)       *atributi       = a.dwFileAttributes;
        if (cas_stvaritve)  *cas_stvaritve  = RDatum::Pretvori(&(a.ftCreationTime));
        if (cas_dostopa)    *cas_dostopa    = RDatum::Pretvori(&(a.ftLastAccessTime));
        if (cas_pisanja)    *cas_pisanja    = RDatum::Pretvori(&(a.ftLastWriteTime));
        if (dolzina)        *dolzina        = ((T_DOLZINA)a.nFileSizeHigh << 32) + a.nFileSizeLow;
        return A_OK;
    } else {
        if (atributi)       *atributi       = 0;
        if (cas_stvaritve)  *cas_stvaritve  = 0;
        if (cas_dostopa)    *cas_dostopa    = 0;
        if (cas_pisanja)    *cas_pisanja    = 0;
        if (dolzina)        *dolzina        = 0;
        return astanje(::GetLastError(), AOSN_NAPAKA_MANJKAJO_ATRIBUTI);
    }
}


T_STANJE RSisTokPodatkov::NastaviAtribute(TK_NIZ ime, T_UINT maska, T_UINT atributi)
{
    assert(ime);

    // Za doloene atribute so potrebni posebni prijemi.
    if (maska & (RDAT_AT_KOMPRESIRAN | RDAT_AT_KRIPTIRAN | RDAT_AT_REDEK)) {
        RFizDatoteka dat(ime, 0);

        if (USPESNO(dat.stanje)) {
#ifdef FSCTL_SET_COMPRESSION
            // e elimo datoteko (za|od)kompresirati, moramo uporabiti sistemsko funkcijo DeviceIoControl(FSCTL_SET_COMPRESSION).
            if (maska & RDAT_AT_KOMPRESIRAN) {
                USHORT algoritem = atributi & RDAT_AT_KOMPRESIRAN ? (USHORT)COMPRESSION_FORMAT_DEFAULT : (USHORT)COMPRESSION_FORMAT_NONE;
                DWORD velikost;
                ::DeviceIoControl(dat, FSCTL_SET_COMPRESSION, &algoritem, sizeof(USHORT), NULL, 0, &velikost, NULL);
            }
#endif

            /* Skomentiran klic sistemske funkcije, ker ne obstaja na platformi Windows 95!!!
#ifdef FILE_ATTRIBUTE_ENCRYPTED
            // e elimo (za|od)kriptirati datoteko, moramo uporabiti sistemski funkciji EncryptFile() in DecryptFile().
            if (maska & RDAT_AT_KRIPTIRAN) {
                if (atributi & RDAT_AT_KRIPTIRAN)
                    ::EncryptFile(ime);
                else
                    ::DecryptFile(ime, 0);
            }
#endif
            */

#if defined(FILE_ATTRIBUTE_SPARSE_FILE) && defined (FSCTL_SET_SPARSE)
            // e elimo nastaviti redko datoteko, moramo uporabiti sistemsko funkcijo DeviceIoControl(FSCTL_SET_SPARSE).
            // Redko datoteko lahko le ustvarimo, ne moremo pa je vrniti v prejnje stanje!!!
            if (maska & atributi & RDAT_AT_REDEK) {
                DWORD velikost;
                ::DeviceIoControl(dat, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &velikost, NULL);
            }
#endif
        }
    }

    // Dodatno maskiramo atribute, ki jih s SetFileAttributes() ne moremo nastavljati.
    maska &= ~(RDAT_AT_KOMPRESIRAN | RDAT_AT_KRIPTIRAN | RDAT_AT_REDEK | RDAT_AT_IMENIK | RDAT_AT_SISTEMSKI);

    // Za atribute, ki jih maskiramo, preberemo staro vrednost, za nemaskirane uporabimo tisto, ki jo je doloil uporabnik.
    atributi = maska & atributi | ~maska & VrniAtribute(ime);

    // Nastavimo atribute.
    return ::SetFileAttributes(ime, (DWORD)atributi) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVLJANJE_ATR);
}


//////////////////////////////////////////////////////////////////////
// RIzrSisTokPodatkov
//////////////////////////////////////////////////////////////////////

RIzrSisTokPodatkov::RIzrSisTokPodatkov(T_ROCICAOS rocica, T_SIZE dolzina_branja, T_SIZE dolzina_pisanja) : RIzravnalnik(NULL, dolzina_branja, dolzina_pisanja)
{
    verify(USPESNO(izvor.Uporabi(rocica)));
    verify(USPESNO(RIzravnalnik::Uporabi(&izvor)));
}


RIzrSisTokPodatkov::~RIzrSisTokPodatkov()
{
    RIzravnalnik::Odpusti();
    izvor.Odpusti();
}


//////////////////////////////////////////////////////////////////////
// RISequentialStream
//////////////////////////////////////////////////////////////////////

RISequentialStream::RISequentialStream() :
    RTokPodatkov(),
    izvor(NULL)
{
}


RISequentialStream::~RISequentialStream()
{
    assert(!izvor);
}


T_STANJE RISequentialStream::Uporabi(ISequentialStream *_izvor)
{
    assert(_izvor);
    assert(!izvor);

    izvor = _izvor;
    izvor->AddRef();
    return stanje = A_OK;
}


ISequentialStream* RISequentialStream::Odpusti()
{
    assert(izvor);
    ISequentialStream *stari_izvor = izvor;

    stanje = A_OK;
    izvor->Release();
    izvor = NULL;
    return stari_izvor;
}


T_SIZE RISequentialStream::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);
    assert(dolzina <= ULONG_MAX);

    HRESULT hr;
    ULONG prebrano = 0;

    __try {
        hr = izvor->Read(podatki, (ULONG)dolzina, &prebrano);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko prebranih.
        // Zato nastavimo stanje napake in vrnemo, da nismo uspeli prebrati niesar, saj podatki v pomnilniku niso zanesljivi.
        stanje = AOSN_NAPAKA_BRANJE_IZJEMA;
        return 0;
    }

    if (hr == S_FALSE) {
        stanje = AOSN_NAPAKA_KONEC_TOKA;
        return prebrano;
    } else if (SUCCEEDED(hr)) {
        stanje = A_OK;
        return prebrano;
    } else {
        stanje = AOSN_NAPAKA_BRANJE;
        return 0;
    }
}


T_SIZE RISequentialStream::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(izvor);
    assert(podatki || !dolzina);
    assert(dolzina <= ULONG_MAX);

    HRESULT hr;
    ULONG zapisano = 0;

    __try {
        hr = izvor->Write(podatki, (ULONG)dolzina, &zapisano);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko zapisanih.
        // Zato nastavimo stanje napake in vrnemo, da nismo uspeli zapisati niesar.
        stanje = AOSN_NAPAKA_PISANJE_IZJEMA;
        return 0;
    }

    if (SUCCEEDED(hr)) {
        stanje = A_OK;
        return zapisano;
    } else {
        stanje = AOSN_NAPAKA_PISANJE;
        return 0;
    }
}


T_STANJE RISequentialStream::Zapri()
{
    if (izvor) {
        // ISequentialStream ne omogoa zapiranja.
        // Odpustimo tok iz uporabe in upamo, da se bo sam zaprl, ko bo njegov tevec referenc padel na 0.
        izvor->Release();
        izvor = NULL;
        return stanje = A_OK;
    } else
        return stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RISequentialStream::Izplakni()
{
    assert(izvor);

    // ISequentialStream ne omogoa izplakovanja.
    return stanje = A_OK;
}


//////////////////////////////////////////////////////////////////////
// RPovezavaASP
//////////////////////////////////////////////////////////////////////

#ifndef WIN32_LEAN_AND_MEAN

RPovezavaASP::RPovezavaASP() :
    RTokPodatkov(),
    zahteva(NULL),
    odgovor(NULL)
{
}


RPovezavaASP::~RPovezavaASP()
{
    assert(!zahteva);
    assert(!odgovor);
}


T_STANJE RPovezavaASP::UporabiZahtevo(IRequest* _zahteva)
{
    assert(_zahteva);
    assert(!zahteva);

    zahteva = _zahteva;
    zahteva->AddRef();
    return stanje = A_OK;
}


IRequest* RPovezavaASP::OdpustiZahtevo()
{
    assert(zahteva);
    IRequest *_zahteva = zahteva;

    stanje = A_OK;
    zahteva->Release();
    zahteva = NULL;
    return _zahteva;
}


T_STANJE RPovezavaASP::UporabiOdgovor(IResponse* _odgovor)
{
    assert(_odgovor);
    assert(!odgovor);

    odgovor = _odgovor;
    odgovor->AddRef();
    return stanje = A_OK;
}


IResponse* RPovezavaASP::OdpustiOdgovor()
{
    assert(odgovor);
    IResponse *_odgovor = odgovor;

    stanje = A_OK;
    odgovor->Release();
    odgovor = NULL;
    return _odgovor;
}


T_SIZE RPovezavaASP::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    VARIANT var_dolzina, var;
    HRESULT hr;

    if (!zahteva) {
        stanje = AOSN_NAPAKA_NI_VZPOSTAVLJEN;
        return 0;
    }

    // Preberemo blok podatkov od odjemalca.
    V_VT(&var_dolzina) = VT_I4;
    V_I4(&var_dolzina) = (LONG)dolzina;
    V_VT(&var) = VT_EMPTY;

    __try {
        hr = zahteva->BinaryRead(&var_dolzina, &var);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko prebranih.
        // Zato nastavimo stanje napake in vrnemo, da nismo uspeli prebrati niesar, saj podatki v pomnilniku niso zanesljivi.
        // Pomnilnika ne sproamo, ker ne vemo, v kaknem stanju je. Raje memory-leak, kot memory-corruption.
        stanje = AOSN_NAPAKA_BRANJE_IZJEMA;
        return 0;
    }

    if (SUCCEEDED(hr)) {
        assert(V_VT(&var_dolzina) == VT_I4 && V_I4(&var_dolzina) <= (LONG)dolzina);
        assert(V_VT(&var) == (VT_ARRAY | VT_UI1));
        void *_podatki;

        if (V_I4(&var_dolzina)) {
            // Prepiemo podatke in sprostimo pomnilnik.
            verify(SUCCEEDED(::SafeArrayAccessData(V_ARRAY(&var), &_podatki)));
            memcpy(podatki, _podatki, V_I4(&var_dolzina));
            verify(SUCCEEDED(::SafeArrayUnaccessData(V_ARRAY(&var))));
            stanje = A_OK;
        } else
            stanje = dolzina ? AOSN_NAPAKA_KONEC_TOKA : A_OK;
        verify(SUCCEEDED(::SafeArrayDestroy(V_ARRAY(&var))));

        return V_I4(&var_dolzina);
    } else {
        stanje = AOSN_NAPAKA_BRANJE;
        return 0;
    }
}


T_SIZE RPovezavaASP::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    VARIANT var;
    HRESULT hr;

    if (!odgovor) {
        stanje = AOSN_NAPAKA_NI_VZPOSTAVLJEN;
        return 0;
    }

    // Poljemo podatkovni blok odjemalcu.
    V_VT(&var) = VT_BSTR;
    V_BSTR(&var) = SysAllocStringByteLen((LPCSTR)podatki, (UINT)dolzina);
    __try {
        hr = odgovor->BinaryWrite(var);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        // Prilo je do izjeme in nimamo pojma, koliko podatkov je bilo dejansko zapisanih.
        // Zato poistimo pomnilnik, nastavimo stanje napake in vrnemo, da nismo uspeli zapisati niesar.
        SysFreeString(V_BSTR(&var));
        stanje = AOSN_NAPAKA_PISANJE_IZJEMA;
        return 0;
    }
    SysFreeString(V_BSTR(&var));

    if (SUCCEEDED(hr)) {
        stanje = A_OK;
        return dolzina;
    } else {
        stanje = AOSN_NAPAKA_PISANJE;
        return 0;
    }
}


T_STANJE RPovezavaASP::Zapri()
{
    if (odgovor) {
        // Poskusimo zapreti odgovor. Napako pri zapiranju ignoriramo.
        __try {
            SUCCEEDED(odgovor->End());
        } __except (EXCEPTION_EXECUTE_HANDLER) {}

        // Odpustimo odgovor iz uporabe.
        odgovor->Release();
        odgovor = NULL;
    }

    if (zahteva) {
        // Odpustimo zahtevo iz uporabe.
        zahteva->Release();
        zahteva = NULL;
    }

    return stanje = A_OK;
}


T_STANJE RPovezavaASP::Izplakni()
{
    assert(odgovor);

    HRESULT hr;

    // Previdno izplaknemo medpomnilnik.
    __try {
        hr = odgovor->Flush();
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        return AOSN_NAPAKA_IZPLAKNI;
    }

    return stanje = SUCCEEDED(hr) ? A_OK : AOSN_NAPAKA_IZPLAKNI;
}

#endif // WIN32_LEAN_AND_MEAN


//////////////////////////////////////////////////////////////////////
// RFizDatoteka
//////////////////////////////////////////////////////////////////////

RFizDatoteka::RFizDatoteka(T_ROCICAOS rocica) :
    RDatoteka(),
    RSisTokPodatkov(rocica)
{
}


RFizDatoteka::RFizDatoteka(TK_NIZ ime, T_UINT nacin) :
    RDatoteka(),
    RSisTokPodatkov()
{
    Odpri(ime, nacin);
}


RFizDatoteka::~RFizDatoteka()
{
    // Datoteko zapremo (e je e nismo prej).
    if (JE_VELJAVNA_ROCICA(rocica))
        verify(USPESNO(Zapri()));
}


T_STANJE RFizDatoteka::Odpri(TK_NIZ ime, T_UINT nacin)
{
    // Datoteka ne sme biti e odprta.
    assert(NI_VELJAVNA_ROCICA(rocica));

    T_U4
        dostop = 0,
        deljenje = 0,
        kako,
        atributi = FILE_ATTRIBUTE_NORMAL;
    SECURITY_ATTRIBUTES varnostne_nastavitve;

    // Nastavimo nain odpiranja.
    if (nacin & RDAT_BRANJE)      dostop |= GENERIC_READ;
    if (nacin & RDAT_PISANJE)     dostop |= GENERIC_WRITE;
    if (nacin & RDAT_UPRAVLJANJE) dostop |= FILE_WRITE_ATTRIBUTES;

    // Nastavimo nain skupne rabe datotek.
    if (nacin & RDAT_SH_BERI)  deljenje |= FILE_SHARE_READ;
    if (nacin & RDAT_SH_PISI)  deljenje |= FILE_SHARE_WRITE;
    if (nacin & RDAT_SH_BRISI) deljenje |= FILE_SHARE_DELETE;

    // Nastavimo varnostne nastavitve.
    varnostne_nastavitve.nLength = sizeof(SECURITY_ATTRIBUTES);
    varnostne_nastavitve.lpSecurityDescriptor = NULL;
    varnostne_nastavitve.bInheritHandle = nacin & RDAT_NE_DEDUJ ? FALSE : TRUE;

    // Nastavimo ali elimo datoteko brezpogojno ustvariti, odpreti obstojeo ...
    if (nacin & RDAT_USTVARI)
        kako = nacin & RDAT_NE_UNICI ? OPEN_ALWAYS : CREATE_ALWAYS;
    else
        kako = OPEN_EXISTING;

    // Nastavimo atribute in nain predpomnenja.
    if (nacin & RDAT_PISI_SKOZI)       atributi |= FILE_FLAG_WRITE_THROUGH;
    if (nacin & RDAT_BREZ_MEDPOMN)     atributi |= FILE_FLAG_NO_BUFFERING;
    if (nacin & RDAT_DOSTOP_NAKLJUCEN) atributi |= FILE_FLAG_RANDOM_ACCESS;
    if (nacin & RDAT_DOSTOP_ZAPOREDEN) atributi |= FILE_FLAG_SEQUENTIAL_SCAN;
    if (nacin & RDAT_ZA_ARHIVIRANJE)   atributi |= FILE_FLAG_BACKUP_SEMANTICS;

    // Naredimo!!!
    rocica = CreateFile(ime, dostop, deljenje, &varnostne_nastavitve, kako, atributi, NULL);

    if (JE_VELJAVNA_ROCICA(rocica)) {
        // e je v igri operacija "Prilepi", se postavimo na konec datoteke.
        if (nacin & RDAT_PRILEPI) {
            Prestavi(0, RDAT_OD_KONCA);
            if (NEUSPESNO(stanje))
                return stanje;
        }

        return stanje = A_OK;
    } else
        return stanje = astanje(::GetLastError(), AOSN_NAPAKA_ODPIRANJE);
}


T_ODMIK RFizDatoteka::Prestavi(T_RODMIK odmik, T_DATNACINODMIKA kako)
{
    if (JE_VELJAVNA_ROCICA(rocica)) {
        LARGE_INTEGER li;

        li.QuadPart = odmik;
        li.LowPart = SetFilePointer(rocica, li.LowPart, &li.HighPart, kako);

        if (li.LowPart == 0xFFFFFFFF && GetLastError() != NO_ERROR) {
            li.QuadPart = -1;
            stanje = AOSN_NAPAKA_PREMIK;
        } else
            stanje = A_OK;

        return li.QuadPart;
    } else {
        stanje = AOSN_NAPAKA_ROCICA;
        return 0;
    }
}


T_ODMIK RFizDatoteka::VrniOdmik()
{
    if (JE_VELJAVNA_ROCICA(rocica)) {
        LARGE_INTEGER li;

        li.QuadPart = 0;
        li.LowPart = SetFilePointer(rocica, 0, &li.HighPart, FILE_CURRENT);

        if (li.LowPart == 0xFFFFFFFF && GetLastError() != NO_ERROR) {
            li.QuadPart = -1;
            stanje = AOSN_NAPAKA_PREMIK;
        } else
            stanje = A_OK;

        return li.QuadPart;
    } else {
        stanje = AOSN_NAPAKA_ROCICA;
        return 0;
    }
}


T_STANJE RFizDatoteka::ZakleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LARGE_INTEGER li_odmik;
    LARGE_INTEGER li_dolzina;

    li_odmik.QuadPart = odmik;
    li_dolzina.QuadPart = dolzina;

    return stanje = LockFile(rocica, li_odmik.LowPart, li_odmik.HighPart, li_dolzina.LowPart, li_dolzina.HighPart) ? A_OK : AOSN_NAPAKA_ZAKLEP;
}


T_STANJE RFizDatoteka::OdkleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LARGE_INTEGER li_odmik;
    LARGE_INTEGER li_dolzina;

    li_odmik.QuadPart = odmik;
    li_dolzina.QuadPart = dolzina;

    return stanje = UnlockFile(rocica, li_odmik.LowPart, li_odmik.HighPart, li_dolzina.LowPart, li_dolzina.HighPart) ? A_OK : AOSN_NAPAKA_ODKLEP;
}


T_DOLZINA RFizDatoteka::VrniDolzino() const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LARGE_INTEGER li;

    li.LowPart = GetFileSize(rocica, (LPDWORD) &li.HighPart);

    if (li.LowPart == 0xFFFFFFFF && GetLastError() != NO_ERROR) {
        li.QuadPart = -1;
        // stanje = AOSN_NAPAKA_DOLZINA;
    }

    return li.QuadPart;
}


T_STANJE RFizDatoteka::PostaviKonec()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    return stanje = SetEndOfFile(rocica) ? A_OK : AOSN_NAPAKA_ODREZI;
}


T_TRENUTEK RFizDatoteka::VrniDatumStvaritve() const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    verify(GetFileTime(rocica, &cas, NULL, NULL));
    return RDatum::Pretvori(&cas);
}


T_TRENUTEK RFizDatoteka::VrniDatumDostopa() const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    verify(GetFileTime(rocica, NULL, &cas, NULL));
    return RDatum::Pretvori(&cas);
}


T_TRENUTEK RFizDatoteka::VrniDatumPisanja() const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    verify(GetFileTime(rocica, NULL, NULL, &cas));
    return RDatum::Pretvori(&cas);
}


T_STANJE RFizDatoteka::NastaviDatumStvaritve(T_TRENUTEK datum)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    RDatum::Pretvori(datum, &cas);
    return ::SetFileTime(rocica, &cas, NULL, NULL) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVITEV_CASA);
}


T_STANJE RFizDatoteka::NastaviDatumDostopa(T_TRENUTEK datum)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    RDatum::Pretvori(datum, &cas);
    return ::SetFileTime(rocica, NULL, &cas, NULL) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVITEV_CASA);
}


T_STANJE RFizDatoteka::NastaviDatumPisanja(T_TRENUTEK datum)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    FILETIME cas;

    RDatum::Pretvori(datum, &cas);
    return ::SetFileTime(rocica, NULL, NULL, &cas) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVITEV_CASA);
}


//////////////////////////////////////////////////////////////////////
// RMedFizDatoteka
//////////////////////////////////////////////////////////////////////

RMedFizDatoteka::RMedFizDatoteka(T_ROCICAOS rocica, T_SIZE dolzina_bloka) : izvor(rocica)
{
    RMedpomnilnik::Uporabi(&izvor, dolzina_bloka);
}


RMedFizDatoteka::RMedFizDatoteka(TK_NIZ ime, T_UINT nacin, T_SIZE dolzina_bloka) : izvor(ime, nacin & RDAT_PISANJE ? nacin | RDAT_BRANJE : nacin)
{
    RMedpomnilnik::Uporabi(&izvor, dolzina_bloka);
}


RMedFizDatoteka::~RMedFizDatoteka()
{
    RMedpomnilnik::Odpusti();

    // Datoteko zapremo (e je e nismo prej).
    if (JE_VELJAVNA_ROCICA((T_ROCICAOS)izvor))
        verify(USPESNO(izvor.Zapri()));
}


T_STANJE RMedFizDatoteka::Zapri()
{
    verify(USPESNO(RazveljaviMedpomnilnik()));
    odmik = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif
    return stanje = izvor.Zapri();
}


//////////////////////////////////////////////////////////////////////
// RPomDatoteka
//////////////////////////////////////////////////////////////////////

RPomDatoteka::RPomDatoteka() : RDatoteka()
{
    podatki = NULL;
    odmik = 0;
    dolzina = fiz_dolzina = 0;
    upravljaj = TRUE;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif
}


RPomDatoteka::RPomDatoteka(T_SIZE _dolzina) : RDatoteka()
{
    // Preverimo 4 GB mejo.
    assert(_dolzina <= T_SIZE_MAX);

    // Pripravimo pomnilnik za delo.
    podatki = (T_U1*)aalloc(fiz_dolzina = _dolzina);
    assert(podatki);

    odmik = 0;
    dolzina = 0;
    upravljaj = TRUE;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif
}


RPomDatoteka::RPomDatoteka(T_KAZALEC _podatki, T_SIZE _dolzina, T_BOOL _upravljaj) : RDatoteka()
{
    // Preverimo 4 GB mejo.
    assert(_dolzina <= T_SIZE_MAX);
    assert(_podatki || !_dolzina);

    // Uporabimo e pripravljene podatke.
    podatki = (T_U1*)_podatki;
    odmik = 0;
    dolzina = fiz_dolzina = _dolzina;
    upravljaj = _upravljaj;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif
}


RPomDatoteka::RPomDatoteka(TK_NIZ ime, T_UINT nacin) : RDatoteka()
{
    podatki = NULL;
    odmik = 0;
    dolzina = fiz_dolzina = 0;
    upravljaj = TRUE;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    Nalozi(ime, nacin);
}


RPomDatoteka::~RPomDatoteka()
{
    // Zapremo pomnilniko datoteko (e je e nismo prej).
    if (podatki)
        verify(USPESNO(Zapri()));

    assert(!podatki);
}


T_STANJE RPomDatoteka::RezervirajDolzino(T_SIZE zeljeno, T_BOOL tesno)
{
    if (zeljeno > fiz_dolzina || tesno && zeljeno < fiz_dolzina) {
        // Ali lahko upravljamo s pomnilnikom?
        if (upravljaj) {
            // Realociramo le, kadar to izrecno elimo (tesno), ali pa je protrebno podaljati dolino.
            fiz_dolzina = tesno ? zeljeno : ((zeljeno + (RDAT_VELIKOST_BLOKA - 1)) / RDAT_VELIKOST_BLOKA) * RDAT_VELIKOST_BLOKA;
            podatki = (T_U1*)arealloc(podatki, fiz_dolzina);
            if (!podatki && fiz_dolzina)
                return stanje = AOSN_NAPAKA_POMNILNIK;
        } else
            return stanje = AOSN_NAPAKA_NE_UPRAVLJAJ;
    }

    // e smo preve skrajali, popravimo tudi dolino.
    if (fiz_dolzina < dolzina)
        dolzina = fiz_dolzina;

    return stanje = A_OK;
}


T_STANJE RPomDatoteka::Uporabi(T_KAZALEC _podatki, T_SIZE _dolzina, T_BOOL _upravljaj)
{
    // Datoteka mora biti zaprta.
    assert(podatki == NULL);
    assert(fiz_dolzina == 0);
    assert(dolzina == 0);
    assert(_podatki || !_dolzina);

    podatki = (T_U1*)_podatki;
    dolzina = fiz_dolzina = _dolzina;
    upravljaj = _upravljaj;
    odmik = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return stanje = A_OK;
}


T_STANJE RPomDatoteka::Uporabi(LPSAFEARRAY sa)
{
    // Datoteka mora biti zaprta.
    assert(podatki == NULL);
    assert(fiz_dolzina == 0);
    assert(dolzina == 0);

    HRESULT rezultat;

    // Dekodiramo tabelo SAFEARRAY.
    rezultat = SafeArrayAccessData(sa, (void HUGEP* FAR*)&podatki);
    if (SUCCEEDED(rezultat)) {
        LONG zgornja_meja = 0, spodnja_meja = 0;

        // Ugotovimo dolino vektorja.
        verify(SUCCEEDED(SafeArrayGetUBound(sa, 1, &zgornja_meja)));
        verify(SUCCEEDED(SafeArrayGetLBound(sa, 1, &spodnja_meja)));
        dolzina = fiz_dolzina = zgornja_meja - spodnja_meja + 1;
        upravljaj = FALSE;
        odmik = 0;
        stanje = A_OK;
    } else {
        podatki = NULL;
        stanje = AOSN_NAPAKA_PARAMETER;
    }

#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return stanje;
}


T_KAZALEC RPomDatoteka::Odpusti()
{
    T_KAZALEC _podatki = podatki;

    podatki = NULL;
    odmik = 0;
    dolzina = fiz_dolzina = 0;
    upravljaj = TRUE;
    stanje = A_OK;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    return _podatki;
}


T_VOID RPomDatoteka::Odpusti(LPSAFEARRAY sa)
{
    podatki = NULL;
    odmik = 0;
    dolzina = fiz_dolzina = 0;
    upravljaj = TRUE;
    stanje = A_OK;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    SafeArrayUnaccessData(sa);
}


T_STANJE RPomDatoteka::Nalozi(TK_NIZ ime, T_UINT nacin)
{
    RFizDatoteka dat;

    stanje = dat.Odpri(ime, nacin | RDAT_BRANJE | RDAT_DOSTOP_ZAPOREDEN);

    if (USPESNO(stanje)) {
        // Preverimo 4 GB mejo.
        assert(dat.VrniDolzino() <= T_SIZE_MAX);

        // Vnaprej rezerviramo ustrezno dolino, da ne bo nepotrebnih realloc-ov.
        if (USPESNO(RezervirajDolzino((T_SIZE)dat.VrniDolzino(), TRUE))) {
            odmik = dolzina = 0;
            ZapisiTok(dat);
            if (USPESNO(stanje))
                odmik = 0;

#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
            datum_stvaritve = dat.VrniDatumStvaritve();
            datum_dostopa   = dat.VrniDatumDostopa();
            datum_pisanja   = dat.VrniDatumPisanja();
#endif
        }

        verify(USPESNO(dat.Zapri()));
    }

    return stanje;
}


T_ODMIK RPomDatoteka::Prestavi(T_RODMIK _odmik, T_DATNACINODMIKA kako)
{
    stanje = A_OK;

    switch (kako) {
    case RDAT_OD_ZACETKA:
        // Preverimo 4 GB mejo.
        assert(_odmik <= T_SIZE_MAX);
        return odmik = (T_SIZE)_odmik;
    case RDAT_OD_TRENUTNEGA:
        // Preverimo 4 GB mejo.
        assert(T_INT_MIN <= _odmik && _odmik <= T_INT_MAX);
        return odmik += (T_SIZE)_odmik;
    case RDAT_OD_KONCA:
        // Preverimo 4 GB mejo.
        assert(T_INT_MIN <= _odmik && _odmik <= T_INT_MAX);
        return odmik = dolzina+(T_SIZE)_odmik;
    default:
        assert(0);
        return T_ODMIK_MAX;
    }
}


T_ODMIK RPomDatoteka::VrniOdmik()
{
    return odmik;
}


T_DOLZINA RPomDatoteka::VrniDolzino() const
{
    return dolzina;
}


T_STANJE RPomDatoteka::PostaviKonec()
{
    dolzina = odmik;
    RezervirajDolzino(odmik);
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif
    return stanje;
}


T_SIZE RPomDatoteka::Preberi(T_KAZALEC _podatki, T_SIZE _dolzina)
{
    // e imamo kaj brati, gremo brati.
    if (_dolzina) {
        // Popravimo dolino, e nimamo na voljo dovolj bajtov v tabeli.
        T_SIZE prebranih = amin(_dolzina, dolzina - odmik);

        if (prebranih) {
            amemcpy(_podatki, podatki + odmik, prebranih);
            odmik += prebranih;
            stanje = A_OK;
        } else
            stanje = AOSN_NAPAKA_KONEC_TOKA;

#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
        datum_dostopa = RDatum::VrniTrenutniCas();
#endif

        return prebranih;
    } else {
        // Ni potrebe po branju, javimo, da je vse v redu.
        stanje = A_OK;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
        datum_dostopa = RDatum::VrniTrenutniCas();
#endif

        return 0;
    }
}


T_SIZE RPomDatoteka::Zapisi(TK_KAZALEC _podatki, T_SIZE _dolzina)
{
    if (NEUSPESNO(RezervirajDolzino(odmik+_dolzina)))
        return 0;

    amemcpy(podatki + odmik, _podatki, _dolzina);
    odmik += _dolzina;
    if (odmik > dolzina) dolzina = odmik;
    stanje = A_OK;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return _dolzina;
}


T_UINT RPomDatoteka::ZapisiObl(TK_ASCII format, T_LOCALE locale, ...)
{
    va_list parametri;
    T_UINT znakov;

    va_start(parametri, locale);
    znakov = ZapisiOblP(format, locale, parametri);
    va_end(parametri);

    return znakov;
}


T_UINT RPomDatoteka::ZapisiObl(TK_UNICODE format, T_LOCALE locale, ...)
{
    va_list parametri;
    T_UINT znakov;

    va_start(parametri, locale);
    znakov = ZapisiOblP(format, locale, parametri);
    va_end(parametri);

    return znakov;
}


T_UINT RPomDatoteka::ZapisiOblP(TK_ASCII format, T_LOCALE locale, va_list parametri)
{
    T_UINT max_dolzina;
    T_SIZE porabljeno;
    T_BOOL premalo = FALSE; // Predpostavimo, da je na zaetku dovolj prostora.
    T_INT zapisano;

    for (;;) {
        max_dolzina = (fiz_dolzina - odmik) / sizeof(T_AZNAK);

        // Po potrebi poveujemo prostor v pomnilniku s korakom RDAT_VELIKOST_BLOKA.
        if (!max_dolzina || premalo) {
            verify(USPESNO(RezervirajDolzino(fiz_dolzina + RDAT_VELIKOST_BLOKA * sizeof(T_AZNAK))));
            max_dolzina += RDAT_VELIKOST_BLOKA;
        }

        zapisano = avsnprintf((T_ASCII)(podatki+odmik), max_dolzina, format, locale, parametri);

        if (zapisano >= 0) break;
        else premalo = TRUE;
    }

    porabljeno = zapisano * sizeof(T_AZNAK);
    odmik += porabljeno;
    if (odmik>dolzina) dolzina = odmik;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return zapisano;
}


T_UINT RPomDatoteka::ZapisiOblP(TK_UNICODE format, T_LOCALE locale, va_list parametri)
{
    T_UINT max_dolzina;
    T_SIZE porabljeno;
    T_BOOL premalo = FALSE; // Predpostavimo, da je na zaetku dovolj prostora.
    T_INT zapisano;

    for (;;) {
        max_dolzina = (fiz_dolzina - odmik) / sizeof(T_UZNAK);

        // Po potrebi poveujemo prostor v pomnilniku s korakom RDAT_VELIKOST_BLOKA.
        if (!max_dolzina || premalo) {
            verify(USPESNO(RezervirajDolzino(fiz_dolzina + RDAT_VELIKOST_BLOKA * sizeof(T_UZNAK))));
            max_dolzina += RDAT_VELIKOST_BLOKA;
        }

        zapisano = avsnprintf((T_UNICODE)(podatki+odmik), max_dolzina, format, locale, parametri);

        if (zapisano >= 0) break;
        else premalo = TRUE;
    }

    porabljeno = zapisano * sizeof(T_UZNAK);
    odmik += porabljeno;
    if (odmik>dolzina) dolzina = odmik;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return zapisano;
}


T_STANJE RPomDatoteka::Zapri()
{
    if (upravljaj && podatki)
        afree(podatki);

    podatki = NULL;
    upravljaj = TRUE;
    odmik = 0;
    dolzina = fiz_dolzina = 0;
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_stvaritve = datum_dostopa = datum_pisanja = T_TRENUTEK_MIN;
#endif

    return stanje = A_OK;
}


T_STANJE RPomDatoteka::Izplakni()
{
#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER
    datum_dostopa = datum_pisanja = RDatum::VrniTrenutniCas();
#endif

    return stanje = A_OK;
}


#ifdef RDAT_BELEZI_CAS_DOSTOPA_VER

T_TRENUTEK RPomDatoteka::VrniDatumStvaritve() const
{
    return datum_stvaritve;
}


T_TRENUTEK RPomDatoteka::VrniDatumDostopa() const
{
    return datum_dostopa;
}


T_TRENUTEK RPomDatoteka::VrniDatumPisanja() const
{
    return datum_pisanja;
}


T_STANJE RPomDatoteka::NastaviDatumStvaritve(T_TRENUTEK datum)
{
    datum_stvaritve = datum;
    return A_OK;
}


T_STANJE RPomDatoteka::NastaviDatumDostopa(T_TRENUTEK datum)
{
    datum_dostopa = datum;
    return A_OK;
}


T_STANJE RPomDatoteka::NastaviDatumPisanja(T_TRENUTEK datum)
{
    datum_pisanja = datum;
    return A_OK;
}

#endif


#ifdef AOSN_FIFO

//////////////////////////////////////////////////////////////////////
// RFifoTokPodatkov
//////////////////////////////////////////////////////////////////////

RFifoTokPodatkov::RFifoTokPodatkov() : RTokPodatkov()
{
    odmik = dolzina = 0;
}


RFifoTokPodatkov::~RFifoTokPodatkov()
{
    verify(USPESNO(Zapri()));
}


T_SIZE RFifoTokPodatkov::Preberi(T_KAZALEC _podatki, T_SIZE _dolzina)
{
    T_UINT odmik = 0;
    T_SIZE ostanek = _dolzina;

    while (ostanek) {
        // Vzamemo prvi element iz vrste.
        T_ROCICA rocica = vrsta.VrniPrvega();
        if (!rocica) break;
        T_BINBLOK &blok = vrsta[rocica];

        // Izraunamo, koliko je ostalo podatkov v prvem elementu.
        T_SIZE razpolozljivo = blok.dolzina - odmik;

        if (razpolozljivo > ostanek) {
            // e je v bloku dovolj podatkov, ga prepiemo v celoti.
            memcpy((T_U1*)_podatki + odmik, (T_U1*)blok.podatki + odmik, ostanek);
            odmik   += ostanek;
            dolzina -= ostanek;
            odmik   += ostanek;
            ostanek = 0;
        } else {
            // Sicer pa prepiemo, kolikor ga pa imamo in sprostimo blok.
            memcpy((T_U1*)_podatki + odmik, (T_U1*)blok.podatki + odmik, razpolozljivo);
            odmik    = 0;
            dolzina -= razpolozljivo;
            odmik   += razpolozljivo;
            ostanek -= razpolozljivo;

            afree(blok.podatki);
            vrsta.Brisi(rocica);
        }
    }

    stanje = _dolzina && odmik == 0 ? AOSN_NAPAKA_KONEC_TOKA : A_OK;

    return odmik;
}


T_SIZE RFifoTokPodatkov::Zapisi(TK_KAZALEC _podatki, T_SIZE _dolzina)
{
    // Ustvarimo vozel.
    T_ROCICA rocica = vrsta.UstvariVozel();
    assert(rocica);
    T_BINBLOK &blok = vrsta[rocica];

    // Nastavimo podatke vozla.
    blok.podatki = aalloc(blok.dolzina = _dolzina);
    if (blok.podatki) {
        memcpy(blok.podatki, _podatki, _dolzina);
        dolzina += _dolzina;

        // Dodamo vozel v seznam.
        vrsta.DodajVozel(rocica);
        stanje = A_OK;

        return _dolzina;
    } else {
        vrsta.UniciVozel(rocica);
        stanje = AOSN_NAPAKA_POMNILNIK;

        return 0;
    }
}


T_STANJE RFifoTokPodatkov::Zapri()
{
    // Sprostimo vse bloke, ki so ostali v vrsti.
    for (;;) {
        T_ROCICA rocica = vrsta.VrniPrvega();
        if (!rocica) break;
        T_BINBLOK &blok = vrsta[rocica];

        afree(blok.podatki);
        vrsta.Brisi(rocica);
    }

    stanje = A_OK;
    dolzina = odmik = 0;

    return stanje;
}


T_STANJE RFifoTokPodatkov::Izplakni()
{
    return stanje = A_OK;
}

#endif // AOSN_FIFO


//////////////////////////////////////////////////////////////////////
// RDiagDatoteka
//////////////////////////////////////////////////////////////////////

#ifdef _DEBUG

RDiagDatoteka::RDiagDatoteka(RDatoteka * const *_datoteke, T_UINT _st_datotek) : RDatoteka()
{
    st_datotek = _st_datotek;
    datoteke = (RDatoteka**)aalloc(st_datotek * sizeof(RDatoteka*));
    assert(datoteke || !st_datotek);
    amemcpy(datoteke, _datoteke, st_datotek * sizeof(RDatoteka*));

    podatki = NULL;
    dolzina = 0;
}


RDiagDatoteka::~RDiagDatoteka()
{
    if (podatki) afree(podatki);
    if (datoteke) afree(datoteke);
}


T_SIZE RDiagDatoteka::Preberi(T_KAZALEC _podatki, T_SIZE _dolzina)
{
    if (st_datotek) {
        T_UINT i;
        T_SIZE rez;

        if (_dolzina > dolzina) {
            // Podaljamo pomoni pomnilniki blok, ki ga uporabljamo za primerjavo prebranih podatkov.
            podatki = arealloc(podatki, _dolzina);
            assert(podatki);
            dolzina = _dolzina;
        }

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->Preberi(_podatki, _dolzina);
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->Preberi(podatki, _dolzina) == rez);
            assert(!memcmp(podatki, _podatki, rez)); // Prebrani podatki se ne smejo razlikovati.
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else {
        stanje = AOSN_NAPAKA_BRANJE;
        return (T_SIZE)0;
    }
}


T_SIZE RDiagDatoteka::Zapisi(TK_KAZALEC _podatki, T_SIZE _dolzina)
{
    if (st_datotek) {
        T_UINT i;
        T_SIZE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->Zapisi(_podatki, _dolzina);
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->Zapisi(_podatki, _dolzina) == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else {
        stanje = AOSN_NAPAKA_PISANJE;
        return (T_SIZE)0;
    }
}


T_STANJE RDiagDatoteka::Izplakni()
{
    if (st_datotek) {
        T_UINT i;
        T_STANJE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->Izplakni();
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->Izplakni() == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        if (podatki) {
            // Izkoristimo trenutek za sprostitev pomonega pomnilnikega bloka.
            afree(podatki);
            podatki = NULL;
            dolzina = 0;
        }

        return rez;
    } else
        return stanje = AOSN_NAPAKA_IZPLAKNI;
}


T_STANJE RDiagDatoteka::Zapri()
{
    if (st_datotek) {
        T_UINT i;
        T_STANJE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->Zapri();
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->Zapri() == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else
        return stanje = AOSN_NAPAKA_ZAPIRANJE;
}


T_ODMIK RDiagDatoteka::Prestavi(T_RODMIK odmik, T_DATNACINODMIKA kako)
{
    if (st_datotek) {
        T_UINT i;
        T_ODMIK rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->Prestavi(odmik, kako);
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->Prestavi(odmik, kako) == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else {
        stanje = AOSN_NAPAKA_PREMIK;
        return (T_ODMIK)0;
    }
}


T_ODMIK RDiagDatoteka::VrniOdmik()
{
    if (st_datotek) {
        T_UINT i;
        T_ODMIK rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->VrniOdmik();
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->VrniOdmik() == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else {
        stanje = AOSN_NAPAKA_PREMIK;
        return (T_ODMIK)0;
    }
}


T_STANJE RDiagDatoteka::ZakleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    if (st_datotek) {
        T_UINT i;
        T_STANJE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->ZakleniObmocje(odmik, dolzina);
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->ZakleniObmocje(odmik, dolzina) == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else
        return stanje = AOSN_NAPAKA_ZAKLEP;
}


T_STANJE RDiagDatoteka::OdkleniObmocje(T_ODMIK odmik, T_DOLZINA dolzina)
{
    if (st_datotek) {
        T_UINT i;
        T_STANJE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->ZakleniObmocje(odmik, dolzina);
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->ZakleniObmocje(odmik, dolzina) == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else
        return stanje = AOSN_NAPAKA_ODKLEP;
}


T_DOLZINA RDiagDatoteka::VrniDolzino() const
{
    if (st_datotek) {
        T_UINT i;
        T_DOLZINA rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->VrniDolzino();
        // stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->VrniDolzino() == rez);
            // verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else {
        // stanje = AOSN_NAPAKA_DOLZINA;
        return (T_DOLZINA)0;
    }
}


T_STANJE RDiagDatoteka::PostaviKonec()
{
    if (st_datotek) {
        T_UINT i;
        T_STANJE rez;

        // Najprej kliemo isto metodo za prvo datoteko.
        assert(datoteke[0]);
        rez = datoteke[0]->PostaviKonec();
        stanje = datoteke[0]->VrniStanje();

        // Sedaj kliemo isto metodo e pri vseh drugih datotekah in preverjamo, da vrnejo enako.
        for (i = 1; i < st_datotek; i++) {
            assert(datoteke[i]);
            verify(datoteke[i]->PostaviKonec() == rez);
            verify(datoteke[i]->VrniStanje() == stanje);
        }

        return rez;
    } else
        return stanje = AOSN_NAPAKA_ODREZI;
}

#endif
