/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_DNEVNIK


//////////////////////////////////////////////////////////////////////
// Definicije lokalnih funkcij
//////////////////////////////////////////////////////////////////////

inline T_STANJE _ReportEvent(HANDLE hEventLog, WORD wType, WORD wCategory, DWORD dwEventID, PSID lpUserSid, WORD wNumStrings, DWORD dwDataSize, LPCTSTR *lpStrings, LPVOID lpRawData);


//////////////////////////////////////////////////////////////////////
// RDnevnik
//////////////////////////////////////////////////////////////////////

RDnevnik::RDnevnik(T_UINT _podrocje, TK_NIZ _identifikacija, HMODULE _modul, T_KODAJEZIKA2 _jezik) :
    RElement(),
    podrocje(_podrocje),
    identifikacija(astrdup<T_ZNAK>(_identifikacija)),
    modul(_modul),
    jezik(_jezik)
{
}

RDnevnik::~RDnevnik()
{
    afree(identifikacija);
}


//////////////////////////////////////////////////////////////////////
// RDnevnikDat
//////////////////////////////////////////////////////////////////////

#ifdef _UNICODE
RDnevnikDat::RDnevnikDat(T_UINT podrocje, TK_NIZ identifikacija, HMODULE modul, T_KODAJEZIKA2 jezik) :
#else
RDnevnikDat::RDnevnikDat(T_UINT podrocje, TK_NIZ identifikacija, HMODULE modul, T_KODAJEZIKA2 jezik, T_KODNASTRAN _kodna_stran) :
#endif
    RDnevnik(podrocje, identifikacija, modul, jezik),
    dat(NULL)
#ifndef _UNICODE
    , kodna_stran(_kodna_stran)
#endif
{
}

RTokPodatkov* RDnevnikDat::Uporabi(RTokPodatkov *dat)
{
    RTokPodatkov *stara = this->dat;
    this->dat = dat;
    return stara;
}

T_STANJE RDnevnikDat::ZapisiP(T_DNEVNIKPRIORITETA prioriteta, TK_NIZ sporocilo, T_LOCALE locale, va_list parametri)
{
    assert(dat);

    RDatum danes;
    T_DATUM datum;
    RNiz sp, k;
    T_UINT n;

    sp.OblikujP(sporocilo, locale, parametri);

    // Ugotovimo datum dogodka.
    danes.NastaviTrenutniCas();
    danes.VrniDatum(&datum);

    // Oblikujemo sporoilo za dnevnik.
    k.Oblikuj(_T("%04") _T(T_4_PRINTF_SIZE_PREFIX) _T("i-%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u-%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u\t%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u:%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u:%02") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u\t%") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u\t%s: %s\n"), T_LOCALE_THREAD_DEFAULT,
        (T_4)datum.leto, (T_U4)datum.mesec, (T_U4)datum.dan,
        (T_U4)datum.ura, (T_U4)datum.minuta, (T_U4)datum.sekunda,
        (T_U4)prioriteta,
        identifikacija,
        (TK_NIZ)sp);

    // Zaklenemo shranjevanje, da ne zapisujeta dve niti hkrati.
    RCakalnik cakalnik(kljucavnica);

    // Zapiemo.
    n = k.VrniDolzino();
    if (dat->ZapisiTab(k.Vrni(), sizeof(T_ZNAK), n) < n)
        return dat->VrniStanje();

    // Izplaknemo tok podatkov.
    // Priporoljiva je uporaba zastavice RDAT_PISI_SKOZI pri odpiranju datoteke za dnevnik (e gre za fiz. datoteko).
    dat->Izplakni();
    return A_OK;
}


//////////////////////////////////////////////////////////////////////
// RDnevnikNT
//////////////////////////////////////////////////////////////////////

RDnevnikNT::RDnevnikNT(T_UINT podrocje, TK_NIZ identifikacija, HMODULE modul, T_KODAJEZIKA2 jezik, T_ROCICAOS rocica) :
    RDnevnik(podrocje, identifikacija, modul, jezik),
    RSisElement(rocica)
{
}


T_STANJE RDnevnikNT::Odpri(TK_NIZ streznik)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    rocica = (T_ROCICAOS)::RegisterEventSource(streznik, identifikacija);
    return JE_VELJAVNA_ROCICA(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODPIRANJE);
}


T_STANJE RDnevnikNT::Zapri()
{
    if (JE_VELJAVNA_ROCICA(rocica)) {
        if (::DeregisterEventSource(rocica)) {
            rocica = NEVELJAVNA_ROCICA;
            return A_OK;
        } else
            return astanje(::GetLastError(), AOSN_NAPAKA_ZAPIRANJE);
    } else
        return AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RDnevnikNT::ZapisiP(T_DNEVNIKPRIORITETA prioriteta, TK_NIZ sporocilo, T_LOCALE locale, va_list parametri)
{
    return Zapisi(prioriteta, 0, RNiz::OblikujSP(sporocilo, locale, parametri).Vrni(), NULL);
}


T_STANJE RDnevnikNT::ZapisiP(T_DNEVNIKPRIORITETA prioriteta, T_KAZALEC podatki, T_SIZE dolzina, T_UINT koda_sporocila, va_list parametri)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    assert(podatki || !dolzina);

    WORD vrsta;
    RDinTabela<TK_NIZ> nizi;

    // Najprej preraunamo resnost napake iz kod AOsn v kode WinNT.
    if (prioriteta <= ADNE_NAPAKA)
        vrsta = RDNE_PRI_NAPAKA;
    else if (prioriteta <= ADNE_OPOZORILO)
        vrsta = RDNE_PRI_OPOZORILO;
    else if (prioriteta <= ADNE_OBVESTILO)
        vrsta = RDNE_PRI_OBVESTILO;
    else
        vrsta = RDNE_PRI_USPEH;

    // Pripraviti moramo tabelo nizov.
    for (;;) {
        TK_NIZ niz = va_arg(parametri, TK_NIZ);
        if (!niz)
            break;

        nizi += niz;
    }

    assert(sizeof(WORD) == sizeof(T_U2));
    assert(podrocje <= T_U2_MAX);
    assert(nizi.VrniStevilo() <= T_U2_MAX);

    return _ReportEvent((HANDLE)rocica, vrsta, (WORD)podrocje, (DWORD)koda_sporocila, NULL, (WORD)nizi.VrniStevilo(), (DWORD)dolzina, (LPCTSTR*)(nizi.Vrni()), podatki);
}


T_STANJE RDnevnikNT::Registriraj(TK_NIZ dnevnik, TK_NIZ identifikacija, T_UINT vrste_prioritet, TK_NIZ dat_sporocil, TK_NIZ dat_parametrov, TK_NIZ dat_podrocij, T_UINT st_podrocij)
{
    assert(identifikacija);

    RRegister reg;
    T_STANJE stanje;
    RNiz kljuc;

    kljuc.Oblikuj(_T("SYSTEM\\CurrentControlSet\\Services\\Eventlog\\%s\\%s"), T_LOCALE_THREAD_DEFAULT, dnevnik, identifikacija);

    stanje = reg.Ustvari(HKEY_LOCAL_MACHINE, kljuc, RREG_PISANJE
#ifndef _WIN64
        | (AOsn::JeProcesWOW64() ? RREG_UPORABI_64 : 0) // Tudi 32-bitni odjemalci naj shranjujejo podatke v 64-bitno vejo registra.
#endif
        );
    if (USPESNO(stanje)) {
        if (dat_sporocil)
            verify(USPESNO(reg.Zapisi(_T("EventMessageFile"), dat_sporocil, TRUE)));

        if (dat_parametrov)
            verify(USPESNO(reg.Zapisi(_T("ParameterMessageFile"), dat_parametrov, TRUE)));

        if (dat_podrocij) {
            verify(USPESNO(reg.Zapisi(_T("CategoryMessageFile"), dat_podrocij, TRUE)));
            verify(USPESNO(reg.Zapisi(_T("CategoryCount"), (T_U4)st_podrocij)));
        }

        verify(USPESNO(reg.Zapisi(_T("TypesSupported"), (T_U4)vrste_prioritet)));

        verify(USPESNO(reg.Zapri()));
    }

    return stanje;
}


T_STANJE RDnevnikNT::Odjavi(TK_NIZ dnevnik, TK_NIZ identifikacija)
{
    assert(identifikacija);

    RRegister reg;
    T_STANJE stanje;
    RNiz kljuc;

    kljuc.Oblikuj(_T("SYSTEM\\CurrentControlSet\\Services\\Eventlog\\%s"), T_LOCALE_THREAD_DEFAULT, dnevnik);

    stanje = reg.Odpri(HKEY_LOCAL_MACHINE, kljuc, RREG_PISANJE
#ifndef _WIN64
        | (AOsn::JeProcesWOW64() ? RREG_UPORABI_64 : 0) // Tudi 32-bitni odjemalci naj shranjujejo podatke v 64-bitno vejo registra.
#endif
        );
    if (USPESNO(stanje)) {
        stanje = reg.BrisiPodkljuc(identifikacija);
        verify(USPESNO(reg.Zapri()));
    }

    return stanje;
}


//////////////////////////////////////////////////////////////////////
// Lokalne funkcije
//////////////////////////////////////////////////////////////////////

inline T_STANJE _ReportEvent(HANDLE hEventLog, WORD wType, WORD wCategory, DWORD dwEventID, PSID lpUserSid, WORD wNumStrings, DWORD dwDataSize, LPCTSTR *lpStrings, LPVOID lpRawData)
{
    // Kliemo sistemsko funkcijo ReportEvent(), vendar smo previdni, ker se je e zgodilo, da je sproila izjemo.
    // Ker pa zapisovanje v sistemski dnevnik ni nekaj takega, kar bi lahko popravljali, ne poiljamo poroila o napaki.
    __try {
        return ::ReportEvent(hEventLog, wType, wCategory, dwEventID, lpUserSid, wNumStrings, dwDataSize, lpStrings, lpRawData) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_SISTEMSKI_DNEVNIK);
    } __except(EXCEPTION_EXECUTE_HANDLER) {}

    return AOSN_NAPAKA_SISTEMSKI_DNEVNIK;
}

#endif // AOSN_DNEVNIK
