/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_HTML


//////////////////////////////////////////////////////////////////////
// Interni podatki
//////////////////////////////////////////////////////////////////////

enum T_HTMLVRSTAZAKLJUCITVE {
    ZAKLJUCITEV_VEDNO = 0,
    ZAKLJUCITEV_VCASIH,
    ZAKLJUCITEV_NIKOLI
};


enum T_HTMLPRIKAZELEMENTA {
    PRIKAZ_VPETO = 0,   // angl. inline
    PRIKAZ_BLOKOVNO,
    PRIKAZ_SKRITO
};


const struct {
    T_HTMLVRSTAZAKLJUCITVE zakljucitev; // kako je element praviloma zakljuen
    T_HTMLPRIKAZELEMENTA prikaz;        // kako je element prikazan
} _vedenje_elementov[HTML_EL_ZADNJI - HTML_EL_PRVI] = {
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // A
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // ABBR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // ACRONYM
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // ADDRESS
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // APPLET
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // AREA
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // B
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // BASE
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // BASEFONT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // BDO
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // BGSOUND
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // BIG
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // BLINK
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // BLOCKQUOTE
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // BODY
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // BR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // BUTTON
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // CAPTION
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // CENTER
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // CITE
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // CODE
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_BLOKOVNO, }, // COL
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // COLGROUP
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_SKRITO,   }, // COMMENT
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // DD
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // DEL
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // DFN
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // DIR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // DIV
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // DL
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // DT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // EM
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_VPETO,    }, // EMBED
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // FIELDSET
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // FONT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // FORM
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_BLOKOVNO, }, // FRAME
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // FRAMESET
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H1
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H2
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H3
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H4
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H5
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // H6
    { ZAKLJUCITEV_VCASIH, PRIKAZ_SKRITO,   }, // HEAD
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_BLOKOVNO, }, // HR
    { ZAKLJUCITEV_VCASIH, PRIKAZ_SKRITO,   }, // HTML
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // I
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // IFRAME
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_VPETO,    }, // IMG
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_VPETO,    }, // INPUT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // INS
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_VPETO,    }, // ISINDEX
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // KBD
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // LABEL
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // LEGEND
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // LI
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // LINK
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // LISTING
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_SKRITO,   }, // MAP
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // MARQUEE
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // MENU
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // META
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // NEXTID
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_SKRITO,   }, // NOBR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // NOEMBED
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // NOFRAMES
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // NOSCRIPT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // OBJECT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // OL
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // OPTGROUP
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // OPTION
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // P
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // PARAM
    { ZAKLJUCITEV_VCASIH, PRIKAZ_VPETO,    }, // PLAINTEXT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // PRE
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // Q
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // RT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // RUBY
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // S
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SAMP
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_SKRITO,   }, // SCRIPT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SELECT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SMALL
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SPAN
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // STRIKE
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // STRONG
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_SKRITO,   }, // STYLE
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SUB
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // SUP
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // TABLE
    { ZAKLJUCITEV_VCASIH, PRIKAZ_SKRITO,   }, // TBODY
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // TD
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // TEXTAREA
    { ZAKLJUCITEV_VCASIH, PRIKAZ_SKRITO,   }, // TFOOT
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // TH
    { ZAKLJUCITEV_VCASIH, PRIKAZ_SKRITO,   }, // THEAD
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // TITLE
    { ZAKLJUCITEV_VCASIH, PRIKAZ_BLOKOVNO, }, // TR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // TT
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // U
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_BLOKOVNO, }, // UL
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // VAR
    { ZAKLJUCITEV_NIKOLI, PRIKAZ_SKRITO,   }, // WBR
    { ZAKLJUCITEV_VEDNO,  PRIKAZ_VPETO,    }, // XMP
};


//////////////////////////////////////////////////////////////////////
// Deklaracija internih funkcij inline
//////////////////////////////////////////////////////////////////////

inline T_HTMLVRSTAZAKLJUCITVE _VrniVrstoZakljucitve(T_HTMLKODAELEMENTA koda);
inline T_HTMLPRIKAZELEMENTA   _VrniPrikazElementa(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_fontstyle(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_phrase(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_special(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_formctrl(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_inline(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_heading(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_list(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_preformatted(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_block(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_flow(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_head_content(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_head_misc(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_pre_exclusion(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_html_content(T_HTMLKODAELEMENTA koda);
inline T_BOOL _is_group(T_HTMLKODAELEMENTA koda);
inline T_BOOL _can_contain(T_HTMLKODAELEMENTA koda_nad, T_HTMLKODAELEMENTA koda_pod);
inline T_BOOL _is_uri(T_HTMLKODAELEMENTA koda, TK_UNICODE ime_lastnosti, T_UINT znakov);
inline T_BOOL _is_localizable(T_HTMLKODAELEMENTA koda, TK_UNICODE ime_lastnosti, T_UINT znakov);


//////////////////////////////////////////////////////////////////////
// RHTMLOdsek
//////////////////////////////////////////////////////////////////////

RHTMLOdsek::RHTMLOdsek(T_HTMLVRSTAODSEKA _vrsta, T_UINT zacetek, T_UINT konec, T_UINT _nadrejeni) :
    RElement(),
    vrsta(_vrsta),
    interval(zacetek, konec),
    nadrejeni(_nadrejeni)
{
}


RHTMLOdsek::~RHTMLOdsek()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLOznakaSamostojna
//////////////////////////////////////////////////////////////////////

RHTMLOznakaSamostojna::~RHTMLOznakaSamostojna()
{
}


inline T_HTMLKODAELEMENTA RHTMLOznakaSamostojna::VrniKodoElementa(TK_UNICODE ime, T_UINT dolzina)
{
    T_UINT i, j, m, i1, i2;
    T_INT r;
    T_UZNAK znak;
    static const struct {
        TK_UNICODE ime;
        T_HTMLKODAELEMENTA koda;
    } kode_elementov[] = {
        { L"a",          HTML_EL_A,          },
        { L"abbr",       HTML_EL_ABBR,       },
        { L"acronym",    HTML_EL_ACRONYM,    },
        { L"address",    HTML_EL_ADDRESS,    },
        { L"applet",     HTML_EL_APPLET,     },
        { L"area",       HTML_EL_AREA,       },
        { L"b",          HTML_EL_B,          },
        { L"base",       HTML_EL_BASE,       },
        { L"basefont",   HTML_EL_BASEFONT,   },
        { L"bdo",        HTML_EL_BDO,        },
        { L"bgsound",    HTML_EL_BGSOUND,    },
        { L"big",        HTML_EL_BIG,        },
        { L"blink",      HTML_EL_BLINK,      },
        { L"blockquote", HTML_EL_BLOCKQUOTE, },
        { L"body",       HTML_EL_BODY,       },
        { L"br",         HTML_EL_BR,         },
        { L"button",     HTML_EL_BUTTON,     },
        { L"caption",    HTML_EL_CAPTION,    },
        { L"center",     HTML_EL_CENTER,     },
        { L"cite",       HTML_EL_CITE,       },
        { L"code",       HTML_EL_CODE,       },
        { L"col",        HTML_EL_COL,        },
        { L"colgroup",   HTML_EL_COLGROUP,   },
        { L"comment",    HTML_EL_COMMENT,    },
        { L"dd",         HTML_EL_DD,         },
        { L"del",        HTML_EL_DEL,        },
        { L"dfn",        HTML_EL_DFN,        },
        { L"dir",        HTML_EL_DIR,        },
        { L"div",        HTML_EL_DIV,        },
        { L"dl",         HTML_EL_DL,         },
        { L"dt",         HTML_EL_DT,         },
        { L"em",         HTML_EL_EM,         },
        { L"embed",      HTML_EL_EMBED,      },
        { L"fieldset",   HTML_EL_FIELDSET,   },
        { L"font",       HTML_EL_FONT,       },
        { L"form",       HTML_EL_FORM,       },
        { L"frame",      HTML_EL_FRAME,      },
        { L"frameset",   HTML_EL_FRAMESET,   },
        { L"h1",         HTML_EL_H1,         },
        { L"h2",         HTML_EL_H2,         },
        { L"h3",         HTML_EL_H3,         },
        { L"h4",         HTML_EL_H4,         },
        { L"h5",         HTML_EL_H5,         },
        { L"h6",         HTML_EL_H6,         },
        { L"head",       HTML_EL_HEAD,       },
        { L"hr",         HTML_EL_HR,         },
        { L"html",       HTML_EL_HTML,       },
        { L"i",          HTML_EL_I,          },
        { L"iframe",     HTML_EL_IFRAME,     },
        { L"img",        HTML_EL_IMG,        },
        { L"input",      HTML_EL_INPUT,      },
        { L"ins",        HTML_EL_INS,        },
        { L"isindex",    HTML_EL_ISINDEX,    },
        { L"kbd",        HTML_EL_KBD,        },
        { L"label",      HTML_EL_LABEL,      },
        { L"legend",     HTML_EL_LEGEND,     },
        { L"li",         HTML_EL_LI,         },
        { L"link",       HTML_EL_LINK,       },
        { L"listing",    HTML_EL_LISTING,    },
        { L"map",        HTML_EL_MAP,        },
        { L"marquee",    HTML_EL_MARQUEE,    },
        { L"menu",       HTML_EL_MENU,       },
        { L"meta",       HTML_EL_META,       },
        { L"nextid",     HTML_EL_NEXTID,     },
        { L"nobr",       HTML_EL_NOBR,       },
        { L"noembed",    HTML_EL_NOEMBED,    },
        { L"noframes",   HTML_EL_NOFRAMES,   },
        { L"noscript",   HTML_EL_NOSCRIPT,   },
        { L"object",     HTML_EL_OBJECT,     },
        { L"ol",         HTML_EL_OL,         },
        { L"optgroup",   HTML_EL_OPTGROUP,   },
        { L"option",     HTML_EL_OPTION,     },
        { L"p",          HTML_EL_P,          },
        { L"param",      HTML_EL_PARAM,      },
        { L"plaintext",  HTML_EL_PLAINTEXT,  },
        { L"pre",        HTML_EL_PRE,        },
        { L"q",          HTML_EL_Q,          },
        { L"rt",         HTML_EL_RT,         },
        { L"ruby",       HTML_EL_RUBY,       },
        { L"s",          HTML_EL_S,          },
        { L"samp",       HTML_EL_SAMP,       },
        { L"script",     HTML_EL_SCRIPT,     },
        { L"select",     HTML_EL_SELECT,     },
        { L"small",      HTML_EL_SMALL,      },
        { L"span",       HTML_EL_SPAN,       },
        { L"strike",     HTML_EL_STRIKE,     },
        { L"strong",     HTML_EL_STRONG,     },
        { L"style",      HTML_EL_STYLE,      },
        { L"sub",        HTML_EL_SUB,        },
        { L"sup",        HTML_EL_SUP,        },
        { L"table",      HTML_EL_TABLE,      },
        { L"tbody",      HTML_EL_TBODY,      },
        { L"td",         HTML_EL_TD,         },
        { L"textarea",   HTML_EL_TEXTAREA,   },
        { L"tfoot",      HTML_EL_TFOOT,      },
        { L"th",         HTML_EL_TH,         },
        { L"thead",      HTML_EL_THEAD,      },
        { L"title",      HTML_EL_TITLE,      },
        { L"tr",         HTML_EL_TR,         },
        { L"tt",         HTML_EL_TT,         },
        { L"u",          HTML_EL_U,          },
        { L"ul",         HTML_EL_UL,         },
        { L"var",        HTML_EL_VAR,        },
        { L"wbr",        HTML_EL_WBR,        },
        { L"xmp",        HTML_EL_XMP,        },
    };


#ifdef _DEBUG
    // Preverimo, ali je tabela imen elementov urejena.
    for (i = 1; i < _countof(kode_elementov); i++)
        assert(astrcmp(kode_elementov[i-1].ime, kode_elementov[i].ime) <= 0);

    // Vsa imena elementov morajo biti zapisana z malimi rkami ali tevilkami.
    for (i = 0; i < _countof(kode_elementov); i++) {
        for (j = 0; kode_elementov[i].ime[j]; j++)
            assert(aislower(kode_elementov[i].ime[j]) || aisdigit(kode_elementov[i].ime[j]));
    }
#endif

    // Poiemo ime v tabeli z bisekcijo.
    for (i = 0, j = _countof(kode_elementov); i < j; ) {
        m = (i + j) / 2;
        // Naredimo primerjavo zaetkov nizov.
        for (i1 = i2 = r = 0;;) {
            if (!kode_elementov[m].ime[i1]) {
                // Niza v tabeli je konec.
                r = i2 >= dolzina || !ime[i2] ? 0 : -1;
                break;
            } else if (i2 >= dolzina || !ime[i2]) {
                // Besede je konec.
                r = 1;
                break;
            }

            // Beseda mora biti zapisana s samimi malimi rkami, saj je spisek imen tudi zapisan s samimi malimi rkami.
            znak = atolower(ime[i2++]);
            if (kode_elementov[m].ime[i1] > znak) {
                r = 1;
                break;
            } else if (kode_elementov[m].ime[i1] < znak) {
                r = -1;
                break;
            }
            i1++;
        }

        if (r < 0)
            i = m + 1;
        else if (r > 0)
            j = m;
        else
            return kode_elementov[m].koda;
    }

    return HTML_EL_NEZNAN;
}


//////////////////////////////////////////////////////////////////////
// RHTMLDeklaracija
//////////////////////////////////////////////////////////////////////

RHTMLDeklaracija::~RHTMLDeklaracija()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLEntiteta
//////////////////////////////////////////////////////////////////////

RHTMLEntiteta::RHTMLEntiteta() : RElement()
{
}


RHTMLEntiteta::~RHTMLEntiteta()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLZeton
//////////////////////////////////////////////////////////////////////

RHTMLZeton::RHTMLZeton(T_HTMLVRSTAZETONA _vrsta, T_UINT _indeks_odseka, T_UINT _instanca) :
    RElement(),
    vrsta(_vrsta),
    indeks_odseka(_indeks_odseka),
    instanca(_instanca)
{
}


RHTMLZeton::~RHTMLZeton()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLZetonBesedilo
//////////////////////////////////////////////////////////////////////

RHTMLZetonBesedilo::RHTMLZetonBesedilo(T_HTMLVRSTAZETONA vrsta, TK_UNICODE _besedilo, T_UINT znakov, T_UINT _vrsta_besedila, T_UINT indeks_odseka, T_UINT instanca) :
    RHTMLZeton(vrsta, indeks_odseka, instanca),
    besedilo(_besedilo, znakov),
    vrsta_besedila(_vrsta_besedila)
{
}


RHTMLZetonBesedilo::~RHTMLZetonBesedilo()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLZetonUvodni
//////////////////////////////////////////////////////////////////////

RHTMLZetonUvodni::RHTMLZetonUvodni(TK_UNICODE besedilo, T_UINT besedilo_znakov, TK_UNICODE ime, T_UINT ime_znakov, T_UINT vrsta_besedila, T_UINT indeks_odseka, T_UINT _indeks_konca, T_UINT instanca) :
    RHTMLZetonBesedilo(HTML_ZETON_UVODNI, besedilo, besedilo_znakov, vrsta_besedila, indeks_odseka, instanca),
    ime_elementa(ime, ime_znakov),
    indeks_konca(_indeks_konca)
{
}


RHTMLZetonUvodni::~RHTMLZetonUvodni()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLZetonURL
//////////////////////////////////////////////////////////////////////

RHTMLZetonURL::RHTMLZetonURL(TK_UNICODE _url, T_UINT znakov, T_HTMLZAPISURLZETONA _zapis, T_UINT indeks_odseka, T_UINT instanca) :
    RHTMLZeton(HTML_ZETON_NASLOVURL, indeks_odseka, instanca),
    url(_url, znakov),
    zapis(_zapis)
{
}


RHTMLZetonURL::~RHTMLZetonURL()
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLZetonNapaka
//////////////////////////////////////////////////////////////////////

RHTMLZetonNapaka::RHTMLZetonNapaka(T_STANJE _stanje, T_UINT instanca) :
    RHTMLZeton(HTML_ZETON_NAPAKA, T_UINT_MAX, instanca),
    stanje(_stanje)
{
}


//////////////////////////////////////////////////////////////////////
// RHTMLDokument
//////////////////////////////////////////////////////////////////////

RHTMLDokument::RHTMLDokument() :
    RElement(),
    znakov_analizirano(0),
    kodna_stran(KODNASTRAN_NEZNANA),

    // Podatki za razpoznavo deklaracij
    st_veljavnih_pogojev(0),
    st_neveljavnih_pogojev(0),
    je_cdata(FALSE),
    je_rcdata(FALSE),

    // Podatki za razpoznavo elementov
    aktivni_element(T_UINT_MAX),
    je_posebni_element(FALSE)
{
}


RHTMLDokument::~RHTMLDokument()
{
    // e se sproi kateri izmed spodnjih assertov, je uporabnik pozabil klicati metodo RHTMLDokument::Sprazni().
    assert(!koda.VrniDolzino());
    assert(!entitete.VrniStevilo());
    assert(!odseki.VrniStevilo());
    assert(!odprte_oznake.VrniStevilo());
}


T_VOID RHTMLDokument::Sprazni()
{
    T_UINT i, n;

    koda.Sprazni();
    znakov_analizirano = 0;
    kodna_stran = KODNASTRAN_NEZNANA;

    // Podatki za razpoznavo deklaracij
    st_veljavnih_pogojev = st_neveljavnih_pogojev = 0;
    je_cdata = je_rcdata = FALSE;
    for (i = 0, n = entitete.VrniStevilo(); i < n; i++) {
        RHTMLEntiteta *&e = entitete[i];
        assert(e);
        delete e;
    }
    entitete.Sprazni();

    // Podatki za razpoznavo elementov
    for (i = 0, n = odseki.VrniStevilo(); i < n; i++) {
        RHTMLOdsek *&o = odseki[i];
        assert(o);
        delete o;
    }
    odseki.Sprazni();

    odprte_oznake.Sprazni();
    aktivni_element = T_UINT_MAX;
    je_posebni_element = FALSE;
}


T_STANJE RHTMLDokument::Prilepi(TK_UNICODE _koda, T_UINT znakov)
{
    T_UINT i, konec;

    konec = koda.Prilepi(_koda, znakov);
    _koda = koda.Vrni();

    for (i = znakov_analizirano; i < konec;) {
        if (je_cdata || je_rcdata) {
            // Smo sredi odseka (R)CDATA.
            if (pogoj_konec.PreveriUjemanje(_koda, i, konec)) {
                // Nali smo konec odseka.
                odseki += new RHTMLOdsek(je_cdata ? HTML_ODSEK_CDATA : HTML_ODSEK_PCDATA, znakov_analizirano, i, aktivni_element);

                je_cdata = je_rcdata = FALSE;

                i = znakov_analizirano = pogoj_konec.interval.konec;
                continue;
            }

            goto naslednji_znak;
        }

        if (st_neveljavnih_pogojev) {
            // Imamo odprtega enega ali ve neveljavnih pogojev.
            if (pogoj_konec.PreveriUjemanje(_koda, i, konec)) {
                // Nali smo konec odseka.
                st_neveljavnih_pogojev--;

                i = znakov_analizirano = pogoj_konec.interval.konec;
                continue;
            }

            goto naslednji_znak;
        }

        if (st_veljavnih_pogojev && pogoj_konec.PreveriUjemanje(_koda, i, konec)) {
            // Nali smo konec odprtega veljavnega pogoja.

            // Po potrebi dodamo vmesno besedilo.
            if (znakov_analizirano < i)
                odseki += new RHTMLOdsek(HTML_ODSEK_BESEDILO, znakov_analizirano, i, aktivni_element);

            st_veljavnih_pogojev--;

            i = znakov_analizirano = pogoj_konec.interval.konec;
            continue;
        }

        if (pogoj_zacetek.PreveriUjemanje(_koda, i, konec)) {
            // Nali smo zaetek pogoja.
            RNizU vrednost_pogoja;

            ZamenjajEntitete(vrednost_pogoja, _koda + pogoj_zacetek.pogoj.zacetek, pogoj_zacetek.pogoj.VrniDolzino());
            if (!vrednost_pogoja.Primerjaj(L"CDATA")) {
                // Nali smo odsek CDATA.
                je_cdata = TRUE;
            } else if (!vrednost_pogoja.Primerjaj(L"RCDATA")) {
                // Nali smo odsek RCDATA.
                je_rcdata = TRUE;
            } if (st_neveljavnih_pogojev) {
                // Smo sredi bloka neveljavnega pogoja, zato bo tudi ta pogoj po definiciji neveljaven.
                st_neveljavnih_pogojev++;
            } else if (!vrednost_pogoja.Primerjaj(L"IGNORE")) {
                // Nali smo pogoj "IGNORE".
                st_neveljavnih_pogojev++;
            } else {
                // Po definiciji so vsi drugi pogoji veljavni.
                st_veljavnih_pogojev++;
            }


            i = znakov_analizirano = pogoj_zacetek.interval.konec;
            continue;
        }

        if (je_posebni_element) {
            // Smo sredi obdelave posebnega elementa.
            RHTMLOznakaUvodna *o_uvodna = (RHTMLOznakaUvodna*)odseki[aktivni_element];
            assert(o_uvodna);

            if (oznaka.PreveriUjemanje(_koda, i, konec) &&
                oznaka.vrsta == HTML_ODSEK_OZNAKA_ZAKLJUCNA &&
                RHTMLOznakaSamostojna::VrniKodoElementa(_koda + oznaka.ime.zacetek, oznaka.ime.VrniDolzino()) == o_uvodna->koda)
            {
                // Nali smo zakljuno oznako posebnega elementa.
                T_UINT j = odprte_oznake.VrniStevilo() - 1;
                RHTMLOznakaZakljucna *o = new RHTMLOznakaZakljucna(oznaka, _koda);
                assert(o);

                // Po potrebi dodamo vmesno besedilo.
                if (znakov_analizirano < i)
                    odseki += new RHTMLOdsek(HTML_ODSEK_BESEDILO, znakov_analizirano, i, aktivni_element);

                o_uvodna->zakljucna = odseki.VrniStevilo();
                o->uvodna = aktivni_element;
                o->nadrejeni = o_uvodna->nadrejeni;
                odseki += o;

                // Mnoico odprtih oznak lahko na koncu skrajamo.
                odprte_oznake.NastaviStevilo(j);
                aktivni_element = j ? odprte_oznake[j - 1] : T_UINT_MAX;

                je_posebni_element = FALSE;

                i = znakov_analizirano = oznaka.interval.konec;
                continue;
            }

            goto naslednji_znak;
        }

        if (oznaka.PreveriUjemanje(_koda, i, konec)) {
            // Nali smo oznako.

            // Po potrebi dodamo vmesno besedilo.
            if (znakov_analizirano < i)
                odseki += new RHTMLOdsek(HTML_ODSEK_BESEDILO, znakov_analizirano, i, aktivni_element);

            if (oznaka.vrsta == HTML_ODSEK_OZNAKA_SAMOSTOJNA ||
                oznaka.vrsta == HTML_ODSEK_OZNAKA_UVODNA)
            {
                T_UINT j, idx = odseki.VrniStevilo();
                RHTMLOznakaSamostojna *o =
                    oznaka.vrsta == HTML_ODSEK_OZNAKA_SAMOSTOJNA ? new RHTMLOznakaSamostojna(oznaka, _koda, aktivni_element) :
                    oznaka.vrsta == HTML_ODSEK_OZNAKA_UVODNA     ? new     RHTMLOznakaUvodna(oznaka, _koda, aktivni_element) :
                    NULL;
                assert(o);
                // Preverimo, ali oznaka prekine katerega izmed odprtih elementov.
                for (j = odprte_oznake.VrniStevilo(); j--; ) {
                    T_UINT &idx_uvodna = odprte_oznake[j];
                    if (idx_uvodna != T_UINT_MAX) {
                        RHTMLOznakaUvodna *o_uvodna = (RHTMLOznakaUvodna*)odseki[idx_uvodna];
                        assert(o_uvodna && o_uvodna->vrsta == HTML_ODSEK_OZNAKA_UVODNA);
                        if (_can_contain(o_uvodna->koda, o->koda)) {
                            // Nadelement lahko vsebuje ta element.
                            o->nadrejeni = idx_uvodna;
                            break;
                        } else {
                            // Zakljuimo odprto oznako.
                            o_uvodna->zakljucna = idx;
                            idx_uvodna = T_UINT_MAX;
                            odprte_oznake.NastaviStevilo(j);
                        }
                    }
                }
                if (oznaka.vrsta == HTML_ODSEK_OZNAKA_UVODNA) {
                    if (_VrniVrstoZakljucitve(o->koda) == ZAKLJUCITEV_NIKOLI) {
                        // Element ni nikoli zakljuen. Zato ga bo zakljuil e naslednji odsek za njim.
                        ((RHTMLOznakaUvodna*)o)->zakljucna = idx + 1;
                    } else {
                        // Element dodamo v mnoico odprtih elementov.
                        odprte_oznake   += idx;
                        aktivni_element  = idx;
                        if (((RHTMLOznakaUvodna*)o)->koda == HTML_EL_CODE ||
                            ((RHTMLOznakaUvodna*)o)->koda == HTML_EL_COMMENT ||
                            ((RHTMLOznakaUvodna*)o)->koda == HTML_EL_SCRIPT ||
                            ((RHTMLOznakaUvodna*)o)->koda == HTML_EL_STYLE)
                        {
                            // Element, ki se je ravnokar zael je poseben in zahteva rokovanje v rokavikah.
                            je_posebni_element = TRUE;
                        }
                    }
                }
                if (o->koda == HTML_EL_META && kodna_stran == KODNASTRAN_NEZNANA) {
                    // Nali smo element META, kodna stran pa e vedno ni znana. Poskusimo jo doloiti.
                    T_UINT i, st_lastnosti = oznaka.lastnosti.VrniStevilo();
                    T_BOOL je_content_type = FALSE;
                    T_UINT i_content = T_UINT_MAX;

                    for (i = 0; i < st_lastnosti; i++) {
                        if (!astrnicmp(_koda + oznaka.lastnosti[i].ime.zacetek,      oznaka.lastnosti[i].ime.VrniDolzino(),      L"HTTP-EQUIV",   T_UINT_MAX) &&
                            !astrnicmp(_koda + oznaka.lastnosti[i].vrednost.zacetek, oznaka.lastnosti[i].vrednost.VrniDolzino(), L"CONTENT-TYPE", T_UINT_MAX))
                        {
                            // Nali smo lastnost http-equiv="Content-Type".
                            je_content_type = TRUE;
                        } else if (!astrnicmp(_koda + oznaka.lastnosti[i].ime.zacetek, oznaka.lastnosti[i].ime.VrniDolzino(), L"CONTENT", T_UINT_MAX)) {
                            // Nali smo lastnost content.
                            i_content = i;
                        }
                    }

                    if (je_content_type && i_content != T_UINT_MAX) {
                        // Nali smo oznako <meta http-equiv="Content-Type" content="...">. Analiziramo jo podrobneje.
                        RMIMERegElContentTypeU vsebina;

                        if (vsebina.PreveriUjemanje(_koda, oznaka.lastnosti[i_content].vrednost.zacetek, oznaka.lastnosti[i_content].vrednost.konec) &&
                            vsebina.kodna_stran.JeVeljaven())
                        {
                            // Prepoznali smo vsebino in kodno stran.
                            RNizA niz;
                            niz.NastaviU(_koda + vsebina.kodna_stran.zacetek, vsebina.kodna_stran.VrniDolzino());
                            kodna_stran = AOsn::VrniKodoKodneStrani(niz.Vrni(), niz.VrniDolzino());
                        }
                    }
                }
                odseki += o;
            } else if (oznaka.vrsta == HTML_ODSEK_OZNAKA_ZAKLJUCNA) {
                T_UINT j, idx = odseki.VrniStevilo();
                T_BOOL je_zadnja = TRUE;
                RHTMLOznakaZakljucna *o = new RHTMLOznakaZakljucna(oznaka, _koda);
                assert(o);
                // Poiemo, kateri element oznaka prekine.
                for (j = odprte_oznake.VrniStevilo(); j--; ) {
                    T_UINT &idx_uvodna = odprte_oznake[j];
                    if (idx_uvodna != T_UINT_MAX) {
                        RHTMLOznakaUvodna *o_uvodna = (RHTMLOznakaUvodna*)odseki[idx_uvodna];
                        assert(o_uvodna && o_uvodna->vrsta == HTML_ODSEK_OZNAKA_UVODNA);
                        if (o_uvodna->koda == o->koda ||
                            o_uvodna->koda == HTML_EL_NEZNAN && o->koda == HTML_EL_NEZNAN && !astrnicmp(_koda + o_uvodna->ime.zacetek, o_uvodna->ime.VrniDolzino(), _koda + o->ime.zacetek, o->ime.VrniDolzino()))
                        {
                            // Nali smo oznako, ki je (najbr) uvodna k nai zakljuni.
                            o_uvodna->zakljucna = idx;
                            o->uvodna = idx_uvodna;
                            o->nadrejeni = o_uvodna->nadrejeni;
                            idx_uvodna = T_UINT_MAX;
                            break;
                        } else {
                            je_zadnja = FALSE;
                        }
                    }
                }
                if (je_zadnja) {
                    if (j != T_UINT_MAX) {
                        // Mnoico odprtih oznak lahko na koncu skrajamo.
                        odprte_oznake.NastaviStevilo(j);
                        aktivni_element = j ? odprte_oznake[j - 1] : T_UINT_MAX;
                    } else {
                        // Mnoico odprtih oznak lahko popolnoma spraznimo.
                        odprte_oznake.Sprazni();
                        aktivni_element = T_UINT_MAX;
                    }
                }
                odseki += o;
            } else if (oznaka.vrsta == HTML_ODSEK_DEKLARACIJA) {
                // Gre za deklaracijo.
                T_UINT st_lastnosti = oznaka.lastnosti.VrniStevilo();
                if (st_lastnosti > 3 &&
                    !astrnicmp(_koda + oznaka.lastnosti[0].ime.zacetek, oznaka.lastnosti[0].ime.VrniDolzino(), L"ENTITY", T_UINT_MAX))
                {
                    if (!astrncmp(_koda + oznaka.lastnosti[1].ime.zacetek, oznaka.lastnosti[1].ime.VrniDolzino(), L"%",      T_UINT_MAX) &&
                            astrncmp(_koda + oznaka.lastnosti[3].ime.zacetek, oznaka.lastnosti[3].ime.VrniDolzino(), L"SYSTEM", T_UINT_MAX) &&
                            astrncmp(_koda + oznaka.lastnosti[3].ime.zacetek, oznaka.lastnosti[3].ime.VrniDolzino(), L"PUBLIC", T_UINT_MAX))
                    {
                        // Nali smo deklaracijo entitete %.
                        RHTMLEntiteta *e = new RHTMLEntiteta;
                        e->ime = oznaka.lastnosti[2].ime;
                        ZamenjajEntitete(e->vrednost, _koda + oznaka.lastnosti[3].ime.zacetek, oznaka.lastnosti[3].ime.VrniDolzino());
                        entitete += e;
                    }

                    // TODO: Obdelaj e entitete vrste & in entitete v zunanjih datotekah SYSTEM ter PUBLIC.
                }
                odseki += new RHTMLDeklaracija(oznaka, aktivni_element);
            } else if (oznaka.vrsta == HTML_ODSEK_KOMENTAR) {
                odseki += new RHTMLKomentar(oznaka, aktivni_element);
            } else if (oznaka.vrsta == HTML_ODSEK_NAVODILO)
                odseki += new RHTMLNavodilo(oznaka, aktivni_element);
            else
                assert(0); // Neznana vrsta oznake.

            i = znakov_analizirano = oznaka.interval.konec;
            continue;
        }

naslednji_znak:
        if (poljuben_znak.PreveriUjemanje(_koda, i, konec)) {
            // Vse druge znake preskoimo, a jih e ne proglasimo za analizirane, saj so lahko sestavni del e nedokonane oznake.
            i = poljuben_znak.interval.konec;
        } else {
            // Niz je bil predasno zakljuen, zato zakljuimo tudi mi.
            break;
        }
    }

    return A_OK;
}


T_STANJE RHTMLDokument::Zakljuci()
{
    T_UINT i = koda.VrniDolzino(), n;

    // Po potrebi dodamo zakljuno besedilo.
    if (znakov_analizirano < i)
        odseki += new RHTMLOdsek(HTML_ODSEK_BESEDILO, znakov_analizirano, i, aktivni_element);

    znakov_analizirano = i;

    // Zakljuimo vse oznake, ki so ostale odprte.
    for (i = 0, n = odprte_oznake.VrniStevilo(); i < n; i++) {
        T_UINT j = odprte_oznake[i];
        if (j != T_UINT_MAX) {
            RHTMLOznakaUvodna *o = (RHTMLOznakaUvodna*)odseki[j];
            assert(o && o->vrsta == HTML_ODSEK_OZNAKA_UVODNA);
            o->zakljucna = T_UINT_MAX;
        }
    }
    odprte_oznake.Sprazni();
    aktivni_element = T_UINT_MAX;

    return A_OK;
}


T_STANJE RHTMLDokument::Nalozi(RDatoteka &dat)
{
    T_STANJE stanje;
    RNizU niz;

    // Najprej spraznimo dokument.
    Sprazni();

    // Poskusimo doloiti kodno stran.
    kodna_stran = dat.DolociKodnoStran(KODNASTRAN_NEZNANA);
    stanje = dat.VrniStanje();
    if (NEUSPESNO(stanje))
        goto napaka1;

    // Dokler ne ugotovimo kodne strani dokumenta, ga beremo znak za znakom.
    while (kodna_stran == KODNASTRAN_NEZNANA) {
        T_AZNAK znak;
        dat >> znak;
        if (dat.JeKonec()) break;

        niz.Nastavi(::aasctouni(&znak, 1, KODNASTRAN_ANSI), 1);
        stanje = Prilepi(niz.Vrni(), 1);
        if (NEUSPESNO(stanje))
            goto napaka1;
    }

    if (kodna_stran != KODNASTRAN_NEZNANA) {
        // Ugotovili smo kodno stran. Dokument zdaj lahko preberemo v blokih.
        for (;;) {
            // Preberemo vrstico.
            dat.PreberiVrstico(niz, kodna_stran);
            if (dat.JeKonec()) break;

            // Dodamo znak za novo vrstico.
            niz += L'\n';

            // Prilepimo vrstico v dokument.
            stanje = Prilepi(niz.Vrni(), niz.VrniDolzino());
            if (NEUSPESNO(stanje))
                goto napaka1;
        }
    }

    verify(USPESNO(Zakljuci()));
    stanje = A_OK;

napaka1:
    return stanje;
}


#ifdef _DEBUG
T_VOID RHTMLDokument::IzpisiDiag(T_UINT zamik) const
{
    T_UINT i, st_odsekov = odseki.VrniStevilo();
    TK_UNICODE html = koda.Vrni();
    for (i = 0; i < st_odsekov; i++) {
        const RHTMLOdsek *o = odseki[i];
        assert(o);
        switch (o->vrsta) {
            case HTML_ODSEK_BESEDILO:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: \"%.*ls\" -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, amin<T_UINT>(o->interval.VrniDolzino(), 10), html + o->interval.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_OZNAKA_SAMOSTOJNA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <%.*ls/> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, ((const RHTMLOznakaSamostojna*)o)->ime.VrniDolzino(), html + ((const RHTMLOznakaSamostojna*)o)->ime.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_OZNAKA_UVODNA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <%.*ls> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u -> %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, ((const RHTMLOznakaUvodna*)o)->ime.VrniDolzino(), html + ((const RHTMLOznakaUvodna*)o)->ime.zacetek, o->nadrejeni, ((const RHTMLOznakaUvodna*)o)->zakljucna);
                break;
            case HTML_ODSEK_OZNAKA_ZAKLJUCNA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u <- </%.*ls> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, ((const RHTMLOznakaZakljucna*)o)->uvodna, ((const RHTMLOznakaZakljucna*)o)->ime.VrniDolzino(), html + ((const RHTMLOznakaZakljucna*)o)->ime.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_DEKLARACIJA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <!%.*ls> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, ((const RHTMLDeklaracija*)o)->ime.VrniDolzino(), html + ((const RHTMLDeklaracija*)o)->ime.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_KOMENTAR:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <!--%.*ls--> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, ((const RHTMLKomentar*)o)->vsebina.VrniDolzino(), html + ((const RHTMLKomentar*)o)->vsebina.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_NAVODILO:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <?%.*ls> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, amin<T_UINT>(((const RHTMLNavodilo*)o)->vsebina.VrniDolzino(), 10), html + ((const RHTMLNavodilo*)o)->vsebina.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_PCDATA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <![PCDATA[%.*ls]]> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, amin<T_UINT>(o->interval.VrniDolzino(), 10), html + o->interval.zacetek, o->nadrejeni);
                break;
            case HTML_ODSEK_CDATA:
                adebug(ADNE_DIAG, ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u: <![CDATA[%.*ls]]> -^ %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u\n"), zamik, _T(""), i, amin<T_UINT>(o->interval.VrniDolzino(), 10), html + o->interval.zacetek, o->nadrejeni);
                break;
            default:
                assert(0); // Neznana vrsta odseka.
        }
    }
}
#endif


//////////////////////////////////////////////////////////////////////
// RHTMLAnalizator
//////////////////////////////////////////////////////////////////////

RHTMLAnalizator::RHTMLAnalizator() :
    RElement(),
    dokument(NULL),
    razcleni_okvirje(FALSE),
    kazalnik_napredka(NULL)
{
}


RHTMLAnalizator::~RHTMLAnalizator()
{
    // e se sproi spodnji assert, je uporabnik pozabil klicati metodo RHTMLAnalizator::Sprazni().
    assert(!zetoni.VrniStevilo());
}


T_VOID RHTMLAnalizator::Sprazni()
{
    T_UINT i, n = zetoni.VrniStevilo();

    dokument          = NULL;
    razcleni_okvirje  = FALSE;
    kazalnik_napredka = NULL;
    naslov_url.Sprazni();

    for (i = 0; i < n; i++) {
        RHTMLZeton *zeton = zetoni[i];
        assert(zeton);
        delete zeton;
    }
    zetoni.Sprazni();

    zacetek = 0;
    koda    = NULL;
}


RHTMLZetonBesedilo* RHTMLAnalizator::Razcleni(const RHTMLDokument *_dokument, TK_UNICODE _naslov_url, T_UINT znakov, T_BOOL _razcleni_okvirje, RKazalnikNapredka *_kazalnik_napredka)
{
    assert(!zetoni.VrniStevilo()); // Analizator je bil e enkrat uporabljen.
    assert(_dokument);

    T_UINT konec;
    RHTMLZetonBesedilo *zeton;

    dokument          = _dokument;
    razcleni_okvirje  = _razcleni_okvirje;
    kazalnik_napredka = _kazalnik_napredka;
    naslov_url.Nastavi(_naslov_url, znakov);

    // Pripravimo za rekurzijo.
    zacetek = 0;
    konec   = dokument->odseki.VrniStevilo();
    koda    = dokument->koda.Vrni();

    if (kazalnik_napredka) {
        // Vzpostavimo kazalnik napredka.
        kazalnik_napredka->NastaviObmocje(zacetek, konec);
        kazalnik_napredka->Nastavi(zacetek);
    }

    zeton = Razcleni(konec);
    if (!zeton)
        Sprazni();

    return zeton;
}


T_VOID RHTMLAnalizator::Sestavi(RNizU &koda, const RHTMLZetonBesedilo *zeton)
{
    assert(zeton);
    assert(
        zeton->vrsta == HTML_ZETON_ZAPRTI ||
        zeton->vrsta == HTML_ZETON_UVODNI ||
        zeton->vrsta == HTML_ZETON_ZAKLJUCNI ||
        zeton->vrsta == HTML_ZETON_ZADNJI);

    if (zeton->vrsta_besedila & HTML_BESEDILO_ZETONA_ZETONI) {
        // eton vsebuje reference na druge etone.
        TK_UNICODE koren = zeton->besedilo.Vrni();
        T_UINT i, dolzina = zeton->besedilo.VrniDolzino();
        const RHTMLZeton *zeton2;

        for (i = 0; i < dolzina && koren[i];) {
            assert(koren[i] != PR_K_KOMENTAR_KONEC);

            zeton2 = RHTMLZeton::AnalizirajOznako(koren, i);
            if (zeton2) {
                // Nali smo referenco etona.

                switch (zeton2->vrsta) {
                    case HTML_ZETON_ZAPRTI:
                    case HTML_ZETON_UVODNI:
                    case HTML_ZETON_ZAKLJUCNI:
                    case HTML_ZETON_ZADNJI:
                        Sestavi(koda, (const RHTMLZetonBesedilo*)zeton2);
                        break;
                    case HTML_ZETON_NASLOVURL:
                        switch (((const RHTMLZetonURL*)zeton2)->zapis) {
                            case HTML_ZAPIS_URL_ZETONA_UNICODE:
                                // Naslov URL je trivialen in ga lahko le prepiemo.
                                koda += ((const RHTMLZetonURL*)zeton2)->url;
                                break;
                            case HTML_ZAPIS_URL_ZETONA_HTML:
                                // Naslov URL pretvorimo v entitete SGML.
                                html_escape(koda, ((const RHTMLZetonURL*)zeton2)->url.Vrni(), ((const RHTMLZetonURL*)zeton2)->url.VrniDolzino());
                                break;
                            case HTML_ZAPIS_URL_ZETONA_CSS:
                                // Naslov URL pretvorimo v ubena zaporedja CSS.
                                css_escape(koda, ((const RHTMLZetonURL*)zeton2)->url.Vrni(), ((const RHTMLZetonURL*)zeton2)->url.VrniDolzino());
                                break;
                            default:
                                // Kaken URL je bil pa to?
                                assert(0);
                        }
                        break;
                    default:
                        // Kaken eton je bil pa to?
                        assert(0);
                }
            } else if (zeton->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) {
                // eton vsebuje tudi besedilo, ki ga moramo pretvoriti v entitete SGML.
                html_escape_min(koda, koren[i]);
                i++;
            } else
                koda += koren[i++];
        }
    } else if (zeton->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) {
        // eton sicer ne vsebuje referenc na druge etone, vsebuje pa besedilo, ki ga moramo pretvoriti v entitete SGML.
        html_escape_min(koda, zeton->besedilo.Vrni(), zeton->besedilo.VrniDolzino());
    } else {
        // eton je trivialen in ga lahko le prepiemo.
        koda += zeton->besedilo;
    }
}


#ifdef _DEBUG
T_VOID RHTMLAnalizator::IzpisiDiag(T_UINT zamik) const
{
    T_UINT i, st_zetonov = zetoni.VrniStevilo();
    for (i = 0; i < st_zetonov; i++) {
        const RHTMLZeton *zeton = zetoni[i];
        assert(zeton);
        switch (zeton->vrsta) {
            case HTML_ZETON_ZAPRTI:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(zaprti): \"%.*ls\", (odsek: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, ima besedilo: %s, ima etone: %s, instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonBesedilo*)zeton)->besedilo.VrniDolzino(), ((RHTMLZetonBesedilo*)zeton)->besedilo.Vrni(),
                    zeton->indeks_odseka,
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) ? _T("da") : _T("ne"),
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_ZETONI)   ? _T("da") : _T("ne"),
                    zeton->instanca);
                break;
            case HTML_ZETON_UVODNI:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(uvodni): \"%.*ls\", (ime elementa: \"%.*ls\", odsek: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, do odseka: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, ima besedilo: %s, ima etone: %s, instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonBesedilo*)zeton)->besedilo.VrniDolzino(), ((RHTMLZetonBesedilo*)zeton)->besedilo.Vrni(),
                    ((RHTMLZetonUvodni*)zeton)->ime_elementa.VrniDolzino(), ((RHTMLZetonUvodni*)zeton)->ime_elementa.Vrni(),
                    zeton->indeks_odseka,
                    ((RHTMLZetonUvodni*)zeton)->indeks_konca,
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) ? _T("da") : _T("ne"),
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_ZETONI)   ? _T("da") : _T("ne"),
                    zeton->instanca);
                break;
            case HTML_ZETON_ZAKLJUCNI:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(zakljucni): \"%.*ls\", (odsek: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, ima besedilo: %s, ima etone: %s, instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonBesedilo*)zeton)->besedilo.VrniDolzino(), ((RHTMLZetonBesedilo*)zeton)->besedilo.Vrni(),
                    zeton->indeks_odseka,
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) ? _T("da") : _T("ne"),
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_ZETONI)   ? _T("da") : _T("ne"),
                    zeton->instanca);
                break;
            case HTML_ZETON_NASLOVURL:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(URL): \"%.*ls\", (zapis: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, odsek: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonURL*)zeton)->url.VrniDolzino(), ((RHTMLZetonURL*)zeton)->url.Vrni(),
                    ((RHTMLZetonURL*)zeton)->zapis,
                    zeton->indeks_odseka,
                    zeton->instanca);
                break;
            case HTML_ZETON_ZADNJI:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(zadnji): \"%.*ls\", (odsek: %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u, ima besedilo: %s, ima etone: %s, instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonBesedilo*)zeton)->besedilo.VrniDolzino(), ((RHTMLZetonBesedilo*)zeton)->besedilo.Vrni(),
                    zeton->indeks_odseka,
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) ? _T("da") : _T("ne"),
                    (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_ZETONI)   ? _T("da") : _T("ne"),
                    zeton->instanca);
                break;
            case HTML_ZETON_NAPAKA:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(napaka): ") _T(T_STANJE_PRINTF) _T(" (instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    ((RHTMLZetonNapaka*)zeton)->stanje,
                    zeton->instanca);
                break;
            case HTML_ZETON_PREKINITEV:
                adebug(ADNE_DIAG,
                    ADNE_AOSN _T("%*s%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u %") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("X(prekinitev): (instanca: %#") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("x)\n"),
                    zamik, _T(""), i, (T_UINT)zeton,
                    zeton->instanca);
                break;
            default:
                assert(0); // Neznana vrsta etona.
        }
    }
}
#endif


RHTMLZeton* RHTMLAnalizator::DodajPrekinitveniZeton(T_UINT indeks_odseka, T_UINT instanca)
{
    return DodajZeton(new RHTMLZeton(HTML_ZETON_PREKINITEV, indeks_odseka, instanca));
}


RHTMLZetonBesedilo* RHTMLAnalizator::DodajZakljucniZeton(TK_UNICODE besedilo, T_UINT znakov, T_UINT vrsta_besedila, T_UINT indeks_odseka, T_UINT instanca)
{
    return (RHTMLZetonBesedilo*)DodajZeton(new RHTMLZetonBesedilo(HTML_ZETON_ZADNJI, besedilo, znakov, vrsta_besedila, indeks_odseka, instanca));
}


RHTMLZetonNapaka* RHTMLAnalizator::DodajZetonNapake(T_STANJE stanje, T_UINT instanca)
{
    return (RHTMLZetonNapaka*)DodajZeton(new RHTMLZetonNapaka(stanje, instanca));
}


T_ROCICA RHTMLAnalizator::ZapriZetone(RNizU &besedilo_prevoda, RHTMLSeznamZetonov2 &odprti_zetoni, const RHTMLSeznamZetonov2 &novi_zetoni)
{
    T_ROCICA rocica1, rocica2;

    // Poiemo, do kje se seznama ujemata.
    for (rocica1 = odprti_zetoni.VrniPrvega(), rocica2 = novi_zetoni.VrniPrvega(); rocica1; rocica1 = odprti_zetoni.VrniNaslednjega(rocica1), rocica2 = novi_zetoni.VrniNaslednjega(rocica2)) {
        if (!rocica2 || odprti_zetoni[rocica1] != novi_zetoni[rocica2]) {
            // Nali smo etona, kjer se seznama ne ujemata ve, ali pa je seznama novi_zetoni konec.
            // Zakljuimo etone, ki niso ve aktualni. V obratnem redu, kot smo jih odpirali.
            T_ROCICA r;

            for (r = odprti_zetoni.VrniZadnjega(); r; ) {
                RHTMLZetonUvodni *z1 = (RHTMLZetonUvodni*)odprti_zetoni[r];
                assert(z1 && z1->vrsta == HTML_ZETON_UVODNI);
                RHTMLZetonBesedilo *z2;

                // Naredimo zakljuni eton.
                z2 = new RHTMLZetonBesedilo(HTML_ZETON_ZAKLJUCNI);
                assert(z2);
                z2->besedilo.Prilepi(L"</");
                z2->besedilo.Prilepi(z1->ime_elementa);
                z2->besedilo.Prilepi(L'>');
                z2->besedilo.SprostiOdvecne();

                // Oznako zakljunega etona dodamo v besedilo prevoda.
                DodajOznakoZetona(z2, besedilo_prevoda);

                // Zbriemo uvodni eton, ki smo ga ravnokar zaprli. e je bil to tudi zadnji eton, ki smo ga morali izbrisati, konamo brisanje.
                odprti_zetoni.Brisi(r);
                if (rocica1 == r) break;
                r = odprti_zetoni.VrniZadnjega();
            }

            break;
        }
    }

    return rocica2;
}


T_VOID RHTMLAnalizator::PretvoriVAbsolutniURL(RNizU &rel)
{
    UNREFERENCED_PARAMETER(rel);

    if (naslov_url.JePrazen())
        return;

    //DWORD
    //    dolzina = (DWORD)(naslov_url.VrniDolzino() + rel.VrniDolzino()) + 0x100,
    //    napaka  = ERROR_SUCCESS;
    //T_UNICODE niz;
    //
    //niz = (T_UNICODE)aalloc(dolzina * sizeof(T_UZNAK));
    //assert(niz);
    //while (
    //    !::InternetCombineUrlW(naslov_url.Vrni(), rel.Vrni(), niz, &dolzina, 0) &&
    //    (napaka = ::GetLastError()) == ERROR_INSUFFICIENT_BUFFER)
    //{
    //    // Zmanjkalo je pomnilnika. V spremenljivki dolzina je zahtevana dolina niza.
    //    niz = (T_UNICODE)aqrealloc(niz, (dolzina + 1) * sizeof(T_UZNAK), FALSE);
    //    assert(niz);
    //    napaka = ERROR_SUCCESS;
    //}
    //
    //if (napaka == ERROR_SUCCESS)
    //    rel.Nastavi(niz, dolzina);
    //afree(niz);
}


RHTMLZetonBesedilo* RHTMLAnalizator::Razcleni(T_UINT konec, T_UINT vrsta_besedila)
{
    RHTMLZetonBesedilo *zeton;

    // Naredimo eton za prevajanje.
    zeton = new RHTMLZetonBesedilo(
        HTML_ZETON_ZAPRTI,
        NULL, 0,
        vrsta_besedila,
        zacetek);
    assert(zeton);

    while (zacetek < konec) {
        if (kazalnik_napredka) {
            if (kazalnik_napredka->PrekiniPostopek()) {
                // Uporabnik je zahteval prekinitev, zato zakljuimo analizo.
                goto napaka;
            }

            // Nastavimo obmoje napredka za fazo analize.
            kazalnik_napredka->Nastavi(zacetek);
        }

        // Pogledamo, kaken odsek je na vrsti.
        const RHTMLOdsek *o = dokument->odseki[zacetek];

        // Besedilo ne sme vsebovati znakov PR_K_KOMENTAR_ZACETEK in PR_K_KOMENTAR_KONEC.
        assert(
            !astrnchr(koda + o->interval.zacetek, (T_UZNAK)PR_K_KOMENTAR_ZACETEK, o->interval.VrniDolzino()) &&
            !astrnchr(koda + o->interval.zacetek, (T_UZNAK)PR_K_KOMENTAR_KONEC,   o->interval.VrniDolzino()));

        if (o->vrsta == HTML_ODSEK_BESEDILO) {
            // Nali smo navadno besedilo, ki ga prepiemo v eton za prevajanje in se pomaknemo naprej.

            // Prilepimo ga na konec niza za prevajanje.
            zeton->besedilo.PrilepiSGML(koda + o->interval.zacetek, o->interval.VrniDolzino());
            if (!(zeton->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO) &&
                !aisspace(koda + o->interval.zacetek, o->interval.VrniDolzino()))
            {
                // Za besedilo smo sprva predpostavljali, da ne bo vsebovalo besedila za prevajanje, vendar smo nali nekaj, kar bi se mogoe dalo prevesti.
                zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_BESEDILO;
            }
            zacetek++;
        } else if (o->vrsta == HTML_ODSEK_OZNAKA_SAMOSTOJNA || o->vrsta == HTML_ODSEK_OZNAKA_UVODNA) {
            // Nali smo samostojno ali uvodno oznako.
            if (((const RHTMLOznakaSamostojna*)o)->koda == HTML_EL_FRAMESET && !razcleni_okvirje) {
                // Nali smo stran z okvirji, ki pa je ne smemo razlenjati/prevajati. Zato prekinemo postopek.
                goto napaka;
            }

            {
                // Naredimo eton, ki bo predstavljal samostojno/uvodno oznako elementa.
                T_UINT i, n = ((const RHTMLOznakaSamostojna*)o)->lastnosti.VrniStevilo();
                T_UINT odmik = o->interval.zacetek;
                RHTMLZetonBesedilo *z;

                // Naredimo eton.
                if (o->vrsta == HTML_ODSEK_OZNAKA_SAMOSTOJNA || _VrniVrstoZakljucitve(((const RHTMLOznakaUvodna*)o)->koda) == ZAKLJUCITEV_NIKOLI) {
                    // Oznaka je samostojna ali pa element s to uvodno oznako nikoli ni zakljuen.
                    z = new RHTMLZetonBesedilo(
                        HTML_ZETON_ZAPRTI,
                        NULL, 0,
                        0,
                        zacetek);
                } else {
                    // Oznaka je uvodna.
                    z = new RHTMLZetonUvodni(
                        NULL, 0,
                        koda + ((const RHTMLOznakaUvodna*)o)->ime.zacetek, ((const RHTMLOznakaUvodna*)o)->ime.VrniDolzino(),
                        0,
                        zacetek,
                        ((const RHTMLOznakaUvodna*)o)->zakljucna);
                }
                assert(z);

                // Prepiemo vsebino samostojne/oznake, vendar pazimo, e ima oznaka kakne posebne lastnosti.
                for (i = 0; i < n; i++) {
                    T_HTMLLASTNOST &l = ((const RHTMLOznakaSamostojna*)o)->lastnosti[i];

                    if (!l.vrednost.VrniDolzino() ||
                        aisspace(koda + l.vrednost.zacetek, l.vrednost.VrniDolzino())) {
                        // Ne zanimajo nas lastnosti, ki nimajo vrednosti, ali pa je ta sestavljena samo iz presledkov.
                        continue;
                    } else if (_is_uri(((const RHTMLOznakaSamostojna*)o)->koda, koda + l.ime.zacetek, l.ime.VrniDolzino())) {
                        // Nali smo lastnost z naslovom URL. Vrednost lastnosti nadomestimo s svojim etonom vrste HTML_ZETON_NASLOVURL.
                        RHTMLZetonURL *z_url;

                        z->besedilo.Prilepi(koda + odmik, l.vrednost.zacetek - odmik);
                        z_url = new RHTMLZetonURL(
                            NULL, 0,
                            HTML_ZAPIS_URL_ZETONA_HTML,
                            zacetek);
                        assert(z_url);
                        html_unescape(z_url->url, koda + l.vrednost.zacetek, l.vrednost.VrniDolzino());
                        z_url->url.SprostiOdvecne();
                        DodajOznakoZetona(z_url, z->besedilo);
                        z->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
                        odmik = l.vrednost.konec;
                    } else if (_is_localizable(((const RHTMLOznakaSamostojna*)o)->koda, koda + l.ime.zacetek, l.ime.VrniDolzino())) {
                        // Nali smo lastnost z opisom, ki je prevedljiv. Vrednost lastnosti nadomestimo s svojim etonom vrste HTML_ZETON_BESEDILO.
                        RHTMLZetonBesedilo *z_vrednost;

                        z->besedilo.Prilepi(koda + odmik, l.vrednost.zacetek - odmik);
                        z_vrednost = new RHTMLZetonBesedilo(
                            HTML_ZETON_ZAPRTI,
                            NULL, 0,
                            HTML_BESEDILO_ZETONA_BESEDILO | HTML_BESEDILO_ZETONA_NASLOV,
                            zacetek);
                        assert(z_vrednost);
                        z_vrednost->besedilo.NastaviSGML(koda + l.vrednost.zacetek, l.vrednost.VrniDolzino());
                        DodajOznakoZetona(z_vrednost, z->besedilo);
                        z->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
                        odmik = l.vrednost.konec;
                    }
                }

                // Dokonamo vsebino etona in njegovo oznako prilepimo na konec niza za prevajanje.
                z->besedilo.Prilepi(koda + odmik, o->interval.konec - odmik);
                z->besedilo.SprostiOdvecne();
                DodajOznakoZetona(z, zeton->besedilo);
                zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
            }
            zacetek++;

            if (o->vrsta == HTML_ODSEK_OZNAKA_UVODNA) {
                if (((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_ADDRESS ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_CODE    ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_COMMENT ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_CITE    ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_KBD     ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_SAMP    ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_SCRIPT  ||
                    ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_STYLE)
                {
                    // Oznaka predstavlja element, ki ga ne smemo prevajati.
                    T_UINT idx_konec = ((const RHTMLOznakaUvodna*)o)->zakljucna;
                    const RHTMLOdsek *o_konec = dokument->odseki[idx_konec];
                    assert(o_konec);

                    if (o->interval.konec < o_konec->interval.zacetek) {
                        // eton ima nekaj vsebine, ki jo zamenjamo z nadomestnim etonom.
                        if (((const RHTMLOznakaUvodna*)o)->koda != HTML_EL_STYLE) {
                            // Naredimo neprevedljivi eton, ki bo predstavljal celotno vsebino tega elementa. S podelementi vred, e jih ima.
                            DodajOznakoZetona(
                                new RHTMLZetonBesedilo(
                                    HTML_ZETON_ZAPRTI,
                                    koda + o->interval.konec, o_konec->interval.zacetek - o->interval.konec,
                                    0,
                                    zacetek),
                                zeton->besedilo);
                        } else {
                            // Oznaka predstavlja zaetek navodila CSS, ki ga delno razlenimo za potrebe absolutizacije naslovov URL.
                            RHTMLZetonBesedilo *z;

                            z = RazcleniCSS(o->interval.konec, o_konec->interval.zacetek);
                            assert(z);

                            // Prilepimo oznako etona k nizu za prevajanje.
                            z->PrilepiOznako(zeton->besedilo);
                        }
                        zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
                    }

                    // Prestavimo se na konec vsebine elementa.
                    zacetek = idx_konec;
                } else if (_is_group(((const RHTMLOznakaUvodna*)o)->koda)) {
                    // Blokovni element obdelamo rekurzivno. Pravzaprav uvedemo razirjeno mnoico elementov, ki jih obravnavamo kot "blokovne".
                    RHTMLZeton *z;

                    // Kliemo rekurzivno razlembo blokovnega elementa.
                    z = Razcleni(
                        amin<T_UINT>(konec, ((const RHTMLOznakaUvodna*)o)->zakljucna),
                        (_is_heading(((const RHTMLOznakaUvodna*)o)->koda) || ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_DT || ((const RHTMLOznakaUvodna*)o)->koda == HTML_EL_TITLE ? HTML_BESEDILO_ZETONA_NASLOV : 0) |
                        (   _is_list(((const RHTMLOznakaUvodna*)o)->koda)  ? HTML_BESEDILO_ZETONA_NASTEVANJE : 0));
                    if (!z) {
                        // Blokovni element je vseboval elemente, ki jih ne smemo prevajati, ali pa je uporabnik prekinil razlembo.
                        goto napaka;
                    }

                    // Prilepimo oznako etona k nizu za prevajanje.
                    z->PrilepiOznako(zeton->besedilo);
                    zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
                }
            }
        } else if (o->vrsta == HTML_ODSEK_OZNAKA_ZAKLJUCNA) {
            // Nali smo zakljuno oznako.

            // Naredimo eton, ki bo predstavljal zakljuno oznako elementa.
            DodajOznakoZetona(
                new RHTMLZetonBesedilo(
                    HTML_ZETON_ZAKLJUCNI,
                    koda + o->interval.zacetek, o->interval.VrniDolzino(),
                    0,
                    zacetek),
                zeton->besedilo);
            zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
            zacetek++;
        } else {
            // Nali smo deklaracijo, navodilo, odsek (P)CDATA, komentar ...

            // Naredimo neprevedljivi eton, ki bo predstavljal ta odsek.
            DodajOznakoZetona(
                new RHTMLZetonBesedilo(
                    HTML_ZETON_ZAPRTI,
                    koda + o->interval.zacetek, o->interval.VrniDolzino(),
                    0,
                    zacetek),
                zeton->besedilo);
            zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
            zacetek++;
        }
    }

    // Tako, imamo eton, ki predstavlja blokovni element.
    zeton->besedilo.SprostiOdvecne();
    return (RHTMLZetonBesedilo*)DodajZeton(zeton);

napaka:
    delete zeton;
    return NULL;
}


RHTMLZetonBesedilo* RHTMLAnalizator::RazcleniCSS(T_UINT zacetek, T_UINT konec)
{
    RHTMLZetonBesedilo *zeton;
    T_INTERVAL<T_UINT> odsek, vsebina;

    // Naredimo eton za prevajanje.
    zeton = new RHTMLZetonBesedilo(
        HTML_ZETON_ZAPRTI,
        NULL, 0,
        0,
        this->zacetek);
    assert(zeton);

    for (;;) {
        if (re_css_komentar.PreveriUjemanje(koda, zacetek, konec)) {
            // Nali smo komentar, ki ga prepiemo v niz za prevod.
            zeton->besedilo.Prilepi(koda + zacetek, re_css_komentar.interval.konec - zacetek);
            zacetek = re_css_komentar.interval.konec;
        } else if (re_css_cdo.PreveriUjemanje(koda, zacetek, konec)) {
            // Nali smo oznako CDO, ki jo prepiemo v niz za prevod.
            zeton->besedilo.Prilepi(koda + zacetek, re_css_cdo.interval.konec - zacetek);
            zacetek = re_css_cdo.interval.konec;
        } else if (re_css_cdc.PreveriUjemanje(koda, zacetek, konec)) {
            // Nali smo oznako CDC, ki jo prepiemo v niz za prevod.
            zeton->besedilo.Prilepi(koda + zacetek, re_css_cdc.interval.konec - zacetek);
            zacetek = re_css_cdc.interval.konec;
        } else if (
            re_css_import.PreveriUjemanje(koda, zacetek, konec) && (odsek = re_css_import.interval, vsebina = re_css_import.vsebina, TRUE) ||
               re_css_uri.PreveriUjemanje(koda, zacetek, konec) && (odsek =    re_css_uri.interval, vsebina =    re_css_uri.vsebina, TRUE))
        {
            // Nali smo navodilo "@import '...'" ali oznako URI "url(...)".
            RHTMLZetonURL *z;

            // Najprej pripravimo eton z URL-jem.
            z = new RHTMLZetonURL(
                NULL, 0,
                HTML_ZAPIS_URL_ZETONA_CSS,
                this->zacetek);
            assert(z);
            css_unescape(z->url, koda + vsebina.zacetek, vsebina.VrniDolzino());
            z->url.SprostiOdvecne();

            // Prilepimo niz do etona, oznako etona in nato od etona naprej.
            zeton->besedilo.Prilepi(koda + zacetek, vsebina.zacetek - zacetek);
            DodajOznakoZetona(z, zeton->besedilo);
            zeton->besedilo.Prilepi(koda + vsebina.konec, odsek.konec - vsebina.konec);

            zeton->vrsta_besedila |= HTML_BESEDILO_ZETONA_ZETONI;
            zacetek = odsek.konec;
        } else if (poljuben_znak.PreveriUjemanje(koda, zacetek, konec)) {
            // Vse druge znake prepisujemo v niz za prevod.
            zeton->besedilo.Prilepi(koda + zacetek, poljuben_znak.interval.konec - zacetek);
            zacetek = poljuben_znak.interval.konec;
        } else {
            // Prili smo do konca bloka CSS.
            break;
        }
    }

    // Tako, imamo eton, ki predstavlja blokovni element.
    zeton->besedilo.SprostiOdvecne();
    return (RHTMLZetonBesedilo*)DodajZeton(zeton);
}


//////////////////////////////////////////////////////////////////////
// Interne funkcije inline
//////////////////////////////////////////////////////////////////////

inline T_HTMLVRSTAZAKLJUCITVE _VrniVrstoZakljucitve(T_HTMLKODAELEMENTA koda)
{
    return HTML_EL_PRVI <= koda && koda < HTML_EL_ZADNJI ?
        ::_vedenje_elementov[koda - HTML_EL_PRVI].zakljucitev :
        ZAKLJUCITEV_VEDNO;
}


inline T_HTMLPRIKAZELEMENTA _VrniPrikazElementa(T_HTMLKODAELEMENTA koda)
{
    return HTML_EL_PRVI <= koda && koda < HTML_EL_ZADNJI ?
        ::_vedenje_elementov[koda - HTML_EL_PRVI].prikaz :
        PRIKAZ_VPETO;
}


inline T_BOOL _is_fontstyle(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_TT:
        case HTML_EL_I:
        case HTML_EL_B:
        case HTML_EL_U:
        case HTML_EL_S:
        case HTML_EL_STRIKE:
        case HTML_EL_BLINK:
        case HTML_EL_BIG:
        case HTML_EL_SMALL:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_phrase(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_EM:
        case HTML_EL_STRONG:
        case HTML_EL_DFN:
        case HTML_EL_CODE:
        case HTML_EL_SAMP:
        case HTML_EL_KBD:
        case HTML_EL_VAR:
        case HTML_EL_CITE:
        case HTML_EL_ABBR:
        case HTML_EL_ACRONYM:
        case HTML_EL_XMP:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_special(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_A:
        case HTML_EL_IMG:
        case HTML_EL_APPLET:
        case HTML_EL_OBJECT:
        case HTML_EL_EMBED:
        case HTML_EL_FONT:
        case HTML_EL_BASEFONT:
        case HTML_EL_BR:
        case HTML_EL_WBR:
        case HTML_EL_RT:
        case HTML_EL_SCRIPT:
        case HTML_EL_MAP:
        case HTML_EL_Q:
        case HTML_EL_SUB:
        case HTML_EL_SUP:
        case HTML_EL_RUBY:
        case HTML_EL_SPAN:
        case HTML_EL_BDO:
        case HTML_EL_IFRAME:
        case HTML_EL_NOBR:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_formctrl(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_INPUT:
        case HTML_EL_SELECT:
        case HTML_EL_TEXTAREA:
        case HTML_EL_LABEL:
        case HTML_EL_BUTTON:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_inline(T_HTMLKODAELEMENTA koda)
{
    return
        koda == HTML_EL_PCDATA ||
        _is_fontstyle(koda) ||
        _is_phrase(koda) ||
        _is_special(koda) ||
        _is_formctrl(koda);
}


inline T_BOOL _is_heading(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_H1:
        case HTML_EL_H2:
        case HTML_EL_H3:
        case HTML_EL_H4:
        case HTML_EL_H5:
        case HTML_EL_H6:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_list(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_UL:
        case HTML_EL_OL:
        case HTML_EL_DIR:
        case HTML_EL_MENU:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_preformatted(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_PRE:
        case HTML_EL_LISTING:
            return TRUE;
    }
    return FALSE;
}


inline T_BOOL _is_block(T_HTMLKODAELEMENTA koda)
{
    if (_is_heading(koda) ||
        _is_list(koda) ||
        _is_preformatted(koda)) return TRUE;
    switch (koda) {
        case HTML_EL_P:
        case HTML_EL_DL:
        case HTML_EL_DIV:
        case HTML_EL_CENTER:
        case HTML_EL_MARQUEE:
        case HTML_EL_NOSCRIPT:
        case HTML_EL_NOFRAMES:
        case HTML_EL_NOEMBED:
        case HTML_EL_BLOCKQUOTE:
        case HTML_EL_FORM:
        case HTML_EL_ISINDEX:
        case HTML_EL_HR:
        case HTML_EL_TABLE:
        case HTML_EL_FIELDSET:
        case HTML_EL_ADDRESS:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_flow(T_HTMLKODAELEMENTA koda)
{
    return _is_block(koda) || _is_inline(koda);
}


inline T_BOOL _is_head_content(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_TITLE:
        case HTML_EL_ISINDEX:
        case HTML_EL_BASE:
        case HTML_EL_NEXTID:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_head_misc(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_SCRIPT:
        case HTML_EL_STYLE:
        case HTML_EL_META:
        case HTML_EL_LINK:
        case HTML_EL_OBJECT:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_pre_exclusion(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_IMG:
        case HTML_EL_OBJECT:
        case HTML_EL_APPLET:
        case HTML_EL_EMBED:
        case HTML_EL_BIG:
        case HTML_EL_SMALL:
        case HTML_EL_SUB:
        case HTML_EL_SUP:
        case HTML_EL_RUBY:
        case HTML_EL_FONT:
        case HTML_EL_BASEFONT:
        case HTML_EL_NOBR:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_html_content(T_HTMLKODAELEMENTA koda)
{
    switch (koda) {
        case HTML_EL_HEAD:
        case HTML_EL_BODY:
        case HTML_EL_FRAMESET:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _is_group(T_HTMLKODAELEMENTA koda)
{
    if (_is_block(koda) ||
        _is_html_content(koda) ||
        _is_head_content(koda)) return TRUE;
    switch (koda) {
        case HTML_EL_COL:
        case HTML_EL_COLGROUP:
        case HTML_EL_DD:
        case HTML_EL_DIR:
        case HTML_EL_DT:
        case HTML_EL_FRAME:
        case HTML_EL_IFRAME:
        case HTML_EL_LEGEND:
        case HTML_EL_TD:
        case HTML_EL_TH:
        case HTML_EL_TR:
            return TRUE;
    };
    return FALSE;
}


inline T_BOOL _can_contain(T_HTMLKODAELEMENTA koda_nad, T_HTMLKODAELEMENTA koda_pod)
{
    if (koda_pod == HTML_EL_NEZNAN || koda_pod == HTML_EL_COMMENT)
        return TRUE; // Za neznane elemente in komentarje privzamemo, da so lahko vsebovani v kateremkoli elementu.
    if (_is_fontstyle(koda_nad) || _is_phrase(koda_nad))
        return _is_inline(koda_pod);
    if (_is_heading(koda_nad))
        return _is_inline(koda_pod);

    switch (koda_nad) {
        case HTML_EL_A:             return _is_inline(koda_pod) && koda_pod != HTML_EL_A;
        case HTML_EL_ADDRESS:       return _is_inline(koda_pod) || koda_pod == HTML_EL_P;
        case HTML_EL_APPLET:        return _is_flow(koda_pod) || koda_pod == HTML_EL_PARAM;
        case HTML_EL_AREA:          return FALSE;
        case HTML_EL_BASE:          return FALSE;
        case HTML_EL_BASEFONT:      return FALSE;
        case HTML_EL_BDO:           return _is_inline(koda_pod);
        case HTML_EL_BLOCKQUOTE:    return _is_flow(koda_pod);
        case HTML_EL_BODY:          return _is_flow(koda_pod) || koda_pod == HTML_EL_INS || koda_pod == HTML_EL_DEL;
        case HTML_EL_BR:            return FALSE;
        case HTML_EL_BUTTON:        return _is_flow(koda_pod) && !_is_formctrl(koda_pod) && koda_pod != HTML_EL_A && koda_pod != HTML_EL_FORM && koda_pod != HTML_EL_ISINDEX && koda_pod != HTML_EL_FIELDSET && koda_pod != HTML_EL_IFRAME;
        case HTML_EL_CAPTION:       return _is_inline(koda_pod);
        case HTML_EL_CENTER:        return _is_flow(koda_pod);
        case HTML_EL_COL:           return FALSE;
        case HTML_EL_COLGROUP:      return koda_pod == HTML_EL_COL;
        case HTML_EL_COMMENT:       return koda_pod == HTML_EL_CDATA;
        case HTML_EL_DD:            return _is_flow(koda_pod);
        case HTML_EL_DEL:           return _is_flow(koda_pod);
        case HTML_EL_DIR:           return koda_pod == HTML_EL_LI;
        case HTML_EL_DIV:           return _is_flow(koda_pod);
        case HTML_EL_DL:            return koda_pod == HTML_EL_DT || koda_pod == HTML_EL_DD;
        case HTML_EL_DT:            return _is_inline(koda_pod);
        case HTML_EL_EMBED:         return _is_flow(koda_pod) || koda_pod == HTML_EL_PARAM;
        case HTML_EL_FIELDSET:      return _is_flow(koda_pod) || koda_pod == HTML_EL_LEGEND || koda_pod == HTML_EL_PCDATA;
        case HTML_EL_FONT:          return _is_inline(koda_pod);
        case HTML_EL_FORM:          return _is_flow(koda_pod) && koda_pod != HTML_EL_FORM;
        case HTML_EL_FRAME:         return FALSE;
        case HTML_EL_FRAMESET:      return koda_pod == HTML_EL_FRAMESET || koda_pod == HTML_EL_FRAME || koda_pod == HTML_EL_NOFRAMES;
        case HTML_EL_HEAD:          return _is_head_content(koda_pod) || _is_head_misc(koda_pod);
        case HTML_EL_HR:            return FALSE;
        case HTML_EL_HTML:          return _is_html_content(koda_pod);
        case HTML_EL_IFRAME:        return _is_flow(koda_pod);
        case HTML_EL_IMG:           return FALSE;
        case HTML_EL_INPUT:         return FALSE;
        case HTML_EL_INS:           return _is_flow(koda_pod);
        case HTML_EL_ISINDEX:       return FALSE;
        case HTML_EL_LABEL:         return _is_inline(koda_pod) && koda_pod != HTML_EL_LABEL;
        case HTML_EL_LEGEND:        return _is_inline(koda_pod);
        case HTML_EL_LI:            return _is_flow(koda_pod);
        case HTML_EL_LINK:          return FALSE;
        case HTML_EL_LISTING:       return koda_pod == HTML_EL_CDATA;
        case HTML_EL_MAP:           return _is_block(koda_pod) || koda_pod == HTML_EL_AREA;
        case HTML_EL_MARQUEE:       return _is_flow(koda_pod);
        case HTML_EL_MENU:          return koda_pod == HTML_EL_LI;
        case HTML_EL_META:          return FALSE;
        case HTML_EL_NOBR:          return _is_inline(koda_pod) || koda_pod == HTML_EL_WBR;
        case HTML_EL_NOFRAMES:      return (_is_flow(koda_pod) || koda_pod == HTML_EL_BODY) && koda_pod != HTML_EL_NOFRAMES;
        case HTML_EL_NOSCRIPT:      return _is_flow(koda_pod);
        case HTML_EL_NOEMBED:       return _is_flow(koda_pod);
        case HTML_EL_OBJECT:        return _is_flow(koda_pod) || koda_pod == HTML_EL_PARAM;
        case HTML_EL_OL:            return koda_pod == HTML_EL_LI;
        case HTML_EL_OPTGROUP:      return koda_pod == HTML_EL_OPTION;
        case HTML_EL_OPTION:        return koda_pod == HTML_EL_PCDATA;
        case HTML_EL_P:             return _is_inline(koda_pod);
        case HTML_EL_PARAM:         return FALSE;
        case HTML_EL_PLAINTEXT:     return _is_flow(koda_pod);
        case HTML_EL_PRE:           return _is_inline(koda_pod) && !_is_pre_exclusion(koda_pod);
        case HTML_EL_Q:             return _is_inline(koda_pod);
        case HTML_EL_RT:            return FALSE;
        case HTML_EL_RUBY:          return _is_inline(koda_pod);
        case HTML_EL_SCRIPT:        return koda_pod == HTML_EL_CDATA;
        case HTML_EL_SELECT:        return koda_pod == HTML_EL_OPTGROUP || koda_pod == HTML_EL_OPTION;
        case HTML_EL_SPAN:          return _is_inline(koda_pod);
        case HTML_EL_STYLE:         return koda_pod == HTML_EL_CDATA;
        case HTML_EL_SUB:           return _is_inline(koda_pod);
        case HTML_EL_SUP:           return _is_inline(koda_pod);
        case HTML_EL_TABLE:         return koda_pod == HTML_EL_CAPTION || koda_pod == HTML_EL_COL || koda_pod == HTML_EL_COLGROUP || koda_pod == HTML_EL_THEAD || koda_pod == HTML_EL_TFOOT || koda_pod == HTML_EL_TBODY;
        case HTML_EL_TBODY:         return koda_pod == HTML_EL_TR;
        case HTML_EL_TD:            return _is_flow(koda_pod);
        case HTML_EL_TEXTAREA:      return koda_pod == HTML_EL_PCDATA;
        case HTML_EL_TFOOT:         return koda_pod == HTML_EL_TR;
        case HTML_EL_TH:            return _is_flow(koda_pod);
        case HTML_EL_THEAD:         return koda_pod == HTML_EL_TR;
        case HTML_EL_TITLE:         return koda_pod == HTML_EL_PCDATA;
        case HTML_EL_TR:            return koda_pod == HTML_EL_TD || koda_pod == HTML_EL_TH;
        case HTML_EL_UL:            return koda_pod == HTML_EL_LI;
        case HTML_EL_WBR:           return FALSE;

        case HTML_EL_NEZNAN:        return TRUE; // Za neznane elemente privzamemo, da lahko vsebujejo katerikoli element.
    }

    return FALSE;
}


inline T_BOOL _is_uri(T_HTMLKODAELEMENTA koda, TK_UNICODE ime_lastnosti, T_UINT znakov)
{
    assert(ime_lastnosti || !znakov);

    switch (koda) {
        case HTML_EL_A:          return !astrnicmp(ime_lastnosti, znakov, L"href",        T_UINT_MAX);
        case HTML_EL_APPLET:     return !astrnicmp(ime_lastnosti, znakov, L"code",        T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"codebase",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_AREA:       return !astrnicmp(ime_lastnosti, znakov, L"href",        T_UINT_MAX);
        case HTML_EL_BASE:       return !astrnicmp(ime_lastnosti, znakov, L"href",        T_UINT_MAX);
        case HTML_EL_BGSOUND:    return !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_BLOCKQUOTE: return !astrnicmp(ime_lastnosti, znakov, L"cite",        T_UINT_MAX);
        case HTML_EL_BODY:       return !astrnicmp(ime_lastnosti, znakov, L"background",  T_UINT_MAX);
        case HTML_EL_COMMENT:    return !astrnicmp(ime_lastnosti, znakov, L"data",        T_UINT_MAX);
        case HTML_EL_DEL:        return !astrnicmp(ime_lastnosti, znakov, L"cite",        T_UINT_MAX);
        case HTML_EL_EMBED:      return !astrnicmp(ime_lastnosti, znakov, L"pluginspage", T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_FORM:       return !astrnicmp(ime_lastnosti, znakov, L"action",      T_UINT_MAX);
        case HTML_EL_FRAME:      return !astrnicmp(ime_lastnosti, znakov, L"longdesc",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_HEAD:       return !astrnicmp(ime_lastnosti, znakov, L"profile",     T_UINT_MAX);
        case HTML_EL_IFRAME:     return !astrnicmp(ime_lastnosti, znakov, L"longdesc",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_IMG:        return !astrnicmp(ime_lastnosti, znakov, L"longdesc",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"lowsrc",      T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"usemap",      T_UINT_MAX);
        case HTML_EL_INPUT:      return !astrnicmp(ime_lastnosti, znakov, L"lowsrc",      T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"usemap",      T_UINT_MAX);
        case HTML_EL_INS:        return !astrnicmp(ime_lastnosti, znakov, L"cite",        T_UINT_MAX);
        case HTML_EL_LINK:       return !astrnicmp(ime_lastnosti, znakov, L"href",        T_UINT_MAX);
        case HTML_EL_OBJECT:     return !astrnicmp(ime_lastnosti, znakov, L"basehref",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"classid",     T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"code",        T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"codebase",    T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"data",        T_UINT_MAX) ||
                                        !astrnicmp(ime_lastnosti, znakov, L"usemap",      T_UINT_MAX);
        case HTML_EL_Q:          return !astrnicmp(ime_lastnosti, znakov, L"cite",        T_UINT_MAX);
        case HTML_EL_SCRIPT:     return !astrnicmp(ime_lastnosti, znakov, L"src",         T_UINT_MAX);
        case HTML_EL_TABLE:      return !astrnicmp(ime_lastnosti, znakov, L"background",  T_UINT_MAX);
        case HTML_EL_TD:         return !astrnicmp(ime_lastnosti, znakov, L"background",  T_UINT_MAX);
        case HTML_EL_TH:         return !astrnicmp(ime_lastnosti, znakov, L"background",  T_UINT_MAX);
    }
    return FALSE;
}


inline T_BOOL _is_localizable(T_HTMLKODAELEMENTA koda, TK_UNICODE ime_lastnosti, T_UINT znakov)
{
    assert(ime_lastnosti || !znakov);

    if (!astrnicmp(ime_lastnosti, znakov, L"title", T_UINT_MAX))
        return TRUE;    // Lastnost title je splona za vse elemente.

    switch (koda) {
        case HTML_EL_APPLET:     return !astrnicmp(ime_lastnosti, znakov, L"alt",     T_UINT_MAX);
        case HTML_EL_AREA:       return !astrnicmp(ime_lastnosti, znakov, L"alt",     T_UINT_MAX);
        case HTML_EL_IMG:        return !astrnicmp(ime_lastnosti, znakov, L"alt",     T_UINT_MAX);
        case HTML_EL_INPUT:      return !astrnicmp(ime_lastnosti, znakov, L"alt",     T_UINT_MAX);
        case HTML_EL_OBJECT:     return !astrnicmp(ime_lastnosti, znakov, L"alt",     T_UINT_MAX);
        case HTML_EL_TABLE:      return !astrnicmp(ime_lastnosti, znakov, L"summary", T_UINT_MAX);
        case HTML_EL_TD:         return !astrnicmp(ime_lastnosti, znakov, L"abbr",    T_UINT_MAX);
        case HTML_EL_TH:         return !astrnicmp(ime_lastnosti, znakov, L"abbr",    T_UINT_MAX);
    }

    return FALSE;
}

#endif // AOSN_HTML
