/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_HTTP


//////////////////////////////////////////////////////////////////////
// RRegElHTTPPrelomVrstice
//////////////////////////////////////////////////////////////////////

RRegElHTTPPrelomVrstice::RRegElHTTPPrelomVrstice() : RRegElementA()
{
}


T_BOOL RRegElHTTPPrelomVrstice::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    // Preverimo, ali gre za enega izmed dovoljenih prelomov.
    if (interval.konec < konec && besedilo[interval.konec]) {
        if (besedilo[interval.konec] == '\r') {
            // Nali smo CR. Pogledamo, e sledi e LF.
            interval.konec++;
            if (interval.konec < konec && besedilo[interval.konec] == '\n') {
                // Nali smo tudi LF.
                interval.zacetek = zacetek;
                interval.konec++;
                return TRUE;
            }
        } else if (besedilo[interval.konec] == '\n') {
            // Nali smo le LF.
            interval.zacetek = zacetek;
            interval.konec++;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPZetonRFC2616
//////////////////////////////////////////////////////////////////////

RRegElHTTPZetonRFC2616::RRegElHTTPZetonRFC2616() : RRegElementA()
{
}


T_BOOL RRegElHTTPZetonRFC2616::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                (unsigned int)besedilo[interval.konec] == 0x7f ||
                besedilo[interval.konec] == '('  ||
                besedilo[interval.konec] == ')'  ||
                besedilo[interval.konec] == '<'  ||
                besedilo[interval.konec] == '>'  ||
                besedilo[interval.konec] == '@'  ||
                besedilo[interval.konec] == ','  ||
                besedilo[interval.konec] == ';'  ||
                besedilo[interval.konec] == ':'  ||
                besedilo[interval.konec] == '\\' ||
                besedilo[interval.konec] == '\"' ||
                besedilo[interval.konec] == '/'  ||
                besedilo[interval.konec] == '['  ||
                besedilo[interval.konec] == ']'  ||
                besedilo[interval.konec] == '?'  ||
                besedilo[interval.konec] == '='  ||
                besedilo[interval.konec] == '{'  ||
                besedilo[interval.konec] == '}'  ||
                aisspace(besedilo[interval.konec]))
            {
                // Nali smo kontrolni znak ali loilo, ki prekinja eton.
                break;
            } else {
                // Preskoimo vsak drug znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (zacetek < interval.konec) {
        // Nali smo eton.
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPNizRFC2616
//////////////////////////////////////////////////////////////////////

RRegElHTTPNizRFC2616::RRegElHTTPNizRFC2616() : RRegElementA()
{
}


T_BOOL RRegElHTTPNizRFC2616::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    // Preverimo, ali se niz zane z uvodnim narekovajem.
    if (interval.konec < konec && besedilo[interval.konec] != '\"')
        goto napaka;

    interval.konec++;
    vsebina.zacetek = interval.konec;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (((unsigned int)besedilo[interval.konec] < 0x20 ||
                 (unsigned int)besedilo[interval.konec] == 0x7f) &&
                !aisspace(besedilo[interval.konec]))
            {
                // Nali smo kontrolni znak, ki ni presledek. Kontrolni znaki morajo biti zapisani zaiteno.
                goto napaka;
            } else if (besedilo[interval.konec] == '\"') {
                // Nali smo zakljuni narekovaj.
                vsebina.konec = interval.konec;
                interval.konec++;
                break;
            } else if (besedilo[interval.konec] == '\\') {
                // Nali smo zaitni znak. Preskoimo njega in znak, ki sledi.
                interval.konec++;
                if (interval.konec < konec && besedilo[interval.konec]) {
                    interval.konec++;
                } else {
                    // Zmanjkalo nam je znakov.
                    goto napaka;
                }
            } else {
                // Preskoimo vsak drug znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je znakov.
            goto napaka;
        }
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    vsebina.zacetek  = 1;
    vsebina.konec    = 0;

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPNizRFC2616::Razveljavi()
{
    // Razveljavimo podelemente.
    vsebina.zacetek  = 1;
    vsebina.konec    = 0;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPVrednostRFC2109
//////////////////////////////////////////////////////////////////////

RRegElHTTPVrednostRFC2109::RRegElHTTPVrednostRFC2109() : RRegElementA()
{
}


T_BOOL RRegElHTTPVrednostRFC2109::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;
    if (niz.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Vrednost je zapisana kot niz.
        zeton.Razveljavi();
        interval.konec   = niz.interval.konec;
        interval.zacetek = zacetek;
        return TRUE;
    } else if (zeton.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Vrednost je zapisana kot eton.
        niz.Razveljavi();
        interval.konec   = zeton.interval.konec;
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        // Ni vrednost.
        niz.Razveljavi();
        zeton.Razveljavi();

        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


T_VOID RRegElHTTPVrednostRFC2109::Razveljavi()
{
    // Razveljavimo podelemente.
    niz.Razveljavi();
    zeton.Razveljavi();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPStreznikURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPStreznikURL::RRegElHTTPStreznikURL() : RRegElementA()
{
}


T_BOOL RRegElHTTPStreznikURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                (unsigned int)besedilo[interval.konec] == 0x7f ||
                besedilo[interval.konec] == ':' ||
                besedilo[interval.konec] == '/' ||
                aisspace(besedilo[interval.konec]))
            {
                // Nali smo kontrolni znak ali loilo, ki prekinja ime strenika.
                break;
            } else {
                // Preskoimo vsak drug znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (zacetek < interval.konec) {
        // Nali smo ime strenika.
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPVrataURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPVrataURL::RRegElHTTPVrataURL() :
    RRegElementA(),
    vrednost(0)
{
}


T_BOOL RRegElHTTPVrataURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    vrednost = 0;
    interval.konec = zacetek;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ('0' <= besedilo[interval.konec] && besedilo[interval.konec] <= '9') {
                // Nali smo tevko.
                T_UINT _vrednost = vrednost * 10 + (besedilo[interval.konec] - '0');
                if (_vrednost > T_U2_MAX) {
                    // tevilka je e prevelika.
                    vrednost = 0;
                    interval.zacetek = (interval.konec = zacetek) + 1;
                    return FALSE;
                }
                vrednost = (T_U2)_vrednost;
                interval.konec++;
            } else {
                // Nali smo znak, ki prekinja tevilko vrat.
                break;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (zacetek < interval.konec) {
        // Nali smo vrata.
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


T_VOID RRegElHTTPVrataURL::Razveljavi()
{
    // Razveljavimo podelemente.
    vrednost = 0;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPSegmentPotiURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPSegmentPotiURL::RRegElHTTPSegmentPotiURL() : RRegElementA()
{
}


T_BOOL RRegElHTTPSegmentPotiURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                (unsigned int)besedilo[interval.konec] == 0x7f ||
                besedilo[interval.konec] == '?' ||
                besedilo[interval.konec] == '/' ||
                aisspace(besedilo[interval.konec]))
            {
                // Nali smo kontrolni znak ali loilo, ki prekinja segment poti.
                break;
            } else {
                // Preskoimo vsak drug znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    interval.zacetek = zacetek;
    return TRUE;
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPPotURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPPotURL::RRegElHTTPPotURL() : RRegElementA()
{
}


T_BOOL RRegElHTTPPotURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    RRegElHTTPSegmentPotiURL s;

    interval.konec = zacetek;
    segmenti.Sprazni();

    // Pot se mora zaeti s poevnico.
    if (interval.konec < konec && besedilo[interval.konec] != '/')
        goto napaka;
    interval.konec++;

    // Dodamo prvi segment poti.
    verify(s.PreveriUjemanje(besedilo, interval.konec, konec, zastavice));
    segmenti += s;
    interval.konec = s.interval.konec;

    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (besedilo[interval.konec] == '/') {
                // Nali smo nov segment poti.
                interval.konec++;
                verify(s.PreveriUjemanje(besedilo, interval.konec, konec, zastavice));
                segmenti += s;
                interval.konec = s.interval.konec;
            } else {
                // Nali smo kontrolni znak ali loilo, ki konuje pot.
                break;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    interval.zacetek = zacetek;
    segmenti.SprostiOdvecne();
    return TRUE;

napaka:
    segmenti.Sprazni();
    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPPotURL::Razveljavi()
{
    // Razveljavimo podelemente.
    segmenti.Sprazni();
    segmenti.SprostiOdvecne();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPParameterURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPParameterURL::RRegElHTTPParameterURL() : RRegElementA()
{
}


T_BOOL RRegElHTTPParameterURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    // Doloimo ime parametra.
    ime.zacetek = interval.konec;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                (unsigned int)besedilo[interval.konec] == 0x7f ||
                besedilo[interval.konec] == '&'  ||
                besedilo[interval.konec] == '='  ||
                aisspace(besedilo[interval.konec]))
            {
                // Nali smo kontrolni znak ali loilo, ki prekinja ime parametra.
                break;
            } else {
                // Preskoimo vsak drug znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (zacetek < interval.konec) {
        ime.konec = interval.konec;
    } else {
        // Nismo nali imena parametra.
        goto napaka;
    }

    if (besedilo[interval.konec] == '=') {
        // Sledi vrednost parametra.
        interval.konec++;

        // Doloimo vrednost parametra.
        vrednost.zacetek = interval.konec;
        for (;;) {
            if (interval.konec < konec && besedilo[interval.konec]) {
                if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                    (unsigned int)besedilo[interval.konec] == 0x7f ||
                    besedilo[interval.konec] == '&'  ||
                    aisspace(besedilo[interval.konec]))
                {
                    // Nali smo kontrolni znak ali loilo, ki prekinja vrednost parametra.
                    break;
                } else {
                    // Preskoimo vsak drug znak.
                    interval.konec++;
                }
            } else {
                // Zmanjkalo nam je znakov.
                break;
            }
        }

        vrednost.konec = interval.konec;
    } else {
        // Nismo nali vrednosti parametra.
        vrednost.zacetek = 1;
        vrednost.konec   = 0;
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    ime.zacetek      = 1;
    ime.konec        = 0;
    vrednost.zacetek = 1;
    vrednost.konec   = 0;

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPParameterURL::Razveljavi()
{
    // Razveljavimo podelemente.
    ime.zacetek      = 1;
    ime.konec        = 0;
    vrednost.zacetek = 1;
    vrednost.konec   = 0;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPURL
//////////////////////////////////////////////////////////////////////

RRegElHTTPURL::RRegElHTTPURL() : RRegElementA()
{
}


RRegElHTTPURL::~RRegElHTTPURL()
{
    parametri.Unici();
}


T_BOOL RRegElHTTPURL::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;

    // Ali se URL zaenja s "http://"?
    static const T_AZNAK uvod[] = "http://";
    if (interval.konec + _countof(uvod) <= konec && astrnicmp(besedilo + interval.konec, _countof(uvod), uvod, _countof(uvod)) == 0) {
        // Pot bo absolutna.
        interval.konec += _countof(uvod);

        // Doloimo ime strenika.
        if (streznik.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            interval.konec = streznik.interval.konec;
        } else {
            // Pot je absolutna, a imena strenika nismo mogli doloiti. URL je napaen.
            goto napaka;
        }

        if (interval.konec < konec && besedilo[interval.konec] == ':') {
            // Imenu strenika sledi zapis vrat.
            interval.konec++;
            if (vrata.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
                // Nali smo vrata.
                interval.konec = vrata.interval.konec;
            } else {
                // Za ":" ni sledil zapis vrat, kar pogojno toleriramo.
            }
        } else {
            // Manjka zapis vrat, zato privzamemo vrata 80.
            vrata.Razveljavi();
            vrata.vrednost = 80;
        }
    } else {
        // Pot bo relativna.
        streznik.Razveljavi();
        vrata.Razveljavi();
        vrata.vrednost = 80;
    }

    // Zdaj mora slediti pot.
    if (pot.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        interval.konec = pot.interval.konec;
    } else {
        // e ni poti, je URL napaen. Imeti moramo vsaj eno poevnico.
        goto napaka;
    }

    parametri.Unici();

    if (interval.konec < konec && besedilo[interval.konec] == '?') {
        // Poti sledijo parametri.
        interval.konec++;

        for (;;) {
            if (interval.konec < konec && besedilo[interval.konec]) {
                if ((unsigned int)besedilo[interval.konec] < 0x20 ||
                    (unsigned int)besedilo[interval.konec] == 0x7f ||
                    aisspace(besedilo[interval.konec]))
                {
                    // Nali smo kontrolni znak ali loilo, ki prekinja seznam parametrov.
                    break;
                } else if (besedilo[interval.konec] == '&') {
                    // Preskoimo loila parametrov.
                    interval.konec++;
                } else {
                    T_ROCICA rocica = parametri.UstvariVozel();
                    RRegElHTTPParameterURL &parameter = parametri.VrniElement(rocica);

                    if (parameter.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
                        // Nali smo parameter.
                        interval.konec = parameter.interval.konec;
                        parametri.DodajVozel(rocica);
                    } else {
                        // Nismo nali parametra.
                        parametri.UniciVozel(rocica);
                        break;
                    }
                }
            } else {
                // Zmanjkalo nam je znakov.
                break;
            }
        }
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    streznik.Razveljavi();
    vrata.Razveljavi();
    pot.Razveljavi();
    parametri.Unici();

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPURL::Razveljavi()
{
    // Razveljavimo podelemente.
    streznik.Razveljavi();
    vrata.Razveljavi();
    pot.Razveljavi();
    parametri.Unici();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPJezikRFC1766
//////////////////////////////////////////////////////////////////////

RRegElHTTPJezikRFC1766::RRegElHTTPJezikRFC1766() : RRegElementA()
{
}


T_BOOL RRegElHTTPJezikRFC1766::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;
    komponente.Sprazni();

    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            T_INTERVAL<T_UINT> k;

            // Doloimo komponento oznake jezika.
            k.konec = interval.konec;
            for (;;) {
                if (k.konec < konec && besedilo[k.konec]) {
                    if (aisalpha(besedilo[k.konec])) {
                        // Nali smo rko.
                        k.konec++;
                    } else {
                        // Nali smo drug znak, zato prekinemo.
                        break;
                    }
                } else {
                    // Zmanjkalo nam je znakov, zato prekinemo.
                    break;
                }
            }

            if (interval.konec < k.konec) {
                // Uspeno smo nali komponento, dodamo jo v tabelo.
                k.zacetek       = interval.konec;
                interval.konec  = k.konec;
                komponente     += k;
            } else {
                // Komponente nismo nali. Ne nadaljujemo iskanja.
                break;
            }

            if (interval.konec < konec && besedilo[interval.konec] == '-') {
                // Nali smo minus, ki povezuje komponente. Preskoimo ga in nadaljujemo.
                interval.konec++;
            } else {
                // Ne sledi minus, zato konamo.
                break;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (komponente.VrniStevilo()) {
        // Nali smo vsaj eno komponento.
        interval.zacetek = zacetek;
        interval.konec   = komponente[komponente.VrniStevilo() - 1].konec;
        komponente.SprostiOdvecne();
        return TRUE;
    } else {
        // Nismo prepoznali nobene komponente.
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


T_VOID RRegElHTTPJezikRFC1766::Razveljavi()
{
    // Razveljavimo podelemente.
    komponente.Sprazni();
    komponente.SprostiOdvecne();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPFaktorUtezi
//////////////////////////////////////////////////////////////////////

RRegElHTTPFaktorUtezi::RRegElHTTPFaktorUtezi() :
    RRegElementA(),
    vrednost(1.0f)
{
}


T_BOOL RRegElHTTPFaktorUtezi::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    T_UINT celi_del = 0, decimalni_del = 0, decimalni_del_n = 1;

    interval.konec = zacetek;

    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if ('0' <= besedilo[interval.konec] && besedilo[interval.konec] <= '9') {
                // Nali smo tevko.
                celi_del = celi_del * 10 + (besedilo[interval.konec] - '0');
                interval.konec++;
            } else if (besedilo[interval.konec] == '.') {
                // Nali smo znak, ki prekinja celi del. Doloimo e vrednost decimalnega dela.
                interval.konec++;
                for (;;) {
                    if (interval.konec < konec && besedilo[interval.konec]) {
                        if ('0' <= besedilo[interval.konec] && besedilo[interval.konec] <= '9') {
                            // Nali smo tevko.
                            decimalni_del    = decimalni_del * 10 + (besedilo[interval.konec] - '0');
                            decimalni_del_n *= 10;
                            interval.konec++;
                        } else if (besedilo[interval.konec] == '.') {
                            // Nali smo znak, ki prekinja decimalni del.
                            break;
                        }
                    } else {
                        // Zmanjkalo nam je znakov.
                        break;
                    }
                }
                break;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    if (zacetek < interval.konec) {
        // Nali smo faktor utei.
        vrednost = (T_FLOAT)((T_DOUBLE)celi_del + (T_DOUBLE)decimalni_del/decimalni_del_n);
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        // Nismo prepoznali nobene komponente.
        vrednost = 1.0f;
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}


T_VOID RRegElHTTPFaktorUtezi::Razveljavi()
{
    // Razveljavimo podelemente.
    vrednost = 1.0f;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPPiskotkiParameterRFC2109
//////////////////////////////////////////////////////////////////////

RRegElHTTPPiskotkiParameterRFC2109::RRegElHTTPPiskotkiParameterRFC2109() : RRegElementA()
{
}


T_BOOL RRegElHTTPPiskotkiParameterRFC2109::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;

    if (interval.konec < konec && besedilo[interval.konec] == '$') {
        // Nali smo znak za parameter.
        interval.konec++;
    } else {
        // Manjka znak za parameter.
        goto napaka;
    }

    if (ime.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Nali smo ime parametra.
        interval.konec = ime.interval.konec;
    } else {
        // Manjka ime parametra.
        goto napaka;
    }

    // Preskoimo presledke.
    while (interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo[interval.konec])) interval.konec++;

    if (interval.konec < konec && besedilo[interval.konec] == '=') {
        // Sledi enaaj.
        interval.konec++;
    } else {
        // Manjka enaaj.
        goto napaka;
    }

    // Preskoimo presledke.
    while (interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo[interval.konec])) interval.konec++;

    if (vrednost.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Nali smo vrednost parametra.
        interval.konec = vrednost.interval.konec;
    } else {
        // Manjka vrednost parametra.
        goto napaka;
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    ime.Razveljavi();
    vrednost.Razveljavi();

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPPiskotkiParameterRFC2109::Razveljavi()
{
    // Razveljavimo podelemente.
    ime.Razveljavi();
    vrednost.Razveljavi();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPPiskotekRFC2109
//////////////////////////////////////////////////////////////////////

RRegElHTTPPiskotekRFC2109::RRegElHTTPPiskotekRFC2109() : RRegElementA()
{
}


RRegElHTTPPiskotekRFC2109::~RRegElHTTPPiskotekRFC2109()
{
    parametri.Unici();
}


T_BOOL RRegElHTTPPiskotekRFC2109::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;

    if (ime.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Nali smo ime pikotka.
        interval.konec = ime.interval.konec;
    } else {
        // Manjka ime pikotka.
        goto napaka;
    }

    // Preskoimo presledke.
    while (interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo[interval.konec])) interval.konec++;

    if (interval.konec < konec && besedilo[interval.konec] == '=') {
        // Sledi enaaj.
        interval.konec++;
    } else {
        // Manjka enaaj.
        goto napaka;
    }

    // Preskoimo presledke.
    while (interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo[interval.konec])) interval.konec++;

    if (vrednost.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        // Nali smo vrednost pikotka.
        interval.konec = vrednost.interval.konec;
    } else {
        // Manjka vrednost pikotka.
        goto napaka;
    }

    // Poiemo e parametre pikotka.
    parametri.Unici();
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Nali smo presledek, ki ga preskoimo.
                interval.konec++;
            } else if (besedilo[interval.konec] == ';') {
                T_ROCICA rocica = parametri.UstvariVozel();
                assert(rocica);
                RRegElHTTPPiskotkiParameterRFC2109 &parameter = parametri.VrniElement(rocica);
                if (parameter.PreveriUjemanje(besedilo, interval.konec + 1, konec, zastavice)) {
                    // Nali smo parameter, ki ga dodamo v seznam.
                    parametri.DodajVozel(rocica);
                    interval.konec = parameter.interval.konec;
                } else {
                    // Parametra nismo nali, zato konamo.
                    parametri.UniciVozel(rocica);
                    break;
                }
            } else {
                // Nali smo neznan znak, zato konamo.
                break;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    interval.zacetek = zacetek;
    interval.konec   = parametri.JePrazen() ? vrednost.interval.konec : parametri.VrniElement(parametri.VrniZadnjega()).interval.konec;
    return TRUE;

napaka:
    ime.Razveljavi();
    vrednost.Razveljavi();
    parametri.Unici();

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}


T_VOID RRegElHTTPPiskotekRFC2109::Razveljavi()
{
    // Razveljavimo podelemente.
    ime.Razveljavi();
    vrednost.Razveljavi();
    parametri.Unici();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPAgent
//////////////////////////////////////////////////////////////////////

RRegElHTTPAgent::RRegElHTTPAgent() :
    RRegElementA(),
    vrsta(1, 0),
    verzija(1, 0)
{
}


T_BOOL RRegElHTTPAgent::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    // Doloimo vrsto protokola.
    vrsta.zacetek = interval.konec;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (besedilo[interval.konec] == '/') {
                // Nali smo poevnico, ki doloa konec vrste protokola in zaetek verzije.
                vrsta.konec = interval.konec;
                interval.konec++;

                // Doloimo verzijo.
                verzija.zacetek = interval.konec;
                for (;;) {
                    if (interval.konec < konec && besedilo[interval.konec]) {
                        if (aisspace(besedilo[interval.konec])) {
                            // Nali smo konec verzije (presledek ali konec vrstice).
                            verzija.konec = interval.konec;
                            break;
                        } else {
                            // Preskoimo vsak drug znak.
                            interval.konec++;
                        }
                    } else {
                        // Zmanjkalo nam je vsebine.
                        verzija.konec = interval.konec;
                        break;
                    }
                }

                break;
            } else if (aisspace(besedilo[interval.konec])) {
                // Zmanjkalo nam je vsebine (presledek ali konec vrstice).
                vrsta.konec = interval.konec;
                break;
            } else {
                // Preskoimo vsak znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je vsebine.
            vrsta.konec = interval.konec;
            break;
        }
    }

    if (zacetek < interval.konec) {
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        vrsta.zacetek   = 1;
        vrsta.konec     = 0;
        verzija.zacetek = 1;
        verzija.konec   = 0;

        interval.zacetek = 1;
        interval.konec   = 0;

        return FALSE;
    }
}


T_VOID RRegElHTTPAgent::Razveljavi()
{
    // Razveljavimo podelemente.
    vrsta.zacetek   = 1;
    vrsta.konec     = 0;
    verzija.zacetek = 1;
    verzija.konec   = 0;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPProtokol
//////////////////////////////////////////////////////////////////////

RRegElHTTPProtokol::RRegElHTTPProtokol() :
    RRegElementA(),
    vrsta(1, 0),
    verzija_maj(1, 0),
    verzija_min(1, 0),
    verzija(0x009)
{
}


T_BOOL RRegElHTTPProtokol::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);
    UNREFERENCED_PARAMETER(zastavice);

    interval.konec = zacetek;

    // Doloimo vrsto protokola.
    vrsta.zacetek = interval.konec;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (besedilo[interval.konec] == '/') {
                // Nali smo poevnico, ki doloa konec vrste protokola.
                vrsta.konec = interval.konec;
                interval.konec++;
                break;
            } else if (aisspace(besedilo[interval.konec])) {
                // Zmanjkalo nam je vsebine (presledek ali konec vrstice).
                goto napaka;
            } else {
                // Preskoimo vsak znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je vsebine.
            vrsta.konec = interval.konec;
            goto napaka;
        }
    }

    // Doloimo verzijo.
    verzija_maj.zacetek = interval.konec;
    for (;;) {
        if (interval.konec < konec && besedilo[interval.konec]) {
            if (besedilo[interval.konec] == '.') {
                // Nali smo piko.
                verzija_maj.konec = interval.konec;
                interval.konec++;

                // Doloimo e podverzijo.
                verzija_min.zacetek = interval.konec;
                for (;;) {
                    if (interval.konec < konec && besedilo[interval.konec]) {
                        if (aisspace(besedilo[interval.konec])) {
                            // Nali smo konec verzije (presledek ali konec vrstice).
                            verzija_min.konec = interval.konec;
                            verzija =
                                (T_U2)astrtoui(besedilo + verzija_maj.zacetek, verzija_maj.VrniDolzino(), NULL, 10) * 0x100 +
                                (T_U2)astrtoui(besedilo + verzija_min.zacetek, verzija_min.VrniDolzino(), NULL, 10);
                            break;
                        } else {
                            // Preskoimo vsak znak.
                            interval.konec++;
                        }
                    } else {
                        // Zmanjkalo nam je vsebine.
                        goto napaka;
                    }
                }

                break;
            } else if (aisspace(besedilo[interval.konec])) {
                // Nali smo konec verzije (presledek ali konec vrstice).
                verzija_maj.konec   = interval.konec;
                verzija_min.zacetek = 1;
                verzija_min.konec   = 0;
                verzija = (T_U2)astrtoui(besedilo + verzija_maj.zacetek, verzija_maj.VrniDolzino(), NULL, 10) * 0x100;
                break;
            } else {
                // Preskoimo vsak znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je vsebine.
            goto napaka;
        }
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    vrsta.zacetek       = 1;
    vrsta.konec         = 0;
    verzija_maj.zacetek = 1;
    verzija_maj.konec   = 0;
    verzija_min.zacetek = 1;
    verzija_min.konec   = 0;
    verzija             = 0x009;

    interval.zacetek    = 1;
    interval.konec      = 0;

    return FALSE;
}


T_VOID RRegElHTTPProtokol::Razveljavi()
{
    // Razveljavimo podelemente.
    vrsta.zacetek       = 1;
    vrsta.konec         = 0;
    verzija_maj.zacetek = 1;
    verzija_maj.konec   = 0;
    verzija_min.zacetek = 1;
    verzija_min.konec   = 0;
    verzija             = 0x009;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPZahteva
//////////////////////////////////////////////////////////////////////

RRegElHTTPZahteva::RRegElHTTPZahteva() :
    RRegElementA(),
    vrsta(1, 0)
    //pot_url(
    //    "http\0https\0",                                                                        // nizi, ki predstavljajo ime sheme "http:" (ASCIIZZ)
    //    "ftp\0",                                                                                // nizi, ki predstavljajo ime sheme "ftp:" (ASCIIZZ)
    //    "mailto\0",                                                                             // nizi, ki predstavljajo ime sheme "mailto:" (ASCIIZZ)
    //    "file\0",                                                                               // nizi, ki predstavljajo ime sheme "file:" (ASCIIZZ)
    //    ":",                                                                                    // niz z znaki, ki loujejo protokol (dvopije)
    //    "/",                                                                                    // niz z znaki, ki dodatno loujejo protokol od nadaljevanja naslova URL (poevnica)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=",       // niz z znaki, ki lahko nastopajo v uporabnikem imenu
    //    ":",                                                                                    // niz z znaki, ki loujejo uporabniko ime od gesla (dvopije)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:",      // niz z znaki, ki lahko nastopajo v geslu
    //    "@",                                                                                    // niz z znaki, ki loujejo uporabniko ime in geslo od nadaljevanja (ajka)
    //    "0",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko ni
    //    "1",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko ena
    //    "2",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko dve
    //    "3",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko tri
    //    "4",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko tiri
    //    "5",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko pet
    //    "6",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko est
    //    "7",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko sedem
    //    "8",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko osem
    //    "9",                                                                                    // niz z znaki, ki predstavljajo desetiko tevko devet
    //    ".",                                                                                    // niz z znaki, ki predstavljajo loila komponent (pika)
    //    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-",                      // niz z znaki, ki lahko nastopajo v komponentah
    //    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789",                       // niz z znaki, ki lahko nastopajo v zadnji komponenti (Top Level Domain)
    //    ":",                                                                                    // niz z znaki, ki loujejo vrata od imena strenika (dvopije)
    //    "/",                                                                                    // niz z znaki, ki loujejo pot od imena in vrat strenika (poevnica)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:@/",    // niz z znaki, ki lahko nastopajo v poti
    //    "?",                                                                                    // niz z znaki, ki loujejo poizvedbo od poti (vpraaj)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:@/?",   // niz z znaki, ki lahko nastopajo v poizvedbi
    //    "#",                                                                                    // niz z znaki, ki loujejo zaznamek od poti (viaj)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:@/?#"   // niz z znaki, ki lahko nastopajo v zaznamku
    //),
    //pot_uri(
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:@/",    // niz z znaki, ki lahko nastopajo v poti
    //    "?",                                                                                    // niz z znaki, ki loujejo poizvedbo od poti (vpraaj)
    //    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~%!$&'()*+,;=:@/?"    // niz z znaki, ki lahko nastopajo v poizvedbi
    //)
{
}


T_BOOL RRegElHTTPZahteva::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;

    // Preskoimo uvodne presledke.
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na prezgodnji konec vrstice.
            goto napaka;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Preskoimo presledek.
                interval.konec++;
            } else
                break;
        } else {
            // Zmanjkalo nam je vsebine.
            goto napaka;
        }
    }

    // Doloimo vrsto zahteve.
    vrsta.zacetek = interval.konec;
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na prezgodnji konec vrstice.
            goto napaka;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Nali smo konec vrste zahteve.
                vrsta.konec = interval.konec;
                interval.konec++;
                break;
            } else {
                // Preskoimo vsak znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je vsebine.
            goto napaka;
        }
    }

    // Preskoimo presledke.
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na prezgodnji konec vrstice.
            goto napaka;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Preskoimo presledek.
                interval.konec++;
            } else
                break;
        } else {
            // Zmanjkalo nam je vsebine.
            goto napaka;
        }
    }

    // Preberemo pot.
    if (url.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
        interval.konec = url.interval.konec;
    } else
        goto napaka;

    protokol.Razveljavi();

    // Preskoimo presledke.
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na konec vrstice. Protokola ne bo.
            interval.konec = prelom_vrstice.interval.konec;
            goto konec;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Preskoimo presledek.
                interval.konec++;
            } else
                break;
        } else {
            // Zmanjkalo nam je vsebine.
            goto konec;
        }
    }

    // Preberemo protokol.
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na konec vrstice. Protokola ne bo.
            interval.konec = prelom_vrstice.interval.konec;
            goto konec;
        } else if (protokol.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Nali smo protokol.
            interval.konec = protokol.interval.konec;
            break;
        } else {
            // Zmanjkalo nam je vsebine.
            goto konec;
        }
    }

    // Poiemo e konec vrstice.
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na prelom vrstice.
            interval.konec = prelom_vrstice.interval.konec;
            break;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            // Preskoimo vsak znak.
            interval.konec++;
        } else {
            // Zmanjkalo nam je vsebine.
            goto konec;
        }
    }

konec:
    interval.zacetek = zacetek;
    return TRUE;

napaka:
    vrsta.zacetek = 1;
    vrsta.konec   = 0;
    url.Razveljavi();
    protokol.Razveljavi();

    interval.zacetek = 1;
    interval.konec   = 0;

    return FALSE;
}


T_VOID RRegElHTTPZahteva::Razveljavi()
{
    // Razveljavimo podelemente.
    vrsta.zacetek = 1;
    vrsta.konec   = 0;
    url.Razveljavi();
    protokol.Razveljavi();

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// RRegElHTTPGlava
//////////////////////////////////////////////////////////////////////

RRegElHTTPGlava::RRegElHTTPGlava() :
    RRegElementA(),
    ime(1, 0),
    vrednost(1, 0)
{
}


T_BOOL RRegElHTTPGlava::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || zacetek < konec);

    interval.konec = zacetek;

    if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice) ||
        interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo, interval.konec))
    {
        // Prazna vrstica ali vrstica, ki se zaenja s presledki, ne more biti glava HTTP.
        goto napaka;
    }

    // Doloimo ime glave HTTP.
    ime.zacetek = interval.konec;
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Naleteli smo na prezgodnji konec vrstice.
            goto napaka;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo, interval.konec)) {
                // Nali smo presledek, ki naznanja konec imena glave.
                ime.konec = interval.konec;
                interval.konec++;

                // Preskoimo presledke na koncu imena glave, e jih je e kaj.
                for (;;) {
                    if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
                        // Naleteli smo na prezgodnji konec vrstice.
                        goto napaka;
                    } else if (interval.konec < konec && besedilo[interval.konec]) {
                        if (aisspace(besedilo[interval.konec])) {
                            // Preskoimo presledek.
                            interval.konec++;
                        } else
                            break;
                    } else {
                        // Zmanjkalo nam je vsebine.
                        goto napaka;
                    }
                }

                // Zdaj mora slediti dvopije.
                if (interval.konec < konec && besedilo[interval.konec] == ':') {
                    interval.konec++;
                    break;
                } else {
                    // Manjka dvopije.
                    goto napaka;
                }

                break;
            } else if (besedilo[interval.konec] == ':') {
                // Nali smo dvopije, ki naznanja konec imena glave.
                ime.konec = interval.konec;
                interval.konec++;
                break;
            } else {
                // Preskoimo vsak znak.
                interval.konec++;
            }
        } else {
            // Zmanjkalo nam je vsebine.
            goto napaka;
        }
    }

    // Doloimo vrednost glave HTTP.
    vrednost.zacetek = T_UINT_MAX;
    vrednost.konec   = 0;
    for (;;) {
        if (prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice)) {
            // Nali smo prelom vrstice.
            interval.konec = prelom_vrstice.interval.konec;

            // Pogledamo, ali naslednja vrstica ni prazna in se zaenja s presledkom.
            if (!prelom_vrstice.PreveriUjemanje(besedilo, interval.konec, konec, zastavice) &&
                interval.konec < konec && besedilo[interval.konec] && aisspace(besedilo[interval.konec]))
            {
                // Nali smo prelom vrednosti glave HTTP, zato nadaljujemo.
                interval.konec++;
            } else
                break;
        } else if (interval.konec < konec && besedilo[interval.konec]) {
            if (aisspace(besedilo[interval.konec])) {
                // Nali smo presledek, ki ga preskoimo.
                interval.konec++;
            } else {
                // Preskoimo tudi druge znake, a jih dodamo v vrednost.
                if (vrednost.zacetek == T_UINT_MAX) vrednost.zacetek = interval.konec;
                vrednost.konec = ++interval.konec;
            }
        } else {
            // Zmanjkalo nam je znakov.
            break;
        }
    }

    interval.zacetek = zacetek;
    return TRUE;

napaka:
    ime.zacetek      = 1;
    ime.konec        = 0;
    vrednost.zacetek = 1;
    vrednost.konec   = 0;

    interval.zacetek = 1;
    interval.konec   = 0;

    return FALSE;
}


T_VOID RRegElHTTPGlava::Razveljavi()
{
    // Razveljavimo podelemente.
    ime.zacetek      = 1;
    ime.konec        = 0;
    vrednost.zacetek = 1;
    vrednost.konec   = 0;

    RRegElementA::Razveljavi();
}


//////////////////////////////////////////////////////////////////////
// ROdjemalecHTTP
//////////////////////////////////////////////////////////////////////

#pragma warning(push)
#pragma warning(disable: 4355)

ROdjemalecHTTP::ROdjemalecHTTP() :
    RElement(),
    casomer(this),
    tok(NULL)
{
}

#pragma warning(pop)


ROdjemalecHTTP::~ROdjemalecHTTP()
{
    glave_zahteve.Unici();
    parametri_piskotkov.Unici();
    piskotki.Unici();
}


T_STANJE ROdjemalecHTTP::ObdelajZahteve(RTokPodatkov *_tok, T_TRENUTEK casovna_kontrola_sprejema_zahteve, T_TRENUTEK casovna_kontrola_obdelave_zahteve)
{
    assert(_tok);

    TK_ASCII podatki;
    T_UINT odmik, dolzina;
    T_BOOL prva_vrstica;
    T_STANJE stanje;
    RRegElHTTPPrelomVrstice prelom_vrstice;

    // Nastavimo tok, kjer obdelujemo zahteve.
    tok = _tok;

    for (;;) {
        // Privzamemo verzijo protokola HTTP 0.9
        verzija = 0x009;

        // Pred zaetkom izpraznimo glavo. ;)
        verify(podatki_glave.PrestaviZac(0) == 0);
        verify(USPESNO(podatki_glave.PostaviKonec()));
        verify(USPESNO(podatki_glave.RezervirajDolzino(RHTTP_MAX_DOLZINA_GLAVE_ZAHTEVE)));

        // Izpraznimo seznam pikotkov.
        parametri_piskotkov.Unici();
        piskotki.Unici();

        // Nastavimo asovno kontrolo, do kdaj moramo sprejeti zahtevo.
        verify(USPESNO(casomer.Nastavi(casovna_kontrola_sprejema_zahteve, 0)));

        // Naprej preberemo glavo v celoti, nato pa jo analiziramo.
        for (prva_vrstica = TRUE;;) {
            T_AZNAK znak;

            // Preberemo en znak iz toka podatkov in ga dodamo v glavo.
            (*tok) >> znak;
            if (NEUSPESNO(tok->VrniStanje())) {
                stanje = AOSN_NAPAKA_HTTP_GLAVA_ZAHTEVE_PREKRATKA;
                goto napaka;
            }
            assert(USPESNO(podatki_glave.VrniStanje()));
            podatki_glave << znak;

            podatki = (TK_ASCII)podatki_glave.VrniPodatke();
            dolzina = (T_UINT)podatki_glave.VrniDolzino();

            // Preverimo, ali smo presegli dovoljeno dolino glave.
            if (dolzina >= RHTTP_MAX_DOLZINA_GLAVE_ZAHTEVE) {
                stanje = AOSN_NAPAKA_HTTP_GLAVA_ZAHTEVE_PREDOLGA;
                goto napaka;
            }

            if (prva_vrstica &&
                (dolzina >= 2 &&
                podatki[dolzina - 2] == '\r' &&
                podatki[dolzina - 1] == '\n' ||
                dolzina >= 1 &&
                podatki[dolzina - 1] == '\n'))
            {
                prva_vrstica = FALSE;

                // Prebrali smo prvo vrstico, ki jo preverimo e med branjem, da prekinemo odjemalce, ki nas zafrkavajo.
                if (!zahteva.PreveriUjemanje(podatki, 0, dolzina) ||
                    astrncmp(podatki + zahteva.protokol.vrsta.zacetek, zahteva.protokol.vrsta.VrniDolzino(), "HTTP", 4) != 0)
                {
                    // Prva vrstica je napana. Prekinemo analizo in se sploh ne pogovarjamo ve z odjemalcem.
                    stanje = AOSN_NAPAKA_HTTP_ZAHTEVA_NAPACNA;
                    goto napaka;
                }

                // Odjemalec je verzije HTTP 1.1 ali noveji. Na strenik pa zna najve HTTP 1.1.
                verzija = amin<T_U2>(zahteva.protokol.verzija, 0x101);

                // Preverimo, kaj na analizator meni o tej zahtevi.
                stanje = SprejmiZahtevo();
                if (NEUSPESNO(stanje)) goto napaka;

                if (zahteva.protokol.verzija >= 0x101) {
                    // Odjemalcem HTTP 1.1 in novejim moramo im prej sporoiti, da jih posluamo in naj nadaljujejo.
                    stanje = PosljiOdgovor(100);
                    if (NEUSPESNO(stanje)) goto napaka;

                    // Zakljuimo glavo.
                    stanje = tok->VrniStanje();
                    if (NEUSPESNO(stanje)) goto napaka;
                    tok->ZapisiObl("\r\n", T_LOCALE_THREAD_DEFAULT);
                } else if (zahteva.protokol.verzija < 0x100) {
                    // Odjemalec je poslal zahtevo HTTP stareje verzije (pred 1.0), zato ne bo vsebovala glav.
                    break;
                }
            } else if (dolzina >= 4 &&
                podatki[dolzina - 4] == '\r' &&
                podatki[dolzina - 3] == '\n' &&
                podatki[dolzina - 2] == '\r' &&
                podatki[dolzina - 1] == '\n' ||
                dolzina >= 2 &&
                podatki[dolzina - 2] == '\n' &&
                podatki[dolzina - 1] == '\n')
            {
                // Prebrali smo zaporedje CRLFCRLF oziroma LFLF, ki naznanja konec glave.
                break;
            }
        }

        // Preden nadaljujemo, sprostimo odveen pomnilnik.
        dolzina = (T_UINT)podatki_glave.VrniDolzino();
        verify(USPESNO(podatki_glave.RezervirajDolzino(dolzina, TRUE)));
        assert((T_UINT)podatki_glave.VrniDolzino() == dolzina);
        podatki = (TK_ASCII)podatki_glave.VrniPodatke();

        // Pripravimo seznam glav.
        glave_zahteve.Unici();
        if (zahteva.protokol.verzija >= 0x100) {
            for (odmik = zahteva.interval.konec; odmik < dolzina;) {
                if (prelom_vrstice.PreveriUjemanje(podatki, odmik, dolzina)) {
                    // Nali smo prazno vrstico, ki pomeni konec glav.
                    odmik = prelom_vrstice.interval.konec;
                    break;
                } else {
                    // Poskusimo, ali smo nali glavo.
                    T_ROCICA rocica = glave_zahteve.UstvariVozel();
                    RRegElHTTPGlava &glava = glave_zahteve.VrniElement(rocica);

                    if (glava.PreveriUjemanje(podatki, odmik, dolzina)) {
                        // Nali smo glavo, ki jo dodamo v seznam.
                        glave_zahteve.DodajVozel(rocica);
                        odmik = glava.interval.konec;
                    } else {
                        // Nismo nali veljavne glave.
                        glave_zahteve.UniciVozel(rocica);

                        // Nali smo vrstico, ki je ne znamo analizirati. Preskoimo jo.
                        for (;;) {
                            if (prelom_vrstice.PreveriUjemanje(podatki, odmik, dolzina)) {
                                // Nali smo preloma vrstice.
                                odmik = prelom_vrstice.interval.konec;
                                break;
                            } else if (odmik < dolzina && podatki[odmik]) {
                                // Nali smo znak, ki ga preskoimo.
                                odmik++;
                            } else {
                                // Zmanjkalo nam je podatkov.
                                break;
                            }
                        }
                    }
                }
            }
        }

        if (zahteva.protokol.verzija == 0x101) {
            // Po definiciji, so odjemalci HTTP 1.1 dolni poslati glavo Host.
            T_ROCICA rocica = NajdiGlavo("Host", _countof("Host"));
            if (!rocica) {
                verify(USPESNO(PosljiStatus(400, NULL, 0, NULL, 0)));
                stanje = AOSN_NAPAKA_HTTP_MANJKA_GLAVA_HOST;
                goto napaka;
            }
        }

        // Nastavimo asovno kontrolo, do kdaj moramo obdelati zahtevo.
        verify(USPESNO(casomer.Nastavi(casovna_kontrola_obdelave_zahteve, 0)));

        // Zdaj obdelamo e podatke zahteve.
        stanje = PreberiZahtevo();
        if (NEUSPESNO(stanje)) goto napaka;

        // Izklopimo asomer, saj sledi le e odloanje, ali povezavo prekinemo, ali sprejmemo e kako zahtevo.
        verify(USPESNO(casomer.Prekini()));

        // Preverimo, ali moramo konati.
        if (zahteva.protokol.verzija < 0x101) {
            // Odjemalec je stareji od HTTP 1.1, zato po obdelavi prve zahteve konamo.
            break;
        } else {
            // Preverimo, ali je odjemalec elel prekiniti povezavo po obdelavi zahteve.
            T_ROCICA rocica = NajdiGlavo("Connection", _countof("Connection"));
            if (rocica) {
                RRegElHTTPGlava &glava = glave_zahteve.VrniElement(rocica);

                if (astrnicmp(podatki + glava.vrednost.zacetek, glava.vrednost.VrniDolzino(), "close", _countof("close")) == 0) {
                    // Odjemalec je elel, da strenik po obdelavi zahteve prekine povezavo.
                    break;
                }
            }
        }
    }

    stanje = A_OK;

napaka:
    verify(USPESNO(casomer.Prekini()));
    tok = NULL;
    return stanje;
}


T_STANJE ROdjemalecHTTP::AnalizirajPiskotke()
{
    T_ROCICA rocica;
    T_UINT odmik;
    TK_ASCII podatki = (TK_ASCII)podatki_glave.VrniPodatke();

    // Preberemo in analiziramo pikotke, ki jih je poslal odjemalec.
    for (rocica = glave_zahteve.VrniPrvega(); rocica; rocica = glave_zahteve.VrniNaslednjega(rocica)) {
        rocica = NajdiGlavo(rocica, "Cookie", _countof("Cookie"));
        if (!rocica) break;
        RRegElHTTPGlava &glava = glave_zahteve.VrniElement(rocica);

        for (odmik = glava.vrednost.zacetek; odmik < glava.vrednost.konec; ) {
            T_ROCICA rocica;

            // Preskoimo vejico/podpije in presledke med zapisi.
            while (odmik < glava.vrednost.konec && podatki[odmik] && aisspace(podatki[odmik])) odmik++;
            if (odmik < glava.vrednost.konec && (podatki[odmik] == ',' || podatki[odmik] == ';')) {
                odmik++;
                while (odmik < glava.vrednost.konec && podatki[odmik] && aisspace(podatki[odmik])) odmik++;
            }

            // Identificiramo parameter.
            rocica = parametri_piskotkov.UstvariVozel();
            assert(rocica);
            RRegElHTTPPiskotkiParameterRFC2109 &parameter = parametri_piskotkov.VrniElement(rocica);

            if (parameter.PreveriUjemanje(podatki, odmik, glava.vrednost.konec)) {
                // Nali smo parameter. Dodamo ga v seznam.
                odmik = parameter.interval.konec;
                parametri_piskotkov.DodajVozel(rocica);
                continue;
            } else {
                // Nismo nali parametra.
                parametri_piskotkov.UniciVozel(rocica);
            }

            // Identificiramo pikotek.
            rocica = piskotki.UstvariVozel();
            assert(rocica);
            RRegElHTTPPiskotekRFC2109 &piskotek = piskotki.VrniElement(rocica);

            if (piskotek.PreveriUjemanje(podatki, odmik, glava.vrednost.konec)) {
                // Nali smo pikotek. Dodamo ga v seznam.
                odmik = piskotek.interval.konec;
                piskotki.DodajVozel(rocica);
                continue;
            } else {
                // Nismo nali pikotka.
                piskotki.UniciVozel(rocica);
            }

            // Nismo nali ne parametra, ne pikotka.
            break;
        }
    }

    return A_OK;
}


T_STANJE ROdjemalecHTTP::SprejmiZahtevo()
{
    assert(tok);

    // Privzeta metoda ROdjemalecHTTP::SprejmiZahtevo() zavrne vse metode, ki jih nismo prestregli v izpeljanem razredu. Sporoimo pa status 501 namesto 405, da se nam ni treba ubadati z glavo "Allow".
    verify(USPESNO(PosljiStatus(501, NULL, 0, NULL, 0)));
    return AOSN_NAPAKA_HTTP_METODA_NI_PODPRTA;
}


T_STANJE ROdjemalecHTTP::PreberiZahtevo()
{
    assert(tok);

    T_ROCICA rocica;

    if (verzija < 0x100) {
        // Pri protokolu HTTP stareje verzije (pred 1.0), zahteva nikoli ne vsebuje podatkov.
        return A_OK;
    }

    // Preberemo vsebino zahteve le zato, da preskoimo na naslednjo zahtevo. Pravo obdelavo vsebine moramo pripraviti v izpeljanem razredu.
    rocica = NajdiGlavo("Content-Type", _countof("Content-Type"));
    if (rocica) {
        // Nali smo glavo "Content-Type", torej bomo imeli nekaj vsebine.
        assert(0); // TODO: Preizkusi ta del kode.

        if (verzija >= 0x101) {
            // Protokol je verzije HTTP 1.1.
            rocica = NajdiGlavo("Transfer-Encoding", _countof("Transfer-Encoding"));
            if (rocica) {
                RRegElHTTPGlava &glava = glave_zahteve.VrniElement(rocica);
                if (astrnicmp((TK_ASCII)podatki_glave.VrniPodatke() + glava.vrednost.zacetek, glava.vrednost.VrniDolzino(), "chunked", _countof("chunked")) == 0) {
                    // Odjemalec je poslal podatke zahteve po kosih.
                    assert(0); // TODO: Preizkusi ta del kode.

                    RPomDatoteka vsebina_vrstice;
                    T_STANJE stanje;
                    T_U8 dolzina_dela;

                    for (;;) {
                        // Preberemo vrstico.
                        verify(vsebina_vrstice.PrestaviZac(0) == 0);
                        verify(USPESNO(vsebina_vrstice.PostaviKonec()));
                        stanje = PrilepiVrstico(*tok, vsebina_vrstice);
                        if (NEUSPESNO(stanje)) return stanje;

                        // Analiziramo vrstico.
                        dolzina_dela = astrtou8((TK_ASCII)vsebina_vrstice.VrniPodatke(), (T_UINT)vsebina_vrstice.VrniDolzino(), NULL, 16);
                        if (dolzina_dela == 0) break;

                        // Preskoimo vsebino dela.
                        verify(USPESNO(tok->Preskoci(dolzina_dela)));

                        // Vsebini dela sledi e en prelom vrstice.
                        verify(vsebina_vrstice.PrestaviZac(0) == 0);
                        verify(USPESNO(vsebina_vrstice.PostaviKonec()));
                        stanje = PrilepiVrstico(*tok, vsebina_vrstice);
                        if (NEUSPESNO(stanje)) return stanje;
                    }

                    return A_OK;
                }
            }
        }

        rocica = NajdiGlavo("Content-Length", _countof("Content-Length"));
        if (rocica) {
            // Preskoimo toliko bajtov podatkov, kot veleva glava "Content-Length".
            RRegElHTTPGlava &glava = glave_zahteve.VrniElement(rocica);
            T_DOLZINA dolzina = astrtou8((TK_ASCII)podatki_glave.VrniPodatke() + glava.vrednost.zacetek, glava.vrednost.VrniDolzino(), NULL, 10);
            return tok->Preskoci(dolzina);
        }
    }

    return A_OK;
}


T_STANJE ROdjemalecHTTP::PosljiStatus(T_UINT status, TK_ASCII vsebina, T_UINT dolzina_vsebine, TK_ASCII dodatne_glave, T_UINT dolzina_dodatnih_glav)
{
    // Parametra vsebina in dolzina_vsebine ignoriramo, ker v tem razredu e ne vemo, ali gre za zahtevo HEAD, kjer ne smemo poiljati vsebine.
    UNREFERENCED_PARAMETER(vsebina);
    UNREFERENCED_PARAMETER(dolzina_vsebine);
    assert(tok);

    T_STANJE stanje;

    if (verzija >= 0x100) {
        // Poljemo odgovor.
        stanje = PosljiOdgovor(status);
        if (NEUSPESNO(stanje)) return stanje;

        // Poljemo skupne glave.
        stanje = PosljiGlave(dodatne_glave, dolzina_dodatnih_glav);
        if (NEUSPESNO(stanje)) return stanje;

        // Zakljuimo glavo.
        stanje = tok->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
        tok->ZapisiObl("\r\n", T_LOCALE_THREAD_DEFAULT);
    } else {
        // Pri protokolu HTTP stareje verzije (pred 1.0), odgovor nikoli ne vsebuje uvodne vrstice in glav.
        RNizA opis;

        // Sestavimo in poljemo opis statusa kot odgovor.
        VrniOpisStatusa(status, opis);
        stanje = tok->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
        tok->ZapisiObl("%u %.*hs\r\n", T_LOCALE_THREAD_DEFAULT, (unsigned int)status, (int)opis.VrniDolzino(), opis.Vrni());
    }

    return A_OK;
}


T_STANJE ROdjemalecHTTP::PosljiGlave(TK_ASCII dodatne_glave, T_UINT dolzina, T_TRENUTEK zadnja_sprememba)
{
    assert(verzija >= 0x100); // Pri protokolu HTTP stareje verzije (pred 1.0), odgovor ne sme vsebovati glav.
    assert(dodatne_glave || !dolzina);
    assert(tok);

    T_STANJE stanje;
    RNizA datum;

    // Zapiemo datum, kdaj je nastal odgovor.
    stanje = tok->VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    RDatum::VrniNizRFC822(RDatum::VrniTrenutniCas(), datum);
    tok->ZapisiObl("Date: %.*hs\r\n", T_LOCALE_THREAD_DEFAULT, (int)datum.VrniDolzino(), datum.Vrni());

    // Zapiemo datum zadnje spremembe vsebine.
    if (zadnja_sprememba) RDatum::VrniNizRFC822(zadnja_sprememba, datum);
    stanje = tok->VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    tok->ZapisiObl("Last-Modified: %.*hs\r\n", T_LOCALE_THREAD_DEFAULT, (int)datum.VrniDolzino(), datum.Vrni());

    if (dodatne_glave && dolzina) {
        // Na koncu zapiemo e dodatne glave.
        stanje = tok->VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
        tok->ZapisiTab(dodatne_glave, sizeof(T_AZNAK), dolzina);
    }

    return A_OK;
}


T_VOID ROdjemalecHTTP::VrniOpisStatusa(T_UINT status, RNizP<T_AZNAK> &opis)
{
    static const TK_ASCII opisi_1x[] = {
        "Continue",                         // 100
        "Switching Protocols",              // 101
    };
    static const TK_ASCII opisi_2x[] = {
        "OK",                               // 200
        "Created",                          // 201
        "Accepted",                         // 202
        "Non-Authoritative Information",    // 203
        "No Content",                       // 204
        "Reset Content",                    // 205
        "Partial Content",                  // 206
    };
    static const TK_ASCII opisi_3x[] = {
        "Multiple Choices",                 // 300
        "Moved Permanently",                // 301
        "Found",                            // 302
        "See Other",                        // 303
        "Not Modified",                     // 304
        "Use Proxy",                        // 305
        "(Unused)",                         // 306
        "Temporary Redirect",               // 307
    };
    static const TK_ASCII opisi_4x[] = {
        "Bad Request",                      // 400
        "Unauthorized",                     // 401
        "Forbidden",                        // 402
        "Bad Request",                      // 403
        "Not Found",                        // 404
        "Method Not Allowed",               // 405
        "Not Acceptable",                   // 406
        "Proxy Authentication Required",    // 407
        "Request Timeout",                  // 408
        "Conflict",                         // 409
        "Gone",                             // 410
        "Length Required",                  // 411
        "Precondition Failed",              // 412
        "Request Entity Too Large",         // 413
        "Request-URI Too Long",             // 414
        "Unsupported Media Type",           // 415
        "Requested Range Not Satisfiable",  // 416
        "Expectation Failed",               // 417
    };
    static const TK_ASCII opisi_5x[] = {
        "Internal Server Error",            // 500
        "Not Implemented",                  // 501
        "Bad Gateway",                      // 502
        "Service Unavailable",              // 503
        "Gateway Timeout",                  // 504
        "HTTP Version Not Supported",       // 505
    };

         if (100 <= status && status < _countof(opisi_1x) + 100)
        opis.Nastavi(opisi_1x[status - 100]);
    else if (200 <= status && status < _countof(opisi_2x) + 200)
        opis.Nastavi(opisi_2x[status - 200]);
    else if (300 <= status && status < _countof(opisi_3x) + 300)
        opis.Nastavi(opisi_3x[status - 300]);
    else if (400 <= status && status < _countof(opisi_4x) + 400)
        opis.Nastavi(opisi_4x[status - 400]);
    else if (500 <= status && status < _countof(opisi_5x) + 500)
        opis.Nastavi(opisi_5x[status - 500]);
    else
        opis.Nastavi("(Unknown)");
}


ROdjemalecHTTP::RCasomerZahteve::RCasomerZahteve(ROdjemalecHTTP *_odjemalec) :
    RCasomer(),
    odjemalec(_odjemalec)
{
}


T_VOID ROdjemalecHTTP::RCasomerZahteve::ObAlarmu(T_8 cas)
{
    assert(odjemalec && odjemalec->tok);
    UNREFERENCED_PARAMETER(cas);

    // Zakljuimo povezavo.
    verify(USPESNO(odjemalec->tok->Zapri()));
}


//////////////////////////////////////////////////////////////////////
// RZapisovalnikHTTP
//////////////////////////////////////////////////////////////////////

RZapisovalnikHTTP::RZapisovalnikHTTP(T_U2 _verzija) :
    RPretvornikToka(),
    verzija(_verzija)
{
}


T_STANJE RZapisovalnikHTTP::Uporabi(RTokPodatkov *izvor)
{
    verify(USPESNO(izvor->VrniStanje()));
    if (verzija >= 0x101) {
        // HTTP/1.1 in noveji omogoa poiljanje po delih.
        // Nastavimo vrsto kodiranja in zakljuimo glave.
        izvor->ZapisiObl("Transfer-Encoding: chunked\r\n\r\n", T_LOCALE_THREAD_DEFAULT);
    } else {
        // Protokol je stareji in ne omogoa poiljanja po delih.
        // Zakljuimo glave da se pripravimo na poiljanje vsebine. Odjemalec bo vedel, kdaj je konec odgovora, ko bo povezava prekinjena.
        izvor->ZapisiObl("\r\n", T_LOCALE_THREAD_DEFAULT);
    }

    return RPretvornikToka::Uporabi(izvor);
}


RTokPodatkov* RZapisovalnikHTTP::Odpusti()
{
    assert(izvor);

    if (verzija >= 0x101) {
        // Na konec zapiemo e del doline 0 ter CRLF, ki naznanja, da ne bo nobenih dodatnih glav ve in da je odgovora konec.
        verify(USPESNO(izvor->VrniStanje()));
        izvor->ZapisiObl("0\r\n\r\n", T_LOCALE_THREAD_DEFAULT);
    }

    return RPretvornikToka::Odpusti();
}


T_SIZE RZapisovalnikHTTP::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    T_SIZE zapisano;

    if (verzija >= 0x101) {
        // Pazimo, da dolina dela ni enaka 0, ker bi tak del prezgodaj naznanil konec odgovora.
        if (dolzina) {
            // Najprej zapiemo glavo dela.
            stanje = izvor->VrniStanje();
            if (NEUSPESNO(stanje)) return 0;
            izvor->ZapisiObl("%" T_SIZE_PRINTF_SIZE_PREFIX "X\r\n", T_LOCALE_THREAD_DEFAULT, dolzina);

            // Nato sledi vsebina dela.
            stanje = izvor->VrniStanje();
            if (NEUSPESNO(stanje)) return 0;
            zapisano = izvor->Zapisi(podatki, dolzina);

            // Na koncu dela zapiemo e CRLF.
            stanje = izvor->VrniStanje();
            if (NEUSPESNO(stanje)) return zapisano;
            izvor->ZapisiObl("\r\n", T_LOCALE_THREAD_DEFAULT);
        } else
            zapisano = 0;
    } else {
        // Pri starejih protokolih zapisujemo vsebino direktno.
        stanje = izvor->VrniStanje();
        if (NEUSPESNO(stanje)) return 0;
        zapisano = izvor->Zapisi(podatki, dolzina);
    }

    return zapisano;
}

#endif // AOSN_HTTP
