/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_IZJEMA

#include "AOsnP.h"
#include <stdlib.h>

#pragma comment(lib, "dbghelp.lib")
#pragma comment(lib, "winhttp.lib")


//////////////////////////////////////////////////////////////////////
// Deklaracije internih funkcij
//////////////////////////////////////////////////////////////////////

inline T_VOID _VrniImeDatotekeModula(HMODULE modul, RNizU &ime_datoteke_modula);


//////////////////////////////////////////////////////////////////////
// ROddajnikHTTP
//////////////////////////////////////////////////////////////////////

ROddajnikHTTP::ROddajnikHTTP() :
    RTokPodatkov(),
    internet(),
    povezava(),
    zahteva(),
    niz((T_UINT)MAX_PATH)
{
}


ROddajnikHTTP::~ROddajnikHTTP()
{
    assert(!internet);
    assert(!povezava);
    assert(!zahteva);
}


T_STANJE ROddajnikHTTP::Odpri(TK_NIZ streznik, T_U2 vrata, TK_NIZ dokument, TK_NIZ up_ime, TK_NIZ geslo, TK_NIZ tip_zahteve, TK_NIZ agent, const TK_NIZ *zelena_vrsta_odgovora)
{
    assert(streznik);
    assert(dokument);
    assert(tip_zahteve);
    assert(!internet);
    assert(!povezava);
    assert(!zahteva);

    DWORD rezultat = ERROR_SUCCESS;

    internet = ::WinHttpOpen(agent ? 
#ifdef _UNICODE
            agent : 
#else
            RNizU(agent) :
#endif
            L"AOsn/" _L(AOSN_VERZIJA_STR),
        WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0);
    if (!internet) {
        stanje = astanje(rezultat = ::GetLastError(), AOSN_NAPAKA_HTTP_NI_INTERNETA);
        goto napaka1;
    }

    // Pripravimo se na povezavo.
    povezava = ::WinHttpConnect(internet,
#ifdef _UNICODE
        streznik,
#else
        RNizU(streznik),
#endif
        vrata, 0);
    if (!povezava) {
        stanje = astanje(rezultat = ::GetLastError(), AOSN_NAPAKA_HTTP_PRIPRAVA_POVEZAVE);
        goto napaka2;
    }

    // Pripravimo zahtevo.
    zahteva = ::WinHttpOpenRequest(
        povezava,
#ifdef _UNICODE
        tip_zahteve,                     // vrsta zahteve
        dokument,                        // zahtevani dokument
#else
        RNizU(tip_zahteve),              // vrsta zahteve
        RNizU(dokument),                 // zahtevani dokument
#endif
        L"HTTP/1.1",                     // verzija HTTP
        WINHTTP_NO_REFERER,              // "referer"
#ifdef _UNICODE
        (LPCTSTR*)zelena_vrsta_odgovora, // zahtevana vrsta datoteke
#else
        WINHTTP_DEFAULT_ACCEPT_TYPES,    // zahtevana vrsta datoteke
#endif
        WINHTTP_FLAG_REFRESH);
    if (!zahteva) {
        stanje = astanje(rezultat = ::GetLastError(), AOSN_NAPAKA_HTTP_POVEZAVA);
        goto napaka3;
    }

    // Nastavimo poverilnice.
    if (up_ime && geslo)
        ::WinHttpSetCredentials(zahteva, WINHTTP_AUTH_TARGET_SERVER, WINHTTP_AUTH_SCHEME_BASIC,
#ifdef _UNICODE
            up_ime,
            geslo,
#else
            RNizU(up_ime),
            RNizU(geslo),
#endif
            NULL);

    // Rono nastavimo pisanje po delih. Pisanje po delih je del standarda HTTP/1.1.
    stanje = DodajGlavo(_T("Transfer-Encoding"), T_UINT_MAX, _T("chunked"), T_UINT_MAX);
    if (NEUSPESNO(stanje)) goto napaka4;

    return stanje = A_OK;

napaka4:
    verify(::WinHttpCloseHandle(zahteva));
    zahteva = NULL;
napaka3:
    verify(::WinHttpCloseHandle(povezava));
    povezava = NULL;
napaka2:
    verify(::WinHttpCloseHandle(internet));
    internet = NULL;
napaka1:
    return stanje;
}


T_STANJE ROddajnikHTTP::DodajGlavo(TK_NIZ ime_glave, T_UINT st_znakov_imena, TK_NIZ vrednost_glave, T_UINT st_znakov_vrednosti)
{
    assert(ime_glave && st_znakov_imena);
    assert(vrednost_glave && st_znakov_vrednosti);
    assert(zahteva);

    // Dodamo parametre v zahtevo.
    niz.Oblikuj(_T("%.*s: %.*s"), T_LOCALE_THREAD_DEFAULT, st_znakov_imena, ime_glave, st_znakov_vrednosti, vrednost_glave);
    if (::WinHttpAddRequestHeaders(zahteva,
#ifdef _UNICODE
        niz.Vrni(),
#else
        RNizU(niz),
#endif
        (DWORD)niz.VrniDolzino(), WINHTTP_ADDREQ_FLAG_ADD))
        stanje = A_OK;
    else {
        DWORD rezultat = ::GetLastError();
        stanje = astanje(rezultat, AOSN_NAPAKA_HTTP_PRIPRAVA_GLAVE);
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::ZacniPosiljanje()
{
    assert(zahteva);

    if (::WinHttpSendRequest(zahteva, WINHTTP_NO_ADDITIONAL_HEADERS, 0, WINHTTP_NO_REQUEST_DATA, 0, 0, 0))
        stanje = A_OK;
    else {
        DWORD rezultat = ::GetLastError();
        stanje = astanje(rezultat, AOSN_NAPAKA_HTTP_ZAHTEVA);
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::KoncajPosiljanje()
{
    assert(zahteva);

    {
        // Poljemo prazen del, da naznanimo streniku, da bo pisanja konec.
        DWORD koliko;
        verify(::WinHttpWriteData(zahteva, "0\r\n\r\n", 5, &koliko));
    }

    if (::WinHttpReceiveResponse(zahteva, NULL))
        stanje = A_OK;
    else {
        DWORD rezultat = ::GetLastError();
        stanje = astanje(rezultat, AOSN_NAPAKA_HTTP_KONEC_POSILJANJA);
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::PreberiInformacijo(T_UINT vrsta, T_UINT &podatek)
{
    assert(zahteva);

    DWORD _podatek = 0, dolzina;

    dolzina = sizeof(DWORD);
    if (::WinHttpQueryHeaders(zahteva, (DWORD)vrsta | WINHTTP_QUERY_FLAG_NUMBER, WINHTTP_HEADER_NAME_BY_INDEX, &_podatek, &dolzina, NULL)) {
        assert(dolzina == sizeof(DWORD));
        podatek = _podatek;
        stanje = A_OK;
    } else {
        podatek = T_UINT_MAX;
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_HTTP_STATUS_ZAHTEVE);
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::PreberiInformacijo(T_UINT vrsta, RNiz &podatek)
{
    assert(zahteva);

    DWORD dolzina = RNIZ_KORAK_POVECEVANJA;
    T_NIZ _podatek;

    for (;;) {
        _podatek = podatek.Odpusti(dolzina, FALSE);
        assert(_podatek);
        dolzina *= sizeof(T_ZNAK);
        if (::WinHttpQueryHeaders(zahteva, (DWORD)vrsta, WINHTTP_HEADER_NAME_BY_INDEX, _podatek, &dolzina, NULL)) {
            dolzina /= sizeof(T_ZNAK);
            podatek.Uporabi(_podatek, dolzina, FALSE);
            podatek.SprostiOdvecne();
            stanje = A_OK;
            break;
        } else {
            DWORD rezultat = ::GetLastError();
            if (rezultat == ERROR_NOT_ENOUGH_MEMORY) {
                // Premalo prostora v pomnilniku. Funkcija WinHttpQueryHeaders() vrne tevilo potrebnega pomnilnika, zato popravimo dolino in poskusimo ponovno.
                dolzina /= sizeof(T_ZNAK);
                dolzina += RNIZ_KORAK_POVECEVANJA;
            } else {
                // Druge napake.
                podatek.Uporabi(_podatek, 0, FALSE);
                stanje = astanje(rezultat, AOSN_NAPAKA_HTTP_STATUS_ZAHTEVE);
                break;
            }
        }
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::PreberiOdgovor(TK_NIZ zelena_vrsta_odgovora, T_UINT &status, RNiz &vrsta_odgovora, RNizU &odgovor)
{
    assert(zelena_vrsta_odgovora);

    T_STANJE stanje;

    // Preberemo status.
    stanje = PreberiInformacijo(WINHTTP_QUERY_STATUS_CODE, status);
    if (USPESNO(stanje)) {
        // Preberemo vrsto odgovora.
        stanje = PreberiInformacijo(WINHTTP_QUERY_CONTENT_TYPE, vrsta_odgovora);
        if (USPESNO(stanje)) {
            T_UINT
                indeks_kodiranja = vrsta_odgovora.Najdi(_T(';')),
                dolzina_vrste_odgovora = vrsta_odgovora.VrniDolzino();

            if (indeks_kodiranja != T_UINT_MAX) {
                T_UINT dolzina = dolzina_vrste_odgovora;

                dolzina_vrste_odgovora = indeks_kodiranja;
                indeks_kodiranja++;
                // Preskoimo vse presledke.
                for (; indeks_kodiranja < dolzina && aisspace(vrsta_odgovora[indeks_kodiranja]); indeks_kodiranja++);
                if (vrsta_odgovora.PrimerjajI(_T("charset"), indeks_kodiranja, 7) == 0) {
                    indeks_kodiranja += 7;
                    // Preskoimo vse presledke.
                    for (; indeks_kodiranja < dolzina && aisspace(vrsta_odgovora[indeks_kodiranja]); indeks_kodiranja++);
                    if (indeks_kodiranja < dolzina && vrsta_odgovora[indeks_kodiranja] == _T('=')) {
                        indeks_kodiranja++;
                        // Preskoimo vse presledke.
                        for (; indeks_kodiranja < dolzina && aisspace(vrsta_odgovora[indeks_kodiranja]); indeks_kodiranja++);
                    } else
                        indeks_kodiranja = T_UINT_MAX;
                } else
                    indeks_kodiranja = T_UINT_MAX;
            }

            // Preverimo, ali je vrnjen odgovor elene vrste.
            if (astrnicmp(vrsta_odgovora.Vrni(), dolzina_vrste_odgovora, zelena_vrsta_odgovora, astrlen(zelena_vrsta_odgovora)) == 0) {
                T_KODNASTRAN kodna_stran_odgovora = KODNASTRAN_NEZNANA;

                // Nastavimo kodno stran.
                if (indeks_kodiranja == T_UINT_MAX || vrsta_odgovora.PrimerjajI(_T("windows-1250"), indeks_kodiranja) == 0)
                    kodna_stran_odgovora = KODNASTRAN_WIN1250;
                else if (vrsta_odgovora.PrimerjajI(_T("utf-7"), indeks_kodiranja) == 0)
                    kodna_stran_odgovora = KODNASTRAN_UTF7;
                else if (vrsta_odgovora.PrimerjajI(_T("utf-8"), indeks_kodiranja) == 0)
                    kodna_stran_odgovora = KODNASTRAN_UTF8;
                else if (vrsta_odgovora.PrimerjajI(_T("utf-16"), indeks_kodiranja) == 0)
                    kodna_stran_odgovora = KODNASTRAN_UTF16;
                else
                    kodna_stran_odgovora = AOSN_PRIVZETI_CP;

                {
                    assert(MAX_PATH * sizeof(T_ZNAK) % sizeof(T_UZNAK) == 0);
                    T_VOID *podatki = niz.Odpusti(MAX_PATH, FALSE);
                    T_SIZE dolzina;

                    assert(podatki);
                    dolzina = Preberi(podatki, MAX_PATH * sizeof(T_ZNAK));
                    if (kodna_stran_odgovora == KODNASTRAN_UTF16)
                        odgovor.Nastavi((TK_UNICODE)podatki, dolzina / sizeof(T_UZNAK));
                    else
                        odgovor.NastaviA((TK_ASCII)podatki, dolzina, kodna_stran_odgovora);
                    niz.Uporabi((T_NIZ)podatki, 0, FALSE);
                }
            } else {
                // Dobili smo neznano napako.
                RFizDatoteka fdat;
                T_NIZ zacasni_niz;

                zacasni_niz = niz.Odpusti(MAX_PATH, FALSE);
                verify(::GetTempPath(MAX_PATH, zacasni_niz));
                niz.Uporabi(zacasni_niz, MAX_PATH, FALSE);
                niz.Prilepi(_T("ODGOVOR.HTM"));
                if (USPESNO(fdat.Odpri(niz, RDAT_PISANJE | RDAT_USTVARI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN))) {
                    fdat.ZapisiTok(*this);
                    verify(USPESNO(fdat.Zapri()));
                }

                stanje = AOSN_NAPAKA_HTTP_ODGOVOR;
            }
        }
    }

    return stanje;
}


T_STANJE ROddajnikHTTP::Zapri()
{
    assert(internet);
    assert(povezava);
    assert(zahteva);

    if (zahteva) {
        verify(::WinHttpCloseHandle(zahteva));
        zahteva = NULL;
    }

    if (povezava) {
        verify(::WinHttpCloseHandle(povezava));
        povezava = NULL;
    }

    if (internet) {
        verify(::WinHttpCloseHandle(internet));
        internet = NULL;
    }

    return stanje = A_OK;
}


T_SIZE ROddajnikHTTP::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(zahteva);

    // e imamo kaj brati, gremo brati.
    if (dolzina) {
        DWORD koliko = 0;

        stanje = ::WinHttpReadData(zahteva, podatki, (DWORD)dolzina, &koliko) ?
            koliko ? A_OK : AOSN_NAPAKA_KONEC_TOKA :
            astanje(::GetLastError(), AOSN_NAPAKA_BRANJE);

        return koliko;
    } else {
        // Ni potrebe po branju, javimo, da je vse v redu.
        stanje = A_OK;
        return 0;
    }
}


T_SIZE ROddajnikHTTP::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(zahteva);

    // Pazimo, da dolina dela ni enaka 0, ker bi tak del prezgodaj naznanil konec pisanja.
    if (dolzina) {
        // Najprej zapiemo glavo dela. V nizu glava mora biti dovolj prostora za najveje 32/64-bitno tevilo + CR + LF + zakljuni znak.
        DWORD koliko = 0, koliko2;
        T_AZNAK glava[
#ifdef _WIN64
                20 + 3
#else
                10 + 3
#endif
            ];
        int n =
#if _MSC_VER < 1400
            _snprintf(
                glava,
                sizeof(glava),
#else
            _snprintf_s(
                glava,
                sizeof(glava),
                sizeof(glava),
#endif
            "%" T_SIZE_PRINTF_SIZE_PREFIX "X\r\n",
            dolzina);
        assert(n > 0);
        if (::WinHttpWriteData(zahteva, glava, n, &koliko)) {
            // Nato zapiemo vsebino dela.
            if (::WinHttpWriteData(zahteva, podatki, (DWORD)dolzina, &koliko)) {
                // Na koncu dela zapiemo e CRLF.
                stanje = ::WinHttpWriteData(zahteva, "\r\n", 2, &koliko2) ?
                    A_OK :
                    astanje(::GetLastError(), AOSN_NAPAKA_PISANJE);

                return koliko;
            } else {
                // Pri zapisovanju vsebine dela je prilo do napake.
                stanje = astanje(::GetLastError(), AOSN_NAPAKA_PISANJE);
                return koliko;
            }
        } else {
            // Pri zapisovanju glave dela je prilo do napake
            stanje = astanje(::GetLastError(), AOSN_NAPAKA_PISANJE);
            return 0;
        }
    } else {
        stanje = A_OK;
        return 0;
    }
}


T_STANJE ROddajnikHTTP::Izplakni()
{
    return A_OK;
}


//////////////////////////////////////////////////////////////////////
// ROddajnikIzjem
//////////////////////////////////////////////////////////////////////

const TK_NIZ ROddajnikIzjem::zelena_vrsta_odgovora[] = {
    _T("text/plain"),
    NULL
};


ROddajnikIzjem::ROddajnikIzjem() :
    ROddajnikHTTP(),
    kodirnik(ZLIB_STOPNJA_KOMPRESIJE_NAJVECJA, AHROSCAR_VELIKOST_BLOKA),
    vrsta_odgovora(AHROSCAR_VELIKOST_VRSTE_ODGOVORA),
    odgovor(AHROSCAR_VELIKOST_ODGOVORA)
{
}


T_STANJE ROddajnikIzjem::PosljiPorocilo(TK_NIZ ime_datoteke_porocila, T_UINT &status)
{
    assert(ime_datoteke_porocila);

    T_STANJE stanje;
    RMedFizDatoteka datoteka(NEVELJAVNA_ROCICA, AHROSCAR_VELIKOST_BLOKA);

    adebug(ADNE_DIAG, ADNE_AOSN _T("Poiljam poroilo o napaki %s ...\n"), ime_datoteke_porocila);
    stanje = datoteka.Odpri(ime_datoteke_porocila, RDAT_BRANJE | RDAT_SH_BERI | RDAT_BINARNO | RDAT_DOSTOP_ZAPOREDEN);
    if (USPESNO(stanje)) {
        // Pripravimo se na zaetek poiljanja.
        verify(USPESNO(DodajGlavo(_T("Content-Type"), T_UINT_MAX, _T("application/x-ahroscarz"), T_UINT_MAX)));
        stanje = ZacniPosiljanje();
        if (USPESNO(stanje)) {
            // Poljemo tehnino poroilo o napaki.
            ZapisiTok(datoteka);

            // Zakljuimo poiljanje.
            stanje = KoncajPosiljanje();
            if (USPESNO(stanje))
                stanje = PreberiOdgovor(status);
            else
                adebug(ADNE_NAPAKA, ADNE_AOSN _T("Poiljanje poroila o napaki ni uspelo (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);
        } else
            adebug(ADNE_NAPAKA, ADNE_AOSN _T("Poiljanje poroila o napaki ni uspelo (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);

        datoteka.Zapri();
        if (USPESNO(stanje)) {
            // Izbriemo datoteko s poroilom o napaki.
            AOsn::ZbrisiDatoteko(ime_datoteke_porocila);
            adebug(ADNE_DIAG, ADNE_AOSN _T("Poiljanje poroila o napaki konano.\n"));
        }
    } else
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Odpiranje datoteke s poroilom o napaki %s ni uspelo (koda") _T(T_STANJE_PRINTF) _T(".\n"), ime_datoteke_porocila, stanje);

    return stanje;
}


//////////////////////////////////////////////////////////////////////
// RRokovalnikIzjem
//////////////////////////////////////////////////////////////////////

RRokovalnikIzjem::RRokovalnikIzjem() :
    RNit(),
    stanje(A_OK),
    kljucavnica(0, _T("AOsnSymInitialize")),
    ep(NULL),
    nit(0),
    ime_datoteke_modula((T_UINT)MAX_PATH),
    ime_datoteke_porocila((T_UINT)MAX_PATH),
    datoteka(NEVELJAVNA_ROCICA, AHROSCAR_VELIKOST_BLOKA),
    izravnalnik(NULL, AHROSCAR_VELIKOST_BLOKA, AHROSCAR_VELIKOST_BLOKA),
    ime_datoteke2((T_UINT)MAX_PATH),
    sporocilo(AHROSCAR_VELIKOST_SPOROCILA)
{
    verify((indeks_tls_dodatnega_opisa = ::TlsAlloc()) != TLS_OUT_OF_INDEXES);

    _VrniImeDatotekeModula(NULL, ime_datoteke_modula);
}


RRokovalnikIzjem::~RRokovalnikIzjem()
{
    if (indeks_tls_dodatnega_opisa != TLS_OUT_OF_INDEXES) ::TlsFree(indeks_tls_dodatnega_opisa);
}


T_STANJE RRokovalnikIzjem::Odpri(const struct _EXCEPTION_POINTERS *_ep)
{
    assert(_ep);
    assert(_ep->ExceptionRecord);

    if (_ep->ExceptionRecord->ExceptionCode == STATUS_SXS_INVALID_DEACTIVATION) {
        // Napake STATUS_SXS_INVALID_DEACTIVATION ne sporoamo, ker nastane v MFC-ju ob razgradnji aplikacije, ki je zaradi predhodne napake postala neuporabna.
        return A_OK;
    }

    if (indeks_tls_dodatnega_opisa != TLS_OUT_OF_INDEXES) {
        // Prepiemo dodatni opis iz TLS-ja.
        TK_UNICODE _niz = (TK_UNICODE)::TlsGetValue(indeks_tls_dodatnega_opisa);
        if (_niz) dodatni_opis.Nastavi(_niz, T_UINT_MAX);
    }

    // Nastavimo parametre za poiljanje.
    ep     = _ep;
    nit    = ::GetCurrentThread();
    proces = ::GetCurrentProcess();

    // Nadaljujemo obdelavo napake v svoji niti.
    stanje = RNit::Odpri();
    if (USPESNO(stanje)) {
        if (USPESNO(pripravljen.Pocakaj(AHROSCAR_ZAGON_NITI))) {
            // Poakamo, da bo poroilo o napaki poslano.
            verify(USPESNO(Pocakaj()));
            verify(USPESNO(Zapri()));
        } else {
            // Zagon niti ni uspel. Verjetno smo v DllMain(), kjer niti ne delujejo. Da ne pride do smrtnega objema, zakljuimo nit.
            verify(USPESNO(Prekini(1)));
            verify(USPESNO(Zapri()));

            // Tukaj bi lahko klicali metodo Izvedi() in poskusili poslati poroilo o napaki kar iz te niti.
            // Toda poiljanje uporablja funkcijo HttpSendRequestEx(), ki se ji prav tako zatakne ob poskusu zagona svoje niti.
            // Zato osnovno poroilo in razirjeno poroilo o napaki raje shranimo v dve datoteki, ki ju poljemo ob naslednjem zagonu programa.
            ShraniPorocilo(FALSE);
        }
    } else {
        // Zagon niti za poiljanje poroila o napaki ni uspel. Poroilo shranimo v datoteko.
        ShraniPorocilo(FALSE);
    }

    return A_OK;
}


T_REZULTAT RRokovalnikIzjem::Izvedi(T_KAZALEC param)
{
    UNREFERENCED_PARAMETER(param);

    // Pri obdelavi izjeme moramo biti izjemno previdni, da ne pride e do kakne,
    // saj je AOsn::kljucavnica_izjem e zaklenjena in bi povzroilo smrtni objem,
    // e bi se AOsn::RokovalnikIzjem() klical ponovno.
    __try {
        verify(USPESNO(pripravljen.Postavi()));

        T_UINT status = T_UINT_MAX; // Status odgovora, ki ga prejmemo od strenika
        T_BOOL je_dodatni_opis = FALSE;
#ifndef AOSN_IZJEMA_NE_OBVESCAJ
        T_BOOL je_interaktiven = AOsn::JeProcesInteraktiven();
#endif

        // Poljemo trenutno poroilo o napaki.
#ifndef AOSN_IZJEMA_NE_OBVESCAJ
        if (je_interaktiven) {
#ifndef AOSN_IZJEMA_DODATNI_OPIS
            if (!dodatni_opis.JePrazen()) {
                // Vpraamo uporabnika, ali dovoli, da priloimo dodaten opis.
                T_UINT znakov = dodatni_opis.VrniDolzino();

                if (znakov <= AHROSCAR_DODATNI_OPIS_UI_MAX) {
                    // Dodatni opis je enak ali kraji od AHROSCAR_DODATNI_OPIS_UI_MAX znakov. Izpiemo celotnega.
                    sporocilo.Oblikuj(
                        _T("Med izvajanjem programa %.*ls je v njem ali enemu izmed njegovih modulov prilo do nepriakovane napake.\r\n")
                        _T("Program bo zdaj poslal poroilo o napaki Amebisovemu streniku. Mu sme priloiti tudi spodnje besedilo, ki bi razvojnemu osebju zelo pomagalo pri odpravi napake?\r\n\r\n%.*ls"), T_LOCALE_THREAD_DEFAULT,
                        ime_datoteke_modula.VrniDolzino(), ime_datoteke_modula.Vrni(),
                        dodatni_opis.VrniDolzino(), dodatni_opis.Vrni());
                } else {
                    // Dodatni opis je dalji od AHROSCAR_DODATNI_OPIS_UI_MAX znakov. Izpiemo samo prvih AHROSCAR_DODATNI_OPIS_UI_MAX.
                    T_UNICODE niz = dodatni_opis.Odpusti(znakov, FALSE);
                    T_UZNAK znak = niz[AHROSCAR_DODATNI_OPIS_UI_MAX - 1];
                    niz[AHROSCAR_DODATNI_OPIS_UI_MAX - 1] = 0;
                    sporocilo.Oblikuj(
                        _T("Med izvajanjem programa %.*ls je v njem ali enemu izmed njegovih modulov prilo do nepriakovane napake.\r\n")
                        _T("Program bo zdaj poslal poroilo o napaki Amebisovemu streniku. Mu sme priloiti tudi spodnje besedilo, ki bi razvojnemu osebju zelo pomagalo pri odpravi napake?\r\n\r\n%.*ls[...]"), T_LOCALE_THREAD_DEFAULT,
                        ime_datoteke_modula.VrniDolzino(), ime_datoteke_modula.Vrni(),
                        AHROSCAR_DODATNI_OPIS_UI_MAX, niz);
                    niz[AHROSCAR_DODATNI_OPIS_UI_MAX - 1] = znak;
                    dodatni_opis.Uporabi(niz, znakov, FALSE);
                }
                if (::MessageBox(NULL, sporocilo.Vrni(), _T("Nepriakovana napaka"), MB_YESNO | MB_DEFBUTTON1 | MB_ICONERROR) == IDYES)
                    je_dodatni_opis = TRUE;
                sporocilo.Sprazni();
            } else
#endif
            {
                sporocilo.Oblikuj(_T("Med izvajanjem programa %.*ls je v njem ali enemu izmed njegovih modulov prilo do nepriakovane napake. "), T_LOCALE_THREAD_DEFAULT, ime_datoteke_modula.VrniDolzino(), ime_datoteke_modula.Vrni());
            }
        }
#endif

#if defined(AOSN_IZJEMA_DODATNI_OPIS) || defined(AOSN_IZJEMA_NE_OBVESCAJ)
        je_dodatni_opis = TRUE;
#endif

        verify(USPESNO(kljucavnica.Zakleni()));

        // Izklopimo obvestila "There is no disk in the drive" med nalaganjem PDB-jev (SYMOPT_FAIL_CRITICAL_ERRORS).
        ::SymSetOptions(::SymGetOptions() | SYMOPT_EXACT_SYMBOLS | SYMOPT_FAIL_CRITICAL_ERRORS);

        // Vzpostavimo knjinico DbgHelp, sicer na 64-bitnih platformah sprehod po skladu ne deluje pravilno.
        /*T_BOOL je_sym_vzpostavljen =*/ ::SymInitialize(proces, NULL, TRUE) ? TRUE : FALSE;

        // Najprej poljemo osnovno poroilo o napaki.
        stanje = ZapisiInPosljiPorocilo(FALSE, je_dodatni_opis, status);
        if (USPESNO(stanje)) {
            if (status == HTTP_STATUS_RETRY_WITH) {
                // Strenik je zaprosil, da mu poljemo e razirjeno poroilo o napaki.
                stanje = ZapisiInPosljiPorocilo(TRUE, je_dodatni_opis, status);
            }

#ifndef AOSN_IZJEMA_NE_OBVESCAJ
            if (je_interaktiven) {
                sporocilo.Prilepi(USPESNO(stanje) && status == HTTP_STATUS_OK ?
                    _T("Program je tehnino poroilo o napaki poslal Amebisovemu streniku, ki je vrnil naslednji odgovor:") :
                    _T("Program je tehnino poroilo o napaki poslal Amebisovemu streniku, a je pri sprejemu prilo do naslednje teave:"));
                sporocilo.PrilepiO(_T("\r\n\r\n%.*ls"), T_LOCALE_THREAD_DEFAULT, oddajnik.odgovor.VrniDolzino(), oddajnik.odgovor.Vrni());
            }
#endif
        } else {
            // Poiljanje preko spleta ni uspelo. Zato shranimo poroilo in upamo na uspeh naslednji.
            stanje = ShraniPorocilo(je_dodatni_opis);
#ifndef AOSN_IZJEMA_NE_OBVESCAJ
            if (je_interaktiven) {
                sporocilo.Prilepi(_T("Program je skual poslati tehnino poroilo o napaki Amebisovemu streniku, a mu to al ni uspelo. "));
                sporocilo.PrilepiO(USPESNO(stanje) ?
                    _T("Poroilo o napaki je uspeno shranil v datoteko in ga bo poskusil poslati ob svojem naslednjem zagonu.") :
                    _T("Poroilo o napaki je poskusil shraniti v datoteko, a mu tudi to ni uspelo (koda ") _T(T_STANJE_PRINTF) _T(")."), T_LOCALE_THREAD_DEFAULT, stanje);
            }
#endif
        }

        //if (je_sym_vzpostavljen) {
        //    // Zakljuimo knjinico DbgHelp.
        //    verify(::SymCleanup(proces));
        //}

        verify(USPESNO(kljucavnica.Odkleni()));

#ifndef AOSN_IZJEMA_NE_OBVESCAJ
        if (je_interaktiven) {
            sporocilo.Prilepi(_T("\r\n\r\nZa nevenosti se vam iskreno opraviujemo. Program se bo zdaj zaprl."));

            // Prikaemo obvestilo o napaki uporabniku.
            ::MessageBox(NULL, sporocilo.Vrni(), _T("Nepriakovana napaka"), MB_OK | MB_ICONERROR);
        }
#endif
    } __except(EXCEPTION_EXECUTE_HANDLER) {}

    return 0;
}


T_STANJE RRokovalnikIzjem::ShraniPorocilo(T_BOOL je_dodatni_opis)
{
    // eprav ime datoteke poroila vsebuje tudi ime nosilnega modula trenutnega procesa, globalno serializiramo zapisovanje poroila o napaki, ker:
    // 1. Ime modula je lahko dllhost.exe, rundll32.exe ali regsvr32.exe ipd. Gre pa v resnici za razline module, ki so napako zakuhali.
    // 2. V isti modul se lahko naloi ve naih samostojnih modulov (npr. Besana in uBesana za InDesign).
    // 3. V niti RIzjemeGlobalnaNit lahko ravnokar poteka poiljanje poroil o izjemah.
    RCakalnik cakalnik(AOsn::izjeme_nit->kljucavnica);

    stanje = A_OK;

    for (idx_datoteke = 0;; idx_datoteke++) {
        zacasni_niz = ime_datoteke_porocila.Odpusti(MAX_PATH, FALSE);
        verify(::GetTempPath(MAX_PATH, zacasni_niz));
        ime_datoteke_porocila.Uporabi(zacasni_niz, MAX_PATH, FALSE);
        ime_datoteke_porocila.PrilepiO(_T("%.*ls-%") _T(T_UINT_PRINTF_SIZE_PREFIX) _T("u.aho"), T_LOCALE_THREAD_DEFAULT,
            ime_datoteke_modula.VrniDolzino(), ime_datoteke_modula.Vrni(),
            idx_datoteke);

        if (AOsn::DatotekaObstaja(ime_datoteke_porocila.Vrni()))
            continue;

        stanje = datoteka.Odpri(ime_datoteke_porocila.Vrni(), RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_BINARNO | RDAT_DOSTOP_ZAPOREDEN);
        if (USPESNO(stanje)) {
            // Shranimo osnovno poroilo o napaki v datoteko.
            verify(USPESNO(oddajnik.kodirnik.Uporabi(&datoteka)));
            stanje = ZapisiPodatkeIzjeme(FALSE, je_dodatni_opis);
            oddajnik.kodirnik.Odpusti();
            verify(USPESNO(datoteka.Zapri()));

            if (USPESNO(stanje)) {
                zacasni_niz = astrrnchr(ime_datoteke_porocila.Vrni(), _T('.'), ime_datoteke_porocila.VrniDolzino());
                if (zacasni_niz)
                    ime_datoteke_porocila.Skrajsaj(zacasni_niz - ime_datoteke_porocila.Vrni());
                ime_datoteke_porocila.Prilepi(_T(".ahr"));

                stanje = datoteka.Odpri(ime_datoteke_porocila.Vrni(), RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_BINARNO | RDAT_DOSTOP_ZAPOREDEN);
                if (USPESNO(stanje)) {
                    // Shranimo e razirjeno poroilo o napaki v datoteko.
                    verify(USPESNO(oddajnik.kodirnik.Uporabi(&datoteka)));
                    stanje = ZapisiPodatkeIzjeme(TRUE, je_dodatni_opis);
                    oddajnik.kodirnik.Odpusti();
                    verify(USPESNO(datoteka.Zapri()));
                }
            }
        }

        break;
    }

    return stanje;
}


T_STANJE RRokovalnikIzjem::ZapisiInPosljiPorocilo(T_BOOL je_razsirjeno, T_BOOL je_dodatni_opis, T_UINT &status)
{
    // Nastavimo eleni jezik za odgovor in vrsto zahteve.
    T_ZNAK jezik[AOSN_KODA_ISO639_MAX]; // elen jezik odgovora
    T_UINT dolzina_jezika;              // Dolina elenega jezika odgovora

    adebug(ADNE_DIAG, ADNE_AOSN _T("Poiljam poroilo o napaki ...\n"));

    stanje = oddajnik.Odpri();
    if (NEUSPESNO(stanje)) {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Priprava oddajnika za poiljanje poroila o napaki ni uspela (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);
        goto napaka1;
    }

    dolzina_jezika = alangtoiso639(KODAJEZIKA2_SISTEMSKI, jezik);
    if (dolzina_jezika)
        verify(USPESNO(oddajnik.DodajGlavo(_T("Accept-Language"), T_UINT_MAX, jezik, dolzina_jezika)));
    verify(USPESNO(oddajnik.DodajGlavo(_T("Content-Type"), T_UINT_MAX, _T("application/x-ahroscarz"), T_UINT_MAX)));

    // Pripravimo se na zaetek poiljanja.
    stanje = oddajnik.ZacniPosiljanje();
    if (NEUSPESNO(stanje)) {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Poiljanje poroila o napaki ni uspelo (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);
        goto napaka2;
    }

    // Poljemo poroilo o napaki.
    verify(USPESNO(izravnalnik.Uporabi(&oddajnik)));
    verify(USPESNO(oddajnik.kodirnik.Uporabi(&izravnalnik)));
    stanje = ZapisiPodatkeIzjeme(je_razsirjeno, je_dodatni_opis);
    oddajnik.kodirnik.Odpusti();
    izravnalnik.Odpusti();

    // Zakljuimo poiljanje.
    stanje = oddajnik.KoncajPosiljanje();
    if (NEUSPESNO(stanje)) {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Zakljuek poiljanja poroila o napaki ni uspel (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);
        goto napaka2;
    }

    // Preberemo odgovor strenika.
    stanje = oddajnik.PreberiOdgovor(status);
    if (USPESNO(stanje))
        adebug(ADNE_DIAG, ADNE_AOSN _T("Poiljanje poroila o napaki konano.\n"));
    else
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Branje odgovora poiljanja poroila o napaki ni uspelo (koda") _T(T_STANJE_PRINTF) _T(".\n"), stanje);

napaka2:
    verify(USPESNO(oddajnik.Zapri()));
napaka1:
    return stanje;
}


T_STANJE RRokovalnikIzjem::ZapisiPodatkeIzjeme(T_BOOL je_razsirjeno, T_BOOL je_dodatni_opis)
{
    assert(ep);

    // Zapiemo ID datoteke.
    stanje = oddajnik.kodirnik.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    if (je_razsirjeno)
        oddajnik.kodirnik << *(const T_U4*)"AHR1";
    else
        oddajnik.kodirnik << *(const T_U4*)"AHO1";

    // Zapiemo platformo.
    stanje = oddajnik.kodirnik.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
#if defined(_M_IX86)
    oddajnik.kodirnik << (T_U2)IMAGE_FILE_MACHINE_I386;
#elif defined(_M_X64)
    oddajnik.kodirnik << (T_U2)IMAGE_FILE_MACHINE_AMD64;
#elif defined(_M_IA64)
    oddajnik.kodirnik << (T_U2)IMAGE_FILE_MACHINE_IA64;
#else
    #error "Neznana platforma."
#endif

    if (je_razsirjeno) {
        // Pretejemo tevilo izjem v zaporedju.
        for (izjema = ep->ExceptionRecord, st_izjem = 0; izjema; izjema = izjema->ExceptionRecord, st_izjem++);

        // Zapiemo tevilo izjem in podatke vseh izjem.
        stanje = oddajnik.kodirnik.VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
        oddajnik.kodirnik << (T_U4)st_izjem;
        for (izjema = ep->ExceptionRecord; izjema; izjema = izjema->ExceptionRecord)
            oddajnik.kodirnik << *izjema;
    } else {
        // Zapiemo le podatke prve izjeme.
        oddajnik.kodirnik << *ep->ExceptionRecord;
    }

    if (je_razsirjeno) {
        // Zapiemo kontekst.
        stanje = oddajnik.kodirnik.VrniStanje();
        if (NEUSPESNO(stanje)) return stanje;
        oddajnik.kodirnik << *ep->ContextRecord;
    }

    if (je_dodatni_opis && !dodatni_opis.JePrazen()) {
        oddajnik.kodirnik << *(const T_U4*)"DOOP";
        _ZapisiNiz(oddajnik.kodirnik, dodatni_opis);
    }

    // Na koncu dodamo e zaporedje klicev.
    return ZapisiSklad(je_razsirjeno);
}


T_STANJE RRokovalnikIzjem::ZapisiSklad(T_BOOL je_razsirjeno)
{
    assert(ep);
    assert(ep->ContextRecord);

    memcpy(&kontekst, ep->ContextRecord, sizeof(CONTEXT));

    // Pripravimo skladovni okvir.
    memset(&sklad_okvir, 0, sizeof(STACKFRAME64));
#if defined(_M_IX86)
    platforma                    = IMAGE_FILE_MACHINE_I386;
    sklad_okvir.AddrPC.Offset    = kontekst.Eip;
    sklad_okvir.AddrPC.Mode      = AddrModeFlat;
    sklad_okvir.AddrFrame.Offset = kontekst.Ebp;
    sklad_okvir.AddrFrame.Mode   = AddrModeFlat;
    sklad_okvir.AddrStack.Offset = kontekst.Esp;
    sklad_okvir.AddrStack.Mode   = AddrModeFlat;
#elif defined(_M_X64)
    platforma                    = IMAGE_FILE_MACHINE_AMD64;
    sklad_okvir.AddrPC.Offset    = kontekst.Rip;
    sklad_okvir.AddrPC.Mode      = AddrModeFlat;
    sklad_okvir.AddrFrame.Offset = kontekst.Rsp;
    sklad_okvir.AddrFrame.Mode   = AddrModeFlat;
    sklad_okvir.AddrStack.Offset = kontekst.Rsp;
    sklad_okvir.AddrStack.Mode   = AddrModeFlat;
//#elif defined(_M_IA64)
//    platforma                    = IMAGE_FILE_MACHINE_IA64;
//    sklad_okvir.AddrPC.Offset    = kontekst.StIIP;
//    sklad_okvir.AddrPC.Mode      = AddrModeFlat;
//    sklad_okvir.AddrFrame.Offset = kontekst.IntSp;
//    sklad_okvir.AddrFrame.Mode   = AddrModeFlat;
//    sklad_okvir.AddrBStore.Offset= kontekst.RsBSP;
//    sklad_okvir.AddrBStore.Mode  = AddrModeFlat;
//    sklad_okvir.AddrStack.Offset = kontekst.IntSp;
//    sklad_okvir.AddrStack.Mode   = AddrModeFlat;
#else
    #error "Neznana platforma."
#endif
    zacetni_naslov_modula = NULL;

    for (je_prvic = TRUE;; je_prvic = FALSE) {
        je_uspesno = ::StackWalk64(platforma, proces, nit, &sklad_okvir, &kontekst, NULL, SymFunctionTableAccess64, SymGetModuleBase64, NULL);
        if (!je_uspesno && !je_prvic) {
            // Mogoe funkciji ni uspelo, mogoe je sklada konec. Konamo.
            break;
        }

        if (sklad_okvir.AddrPC.Offset != 0) {
            // SymGetModuleBase64() v doloenih scenarijih ne deluje: vrne NULL, GetLastError() vrne 0. Zato zaetni naslov modula doloimo z VirtualQuery().
            if (::VirtualQuery((LPCVOID)sklad_okvir.AddrPC.Offset, &mbi, sizeof(MEMORY_BASIC_INFORMATION)) < 2*sizeof(PVOID))
                break;

            modul_zacetek = mbi.AllocationBase;
            if (!modul_zacetek)
                break;

            naslov_ukaza  = (TK_KAZALEC)sklad_okvir.AddrPC.Offset;

            // Ali dani naslov sploh pripada kodnemu segmentu?
            if (!naslov_ukaza || ::IsBadCodePtr((FARPROC)naslov_ukaza))
                break;

            // Dodamo diagnostino informacijo.
            if (zacetni_naslov_modula != modul_zacetek) {
                // Modul se je spremenil.
                // Da ne odpiramo, preslikujemo in iemo diagnostinih informacij vsaki znova, si zapomnimo podatke za naslednji.

                // Zapiemo oznako modula v poroilo.
                oddajnik.kodirnik << *(const T_U4*)"MOD1";
                stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

                zacetni_naslov_modula = modul_zacetek;
                glava_NT              = ::ImageNtHeader((PVOID)modul_zacetek);

                // Zapiemo ime modula v poroilo.
                _VrniImeDatotekeModula((HMODULE)modul_zacetek, ime_datoteke2);
                oddajnik.kodirnik << (T_U4)ime_datoteke2.VrniDolzino();
                stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;
                oddajnik.kodirnik.ZapisiTab(ime_datoteke2.Vrni(), sizeof(T_UZNAK), ime_datoteke2.VrniDolzino());
                stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

                // Zapiemo dodatne podatke modula v poroilo.
                oddajnik.kodirnik << (T_U4)(glava_NT ? glava_NT->FileHeader.TimeDateStamp : 0);
                stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

                if (je_razsirjeno) {
                    oddajnik.kodirnik << (T_U4)(glava_NT ? glava_NT->OptionalHeader.CheckSum : 0);
                    stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;
                }

                // Zapiemo zaetni naslov modula v poroilo.
                oddajnik.kodirnik << (T_UINT)modul_zacetek;
                stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;
            }

            // Zapiemo oznako funkcije v poroilo.
            oddajnik.kodirnik << *(const T_U4*)"FUN1";
            stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

            odmik = (TK_KAZALEC)((T_SIZE)naslov_ukaza - (T_SIZE)modul_zacetek);

            // Zapiemo relativni naslov ukaza v poroilo.
            oddajnik.kodirnik << (T_UINT)odmik;
            stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

            if (je_razsirjeno) {
                sklad_od = (TK_KAZALEC)sklad_okvir.AddrStack.Offset;
                sklad_do = (TK_KAZALEC)sklad_okvir.AddrFrame.Offset;

                if (je_uspesno && sklad_do && sklad_od && sklad_od < sklad_do) {
                    // Zapiemo e vsebino sklada.
                    dolzina_sklada = (T_SIZE)sklad_do - (T_SIZE)sklad_od;

                    oddajnik.kodirnik << dolzina_sklada;
                    stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

                    oddajnik.kodirnik << (T_SIZE)sklad_od;
                    stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;

                    oddajnik.kodirnik.Zapisi(sklad_od, dolzina_sklada);
                    stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;
                } else {
                    // Sklada ne zapisujemo.
                    oddajnik.kodirnik << (T_SIZE)0;
                    stanje = oddajnik.kodirnik.VrniStanje(); if (NEUSPESNO(stanje)) return stanje;
                }
            }
        } else {
            // Prili smo do konca sklada.
            break;
        }

        if (sklad_okvir.AddrPC.Offset == sklad_okvir.AddrReturn.Offset) {
            // Naleteli smo na zanko. e nadaljujemo, se bomo veno vrteli v krogu.
            break;
        }
    }

    return A_OK;
}


//////////////////////////////////////////////////////////////////////
// RIzjemeGlobalnaNit
//////////////////////////////////////////////////////////////////////

RIzjemeGlobalnaNit::RIzjemeGlobalnaNit() :
    RNit(),
    kljucavnica(0, _T("AOsnShranjevanjeInPosiljanjeIzjem"))
{
}


T_REZULTAT RIzjemeGlobalnaNit::Izvedi(T_KAZALEC param)
{
    UNREFERENCED_PARAMETER(param);

    adebug(ADNE_DIAG, ADNE_AOSN _T("Vzpostavljam prestrezanje in poiljanje izjem ...\n"));

    // Poiljanje shranjenih izjem mora soasno opravljati samo eden izmed modulov procesa, saj so datoteke poimenovane z imenom procesa, ne modula in so torej skupne za cel proces.
    // Pravzaprav je isto sprejemljivo, e samo en proces na raunalniku poilja naenkrat. Zato je globalna kljuavnica sprejemljiva.
    RCakalnik cakalnik(kljucavnica);

    RNizU ime_datoteke_modula;  // Ime datoteke trenutno izvajajoega modula
    RNiz ime_datoteke_porocila; // Pot datoteke poroila o napaki
    T_NIZ zacasni_niz;          // Pomoni kazalec na niz
    WIN32_FIND_DATA podatki;    // Podatki o najdeni datoteki
    HANDLE rocica;              // Roica iskalnika datotek

    _VrniImeDatotekeModula(NULL, ime_datoteke_modula);

    adebug(ADNE_DIAG, ADNE_AOSN _T("Iem prejnja poroila o napakah ...\n"));
    zacasni_niz = ime_datoteke_porocila.Odpusti(MAX_PATH, FALSE);
    verify(::GetTempPath(MAX_PATH, zacasni_niz));
    ime_datoteke_porocila.Uporabi(zacasni_niz, MAX_PATH, FALSE);
    ime_datoteke_porocila.PrilepiO(_T("%.*ls-*.aho"), T_LOCALE_THREAD_DEFAULT, ime_datoteke_modula.VrniDolzino(), ime_datoteke_modula.Vrni());

    // Poiemo vsa poroila o napakah.
    rocica = ::FindFirstFileEx(ime_datoteke_porocila.Vrni(), (FINDEX_INFO_LEVELS)1 /*FindExInfoBasic*/, &podatki, FindExSearchNameMatch, NULL, 0);
    if (rocica != INVALID_HANDLE_VALUE) {
        T_STANJE stanje;
        ROddajnikIzjem oddajnik; // Oddajnik za poiljanje izjem

        // Pripravimo oddajnik.
        stanje = oddajnik.Odpri();
        if (USPESNO(stanje)) {
            do {
                T_UINT status = T_UINT_MAX; // Status odgovora, ki ga prejmemo od strenika

                zacasni_niz = ime_datoteke_porocila.Odpusti(MAX_PATH, FALSE);
                verify(::GetTempPath(MAX_PATH, zacasni_niz));
                ime_datoteke_porocila.Uporabi(zacasni_niz, MAX_PATH, FALSE);
                ime_datoteke_porocila.Prilepi(podatki.cFileName, _countof(podatki.cFileName));

                // Poljemo osnovno poroilo o napaki.
                stanje = oddajnik.PosljiPorocilo(ime_datoteke_porocila.Vrni(), status);
                zacasni_niz = astrrnchr(ime_datoteke_porocila.Vrni(), _T('.'), ime_datoteke_porocila.VrniDolzino());
                if (zacasni_niz)
                    ime_datoteke_porocila.Skrajsaj(zacasni_niz - ime_datoteke_porocila.Vrni());
                ime_datoteke_porocila.Prilepi(_T(".ahr"));
                if (USPESNO(stanje)) {
                    if (status == HTTP_STATUS_RETRY_WITH) {
                        // Poljemo e razirjeno poroilo.
                        stanje = oddajnik.PosljiPorocilo(ime_datoteke_porocila.Vrni(), status);
                    } else {
                        // Izbriemo razirjeno poroilo o napaki, saj smo streniku sicer uspeli poslati osnovno poroilo, a ga razirjeno ve ne zanima.
                        AOsn::ZbrisiDatoteko(ime_datoteke_porocila.Vrni());
                    }
                }
            } while (::FindNextFile(rocica, &podatki));

            verify(USPESNO(oddajnik.Zapri()));
        }

        verify(::FindClose(rocica));
    }

    adebug(ADNE_DIAG, ADNE_AOSN _T("Prestrezanje in poiljanje izjem vzpostavljeno.\n"));
    return 0;
}


//////////////////////////////////////////////////////////////////////
// Interne funkcije
//////////////////////////////////////////////////////////////////////

inline T_VOID _VrniImeDatotekeModula(HMODULE modul, RNizU &ime_datoteke_modula)
{
    T_UNICODE pot_datoteke;
    T_UINT zacetek_imena_datoteke;

    pot_datoteke = ime_datoteke_modula.Odpusti(MAX_PATH, FALSE);
    assert(pot_datoteke);
    ime_datoteke_modula.Uporabi(pot_datoteke, ::GetModuleFileNameW(modul, pot_datoteke, MAX_PATH), FALSE);
    zacetek_imena_datoteke = ime_datoteke_modula.NajdiNazaj(_T(LOCILO_IMENIKOV));
    if (zacetek_imena_datoteke != T_UINT_MAX) {
        zacetek_imena_datoteke++;
        ime_datoteke_modula.Zbrisi(0, zacetek_imena_datoteke);
    }
}

#endif // AOSN_IZJEMA
