/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_KANAL


//////////////////////////////////////////////////////////////////////
// RKanal
//////////////////////////////////////////////////////////////////////

RKanal::RKanal(T_ROCICAOS vh, T_ROCICAOS iz) :
    RElement(),
    vhod(vh),
    izhod(iz)
{
}


T_STANJE RKanal::Odpri(T_SIZE dolzina_medpomnilnika)
{
    return RKanal::Odpri(vhod, izhod, dolzina_medpomnilnika);
}


T_STANJE RKanal::Odpri(RSisTokPodatkov &vhod, RSisTokPodatkov &izhod, T_SIZE dolzina_medpomnilnika)
{
    T_ROCICAOS rocica_vhoda, rocica_izhoda;

    if (::CreatePipe(&rocica_izhoda, &rocica_vhoda, NULL, (DWORD)dolzina_medpomnilnika)) {
        verify(USPESNO(vhod.Uporabi(rocica_vhoda)));
        verify(USPESNO(izhod.Uporabi(rocica_izhoda)));

        return A_OK;
    } else
        return AOSN_NAPAKA_ODPIRANJE;
}


//////////////////////////////////////////////////////////////////////
// RImenovanKanal
//////////////////////////////////////////////////////////////////////

RImenovanKanal::RImenovanKanal(T_ROCICAOS rocica) : RSisTokPodatkov(rocica)
{
}


RImenovanKanal::~RImenovanKanal()
{
}


T_STANJE RImenovanKanal::Ustvari(TK_NIZ ime, T_UINT nacin, T_UINT nacin_instance, T_UINT najvec_klientov, T_SIZE izhodni_medpomnilnik, T_SIZE vhodni_medpomnilnik, T_UINT naj_cas)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    rocica = ::CreateNamedPipe(
        ime,
        (DWORD)nacin,
        (DWORD)(nacin_instance | PIPE_WAIT),
        (DWORD)najvec_klientov,
        (DWORD)izhodni_medpomnilnik,
        (DWORD)vhodni_medpomnilnik,
        (DWORD)naj_cas,
        NULL
    );

    if (rocica != INVALID_HANDLE_VALUE && (T_UINT)rocica != ERROR_INVALID_PARAMETER)
        return stanje = A_OK;
    else
        return stanje = AOSN_NAPAKA_ODPIRANJE;
}


T_STANJE RImenovanKanal::Sprejmi()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    if (::ConnectNamedPipe(rocica, NULL))
        return A_OK;
    else {
        DWORD rezultat = ::GetLastError();
        return rezultat == ERROR_PIPE_CONNECTED ? A_OK : astanje(rezultat, AOSN_NAPAKA_SPREJMI);
    }
}


T_STANJE RImenovanKanal::Zakljuci()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    return ::DisconnectNamedPipe(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ZAKLJUCI);
}


T_STANJE RImenovanKanal::VrniNaslovVrstnika(RNiz &naslov) const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    T_ZNAK uporabnik[UNLEN];

    if (::GetNamedPipeHandleState(rocica, NULL, NULL, NULL, NULL, uporabnik, UNLEN)) {
        naslov.Nastavi(uporabnik, UNLEN);
        return A_OK;
    } else {
        naslov.Sprazni();
        return astanje(::GetLastError(), AOSN_NAPAKA_NASLOV_VTICNICE);
    }
}

#endif // AOSN_KANAL
