/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_NIT


struct T_RDOGNITPARAM {
    T_ROCICAOS dogodek; // roica dogodka, ki ga sproi dogodkovna nit, ko je pripravljena na prevzem sporoil
    T_KAZALEC param;    // parameter niti
};


//////////////////////////////////////////////////////////////////////
// RNit
//////////////////////////////////////////////////////////////////////

RNit::RNit(T_ROCICAOS r) :
    koda_niti(0),
    param(NULL),
    RSinhronizator(r)
{
}


T_STANJE RNit::Odpri(T_KAZALEC _param, T_NITZACETNOSTANJE zacetno_stanje, T_UINT sklad)
{
    assert(NI_VELJAVNA_ROCICA(rocica));
    param = _param;
    rocica = (T_ROCICAOS)::_beginthreadex(NULL, (unsigned int)sklad, RNit::VstopnaTocka, this, zacetno_stanje, (unsigned*)&koda_niti);
    return JE_VELJAVNA_ROCICA(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_USTVARI);
}


T_STANJE RNit::Zamrzni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::SuspendThread(rocica)!=-1 ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ZAMRZNI);
}


T_STANJE RNit::Odmrzni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::ResumeThread(rocica)!=-1 ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODMRZNI);
}


T_STANJE RNit::Prekini(T_REZULTAT rezultat)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::TerminateThread(rocica, rezultat) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREKINI);
}


T_STANJE RNit::NastaviPrioriteto(T_NITPRIORITETA prioriteta)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::SetThreadPriority(rocica, prioriteta) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVI_PRI);
}


T_NITPRIORITETA RNit::VrniPrioriteto()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return (T_NITPRIORITETA)::GetThreadPriority(rocica);
}


T_STANJE RNit::VrniRezultat(T_REZULTAT *rezultat)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::GetExitCodeThread(rocica, (LPDWORD)rezultat) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_REZULTAT);
}


T_STANJE RNit::Uporabi(T_ROCICAOS rocica, T_UINT koda_niti)
{
    assert(!this->koda_niti);

    T_STANJE stanje;

    stanje = RSinhronizator::Uporabi(rocica);
    if (USPESNO(stanje)) {
        this->koda_niti = koda_niti;
        return A_OK;
    } else
        return stanje;
}


T_ROCICAOS RNit::Odpusti()
{
    koda_niti = 0;
    return RSinhronizator::Odpusti();
}


T_STANJE RNit::Zapri()
{
    koda_niti = 0;
    return RSinhronizator::Zapri();
}


T_STANJE RNit::Odkleni()
{
    return Odpri();
}


T_REZULTAT RNIT_API RNit::VstopnaTocka(T_KAZALEC param)
{
    // Preko te statine funkcije vedno vstopamo v nit.
    // Prekopiramo parameter niti in original sprostimo.
    assert(param);
    RNit *nit = (RNit*)param;

    // Sedaj izvedemo telo niti.
    T_REZULTAT rezultat = nit->Izvedi(nit->param);
    return rezultat;
}


//////////////////////////////////////////////////////////////////////
// RDogNit
//////////////////////////////////////////////////////////////////////

T_STANJE RDogNit::Odpri(T_UINT dolzina_vrste, T_KAZALEC param, T_NITZACETNOSTANJE zacetno_stanje, T_UINT sklad)
{
    T_STANJE stanje;

#ifdef RDOGNIT_KANAL_VER
    stanje = vrsta.Odpri(dolzina_vrste * sizeof(T_SPOROCILO));
    return USPESNO(stanje) ? RNit::Odpri(param, zacetno_stanje, sklad) : stanje;
#else
    UNREFERENCED_PARAMETER(dolzina_vrste);

    RDogodek pripravljen;
    T_RDOGNITPARAM p = {
        pripravljen.VrniRocico(),
        param
    };

    stanje = RNit::Odpri(&p, zacetno_stanje, sklad);
    if (USPESNO(stanje)) {
        // Poakamo, da se nit pripravi na sprejem sporoil.
        verify(USPESNO(pripravljen.Pocakaj()));
        return A_OK;
    } else
        return stanje;
#endif
}


#ifdef RDOGNIT_KANAL_VER

T_STANJE RDogNit::Zapri()
{
    T_STANJE stanje = RNit::Zapri();

    verify(USPESNO(vrsta.vhod.Zapri()));
    verify(USPESNO(vrsta.izhod.Zapri()));

    return stanje;
}


T_STANJE RDogNit::Uporabi(T_ROCICAOS rocica, T_UINT koda_niti)
{
    // Dogodkovne niti ne moremo kar pripeti na razred.
    // Samo iz njene roice in/ali kode niti ne moremo priti do njene vrste za sporoila. e pa ji ne moremo poiljati sporoil, je vseeno, e nit pripnemo na RNit.
    assert(0);
    return AOSN_NAPAKA_NI_PODPRTO;
}


T_ROCICAOS RDogNit::Odpusti()
{
    T_ROCICAOS rocica = RNit::Odpusti();

    verify(USPESNO(vrsta.vhod.Zapri()));
    verify(USPESNO(vrsta.izhod.Zapri()));

    return rocica;
}


T_STANJE RDogNit::Sporoci(T_UINT koda_sporocila, T_UINT param1, T_KAZALEC param2)
{
    T_SPOROCILO sporocilo = {
        koda_sporocila,
        param1,
        param2
    };

    // Zapiemo sporoilo.
    return vrsta.vhod.Zapisi(&sporocilo, sizeof(T_SPOROCILO)) == (T_DOLZINA)sizeof(T_SPOROCILO) ? A_OK : vrsta.vhod.VrniStanje();
}

#endif


T_REZULTAT RDogNit::Pozeni(T_KAZALEC param)
{
    T_REZULTAT rezultat;

    // Poenemo telo niti.
#ifdef RDOGNIT_KANAL_VER
    rezultat = Izvedi(param);
#else
    T_RDOGNITPARAM p = {
        NEVELJAVNA_ROCICA,
        param
    };

    rezultat = Izvedi(&p);
#endif

    return rezultat;
}


T_REZULTAT RDogNit::Izvedi(T_KAZALEC param)
{
    T_REZULTAT rezultat;

#ifdef RDOGNIT_KANAL_VER
    T_SPOROCILO sporocilo;
    T_SIZE dolzina;

    rezultat = ObZagonu(param);

    while (!rezultat) {
        // Ne zaklepamo kljuavnice za dostop do vrste, saj predpostavimo, da iz vrste bere samo ta nit.
        dolzina = vrsta.izhod.PreberiTab(&sporocilo, sizeof(T_SPOROCILO), 1);

        // Ali je zmanjkalo podatkov?
        if (NEUSPESNO(vrsta.izhod.VrniStanje()) || dolzina < 1)
            break;

        // e rokovalnik dogodkov vrne kodo razlino od 0 smatramo, da se nit eli zakljuiti.
        rezultat = RokovalnikDogodkov(sporocilo.koda, sporocilo.param1, sporocilo.param2);
        if (rezultat)
            break;

        // e je koda sporoila SP_KONEC, zakljuimo nit.
        if (sporocilo.koda == SP_KONEC)
            break;
    }

    ObZakljucku(param);
#else
    assert(param);

    T_RDOGNITPARAM *p = (T_RDOGNITPARAM*)param;
    T_KAZALEC _param = p->param;  // Vrednost parametra si moramo zapomniti, saj je struktura p na skladu metode RDogNit::Odpri() in bo njena usoda negotova takoj, ko bomo sproili dogodek "pripravljen".
    MSG sporocilo;
    BOOL bRes;
    T_UINT koda_sporocila;

    // Potipamo vrsto sporoil, da jo OS ustvari.
    PeekMessage(&sporocilo, NULL, WM_USER, WM_USER, PM_NOREMOVE);

    // Sporoimo niti, ki je ustvarila dogodkovno nit, da je ta pripravljena na sprejem sporoil.
    if (JE_VELJAVNA_ROCICA(p->dogodek)) {
        RDogodek pripravljen(p->dogodek);
        verify(USPESNO(pripravljen.Postavi()));
        pripravljen.Odpusti();
    }

    rezultat = ObZagonu(_param);

    while (!rezultat) {
        bRes = ::GetMessage(&sporocilo, NULL, 0, 0);
        if (bRes == -1) {
            rezultat = (T_REZULTAT)-1;
            break;
        }

        if (RDOGNIT_KODA_SPOROCILA_ZACETEK <= sporocilo.message && sporocilo.message < RDOGNIT_KODA_SPOROCILA_KONEC) {
            koda_sporocila = sporocilo.message - RDOGNIT_KODA_SPOROCILA_ZACETEK;

            rezultat = RokovalnikDogodkov(koda_sporocila, (T_UINT)sporocilo.wParam, (T_VOID*)sporocilo.lParam);
            if (rezultat) {
                // e rokovalnik dogodkov vrne kodo razlino od 0 smatramo, da se nit eli zakljuiti.
                break;
            }

            // e je koda sporoila SP_KONEC, zakljuimo nit.
            if (koda_sporocila == SP_KONEC)
                break;
        } else {
            // Koda sporoila ni v okviru naih kod. Najbr je prila od Windowsov, zato jo ignoriramo.
            // Vendar se to ne bi smelo zgoditi. Je kdo odprl uporabniki vmesnik v tej niti? V rokovalniku dogodkov?
            // Mogoe se je zgodil kak assert v delovni niti?
            // Sporoilo polje tudi arecv(), kadar mu zmanjka podatkov na vhodu in gre v akajoe stanje.
            //assert(0);
        }
    }

    ObZakljucku(_param);
#endif

    return rezultat;
}


T_REZULTAT RDogNit::ObZagonu(T_VOID*)
{
    return 0;
}


T_VOID RDogNit::ObZakljucku(T_VOID*)
{
}


T_REZULTAT RDogNit::RokovalnikDogodkov(T_UINT, T_UINT, T_VOID*)
{
    return 0;
}


//////////////////////////////////////////////////////////////////////
// RDogNitVN
//////////////////////////////////////////////////////////////////////

#ifdef RDOGNIT_KANAL_VER

T_STANJE RDogNitVN::Sporoci(T_UINT koda_sporocila, T_UINT param1, T_KAZALEC param2)
{
    RCakalnik cakalnik(kljucavnica);
    return RDogNit::Sporoci(koda_sporocila, param1, param2);
}

#endif


//////////////////////////////////////////////////////////////////////
// RStopnjaCevovoda
//////////////////////////////////////////////////////////////////////

RStopnjaCevovoda::RStopnjaCevovoda(RDogNit *naslednja)
{
    this->naslednja = naslednja;
}


//////////////////////////////////////////////////////////////////////
// RCasovnaPast
//////////////////////////////////////////////////////////////////////

#ifndef _WIN32_WCE

T_STANJE RCasovnaPast::Odpri(T_KAZALEC param, T_NITZACETNOSTANJE zacetno_stanje, T_UINT sklad)
{
    koncaj = FALSE;
    return RNit::Odpri(param, zacetno_stanje, sklad);
}


T_REZULTAT RCasovnaPast::ObAlarmu(T_VOID*)
{
    return 0;
}


T_REZULTAT RCasovnaPast::Izvedi(T_KAZALEC param)
{
    T_REZULTAT rezultat = (T_REZULTAT)-1;

    for (;;) {
        if (USPESNO(casomer.Pocakaj())) {
            if (!koncaj)
                rezultat = ObAlarmu(param);
            else
                verify(USPESNO(casomer.Prekini()));
                break;
        }
    }

    return rezultat;
}

#endif


//////////////////////////////////////////////////////////////////////
// RDiagNit
//////////////////////////////////////////////////////////////////////

RDiagNit::RDiagNit() :
    RNit(),
    dogodek_pripravljen(RSEM_ROCNI_RESET),
    dogodek_koncaj(RSEM_ROCNI_RESET),
    dogodek_pomnilnik_pripravljen(RSEM_POSTAVLJEN, _T("DBWIN_BUFFER_READY")),
    dogodek_podatki_pripravljeni(0, _T("DBWIN_DATA_READY")),
    rocica_podatkov(NULL)
{
}

RDiagNit::~RDiagNit()
{
    assert(!rocica_podatkov);
}


T_STANJE RDiagNit::Odpri(T_KAZALEC param, T_NITZACETNOSTANJE zacetno_stanje, T_UINT sklad)
{
    T_STANJE stanje;

    // Odpremo/ustvarimo skupni pomnilnik DBWIN_BUFFER.
    rocica_podatkov = (T_ROCICAOS)::OpenFileMapping(FILE_MAP_READ, FALSE, _T("DBWIN_BUFFER"));
    if (!rocica_podatkov) {
        rocica_podatkov = (T_ROCICAOS)::CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, sizeof(T_DIAGNIT_PODATKI), _T("DBWIN_BUFFER"));
        if (!rocica_podatkov) {
            stanje = astanje(::GetLastError(), AOSN_NAPAKA_USTVARI);
            goto napaka1;
        }
    }

    // Odpremo nit.
    stanje = RNit::Odpri(param, zacetno_stanje, sklad);
    if (NEUSPESNO(stanje))
        goto napaka2;

    // Nastavimo najvijo mono prioriteto.
    //verify(::SetPriorityClass(::GetCurrentProcess(), REALTIME_PRIORITY_CLASS));
    verify(::SetThreadPriority((HANDLE)rocica, THREAD_PRIORITY_TIME_CRITICAL));

    return A_OK;

napaka2:
    verify(::CloseHandle(rocica_podatkov));
    rocica_podatkov = NULL;
napaka1:
    return stanje;
}


T_STANJE RDiagNit::Zapri()
{
    T_STANJE stanje = RNit::Zapri();

    verify(::CloseHandle(rocica_podatkov));
    rocica_podatkov = NULL;

    return stanje;
}


T_REZULTAT RDiagNit::ObZagonu(T_KAZALEC param)
{
    UNREFERENCED_PARAMETER(param);

    return 0;
}


T_REZULTAT RDiagNit::ObSporocilu(const T_DIAGNIT_PODATKI *podatki)
{
    UNREFERENCED_PARAMETER(podatki);

    return 0;
}


T_VOID RDiagNit::ObZakljucku(T_KAZALEC param)
{
    UNREFERENCED_PARAMETER(param);
}


T_REZULTAT RDiagNit::Izvedi(T_KAZALEC param)
{
    T_REZULTAT rez;
    HANDLE dogodki[] = {
        dogodek_podatki_pripravljeni.VrniRocico(),
        dogodek_koncaj.VrniRocico()
    };
    T_DIAGNIT_PODATKI *podatki;

    rez = ObZagonu(param);
    // Naznanimo, da smo pripravljeni za zaetek, oziroma sprostimo na ta dogodek akajoe niti, e je med vzpostavitvijo prilo do napake in bomo to nit zakljuili.
    verify(USPESNO(dogodek_pripravljen.Postavi()));
    if (rez)
        return rez;

    for (;;) {
        // Poakamo na podatke ali na konec.
        rez = ::WaitForMultipleObjects(_countof(dogodki), dogodki, FALSE, INFINITE);
        if (rez == WAIT_OBJECT_0) {
            // Preslikamo pomnilnik.
            podatki = (T_DIAGNIT_PODATKI*)::MapViewOfFile(rocica_podatkov, SECTION_MAP_READ, 0, 0, 0);
            if (podatki) {
                // Posredujemo podatke navidezni niti.
                rez = ObSporocilu(podatki);
                verify(::UnmapViewOfFile(podatki));
            } else
                rez = 1;

            // Naznanimo, da smo pripravljeni za sprejem sporoila.
            verify(USPESNO(dogodek_pomnilnik_pripravljen.Postavi()));

            if (rez) break;
        } else {
            rez = 0;
            break;
        }
    }

    ObZakljucku(param);
    return rez;
}

#endif // AOSN_NIT
