/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"


#ifndef _DLL

// Definiramo funkciji _input() in _winput(), ki sta sicer interni v knjinici CRT.
#if _MSC_VER >= 1500

#ifdef __cplusplus
extern "C"
#endif
int __cdecl _input_l(_Inout_ FILE * _File, _In_z_ __format_string const unsigned char *, _In_opt_ _locale_t _Locale, va_list _ArgList);

#ifdef __cplusplus
extern "C"
#endif
int __cdecl _winput_l(_Inout_ FILE * _File, _In_z_ __format_string const wchar_t *, _In_opt_ _locale_t _Locale, va_list _ArgList);

#else

#define _input_l(infile, format, plocinfo, arglist)  _input(infile, format, arglist)
#define _winput_l(infile, format, plocinfo, arglist) _winput(infile, format, arglist)

#ifdef __cplusplus
extern "C"
#endif
int __cdecl _input(FILE *, const unsigned char *, va_list);

#ifdef __cplusplus
extern "C"
#endif
int __cdecl _winput(FILE *, const wchar_t *, va_list);

#endif

#endif


//////////////////////////////////////////////////////////////////////
// Funkcije
//////////////////////////////////////////////////////////////////////

#ifndef _DLL

T_INT ainput(FILE *dat, TK_ASCII format, T_LOCALE locale, va_list parametri)
{
    return _input_l(dat, (const unsigned char*)format, locale, parametri);
}


T_INT ainput(FILE *dat, TK_UNICODE format, T_LOCALE locale, va_list parametri)
{
    return _winput_l(dat, format, locale, parametri);
}

#endif


//////////////////////////////////////////////////////////////////////
// Konstruktorji / Destruktor RNizA/RNizU
//////////////////////////////////////////////////////////////////////

RNizA::RNizA(T_UINT koliko) : RNizVP<T_AZNAK>(koliko)
{
}

RNizU::RNizU(T_UINT koliko) : RNizVP<T_UZNAK>(koliko)
{
}

RNizA::RNizA(T_AZNAK znak, T_UINT koliko) : RNizVP<T_AZNAK>(znak, koliko)
{
}

RNizU::RNizU(T_AZNAK znak, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_UZNAK>()
{
    assert(0); // TODO: Preizkusi ta del kode.
    NastaviA(znak, koliko, kodna_stran);
}

RNizA::RNizA(T_UZNAK znak, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_AZNAK>()
{
    assert(0); // TODO: Preizkusi ta del kode.
    NastaviU(znak, koliko, kodna_stran);
}

RNizU::RNizU(T_UZNAK znak, T_UINT koliko) : RNizVP<T_UZNAK>(znak, koliko)
{
}

RNizA::RNizA(TK_ASCII niz, T_UINT koliko) : RNizVP<T_AZNAK>(niz, koliko)
{
}

RNizU::RNizU(TK_ASCII niz, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_UZNAK>()
{
    NastaviA(niz, koliko, kodna_stran);
}

RNizA::RNizA(TK_UNICODE niz, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_AZNAK>()
{
    NastaviU(niz, koliko, kodna_stran);
}

RNizU::RNizU(TK_UNICODE niz, T_UINT koliko) : RNizVP<T_UZNAK>(niz, koliko)
{
}

RNizA::RNizA(const RNizP<T_AZNAK> &niz, T_UINT koliko) : RNizVP<T_AZNAK>(niz, koliko)
{
    assert(0); // TODO: Preizkusi ta del kode.
}

RNizU::RNizU(const RNizP<T_AZNAK> &niz, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_UZNAK>()
{
    assert(0); // TODO: Preizkusi ta del kode.
    NastaviA(niz, koliko, kodna_stran);
}

RNizA::RNizA(const RNizP<T_UZNAK> &niz, T_UINT koliko, T_KODNASTRAN kodna_stran) : RNizVP<T_AZNAK>()
{
    assert(0); // TODO: Preizkusi ta del kode.
    NastaviU(niz, koliko, kodna_stran);
}

RNizU::RNizU(const RNizP<T_UZNAK> &niz, T_UINT koliko) : RNizVP<T_UZNAK>(niz, koliko)
{
}

RNizA::RNizA(HMODULE modul, T_UINT koda, T_KODAJEZIKA2 jezik, T_KODNASTRAN kodna_stran) : RNizVP<T_AZNAK>()
{
    NastaviVU(modul, koda, jezik, kodna_stran);
}

RNizU::RNizU(HMODULE modul, T_UINT koda, T_KODAJEZIKA2 jezik) : RNizVP<T_UZNAK>(modul, koda, jezik)
{
}


//////////////////////////////////////////////////////////////////////
// Funkcije RNizA/RNizU
//////////////////////////////////////////////////////////////////////

T_UINT RNizU::PrilepiURLEncoded(TK_ASCII urlencoded, T_UINT koliko, T_KODNASTRAN kodna_stran)
{
    assert(urlencoded || !koliko);

    T_UINT i, koliko2 = astrnlen(urlencoded, koliko, kodna_stran);
    RNizA niz(koliko2);

    // Najprej izloimo znake %.
    for (i = 0; i < koliko;) {
        if (urlencoded[i] == '%') {
            // Nali smo simbol za poseben znak.
            T_AZNAK z = 0;

            i++;
            // Analiziramo prvi znak.
            if (i < koliko) {
                     if ('0' <= urlencoded[i] && urlencoded[i] <= '9') z = urlencoded[i++] - '0';
                else if ('a' <= urlencoded[i] && urlencoded[i] <= 'f') z = urlencoded[i++] - 'a' + 10;
                else if ('A' <= urlencoded[i] && urlencoded[i] <= 'F') z = urlencoded[i++] - 'A' + 10;
                else {
                    // Prvi znak ni bila esnajstika tevka.
                    niz += '%';
                    continue;
                }
            }

            // Analiziramo drugi znak.
            if (i < koliko) {
                     if ('0' <= urlencoded[i] && urlencoded[i] <= '9') z = z * 16 + urlencoded[i++] - '0';
                else if ('a' <= urlencoded[i] && urlencoded[i] <= 'f') z = z * 16 + urlencoded[i++] - 'a' + 10;
                else if ('A' <= urlencoded[i] && urlencoded[i] <= 'F') z = z * 16 + urlencoded[i++] - 'A' + 10;
            }

            niz += z;
        } else if (urlencoded[i] == '+') {
            // Nali smo presledek zapisan s plusom.
            niz += ' ';
            i++;
        } else {
            // Druge znake prepiemo.
            niz += urlencoded[i];
            i++;
        }
    }

    // Prilepimo dekodiran niz na konec naega niza.
    return PrilepiA(niz, koliko2, kodna_stran);
}


BSTR RNizA::VrniBSTR(T_UINT koliko, T_KODNASTRAN kodna_stran) const
{
    assert(podatki);

    koliko = amin(st_znakov, koliko);
    // Ugotoviti moramo, koliko znakov Unicode bo dejansko potrebnih za niz BSTR.
    // Kadar je izvorni niz v zapisu UTF-8, je lahko njegova dolina dalja od tevila znakov.
    // ActiveX pa ne mara, da se dolini strlen(bstr) in SysStringLen(bstr) razlikujeta.
    T_UINT koliko2 = astrnlen(podatki, koliko, kodna_stran);
    BSTR bstr = ::SysAllocStringLen(NULL, (UINT)koliko2);

    if (bstr) {
        verify(astrncpy2(bstr, koliko2, podatki, koliko, kodna_stran) == koliko2);
        bstr[koliko2] = 0;
    }

    return bstr;
}

BSTR RNizU::VrniBSTR(T_UINT koliko) const
{
    assert(podatki);

    koliko = amin(st_znakov, koliko);
    return ::SysAllocStringLen(podatki, (UINT)koliko);
}

#ifndef AOSN_BREZ_SGML

#define KODA_MAX (sizeof(T_UZNAK)*2 + 1)

RNizA RNizA::VrniSGML(TK_ASCII niz, T_UINT koliko, T_UINT kateri, T_KODNASTRAN kodna_stran)
{
    assert(niz || !koliko);

    T_UINT i, n;
    T_AZNAK koda[KODA_MAX];
    T_UZNAK znak;
    TK_ASCII entiteta;
    RNizA sgml;
    T_BOOL
        pusti_ascii        = (kateri & AOSN_SGML_POLNI       ) == 0,
        pusti_narekovaj    = (kateri & AOSN_SGML_NAREKOVAJ   ) == 0,
        pusti_apostrof     = (kateri & AOSN_SGML_APOSTROF    ) == 0,
        pusti_manjse_vecje = (kateri & AOSN_SGML_MANJSE_VECJE) == 0,
        pusti_posevnico    = (kateri & AOSN_SGML_POSEVNICA   ) == 0,
        pusti_dolar        = (kateri & AOSN_SGML_DOLAR       ) == 0,
        pusti_odstotek     = (kateri & AOSN_SGML_ODSTOTEK    ) == 0,
        pusti_ajko         = (kateri & AOSN_SGML_AJKA        ) == 0,
        pusti_lojtro       = (kateri & AOSN_SGML_LOJTRA      ) == 0,
        pusti_oklepaja     = (kateri & AOSN_SGML_OKLEPAJA    ) == 0,
        pusti_zav_oklepaja = (kateri & AOSN_SGML_ZAV_OKLEPAJA) == 0,
        pusti_ogl_oklepaja = (kateri & AOSN_SGML_OGL_OKLEPAJA) == 0;

    for (i = 0; i < koliko && niz[i];) {
        n = achrlen(niz + i, koliko - i, kodna_stran);
        if (n == 1 &&
            pusti_ascii && (T_U2)niz[i] < 128 &&
                                    niz[i] != '&'                     &&
            (pusti_narekovaj    || (niz[i] != '"'                  )) &&
            (pusti_apostrof     || (niz[i] != '\''                 )) &&
            (pusti_manjse_vecje || (niz[i] != '<'  && niz[i] != '>')) &&
            (pusti_posevnico    || (niz[i] != '\\'                 )) &&
            (pusti_dolar        || (niz[i] != '$'                  )) &&
            (pusti_odstotek     || (niz[i] != '%'                  )) &&
            (pusti_ajko         || (niz[i] != '@'                  )) &&
            (pusti_lojtro       || (niz[i] != '#'                  )) &&
            (pusti_oklepaja     || (niz[i] != '('  && niz[i] != ')')) &&
            (pusti_zav_oklepaja || (niz[i] != '{'  && niz[i] != '}')) &&
            (pusti_ogl_oklepaja || (niz[i] != '['  && niz[i] != ']')))
        {
            // Znak je 7-bitni ASCII in nimamo posebne elje, da bi ga pretvorili v entiteto. Zato ga le prepiemo.
            sgml += (T_AZNAK)niz[i++];
        } else {
            znak = aasctouni(niz + i, n, kodna_stran);
            entiteta = achrtosgml(&znak, 1);
            if (entiteta) {
                // Celoten znak lahko zapiemo z eno entiteto.
                sgml += '&';
                sgml += entiteta;
                sgml += ';';
                i += n;
            } else if (n == 1) {
                // Znak je trivialen (doline ena), entitete SGML zanj ne poznamo.
                if ((T_U2)niz[i] < 128) {
                    // Znak je 7-bitni ASCII. Zato ga prepiemo.
                    sgml += (T_AZNAK)niz[i++];
                } else {
                    // Zapiemo numerino.
                    sgml += "&#x";
#if _MSC_VER >= 1500
                    _itoa_s(znak, koda, KODA_MAX, 16);
#else
                    _itoa(znak, koda, 16);
#endif
                    sgml.Prilepi(koda, KODA_MAX);
                    sgml += ';';
                    i++;
                }
            } else {
                assert(0); // TODO: Preizkusi ta del kode!
                // Znak je sestavljen, zato ga pretvorimo po delih.
                T_UINT konec = i + n;
                while (i < konec) {
                    znak = aasctouni(niz + i, 1, kodna_stran);
                    entiteta = achrtosgml(&znak, 1);
                    if (entiteta) {
                        // Zapiemo z entiteto.
                        sgml += '&';
                        sgml += entiteta;
                        sgml += ';';
                        i++;
                    } else if ((T_U2)niz[i] < 128) {
                        // Prepiemo.
                        sgml += (T_AZNAK)niz[i++];
                    } else {
                        // Zapiemo numerino.
                        sgml += "&#x";
#if _MSC_VER >= 1500
                        _itoa_s(znak, koda, KODA_MAX, 16);
#else
                        _itoa(znak, koda, 16);
#endif
                        sgml.Prilepi(koda, KODA_MAX);
                        sgml += ';';
                        i++;
                    }
                }
            }
        }
    }

    return sgml;
}


RNizA RNizU::VrniSGML(TK_UNICODE niz, T_UINT koliko, T_UINT kateri)
{
    assert(niz || !koliko);

    T_UINT i, n;
    T_AZNAK koda[KODA_MAX];
    TK_ASCII entiteta;
    RNizA sgml;
    T_BOOL
        pusti_ascii        = (kateri & AOSN_SGML_POLNI       ) == 0,
        pusti_narekovaj    = (kateri & AOSN_SGML_NAREKOVAJ   ) == 0,
        pusti_apostrof     = (kateri & AOSN_SGML_APOSTROF    ) == 0,
        pusti_manjse_vecje = (kateri & AOSN_SGML_MANJSE_VECJE) == 0,
        pusti_posevnico    = (kateri & AOSN_SGML_POSEVNICA   ) == 0,
        pusti_dolar        = (kateri & AOSN_SGML_DOLAR       ) == 0,
        pusti_odstotek     = (kateri & AOSN_SGML_ODSTOTEK    ) == 0,
        pusti_ajko         = (kateri & AOSN_SGML_AJKA        ) == 0,
        pusti_lojtro       = (kateri & AOSN_SGML_LOJTRA      ) == 0,
        pusti_oklepaja     = (kateri & AOSN_SGML_OKLEPAJA    ) == 0,
        pusti_zav_oklepaja = (kateri & AOSN_SGML_ZAV_OKLEPAJA) == 0,
        pusti_ogl_oklepaja = (kateri & AOSN_SGML_OGL_OKLEPAJA) == 0;

    for (i = 0; i < koliko && niz[i];) {
        n = achrlen(niz + i, koliko - i);
        if (n == 1 &&
            pusti_ascii && (T_U2)niz[i] < 128 &&
                                    niz[i] != L'&'                      &&
            (pusti_narekovaj    || (niz[i] != L'"'                   )) &&
            (pusti_apostrof     || (niz[i] != L'\''                  )) &&
            (pusti_manjse_vecje || (niz[i] != L'<'  && niz[i] != L'>')) &&
            (pusti_posevnico    || (niz[i] != L'\\'                  )) &&
            (pusti_dolar        || (niz[i] != L'$'                   )) &&
            (pusti_odstotek     || (niz[i] != L'%'                   )) &&
            (pusti_ajko         || (niz[i] != L'@'                   )) &&
            (pusti_lojtro       || (niz[i] != L'#'                   )) &&
            (pusti_oklepaja     || (niz[i] != L'('  && niz[i] != L')')) &&
            (pusti_zav_oklepaja || (niz[i] != L'{'  && niz[i] != L'}')) &&
            (pusti_ogl_oklepaja || (niz[i] != L'['  && niz[i] != L']')))
        {
            // Znak je 7-bitni ASCII in nimamo posebne elje, da bi ga pretvorili v entiteto. Zato ga le prepiemo.
            sgml += (T_AZNAK)niz[i++];
        } else {
            entiteta = achrtosgml(niz + i, n);
            if (entiteta) {
                // Celoten znak lahko zapiemo z eno entiteto.
                sgml += '&';
                sgml += entiteta;
                sgml += ';';
                i += n;
            } else if (n == 1) {
                // Znak je trivialen (doline ena), entitete SGML zanj ne poznamo.
                if ((T_U2)niz[i] < 128) {
                    // Znak je 7-bitni ASCII. Zato ga prepiemo.
                    sgml += (T_AZNAK)niz[i++];
                } else {
                    // Zapiemo numerino.
                    sgml += "&#x";
#if _MSC_VER >= 1500
                    _itoa_s(niz[i], koda, KODA_MAX, 16);
#else
                    _itoa(niz[i], koda, 16);
#endif
                    sgml.Prilepi(koda, KODA_MAX);
                    sgml += ';';
                    i++;
                }
            } else {
                // Znak je sestavljen, zato ga pretvorimo po delih.
                T_UINT konec = i + n;
                while (i < konec) {
                    entiteta = achrtosgml(niz + i, 1);
                    if (entiteta) {
                        // Zapiemo z entiteto.
                        sgml += '&';
                        sgml += entiteta;
                        sgml += ';';
                        i++;
                    } else if ((T_U2)niz[i] < 128) {
                        // Prepiemo.
                        sgml += (T_AZNAK)niz[i++];
                    } else {
                        // Zapiemo numerino.
                        sgml += "&#x";
#if _MSC_VER >= 1500
                        _itoa_s(niz[i], koda, KODA_MAX, 16);
#else
                        _itoa(niz[i], koda, 16);
#endif
                        sgml.Prilepi(koda, KODA_MAX);
                        sgml += ';';
                        i++;
                    }
                }
            }
        }
    }

    return sgml;
}

#undef KODA_MAX

#endif


RNizA RNizA::VrniUTF8(TK_ASCII niz, T_UINT koliko, T_KODNASTRAN kodna_stran)
{
    RNizU pom;
    pom.NastaviA(niz, koliko, kodna_stran);
    return pom.VrniUTF8();
}


RNizA RNizA::VrniHEX(TK_KAZALEC pom, T_SIZE dolzina)
{
    assert(pom || !dolzina);

    T_SIZE i;
    RNizA izhod;

    for (i = 0; i < dolzina; i++)
        izhod.PrilepiO("%02x", T_LOCALE_THREAD_DEFAULT, ((const T_U1*)pom)[i]);

    return izhod;
}


RNizA RNizU::VrniUTF8(TK_UNICODE niz, T_UINT koliko)
{
    koliko = astrnlen(niz, koliko);

    // Predpostavimo, da noben znak Unicode ne bo potreboval ve kot pet bytov za svoj zapis.
    T_UINT dolzina = koliko * 5;
    RNizA utf8(dolzina);

    utf8.st_znakov = astrncpy2(utf8.podatki, dolzina, niz, koliko, KODNASTRAN_UTF8);
    utf8.podatki[utf8.st_znakov] = 0;
    utf8.SprostiOdvecne();

    return utf8;
}
