/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_ODTISNIK


//////////////////////////////////////////////////////////////////////
// Makri za racunanje CRC, MD5 in SHA1
//////////////////////////////////////////////////////////////////////

#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

// Rotate x koncni to the left
#ifndef ROL32
#ifdef _MSC_VER
#define ROL32(_val32, _nBits) _rotl(_val32, _nBits)
#else
#define ROL32(_val32, _nBits) (((_val32)<<(_nBits))|((_val32)>>(32-(_nBits))))
#endif
#endif

#ifdef DEBELEJSI_KONEC_VER
#define SHA1BLK0(i) (delovni[i])
#else
#define SHA1BLK0(i) (delovni[i] = \
    (ROL32(delovni[i],24) & 0xFF00FF00) | (ROL32(delovni[i],8) & 0x00FF00FF))
#endif

#define SHA1BLK(i) (delovni[i&15] = ROL32(delovni[(i+13)&15] ^ delovni[(i+8)&15] \
    ^ delovni[(i+2)&15] ^ delovni[i&15],1))

// MD5 rounds
#define MD5_R1(a, b, c, d, i, s, ac) { (a) += (((b) & (c)) | ((~b) & (d))) + delovni[(i)] + (T_U4)(ac); (a) = ROL32((a), (s)); (a) += (b); }
#define MD5_R2(a, b, c, d, i, s, ac) { (a) += (((b) & (d)) | ((c) & (~d))) + delovni[(i)] + (T_U4)(ac); (a) = ROL32((a), (s)); (a) += (b); }
#define MD5_R3(a, b, c, d, i, s, ac) { (a) += ((b) ^ (c) ^ (d)) + delovni[(i)] + (T_U4)(ac); (a) = ROL32((a), (s)); (a) += (b); }
#define MD5_R4(a, b, c, d, i, s, ac) { (a) += ((c) ^ ((b) | (~d))) + delovni[(i)] + (T_U4)(ac); (a) = ROL32((a), (s)); (a) += (b); }

// SHA1 rounds
#define SHA1_R0(v, w, x, y, z, i) { (z) += (((w)&((x)^(y)))^(y))+SHA1BLK0((i))+0x5A827999+ROL32((v),5); (w)=ROL32((w),30); }
#define SHA1_R1(v, w, x, y, z, i) { (z) += (((w)&((x)^(y)))^(y))+SHA1BLK((i))+0x5A827999+ROL32((v),5); (w)=ROL32((w),30); }
#define SHA1_R2(v, w, x, y, z, i) { (z) += ((w)^(x)^(y))+SHA1BLK((i))+0x6ED9EBA1+ROL32((v),5); (w)=ROL32((w),30); }
#define SHA1_R3(v, w, x, y, z, i) { (z) += ((((w)|(x))&(y))|((w)&(x)))+SHA1BLK((i))+0x8F1BBCDC+ROL32((v),5); (w)=ROL32((w),30); }
#define SHA1_R4(v, w, x, y, z, i) { (z) += ((w)^(x)^(y))+SHA1BLK((i))+0xCA62C1D6+ROL32((v),5); (w)=ROL32((w),30); }



//////////////////////////////////////////////////////////////////////
// Lokalne spremenljivke
//////////////////////////////////////////////////////////////////////

const T_U4 tabela_crc32[256] = {
  0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
  0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
  0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
  0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
  0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
  0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
  0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
  0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
  0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
  0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
  0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
  0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
  0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
  0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
  0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
  0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
  0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
  0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
  0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
  0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
  0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
  0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
  0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
  0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
  0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
  0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
  0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
  0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
  0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
  0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
  0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
  0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
  0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
  0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
  0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
  0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
  0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
  0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
  0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
  0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
  0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
  0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
  0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
  0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
  0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
  0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
  0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
  0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
  0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
  0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
  0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
  0x2d02ef8dL
};


const T_U1 tabela_polnilo_md5[64] = {
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


//////////////////////////////////////////////////////////////////////
// ROdtisnikToka
//////////////////////////////////////////////////////////////////////

ROdtisnikToka::ROdtisnikToka(ROdtisnik &o, RTokPodatkov *izvor) :
    RPretvornikToka(izvor),
    odtisnik(o)
{
}


T_STANJE ROdtisnikToka::Uporabi(RTokPodatkov *izvor)
{
    odtisnik.Resetiraj();
    return RPretvornikToka::Uporabi(izvor);

}


RTokPodatkov* ROdtisnikToka::Odpusti()
{
    odtisnik.Zakljuci();
    return RPretvornikToka::Odpusti();
}


T_SIZE ROdtisnikToka::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);

    T_SIZE prebrano;

    // Najprej podatke preberemo.
    prebrano = RPretvornikToka::Preberi(podatki, dolzina);
    if (USPESNO(stanje))
        // Izraunamo kontrolno vsoto.
        odtisnik.Izracunaj(podatki, prebrano);

    return prebrano;
}


T_SIZE ROdtisnikToka::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);

    // Najprej izraunamo kontrolno vsoto.
    odtisnik.Izracunaj(podatki, dolzina);

    // Sedaj zapiemo podatke.
    return RPretvornikToka::Zapisi(podatki, dolzina);
}


//////////////////////////////////////////////////////////////////////
// RCRC32
//////////////////////////////////////////////////////////////////////

RCRC32::RCRC32(T_ODTIS_CRC32 crc) : ROdtisnik()
{
    this->crc = ~crc;
}


T_VOID RCRC32::Resetiraj()
{
    this->crc = 0xffffffffL;
}


T_VOID RCRC32::Izracunaj(TK_KAZALEC blok, T_SIZE dolzina)
{
    assert(blok || !dolzina);

    T_UINT i;

    for (i = 0; i < dolzina; i++)
        crc = ::tabela_crc32[(crc ^ ((T_U1*)blok)[i]) & 0xff] ^ (crc >> 8);
}


T_VOID RCRC32::Zakljuci()
{
    crc = ~crc;
}


T_STANJE RCRC32::Zapisi(RTokPodatkov &tok) const
{
    T_STANJE stanje;

    stanje = tok.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    // Resetiraj(); (resetiranje ni potrebno, saj je polje crc edino notranje stanje razreda RCRC32.
    tok << crc;

    return tok.VrniStanje();
}


T_STANJE RCRC32::Preberi(RTokPodatkov &tok)
{
    tok >> crc;
    return tok.VrniStanje();
}


T_SIZE RCRC32::VrniDolzinoOdtisa()
{
    return sizeof(T_ODTIS_CRC32);
}


const T_VOID* RCRC32::VrniOdtis()
{
    return (const T_VOID*)&crc;
}


//////////////////////////////////////////////////////////////////////
// RBlokovniOdtisnik
//////////////////////////////////////////////////////////////////////

T_VOID RBlokovniOdtisnik::Resetiraj()
{
  stevec[0] = stevec[1] = 0;
}


T_VOID RBlokovniOdtisnik::Izracunaj(TK_KAZALEC blok, T_SIZE dolzina)
{
    assert(blok || !dolzina);

    T_UINT i, j, ostanek;

    // Compute number of bytes mod 64
    j = (T_UINT)((stevec[0] >> 3) & 63);

    // Update number of koncni
    if ((stevec[0] += ((T_U4)dolzina << 3)) < ((T_U4)dolzina << 3))
        stevec[1]++;
    stevec[1] += ((T_U4)dolzina >> 29);

    ostanek = 64 - j;

    // Transform as many times as possible.
    if (dolzina >= ostanek) {
        assert(j < 64 && j + ostanek <= 64);
        assert(ostanek <= dolzina);
        memcpy(vrsta + j, blok, ostanek);
        IzracunajBlok();

        for (i = ostanek; i + 63 < dolzina; i += 64) {
            memcpy(vrsta, (const T_U1*)blok + i, 64);
            IzracunajBlok();
        }

        j = 0;
    } else
        i = 0;

    // Buffer remaining input
    assert(j < 64 && j + dolzina - i <= 64);
    assert(i <= dolzina);
    memcpy(vrsta + j, (const T_U1*)blok + i, dolzina - i);
}


//////////////////////////////////////////////////////////////////////
// RMD5
//////////////////////////////////////////////////////////////////////

RMD5::RMD5() : RBlokovniOdtisnik()
{
    Resetiraj();
}


T_VOID RMD5::Resetiraj()
{
    RBlokovniOdtisnik::Resetiraj();
    stanje[0] = 0x67452301;
    stanje[1] = 0xefcdab89;
    stanje[2] = 0x98badcfe;
    stanje[3] = 0x10325476;
}


T_VOID RMD5::Zakljuci()
{
    T_UINT index, ostanek;
    T_U1 koncni[8];

    // Save number of koncni
    memcpy(koncni, stevec, sizeof(stevec));

    // Pad out to 56 mod 64.
    index = (T_UINT)((stevec[0] >> 3) & 0x3f);
    ostanek = (index < 56) ? (56 - index) : (120 - index);
    Izracunaj(tabela_polnilo_md5, ostanek);

    // Append length (before padding)
    Izracunaj(koncni, 8);

    // Store stanje in digest
    memcpy(&odtis, stanje, sizeof(T_ODTIS_MD5));
}


T_STANJE RMD5::Zapisi(RTokPodatkov &tok) const
{
    T_STANJE stanje;
    stanje = tok.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    tok << odtis;

    return tok.VrniStanje();
}


T_STANJE RMD5::Preberi(RTokPodatkov &tok)
{
    Resetiraj();
    tok >> odtis;
    return tok.VrniStanje();
}


T_VOID RMD5::IzracunajBlok()
{
    // Copy stanje[] to working vars
    T_U4 a = stanje[0], b = stanje[1], c = stanje[2], d = stanje[3];

    // 4 rounds of 16 operations each. Loop unrolled.
    MD5_R1(a, b, c, d,  0, S11, 0xd76aa478);
    MD5_R1(d, a, b, c,  1, S12, 0xe8c7b756);
    MD5_R1(c, d, a, b,  2, S13, 0x242070db);
    MD5_R1(b, c, d, a,  3, S14, 0xc1bdceee);
    MD5_R1(a, b, c, d,  4, S11, 0xf57c0faf);
    MD5_R1(d, a, b, c,  5, S12, 0x4787c62a);
    MD5_R1(c, d, a, b,  6, S13, 0xa8304613);
    MD5_R1(b, c, d, a,  7, S14, 0xfd469501);
    MD5_R1(a, b, c, d,  8, S11, 0x698098d8);
    MD5_R1(d, a, b, c,  9, S12, 0x8b44f7af);
    MD5_R1(c, d, a, b, 10, S13, 0xffff5bb1);
    MD5_R1(b, c, d, a, 11, S14, 0x895cd7be);
    MD5_R1(a, b, c, d, 12, S11, 0x6b901122);
    MD5_R1(d, a, b, c, 13, S12, 0xfd987193);
    MD5_R1(c, d, a, b, 14, S13, 0xa679438e);
    MD5_R1(b, c, d, a, 15, S14, 0x49b40821);
    MD5_R2(a, b, c, d,  1, S21, 0xf61e2562);
    MD5_R2(d, a, b, c,  6, S22, 0xc040b340);
    MD5_R2(c, d, a, b, 11, S23, 0x265e5a51);
    MD5_R2(b, c, d, a,  0, S24, 0xe9b6c7aa);
    MD5_R2(a, b, c, d,  5, S21, 0xd62f105d);
    MD5_R2(d, a, b, c, 10, S22,  0x2441453);
    MD5_R2(c, d, a, b, 15, S23, 0xd8a1e681);
    MD5_R2(b, c, d, a,  4, S24, 0xe7d3fbc8);
    MD5_R2(a, b, c, d,  9, S21, 0x21e1cde6);
    MD5_R2(d, a, b, c, 14, S22, 0xc33707d6);
    MD5_R2(c, d, a, b,  3, S23, 0xf4d50d87);
    MD5_R2(b, c, d, a,  8, S24, 0x455a14ed);
    MD5_R2(a, b, c, d, 13, S21, 0xa9e3e905);
    MD5_R2(d, a, b, c,  2, S22, 0xfcefa3f8);
    MD5_R2(c, d, a, b,  7, S23, 0x676f02d9);
    MD5_R2(b, c, d, a, 12, S24, 0x8d2a4c8a);
    MD5_R3(a, b, c, d,  5, S31, 0xfffa3942);
    MD5_R3(d, a, b, c,  8, S32, 0x8771f681);
    MD5_R3(c, d, a, b, 11, S33, 0x6d9d6122);
    MD5_R3(b, c, d, a, 14, S34, 0xfde5380c);
    MD5_R3(a, b, c, d,  1, S31, 0xa4beea44);
    MD5_R3(d, a, b, c,  4, S32, 0x4bdecfa9);
    MD5_R3(c, d, a, b,  7, S33, 0xf6bb4b60);
    MD5_R3(b, c, d, a, 10, S34, 0xbebfbc70);
    MD5_R3(a, b, c, d, 13, S31, 0x289b7ec6);
    MD5_R3(d, a, b, c,  0, S32, 0xeaa127fa);
    MD5_R3(c, d, a, b,  3, S33, 0xd4ef3085);
    MD5_R3(b, c, d, a,  6, S34,  0x4881d05);
    MD5_R3(a, b, c, d,  9, S31, 0xd9d4d039);
    MD5_R3(d, a, b, c, 12, S32, 0xe6db99e5);
    MD5_R3(c, d, a, b, 15, S33, 0x1fa27cf8);
    MD5_R3(b, c, d, a,  2, S34, 0xc4ac5665);
    MD5_R4(a, b, c, d,  0, S41, 0xf4292244);
    MD5_R4(d, a, b, c,  7, S42, 0x432aff97);
    MD5_R4(c, d, a, b, 14, S43, 0xab9423a7);
    MD5_R4(b, c, d, a,  5, S44, 0xfc93a039);
    MD5_R4(a, b, c, d, 12, S41, 0x655b59c3);
    MD5_R4(d, a, b, c,  3, S42, 0x8f0ccc92);
    MD5_R4(c, d, a, b, 10, S43, 0xffeff47d);
    MD5_R4(b, c, d, a,  1, S44, 0x85845dd1);
    MD5_R4(a, b, c, d,  8, S41, 0x6fa87e4f);
    MD5_R4(d, a, b, c, 15, S42, 0xfe2ce6e0);
    MD5_R4(c, d, a, b,  6, S43, 0xa3014314);
    MD5_R4(b, c, d, a, 13, S44, 0x4e0811a1);
    MD5_R4(a, b, c, d,  4, S41, 0xf7537e82);
    MD5_R4(d, a, b, c, 11, S42, 0xbd3af235);
    MD5_R4(c, d, a, b,  2, S43, 0x2ad7d2bb);
    MD5_R4(b, c, d, a,  9, S44, 0xeb86d391);

    // Add the working vars back into stanje
    stanje[0] += a;
    stanje[1] += b;
    stanje[2] += c;
    stanje[3] += d;
}


T_SIZE RMD5::VrniDolzinoOdtisa()
{
    return sizeof(T_ODTIS_MD5);
}


const T_VOID* RMD5::VrniOdtis()
{
    return (const T_VOID*)&odtis;
}


//////////////////////////////////////////////////////////////////////
// RSHA1
//////////////////////////////////////////////////////////////////////

RSHA1::RSHA1() : RBlokovniOdtisnik()
{
    Resetiraj();
}


T_VOID RSHA1::Resetiraj()
{
    RBlokovniOdtisnik::Resetiraj();

    // SHA1 initialization constants
    stanje[0] = 0x67452301;
    stanje[1] = 0xEFCDAB89;
    stanje[2] = 0x98BADCFE;
    stanje[3] = 0x10325476;
    stanje[4] = 0xC3D2E1F0;
}


T_VOID RSHA1::Zakljuci()
{
    T_UINT i;
    T_U1 koncni[8];

    // Save number of koncni
    for (i = 0; i < 8; i++)
        koncni[i] = (T_U1)((stevec[((i >= 4) ? 0 : 1)] >> ((3 - (i & 3)) * 8) ) & 255); // Endian independent

    Izracunaj((T_U1 *)"\200", 1);

    while ((stevec[0] & 504) != 448)
        Izracunaj((T_U1 *)"\0", 1);

    Izracunaj(koncni, 8); // Cause a SHA1Transform()

    // Store stanje in digest
    for (i = 0; i < 20; i++)
        odtis.podatki[i] = (T_U1)((stanje[i >> 2] >> ((3 - (i & 3)) * 8) ) & 255);
}


T_STANJE RSHA1::Zapisi(RTokPodatkov &tok) const
{
    T_STANJE stanje;
    stanje = tok.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    tok << odtis;

    return tok.VrniStanje();
}


T_STANJE RSHA1::Preberi(RTokPodatkov &tok)
{
    Resetiraj();
    tok >> odtis;
    return tok.VrniStanje();
}


T_VOID RSHA1::IzracunajBlok()
{
    // Copy stanje[] to working vars
    T_U4 a = stanje[0], b = stanje[1], c = stanje[2], d = stanje[3], e = stanje[4];

    // 5 rounds of 16 operations each. Loop unrolled.
    SHA1_R0(a,b,c,d,e, 0); SHA1_R0(e,a,b,c,d, 1); SHA1_R0(d,e,a,b,c, 2); SHA1_R0(c,d,e,a,b, 3);
    SHA1_R0(b,c,d,e,a, 4); SHA1_R0(a,b,c,d,e, 5); SHA1_R0(e,a,b,c,d, 6); SHA1_R0(d,e,a,b,c, 7);
    SHA1_R0(c,d,e,a,b, 8); SHA1_R0(b,c,d,e,a, 9); SHA1_R0(a,b,c,d,e,10); SHA1_R0(e,a,b,c,d,11);
    SHA1_R0(d,e,a,b,c,12); SHA1_R0(c,d,e,a,b,13); SHA1_R0(b,c,d,e,a,14); SHA1_R0(a,b,c,d,e,15);
    SHA1_R1(e,a,b,c,d,16); SHA1_R1(d,e,a,b,c,17); SHA1_R1(c,d,e,a,b,18); SHA1_R1(b,c,d,e,a,19);
    SHA1_R2(a,b,c,d,e,20); SHA1_R2(e,a,b,c,d,21); SHA1_R2(d,e,a,b,c,22); SHA1_R2(c,d,e,a,b,23);
    SHA1_R2(b,c,d,e,a,24); SHA1_R2(a,b,c,d,e,25); SHA1_R2(e,a,b,c,d,26); SHA1_R2(d,e,a,b,c,27);
    SHA1_R2(c,d,e,a,b,28); SHA1_R2(b,c,d,e,a,29); SHA1_R2(a,b,c,d,e,30); SHA1_R2(e,a,b,c,d,31);
    SHA1_R2(d,e,a,b,c,32); SHA1_R2(c,d,e,a,b,33); SHA1_R2(b,c,d,e,a,34); SHA1_R2(a,b,c,d,e,35);
    SHA1_R2(e,a,b,c,d,36); SHA1_R2(d,e,a,b,c,37); SHA1_R2(c,d,e,a,b,38); SHA1_R2(b,c,d,e,a,39);
    SHA1_R3(a,b,c,d,e,40); SHA1_R3(e,a,b,c,d,41); SHA1_R3(d,e,a,b,c,42); SHA1_R3(c,d,e,a,b,43);
    SHA1_R3(b,c,d,e,a,44); SHA1_R3(a,b,c,d,e,45); SHA1_R3(e,a,b,c,d,46); SHA1_R3(d,e,a,b,c,47);
    SHA1_R3(c,d,e,a,b,48); SHA1_R3(b,c,d,e,a,49); SHA1_R3(a,b,c,d,e,50); SHA1_R3(e,a,b,c,d,51);
    SHA1_R3(d,e,a,b,c,52); SHA1_R3(c,d,e,a,b,53); SHA1_R3(b,c,d,e,a,54); SHA1_R3(a,b,c,d,e,55);
    SHA1_R3(e,a,b,c,d,56); SHA1_R3(d,e,a,b,c,57); SHA1_R3(c,d,e,a,b,58); SHA1_R3(b,c,d,e,a,59);
    SHA1_R4(a,b,c,d,e,60); SHA1_R4(e,a,b,c,d,61); SHA1_R4(d,e,a,b,c,62); SHA1_R4(c,d,e,a,b,63);
    SHA1_R4(b,c,d,e,a,64); SHA1_R4(a,b,c,d,e,65); SHA1_R4(e,a,b,c,d,66); SHA1_R4(d,e,a,b,c,67);
    SHA1_R4(c,d,e,a,b,68); SHA1_R4(b,c,d,e,a,69); SHA1_R4(a,b,c,d,e,70); SHA1_R4(e,a,b,c,d,71);
    SHA1_R4(d,e,a,b,c,72); SHA1_R4(c,d,e,a,b,73); SHA1_R4(b,c,d,e,a,74); SHA1_R4(a,b,c,d,e,75);
    SHA1_R4(e,a,b,c,d,76); SHA1_R4(d,e,a,b,c,77); SHA1_R4(c,d,e,a,b,78); SHA1_R4(b,c,d,e,a,79);

    // Add the working vars back into stanje
    stanje[0] += a;
    stanje[1] += b;
    stanje[2] += c;
    stanje[3] += d;
    stanje[4] += e;
}


T_SIZE RSHA1::VrniDolzinoOdtisa()
{
    return sizeof(T_ODTIS_SHA1);
}


const T_VOID* RSHA1::VrniOdtis()
{
    return (const T_VOID*)&odtis;
}

#endif // AOSN_ODTISNIK
