/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_REGIZR


//////////////////////////////////////////////////////////////////////
// Deklaracije funkcij inline
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML
inline TK_UNICODE _VrniSGMLZnakLevo(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT &zacetek_znaka, T_UZNAK (&pom)[2]);
inline TK_UNICODE _VrniSGMLZnakDesno(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT &konec_znaka, T_UZNAK (&pom)[2]);
#endif


//////////////////////////////////////////////////////////////////////
// RRegElPoljubenZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElPoljubenZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        if (besedilo[zacetek] == '&') {
            // Znak je zapisan kot entiteta SGML.

            // Preskoimo vse znake do zakljuka entitete.
            for (interval.konec = zacetek + 1; interval.konec < konec && besedilo[interval.konec]; interval.konec++)
                if (besedilo[interval.konec] == ';') {
                    interval.konec++;
                    interval.zacetek = zacetek;
                    return TRUE;
                } else if (besedilo[interval.konec] == '&' || aisspace(besedilo[interval.konec]))
                    break;

            // Entiteta ni zakljuena.
        }

        // Znak je trivialen.
        interval.konec = (interval.zacetek = zacetek) + 1;
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

RRegElZnakSGML::RRegElZnakSGML(TK_ASCII _znak, T_UINT dolzina, T_BOOL _je_negirano) :
    RRegElementP<T_AZNAK>(),
    je_negirano(_je_negirano)
{
    assert(_znak || !dolzina);

    T_UZNAK pom[2];
    T_UINT konec_znaka;

    znak.Nastavi(dolzina ? _VrniSGMLZnakDesno(_znak, 0, dolzina, konec_znaka, pom) : L"");
}


T_BOOL RRegElZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE _znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        T_BOOL r = (zastavice & RREGIZR_NE_LOCUJ_VELIKIH_MALIH) ?
            astrnicmp(_znak, T_UINT_MAX, znak.Vrni(), znak.VrniDolzino()) == 0 :
             astrncmp(_znak, T_UINT_MAX, znak.Vrni(), znak.VrniDolzino()) == 0;

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElAlfanumericniZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElAlfanumericniZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        T_BOOL r = aisalnum(znak[0]);

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElPresledekSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElPresledekSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        T_BOOL r = (zastavice & RREGIZR_VECVRSTICNO) ?
                                   aisspace(znak[0]) :
            !aislbreak(znak[0]) && aisspace(znak[0]);

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElLociloSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElLociloSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        T_BOOL r = aispunct(znak[0]);

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElPresledekAliLociloSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElPresledekAliLociloSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        T_BOOL r = (zastavice & RREGIZR_VECVRSTICNO) ?
                                    aisspace(znak[0]) || aispunct(znak[0]) :
            !aislbreak(znak[0]) && (aisspace(znak[0]) || aispunct(znak[0]));

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElZacetekBesedeSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElZacetekBesedeSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT _zastavice)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(_zastavice);

    T_UZNAK
        pom_l[2],
        pom_d[2];
    TK_UNICODE
        znak_l = zacetek                              ? _VrniSGMLZnakLevo (besedilo, zacetek, konec, interval.zacetek, pom_l) : NULL,
        znak_d = zacetek < konec && besedilo[zacetek] ? _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec,   pom_d) : NULL;
    T_BOOL r =
         (!znak_l ||
          (zastavice & RREGIZR_LOCILO_BESED_PRESLEDKI) && aisspace(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_PIKE)      && aispunct(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_STEVKE)    && aisdigit(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_SIMBOLI)   && aisgraph(znak_l[0]) && !aisalnum(znak_l[0])) &&
        !(!znak_d ||
          (zastavice & RREGIZR_LOCILO_BESED_PRESLEDKI) && aisspace(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_PIKE)      && aispunct(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_STEVKE)    && aisdigit(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_SIMBOLI)   && aisgraph(znak_d[0]) && !aisalnum(znak_d[0]));

    if (r && !je_negirano || !r && je_negirano) {
        // Prejnji znak je bil presledek, loilo, tevka ipd., trenutni pa ni od tega.
        interval.konec = (interval.zacetek = zacetek);
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElKonecBesedeSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElKonecBesedeSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT _zastavice)
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(_zastavice);

    T_UZNAK
        pom_l[2],
        pom_d[2];
    TK_UNICODE
        znak_l = zacetek                              ? _VrniSGMLZnakLevo (besedilo, zacetek, konec, interval.zacetek, pom_l) : 0,
        znak_d = zacetek < konec && besedilo[zacetek] ? _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec,   pom_d) : 0;
    T_BOOL r =
        !(!znak_l ||
          (zastavice & RREGIZR_LOCILO_BESED_PRESLEDKI) && aisspace(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_PIKE)      && aispunct(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_STEVKE)    && aisdigit(znak_l[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_SIMBOLI)   && aisgraph(znak_l[0]) && !aisalnum(znak_l[0])) &&
         (!znak_d ||
          (zastavice & RREGIZR_LOCILO_BESED_PRESLEDKI) && aisspace(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_PIKE)      && aispunct(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_STEVKE)    && aisdigit(znak_d[0]) ||
          (zastavice & RREGIZR_LOCILO_BESED_SIMBOLI)   && aisgraph(znak_d[0]) && !aisalnum(znak_d[0]));

    if (r && !je_negirano || !r && je_negirano) {
        // Trenutni znak je presledek, loilo, tevka ipd., prejnji pa ni od tega.
        interval.konec = (interval.zacetek = zacetek);
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElMaskaSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

RRegElMaskaSGML::RRegElMaskaSGML(TK_ASCII _maska, T_UINT dolzina, T_BOOL je_negirano) : RRegElMaskaOsnP<T_AZNAK>(je_negirano)
{
    if (_maska)
        maska.NastaviSGML(_maska, dolzina);
}


T_BOOL RRegElMaskaSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE
            znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom),
            _maska = maska.Vrni(),
            r = (zastavice & RREGIZR_NE_LOCUJ_VELIKIH_MALIH) ?
                astrnistr(_maska, znak, maska.VrniDolzino()) :
                astrnstr (_maska, znak, maska.VrniDolzino());
        odmik_zadetka = r ? r - _maska : T_UINT_MAX;

        if (r && !je_negirano || !r && je_negirano) {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    odmik_zadetka = T_UINT_MAX;
    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElNizSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

RRegElNizSGML::RRegElNizSGML(TK_ASCII _niz, T_UINT dolzina) : RRegElementP<T_AZNAK>()
{
    niz.NastaviSGML(_niz, dolzina);
}


T_BOOL RRegElNizSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);

    T_UINT i, j, n = niz.VrniDolzino();
    T_BOOL r;
    TK_UNICODE _niz = niz.Vrni();
    T_UZNAK pom[2];
    TK_UNICODE znak;
    T_BOOL ne_locuj_velikih_malih = zastavice & RREGIZR_NE_LOCUJ_VELIKIH_MALIH ? TRUE : FALSE;
    T_UZNAK znak1, znak2;

    for (i = 0, interval.konec = zacetek, r = TRUE; i < n && interval.konec < konec && besedilo[interval.konec];) {
        znak = _VrniSGMLZnakDesno(besedilo, interval.konec, konec, interval.konec, pom);

        // Primerjamo.
        for (j = 0; i < n && znak[j]; i++, j++) {
            if (ne_locuj_velikih_malih) {
                znak1 = atolower(_niz[i]);
                znak2 = atolower(znak[j]);
            } else {
                znak1 = _niz[i];
                znak2 = znak[j];
            }

            if (znak1 != znak2) {
                r = FALSE;
                break;
            }
        }
        if (!r) break;
    }

    if (r && i >= n) {
        interval.zacetek = zacetek;
        return TRUE;
    } else {
        interval.zacetek = (interval.konec = zacetek) + 1;
        return FALSE;
    }
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElVejitevNizovSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

RRegElVejitevNizovSGML::RRegElVejitevNizovSGML(TK_ASCII niziz, T_UINT dolzina) : RRegElVejitevP<T_AZNAK>()
{
    ZgradiTabelo(niziz, dolzina);
}


RRegElVejitevNizovSGML::RRegElVejitevNizovSGML(TK_ASCII niz, ...) : RRegElVejitevP<T_AZNAK>()
{
    va_list parametri;
    va_start(parametri, niz);
    ZgradiTabelo(niz, parametri);
    va_end(parametri);
}


RRegElVejitevNizovSGML::~RRegElVejitevNizovSGML()
{
    T_UINT i, stevilo = tab.VrniStevilo();

    // Nize smo alocirali mi v metodi ZgradiTabelo(), zato jih moramo tudi uniiti.
    for (i = 0; i < stevilo; i++) {
        RRegElNizSGML *el = (RRegElNizSGML*)tab[i];
        assert(el);
        delete (RRegElNizSGML*)el;
    }
}


T_VOID RRegElVejitevNizovSGML::ZgradiTabelo(TK_ASCII niziz, T_UINT dolzina)
{
    assert(niziz || !dolzina);

    if (dolzina) {
        T_UINT i, zacetek, stevilo;
        RRegElementP<T_AZNAK> **p;

        // Najprej pretejemo tevilo nizov.
        for (zacetek = stevilo = 0; zacetek < dolzina && niziz[zacetek]; zacetek += astrnlen(niziz + zacetek, dolzina - zacetek) + 1, stevilo++);

        // Ustvarimo tabelo in naredimo nize.
        p = tab.Odpusti(stevilo, FALSE);
        assert(p || !stevilo);
        for (zacetek = i = 0; i < stevilo; zacetek += astrnlen(niziz + zacetek, dolzina - zacetek) + 1, i++) {
            // Vzpostavimo nize z alternativami.
            p[i] = new RRegElNizSGML(niziz + zacetek, dolzina - zacetek);
        }
        tab.Uporabi(p, stevilo, FALSE);
    }
}


T_VOID RRegElVejitevNizovSGML::ZgradiTabelo(TK_ASCII niz, va_list parametri)
{
    assert(tab.JePrazna());

    TK_ASCII p;

    // Nize prepiemo v tabelo.
    for (tab += new RRegElNizSGML(niz); (p = va_arg(parametri, TK_ASCII)) != NULL; tab += new RRegElNizSGML(p));
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElDNSNaslovZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

RRegElDNSNaslovZnakSGML::RRegElDNSNaslovZnakSGML(T_BOOL je_lahko_IDN) : RRegElDNSNaslovZnakP<T_AZNAK>(je_lahko_IDN)
{
}


T_BOOL RRegElDNSNaslovZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);

        if ((('A' <= znak[0] && znak[0] <= 'Z') ||
             ('a' <= znak[0] && znak[0] <= 'z') ||
             ('0' <= znak[0] && znak[0] <= '9')) && znak[1] == 0)
        {
            je_lahko_robni = TRUE;
        } else if (znak[0] == '-' && znak[1] == 0) {
            je_lahko_robni = FALSE;
        } else if (je_lahko_IDN && aisalnum(znak[0])) {
            je_lahko_robni = TRUE;
        } else {
            interval.zacetek = (interval.konec = zacetek) + 1;
            return FALSE;
        }

        interval.zacetek = zacetek;
        return TRUE;
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElURLUporabnikZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElURLUporabnikZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        if ((znak[0] == L'-'  ||
             znak[0] == L'.'  ||
             znak[0] == L'_'  ||
             znak[0] == L'~'  ||
             znak[0] == L'%'  ||
             znak[0] == L'!'  ||
             znak[0] == L'$'  ||
             znak[0] == L'&'  ||
             znak[0] == L'\'' ||
             znak[0] == L'('  ||
             znak[0] == L')'  ||
             znak[0] == L'*'  ||
             znak[0] == L'+'  ||
             znak[0] == L','  ||
             znak[0] == L';'  ||
             znak[0] == L'=') && znak[1] == 0 ||
            aisalnum(znak[0]))
        {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElURLGesloZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElURLGesloZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        if ((znak[0] == L'-'  ||
             znak[0] == L'.'  ||
             znak[0] == L'_'  ||
             znak[0] == L'~'  ||
             znak[0] == L'%'  ||
             znak[0] == L'!'  ||
             znak[0] == L'$'  ||
             znak[0] == L'&'  ||
             znak[0] == L'\'' ||
             znak[0] == L'('  ||
             znak[0] == L')'  ||
             znak[0] == L'*'  ||
             znak[0] == L'+'  ||
             znak[0] == L','  ||
             znak[0] == L';'  ||
             znak[0] == L'='  ||
             znak[0] == L':') && znak[1] == 0 ||
            aisalnum(znak[0]))
        {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// RRegElURLPotZnakSGML
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

T_BOOL RRegElURLPotZnakSGML::PreveriUjemanje(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT zastavice)
{
    assert(besedilo || !konec);
    UNREFERENCED_PARAMETER(zastavice);

    if (zacetek < konec && besedilo[zacetek]) {
        T_UZNAK pom[2];
        TK_UNICODE znak = _VrniSGMLZnakDesno(besedilo, zacetek, konec, interval.konec, pom);
        if ((znak[0] == L'/'  ||
             znak[0] == L'-'  ||
             znak[0] == L'.'  ||
             znak[0] == L'_'  ||
             znak[0] == L'~'  ||
             znak[0] == L'%'  ||
             znak[0] == L'!'  ||
             znak[0] == L'$'  ||
             znak[0] == L'&'  ||
             znak[0] == L'\'' ||
             znak[0] == L'('  ||
             znak[0] == L')'  ||
             znak[0] == L'*'  ||
             znak[0] == L'+'  ||
             znak[0] == L','  ||
             znak[0] == L';'  ||
             znak[0] == L'='  ||
             znak[0] == L':'  ||
             znak[0] == L'@'  ||
             znak[0] == L'?'  ||
             znak[0] == L'#') && znak[1] == 0 ||
            aisalnum(znak[0]))
        {
            interval.zacetek = zacetek;
            return TRUE;
        }
    }

    interval.zacetek = (interval.konec = zacetek) + 1;
    return FALSE;
}

#endif


//////////////////////////////////////////////////////////////////////
// Funkcije inline
//////////////////////////////////////////////////////////////////////

#ifndef AOSN_BREZ_SGML

inline TK_UNICODE _VrniSGMLZnakLevo(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT &zacetek_znaka, T_UZNAK (&pom)[2])
{
    assert(0); // TODO: Preizkusi ta del kode.
    assert(besedilo || !konec);
    assert(zacetek);
    UNREFERENCED_PARAMETER(konec);

    TK_UNICODE entiteta_uni;

    if (besedilo[zacetek - 1] == ';') {
        // Nali smo moen konec entitete.
        for (zacetek_znaka = zacetek - 1, entiteta_uni = 0; zacetek_znaka--;) {
            if (besedilo[zacetek_znaka] == '&') {
                // Nali smo zaetek entitete.
                T_UINT n = zacetek - zacetek_znaka - 1;

                if (n >= 2 && besedilo[zacetek_znaka + 1] == '#') {
                    // Gre za numerino entiteto.
                    if (besedilo[zacetek_znaka + 2] == 'x' || besedilo[zacetek_znaka + 2] == 'X')
                        pom[0] = (T_UZNAK)astrtoui(besedilo + zacetek_znaka + 3, n - 2, NULL, 16);
                    else
                        pom[0] = (T_UZNAK)astrtoui(besedilo + zacetek_znaka + 2, n - 1, NULL, 10);
                    pom[1] = 0;
                    entiteta_uni = pom;
                } else {
                    // Poiemo entiteto v tabeli entitet.
                    entiteta_uni = asgmltouni(besedilo + zacetek_znaka + 1, n);
                }

                zacetek_znaka--;
                break;
            } else if (besedilo[zacetek_znaka] == ';' || aisspace(besedilo[zacetek_znaka])) {
                // Nali smo znak, ki ne more biti del entitete.
                break;
            }
        }

        if (!entiteta_uni) {
            // Entiteta ni imela uvoda, ali pa je bila neznana.
            pom[0] = ';';
            pom[1] = 0;
            entiteta_uni = pom;
            zacetek_znaka = zacetek - 1;
        }
    } else {
        pom[0] = besedilo[zacetek - 1];
        pom[1] = 0;
        entiteta_uni = pom;

        assert(0); // TODO: Preveri, ali je prav, da sem dodal e spodnjo vrstico.
        zacetek_znaka = zacetek - 1;
    }

    return entiteta_uni;
}


inline TK_UNICODE _VrniSGMLZnakDesno(const T_AZNAK *besedilo, T_UINT zacetek, T_UINT konec, T_UINT &konec_znaka, T_UZNAK (&pom)[2])
{
    assert(besedilo || !konec);
    assert(zacetek < konec && besedilo[zacetek]);

    TK_UNICODE entiteta_uni;

    if (besedilo[zacetek] == '&') {
        // Nali smo moen zaetek entitete.
        for (konec_znaka = zacetek + 1, entiteta_uni = 0; konec_znaka < konec && besedilo[konec_znaka]; konec_znaka++) {
            if (besedilo[konec_znaka] == ';') {
                // Nali smo konec entitete.
                T_UINT n = konec_znaka - zacetek - 1;

                if (n >= 2 && besedilo[zacetek + 1] == '#') {
                    // Gre za numerino entiteto.
                    if (besedilo[zacetek + 2] == 'x' || besedilo[zacetek + 2] == 'X')
                        pom[0] = (T_UZNAK)astrtoui(besedilo + zacetek + 3, n - 2, NULL, 16);
                    else
                        pom[0] = (T_UZNAK)astrtoui(besedilo + zacetek + 2, n - 1, NULL, 10);
                    pom[1] = 0;
                    entiteta_uni = pom;
                } else {
                    // Poiemo entiteto v tabeli entitet.
                    entiteta_uni = asgmltouni(besedilo + zacetek + 1, n);
                }

                konec_znaka++;
                break;
            } else if (besedilo[konec_znaka] == '&' || aisspace(besedilo[konec_znaka])) {
                // Nali smo znak, ki ne more biti del entitete.
                break;
            }
        }

        if (!entiteta_uni) {
            // Entiteta ni bila zakljuena, ali pa je bila neznana.
            pom[0] = '&';
            pom[1] = 0;
            entiteta_uni = pom;
            konec_znaka = zacetek + 1;
        }
    } else {
        pom[0] = besedilo[zacetek];
        pom[1] = 0;
        entiteta_uni = pom;
        konec_znaka = zacetek + 1;
    }

    return entiteta_uni;
}

#endif

#endif // AOSN_REGIZR
