/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_REGISTER

#pragma comment(lib, "advapi32.lib")


#define RREGISTER_KORAK_POVECEVANJA 0x1000


RRegister::RRegister(T_ROCICAOS r) :
    RSisElement(r),
    RIzpElement()
{
}


T_STANJE RRegister::Odpri(T_ROCICAOS rocica_korena, TK_NIZ kljuc, T_UINT nacin)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegOpenKeyEx((HKEY)rocica_korena, kljuc, 0, (REGSAM)nacin, (PHKEY)&rocica);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_ODPIRANJE);
}


T_STANJE RRegister::Odpri(T_ROCICAOS rocica_korena, TK_NIZ kljuc, T_UINT nacin, TK_NIZ vrednost)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    // Statine tabele konstant
    static const HKEY _rocice[] = {
        HKEY_CURRENT_USER,
        HKEY_LOCAL_MACHINE,
    };
    static const REGSAM _prazna_zastavica = 0;
#ifdef _WIN64
    static const REGSAM _dodatne_zastavice_x64[] = {
        0,
        RREG_UPORABI_32,
    };
#else
    static const REGSAM _dodatne_zastavice_wow64[] = {
        RREG_UPORABI_64,
        RREG_UPORABI_32,
    };
#endif

    LONG rezultat;
    const HKEY *rocice;
    T_UINT idx_rocica, st_rocic;
    const REGSAM *dodatne_zastavice;
    T_UINT idx_dodatne_zastavice, st_dodatnih_zastavic;

    if (rocica_korena) {
        // Uporabnik je specifino navedel katero vejo v registru eli odpreti.
        rocice   = (const HKEY*)&rocica_korena;
        st_rocic = 1;
    } else {
        // Uporabnik ni navedel veje, zato poskusimo najprej v uporabnikovi, nato pa e raunalnikovi veji.
        rocice   = _rocice;
        st_rocic = _countof(_rocice);
    }

    if ((nacin & RREG_UPORABI_64IN32) != RREG_UPORABI_64IN32) {
        // Uporabnik je specifino navedel zastavice, s katerimi eli odpreti klju.
        dodatne_zastavice     = &_prazna_zastavica;
        st_dodatnih_zastavic  = 1;
    } else {
        // Uporabnik je navedel obe zastavici (RREG_UPORABI_64IN32), zato poskusimo na dva naina.
#ifdef _WIN64
        dodatne_zastavice    = _dodatne_zastavice_x64;
        st_dodatnih_zastavic = _countof(_dodatne_zastavice_x64);
#else
        T_BOOL je_proces_wow64 = AOsn::JeProcesWOW64();
        dodatne_zastavice      = je_proces_wow64 ?          _dodatne_zastavice_wow64  : &_prazna_zastavica;
        st_dodatnih_zastavic   = je_proces_wow64 ? _countof(_dodatne_zastavice_wow64) : 1;
#endif

        // Odstranimo zastavici RREG_UPORABI_32 in RREG_UPORABI_64, ker bomo sami upravljali z njima.
        nacin &= ~RREG_UPORABI_64IN32;
    }

    for (idx_rocica = 0; idx_rocica < st_rocic; idx_rocica++) {
        for (idx_dodatne_zastavice = 0; idx_dodatne_zastavice < st_dodatnih_zastavic; idx_dodatne_zastavice++) {
            // Poskusimo odpreti klju.
            rezultat = ::RegOpenKeyEx(rocice[idx_rocica], kljuc, 0, (REGSAM)nacin | dodatne_zastavice[idx_dodatne_zastavice], (PHKEY)&rocica);
            if (rezultat == ERROR_SUCCESS) {
                // Klju smo uspeno odprli.

                if (vrednost) {
                    // Uporabnika zanima tudi, ali je v kljuu prisotna doloena vrednost.
                    DWORD tip, dolzina;
                    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &tip, NULL, &dolzina);
                    if (rezultat == ERROR_SUCCESS) {
                        // Vrednost obstaja.
                        return A_OK;
                    }
                } else {
                    // Klju obstaja in to je za uporabnika dovolj.
                    return A_OK;
                }

                // Klju zapremo, ker uporabnika ne zanima.
                ::RegCloseKey((HKEY)rocica);
            }
        }
    }

    rocica = NULL;
    return AOSN_NAPAKA_NI_NAJDENO;
}


T_STANJE RRegister::Ustvari(T_ROCICAOS rocica_korena, TK_NIZ kljuc, T_UINT nacin)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    DWORD dispozicija;
    LONG rezultat;

    rezultat = ::RegCreateKeyEx((HKEY)rocica_korena, kljuc, 0, NULL, REG_OPTION_NON_VOLATILE, (REGSAM)nacin, NULL, (PHKEY)&rocica, &dispozicija);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_USTVARI);
}


T_STANJE RRegister::Zapri()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    T_STANJE stanje;

    if (rocica) {
        LONG rezultat;

        rezultat = ::RegCloseKey((HKEY)rocica);
        stanje = rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_ZAPIRANJE);
        rocica = NULL;
        return stanje;
    } else
        return AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RRegister::Izplakni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegFlushKey((HKEY)rocica);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_IZPLAKNI);
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, T_U4 &podatek)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD tip, dolzina;
    T_STANJE stanje;
    LONG rezultat;

    // Najprej ugotovimo tip in dolino vrednosti.
    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &tip, NULL, &dolzina);
    if (rezultat == ERROR_SUCCESS) {
        if (tip == REG_DWORD) {
            // Vrednost je res vrste REG_DWORD.
            assert(dolzina == sizeof(T_U4));
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&podatek, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _LE2HE(&podatek);
                stanje = A_OK;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_DWORD_BIG_ENDIAN) {
            // Vrednost je vrste REG_DWORD_BIG_ENDIAN.
            assert(0); // TODO: Preizkusi ta del kode.
            assert(dolzina == sizeof(T_U4));
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&podatek, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _BE2HE(&podatek);
                stanje = A_OK;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_SZ || tip == REG_EXPAND_SZ || tip == REG_MULTI_SZ) {
            // Vrednost je niz. Preberemo jo in jo pretvorimo v tevilo.
            RNiz niz;
            stanje = Preberi(vrednost, niz);
            if (USPESNO(stanje)) {
                podatek = astrtou4(niz.Vrni(), niz.VrniDolzino(), NULL, 0);
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            }
        } else
            stanje = AOSN_NAPAKA_NAPACEN_TIP;
    } else
        stanje = astanje(rezultat, AOSN_NAPAKA_NI_DATOTEKE);

    return stanje;
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, T_U8 &podatek)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD tip, dolzina;
    T_STANJE stanje;
    LONG rezultat;

    // Najprej ugotovimo tip in dolino vrednosti.
    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &tip, NULL, &dolzina);
    if (rezultat == ERROR_SUCCESS) {
        if (tip == REG_QWORD) {
            // Vrednost je res vrste REG_QWORD.
            assert(dolzina == sizeof(T_U8));
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&podatek, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _LE2HE(&podatek);
                stanje = A_OK;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_BINARY && dolzina == sizeof(T_U8)) {
            // Preberemo binarne podatke.
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&podatek, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _LE2HE(&podatek);
                stanje = A_OK;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_DWORD) {
            // Vrednost je vrste REG_DWORD. Preberemo v pomoni T_U4 in prepiemo v T_U8.
            assert(dolzina == sizeof(T_U4));
            T_U4 pom4;
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&pom4, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _LE2HE(&pom4);
                podatek = pom4;
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_DWORD_BIG_ENDIAN) {
            // Vrednost je vrste REG_DWORD_BIG_ENDIAN. Preberemo v pomoni T_U4 in prepiemo v T_U8.
            assert(0); // TODO: Preizkusi ta del kode.
            assert(dolzina == sizeof(T_U4));
            T_U4 pom4;
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)&pom4, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                _BE2HE(&pom4);
                podatek = pom4;
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_SZ || tip == REG_EXPAND_SZ || tip == REG_MULTI_SZ) {
            // Vrednost je niz. Preberemo jo in jo pretvorimo v tevilo.
            RNiz niz;
            stanje = Preberi(vrednost, niz);
            if (USPESNO(stanje)) {
                podatek = astrtou8(niz.Vrni(), niz.VrniDolzino(), NULL, 0);
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            }
        } else
            stanje = AOSN_NAPAKA_NAPACEN_TIP;
    } else
        stanje = astanje(rezultat, AOSN_NAPAKA_NI_DATOTEKE);

    return stanje;
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, RNiz &niz)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD tip, dolzina;
    T_NIZ temp;
    LONG rezultat;
    T_STANJE stanje;

    // Najprej ugotovimo tip in dolino vrednosti.
    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &tip, NULL, &dolzina);
    if (rezultat == ERROR_SUCCESS) {
        if (tip == REG_SZ || tip == REG_MULTI_SZ) {
            temp = niz.Odpusti(dolzina / sizeof(T_ZNAK), FALSE);
            assert(temp);
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)temp, &dolzina);
            niz.Uporabi(temp, dolzina / sizeof(T_ZNAK), FALSE);
            stanje = rezultat == ERROR_SUCCESS ? (tip != REG_MULTI_SZ ? A_OK : AOSN_OPOZORILO_DRUGACEN_TIP) : astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else if (tip == REG_EXPAND_SZ) {
            temp = (T_NIZ)aalloc(dolzina);
            assert(temp);
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)temp, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                // V nizu so reference na spremenljivkje okolja (ENVIRONMENT), ki jih je treba nadomestiti.
                DWORD dolzina2, prava_dolzina;
                T_NIZ temp2 = NULL;

                // Za zaetek ugibamo pravo dolino.
                prava_dolzina = dolzina / sizeof(T_ZNAK) + RREGISTER_KORAK_POVECEVANJA;
                do {
                    dolzina2 = prava_dolzina;
                    temp2 = (T_NIZ)aqrealloc(temp2, dolzina2*sizeof(T_ZNAK), TRUE);
                    prava_dolzina = ::ExpandEnvironmentStrings(temp, temp2, dolzina2);
                } while (prava_dolzina > dolzina2);

                niz.Nastavi(temp2, prava_dolzina);
                afree(temp2);
                afree(temp);
                stanje = A_OK;
            } else {
                afree(temp);
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
            }
        } else if (tip==REG_DWORD) {
            // v registru je zapisan podatek, kot tevilka, zato ga pretvorimo in vrnemo kot niz.
            T_U4 pom;
            stanje = Preberi(vrednost, pom);
            if (USPESNO(stanje)) {
                niz.Oblikuj(_T("%") _T(T_U4_PRINTF_SIZE_PREFIX) _T("u"), T_LOCALE_THREAD_DEFAULT, pom);
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            }
        } else {
            // Gre za neznan tip podatka.
            stanje = AOSN_NAPAKA_NAPACEN_TIP;
        }
    } else
        stanje = astanje(rezultat, AOSN_NAPAKA_NI_DATOTEKE);

    return stanje;
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, RDinTabela<T_ZNAK> &nizz)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD tip, dolzina;
    T_NIZ temp;
    LONG rezultat;
    T_STANJE stanje;

    // Najprej ugotovimo tip in dolino vrednosti.
    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &tip, NULL, &dolzina);
    if (rezultat == ERROR_SUCCESS) {
        if (tip == REG_MULTI_SZ) {
            temp = nizz.Odpusti(dolzina / sizeof(T_ZNAK));
            rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)temp, &dolzina);
            if (rezultat == ERROR_SUCCESS) {
                nizz.Uporabi(temp, dolzina / sizeof(T_ZNAK));
                stanje = A_OK;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);
        } else {
            // e ne moremo neposredno prebrati te vrednosti, kliemo metodo za branje nizov in pretvorimo dobljen niz v ZZ.
            RNiz pom;
            stanje = Preberi(vrednost, pom);
            if (USPESNO(stanje)) {
                nizz.Nastavi(pom.Vrni(), pom.VrniDolzino() + 1);
                nizz += (T_ZNAK)0; // Dodamo e en terminator, da bo ZZ.
                stanje = AOSN_OPOZORILO_DRUGACEN_TIP;
            }
        }
    } else
        stanje = astanje(rezultat, AOSN_NAPAKA_NI_DATOTEKE);

    return stanje;
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, T_REGTIP *tip, RTokPodatkov &datoteka)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD _tip, dolzina;
    T_KAZALEC temp;
    LONG rezultat;
    T_STANJE stanje;

    // Najprej ugotovimo tip in dolino vrednosti.
    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &_tip, NULL, &dolzina);
    if (rezultat == ERROR_SUCCESS) {
        temp = (T_NIZ)aalloc(dolzina);
        assert(temp);

        rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, NULL, (LPBYTE)temp, &dolzina);
        if (rezultat == ERROR_SUCCESS) {
            datoteka.Zapisi(temp, dolzina);
            if (tip) *tip = (T_REGTIP)_tip;
            stanje = datoteka.VrniStanje();
        } else
            stanje = astanje(rezultat, AOSN_NAPAKA_BRANJE);

        afree(temp);
    } else
        stanje = astanje(rezultat, AOSN_NAPAKA_NI_DATOTEKE);

    return stanje;
}


T_STANJE RRegister::Preberi(TK_NIZ vrednost, T_KAZALEC podatki, T_SIZE dolzina, T_REGTIP *tip, T_SIZE *prebrano)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    assert(vrednost);
    assert(podatki || !dolzina);

    DWORD _tip, _dolzina = (DWORD)dolzina;
    LONG rezultat;

    rezultat = ::RegQueryValueEx((HKEY)rocica, vrednost, NULL, &_tip, (LPBYTE)podatki, &_dolzina);
    if (rezultat == ERROR_SUCCESS) {
        if (tip)      *tip = (T_REGTIP)_tip;
        if (prebrano) *prebrano = _dolzina;
        return A_OK;
    } else
        return astanje(rezultat, AOSN_NAPAKA_BRANJE);
}


T_STANJE RRegister::Zapisi(TK_NIZ vrednost, T_U4 podatek)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegSetValueEx((HKEY)rocica, vrednost, 0,
#ifdef DEBELEJSI_KONEC_VER
        REG_DWORD_BIG_ENDIAN,
#else
        REG_DWORD_LITTLE_ENDIAN,
#endif
        (LPBYTE)&podatek, sizeof(T_U4));
    return rezultat == ERROR_SUCCESS ? A_OK: astanje(rezultat, AOSN_NAPAKA_PISANJE);
}


T_STANJE RRegister::Zapisi(TK_NIZ vrednost, T_U8 podatek)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    _HE2LE(podatek);
    rezultat = ::RegSetValueEx((HKEY)rocica, vrednost, 0,
#ifdef _WIN64
        REG_QWORD,
#else
        REG_BINARY,
#endif
        (LPBYTE)&podatek, sizeof(T_U8));
    return rezultat == ERROR_SUCCESS ? A_OK: astanje(rezultat, AOSN_NAPAKA_PISANJE);
}


T_STANJE RRegister::Zapisi(TK_NIZ vrednost, TK_NIZ niz, T_BOOL reference)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegSetValueEx((HKEY)rocica, vrednost, 0, reference ? REG_EXPAND_SZ : REG_SZ, (LPBYTE)niz, (DWORD)((astrlen(niz) + 1) * sizeof(T_ZNAK)));
    return rezultat == ERROR_SUCCESS ? A_OK: astanje(rezultat, AOSN_NAPAKA_PISANJE);
}


T_STANJE RRegister::Zapisi(TK_NIZ vrednost, TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegSetValueEx((HKEY)rocica, vrednost, 0, REG_BINARY, (LPBYTE)podatki, (DWORD)dolzina);
    return rezultat == ERROR_SUCCESS ? A_OK: astanje(rezultat, AOSN_NAPAKA_PISANJE);
}


T_STANJE RRegister::ZapisiSeznam(TK_NIZ vrednost, TK_NIZ nizz)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    T_UINT dolzina = 0, n;
    TK_NIZ p = nizz;
    LONG rezultat;

    while (*p) {
        n = astrlen(p) + 1;
        dolzina += n;
        p += n;
    }

    rezultat = ::RegSetValueEx((HKEY)rocica, vrednost, 0, REG_MULTI_SZ, (LPBYTE)nizz, (DWORD)((dolzina + 1) * sizeof(T_ZNAK)));
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_PISANJE);
}


T_STANJE RRegister::BrisiVrednost(TK_NIZ vrednost)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegDeleteValue((HKEY)rocica, vrednost);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_BRISANJE);
}


T_STANJE RRegister::BrisiPodkljuc(TK_NIZ podKljuc)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    RRegister pod;
    T_STANJE stanje;

    stanje = pod.Odpri(rocica, podKljuc, RREG_VRACANJE_PODKLJUCEV | RREG_NASTAVLJANJE);
    if (USPESNO(stanje)) {
        T_U4 dolzina;
        T_NIZ str = (T_NIZ)aalloc((dolzina = MAX_PATH + 1) * sizeof(T_U4));
        FILETIME datum;
        LONG rezultat = ERROR_SUCCESS;

        for (T_U4 i = 0; rezultat!=ERROR_NO_MORE_ITEMS; ) {
            rezultat = ::RegEnumKeyEx((HKEY)pod.rocica, i, str, &dolzina, NULL, NULL, NULL, &datum);
            if (rezultat==ERROR_SUCCESS)
                verify(USPESNO(pod.BrisiPodkljuc(str)));
            else
                i++;
        }

        afree(str);

        verify(USPESNO(pod.Zapri()));
        rezultat = ::RegDeleteKey((HKEY)rocica, podKljuc);
        stanje = rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_BRISANJE);
    }

    return stanje;
}


T_STANJE RRegister::VrniPodkljuc(T_UINT indeks, RNiz &ime)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD dolzina = RREGISTER_KORAK_POVECEVANJA;
    T_NIZ _ime = ime.Odpusti(dolzina, FALSE);
    FILETIME cas_spremembe;
    LONG rezultat;

    rezultat = ::RegEnumKeyEx((HKEY)rocica, (DWORD)indeks, _ime, &dolzina, NULL, NULL, NULL, &cas_spremembe);
    if (rezultat == ERROR_SUCCESS) {
        ime.Uporabi(_ime, dolzina, FALSE);
        return A_OK;
    } else {
        ime.Uporabi(_ime, 0, FALSE);
        return astanje(rezultat, AOSN_NAPAKA_ISKANJE);
    }
}


T_STANJE RRegister::SpremljajSpremembe(T_UINT filter)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegNotifyChangeKeyValue((HKEY)rocica, (filter & RREG_SPR_SPREMLJAJ_PODKLJUCE) != 0, (DWORD)(filter & RREG_SPR_SIS_MASKA), NULL, FALSE);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_SPREMLJANJE);
}


T_STANJE RRegister::SpremljajSpremembe(T_UINT filter, T_ROCICAOS dogodek)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LONG rezultat;

    rezultat = ::RegNotifyChangeKeyValue((HKEY)rocica, (filter & RREG_SPR_SPREMLJAJ_PODKLJUCE) != 0, (DWORD)(filter & RREG_SPR_SIS_MASKA), dogodek, TRUE);
    return rezultat == ERROR_SUCCESS ? A_OK : astanje(rezultat, AOSN_NAPAKA_SPREMLJANJE);
}

#endif // AOSN_REGISTER
