/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_SEMAFOR


//////////////////////////////////////////////////////////////////////
// RSinhronizator
//////////////////////////////////////////////////////////////////////

RSinhronizator::RSinhronizator(T_ROCICAOS r) : RSisElement(r)
{
}


RSinhronizator::~RSinhronizator()
{
    verify(USPESNO(Zapri()));
}


T_STANJE RSinhronizator::Zakleni(T_UINT naj_cas) const
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD rezultat;

    // Funkcijo WaitForSingleObject() kliemo previdno, saj v primeru, kadar je roica pokvarjena,
    // pride do izjeme 0xe0000002 v KERNELBASE.dll.
    __try {
        rezultat = ::WaitForSingleObject(rocica, (DWORD)naj_cas);
    } __except(EXCEPTION_EXECUTE_HANDLER) {
        return AOSN_NAPAKA_ZAKLEP;
    }
    switch (rezultat) {
        case WAIT_OBJECT_0:
            return A_OK;
        case WAIT_ABANDONED:
            return AOSN_OPOZORILO_ZE_ZAPRTA;
        case WAIT_TIMEOUT:
            return AOSN_NAPAKA_CAS_POTEKEL;
        default:
            return astanje(::GetLastError(), AOSN_NAPAKA_ZAKLEP);
    };
}


//////////////////////////////////////////////////////////////////////
// RDogodek
//////////////////////////////////////////////////////////////////////

RDogodek::RDogodek(T_ROCICAOS r) : RSinhronizator(r)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
}


RDogodek::RDogodek(T_UINT nacin, TK_NIZ ime) : RSinhronizator()
{
    rocica = ::CreateEvent(
        NULL,
        (nacin & RSEM_ROCNI_RESET) == RSEM_ROCNI_RESET,
        (nacin & RSEM_POSTAVLJEN) == RSEM_POSTAVLJEN,
        ime);
    assert(JE_VELJAVNA_ROCICA(rocica));
}


T_STANJE RDogodek::Resetiraj()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::ResetEvent(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_RESETIRAJ);
}


T_STANJE RDogodek::Impulz()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::PulseEvent(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_IMPULZ);
}


T_STANJE RDogodek::Odkleni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::SetEvent(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODKLEP);
}


//////////////////////////////////////////////////////////////////////
// RNevarenOdsek
//////////////////////////////////////////////////////////////////////

RNevarenOdsek::RNevarenOdsek(T_ROCICAOS r) : RSinhronizator(r)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
}


RNevarenOdsek::RNevarenOdsek(T_UINT nacin, TK_NIZ ime) : RSinhronizator()
{
    rocica = ::CreateMutex(
        NULL,
        (nacin & RSEM_POSTAVLJEN) == RSEM_POSTAVLJEN,
        ime);
    assert(JE_VELJAVNA_ROCICA(rocica));
}


T_STANJE RNevarenOdsek::Odkleni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::ReleaseMutex(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODKLEP);
}


//////////////////////////////////////////////////////////////////////
// RSemafor
//////////////////////////////////////////////////////////////////////

RSemafor::RSemafor(T_ROCICAOS r) : RSinhronizator(r)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
}


RSemafor::RSemafor(T_UINT stevilo, T_UINT maksimum, TK_NIZ ime) : RSinhronizator()
{
    assert(stevilo<=maksimum);
    rocica = ::CreateSemaphore(NULL, (LONG)stevilo, (LONG)maksimum, ime);
    assert(JE_VELJAVNA_ROCICA(rocica));
}


T_STANJE RSemafor::Odkleni()
{
    return Odkleni(1);
}


//////////////////////////////////////////////////////////////////////
// RNadzornikImenika
//////////////////////////////////////////////////////////////////////

RNadzornikImenika::RNadzornikImenika(T_ROCICAOS r) : RSinhronizator(r)
{
}


RNadzornikImenika::RNadzornikImenika(TK_NIZ pot, T_UINT nacin) : RSinhronizator()
{
    assert(pot);

    DWORD filter = 0;

    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_IMENA_DATOTEKE)   filter |= FILE_NOTIFY_CHANGE_FILE_NAME;
    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_IMENA_PODIMENIKA) filter |= FILE_NOTIFY_CHANGE_DIR_NAME;
    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_ATRIBUTOV)        filter |= FILE_NOTIFY_CHANGE_ATTRIBUTES;
    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_DOLZINE)          filter |= FILE_NOTIFY_CHANGE_SIZE;
    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_VSEBINE)          filter |= FILE_NOTIFY_CHANGE_LAST_WRITE;
    if (nacin & RSEM_SPREMLJAJ_SPREMEMBO_PRAVIC)           filter |= FILE_NOTIFY_CHANGE_SECURITY;

    rocica = ::FindFirstChangeNotification(pot, nacin & RSEM_SPREMLJAJ_PODIMENIKE ? TRUE : FALSE, filter);
}


RNadzornikImenika::~RNadzornikImenika()
{
    verify(USPESNO(Zapri()));
}


T_STANJE RNadzornikImenika::Zapri()
{
    // e je roica veljavna, jo zapremo in nastavimo na NEVELJAVNA_ROCICA, sicer vrnemo opozorilo.
    if (JE_VELJAVNA_ROCICA(rocica)) {
        if (::FindCloseChangeNotification(rocica)) {
            rocica = NEVELJAVNA_ROCICA;
            return A_OK;
        } else
            return astanje(::GetLastError(), AOSN_NAPAKA_ZAPIRANJE);
    } else
        return AOSN_OPOZORILO_ZE_ZAPRTA;
}


T_STANJE RNadzornikImenika::Odkleni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::FindNextChangeNotification(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREKINI);
}


//////////////////////////////////////////////////////////////////////
// RCasomer
//////////////////////////////////////////////////////////////////////

#if !defined(_WIN32_WCE) && _MSC_VER >= 1200

RCasomer::RCasomer(T_ROCICAOS r) : RSinhronizator(r)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
}


RCasomer::RCasomer(T_UINT nacin, TK_NIZ ime) : RSinhronizator()
{
   // e ne prevede, dodaj #define _WIN32_WINNT 0x0400
    rocica = ::CreateWaitableTimer(
        NULL,
        (nacin & RSEM_ROCNI_RESET) == RSEM_ROCNI_RESET,
        ime);
    assert(JE_VELJAVNA_ROCICA(rocica));
}


RCasomer::~RCasomer()
{
    verify(USPESNO(Zapri()));
}


T_STANJE RCasomer::Nastavi(T_TRENUTEK kdaj, T_UINT perioda)
{
    assert((LONG)perioda >= 0);
    assert(JE_VELJAVNA_ROCICA(rocica));

    LARGE_INTEGER cas;

    // Preracunamo 1 ms v 100 ns.
    cas.QuadPart = kdaj * 10000;

    return ::SetWaitableTimer(rocica, &cas, (LONG)perioda, Rokovalnik, this, FALSE) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVI_CASOVNIK);
}


T_STANJE RCasomer::Prekini()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::CancelWaitableTimer(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_PREKINI);
}



T_STANJE RCasomer::Odkleni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    LARGE_INTEGER cas;

    cas.QuadPart = -1;

    return ::SetWaitableTimer(rocica, &cas, 0, Rokovalnik, this, FALSE) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODKLEP);
}


T_STANJE RCasomer::Zapri()
{
    if (JE_VELJAVNA_ROCICA(rocica)) {
        // Da Windows po unienju asomera po nesrei ne klie ve rokovalnik.
        verify(::CancelWaitableTimer(rocica));
    }

    return RSinhronizator::Zapri();
}


T_VOID RCasomer::ObAlarmu(T_TRENUTEK)
{
}


VOID CALLBACK RCasomer::Rokovalnik(LPVOID lpArgToCompletionRoutine, DWORD dwTimerLowValue, DWORD dwTimerHighValue)
{
    assert(lpArgToCompletionRoutine);
    ((RCasomer*)lpArgToCompletionRoutine)->ObAlarmu(((T_TRENUTEK)dwTimerHighValue << 32) | dwTimerLowValue);
}

#endif


//////////////////////////////////////////////////////////////////////
// RCakalnik
//////////////////////////////////////////////////////////////////////

RCakalnik::RCakalnik(RSinhronizator &sinhr, T_BOOL takoj_zakleni) :
    RElement(),
    sinhronizator(sinhr)
{
    zaklenjen = 0;
    if (takoj_zakleni)
        verify(USPESNO(Zakleni()));
}


RCakalnik::~RCakalnik()
{
    while (zaklenjen)
        verify(USPESNO(Odkleni()));
}


//////////////////////////////////////////////////////////////////////
// RMultiCakalnik
//////////////////////////////////////////////////////////////////////

RMultiCakalnik::RMultiCakalnik(RSinhronizator **sinhronizator, T_UINT stevilo, T_BOOL takoj_zakleni) : RElement()
{
    assert(sinhronizator);
    assert(stevilo);

    this->sinhronizator = (RSinhronizator**)aalloc(stevilo * sizeof(RSinhronizator*)); assert(this->sinhronizator);
    this->rocica        = (T_ROCICAOS*)aalloc(stevilo * sizeof(T_ROCICAOS)); assert(this->rocica);
    this->zaklenjen     = (T_BOOL*)aalloc(stevilo * sizeof(T_BOOL)); assert(this->zaklenjen);
    this->stevilo       = stevilo;

    for (T_UINT i = 0; i < stevilo; i++) {
        assert(sinhronizator[i]);
        this->sinhronizator[i] = sinhronizator[i];
        this->rocica[i]        = *sinhronizator[i];
        this->zaklenjen[i]     = FALSE;
    }

    if (takoj_zakleni)
        verify(USPESNO(Zakleni()));
}


RMultiCakalnik::~RMultiCakalnik()
{
    verify(USPESNO(Odkleni()));

    afree(zaklenjen);
    afree(rocica);
    afree(sinhronizator);
}

#endif // AOSN_SEMAFOR
