/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_SERVIS

#include <fcntl.h>
#include <io.h>

#pragma comment(lib, "advapi32.lib")

#define MAX_DOLZINA_UKAZA   0x100


//////////////////////////////////////////////////////////////////////
// ROskrbnikServisov
//////////////////////////////////////////////////////////////////////

T_STANJE ROskrbnikServisov::Odpri(TK_NIZ racunalnik, TK_NIZ baza, T_UINT dostop)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    // Odpremo oskrbnika servisov.
    rocica = (SC_HANDLE)::OpenSCManager(racunalnik, baza, (DWORD)dostop);
    return JE_VELJAVNA_ROCICA(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_OSKRBNIK_SERVISOV);
}


T_STANJE ROskrbnikServisov::Zapri()
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    if (::CloseServiceHandle((SC_HANDLE)rocica)) {
        rocica = NEVELJAVNA_ROCICA;
        return A_OK;
    } else
        return astanje(::GetLastError(), AOSN_NAPAKA_ZAPIRANJE);
}


T_STANJE ROskrbnikServisov::VrniDolgoIme(TK_NIZ ime, RNiz &dolgo_ime)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD dolzina = 0;
    T_NIZ buf;
    T_STANJE stanje;

    // Ne preverjamo statusa, ker le "tipamo" dolino.
    ::GetServiceDisplayName((SC_HANDLE)rocica, ime, NULL, &dolzina);

    buf = dolgo_ime.Odpusti(dolzina, FALSE);
    assert(buf);
    if (::GetServiceDisplayName((SC_HANDLE)rocica, ime, buf, &dolzina)) {
        dolgo_ime.Uporabi(buf, dolzina, FALSE);
        stanje = A_OK;
    } else {
        dolgo_ime.Uporabi(buf, 0, FALSE);
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VRNI_IME);
    }

    return stanje;
}


T_STANJE ROskrbnikServisov::VrniIme(TK_NIZ dolgo_ime, RNiz &ime)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    DWORD dolzina = 0;
    T_NIZ buf;
    T_STANJE stanje;

    // Ne preverjamo statusa, ker le "tipamo" dolino.
    ::GetServiceKeyName((SC_HANDLE)rocica, dolgo_ime, NULL, &dolzina);

    buf = ime.Odpusti(dolzina, FALSE);
    assert(buf);
    if (::GetServiceKeyName((SC_HANDLE)rocica, dolgo_ime, buf, &dolzina)) {
        ime.Uporabi(buf, dolzina, FALSE);
        stanje = A_OK;
    } else {
        ime.Uporabi(buf, 0, FALSE);
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_VRNI_IME);
    }

    return stanje;
}


//////////////////////////////////////////////////////////////////////
// RSistemskiServis
//////////////////////////////////////////////////////////////////////

RSistemskiServis::RSistemskiServis(T_ROCICAOS r) : RSisElement(r)
{
}


T_STANJE RSistemskiServis::Odpri(T_ROCICAOS oskrbnik, TK_NIZ ime, T_UINT dostop)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    // Odpremo servis.
    rocica = (SC_HANDLE)::OpenService((SC_HANDLE)oskrbnik, ime, (DWORD)dostop);
    return JE_VELJAVNA_ROCICA(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ODPRI_SERVIS);
}


T_STANJE RSistemskiServis::Ustvari(T_ROCICAOS oskrbnik, TK_NIZ ime, TK_NIZ dolgo_ime, TK_NIZ ukazna_vrstica, T_UINT dostop, T_UINT vrsta, T_ZAGONSERVISA zagon, T_OBRAVNAVANAPAKESERVISA obravnava_napake, TK_NIZ skupina, T_U4 *id_skupine, TK_NIZ odvisnosti, TK_NIZ up_ime, TK_NIZ geslo)
{
    assert(NI_VELJAVNA_ROCICA(rocica));
    assert(sizeof(DWORD) == sizeof(T_U4));

    // Ustvarimo servis.
    rocica = (T_ROCICAOS)::CreateService(
        (SC_HANDLE)oskrbnik,
        ime,
        dolgo_ime,
        (DWORD)dostop,
        (DWORD)vrsta,
        zagon,
        obravnava_napake,
        ukazna_vrstica,
        skupina,
        (LPDWORD)id_skupine,
        odvisnosti,
        up_ime,
        geslo
    );
    return JE_VELJAVNA_ROCICA(rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_USTVARI_SERVIS);
}


T_STANJE RSistemskiServis::Zbrisi()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::DeleteService((SC_HANDLE)rocica) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_BRISI_SERVIS);
}


T_STANJE RSistemskiServis::Zapri()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    if (::CloseServiceHandle((SC_HANDLE)rocica)) {
        rocica = NEVELJAVNA_ROCICA;
        return A_OK;
    } else
        return astanje(::GetLastError(), AOSN_NAPAKA_ZAPIRANJE);
}


T_STANJE RSistemskiServis::Pozeni(T_UINT argc, TK_NIZ argv[])
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::StartService((SC_HANDLE)rocica, (DWORD)argc, (LPCTSTR*)argv) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_ZAGON_SERVISA);
}


T_STANJE RSistemskiServis::Sporoci(T_UINT koda, SERVICE_STATUS *status)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    return ::ControlService((SC_HANDLE)rocica, (DWORD)koda, status) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_SPOROCI);
}


T_STANJE RSistemskiServis::Nastavi(TK_NIZ dolgo_ime, TK_NIZ ukazna_vrstica, T_UINT vrsta, T_ZAGONSERVISA zagon, T_OBRAVNAVANAPAKESERVISA obravnava_napake, TK_NIZ skupina, T_U4 *id_skupine, TK_NIZ odvisnosti, TK_NIZ up_ime, TK_NIZ geslo)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    // Nastavimo servis.
    return ::ChangeServiceConfig(
        (SC_HANDLE)rocica,
        (DWORD)vrsta,
        zagon,
        obravnava_napake,
        ukazna_vrstica,
        skupina,
        (LPDWORD)id_skupine,
        odvisnosti,
        up_ime,
        geslo,
        dolgo_ime) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVI);
}


T_STANJE RSistemskiServis::VrniStatus(SERVICE_STATUS_PROCESS *status)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    // Zahtevamo stanje servisa.
    DWORD potrebno;
    return ::QueryServiceStatusEx(
        (SC_HANDLE)rocica,
        SC_STATUS_PROCESS_INFO,
        (LPBYTE)status,
        sizeof(SERVICE_STATUS_PROCESS),
        &potrebno) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_VRNI_STATUS);
}


T_STANJE RSistemskiServis::NastaviOpis(TK_NIZ opis)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    // Nastavimo opis servisa.
    SERVICE_DESCRIPTION sd = {
        opis ?
#if _MSC_VER < 1300
            (LPTSTR)opis
#else
            opis
#endif
            : _T("")
    };
    return ::ChangeServiceConfig2((SC_HANDLE)rocica, SERVICE_CONFIG_DESCRIPTION, (LPVOID)&sd) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVI);
}


T_STANJE RSistemskiServis::NastaviRavnanjeObOkvari(T_UINT perioda, TK_NIZ sporocilo, TK_NIZ ukazna_vrstica, T_UINT st_ravnanj, const T_RAVNANJEOBOKVARISERVISA *ravnanja)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    assert(st_ravnanj || !ravnanja);

    SERVICE_FAILURE_ACTIONS sfa = {
        (DWORD)(perioda / RDTM_FAKTOR_SEKUNDA),
        sporocilo ?
#if _MSC_VER < 1300
        (LPTSTR)sporocilo
#else
        sporocilo
#endif
        : _T(""),
        ukazna_vrstica ?
#if _MSC_VER < 1300
        (LPTSTR)ukazna_vrstica
#else
        ukazna_vrstica
#endif
        : _T(""),
        (DWORD)st_ravnanj
    };
    T_UINT i;
    T_STANJE stanje;

    // Pripraviti moramo tabelo ravnanj.
    sfa.lpsaActions = (SC_ACTION*)aalloc(sizeof(SC_ACTION) * st_ravnanj);
    assert(sfa.lpsaActions);
    for (i = 0; i < st_ravnanj; i++) {
        sfa.lpsaActions[i].Type  = (SC_ACTION_TYPE)ravnanja[i].vrsta_dejanja;
        sfa.lpsaActions[i].Delay = (UINT)ravnanja[i].zacetni_premor;
    }

    // Nastavimo ravnanje ob okvari.
    stanje = ::ChangeServiceConfig2((SC_HANDLE)rocica, SERVICE_CONFIG_FAILURE_ACTIONS, (LPVOID)&sfa) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_NASTAVI);
    afree(sfa.lpsaActions);
    return stanje;
}


//////////////////////////////////////////////////////////////////////
// RRegisterServisov
//////////////////////////////////////////////////////////////////////

RRegisterServisov::RRegisterServisov() :
    RElement(),
    rocica_modula(NULL)
{
    // Nastavimo globalni kazalec na register servisov.
    assert(!register_servisov); // Register servisov je lahko en sam.
    register_servisov = this;
}


RRegisterServisov::~RRegisterServisov()
{
    if (register_servisov == this)
        register_servisov = NULL;
}


T_STANJE RRegisterServisov::ObRegistraciji()
{
    T_STANJE stanje;
    ROskrbnikServisov oskrbnik_servisov;

    stanje = oskrbnik_servisov.Odpri(NULL, NULL, RSER_OSKRBNIK_PISANJE);
    if (USPESNO(stanje)) {
        T_ROCICA rocica;

        stanje = A_OK;
        for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
            if (NEUSPESNO(servisi[rocica]->ObRegistraciji(*this, oskrbnik_servisov)))
                stanje = AOSN_OPOZORILO_NEKAJ_NAPAK;
        }

        verify(USPESNO(oskrbnik_servisov.Zapri()));
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri dostopu do oskrbnika servisov (koda ") _T(T_STANJE_PRINTF) _T(").\n"), stanje);
    }

    return stanje;
}


T_STANJE RRegisterServisov::ObOdjavi()
{
    T_STANJE stanje;
    ROskrbnikServisov oskrbnik_servisov;

    stanje = oskrbnik_servisov.Odpri(NULL, NULL, RSER_OSKRBNIK_PISANJE);
    if (USPESNO(stanje)) {
        T_ROCICA rocica;

        stanje = A_OK;
        for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
            if (NEUSPESNO(servisi[rocica]->ObOdjavi(*this, oskrbnik_servisov)))
                stanje = AOSN_OPOZORILO_NEKAJ_NAPAK;
        }

        verify(USPESNO(oskrbnik_servisov.Zapri()));
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri dostopu do oskrbnika servisov (koda ") _T(T_STANJE_PRINTF) _T(").\n"), stanje);
    }

    return stanje;
}


T_STANJE RRegisterServisov::ObZagonu()
{
    return A_OK;
}


T_STANJE RRegisterServisov::ObZakljucku()
{
    return A_OK;
}


T_INT RRegisterServisov::Pozeni(T_ROCICAOS modul, T_ROCICAOS, TK_NIZ ukazna_vrstica, T_INT)
{
    T_STANJE stanje;
    T_UINT dolzina;

    rocica_modula = modul;

    // Doloimo pot modula.
    verify(USPESNO(AOsn::VrniPotModula(modul, pot_modula)));

    dolzina = pot_modula.NajdiNazaj(_T(LOCILO_IMENIKOV));
    if (dolzina != T_UINT_MAX)
        imenik_modula.Nastavi(pot_modula, dolzina + 1);
    else
        imenik_modula.Sprazni();

    if (!astricmp(ukazna_vrstica, _T("reg"))) {
        // Uporabnik eli registrirati servise.
        stanje = ObRegistraciji();
    } else if (!astricmp(ukazna_vrstica, _T("unreg"))) {
        // Uporabnik eli odjaviti servise.
        stanje = ObOdjavi();
    } else if (!astricmp(ukazna_vrstica, _T("fg"))) {
        // Poenemo servise kot konzolno aplikacijo.
        T_ROCICA rocica;

        // Zviamo prioriteto te niti, saj predstavlja uporabniki vmesnik.
        verify(USPESNO(RNit::NastaviPrioritetoTrenutneNiti(RNIT_PRI_NAJVISJA)));

        // Odpremo konzolno okno.
        verify(::AllocConsole());

#if _MSC_VER < 1300
        typedef long intptr_t;
#endif

        // Pripravimo tokove podatkov
        int
            rocica_konzole_vh = _open_osfhandle((intptr_t)::GetStdHandle(STD_INPUT_HANDLE),  _O_TEXT | _O_RDONLY),
            rocica_konzole_iz = _open_osfhandle((intptr_t)::GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT),
            rocica_konzole_na = _open_osfhandle((intptr_t)::GetStdHandle(STD_ERROR_HANDLE),  _O_TEXT);

        *stdin = *(::_fdopen(rocica_konzole_vh, "r"));
        ::setvbuf(stdin, NULL, _IONBF, 0);
        *stdout = *(::_fdopen(rocica_konzole_iz, "w"));
        ::setvbuf(stdout, NULL, _IONBF, 0);
        *stderr = *(::_fdopen(rocica_konzole_na, "w"));
        ::setvbuf(stderr, NULL, _IONBF, 0);

        verify(::SetConsoleCP(AOSN_PRIVZETI_CP));
        verify(::SetConsoleOutputCP(AOSN_PRIVZETI_CP));

        _ftprintf(stdout, _T("Amebis AOsn v%hs Konzola za simulacijo servisov\n"), AOSN_VERZIJA_STR);

        stanje = ObZagonu();
        if (USPESNO(stanje)) {
            // Poenemo vse servise.
            for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
                RServisImpl *&servis = servisi[rocica];
                TK_NIZ ime = servis->VrniIme();
                TK_NIZ argv[] = {
                    ime,
                };

                _ftprintf(stdout, _T("Zaganjam %s...\n"), ime);
                verify(USPESNO(servis->Odpri(_countof(argv), argv)));
            }

            _ftprintf(stdout, _T("Vnesite ukaz ali poskusite 'help'.\n"));
            for (;;) {
                T_ZNAK ukaz[MAX_DOLZINA_UKAZA];
                T_UINT dolzina_ukaza;

                _ftprintf(stdout, _T("> "));
                if (!_fgetts(ukaz, MAX_DOLZINA_UKAZA, stdin))
                    break;

                dolzina_ukaza = astrnlen(ukaz, MAX_DOLZINA_UKAZA);
                if (ukaz[dolzina_ukaza - 1] == _T('\n'))
                    ukaz[--dolzina_ukaza] = 0;

                if (!astrnicmp(ukaz, _T("list"), MAX_DOLZINA_UKAZA)) {
                    _ftprintf(stdout,
                        _T("Servis       Stanje\n")
                        _T("------       ------\n"));
                    for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
                        RServisImpl *&servis = servisi[rocica];

                        _ftprintf(stdout, _T("%-12s %-12s\n"), servis->VrniIme(),
                            servis->status.dwCurrentState == SERVICE_START_PENDING    ? _T("Zaganja") :
                            servis->status.dwCurrentState == SERVICE_RUNNING          ? _T("Zagnan") :
                            servis->status.dwCurrentState == SERVICE_STOP_PENDING     ? _T("Zaustavlja") :
                            servis->status.dwCurrentState == SERVICE_STOPPED          ? _T("Ustavljen") :
                            servis->status.dwCurrentState == SERVICE_PAUSE_PENDING    ? _T("Prekinja") :
                            servis->status.dwCurrentState == SERVICE_PAUSED           ? _T("Prekinjen") :
                            servis->status.dwCurrentState == SERVICE_CONTINUE_PENDING ? _T("Nadaljuje") :
                                _T("Neznan"));
                    }
                } else if (!astrnicmp(ukaz, _T("exit"), MAX_DOLZINA_UKAZA)) {
                    break;
                } else if (!astrnicmp(ukaz, _T("stop "), 5)) {
                    TK_NIZ ime = ukaz + 5;
                    RServisImpl *servis = VrniServis(ime);
                    if (servis) {
                        stanje = servis->Sporoci(SERVICE_CONTROL_STOP, 0, NULL);
                        if (USPESNO(stanje)) {
                            _ftprintf(stderr, _T("Ukaz stop uspeno poslan servisu %s.\n"), ime);
                        } else
                            _ftprintf(stderr, _T("Pri poiljanju ukaza stop servisu %s je prilo do napake (koda ") _T(T_STANJE_PRINTF) _T(").\n"), ime, stanje);
                    } else
                        _ftprintf(stderr, _T("Neznan servis %s.\n"), ime);
                } else if (!astrnicmp(ukaz, _T("pause "), 6)) {
                    TK_NIZ ime = ukaz + 6;
                    RServisImpl *servis = VrniServis(ime);
                    if (servis) {
                        stanje = servis->Sporoci(SERVICE_CONTROL_PAUSE, 0, NULL);
                        if (USPESNO(stanje)) {
                            _ftprintf(stderr, _T("Ukaz pause uspeno poslan servisu %s.\n"), ime);
                        } else
                            _ftprintf(stderr, _T("Pri poiljanju ukaza pause servisu %s je prilo do napake (koda ") _T(T_STANJE_PRINTF) _T(").\n"), ime, stanje);
                    } else
                        _ftprintf(stderr, _T("Neznan servis %s.\n"), ime);
                } else if (!astrnicmp(ukaz, _T("continue "), 9)) {
                    TK_NIZ ime = ukaz + 9;
                    RServisImpl *servis = VrniServis(ime);
                    if (servis) {
                        stanje = servis->Sporoci(SERVICE_CONTROL_CONTINUE, 0, NULL);
                        if (USPESNO(stanje)) {
                            _ftprintf(stderr, _T("Ukaz continue uspeno poslan servisu %s.\n"), ime);
                        } else
                            _ftprintf(stderr, _T("Pri poiljanju ukaza continue servisu %s je prilo do napake (koda ") _T(T_STANJE_PRINTF) _T(").\n"), ime, stanje);
                    } else
                        _ftprintf(stderr, _T("Neznan servis %s.\n"), ime);
                } else if (!astrnicmp(ukaz, _T("shutdown "), 9)) {
                    TK_NIZ ime = ukaz + 9;
                    RServisImpl *servis = VrniServis(ime);
                    if (servis) {
                        stanje = servis->Sporoci(SERVICE_CONTROL_SHUTDOWN, 0, NULL);
                        if (USPESNO(stanje)) {
                            _ftprintf(stderr, _T("Ukaz shutdown uspeno poslan servisu %s.\n"), ime);
                        } else
                            _ftprintf(stderr, _T("Pri poiljanju ukaza shutdown servisu %s je prilo do napake (koda ") _T(T_STANJE_PRINTF) _T(").\n"), ime, stanje);
                    } else
                        _ftprintf(stderr, _T("Neznan servis %s.\n"), ime);
                } else if (!astrnicmp(ukaz, _T("help"), MAX_DOLZINA_UKAZA)) {
                    _ftprintf(stdout,
                        _T("Ukaz                      Opis\n")
                        _T("----                      ----\n")
                        _T("exit                      Ustavi vse servise in kona\n")
                        _T("help                      Izpie ta navodila\n")
                        _T("list                      Izpie vse servise v modulu in njihovo stanje\n")
                        _T("stop <servis>             Polje servisu ukaz za zaustavitev\n")
                        _T("pause <servis>            Polje servisu ukaz za prekinitev\n")
                        _T("continue <servis>         Polje servisu ukaz za nadaljevanje\n")
                        _T("shutdown <servis>         Polje servisu ukaz za zaustavitev\n")
                        _T("                          (simulira zaustavitev sistema)\n")
                        );
                } else {
                    _ftprintf(stdout, _T("Neznan ukaz. Poskusite 'help'.\n"));
                }
            }

            // Zakljuimo vse servise.
            for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
                RServisImpl *&servis = servisi[rocica];
                if (servis->status.dwCurrentState != SERVICE_STOPPED && servis->status.dwCurrentState != SERVICE_STOP_PENDING) {
                    _ftprintf(stdout, _T("Ustavljam %s...\n"), servis->VrniIme());
                    verify(USPESNO(servis->Sporoci(SERVICE_CONTROL_STOP, 0, NULL)));
                }
                verify(USPESNO(servis->Pocakaj()));
                verify(USPESNO(servis->Zapri()));
            }
        }

        verify(USPESNO(ObZakljucku()));

        verify(::FreeConsole());

        // Vrnemo prioriteto te niti na normalno.
        verify(USPESNO(RNit::NastaviPrioritetoTrenutneNiti(RNIT_PRI_NORMALNA)));
    } else {
        // Poenemo kot servis.
        BOOL rezultat;
        SERVICE_TABLE_ENTRY *servisi_win;
        T_ROCICA rocica;
        T_UINT i, st_servisov = servisi.VrniStevilo();

#ifdef _DEBUG
        // Zaspimo 10 sek, da imamo as priklopiti debugger na proces.
        // RNit::Zaspi(10000);
#endif

        stanje = ObZagonu();
        if (USPESNO(stanje)) {
            // Ustvarimo tabelo za klic funkcije StartServiceCtrlDispatcher().
            servisi_win = (SERVICE_TABLE_ENTRY*)aalloc((st_servisov + 1) * sizeof(SERVICE_TABLE_ENTRY));
            assert(servisi_win);

            // Sestavimo tabelo servisov za StartServiceCtrlDispatcher().
            for (rocica = servisi.VrniPrvega(), i = 0; rocica; rocica = servisi.VrniNaslednjega(rocica), i++) {
                RServisImpl *&servis = servisi[rocica];

                servisi_win[i].lpServiceName = (LPTSTR)servis->VrniIme();
                servisi_win[i].lpServiceProc = RokovalnikServisov;
            }
            servisi_win[i].lpServiceName = NULL;
            servisi_win[i].lpServiceProc = NULL;

            rezultat = ::StartServiceCtrlDispatcher(servisi_win);
            afree(servisi_win);
            if (rezultat) {
                // Poakamo vse servise, da se zakljuijo.
                for (rocica = servisi.VrniPrvega(); rocica; rocica = servisi.VrniNaslednjega(rocica)) {
                    RServisImpl *&servis = servisi[rocica];
                    if (servis->JeVeljaven())
                        verify(USPESNO(servis->Pocakaj()));
                }

                stanje = A_OK;
            } else
                stanje = astanje(::GetLastError(), AOSN_NAPAKA_ZAGON_SERVISA);
        }

        verify(USPESNO(ObZakljucku()));
    }

    if (USPESNO(stanje)) {
        return 0;
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Med delom servisa je prilo do napake (koda ") _T(T_STANJE_PRINTF) _T(")\n"), stanje);
        return 1;
    }
}


VOID WINAPI RRegisterServisov::RokovalnikServisov(DWORD argc, LPTSTR *argv)
{
    assert(register_servisov);

    if (argc > 0) {
        // Poiemo servis po imenu.
        RServisImpl *servis = register_servisov->VrniServis(argv[0]);
        if (servis) {
            T_ROCICAOS nit = RNit::VrniTrenutnoNit();
            T_UINT koda_niti = RNit::VrniKodoTrenutneNiti();

            // Pripnemo na razred RServisImpl na trenutno nit.
            verify(USPESNO(servis->Uporabi(nit, koda_niti)));

            // Izvedemo telo servisa.
            servis->Pozeni(argc, (TK_NIZ*)argv);

            // Odpustimo roico niti iz uporabe (nismo je odprli mi, ne bomo je zaprli mi).
            servis->Odpusti();

            // Zapremo roico, ki jo je vrnila metoda RNit::VrniTrenutnoNit().
            verify(USPESNO(RSisElement::Zapri(nit)));
        } else {
            // Servis s tem imenom ne obstaja.
            assert(0);
        }
    }
}


RRegisterServisov *RRegisterServisov::register_servisov = NULL;


//////////////////////////////////////////////////////////////////////
// RServisImpl
//////////////////////////////////////////////////////////////////////

RServisImpl::RServisImpl() : RDogNitVN()
{
    parametri.stevilo = 0;
    parametri.tabela  = NULL;
}


T_STANJE RServisImpl::Odpri(T_UINT argc, TK_NIZ argv[], T_UINT dolzina_vrste, T_NITZACETNOSTANJE stanje, T_UINT sklad)
{
    assert(argc > 0 && argv);

    // Shranimo parametre.
    parametri.stevilo = argc;
    parametri.tabela  = argv;

    return RDogNit::Odpri(dolzina_vrste, (T_VOID*)FALSE, stanje, sklad);
}


T_VOID RServisImpl::Pozeni(T_UINT argc, TK_NIZ argv[])
{
    assert(argc > 0 && argv);

    // Shranimo parametre.
    parametri.stevilo = argc;
    parametri.tabela  = argv;

    RDogNit::Pozeni((T_VOID*)TRUE);
}


T_VOID RServisImpl::PredRegistracijoServisa()
{
}


T_STANJE RServisImpl::ObZagonu()
{
    return A_OK;
}


T_STANJE RServisImpl::ObZakljucku()
{
    return A_OK;
}


T_STANJE RServisImpl::ObZaustavitvi(T_BOOL)
{
    return A_OK;
}


T_STANJE RServisImpl::ObPrekinitvi()
{
    return A_OK;
}


T_STANJE RServisImpl::ObNadaljevanju()
{
    return A_OK;
}


T_STANJE RServisImpl::ObStatusu()
{
    return A_OK;
}


T_REZULTAT RServisImpl::ObZagonu(T_KAZALEC param)
{
    T_STANJE stanje;

    adebug(ADNE_DIAG, ADNE_AOSN _T("Zaganjam servis %s ...\n"), parametri.tabela[0]);

    // Inicializiramo stanje servisa (ni treba zaklepati, ker nas tukaj SCM e ne bo nadlegoval).
    memset(&status, 0, sizeof(SERVICE_STATUS));
    status.dwServiceType        = SERVICE_WIN32;
    status.dwCurrentState       = SERVICE_START_PENDING;
    status.dwControlsAccepted   = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_PAUSE_CONTINUE;
    status.dwWin32ExitCode      = NO_ERROR;
    status.dwServiceSpecificExitCode = 0;
    status.dwCheckPoint         = 0;
    status.dwWaitHint           = 0;

    PredRegistracijoServisa();

    // Registriramo rokovalnik dogodkov.
    statusna_rocica = param ? ::RegisterServiceCtrlHandlerEx(VrniIme(), RokovalnikUkazov, this) : NULL;
    if (!param || statusna_rocica) {
        adebug(ADNE_DIAG, ADNE_AOSN _T("Rokovalnik dogodkov servisa %s upesno registriran.\n"), parametri.tabela[0]);

        stanje = ObZagonu();
        if (USPESNO(stanje)) {
            adebug(ADNE_DIAG, ADNE_AOSN _T("Servis %s uspeno zagnan.\n"), parametri.tabela[0]);
            verify(USPESNO(SporociStatus(RSER_ZAGNAN)));
            return 0;
        } else
            goto napaka;
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri registraciji rokovalnika servisa (koda MS ") _T(T_STANJE_PRINTF) _T(").\n"), ::GetLastError());
        stanje = astanje(::GetLastError(), AOSN_NAPAKA_ZAGON_SERVISA);
        goto napaka;
    }

napaka:
    status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    status.dwServiceSpecificExitCode = stanje;
    return 1;
}


T_VOID RServisImpl::ObZakljucku(T_VOID*)
{
    T_STANJE stanje;

    adebug(ADNE_DIAG, ADNE_AOSN _T("Zakljuujem servis %s ...\n"), parametri.tabela[0]);
    stanje = ObZakljucku();

    status.dwCurrentState = RSER_ZAUSTAVLJEN;

    if (status.dwWin32ExitCode == NO_ERROR) {
        // Servis ni javil nobene napake.
        if (USPESNO(stanje)) {
            // Servis se je uspesno zakljuil.
            adebug(ADNE_DIAG, ADNE_AOSN _T("Servis %s uspeno zakljuen.\n"), parametri.tabela[0]);
            status.dwWin32ExitCode = NO_ERROR;
            status.dwServiceSpecificExitCode = 0;
        } else {
            // Servis sicer ni javil nobene napake, pa pa je prilo do napake pri zakljuitvi.
            adebug(ADNE_DIAG, ADNE_AOSN _T("Napaka pri zakljuitvi servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), parametri.tabela[0], stanje);
            status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
            status.dwServiceSpecificExitCode = stanje;
        }
    }

    // Sporoimo status.
    verify(USPESNO(SporociStatus()));
}


T_REZULTAT RServisImpl::RokovalnikDogodkov(T_UINT koda, T_UINT param1, T_KAZALEC param2)
{
    T_STANJE stanje;

    switch (koda) {
        case SERVICE_CONTROL_STOP:
        case SERVICE_CONTROL_SHUTDOWN:
            adebug(ADNE_DIAG, ADNE_AOSN _T("Zaustavljam servis %s ...\n"), parametri.tabela[0]);
            status.dwCurrentState = RSER_ZAUSTAVLJA;
            stanje = ObZaustavitvi(koda == SERVICE_CONTROL_SHUTDOWN);
            if (USPESNO(stanje)) {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Servis %s zaustavljen.\n"), parametri.tabela[0]);
                return 0;
            } else {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Napaka pri zaustavitvi servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), parametri.tabela[0], stanje);
                status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
                status.dwServiceSpecificExitCode = stanje;
                return 1;
            }
            break;

        case SERVICE_CONTROL_PAUSE:
            adebug(ADNE_DIAG, ADNE_AOSN _T("Prekinjam servis %s ...\n"), parametri.tabela[0]);
            status.dwCurrentState = RSER_PREKINJA;
            stanje = ObPrekinitvi();
            if (USPESNO(stanje)) {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Servis %s prekinjen.\n"), parametri.tabela[0]);
                verify(USPESNO(SporociStatus(RSER_PREKINJEN)));
                return 0;
            } else {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Napaka pri prekinitvi servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), parametri.tabela[0], stanje);
                status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
                status.dwServiceSpecificExitCode = stanje;
                return 1;
            }
            break;

        case SERVICE_CONTROL_CONTINUE:
            adebug(ADNE_DIAG, ADNE_AOSN _T("Nadaljujem servis %s ...\n"), parametri.tabela[0]);
            status.dwCurrentState = RSER_NADALJUJE;
            stanje = ObNadaljevanju();
            if (USPESNO(stanje)) {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Servis %s ponovno zagnan.\n"), parametri.tabela[0]);
                verify(USPESNO(SporociStatus(RSER_ZAGNAN)));
                return 0;
            } else {
                adebug(ADNE_DIAG, ADNE_AOSN _T("Napaka pri nadaljevanju servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), parametri.tabela[0], stanje);
                status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
                status.dwServiceSpecificExitCode = stanje;
                return 1;
            }
            break;

        case SERVICE_CONTROL_POWEREVENT:
            // TODO: naredi do konca.
            break;

        case SERVICE_CONTROL_INTERROGATE:
            // Ne naredimo ni, le sporoimo SCM-ju trenutno stanje servisa.
            adebug(ADNE_DIAG, ADNE_AOSN _T("Sporoam stanje servisa %s ...\n"), parametri.tabela[0]);
            verify(USPESNO(SporociStatus()));
            return 0;
    }

    return RDogNit::RokovalnikDogodkov(koda, param1, param2);
}


T_STANJE RServisImpl::ObRegistraciji(RRegisterServisov &register_servisov, ROskrbnikServisov &oskrbnik_servisov)
{
    T_STANJE stanje;
    RSistemskiServis servis;
    RNiz ukazna_vrstica;
    T_UINT i;
    T_REGISTRACIJASERVISA reg;
    TK_NIZ ime = VrniIme();

    ukazna_vrstica.Oblikuj(_T("\"%s\""), T_LOCALE_THREAD_DEFAULT, register_servisov.pot_modula.Vrni());
    for (i = 0; i < parametri.stevilo; i++) {
        assert(parametri.tabela[i]);
        ukazna_vrstica.PrilepiO(_T(" \"%s\""), T_LOCALE_THREAD_DEFAULT, parametri.tabela[i]);
    }

    // Napolnimo podatke za registracijo servisa.
    IzpolniPodatkeServisa(&reg);

    stanje = servis.Ustvari(
        oskrbnik_servisov,
        ime,
        reg.polno_ime,
        ukazna_vrstica,
        RSER_SERVIS_VES_DOSTOP,
        reg.vrsta,
        reg.zagon,
        reg.obravnava_napake,
        reg.skupina,
        reg.id_skupine,
        reg.odvisnosti,
        reg.up_ime,
        reg.geslo);

    if (USPESNO(stanje)) {
        adebug(ADNE_INFO, ADNE_AOSN _T("Servis %s uspeno nameen.\n"), ime);
        verify(USPESNO(servis.NastaviOpis(reg.opis)));
        verify(USPESNO(servis.NastaviRavnanjeObOkvari(reg.ravnanje_ob_okvari.perioda, reg.ravnanje_ob_okvari.sporocilo, reg.ravnanje_ob_okvari.ukazna_vrstica, reg.ravnanje_ob_okvari.st_ravnanj, reg.ravnanje_ob_okvari.ravnanja)));
        verify(USPESNO(servis.Zapri()));
        return A_OK;
    } else {
        // Mogoe pa je servis e nameen?
        stanje = servis.Odpri(oskrbnik_servisov, ime, RSER_SERVIS_NASTAVI | RSER_SERVIS_POZENI);
        if (USPESNO(stanje)) {
            stanje = servis.Nastavi(
                reg.polno_ime,
                ukazna_vrstica,
                reg.vrsta,
                reg.zagon,
                reg.obravnava_napake,
                reg.skupina,
                reg.id_skupine,
                reg.odvisnosti,
                reg.up_ime,
                reg.geslo);
            verify(USPESNO(servis.NastaviOpis(reg.opis)));
            verify(USPESNO(servis.NastaviRavnanjeObOkvari(reg.ravnanje_ob_okvari.perioda, reg.ravnanje_ob_okvari.sporocilo, reg.ravnanje_ob_okvari.ukazna_vrstica, reg.ravnanje_ob_okvari.st_ravnanj, reg.ravnanje_ob_okvari.ravnanja)));
            verify(USPESNO(servis.Zapri()));
            return stanje;
        } else {
            adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri nameanju servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), ime, stanje);
            return stanje;
        }
    }
}


T_STANJE RServisImpl::ObOdjavi(RRegisterServisov&, ROskrbnikServisov &oskrbnik_servisov)
{
    T_STANJE stanje;
    RSistemskiServis servis;

    stanje = servis.Odpri(oskrbnik_servisov, VrniIme(), RSER_SERVIS_BRISI);
    if (USPESNO(stanje)) {
        stanje = servis.Zbrisi();
        verify(USPESNO(servis.Zapri()));
        if (USPESNO(stanje)) {
            adebug(ADNE_INFO, ADNE_AOSN _T("Servis %s uspeno odstranjen.\n"), VrniIme());
            return A_OK;
        } else {
            adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri odstranjanju servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), VrniIme(), stanje);
            return stanje;
        }
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Napaka pri odpiranju servisa %s (koda ") _T(T_STANJE_PRINTF) _T(").\n"), VrniIme(), stanje);
        return stanje;
    }
}


T_VOID RServisImpl::IzpolniPodatkeServisa(T_REGISTRACIJASERVISA *reg) const
{
    assert(reg);

    memset(reg, 0, sizeof(T_REGISTRACIJASERVISA));

    reg->polno_ime                         = NULL;
    reg->opis                              = NULL;
    reg->vrsta                             = RSER_SAMOSTOJNI_PROCES;
    reg->zagon                             = RSER_SAMODEJNO;
    reg->obravnava_napake                  = RSER_OBICAJNO;
    reg->skupina                           = NULL;
    reg->id_skupine                        = NULL;
    reg->odvisnosti                        = NULL;
    reg->up_ime                            = NULL;
    reg->geslo                             = NULL;
    reg->ravnanje_ob_okvari.perioda        = NESKONCNO;
    reg->ravnanje_ob_okvari.sporocilo      = NULL;
    reg->ravnanje_ob_okvari.ukazna_vrstica = NULL;
    reg->ravnanje_ob_okvari.st_ravnanj     = 0;
    reg->ravnanje_ob_okvari.ravnanja       = NULL;
}


DWORD WINAPI RServisImpl::RokovalnikUkazov(DWORD ukaz, DWORD tip, LPVOID podatki, LPVOID instanca)
{
    assert(instanca);
    return ukaz < SP_RSERVISIMPL_KONEC && USPESNO(((RServisImpl*)instanca)->Sporoci(ukaz, tip, podatki)) ? NO_ERROR : ERROR_CALL_NOT_IMPLEMENTED;
}

#endif // AOSN_SERVIS
