/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_VIRI

#pragma comment(lib, "gdi32.lib")


//////////////////////////////////////////////////////////////////////
// RBazaVirov
//////////////////////////////////////////////////////////////////////

RBazaVirov::RBazaVirov()
{
}


RBazaVirov::~RBazaVirov()
{
}


T_STANJE RBazaVirov::Dodaj(TK_NIZ pot_modula, T_KODAJEZIKA jezik, T_ROCICA &rocica)
{
    T_STANJE stanje;
    T_UINT dolzina;
    VrniJezikModulaF VrniJezikModula;
    rocica = UstvariVozel();
    T_KNJIZNICAVIROV &k = VrniElement(rocica);

    k.rocica = ::LoadLibraryEx(pot_modula, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
    if (!k.rocica) {
        stanje = AOSN_NAPAKA_NALAGANJE;
        goto napaka1;
    }

    // poiscemo vstopno tocko za funkcijo, ki vrne jezik modula
    VrniJezikModula = (VrniJezikModulaF)GetProcAddress(k.rocica, "VrniJezikModula");
    if (!VrniJezikModula) {
        stanje = AOSN_NAPAKA_MODUL;
        goto napaka2;
    }

    // zapomnimo si tudi pot do modula (ce ga bomo kdaj ponovno nalagali)
    k.pot_modula = astrdup<T_ZNAK>(pot_modula);
    if (!k.pot_modula) {
        stanje = AOSN_NAPAKA_POMNILNIK;
        goto napaka2;
    }

    // ugotovimo jezik modula
    if (NEUSPESNO(VrniJezikModula(&k.jezik, &k.jezik_win))) {
        stanje = AOSN_NAPAKA_JEZIK_MODULA;
        goto napaka3;
    }

    // nalozimo ime jezika iz modula
    dolzina = RNIZ_KORAK_POVECEVANJA;
    k.ime_jezika = NULL;
    for (;;) {
        k.ime_jezika = (T_NIZ)aqrealloc(k.ime_jezika, dolzina*sizeof(T_ZNAK), FALSE);
        if (!k.ime_jezika) {
            stanje = AOSN_NAPAKA_POMNILNIK;
            goto napaka3;
        }

        T_UINT prebrano = ::LoadString(k.rocica, IDS_JEZIK, k.ime_jezika, (int)dolzina);

        if (!prebrano) {
            // nismo nasli niza
            stanje = AOSN_NAPAKA_MANJKA_JEZIK;
            goto napaka4;
        } else if (prebrano == dolzina - 1) {
            // uff, tole je slo pa v zivi rob, niz je najbrz daljsi, zato podaljsamo pomnilnik in poskusimo ponovno
            dolzina += RNIZ_KORAK_POVECEVANJA;
        } else {
            // dovolj bajtov

            // Ce so viri prevedeni z /n (nizi imajo dodano zakljucno niclo), odstejemo odvecno niclo.
            if (!k.ime_jezika[prebrano - 1])
                prebrano--;

            dolzina = prebrano + 1;
            break;
        }
    }

    // skrajsamo niz, da zmanjsamo porabo pomnilnika
    k.ime_jezika = (T_NIZ)arealloc(k.ime_jezika, dolzina*sizeof(T_ZNAK));

    // ce tega modula ne bomo uporabljali kot privzet jezik, ga raje sprostimo
    if (jezik!=k.jezik) {
        verify(::FreeLibrary(k.rocica));
        k.rocica = NULL;
        stanje = AOSN_OPOZORILO_NI_PRIVZET;
    } else
        stanje = A_OK;

    // dodamo modul v seznam
    DodajVozel(rocica);

    return stanje;

napaka4:
    afree(k.ime_jezika);
napaka3:
    afree(k.pot_modula);
napaka2:
    if (k.rocica) verify(::FreeLibrary(k.rocica));
napaka1:
    UniciVozel(rocica);
    rocica = NULL;
    return stanje;
}


T_STANJE RBazaVirov::Dodaj(TK_NIZ pot, TK_NIZ maska, T_KODAJEZIKA jezik, T_ROCICA &rocica)
{
    RNiz iskalna_pot;
    RIskalnikDatotek iskalnik;
    T_ZNAK ime_jezika[5];
    T_KODAJEZIKA2 jezik_win;
    T_STANJE stanje;
    T_BOOL najden;
    T_ROCICA r;

    if (pot[0]) {
        iskalna_pot = pot;

        // ce se pot ne konca z '\', ga dodamo
        if (iskalna_pot[iskalna_pot.VrniDolzino()-1] != _T(LOCILO_IMENIKOV))
            iskalna_pot += _T(LOCILO_IMENIKOV);
    }

    iskalna_pot += maska;

    najden = FALSE;

    stanje = iskalnik.Odpri(iskalna_pot);
    if (USPESNO(stanje)) {
        T_NAJDENADATOTEKA podatki;
        while (USPESNO(iskalnik.Poisci(&podatki))) {
            stanje = Dodaj(podatki.pot_datoteke, jezik, r);

            // ali smo nasli knjiznico v jeziku, ki ga iscemo?
            if (BREZHIBNO(stanje)) {
                najden = TRUE;
                jezik = KODAJEZIKA_NEZNANA;    // poneverimo kodo jezika, da ne nalagamo vec modulov, ki bi vsebovali isti jezik
                rocica = r;
            }
        }

        verify(USPESNO(iskalnik.Zapri()));
        stanje = A_OK;
    }

    // preverimo, ali smo nasli modul izbranega jezika in ali smo sploh nasli kaksen modul
    if (!st_elementov) {
        stanje = AOSN_NAPAKA_NI_MODULOV;
    } else if (!najden) {
        // nismo nasli modula, ki ga bi zelel uporabnik, zato poiscemo, ali obstaja kaksen modul, ki bi podpiral jezik, ki ga ima izbranega uporabnik v Windowsih
        ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_ILANGUAGE, ime_jezika, _countof(ime_jezika));
        jezik_win = (T_KODAJEZIKA2)astrtoui(ime_jezika, _countof(ime_jezika), NULL, 16);

        for (r = VrniPrvega(); r; r = VrniNaslednjega(r)) {
            if (VrniElement(r).jezik_win == jezik_win) {
                najden = TRUE;
                rocica = r;
                break;
            }
        }

        if (!najden && jezik_win != OSNOVNI_JEZIK(jezik_win)) {
            // Mogoe pa so na voljo viri v osnovi enakem jeziku, le v drugi podroni razliici.
            jezik_win = OSNOVNI_JEZIK(jezik_win);
            for (r = VrniPrvega(); r; r = VrniNaslednjega(r)) {
                if (OSNOVNI_JEZIK(VrniElement(r).jezik_win) == jezik_win) {
                    najden = TRUE;
                    rocica = r;
                    break;
                }
            }
        }

        if (!najden) {
            // Ce nismo nasli niti modula s privzetim jezikom uporabnika, poiscemo modul z anglescino. Katerokoli!
            jezik_win = KODAJEZIKA2_ANGLESCINA;
            for (r = VrniPrvega(); r; r = VrniNaslednjega(r)) {
                if (OSNOVNI_JEZIK(VrniElement(r).jezik_win) == jezik_win) {
                    najden = TRUE;
                    rocica = r;
                    break;
                }
            }
        }

        // Ce pa nimamo niti anglescine, v obupu izberemo pac prvo knjiznico z resourci. Ce bodo kitajski, bodo pa kitajski!!!
        if (!najden)
            rocica = VrniPrvega();

        // privzamemo in nalozimo prvi najden modul
        T_KNJIZNICAVIROV &k = VrniElement(rocica);
        k.rocica = ::LoadLibraryEx(k.pot_modula, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
        stanje = AOSN_OPOZORILO_NI_ZELJEN;
    } else
        stanje = A_OK;

    return stanje;
}


T_STANJE RBazaVirov::Sprosti()
{
    T_ROCICA r;

    // sprostimo vse knjiznice
    while ((r = VrniPrvega()) != 0) {
        T_KNJIZNICAVIROV &k = VrniElement(r);

        if (k.rocica)
            // Ne preverjamo uspesnosti, ker v Win98 javlja napako
            ::FreeLibrary(k.rocica);
        if (k.ime_jezika)
            afree(k.ime_jezika);
        if (k.pot_modula)
            afree(k.pot_modula);

        Brisi(r);
    }

    return A_OK;
}


T_STANJE RBazaVirov::Nalozi(T_ROCICA rocica, T_BOOL nalozi)
{
    assert(rocica);
    T_KNJIZNICAVIROV &k = VrniElement(rocica);

    if (nalozi) {
        // knjiznico bomo nalozili
        if (!k.rocica) {
            k.rocica = ::LoadLibraryEx(k.pot_modula, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
            return k.rocica ? A_OK : AOSN_NAPAKA_NALAGANJE;
        } else
            return AOSN_OPOZORILO_ZE_NALOZEN;
    } else {
        // knjiznico bomo sprostili
        if (k.rocica) {
            if (::FreeLibrary(k.rocica)) {
                k.rocica = NULL;
                return A_OK;
            } else
                return AOSN_NAPAKA_SPROSCANJE;
        } else
            return AOSN_OPOZORILO_ZE_SPROSCEN;
    }
}


T_INT RBazaVirov::Primerjaj(T_KNJIZNICAVIROV const &prvi, T_KNJIZNICAVIROV const &drugi) const
{
    // Pri primerjavi imen jezikov uporabimo sistemski locale, ker predvidevamo, da uporabnik razume jezik svojega operacijskega sistema in mu njegov vrstni red urejanja ni tuj.
    return astricoll(prvi.ime_jezika, drugi.ime_jezika, KODALOCALE_SISTEMSKI);
}


T_STANJE RBazaVirov::NaloziPodatke(HMODULE rocica, TK_NIZ vrsta, UINT id, TK_KAZALEC *podatki, T_SIZE *dolzina, T_KODAJEZIKA2 jezik)
{
    assert(podatki && dolzina);

    T_STANJE stanje;
    HRSRC izvor;
    HGLOBAL vir;

    // Poiscemo rocico vira.
    izvor = ::FindResourceEx(rocica, vrsta, MAKEINTRESOURCE(id), (WORD)jezik);
    if (!izvor) {
        stanje = AOSN_NAPAKA_ISKANJE;
        goto napaka;
    }

    // Poiscemo dolzino vira.
    *dolzina = ::SizeofResource(rocica, izvor);
    if (!*dolzina && ::GetLastError() != ERROR_SUCCESS) {
        stanje = AOSN_NAPAKA_VIRI;
        goto napaka;
    }

    // Poiscemo rocico vira v pomnilniku.
    vir = ::LoadResource(rocica, izvor);
    if (!vir) {
        stanje = AOSN_NAPAKA_NALAGANJE;
        goto napaka;
    }

    // Poiscemo kazalec na podatke v pomnilniku.
    *podatki = (const T_VOID*)::LockResource(vir);
    if (!*podatki) {
        stanje = AOSN_NAPAKA_VIRI;
        goto napaka;
    }

    stanje = A_OK;

napaka:
    return stanje;
}


T_STANJE RBazaVirov::NaloziPodatke(HMODULE rocica, TK_NIZ vrsta, UINT id, RTokPodatkov &dat, T_KODAJEZIKA2 jezik)
{
    T_STANJE stanje;
    TK_KAZALEC podatki;
    T_SIZE dolzina;

    stanje = NaloziPodatke(rocica, vrsta, id, &podatki, &dolzina, jezik);
    if (NEUSPESNO(stanje)) {
        goto napaka;
    }

    stanje = dat.Zapisi(podatki, dolzina)==dolzina ? A_OK : AOSN_NAPAKA_POMNILNIK;

napaka:
    return stanje;
}


T_STANJE RBazaVirov::NaloziNiz(HMODULE rocica, UINT id, TK_UNICODE &niz, T_KODAJEZIKA2 jezik)
{
    return aloadstringz(rocica, id, &niz, jezik) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_VIRI);
}


T_STANJE RBazaVirov::NaloziNiz(HMODULE rocica, UINT id, RNizP<T_UZNAK> &niz, T_KODAJEZIKA2 jezik)
{
    TK_UNICODE zaporedje;
    T_UINT dolzina;

    dolzina = aloadstring(rocica, id, &zaporedje,jezik);
    if (dolzina) {
        niz.Nastavi(zaporedje, dolzina);
        return A_OK;
    } else
        return astanje(::GetLastError(), AOSN_NAPAKA_VIRI);
}


T_STANJE RBazaVirov::NaloziNiz(HMODULE rocica, UINT id, T_ASCII niz, T_UINT znakov, T_KODAJEZIKA2 jezik, T_KODNASTRAN kodna_stran)
{
    return aloadstring(rocica, id, niz, znakov, jezik, kodna_stran) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_VIRI);
}


T_STANJE RBazaVirov::NaloziNiz(HMODULE rocica, UINT id, T_UNICODE niz, T_UINT znakov, T_KODAJEZIKA2 jezik)
{
    return aloadstring(rocica, id, niz, znakov, jezik) ? A_OK : astanje(::GetLastError(), AOSN_NAPAKA_VIRI);
}


UINT RBazaVirov::PoisciNajblizjoIkono(LPCGRPICONDIR zbirka, T_UINT sirina, T_UINT visina, T_UINT bitov_na_piko)
{
    T_UINT i, diff, i_min, diff_min;

    if (!bitov_na_piko) {
        // Uporabnik eli naloiti ikono v enaki barvni globini, kot je zaslon.
        HDC dc = ::CreateDC(_T("DISPLAY"), _T(""), _T(""), NULL);
        if (dc) {
            bitov_na_piko = ::GetDeviceCaps(dc, BITSPIXEL);
            verify(::DeleteDC(dc));
        } else
            bitov_na_piko = 24;
    }

    assert(zbirka->idType == 1);

    for (i = 0, i_min = T_UINT_MAX, diff_min = T_UINT_MAX; i < zbirka->idCount; i++) {
        // Nove ikone PNG preskoimo. Z njimi se bomo ubadali na Visti.
        if (!zbirka->idEntries[i].bWidth)
            continue;

        if (zbirka->idEntries[i].wBitCount == bitov_na_piko) {
            // Kot kriterij ikone bomo upotevali vsoto razlike v obeh dimenzijah.
            diff  = (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bWidth)  - (T_INT)sirina);
            diff += (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bHeight) - (T_INT)visina);
            if (diff < diff_min) {
                i_min = i;
                diff_min = diff;
            }
        }
    }

    if (i_min != T_UINT_MAX) {
        // Nali smo ikono, ki je elene barvne globine in ustreznih dimenzij. Vrnemo njen ID.
        return zbirka->idEntries[i_min].nID;
    }

    // Ponovimo vajo, toda sedaj je dovolj, da je tevilo bitov na piko veje ali enako elenemu.
    for (i = 0, i_min = T_UINT_MAX, diff_min = T_UINT_MAX; i < zbirka->idCount; i++) {
        // Nove ikone PNG preskoimo. Z njimi se bomo ubadali na Visti.
        if (!zbirka->idEntries[i].bWidth)
            continue;

        if (zbirka->idEntries[i].wBitCount >= bitov_na_piko) {
            // Kot kriterij ikone bomo upotevali vsoto razlike v obeh dimenzijah.
            diff  = (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bWidth)  - (T_INT)sirina);
            diff += (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bHeight) - (T_INT)visina);
            if (diff < diff_min) {
                i_min = i;
                diff_min = diff;
            }
        }
    }

    if (i_min != T_UINT_MAX) {
        // Nali smo ikono, ki je elene barvne globine in ustreznih dimenzij. Vrnemo njen ID.
        return zbirka->idEntries[i_min].nID;
    }

    // Ponovimo vajo, ampak sedaj nam je pa povsem vseeno, kakna je barvna globina ikone.
    for (i = 0, i_min = T_UINT_MAX, diff_min = T_UINT_MAX; i < zbirka->idCount; i++) {
        // Nove ikone PNG preskoimo. Z njimi se bomo ubadali na Visti.
        if (!zbirka->idEntries[i].bWidth)
            continue;

        // Kot kriterij ikone bomo upotevali vsoto razlike v obeh dimenzijah.
        diff  = (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bWidth)  - (T_INT)sirina);
        diff += (T_UINT)aabs<T_INT>((T_INT)(zbirka->idEntries[i].bHeight) - (T_INT)visina);
        if (diff < diff_min) {
            i_min = i;
            diff_min = diff;
        }
    }

    if (i_min != T_UINT_MAX) {
        // Nali smo ikono, ki je elene barvne globine in ustreznih dimenzij. Vrnemo njen ID.
        return zbirka->idEntries[i_min].nID;
    }

    return 0;
}

#endif // AOSN_VIRI
