/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_VTIC


//////////////////////////////////////////////////////////////////////
// RVticnica
//////////////////////////////////////////////////////////////////////

RVticnica::RVticnica(T_VTICNICA vticnica) : RTokPodatkov()
{
    this->vticnica = vticnica;
}


RVticnica::~RVticnica()
{
    assert(NI_VELJAVNA_VTICNICA(vticnica));
}


T_STANJE RVticnica::Odpri(T_VTICDOMENA domena, T_VTICVRSTA vrsta, T_INT protokol)
{
    assert(NI_VELJAVNA_VTICNICA(vticnica));
    vticnica = asocket(domena, vrsta, protokol);
    return stanje = JE_VELJAVNA_VTICNICA(vticnica) ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_ODPIRANJE);
}


T_STANJE RVticnica::Uporabi(T_VTICNICA vticnica)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    assert(NI_VELJAVNA_VTICNICA(this->vticnica));

    this->vticnica = vticnica;
    return stanje = A_OK;
}


T_VTICNICA RVticnica::Odpusti()
{
    T_VTICNICA v = vticnica;

    vticnica = NEVELJAVNA_VTICNICA;
    stanje = A_OK;
    return v;
}


T_STANJE RVticnica::Privezi(const T_SOCKADDR *naslov, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = abind(vticnica, naslov, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_PRIVEZI);
}


T_STANJE RVticnica::Povezi(const T_SOCKADDR *naslov, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = aconnect(vticnica, naslov, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_POVEZI);
}


T_STANJE RVticnica::Poslusaj(T_UINT najvec_klientov)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = alisten(vticnica, najvec_klientov) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_POSLUSAJ);
}


T_STANJE RVticnica::Sprejmi(RVticnica &odjemalec, T_SOCKADDR *naslov, T_SIZE *dolzina)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    assert(NI_VELJAVNA_VTICNICA(odjemalec.Vrni()));

    T_VTICNICA nova_vticnica = aaccept(vticnica, naslov, dolzina);
    return stanje = JE_VELJAVNA_VTICNICA(nova_vticnica) ? odjemalec.Uporabi(nova_vticnica): astanje(::WSAGetLastError(), AOSN_NAPAKA_SPREJMI);
}


T_STANJE RVticnica::Zakljuci(T_VTICNACINZAPIRANJA nacin)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));

    return stanje = ashutdown(vticnica, nacin) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_ZAKLJUCI);
}


T_SIZE RVticnica::Prejmi(T_KAZALEC podatki, T_SIZE dolzina, T_UINT nacin)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    T_SIZE prebrano = arecv(vticnica, podatki, dolzina, nacin);

    if (prebrano == 0) {
        stanje = AOSN_NAPAKA_KONEC_TOKA;
        return 0;
    } else if (prebrano != SOCKET_ERROR) {
        stanje = A_OK;
        return prebrano;
    } else {
        stanje = astanje(::WSAGetLastError(), AOSN_NAPAKA_BRANJE);
        return 0;
    }
}


T_SIZE RVticnica::Poslji(TK_KAZALEC podatki, T_SIZE dolzina, T_UINT nacin)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    T_SIZE poslano = asend(vticnica, podatki, dolzina, nacin);

    if (poslano != SOCKET_ERROR) {
        stanje = A_OK;
        return poslano;
    } else {
        stanje = astanje(::WSAGetLastError(), AOSN_NAPAKA_PISANJE);
        return 0;
    }
}


T_STANJE RVticnica::VrniNaslovVticnice(T_SOCKADDR *naslov, T_SIZE *dolzina)
{
    assert(naslov);
    assert(dolzina);
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = agetsockname(vticnica, naslov, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_NASLOV_VTICNICE);
}


T_STANJE RVticnica::VrniNaslovVrstnika(T_SOCKADDR *naslov, T_SIZE *dolzina)
{
    assert(naslov);
    assert(dolzina);
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = agetpeername(vticnica, naslov, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_NASLOV_VRSTNIKA);
}


T_STANJE RVticnica::NastaviParameter(T_VTICPARAMETER parameter, const T_VOID* vrednost, T_SIZE dolzina, T_VTICNIVOPARAM nivo)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = asetsockopt(vticnica, nivo, parameter, vrednost, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_NASTAVI_PARAM_VTIC);
}


T_STANJE RVticnica::VrniParameter(T_VTICPARAMETER parameter, T_VOID* vrednost, T_UINT* dolzina, T_VTICNIVOPARAM nivo)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));
    return stanje = agetsockopt(vticnica, nivo, parameter, vrednost, dolzina) == 0 ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_VRNI_PARAM_VTIC);
}


T_SIZE RVticnica::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));

    T_SIZE
        blok,
        prebranih = 0,
        ostalo    = dolzina;

    stanje = A_OK;

    while (ostalo) {
        blok = arecv(vticnica, (T_U1*)podatki+prebranih, ostalo, 0);

        if (blok == 0) {
            stanje = AOSN_NAPAKA_KONEC_TOKA;
            break;
        } else if (blok == SOCKET_ERROR) {
            stanje = astanje(::WSAGetLastError(), AOSN_NAPAKA_BRANJE);
            break;
        }

        prebranih += blok;
        ostalo -= blok;
    };

    return prebranih;
}


T_SIZE RVticnica::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(JE_VELJAVNA_VTICNICA(vticnica));

    T_SIZE
        blok,
        zapisanih = 0,
        ostalo    = dolzina;

    stanje = A_OK;

    while (ostalo) {
        blok = asend(vticnica, (T_U1*)podatki+zapisanih, ostalo, 0);

        if (blok == SOCKET_ERROR) {
            stanje = astanje(::WSAGetLastError(), AOSN_NAPAKA_PISANJE);
            break;
        }

        zapisanih += blok;
        ostalo -= blok;
    };

    return zapisanih;
}


T_STANJE RVticnica::Zapri()
{
    if (JE_VELJAVNA_VTICNICA(vticnica)) {
        if (aclosesocket(vticnica) == 0) {
            vticnica = NEVELJAVNA_VTICNICA;
            stanje = A_OK;
        } else {
            int rezultat = ::WSAGetLastError();
            if (rezultat == WSAENOTSOCK) {
                // Vtinica s to roico v bistvu ne obstaja (ve), oz. je e zaprta. Pravzaprav je stanje, ki smo ga s klicem te funkcije eleli dosei, e doseeno.
                vticnica = NEVELJAVNA_VTICNICA;
                stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
            } else
                stanje = astanje(rezultat, AOSN_NAPAKA_ZAPIRANJE);
        }
    } else
        stanje = AOSN_OPOZORILO_ZE_ZAPRTA;
    return stanje;
}


T_STANJE RVticnica::Izplakni()
{
    return stanje = A_OK;
}


T_STANJE RVticnica::VrniIme(const T_SOCKADDR_IN *naslov, RNiz &vmesnik, T_UINT *vrata)
{
    assert(vrata);
#ifdef _UNICODE
    vmesnik.NastaviA(ainet_ntoa(naslov->sin_addr));
#else
    vmesnik.Nastavi(ainet_ntoa(naslov->sin_addr));
#endif
    *vrata = antohs(naslov->sin_port);
    return A_OK;
}


T_STANJE RVticnica::VrniNaslov(TK_NIZ vmesnik, T_UINT vrata, T_SOCKADDR_IN *naslov)
{
    assert(naslov);

    memset(naslov, 0, sizeof(T_SOCKADDR_IN));

    naslov->sin_family = AF_INET;
    naslov->sin_port = ahtons((u_short)vrata);
    naslov->sin_addr.s_addr = vmesnik ? ainet_addr(vmesnik): ahtonl(INADDR_ANY);

    return naslov->sin_addr.s_addr != INADDR_NONE ? A_OK : astanje(::WSAGetLastError(), AOSN_NAPAKA_NASLOV_VTICNICE);
}


T_STANJE RVticnica::PoisciNaslov(TK_NIZ ime, T_U4 *naslov)
{
    assert(ime);
    assert(naslov);

    T_HOSTENT *he = agethostbyname(ime);
    if (he!=NULL) {
        *naslov = *((T_U4*)he->h_addr_list[0]);
        return A_OK;
    } else
        return astanje(::WSAGetLastError(), AOSN_NAPAKA_NASLOV_VTICNICE);
}

#endif // AOSN_VTIC
