/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#if defined(AOSN_ZLIB_DEFLATE) || defined(AOSN_ZLIB_INFLATE)

#define ZLIB_WINAPI
#define NO_GZIP
#define NO_DUMMY_DECL
#include <zlib.h>


#ifdef AOSN_ZLIB_DEFLATE

//////////////////////////////////////////////////////////////////////
// RZLibKodirnikTokaPodatkov
//////////////////////////////////////////////////////////////////////

RZLibKodirnikTokaPodatkov::RZLibKodirnikTokaPodatkov(T_ZLIB_STOPNJA_KOMPRESIJE stopnja, T_U4 _dolzina_bloka, RTokPodatkov *izvor) :
    dolzina_bloka(_dolzina_bloka),
    RPretvornikToka(izvor)
{
    pom = aalloc(dolzina_bloka);
    assert(pom);

    kodirnik = aalloc(sizeof(z_stream));
    assert(kodirnik);

    memset(kodirnik, 0, sizeof(z_stream));
    //((z_streamp)kodirnik)->zalloc = Z_NULL;
    //((z_streamp)kodirnik)->zfree  = Z_NULL;

    // Nastavimo dolino in kazalec na izhodne podatke.
    ((z_streamp)kodirnik)->avail_out = dolzina_bloka;
    ((z_streamp)kodirnik)->next_out  = (unsigned char*)pom;

    // Vzpostavimo kodirnik.
    verify(deflateInit((z_streamp)kodirnik, stopnja) == Z_OK);
}


RZLibKodirnikTokaPodatkov::~RZLibKodirnikTokaPodatkov()
{
    deflateEnd((z_streamp)kodirnik);
    afree(kodirnik);
    afree(pom);
}


RTokPodatkov* RZLibKodirnikTokaPodatkov::Odpusti()
{
    assert(kodirnik);
    assert(izvor);

    // Nastavimo dolino in kazalec na vhodne podatke.
    ((z_streamp)kodirnik)->avail_in = 0;
    ((z_streamp)kodirnik)->next_in  = NULL;

    do {
        // Resetiramo medpomnilnik.
        ((z_streamp)kodirnik)->avail_out = dolzina_bloka;
        ((z_streamp)kodirnik)->next_out  = (unsigned char*)pom;

        // Zakodiramo podatke.
        verify(deflate((z_streamp)kodirnik, Z_FINISH) != Z_STREAM_ERROR);

        // Zapiemo podatke.
        izvor->Zapisi(pom, dolzina_bloka - ((z_streamp)kodirnik)->avail_out);
        stanje = izvor->VrniStanje();
        if (NEUSPESNO(stanje))
            break;
    } while (((z_streamp)kodirnik)->avail_out == 0);

    // Resetiramo kodirnik in odpustimo izvorni tok.
    ((z_streamp)kodirnik)->avail_out = dolzina_bloka;
    ((z_streamp)kodirnik)->next_out  = (unsigned char*)pom;
    deflateReset((z_streamp)kodirnik);

    return RPretvornikToka::Odpusti();
}


T_SIZE RZLibKodirnikTokaPodatkov::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);
    assert(kodirnik);
    assert(izvor);

    if (dolzina) {
        // Nastavimo dolino in kazalec na izhodne podatke.
        ((z_streamp)kodirnik)->avail_out = (unsigned int)dolzina;
        ((z_streamp)kodirnik)->next_out  = (unsigned char*)podatki;

        stanje = A_OK;

        do {
            if (((z_streamp)kodirnik)->avail_in == 0) {
                // Ni (ve) vhodnih podatkov, zato jih preberemo.
                T_SIZE prebranih = izvor->Preberi(pom, dolzina_bloka);
                stanje = izvor->VrniStanje();
                if (NEUSPESNO(stanje)) {
                    T_SIZE zakodirano = dolzina - ((z_streamp)kodirnik)->avail_out;
                    if (zakodirano) {
                        // Na vhodnem toku je zmanjkalo podatkov. Nekaj smo vendarle uspeno prebrali in zakodirali, zato vrnemo, kar pa imamo.
                        stanje = A_OK;
                        return zakodirano;
                    } else {
                        // Nismo uspeli prebrati in/ali zakodirati niesar.
                        return 0;
                    }
                } else if (!prebranih) {
                    // Nimamo novih prebranih podatkov, zato vrnemo, kar pa imamo.
                    break;
                }

                // Resetiramo medpomnilnik.
                ((z_streamp)kodirnik)->avail_in = (unsigned int)prebranih;
                ((z_streamp)kodirnik)->next_in  = (unsigned char*)pom;
            }

            // Zakodiramo blok vhodnih podatkov.
            verify(deflate((z_streamp)kodirnik, Z_NO_FLUSH) != Z_STREAM_ERROR);
        } while (((z_streamp)kodirnik)->avail_out);

        return dolzina - ((z_streamp)kodirnik)->avail_out;
    } else {
        stanje = A_OK;
        return 0;
    }
}


T_SIZE RZLibKodirnikTokaPodatkov::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);
    assert(kodirnik);
    assert(izvor);

    if (dolzina) {
        T_SIZE zakodiranih = 0, koliko;

        // Nastavimo dolino in kazalec na vhodne podatke.
        ((z_streamp)kodirnik)->avail_in = (unsigned int)dolzina;
        ((z_streamp)kodirnik)->next_in  = (unsigned char*)podatki;

        stanje = A_OK;

        do {
            // Resetiramo medpomnilnik.
            ((z_streamp)kodirnik)->avail_out = dolzina_bloka;
            ((z_streamp)kodirnik)->next_out  = (unsigned char*)pom;

            // Zakodiramo podatke.
            verify(deflate((z_streamp)kodirnik, Z_NO_FLUSH) != Z_STREAM_ERROR);

            // Zapiemo podatke.
            koliko = dolzina_bloka - ((z_streamp)kodirnik)->avail_out;
            if (koliko) {
                izvor->Zapisi(pom, koliko);
                stanje = izvor->VrniStanje();
                if (NEUSPESNO(stanje))
                    break;
            }

            zakodiranih += dolzina - ((z_streamp)kodirnik)->avail_in;
        } while (((z_streamp)kodirnik)->avail_out == 0);

        return zakodiranih;
    } else {
        stanje = A_OK;
        return 0;
    }
}


T_STANJE RZLibKodirnikTokaPodatkov::Izplakni()
{
    assert(kodirnik);
    assert(izvor);

    // Nastavimo dolino in kazalec na vhodne podatke.
    ((z_streamp)kodirnik)->avail_in = 0;
    ((z_streamp)kodirnik)->next_in  = NULL;

    do {
        // Resetiramo medpomnilnik.
        ((z_streamp)kodirnik)->avail_out = dolzina_bloka;
        ((z_streamp)kodirnik)->next_out  = (unsigned char*)pom;

        // Zakodiramo podatke.
        verify(deflate((z_streamp)kodirnik, Z_SYNC_FLUSH) != Z_STREAM_ERROR);

        // Zapiemo podatke.
        izvor->Zapisi(pom, dolzina_bloka - ((z_streamp)kodirnik)->avail_out);
        stanje = izvor->VrniStanje();
        if (NEUSPESNO(stanje))
            break;
    } while (((z_streamp)kodirnik)->avail_out == 0);

    return izvor->Izplakni();
}

#endif // AOSN_ZLIB_DEFLATE


#ifdef AOSN_ZLIB_INFLATE

//////////////////////////////////////////////////////////////////////
// RZLibDekodirnikTokaPodatkov
//////////////////////////////////////////////////////////////////////

RZLibDekodirnikTokaPodatkov::RZLibDekodirnikTokaPodatkov(T_U4 _dolzina_bloka, RTokPodatkov *izvor) :
    dolzina_bloka(_dolzina_bloka),
    RPretvornikToka(izvor)
{
    pom = aalloc(dolzina_bloka);
    assert(pom);

    dekodirnik = aalloc(sizeof(z_stream));
    assert(dekodirnik);

    memset(dekodirnik, 0, sizeof(z_stream));
    //((z_streamp)dekodirnik)->zalloc = Z_NULL;
    //((z_streamp)dekodirnik)->zfree  = Z_NULL;

    // Vzpostavimo dekodirnik.
    verify(inflateInit((z_streamp)dekodirnik) == Z_OK);
}


RZLibDekodirnikTokaPodatkov::~RZLibDekodirnikTokaPodatkov()
{
    inflateEnd((z_streamp)dekodirnik);
    afree(dekodirnik);
    afree(pom);
}


T_SIZE RZLibDekodirnikTokaPodatkov::Preberi(T_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);
    assert(dekodirnik);
    assert(izvor);

    if (dolzina) {
        // Nastavimo dolino in kazalec na izhodne podatke.
        ((z_streamp)dekodirnik)->avail_out = (unsigned int)dolzina;
        ((z_streamp)dekodirnik)->next_out  = (unsigned char*)podatki;

        stanje = A_OK;

        do {
            if (((z_streamp)dekodirnik)->avail_in == 0) {
                // Ni (ve) vhodnih podatkov, zato jih preberemo.
                T_SIZE prebranih = izvor->Preberi(pom, dolzina_bloka);
                stanje = izvor->VrniStanje();
                if (NEUSPESNO(stanje)) {
                    T_SIZE dekodirano = dolzina - ((z_streamp)dekodirnik)->avail_out;
                    if (dekodirano) {
                        // Na vhodnem toku je zmanjkalo podatkov. Nekaj smo vendarle uspeno prebrali in dekodirali, zato vrnemo, kar pa imamo.
                        stanje = A_OK;
                        return dekodirano;
                    } else {
                        // Nismo uspeli prebrati in/ali dekodirati niesar.
                        return 0;
                    }
                } else if (!prebranih) {
                    // Nimamo novih prebranih podatkov, zato vrnemo, kar pa imamo.
                    break;
                }

                // Resetiramo medpomnilnik.
                ((z_streamp)dekodirnik)->avail_in = (unsigned int)prebranih;
                ((z_streamp)dekodirnik)->next_in  = (unsigned char*)pom;
            }

            // Odkodiramo blok vhodnih podatkov.
            verify(inflate((z_streamp)dekodirnik, Z_NO_FLUSH) != Z_STREAM_ERROR);
        } while (((z_streamp)dekodirnik)->avail_out);

        return dolzina - ((z_streamp)dekodirnik)->avail_out;
    } else {
        stanje = A_OK;
        return 0;
    }
}


T_SIZE RZLibDekodirnikTokaPodatkov::Zapisi(TK_KAZALEC podatki, T_SIZE dolzina)
{
    assert(podatki || !dolzina);
    assert(dekodirnik);
    assert(izvor);

    if (dolzina) {
        T_SIZE odkodiranih = 0;

        // Nastavimo dolino in kazalec na vhodne podatke.
        ((z_streamp)dekodirnik)->avail_in = (unsigned int)dolzina;
        ((z_streamp)dekodirnik)->next_in  = (unsigned char*)podatki;

        stanje = A_OK;

        do {
            // Resetiramo medpomnilnik.
            ((z_streamp)dekodirnik)->avail_out = dolzina_bloka;
            ((z_streamp)dekodirnik)->next_out  = (unsigned char*)pom;

            // Odkodiramo podatke.
            verify(inflate((z_streamp)dekodirnik, Z_NO_FLUSH) != Z_STREAM_ERROR);

            // Zapiemo podatke.
            izvor->Zapisi(pom, dolzina_bloka - ((z_streamp)dekodirnik)->avail_out);
            stanje = izvor->VrniStanje();
            if (NEUSPESNO(stanje))
                break;

            odkodiranih += dolzina - ((z_streamp)dekodirnik)->avail_in;
        } while (((z_streamp)dekodirnik)->avail_out == 0);

        return odkodiranih;
    } else {
        stanje = A_OK;
        return 0;
    }
}

#endif // AOSN_ZLIB_INFLATE
#endif // AOSN_ZLIB_DEFLATE || AOSN_ZLIB_INFLATE
