/*
    Copyright 1991-2015 Amebis

    This file is part of AOsn.

    AOsn is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AOsn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AOsn. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"
#include "AOsn.h"

#ifdef AOSN_ZVOK

#pragma comment(lib, "winmm.lib")


#define MM_WOM_ADD      (WM_USER+0x10)
#define MM_WIM_ADD      (WM_USER+0x10)


//////////////////////////////////////////////////////////////////////
// T_BLOKZVOKA
//////////////////////////////////////////////////////////////////////

T_BLOKZVOKA* T_BLOKZVOKA::Ustvari(T_SIZE dolzina, T_UINT instanca, T_BOOL ne_unici)
{
    T_BLOKZVOKA *blok;

    blok = (T_BLOKZVOKA*)aalloc(sizeof(T_BLOKZVOKA) + dolzina);
    if (blok) {
        blok->podatki  = (T_VOID*)(blok + 1);   // kazalec podatki kaze na konec strukture
        blok->dolzina  = dolzina;
        blok->instanca = instanca;
        blok->ne_unici = ne_unici;

        // nastavimo vsa polja opisnika na 0, da ne bo zmede
        memset(&(blok->opisnik), 0, sizeof(WAVEHDR));

        // napolnimo polja opisnika
        blok->opisnik.dwBufferLength = (DWORD)blok->dolzina;
        blok->opisnik.lpData = (LPSTR)blok->podatki;
        blok->opisnik.dwUser = (DWORD_PTR)blok;
    }

    return blok;
}


T_VOID T_BLOKZVOKA::Unici(T_BLOKZVOKA *blok)
{
    assert(blok);
    afree(blok);
}


//////////////////////////////////////////////////////////////////////
// RPredvajalnikZvoka
//////////////////////////////////////////////////////////////////////

RPredvajalnikZvoka::RPredvajalnikZvoka(T_ROCICAOS r, T_UINT max_bl) :
    RNit(),
    rocica(r),
    max_blokov(max_bl)
{
}


RPredvajalnikZvoka::~RPredvajalnikZvoka()
{
    assert(NI_VELJAVNA_ROCICA(rocica));
}


T_STANJE RPredvajalnikZvoka::Odpri(T_UINT vzorcevalna_frekvenca, T_UINT bitov, T_UINT kanalov)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    T_UINT dolzina_vzorca = kanalov * (bitov+7) / 8;
    WAVEFORMATEX format;             // struktura, ki opisuje zapis zvoka

    // ustvarimo nit
    verify(USPESNO(RNit::Odpri()));

    // Pocakamo, da je nit pripravljena
    verify(USPESNO(pripravljen.Pocakaj()));

    // odpremo predvajalnik
    memset(&format, 0, sizeof(WAVEFORMATEX));

    format.wFormatTag      = WAVE_FORMAT_PCM;
    format.nChannels       = (WORD)kanalov;
    format.nSamplesPerSec  = (DWORD)vzorcevalna_frekvenca;
    format.nAvgBytesPerSec = (DWORD)(vzorcevalna_frekvenca * dolzina_vzorca);
    format.nBlockAlign     = (WORD)dolzina_vzorca;
    format.wBitsPerSample  = (WORD)bitov;
    format.cbSize          = 0;

    if (::waveOutOpen((LPHWAVEOUT)&rocica, WAVE_MAPPER, &format, koda_niti, (DWORD_PTR)this, CALLBACK_THREAD) == MMSYSERR_NOERROR) {
        adebug(ADNE_INFO, ADNE_AOSN _T("Odpiranje predvajalnika zvoka uspeno.\n"));
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Odpiranje predvajalnika zvoka neuspeno.\n"));
        rocica = NEVELJAVNA_ROCICA;
        return AOSN_NAPAKA_ODPIRANJE;
    }

    return A_OK;
}


T_STANJE RPredvajalnikZvoka::Zapri(T_BOOL pocakaj)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    // ce ne zelimo cakati na konec, prekinemo predvajanje
    if (!pocakaj) {
        ::waveOutPause((HWAVEOUT)rocica);
        ::waveOutReset((HWAVEOUT)rocica);
    }

    // pocakamo, da se vrsta izprazni
    verify(USPESNO(prazen.Pocakaj()));

    if (::waveOutClose((HWAVEOUT)rocica) == MMSYSERR_NOERROR) {
        adebug(ADNE_INFO, ADNE_AOSN _T("Zapiranje predvajalnika zvoka uspeno.\n"));
        rocica = NEVELJAVNA_ROCICA;
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Zapiranje predvajalnika zvoka neuspeno.\n"));
        return AOSN_NAPAKA_ZAPIRANJE;
    }

    // pocakamo, da je konec niti in jo unicimo
    verify(USPESNO(RNit::Pocakaj()));
    verify(USPESNO(RNit::Zapri()));

    return A_OK;
}


T_STANJE RPredvajalnikZvoka::Predvajaj(T_BLOKZVOKA *blok)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    assert(blok);

    // Pocakamo, da je nit pripravljena
    verify(USPESNO(prost.Pocakaj()));

    ::PostThreadMessage((DWORD)koda_niti, MM_WOM_ADD, 0, (LPARAM)&(blok->opisnik));
    return A_OK;
}


T_REZULTAT RPredvajalnikZvoka::Izvedi(T_VOID*)
{
    MSG sporocilo;
    T_INT rezultat;
    T_UINT blokov;
    T_BOOL konec = FALSE;
    T_BLOKZVOKA *blok;

    // klicemo PeekMessage, da OS ustvari vrsto za sporocila
    rezultat = PeekMessage(&sporocilo, NULL, 0, 0, PM_NOREMOVE);
    blokov = 0;

    // postavimo dogodek, ki signalizira, da je predvajalnik pripravljen
    pripravljen.Postavi();
    prost.Postavi();

    while (!konec) {
        rezultat = GetMessage(&sporocilo, NULL, 0, 0);

        if (rezultat<=0) konec = TRUE;

        switch (sporocilo.message) {
        case MM_WOM_OPEN: // odprtje predvajalnika
            break;
        case MM_WOM_ADD: // nov blok
            assert(rocica);
            blok = (T_BLOKZVOKA*)(((LPWAVEHDR)sporocilo.lParam)->dwUser);
            assert(blok);

            // pripravimo in posljemo blok v predvajanje
            ::waveOutPrepareHeader((HWAVEOUT)rocica, &(blok->opisnik), sizeof(WAVEHDR));
            ::waveOutWrite((HWAVEOUT)rocica, &(blok->opisnik), sizeof(WAVEHDR));
            blokov++;

            if (blokov < max_blokov)
                verify(USPESNO(prost.Postavi()));

            verify(USPESNO(prazen.Resetiraj()));

            break;
        case MM_WOM_DONE: // konec bloka
            // odregistriramo blok
            assert(rocica);
            blok = (T_BLOKZVOKA*)(((LPWAVEHDR)sporocilo.lParam)->dwUser);
            assert(blok);

            // sprostimo pomnilnik
            ::waveOutUnprepareHeader((HWAVEOUT)rocica, &(blok->opisnik), sizeof(WAVEHDR));

            // naznanimo konec bloka
            verify(USPESNO(ObKoncuBloka(blok)));

            // unicimo blok
            if (!blok->ne_unici) T_BLOKZVOKA::Unici(blok);
            blokov--;

            verify(USPESNO(prost.Postavi()));

            // ali je bil to zadnji blok?
            if (!blokov)
                verify(USPESNO(prazen.Postavi()));

            break;
        case MM_WOM_CLOSE: // zaprtje predvajalnika
            konec = TRUE;
            break;
        default: // neprepoznano sporocilo
            break;
        }
    }

    return 0;
}


T_STANJE RPredvajalnikZvoka::ObKoncuBloka(T_BLOKZVOKA*)
{
    return A_OK;
}


//////////////////////////////////////////////////////////////////////
// RSnemalnikZvoka
//////////////////////////////////////////////////////////////////////

RSnemalnikZvoka::RSnemalnikZvoka(T_ROCICAOS r, T_UINT max_bl) :
    RNit(),
    rocica(r),
    max_blokov(max_bl)
{
}


RSnemalnikZvoka::~RSnemalnikZvoka()
{
    assert(NI_VELJAVNA_ROCICA(rocica));
}


T_STANJE RSnemalnikZvoka::Odpri(T_UINT vzorcevalna_frekvenca, T_UINT bitov, T_UINT kanalov)
{
    assert(NI_VELJAVNA_ROCICA(rocica));

    T_UINT dolzina_vzorca = kanalov * (bitov/8);
    WAVEFORMATEX format;    // struktura, ki opisuje zapis zvoka

    // ustvarimo nit
    verify(USPESNO(RNit::Odpri()));

    // Pocakamo, da je nit pripravljena
    verify(USPESNO(pripravljen.Pocakaj()));

    // odpremo snemalnik
    memset(&format, 0, sizeof(WAVEFORMATEX));

    format.wFormatTag      = WAVE_FORMAT_PCM;
    format.nChannels       = (WORD)kanalov;
    format.nSamplesPerSec  = (DWORD)vzorcevalna_frekvenca;
    format.nAvgBytesPerSec = (DWORD)(vzorcevalna_frekvenca * dolzina_vzorca);
    format.nBlockAlign     = (WORD)dolzina_vzorca;
    format.wBitsPerSample  = (WORD)bitov;
    format.cbSize          = 0;

    if (::waveInOpen((LPHWAVEIN)&rocica, WAVE_MAPPER, &format, koda_niti, (DWORD_PTR)this, CALLBACK_THREAD) == MMSYSERR_NOERROR) {
        adebug(ADNE_INFO, ADNE_AOSN _T("Odpiranje snemalnika zvoka uspeno.\n"));
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Odpiranje snemalnika zvoka neuspeno.\n"));
        rocica = NEVELJAVNA_ROCICA;
        return AOSN_NAPAKA_ODPIRANJE;
    }

    return A_OK;
}


T_STANJE RSnemalnikZvoka::Zapri(T_BOOL pocakaj)
{
    assert(JE_VELJAVNA_ROCICA(rocica));

    // ce ne zelimo cakati na konec, prekinemo snemanje
    if (!pocakaj)
        ::waveInReset((HWAVEIN)rocica);

    // pocakamo, da se vrsta izprazni
    verify(USPESNO(prazen.Pocakaj()));

    if (::waveInClose((HWAVEIN)rocica) == MMSYSERR_NOERROR) {
        adebug(ADNE_INFO, ADNE_AOSN _T("Zapiranje snemalnika zvoka uspeno.\n"));
        rocica = NEVELJAVNA_ROCICA;
    } else {
        adebug(ADNE_NAPAKA, ADNE_AOSN _T("Zapiranje snemalnika zvoka neuspeno.\n"));
        return AOSN_NAPAKA_ZAPIRANJE;
    }

    // pocakamo, da je konec niti in jo unicimo
    verify(USPESNO(RNit::Pocakaj()));
    verify(USPESNO(RNit::Zapri()));

    return A_OK;
}


T_STANJE RSnemalnikZvoka::Posnemi(T_BLOKZVOKA *blok)
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    assert(blok);

    // Pocakamo, da je nit pripravljena
    verify(USPESNO(prost.Pocakaj()));

    ::PostThreadMessage((DWORD)koda_niti, MM_WIM_ADD, 0, (LPARAM)&(blok->opisnik));
    return A_OK;
}


T_STANJE RSnemalnikZvoka::Zacni()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    if (::waveInStart((HWAVEIN)rocica) == MMSYSERR_NOERROR)
        return A_OK;
    else
        return AOSN_NAPAKA_SNEMALNIK;
}


T_STANJE RSnemalnikZvoka::Koncaj()
{
    assert(JE_VELJAVNA_ROCICA(rocica));
    if (::waveInStop((HWAVEIN)rocica) == MMSYSERR_NOERROR)
        return A_OK;
    else
        return AOSN_NAPAKA_SNEMALNIK;
}


T_REZULTAT RSnemalnikZvoka::Izvedi(T_VOID*)
{
    MSG sporocilo;
    T_INT rezultat;
    T_UINT blokov;
    T_BOOL konec = FALSE;
    T_BLOKZVOKA *blok;

    // klicemo PeekMessage, da OS ustvari vrsto za sporocila
    rezultat = PeekMessage(&sporocilo, NULL, 0, 0, PM_NOREMOVE);
    blokov = 0;

    // postavimo dogodek, ki signalizira, da je predvajalnik pripravljen
    pripravljen.Postavi();
    prost.Postavi();

    while (!konec) {
        rezultat = GetMessage(&sporocilo, NULL, 0, 0);

        if (rezultat<=0) konec = TRUE;

        switch (sporocilo.message) {
        case MM_WIM_OPEN: // odprtje snemalnika
            break;
        case MM_WIM_ADD: // nov blok
            assert(rocica);
            blok = (T_BLOKZVOKA*)(((LPWAVEHDR)sporocilo.lParam)->dwUser);
            assert(blok);

            // pripravimo in posljemo blok v snemanje
            ::waveInPrepareHeader((HWAVEIN)rocica, &(blok->opisnik), sizeof(WAVEHDR));
            ::waveInAddBuffer((HWAVEIN)rocica, &(blok->opisnik), sizeof(WAVEHDR));
            blokov++;

            if (blokov < max_blokov)
                verify(USPESNO(prost.Postavi()));

            verify(USPESNO(prazen.Resetiraj()));

            break;
        case MM_WIM_DATA: // konec bloka
            // odregistriramo blok
            assert(rocica);
            blok = (T_BLOKZVOKA*)(((LPWAVEHDR)sporocilo.lParam)->dwUser);
            assert(blok);

            // sprostimo pomnilnik
            ::waveInUnprepareHeader((HWAVEIN)rocica, &(blok->opisnik), sizeof(WAVEHDR));

            // naznanimo konec bloka
            verify(USPESNO(ObKoncuBloka(blok)));

            // unicimo blok
            if (!blok->ne_unici) T_BLOKZVOKA::Unici(blok);
            blokov--;

            verify(USPESNO(prost.Postavi()));

            // ali je bil to zadnji blok?
            if (!blokov)
                verify(USPESNO(prazen.Postavi()));

            break;
        case MM_WIM_CLOSE: // zaprtje snemalnika
            konec = TRUE;
            break;
        default: // neprepoznano sporocilo
            break;
        }
    }

    return 0;
}


T_STANJE RSnemalnikZvoka::ObKoncuBloka(T_BLOKZVOKA*)
{
    return A_OK;
}

#endif // AOSN_ZVOK
