/*
    Copyright 1991-2015 Amebis

    This file is part of MSIMSM.

    MSIMSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    MSIMSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MSIMSM. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "AOsn.h"

#define RSEZNAST_DOLZINA_ID 0x100   // najdalji moni ID


////////////////////////////////////////////////////////////////////////////
// T_NASTZAPIS
////////////////////////////////////////////////////////////////////////////

struct T_NASTZAPIS
{
    inline T_NASTZAPIS(TK_UNICODE id = NULL, RElement *element = NULL);

    T_UZNAK id[RSEZNAST_DOLZINA_ID];    // ID zapisa
    RElement *element;                  // kazalec na element (mora biti ustvarjen z operatorjem new)
};


////////////////////////////////////////////////////////////////////////////
// RSeznamNastavitev
////////////////////////////////////////////////////////////////////////////

class RSeznamNastavitev : public RUrejenSeznam<T_NASTZAPIS>
{
public:
    RSeznamNastavitev(T_UINT korak_povecevanja = RSEZNAM_KORAK_POVECEVANJA);
    virtual ~RSeznamNastavitev();

    inline T_VOID Nastavi(TK_UNICODE id, RElement *element);
    inline RElement* Vrni(TK_UNICODE id) const;

    virtual T_VOID Unici();                                                             // Unici vse elemente v seznamu
    virtual T_VOID UniciVozel(T_ROCICA rocica);                                         // Unici vozlisce (mora biti nepovezano)
    virtual T_INT Primerjaj(const T_NASTZAPIS &prvi, const T_NASTZAPIS &drugi) const;   // Primerja enakost dveh elementov
};


////////////////////////////////////////////////////////////////////////////
// Metode inline
////////////////////////////////////////////////////////////////////////////

inline T_NASTZAPIS::T_NASTZAPIS(TK_UNICODE _id, RElement *_element) : element(_element)
{
    if (_id)
        astrncpy(id, _id, RSEZNAST_DOLZINA_ID);
    else
        id[0] = 0;
}


inline T_VOID RSeznamNastavitev::Nastavi(TK_UNICODE id, RElement *element)
{
    assert(id);
    T_ROCICA rocica;

    if (element) {
        T_ROCICA r_n;

        r_n = UstvariVozel();
        assert(r_n);
        T_NASTZAPIS &el_n = VrniElement(r_n);
        astrncpy(el_n.id, id, RSEZNAST_DOLZINA_ID);
        
        // Vedno dodamo element pogojno. e je element s tem ID e v seznamu, vrne njegovo roico.
        rocica = DodajVozelPogojno(r_n);
        if (rocica != r_n) {
            // Zbrisemo novi vozel, ker ga ne potrebujemo ve.
            UniciVozel(r_n);

            // Zbriemo prejni zapis.
            T_NASTZAPIS &el = VrniElement(rocica);
            if (el.element)
                delete el.element;

            // Nastavimo staremu elementu nov zapis.
            el.element = element;
        } else {
            // Nastavimo novemu elementu zapis.
            el_n.element = element;
        }
    } else {
        T_NASTZAPIS iskani(id);

        // Poiemo element in e ga najdemo, ga izbriemo iz seznama.
        rocica = PoisciOdPrvega(iskani);
        if (rocica)
            Brisi(rocica);
    }
}


inline RElement* RSeznamNastavitev::Vrni(TK_UNICODE id) const
{
    assert(id);
    T_NASTZAPIS iskani(id);
    T_ROCICA rocica;

    // Poiemo element.
    rocica = PoisciOdPrvega(iskani);
    return rocica ? VrniElement(rocica).element : NULL;
}
