/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


//////////////////////////////////////////////////////////////////////
// RDatotekaIDT
//////////////////////////////////////////////////////////////////////

RDatotekaIDT::RDatotekaIDT() : RUrejenSeznam<RZapisIDT>()
{
}


T_STANJE RDatotekaIDT::ShraniVDatoteko(TK_NIZ _ime_datoteke, T_UINT dolzina)
{
    assert(_ime_datoteke || !dolzina);

    T_STANJE stanje;
    RMedFizDatoteka izhod;
    RNiz ime_datoteke(_ime_datoteke, dolzina);
    RNizU vrstica;
    T_UINT i, n;
    T_ROCICA rocica;

    // Odpremo izhodno datoteko.
    stanje = izhod.Odpri(_ime_datoteke, RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) goto konec1;

    izhod.ZapisiKodnoStran(kodna_stran);

    // Najprej zapiemo stolpce.
    idlmerge(vrstica, stolpci);
    vrstica += L'\n';
    izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);

    // Nato sledijo podatkovni tipi stolpcev.
    idlmerge(vrstica, podatkovni_tipi);
    vrstica += L'\n';
    izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);

    // Nato sledi kodna stran, ime tabele in kljui.
    vrstica.Oblikuj(L"%Iu\t%.*ls", AOsn::locale_C, (T_UINT)kodna_stran, ime_tabele.VrniDolzino(), ime_tabele.Vrni());
    for (i = 0, n = kljuci.VrniStevilo(); i < n; i++) {
        T_UINT &i_st = kljuci[i];
        assert(i_st < stolpci.VrniStevilo());
        vrstica += L'\t';
        vrstica += stolpci[i_st];
    }
    vrstica += L'\n';
    izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);

    // Sedaj pa e vse vrednosti.
    for (rocica = VrniPrvega(); rocica; rocica = VrniNaslednjega(rocica)) {
        idlmerge(vrstica, VrniElement(rocica));
        vrstica += L'\n';
        izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);
    }

    stanje = A_OK;

    verify(USPESNO(izhod.Zapri()));
konec1:
    return stanje;
}


T_STANJE RDatotekaIDT::NaloziIzDatoteke(TK_NIZ _ime_datoteke, T_UINT dolzina)
{
    assert(_ime_datoteke || !dolzina);

    T_STANJE stanje;
    RMedFizDatoteka vhod;
    T_UINT st_vrstice;
    RNizU vrstica;
    RZapisIDT zapis;
    T_ROCICA rocica;
    RRegElementP<T_UZNAK>
        *re_kodna_stran = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElZacetekVrsticeP<T_UZNAK>(),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"0123456789"), 1),
            new RRegElKonecVrsticeP<T_UZNAK>(),
            NULL);

    // Odpremo vhodno datoteko.
    ime_datoteke.Nastavi(_ime_datoteke, dolzina);
    stanje = vhod.Odpri(ime_datoteke, RDAT_BRANJE | RDAT_SH_BERI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) goto konec1;

    // Doloimo kodno tabelo vhodne datoteke.
    kodna_stran = vhod.DolociKodnoStran();

    for (st_vrstice = 0; ; st_vrstice++) {
        vhod.PreberiVrstico(vrstica, kodna_stran);
        if (vhod.JeKonec()) break;

        if (st_vrstice == 0) {
            // Prva vrstica so imena stolpcev.
            idlsplit(stolpci, vrstica);
        } else if (st_vrstice == 1) {
            // Druga vrstica so podatkovni tipi stolpcev.
            idlsplit(podatkovni_tipi, vrstica);
        } else if (st_vrstice == 2) {
            // Tretja vrstica so podatki o kodni strani, imenu tabele in kljuih.
            T_UINT i, j, m, n;
            RDinTabela<T_UINT> za_prevajanje;

            idlsplit(zapis, vrstica);

            // Analiziramo kodno stran in ime tabele.
            if (zapis.VrniStevilo() && re_kodna_stran->PreveriUjemanje(zapis[0].Vrni(), 0, zapis[0].VrniDolzino())) {
                // Nali smo podatek o vhodni kodni strani.
                if (kodna_stran != KODNASTRAN_UTF8 && kodna_stran != KODNASTRAN_UTF16)
                    kodna_stran = (T_KODNASTRAN)astrtoui(zapis[0].Vrni(), zapis[0].VrniDolzino(), NULL, 10);

                ime_tabele = zapis[1];
                i = 2;
            } else {
                ime_tabele = zapis[0];
                i = 1;
            }

            // Analiziramo kljue.
            kljuci.Sprazni();
            for (n = zapis.VrniStevilo(); i < n; i++) {
                T_UINT j = idlfind(stolpci, zapis[i].Vrni(), zapis[i].VrniDolzino());
                if (j >= stolpci.VrniStevilo()) {
                    stanje = AOSN_NAPAKA_ISKANJE;
                    goto konec2;
                }
                kljuci += j;
            }

            VrniStolpceZaPrevajanje(za_prevajanje);
            n = za_prevajanje.VrniStevilo();
            m = kljuci.VrniStevilo();
            for (i = 0; i < n; i++) {
                for (j = 0; j < m; j++) {
                    if (kljuci[j] == za_prevajanje[i]) {
                        // Klju vsebuje stolpec, ki je namenjen prevajanju.
                        stanje = AOSN_NAPAKA_KROZNA_ODVISNOST;
                        goto konec2;
                    }
                }
            }
        } else {
            T_ROCICA rocica2;

            rocica = UstvariVozel();
            assert(rocica);
            RZapisIDT &zapis = VrniElement(rocica);

            idlsplit(zapis, vrstica);
            rocica2 = DodajVozelPogojno(rocica);
            if (rocica != rocica2) {
                // Nali smo zapis z enakim kljuem, kar pomeni napako.
                UniciVozel(rocica);
                stanje = AOSN_NAPAKA_PODVAJANJE_ROCICE;
                goto konec2;
            }
        }
    }

    // Datoteko smo prebrali. Sedaj zgradimo indeks.
    indeks.Sprazni();
    for (rocica = VrniPrvega(); rocica; rocica = VrniNaslednjega(rocica))
        indeks += rocica;

    stanje = A_OK;

konec2:
    verify(USPESNO(vhod.Zapri()));
konec1:
    re_kodna_stran->Unici();
    return stanje;
}


T_STANJE RDatotekaIDT::ZgradiSeznamPrevodov(RSeznamPrevodov &prevodi, const RDatotekaIDT &prevod) const
{
    RDinTabela<T_UINT> za_prevajanje;
    T_ROCICA rocica;
    T_UINT i, n;

    if (astrnicmp(ime_tabele.Vrni(), ime_tabele.VrniDolzino(), prevod.ime_tabele.Vrni(), prevod.ime_tabele.VrniDolzino())) {
        // Tabeli nista iste vrste.
        return AOSN_NAPAKA_IME;
    }

    VrniStolpceZaPrevajanje(za_prevajanje);
    n = za_prevajanje.VrniStevilo();
    if (n) {
        for (rocica = VrniPrvega(); rocica; rocica = VrniNaslednjega(rocica)) {
            RZapisIDT &el = VrniElement(rocica);
            T_UINT l, d, m;
            T_INT r;
            T_ROCICA rocica2;

            // Z bisekcijo poiemo istoleni zapis v datoteki prevodov.
            for (l = 0, d = prevod.indeks.VrniStevilo(), rocica2 = NULL; l < d; ) {
                m = (l + d) / 2;
                rocica2 = prevod.indeks[m];
                RZapisIDT &el_prevod = prevod.VrniElement(rocica2);

                r = Primerjaj(el, el_prevod);
                if (r < 0) d = m;
                else if (r > 0) l = m + 1;
                else break;
            }

            if (rocica2) {
                // Nali smo zapis s prevodom.
                RZapisIDT &el_prevod = prevod.VrniElement(rocica2);

                for (i = 0; i < n; i++) {
                    T_UINT i_st = za_prevajanje[i];

                    // Oba zapisa morata imeti stolpec s podatki.
                    assert(i_st < el.VrniStevilo());
                    assert(i_st < el_prevod.VrniStevilo());

                    if (!el[i_st].JePrazen())
                        prevodi.Dodaj(el[i_st].Vrni(), el[i_st].VrniDolzino(), el_prevod[i_st].Vrni(), el_prevod[i_st].VrniDolzino(), NULL, 0);
                }
            }
        }
    }

    return A_OK;
}


T_INT RDatotekaIDT::Primerjaj(const RZapisIDT &prvi, const RZapisIDT &drugi) const
{
    T_UINT n = kljuci.VrniStevilo();

    if (n) {
        T_UINT i;
        T_INT r;

        for (i = 0; i < n; i++) {
            T_UINT &i_st = kljuci[i];

            // Oba zapisa morata imeti stolpec s podatki.
            assert(i_st < prvi.VrniStevilo());
            assert(i_st < drugi.VrniStevilo());

            r = astrnicmp(prvi[i_st].Vrni(), prvi[i_st].VrniDolzino(), drugi[i_st].Vrni(), drugi[i_st].VrniDolzino());
            if (r) return r;
        }

        return 0;
    } else {
        // Datoteka ne uporablja kljuev. Vse zapise dodamo vedno na konec.
        return -1;
    }
}
