/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "AOsn.h"


//////////////////////////////////////////////////////////////////////
// RZapisIDT
//////////////////////////////////////////////////////////////////////

typedef RDinTabela<RNizU> RZapisIDT;


//////////////////////////////////////////////////////////////////////
// RDatotekaIDT
//////////////////////////////////////////////////////////////////////

class RDatotekaIDT : public RUrejenSeznam<RZapisIDT>
{
public:
    RDatotekaIDT();

    T_STANJE ShraniVDatoteko(TK_NIZ ime_datoteke, T_UINT dolzina = T_UINT_MAX);
    T_STANJE NaloziIzDatoteke(TK_NIZ ime_datoteke, T_UINT dolzina = T_UINT_MAX);
    T_STANJE ZgradiSeznamPrevodov(RSeznamPrevodov &prevodi, const RDatotekaIDT &prevod) const;
    virtual T_INT Primerjaj(const RZapisIDT &prvi, const RZapisIDT &drugi) const;

    inline T_VOID VrniStolpceZaPrevajanje(RDinTabela<T_UINT> &za_prevajanje) const;
    inline static T_VOID VrniStolpceZaPrevajanje(RDinTabela<T_UINT> &za_prevajanje, TK_UNICODE ime_tabele, T_UINT dolzina, const RZapisIDT &stolpci);

protected:
    RNiz ime_datoteke;
    T_KODNASTRAN kodna_stran;

    RNizU ime_tabele;
    RZapisIDT stolpci;              // Naslovi stolpcev
    RZapisIDT podatkovni_tipi;      // Podatkovni tipi stolpcev
    RDinTabela<T_UINT> kljuci;      // Indeksi stolpcev, ki tvorijo klju
    RDinTabela<T_ROCICA> indeks;    // Indeks urejen po kljuu za iskanje po zbirki z bisekcijo
};


//////////////////////////////////////////////////////////////////////
// Funkcije inline
//////////////////////////////////////////////////////////////////////

inline T_VOID idtunescape(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina)
{
    assert(vhod || !dolzina);
    T_UINT i;

    // Pretvorimo posebne znake v besedilu.
    for (i = 0; i < dolzina && vhod[i]; ) {
        if (vhod[i] == L'\\' && i + 1 < dolzina) {
            switch (vhod[++i]) {
                case L'n':  izhod += L'\n'; i++; break;
                case L't':  izhod += L'\t'; i++; break;
                default:    izhod += L'\\';
            }
        } else
            izhod += vhod[i++];
    }
}


inline T_VOID idtescape(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina)
{
    assert(vhod || !dolzina);
    T_UINT i;

    // Pretvorimo posebne znake v besedilu.
    for (i = 0; i < dolzina && vhod[i]; i++) {
        switch (vhod[i]) {
            case L'\n': izhod += L"\\n";  break;
            case L'\t': izhod += L"\\t";  break;
            default:    izhod += vhod[i];
        }
    }
}


inline T_VOID idlsplit(RZapisIDT &zapis, const RNizU &vrstica)
{
    T_UINT zacetek, i, m = vrstica.VrniDolzino(), n;

    zapis.Sprazni();
    for (zacetek = 0; zacetek <= m; zacetek = i + 1) {
        i = vrstica.Najdi(L'\t', zacetek);
        if (i == RNIZ_NEVELJAVEN_INDEKS) i = m;

        n = zapis.VrniStevilo();
        zapis.NastaviStevilo(n + 1);
        RNizU &el = zapis[n];
        el.Sprazni();
        idtunescape(el, vrstica.Vrni() + zacetek, i - zacetek);
    }
}


inline T_VOID idlmerge(RNizU &vrstica, const RZapisIDT &zapis)
{
    T_UINT i, n;

    vrstica.Sprazni();
    for (i = 0, n = zapis.VrniStevilo(); i < n; i++) {
        if (i) vrstica += L'\t';
        vrstica += zapis[i];
    }
}


inline T_UINT idlfind(const RZapisIDT &zapis, TK_UNICODE naziv, T_UINT dolzina = T_UINT_MAX)
{
    T_UINT i, n;

    for (i = 0, n = zapis.VrniStevilo(); i < n; i++) {
        if (!astrnicmp(zapis[i].Vrni(), zapis[i].VrniDolzino(), naziv, dolzina))
            return i;
    }

    return T_UINT_MAX;
}


//////////////////////////////////////////////////////////////////////
// Metode inline
//////////////////////////////////////////////////////////////////////

inline T_VOID RDatotekaIDT::VrniStolpceZaPrevajanje(RDinTabela<T_UINT> &za_prevajanje) const
{
    VrniStolpceZaPrevajanje(za_prevajanje, ime_tabele.Vrni(), ime_tabele.VrniDolzino(), stolpci);
}


inline T_VOID RDatotekaIDT::VrniStolpceZaPrevajanje(RDinTabela<T_UINT> &za_prevajanje, TK_UNICODE ime_tabele, T_UINT dolzina, const RZapisIDT &stolpci)
{
    assert(ime_tabele || !dolzina);

    za_prevajanje.Sprazni();
    if (!astrnicmp(ime_tabele, dolzina, L"ActionText", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Description");
        za_prevajanje += idlfind(stolpci, L"Template");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Control", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Text");
        za_prevajanje += idlfind(stolpci, L"Help");
    } else if (!astrnicmp(ime_tabele, dolzina, L"ControlEvent", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Argument");
    } else if (!astrnicmp(ime_tabele, dolzina, L"CustomAction", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Target");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Dialog", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Title");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Directory", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"DefaultDir");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Error", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Message");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Feature", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Title");
        za_prevajanje += idlfind(stolpci, L"Description");
    } else if (!astrnicmp(ime_tabele, dolzina, L"File", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"FileName");
        za_prevajanje += idlfind(stolpci, L"Language");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Media", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"DiskPrompt");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Property", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Value");
    } else if (!astrnicmp(ime_tabele, dolzina, L"PublishComponent", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"AppData");
    } else if (!astrnicmp(ime_tabele, dolzina, L"RadioButton", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Text");
        za_prevajanje += idlfind(stolpci, L"Help");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Registry", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Value");
    } else if (!astrnicmp(ime_tabele, dolzina, L"ScheduledTask", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"DisplayName");
        za_prevajanje += idlfind(stolpci, L"Description");
    } else if (!astrnicmp(ime_tabele, dolzina, L"ServiceInstall", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"DisplayName");
        za_prevajanje += idlfind(stolpci, L"Description");
    } else if (!astrnicmp(ime_tabele, dolzina, L"Shortcut", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Name");
        za_prevajanje += idlfind(stolpci, L"Description");
    } else if (!astrnicmp(ime_tabele, dolzina, L"UIText", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Text");
    } else if (!astrnicmp(ime_tabele, dolzina, L"WLANProfile", T_UINT_MAX)) {
        za_prevajanje += idlfind(stolpci, L"Name");
    }
}
