/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


//////////////////////////////////////////////////////////////////////
// Lokalni podatki
//////////////////////////////////////////////////////////////////////

struct T_KODANAZIV {
    T_KODAJEZIKA2 koda;
    TK_UNICODE jezik;
    TK_UNICODE drzava;
};

const T_KODANAZIV jeziki[] = {
    { KODAJEZIKA2_ANGLESCINA_PRIVZETA,    L"English",   L"" },
    { KODAJEZIKA2_ANGLESCINA_UK,          L"English",   L"United Kingdom" },
    { KODAJEZIKA2_ANGLESCINA_US,          L"English",   L"United States" },
    { KODAJEZIKA2_SLOVENSCINA_PRIVZETA,   L"Slovenian", L"" },
    { KODAJEZIKA2_SLOVENSCINA_SI,         L"Slovenian", L"Slovenia" },
    { KODAJEZIKA2_ALBANSCINA_PRIVZETA,    L"Albanian",  L"" },
    { KODAJEZIKA2_ALBANSCINA_AL,          L"Albanian",  L"Albania" },
    { KODAJEZIKA2_NEMSCINA_PRIVZETA,      L"German",    L"" },
    { KODAJEZIKA2_NEMSCINA_DE,            L"German",    L"Germany" },
    { KODAJEZIKA2_FRANCOSCINA_PRIVZETA,   L"French",    L"" },
    { KODAJEZIKA2_FRANCOSCINA_FR,         L"French",    L"France" },
    { KODAJEZIKA2_ITALIJANSCINA_PRIVZETA, L"Italian",   L"" },
    { KODAJEZIKA2_ITALIJANSCINA_IT,       L"Italian",   L"Italy" },
};


//////////////////////////////////////////////////////////////////////
// RSeznamDatotek
//////////////////////////////////////////////////////////////////////

T_INT RSeznamDatotek::Primerjaj(const RNizU &prvi, const RNizU &drugi) const
{
    return astrncoll(prvi.Vrni(), prvi.VrniDolzino(), drugi.Vrni(), drugi.VrniDolzino(), KODALOCALE_UPORABNIKOV);
}


//////////////////////////////////////////////////////////////////////
// RSeznamPrevodov
//////////////////////////////////////////////////////////////////////

RSeznamPrevodov::RSeznamPrevodov() :
    RUrejenSeznam<T_PREVOD>(),
    kodna_stran(KODNASTRAN_UTF8),
    koda_jezika(KODAJEZIKA2_NEZNANA)
{
}


T_VOID RSeznamPrevodov::Dodaj(TK_UNICODE besedilo, T_UINT dolzina_bes, TK_UNICODE prevod, T_UINT dolzina_pre, TK_UNICODE datoteka, T_UINT dolzina_dat)
{
    assert(besedilo || !dolzina_bes);
    assert(prevod   || !dolzina_pre);
    assert(datoteka || !dolzina_dat);

    T_ROCICA rocica, rocica2;

    rocica = UstvariVozel();
    assert(rocica);
    T_PREVOD &el = VrniElement(rocica);

    el.besedilo.Nastavi(besedilo, dolzina_bes);
    rocica2 = DodajVozelPogojno(rocica);
    if (rocica != rocica2) {
        // To besedilo je e v seznamu. Dodamo zapis o datoteki k obstojeemu vozlu, novega pa uniimo.
        T_PREVOD &el2 = VrniElement(rocica2);
        if (el2.prevod.JePrazen() || el2.je_uganjen && !el.je_uganjen && dolzina_pre && prevod[0]) {
            // Prejnji zapis e ni imel prevoda, ali pa je imel uganjen prevod, novi pa je pravi. Zato mu nastavimo novi prevod.
            el2.prevod.Nastavi(prevod, dolzina_pre);
        } else {
            // TODO: Razmisli, e bi bilo smiselno primerjati prevode, e so razlini.
        }
        if (dolzina_dat && datoteka[0])
            el2.datoteke.DodajPogojno(RNizU(datoteka, dolzina_dat));
        UniciVozel(rocica);
    } else {
        // Dodamo e podatek o prevodu in datoteki ter konamo.
        el.prevod.Nastavi(prevod, dolzina_pre);
        if (dolzina_dat && datoteka[0])
            el.datoteke.Dodaj(RNizU(datoteka, dolzina_dat));
    }
}


T_STANJE RSeznamPrevodov::ShraniVDatoteko(TK_NIZ _ime_datoteke)
{
    assert(_ime_datoteke);

    RMedFizDatoteka izhod;
    T_STANJE stanje;
    T_ROCICA rocica;
    RNizU vrstica;

    // Odpremo izhodno datoteko.
    stanje = izhod.Odpri(_ime_datoteke, RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) goto napaka1;

    // Zapiemo glavo.
    stanje = izhod.VrniStanje();
    if (NEUSPESNO(stanje)) goto napaka2;
    vrstica.Oblikuj(
        L"# SOME DESCRIPTIVE TITLE.\n"
        L"# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER\n"
        L"# This file is distributed under the same license as the PACKAGE package.\n"
        L"# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.\n"
        L"#\n"
        L"#, fuzzy\n"
        L"msgid \"\"\n"
        L"msgstr \"\"\n"
        L"\"Project-Id-Version: PACKAGE VERSION\\n\"\n"
        L"\"Report-Msgid-Bugs-To: \\n\"\n"
        L"\"POT-Creation-Date: YEAR-MO-DA HO:MI+ZONE\\n\"\n"
        L"\"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n\"\n"
        L"\"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n\"\n"
        L"\"Language-Team: LANGUAGE <LL@li.org>\\n\"\n"
        L"\"MIME-Version: 1.0\\n\"\n"
        L"\"Content-Type: text/plain; charset=%hs\\n\"\n"
        L"\"Content-Transfer-Encoding: 8bit\\n\"\n"
        L"\"X-Poedit-Language: %ls\\n\"\n"
        L"\"X-Poedit-Country: %ls\\n\"\n\n", AOsn::locale_C,
        AOsn::VrniNazivKodneStrani(kodna_stran),
        VrniImeJezika(koda_jezika),
        VrniImeDrzave(koda_jezika)
    );
    izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);

    // Zapiemo vse prevode.
    for (rocica = VrniPrvega(); rocica; rocica = VrniNaslednjega(rocica)) {
        T_PREVOD &el = VrniElement(rocica);
        T_ROCICA rocica2;

        for (rocica2 = el.datoteke.VrniPrvega(); rocica2; rocica2 = el.datoteke.VrniNaslednjega(rocica2)) {
            RNizU &datoteka = el.datoteke.VrniElement(rocica2);
            stanje = izhod.VrniStanje();
            if (NEUSPESNO(stanje)) goto napaka2;
            vrstica  = L"#: ";
            vrstica += datoteka;
            vrstica += L'\n';
            izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran);
        }
        if (el.je_uganjen) {
            stanje = izhod.VrniStanje();
            if (NEUSPESNO(stanje)) goto napaka2;
            izhod.ZapisiNiz(L"#, fuzzy\n", T_UINT_MAX, kodna_stran);
        }

        stanje = ZapisiElement(izhod, L"msgid", T_UINT_MAX, el.besedilo.Vrni(), el.besedilo.VrniDolzino());
        if (NEUSPESNO(stanje)) goto napaka2;

        stanje = ZapisiElement(izhod, L"msgstr", T_UINT_MAX, el.prevod.Vrni(), el.prevod.VrniDolzino());
        if (NEUSPESNO(stanje)) goto napaka2;

        stanje = izhod.VrniStanje();
        if (NEUSPESNO(stanje)) goto napaka2;
        izhod.ZapisiNiz(L"\n", T_UINT_MAX, kodna_stran);
    }

    ime_datoteke.Nastavi(_ime_datoteke);

    verify(USPESNO(izhod.Zapri()));
    return A_OK;

napaka2:
    verify(USPESNO(izhod.Zapri()));
    verify(USPESNO(AOsn::ZbrisiDatoteko(ime_datoteke)));
napaka1:
    return stanje;
}


T_STANJE RSeznamPrevodov::NaloziIzDatoteke(TK_NIZ _ime_datoteke)
{
    assert(_ime_datoteke);

    RMedFizDatoteka vhod;
    T_STANJE stanje;
    RNizU vrstica, id_uni;
    RRegElementP<T_UZNAK>
        *re_prazna = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElZacetekVrsticeP<T_UZNAK>(),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElKonecVrsticeP<T_UZNAK>(),
            NULL),
        *re_medel_prevod = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElNizP<T_UZNAK>(L"#,"),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElNizP<T_UZNAK>(L"fuzzy"),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            NULL),
        *re_datoteka_ime = new RRegElPonavljanjeP<T_UZNAK>(new RRegElPoljubenZnakP<T_UZNAK>()),
        *re_datoteka = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElNizP<T_UZNAK>(L"#:"),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            re_datoteka_ime,
            NULL),
        *re_komentar = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElZnakP<T_UZNAK>(L'#'),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPoljubenZnakP<T_UZNAK>()),
            NULL),
        *re_zapis_id = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_")),
        *re_zapis_vrednost = new RRegElPonavljanjeP<T_UZNAK>(
            new RRegElVejitevP<T_UZNAK>(
                new RRegElNizP<T_UZNAK>(L"\\\\"),
                new RRegElNizP<T_UZNAK>(L"\\\""),
                new RRegElMaskaP<T_UZNAK>(L"\"", T_UINT_MAX, TRUE),
                NULL)),
        *re_zapis = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            re_zapis_id,
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElZnakP<T_UZNAK>(L'\"'),
            re_zapis_vrednost,
            new RRegElZnakP<T_UZNAK>(L'\"'),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            NULL),
        *re_glava_id = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-")),
        *re_glava_vrednost = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"\n", T_UINT_MAX, TRUE)),
        *re_glava = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElZacetekVrsticeP<T_UZNAK>(),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            re_glava_id,
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElZnakP<T_UZNAK>(L':'),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            re_glava_vrednost,
            new RRegElKonecVrsticeP<T_UZNAK>(),
            NULL),
        *re_content_type_mime = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-")),
            new RRegElZnakP<T_UZNAK>(L'/'),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-")),
            NULL),
        *re_content_type_kodna_stran = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-")),
        *re_content_type = new RRegElZaporedjeP<T_UZNAK>(
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            re_content_type_mime,
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElZnakP<T_UZNAK>(L';'),
            new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
            new RRegElVejitevP<T_UZNAK>(
                new RRegElZaporedjeP<T_UZNAK>(
                    new RRegElNizP<T_UZNAK>(L"charset="),
                    re_content_type_kodna_stran,
                    NULL),
                new RRegElNOPP<T_UZNAK>(),
                NULL),
            NULL);

    // Odpremo vhodno datoteko.
    stanje = vhod.Odpri(_ime_datoteke, RDAT_BRANJE | RDAT_SH_BERI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) goto napaka1;

    Unici();

    // Doloimo zaetno kodno stran. Kasneje jo bomo popravili glede na glavo Content-Type.
    kodna_stran = vhod.DolociKodnoStran();

    for (;;) {
        T_ROCICA rocica = UstvariVozel();
        T_PREVOD &zapis = VrniElement(rocica);

        for (;;) {
            vhod.PreberiVrstico(vrstica, kodna_stran);
            if (vhod.JeKonec()) break;

            // e je vrstica prazna, je konec zapisa.
            if (re_prazna->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino()))
                break;

            // e vrstica oznauje medel prevod, si to zapomnimo in preskoimo.
            if (re_medel_prevod->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino(), RREGIZR_NE_LOCUJ_VELIKIH_MALIH)) {
                zapis.je_uganjen = TRUE;
                continue;
            }

            // e vrstica oznauje ime datoteke, si to zapomnimo in preskoimo.
            if (re_datoteka->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino())) {
                T_ROCICA rocica = zapis.datoteke.UstvariVozel();
                RNizU &datoteka = zapis.datoteke.VrniElement(rocica);
                datoteka.Nastavi(vrstica.Vrni() + re_datoteka_ime->interval.zacetek, re_datoteka_ime->interval.VrniDolzino());
                zapis.datoteke.DodajVozel(rocica);
                continue;
            }

            // e vrstica predstavlja komentar, jo preskoimo.
            if (re_komentar->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino()))
                continue;

            // e vrstica predstavlja zapis, ga analiziramo.
            if (re_zapis->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino())) {
                if (re_zapis_id->interval.zacetek < re_zapis_id->interval.konec)
                    id_uni.Nastavi(vrstica.Vrni() + re_zapis_id->interval.zacetek, re_zapis_id->interval.VrniDolzino());

                if (id_uni.PrimerjajI(L"msgid") == 0)
                    OdscitiPosebneZnake(zapis.besedilo, vrstica.Vrni() + re_zapis_vrednost->interval.zacetek, re_zapis_vrednost->interval.VrniDolzino());
                else if (id_uni.PrimerjajI(L"msgstr") == 0)
                    OdscitiPosebneZnake(zapis.prevod,   vrstica.Vrni() + re_zapis_vrednost->interval.zacetek, re_zapis_vrednost->interval.VrniDolzino());

                continue;
            }
        }

        if (!zapis.prevod.JePrazen()) {
            if (zapis.besedilo.JePrazen()) {
                // Nali smo glavo. Preberemo podatke, ki nas zanimajo.
                T_UINT i, n = zapis.prevod.VrniDolzino();
                RNizU jezik, drzava;
                RNizA jezik_iso;

                for (i = 0; i < n;) {
                    if (re_glava->PreveriUjemanje(zapis.prevod.Vrni(), i, n)) {
                        if (!astrnicmp(L"Content-Type", zapis.prevod.Vrni() + re_glava_id->interval.zacetek, re_glava_id->interval.VrniDolzino())) {
                            // Analiziramo glavo "Content-Type", da ugotovimo kodno stran.
                            if (re_content_type->PreveriUjemanje(zapis.prevod.Vrni(), re_glava_vrednost->interval.zacetek, re_glava_vrednost->interval.konec, RREGIZR_NE_LOCUJ_VELIKIH_MALIH) &&
                                re_content_type_kodna_stran->interval.zacetek < re_content_type_kodna_stran->interval.konec)
                            {
                                RNizA naziv(zapis.prevod.Vrni() + re_content_type_kodna_stran->interval.zacetek, re_content_type_kodna_stran->interval.VrniDolzino());
                                T_KODNASTRAN kodna_stran2 = AOsn::VrniKodoKodneStrani(naziv.Vrni(), naziv.VrniDolzino());
                                if (kodna_stran2 != KODNASTRAN_NEZNANA) {
                                    // Prepoznali smo kodno stran.
                                    kodna_stran = kodna_stran2;
                                }
                            }
                        } else if (!astrnicmp(L"X-Poedit-Language", zapis.prevod.Vrni() + re_glava_id->interval.zacetek, re_glava_id->interval.VrniDolzino()))
                            jezik.Nastavi(zapis.prevod.Vrni() + re_glava_vrednost->interval.zacetek, re_glava_vrednost->interval.VrniDolzino());
                        else if (!astrnicmp(L"X-Poedit-Country", zapis.prevod.Vrni() + re_glava_id->interval.zacetek, re_glava_id->interval.VrniDolzino()))
                            drzava.Nastavi(zapis.prevod.Vrni() + re_glava_vrednost->interval.zacetek, re_glava_vrednost->interval.VrniDolzino());
                        else if (!astrnicmp(L"Language", zapis.prevod.Vrni() + re_glava_id->interval.zacetek, re_glava_id->interval.VrniDolzino()))
                            jezik_iso.NastaviU(zapis.prevod.Vrni() + re_glava_vrednost->interval.zacetek, re_glava_vrednost->interval.VrniDolzino(), kodna_stran);
                        i = re_glava->interval.konec;
                    } else
                        i++;
                }

                if (!jezik_iso.JePrazen())
                    koda_jezika = AOsn::VrniKodoJezika(jezik_iso.Vrni(), jezik_iso.VrniDolzino());
                else
                    koda_jezika = VrniKodoJezika(jezik.Vrni(), drzava.Vrni());
            }

            T_ROCICA rocica2 = DodajVozelPogojno(rocica);
            if (rocica2 != rocica) {
                T_PREVOD &zapis2 = VrniElement(rocica2);
                if (zapis2.je_uganjen && !zapis.je_uganjen) {
                    // Ha! Prejnji zapis je uganjen, novi pa ne. Prenesemo podatke v prejnji zapis.
                    zapis2.prevod = zapis.prevod;
                    zapis2.je_uganjen = FALSE;
                    for (T_ROCICA rocica3 = zapis.datoteke.VrniPrvega(); rocica3; rocica3 = zapis.datoteke.VrniNaslednjega(rocica3))
                        zapis2.datoteke.DodajPogojno(zapis.datoteke.VrniElement(rocica3));
                }

                UniciVozel(rocica);
            }
        } else {
            UniciVozel(rocica);
            if (vhod.JeKonec())
                break;
        }
    }

    ime_datoteke.Nastavi(_ime_datoteke);

    stanje = A_OK;

    verify(USPESNO(vhod.Zapri()));
napaka1:
    re_content_type->Unici();
    re_glava->Unici();
    re_zapis->Unici();
    re_komentar->Unici();
    re_datoteka->Unici();
    re_medel_prevod->Unici();
    re_prazna->Unici();

    return stanje;
}


T_ROCICA RSeznamPrevodov::NajdiPrevod(TK_UNICODE besedilo, T_UINT dolzina) const
{
    assert(besedilo || !dolzina);

    T_ROCICA rocica;

    for (rocica = VrniPrvega(); rocica; rocica = VrniNaslednjega(rocica)) {
        T_PREVOD &prevod = VrniElement(rocica);

        // Uganjene in prazne prevode ignoriramo.
        if (prevod.je_uganjen || prevod.prevod.JePrazen()) continue;

        T_INT r = astrncoll(besedilo, dolzina, prevod.besedilo.Vrni(), prevod.besedilo.VrniDolzino(), KODALOCALE_UPORABNIKOV);
        if (!r) {
            return rocica;
        } else if (r < 0) {
            // Prevoda ne bomo nali.
            break;
        }
    }

    return NULL;
}


T_INT RSeznamPrevodov::Primerjaj(const T_PREVOD &prvi, const T_PREVOD &drugi) const
{
    return astrncoll(prvi.besedilo.Vrni(), prvi.besedilo.VrniDolzino(), drugi.besedilo.Vrni(), drugi.besedilo.VrniDolzino(), KODALOCALE_UPORABNIKOV);
}


TK_UNICODE RSeznamPrevodov::VrniImeJezika(T_KODAJEZIKA2 jezik)
{
    T_UINT i;

    for (i = 0; i < sizeof(jeziki) / sizeof(T_KODANAZIV); i++)
        if (jeziki[i].koda == jezik)
            return jeziki[i].jezik;

    return L"";
}


TK_UNICODE RSeznamPrevodov::VrniImeDrzave(T_KODAJEZIKA2 jezik)
{
    T_UINT i;

    for (i = 0; i < sizeof(jeziki) / sizeof(T_KODANAZIV); i++)
        if (jeziki[i].koda == jezik)
            return jeziki[i].drzava;

    return L"";
}


T_KODAJEZIKA2 RSeznamPrevodov::VrniKodoJezika(TK_UNICODE jezik, TK_UNICODE drzava)
{
    T_UINT i;

    for (i = 0; i < sizeof(jeziki) / sizeof(T_KODANAZIV); i++)
        if (!astricmp(jezik, jeziki[i].jezik) && !astricmp(drzava, jeziki[i].drzava))
            return jeziki[i].koda;

    return KODAJEZIKA2_NEZNANA;
}
