/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "AOsn.h"


//////////////////////////////////////////////////////////////////////
// RSeznamDatotek
//////////////////////////////////////////////////////////////////////

class RSeznamDatotek : public RUrejenSeznam<RNizU>
{
public:
    virtual T_INT Primerjaj(const RNizU &prvi, const RNizU &drugi) const;
};


//////////////////////////////////////////////////////////////////////
// T_PREVOD
//////////////////////////////////////////////////////////////////////

struct T_PREVOD
{
    inline T_PREVOD();

    RNizU besedilo;             // besedilo niza za prevajanje
    RNizU prevod;               // prevod niza
    RSeznamDatotek datoteke;    // datoteke in vrstice, kjer se niz za prevajanje pojavlja
    T_BOOL je_uganjen;          // ali je prevod uganjen
};


//////////////////////////////////////////////////////////////////////
// RSeznamPrevodov
//////////////////////////////////////////////////////////////////////

class RSeznamPrevodov : public RUrejenSeznam<T_PREVOD>
{
public:
    RSeznamPrevodov();

    inline T_VOID Dodaj(TK_UNICODE besedilo, T_UINT dolzina_bes, TK_UNICODE datoteka, T_UINT dolzina_dat);
    T_VOID Dodaj(TK_UNICODE besedilo, T_UINT dolzina_bes, TK_UNICODE prevod, T_UINT dolzina_pre, TK_UNICODE datoteka, T_UINT dolzina_dat);
    T_STANJE ShraniVDatoteko(TK_NIZ ime_datoteke);
    T_STANJE NaloziIzDatoteke(TK_NIZ ime_datoteke);
    T_ROCICA NajdiPrevod(TK_UNICODE besedilo, T_UINT dolzina) const;

    virtual T_INT Primerjaj(const T_PREVOD &prvi, const T_PREVOD &drugi) const;

    static TK_UNICODE VrniImeJezika(T_KODAJEZIKA2 jezik);
    static TK_UNICODE VrniImeDrzave(T_KODAJEZIKA2 jezik);
    static T_KODAJEZIKA2 VrniKodoJezika(TK_UNICODE jezik, TK_UNICODE drzava);

protected:
    inline T_STANJE ZapisiElement(RTokPodatkov &izhod, TK_UNICODE id, T_UINT dolzina_id, TK_UNICODE niz, T_UINT dolzina_niz);
    inline static T_VOID OdscitiPosebneZnake(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina = T_UINT_MAX);

public:
    T_KODNASTRAN kodna_stran;
    T_KODAJEZIKA2 koda_jezika;

    RNiz ime_datoteke;
};


//////////////////////////////////////////////////////////////////////
// Metode inline
//////////////////////////////////////////////////////////////////////

inline T_PREVOD::T_PREVOD()
{
    je_uganjen = FALSE;
}


inline T_VOID RSeznamPrevodov::Dodaj(TK_UNICODE besedilo, T_UINT dolzina_bes, TK_UNICODE datoteka, T_UINT dolzina_dat)
{
    Dodaj(besedilo, dolzina_bes, NULL, 0, datoteka, dolzina_dat);
}


inline T_STANJE RSeznamPrevodov::ZapisiElement(RTokPodatkov &izhod, TK_UNICODE id, T_UINT dolzina_id, TK_UNICODE niz, T_UINT dolzina_niz)
{
    assert(id || !dolzina_id);
    assert(niz || !dolzina_niz);

    RNizU vrstica_uni;
    T_UINT i;
    T_STANJE stanje;

    vrstica_uni.Nastavi(id, dolzina_id);
    vrstica_uni += L" \"";
    for (i = 0; i < dolzina_niz; i++) {
        if (niz[i] == L'\\')
            vrstica_uni += L"\\\\";
        else if (niz[i] == L'\n')
            vrstica_uni += L"\\n\"\n\"";
        else if (niz[i] == L'\t')
            vrstica_uni += L"\\t";
        else if (niz[i] == L'\"')
            vrstica_uni += L"\\\"";
        else
            vrstica_uni += niz[i];
    }
    vrstica_uni += L"\"\n";

    stanje = izhod.VrniStanje();
    if (NEUSPESNO(stanje)) return stanje;
    izhod.ZapisiNiz(vrstica_uni.Vrni(), vrstica_uni.VrniDolzino(), kodna_stran);

    return A_OK;
}


inline T_VOID RSeznamPrevodov::OdscitiPosebneZnake(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina)
{
    T_UINT i;

    for (i = 0; i < dolzina; i++) {
        if (vhod[i] == L'\\') {
            if (i + 1 < dolzina) {
                switch(vhod[i + 1]) {
                    case L'\\': izhod += L'\\'; i++; continue;
                    case L'n':  izhod += L'\n'; i++; continue;
                    case L't':  izhod += L'\t'; i++; continue;
                    case L'\"': izhod += L'\"'; i++; continue;
                    default:    break;
                }
            }
        }

        izhod += vhod[i];
    }
}
