/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


class RAnalizatorHTML2 : public RAnalizator {
public:
    RAnalizatorHTML2();
    virtual ~RAnalizatorHTML2();

protected:
    virtual T_STANJE Obdelaj(T_BOOL prevedi);

protected:
};


RAnalizatorHTML2::RAnalizatorHTML2() : RAnalizator()
{
}


RAnalizatorHTML2::~RAnalizatorHTML2()
{
}


T_STANJE RAnalizatorHTML2::Obdelaj(T_BOOL prevedi)
{
    T_BOOL ima_BOM = TRUE;
    T_KODNASTRAN kodna_stran;
    T_UINT st_vrstice;
    RHTMLDokument dokument;
    RNizU vrstica, datoteka, besedilo_zetoni, besedilo;
    RHTMLAnalizator analizator;
    const RHTMLZetonBesedilo *zeton;
    RDinTabela<const RHTMLZeton*> tabela_zetonov;

    // Doloimo kodno tabelo vhodne datoteke.
    kodna_stran = vhod.DolociKodnoStran();
    if (kodna_stran != KODNASTRAN_UTF7 && kodna_stran != KODNASTRAN_UTF8 && kodna_stran != KODNASTRAN_UTF16) {
        kodna_stran = KODNASTRAN_UTF8;
        ima_BOM = FALSE;
    }

    // Najprej preberemo dokument v pomnilnik.
    for (st_vrstice = 0; ; st_vrstice++) {
        vhod.PreberiVrstico(vrstica, kodna_stran);
        if (vrstica.JePrazen() && vhod.JeKonec()) break;
        vrstica += L'\n';
        verify(USPESNO(dokument.Prilepi(vrstica, vrstica.VrniDolzino())));
    }
    verify(USPESNO(dokument.Zakljuci()));

    zeton = analizator.Razcleni(&dokument, NULL, 0, TRUE);
    assert(zeton);

    const RHTMLTabelaZetonov &zetoni = analizator.VrniTabeloZetonov();
    T_UINT i, n = zetoni.VrniStevilo();

    for (i = 0; i < n; i++) {
        RHTMLZeton *&zeton = zetoni[i];
        if ((zeton->vrsta == HTML_ZETON_ZAPRTI || zeton->vrsta == HTML_ZETON_ZADNJI) &&
            (((RHTMLZetonBesedilo*)zeton)->vrsta_besedila & HTML_BESEDILO_ZETONA_BESEDILO))
        {
            T_UINT j, m, zacetek, konec;
            TK_UNICODE niz;
            RRegElPresledekU re_presledek;
            RRegElPoljubenZnakU re_poljuben_znak;

            // Poiemo presledke na zaetku in koncu.
            besedilo_zetoni.Nastavi(((const RHTMLZetonBesedilo*)zeton)->besedilo.Vrni(), ((const RHTMLZetonBesedilo*)zeton)->besedilo.VrniDolzino());
            niz = besedilo_zetoni.Vrni();
            m = besedilo_zetoni.VrniDolzino();
            for (zacetek = 0; zacetek < m;) {
                if (re_presledek.PreveriUjemanje(niz, zacetek, m)) {
                    zacetek = re_presledek.interval.konec;
                } else
                    break;
            }
            for (j = konec = zacetek; j < m;) {
                if (re_presledek.PreveriUjemanje(niz, j, m)) {
                    j = re_presledek.interval.konec;
                } else if (re_poljuben_znak.PreveriUjemanje(niz, j, m)) {
                    j = konec = re_poljuben_znak.interval.konec;
                } else
                    assert(0);
            }

            if (zacetek >= konec) {
                // Praznih nizov ne prevajamo. Pogosto se pojavlja &nbsp;.
                continue;
            }

            // Zamenjamo etone. Ker ti vsebujejo kazalce, ki se naeloma spreminjajo, jih nadomestimo z zaporednimi tevilkami etonov.
            // Hkrati izloimo zaetne in konne presledke.
            niz = besedilo_zetoni.Vrni();
            besedilo.Sprazni();
            tabela_zetonov.Sprazni();
            for (j = zacetek; j < konec && niz[j];) {
                assert(niz[j] != PR_K_KOMENTAR_KONEC);
                const RHTMLZeton *zeton = RHTMLZeton::AnalizirajOznako(niz, j);

                if (zeton) {
                    // Nali smo referenco etona. Nadomestimo jo.
                    tabela_zetonov += zeton;
                    besedilo.PrilepiO(L"%%%" _L(T_UINT_PRINTF_SIZE_PREFIX) L"u%%", AOsn::locale_C, tabela_zetonov.VrniStevilo());
                } else if (niz[j] == L'%') {
                    // Znake "%" nadomestimo z "%%", da razlikujemo znak "%" od oznake etonov "%x%".
                    besedilo += L"%%";
                    j++;
                } else
                    besedilo += niz[j++];
            }

            if (prevedi) {
                T_ROCICA rocica;
                if (katalogi.VrniStevilo()) {
                    rocica = katalogi.VrniPrvega();
                    assert(rocica);

                    RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                    T_ROCICA rocica2 = prevodi.NajdiPrevod(besedilo.Vrni(), besedilo.VrniDolzino());

                    if (rocica2) {
                        // Nali smo prevod.
                        T_PREVOD &prevod = prevodi.VrniElement(rocica2);

                        m = prevod.prevod.VrniDolzino();
                        niz = prevod.prevod.Vrni();
                        ((RHTMLZetonBesedilo*)zeton)->besedilo.Sprazni();
                        ((RHTMLZetonBesedilo*)zeton)->besedilo.Prilepi(besedilo_zetoni.Vrni(), zacetek);
                        for (j = 0; j < m && niz[j];) {
                            if (niz[j] == L'%') {
                                if (j + 1 < m && niz[j + 1] == L'%') {
                                    // V prevodu smo nali niz "%%", ki ga nadomestimo z enim "%".
                                    ((RHTMLZetonBesedilo*)zeton)->besedilo += L'%';
                                    j += 2;
                                } else {
                                    // V nadomestni oznaki je zapisan indeks etona.
                                    T_UINT idx, dolzina;

                                    idx = astrtoui(niz + j + 1, m - j - 1, &dolzina, 10);
                                    if (!idx || idx > tabela_zetonov.VrniStevilo()) {
                                        // Indeks etona je napaen.
                                        _ftprintf(stderr, _T("%s : warning X0002: Prevod \"%.*ls\" vsebuje napaen indeks etona.\n"), ime_vhodne_datoteke.Vrni(), prevod.prevod.VrniDolzino(), prevod.prevod.Vrni());
                                        ((RHTMLZetonBesedilo*)zeton)->besedilo += L'%';
                                        j++;
                                    } else {
                                        // Indeks etona je pravilen. Vstavimo njegovo oznako.
                                        tabela_zetonov[idx - 1]->PrilepiOznako(((RHTMLZetonBesedilo*)zeton)->besedilo);
                                        j += 1 + dolzina;
                                        if (j < m && niz[j] == L'%')
                                            j++;
                                    }
                                }
                            // Posebnih znakov HTML ne pretvarjamo v entitete SGML, ker jih RHTMLAnalizator::Sestavi().
                            //} else if (niz[j] == L'<') {
                            //    ((RHTMLZetonBesedilo*)zeton)->besedilo += L"&lt;";
                            //    j++;
                            //} else if (niz[j] == L'>') {
                            //    ((RHTMLZetonBesedilo*)zeton)->besedilo += L"&gt;";
                            //    j++;
                            //} else if (niz[j] == L'"') {
                            //    ((RHTMLZetonBesedilo*)zeton)->besedilo += L"&quot;";
                            //    j++;
                            //} else if (niz[j] == L'&') {
                            //    ((RHTMLZetonBesedilo*)zeton)->besedilo += L"&amp;";
                            //    j++;
                            //} else if (niz[j] == 0x00a0) {
                            //    ((RHTMLZetonBesedilo*)zeton)->besedilo += L"&nbsp;";
                            //    j++;
                            } else
                                ((RHTMLZetonBesedilo*)zeton)->besedilo += niz[j++];
                        }
                        ((RHTMLZetonBesedilo*)zeton)->besedilo.Prilepi(besedilo_zetoni.Vrni() + konec, besedilo_zetoni.VrniDolzino() - konec);
                    } else {
                        // Prevoda nismo nali, zato bomo pustili kar izvorno besedilo.
                        _ftprintf(stderr, _T("%s : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), prevodi.ime_datoteke.Vrni());
                    }
                } else {
                    // Nimamo nobenega kataloga, zato pustimo samo prvotno besedilo.
                }
            } else {
                datoteka.Oblikuj(
#ifdef _UNICODE
                    L"%ls",
#else
                    L"%hs",
#endif
                    AOsn::locale_ACP, ime_vhodne_datoteke.Vrni());
                prevodi.Dodaj(
                    besedilo.Vrni(), besedilo.VrniDolzino(),
                    datoteka.Vrni(), datoteka.VrniDolzino());
            }
        }
    }

    if (prevedi) {
        RNizU koda;

        // Sestavimo prevod HTML in ga zapiemo na izhod.
        analizator.Sestavi(koda, zeton);
        if (ima_BOM) izhod.ZapisiKodnoStran(kodna_stran);
        izhod.ZapisiNiz(koda.Vrni(), koda.VrniDolzino(), kodna_stran);
    }

    analizator.Sprazni();
    dokument.Sprazni();
    return A_OK;
}


T_REZULTAT html(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorHTML2 analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext html <izhodna datoteka.po> <vhodna datoteka.html> ...\n"));
        rezultat = 1;
        goto konec1;
    }

    for (i = 2; i < argc; i++) {
        stanje = analizator.Analiziraj(argv[i]);
        if (NEUSPESNO(stanje)) {
            rezultat = 2;
            goto konec1;
        }
    }

    // Shranimo katalog.
    stanje = analizator.ShraniKatalog(argv[1]);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}


T_REZULTAT htmlp(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_STANJE stanje;
    RAnalizatorHTML2 analizator;

    if (argc < 4) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext htmlp <izhodna datoteka.html> <vhodna datoteka.html> <katalog prevodov.po>\n"));
        rezultat = 1;
        goto konec1;
    }

    // Naloimo katalog.
    stanje = analizator.NaloziKatalog(argv[3]);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

    stanje = analizator.Prevedi(argv[1], T_UINT_MAX, argv[2], T_UINT_MAX);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}
