/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


class RAnalizatorIDT : public RAnalizator {
public:
    RAnalizatorIDT();
    virtual ~RAnalizatorIDT();

protected:
    virtual T_STANJE Obdelaj(T_BOOL prevedi);

protected:
    RRegElementP<T_UZNAK>
        *re_kodna_stran;
};


RAnalizatorIDT::RAnalizatorIDT() : RAnalizator()
{
    re_kodna_stran = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElZacetekVrsticeP<T_UZNAK>(),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"0123456789"), 1),
        new RRegElKonecVrsticeP<T_UZNAK>(),
        NULL);
    assert(re_kodna_stran);
}


RAnalizatorIDT::~RAnalizatorIDT()
{
    re_kodna_stran->Unici();
}


T_STANJE RAnalizatorIDT::Obdelaj(T_BOOL prevedi)
{
    T_KODNASTRAN kodna_stran_vh, kodna_stran_iz;
    RNizU vrstica, besedilo, datoteka, ime_jezika, ime_tabele;
    T_UINT st_vrstice;
    RDinTabela<RNizU> stolpci, kljuci, vrednosti;
    RDinTabela<T_UINT> za_prevajanje;
    T_BOOL je_spremenjena;

    // Doloimo kodno tabelo vhodne datoteke.
    kodna_stran_vh = kodna_stran_iz = vhod.DolociKodnoStran();

    for (st_vrstice = 0; ; st_vrstice++) {
        vhod.PreberiVrstico(vrstica, kodna_stran_vh);
        if (vhod.JeKonec()) break;

        je_spremenjena = FALSE;

        if (st_vrstice == 0) {
            idlsplit(stolpci, vrstica);
            stolpci.SprostiOdvecne();
        } else if (st_vrstice == 1) {
            // Vrstica s tipi podatkov nas ne zanima.
        } else if (st_vrstice == 2) {
            idlsplit(kljuci, vrstica);
            kljuci.SprostiOdvecne();

            if (kljuci.VrniStevilo() && re_kodna_stran->PreveriUjemanje(kljuci[0].Vrni(), 0, kljuci[0].VrniDolzino())) {
                // Nali smo podatek o vhodni kodni strani.
                if (kodna_stran_vh != KODNASTRAN_UTF8 && kodna_stran_vh != KODNASTRAN_UTF16) {
                    kodna_stran_vh = (T_KODNASTRAN)astrtoui(kljuci[0].Vrni(), kljuci[0].VrniDolzino(), NULL, 10);
                    if (prevedi) {
                        T_ROCICA rocica;

                        kodna_stran_iz = kodna_stran_vh;
                        if (katalogi.VrniStevilo()) {
                            rocica = katalogi.VrniPrvega();
                            assert(rocica);

                            RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                            T_ROCICA rocica2 = prevodi.NajdiPrevod(kljuci[0].Vrni(), kljuci[0].VrniDolzino());

                            if (rocica2) {
                                // Nali smo prevod.
                                T_PREVOD &prevod = prevodi.VrniElement(rocica2);
                                kodna_stran_iz = (T_KODNASTRAN)astrtoui(prevod.prevod.Vrni(), prevod.prevod.VrniDolzino(), NULL, 10);
                                kljuci[0].Sprazni();
                                idtescape(kljuci[0], prevod.prevod.Vrni(), prevod.prevod.VrniDolzino());
                                vrednosti = kljuci;
                                je_spremenjena = TRUE;
                            } else {
                                // Prevoda nismo nali, zato bomo pustili kar izvorno besedilo.
                                _ftprintf(stderr, _T("%s(%Iu) : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice + 1, prevodi.ime_datoteke.Vrni());
                            }
                        } else {
                            // Nimamo nobenega kataloga, zato pustimo samo prvotno kodno stran.
                        }
                    } else {
                        // Tudi kodna tabela je niz za prevajanje.
                        datoteka.Oblikuj(
#ifdef _UNICODE
                            L"%ls:%Iu",
#else
                            L"%hs:%Iu",
#endif
                            AOsn::locale_ACP, ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                        prevodi.Dodaj(
                            kljuci[0].Vrni(), kljuci[0].VrniDolzino(),
                            datoteka.Vrni(), datoteka.VrniDolzino());
                    }
                }

                ime_tabele = kljuci[1];
            } else
                ime_tabele = kljuci[0];

            RDatotekaIDT::VrniStolpceZaPrevajanje(za_prevajanje, ime_tabele.Vrni(), ime_tabele.VrniDolzino(), stolpci);
            za_prevajanje.SprostiOdvecne();
        } else {
            T_UINT i, n;

            if (za_prevajanje.VrniStevilo()) {
                idlsplit(vrednosti, vrstica);

                for (i = 0, n = za_prevajanje.VrniStevilo(); i < n; i++) {
                    if (za_prevajanje[i] < vrednosti.VrniStevilo()) {
                        RNizU &besedilo = vrednosti[za_prevajanje[i]];
                        if (!besedilo.JePrazen()) {
                            if (prevedi) {
                                T_ROCICA rocica;
                                if (katalogi.VrniStevilo()) {
                                    rocica = katalogi.VrniPrvega();
                                    assert(rocica);

                                    RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                                    T_ROCICA rocica2 = prevodi.NajdiPrevod(besedilo.Vrni(), besedilo.VrniDolzino());

                                    if (rocica2) {
                                        // Nali smo prevod.
                                        T_PREVOD &prevod = prevodi.VrniElement(rocica2);
                                        besedilo.Sprazni();
                                        idtescape(besedilo, prevod.prevod.Vrni(), prevod.prevod.VrniDolzino());
                                        je_spremenjena = TRUE;
                                    } else {
                                        // Prevoda nismo nali, zato bomo pustili kar izvorno besedilo.
                                        _ftprintf(stderr, _T("%s(%Iu) : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice + 1, prevodi.ime_datoteke.Vrni());
                                    }
                                } else {
                                    // Nimamo nobenega kataloga, zato pustimo samo prvotno besedilo.
                                }
                            } else {
                                datoteka.Oblikuj(
#ifdef _UNICODE
                                    L"%ls:%Iu",
#else
                                    L"%hs:%Iu",
#endif
                                    AOsn::locale_ACP, ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                                prevodi.Dodaj(
                                    besedilo.Vrni(), besedilo.VrniDolzino(),
                                    datoteka.Vrni(), datoteka.VrniDolzino());
                            }
                        }
                    }
                }
            }
        }

        // Prepiemo vrstico na izhod.
        if (prevedi) {
            if (je_spremenjena)
                idlmerge(vrstica, vrednosti);

            vrstica += L'\n';
            izhod.ZapisiNiz(vrstica.Vrni(), vrstica.VrniDolzino(), kodna_stran_iz);
        }
    }

    return A_OK;
}


T_REZULTAT idt(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorIDT analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext idt <izhodna datoteka.po> <vhodna datoteka.idt> ...\n"));
        rezultat = 1;
        goto konec1;
    }

    for (i = 2; i < argc; i++) {
        stanje = analizator.Analiziraj(argv[i]);
        if (NEUSPESNO(stanje)) {
            rezultat = 2;
            goto konec1;
        }
    }

    // Shranimo katalog.
    stanje = analizator.ShraniKatalog(argv[1]);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}


T_REZULTAT idtp(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_STANJE stanje;
    RAnalizatorIDT analizator;

    if (argc < 4) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext idtp <izhodna datoteka.idt> <vhodna datoteka.idt> <katalog prevodov.po>\n"));
        rezultat = 1;
        goto konec1;
    }

    // Naloimo katalog.
    stanje = analizator.NaloziKatalog(argv[3]);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

    stanje = analizator.Prevedi(argv[1], T_UINT_MAX, argv[2], T_UINT_MAX);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}


T_REZULTAT idta(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_STANJE stanje;
    RDatotekaIDT vhod, prevod;
    RSeznamPrevodov prevodi;

    if (argc < 4) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext idta <izhodna datoteka.po> <vhodna datoteka.idt> <prevod vhodne datoteke.idt>\n"));
        rezultat = 1;
        goto konec1;
    }

    stanje = vhod.NaloziIzDatoteke(argv[2]);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0002: Napaka pri odpiranju vhodne datoteke (koda %#010I32x).\n"), argv[2], stanje);
        rezultat = 2;
        goto konec1;
    }

    stanje = prevod.NaloziIzDatoteke(argv[3]);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0003: Napaka pri odpiranju prevoda vhodne datoteke (koda %#010I32x).\n"), argv[3], stanje);
        rezultat = 2;
        goto konec1;
    }

    stanje = vhod.ZgradiSeznamPrevodov(prevodi, prevod);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0004: Pri izgradnji kataloga prevodov je prilo do napake (koda %#010I32x). Sta tabeli zdruljivi?\n"), argv[2], stanje);
        rezultat = 2;
        goto konec1;
    }

    // Shranimo katalog.
    stanje = prevodi.ShraniVDatoteko(argv[1]);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0005: Napaka pri shranjevanju v datoteko (koda %#010I32x).\n"), argv[1], stanje);
        rezultat = 3;
        goto konec1;
    }

konec1:
    return rezultat;
}
