/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


class RAnalizatorMC : public RAnalizator {
public:
    RAnalizatorMC();
    virtual ~RAnalizatorMC();

protected:
    virtual T_STANJE Obdelaj(T_BOOL prevedi);

protected:
    RRegElementP<T_UZNAK>
        *re_language_names_ime,
        *re_language_names,
        *re_language_ime,
        *re_language;
};


RAnalizatorMC::RAnalizatorMC() : RAnalizator()
{
    re_language_names_ime = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"), 1);
    assert(re_language_names_ime);

    re_language_names = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElNizP<T_UZNAK>(L"LanguageNames"),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElZnakP<T_UZNAK>(L'='),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElZnakP<T_UZNAK>(L'('),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        re_language_names_ime,
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElZnakP<T_UZNAK>(L'='),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPoljubenZnakP<T_UZNAK>()),
        NULL);
    assert(re_language_names);

    re_language_ime = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"), 1);
    assert(re_language_ime);

    re_language = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElNizP<T_UZNAK>(L"Language"),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElZnakP<T_UZNAK>(L'='),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        re_language_ime,
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPoljubenZnakP<T_UZNAK>()),
        NULL);
    assert(re_language);
}


RAnalizatorMC::~RAnalizatorMC()
{
    re_language_names->Unici();
    re_language->Unici();
}


T_STANJE RAnalizatorMC::Obdelaj(T_BOOL prevedi)
{
    T_KODNASTRAN kodna_stran;
    RNizU vrstica, vrstica2, besedilo, datoteka, ime_jezika;
    T_UINT st_vrstice;
    T_BOOL je_besedilo = FALSE;

    // Doloimo kodno tabelo vhodne datoteke.
    kodna_stran = vhod.DolociKodnoStran();
    if (prevedi) izhod.ZapisiKodnoStran(kodna_stran);

    for (st_vrstice = 0; ; st_vrstice++) {
        vhod.PreberiVrstico(vrstica, kodna_stran);
        if (vhod.JeKonec()) break;

        // Odstranimo odvene presledke na zaetku in koncu vrstice.
        vrstica.Poravnaj();

        // Prepiemo vrstico na izhod.
        if (prevedi) {
            vrstica2  = vrstica;
            vrstica2 += L'\n';
            izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
        }

        // Preskoimo prazne vrstice in komentarje
        if (vrstica.JePrazen() ||
            vrstica.VrniZnak(0) == L';') continue;

        if (re_language_names->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino(), RREGIZR_NE_LOCUJ_VELIKIH_MALIH)) {
            // Nali smo vrstico LanguageNames = xy.

            if (ime_jezika.JePrazen()) {
                // To je primarni jezik datoteke.
                ime_jezika.Nastavi(vrstica.Vrni() + re_language_names_ime->interval.zacetek, re_language_names_ime->interval.VrniDolzino());

                if (prevedi) {
                    // Dodamo e svoje jezike.
                    T_ROCICA rocica;

                    for (rocica = katalogi.VrniPrvega(); rocica; rocica = katalogi.VrniNaslednjega(rocica)) {
                        RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                        T_UZNAK iso639[AOSN_KODA_ISO639_MAX];
                        T_UZNAK iso3166[AOSN_KODA_ISO3166_MAX];

                        alangtoiso639 (prevodi.koda_jezika, iso639 );
                        alangtoiso3166(prevodi.koda_jezika, iso3166);
                        vrstica2.Oblikuj(L"LanguageNames=(%.*ls_%.*ls=%#Ix:MSG%05Ix)\n", AOsn::locale_C,
                            AOSN_KODA_ISO639_MAX,  iso639,
                            AOSN_KODA_ISO3166_MAX, iso3166,
                            (T_UINT)prevodi.koda_jezika,
                            (T_UINT)prevodi.koda_jezika);
                        izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
                    }
                }
            }
        }

        if (re_language->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino(), RREGIZR_NE_LOCUJ_VELIKIH_MALIH)) {
            // Nali smo vrstico Language = xy.
            if (!astrnicmp(ime_jezika.Vrni(), ime_jezika.VrniDolzino(), vrstica.Vrni() + re_language_ime->interval.zacetek, re_language_ime->interval.VrniDolzino())) {
                // To je primarni jezik datoteke.
                besedilo.Sprazni();
                datoteka.Oblikuj(
#ifdef _UNICODE
                    L"%ls:%Iu",
#else
                    L"%hs:%Iu",
#endif
                    AOsn::locale_ACP, ime_vhodne_datoteke.Vrni(), st_vrstice + 2);
                je_besedilo = TRUE;
            }
        } else if (je_besedilo && !vrstica.Primerjaj(L".")) {
            // Nali smo konec besedila.

            if (!besedilo.JePrazen()) {
                if (prevedi) {
                    T_ROCICA rocica;
                    if (katalogi.VrniStevilo()) {
                        for (rocica = katalogi.VrniPrvega(); rocica; rocica = katalogi.VrniNaslednjega(rocica)) {
                            RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                            T_ROCICA rocica2 = prevodi.NajdiPrevod(besedilo.Vrni(), besedilo.VrniDolzino());
                            T_UZNAK iso639[AOSN_KODA_ISO639_MAX];
                            T_UZNAK iso3166[AOSN_KODA_ISO3166_MAX];

                            alangtoiso639 (prevodi.koda_jezika, iso639 );
                            alangtoiso3166(prevodi.koda_jezika, iso3166);

                            if (rocica2) {
                                // Nali smo prevod.
                                T_PREVOD &prevod = prevodi.VrniElement(rocica2);
                                vrstica = prevod.prevod;
                            } else {
                                // Prevoda nismo nali, zato bomo namesto prevoda zapisali kar izvorno besedilo.
                                _ftprintf(stderr, _T("%s(%Iu) : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice + 2, prevodi.ime_datoteke.Vrni());
                                vrstica = besedilo;
                            }

                            // Zapiemo prevod.
                            vrstica2.Oblikuj(
                                L"Language=%.*ls_%.*ls\n"
                                    L"%ls\n"
                                    L".\n", AOsn::locale_C,
                                    AOSN_KODA_ISO639_MAX,  iso639,
                                    AOSN_KODA_ISO3166_MAX, iso3166,
                                    vrstica.Vrni());
                            izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
                        }
                    } else {
                        // Nimamo nobenega kataloga, zato pustimo samo prvotno besedilo.
                    }
                } else
                    prevodi.Dodaj(besedilo.Vrni(), besedilo.VrniDolzino(), datoteka.Vrni(), datoteka.VrniDolzino());
            }

            je_besedilo = FALSE;
        } else if (je_besedilo) {
            // Smo sredi besedila.
            if (!besedilo.JePrazen()) besedilo += L'\n';
            besedilo += vrstica;
        }
    }

    return A_OK;
}


T_REZULTAT mc(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezutat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorMC analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext mc <izhodna datoteka.po> <vhodna datoteka.rc> ...\n"));
        rezutat = 1;
        goto konec1;
    }

    for (i = 2; i < argc; i++) {
        stanje = analizator.Analiziraj(argv[i], T_UINT_MAX);
        if (NEUSPESNO(stanje)) {
            rezutat = 2;
            goto konec1;
        }
    }

    // Shranimo katalog.
    stanje = analizator.ShraniKatalog(argv[1]);
    if (NEUSPESNO(stanje)) {
        rezutat = 2;
        goto konec1;
    }

konec1:
    return rezutat;
}


T_REZULTAT mcp(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorMC analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext mcp <izhodna datoteka.rc> <vhodna datoteka.rc> [<katalog prevodov.po> ...]\n"));
        rezultat = 1;
        goto konec1;
    }

    // Naloimo kataloge.
    for (i = 3; i < argc; i++) {
        stanje = analizator.NaloziKatalog(argv[i]);
        if (NEUSPESNO(stanje)) {
            rezultat = 2;
            goto konec1;
        }
    }

    stanje = analizator.Prevedi(argv[1], T_UINT_MAX, argv[2], T_UINT_MAX);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}

