/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


class RAnalizatorRC : public RAnalizator {
public:
    RAnalizatorRC();
    virtual ~RAnalizatorRC();

protected:
    virtual T_STANJE Obdelaj(T_BOOL prevedi);

protected:
    RRegElementP<T_UZNAK>
        *re_zacetek_bloka,
        *re_konec_bloka,
        *re_zacetek_oznake,
        *re_presledki,
        *re_zapis_niz,
        *re_zapis_simbol;
};


inline T_VOID rcunescape(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina)
{
    assert(vhod || !dolzina);
    T_UINT i;

    // Pretvorimo posebne znake v besedilu.
    for (i = 0; i < dolzina && vhod[i]; ) {
        if (vhod[i] == L'\\' && i + 1 < dolzina) {
            switch (vhod[++i]) {
                case L'n':  izhod += L'\n'; i++; break;
                case L't':  izhod += L'\t'; i++; break;
                case L'\\': izhod += L'\\'; i++; break;
                default:    izhod += L'\\';
            }
        } else if (vhod[i] == L'"' && i + 1 < dolzina && vhod[i + 1] == L'"') {
            izhod += L'"';
            i += 2;
        } else
            izhod += vhod[i++];
    }
}


inline T_VOID rcescape(RNizU &izhod, TK_UNICODE vhod, T_UINT dolzina)
{
    assert(vhod || !dolzina);
    T_UINT i;

    // Pretvorimo posebne znake v besedilu.
    for (i = 0; i < dolzina && vhod[i]; i++) {
        switch (vhod[i]) {
            case L'\n': izhod += L"\\n";  break;
            case L'\t': izhod += L"\\t";  break;
            case L'\"': izhod += L"\"\""; break;
            default:    izhod += vhod[i];
        }
    }
}


RAnalizatorRC::RAnalizatorRC() : RAnalizator()
{
    re_zacetek_bloka = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElNizP<T_UZNAK>(L"//{{LOCALIZE_RESOURCES_START}}"),
        NULL);
    assert(re_zacetek_bloka);

    re_konec_bloka = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElNizP<T_UZNAK>(L"//{{LOCALIZE_RESOURCES_END}}"),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        NULL);
    assert(re_konec_bloka);

    re_zacetek_oznake = new RRegElZaporedjeP<T_UZNAK>(
        new RRegElNizP<T_UZNAK>(L"__"),
        new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>()),
        new RRegElZnakP<T_UZNAK>(L'('),
        NULL);
    assert(re_zacetek_oznake);

    re_presledki = new RRegElPonavljanjeP<T_UZNAK>(new RRegElPresledekP<T_UZNAK>());
    assert(re_presledki);

    re_zapis_niz = new RRegElPonavljanjeP<T_UZNAK>(
        new RRegElVejitevP<T_UZNAK>(
            new RRegElNizP<T_UZNAK>(L"\"\""),
            new RRegElMaskaP<T_UZNAK>(L"\"", T_UINT_MAX, TRUE),
            NULL));
    assert(re_zapis_niz);

    re_zapis_simbol = new RRegElPonavljanjeP<T_UZNAK>(new RRegElMaskaP<T_UZNAK>(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"), 1);
    assert(re_zapis_simbol);
}


RAnalizatorRC::~RAnalizatorRC()
{
    re_zapis_simbol->Unici();
    re_zapis_niz->Unici();
    re_presledki->Unici();
    re_zacetek_oznake->Unici();
    re_konec_bloka->Unici();
    re_zacetek_bloka->Unici();
}


T_STANJE RAnalizatorRC::Obdelaj(T_BOOL prevedi)
{
    T_KODNASTRAN kodna_stran;
    RNizU vrstica, vrstica2, besedilo, datoteka;
    T_ODMIK zacetek_bloka = 0;
    T_UINT zacetek, konec, st_vrstice, zacetek_st_vrstice, st_vrstice2 = 0;
    T_BOOL je_blok = FALSE;

    // Doloimo kodno tabelo vhodne datoteke.
    kodna_stran = vhod.DolociKodnoStran();
    if (prevedi)
        izhod.ZapisiKodnoStran(kodna_stran);

    for (st_vrstice = zacetek_st_vrstice = 0; ; st_vrstice++) {
        vhod.PreberiVrstico(vrstica, kodna_stran);
        if (vhod.JeKonec()) break;

        if (re_zacetek_bloka->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino())) {
            // Nali smo zaetek bloka za prevajanje.
            je_blok = TRUE;
            zacetek_bloka = vhod.VrniOdmik();
            zacetek_st_vrstice = st_vrstice;
        } else if (re_konec_bloka->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino())) {
            if (prevedi) {
                T_ROCICA rocica, rocica2;

                for (rocica = katalogi.VrniPrvega(); rocica; rocica = katalogi.VrniNaslednjega(rocica)) {
                    RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);
                    verify(vhod.Prestavi(zacetek_bloka) == zacetek_bloka);

                    for (st_vrstice2 = zacetek_st_vrstice + 1; ; st_vrstice2++) {
                        vhod.PreberiVrstico(vrstica, kodna_stran);
                        if (vhod.JeKonec()) break;

                        if (re_konec_bloka->PreveriUjemanje(vrstica.Vrni(), 0, vrstica.VrniDolzino())) {
                            // Nali smo konec bloka.
                            break;
                        }

                        for (zacetek = 0, konec = vrstica.VrniDolzino(); zacetek < konec;) {
                            if (re_zacetek_oznake->PreveriUjemanje(vrstica.Vrni(), zacetek, konec)) {
                                // Prepiemo vsebino vrstice do zaetka oznake.
                                izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_zacetek_oznake->interval.zacetek - zacetek, kodna_stran);
                                zacetek = re_zacetek_oznake->interval.konec;

                                while (zacetek < konec) {
                                    // Preskoimo presledke.
                                    verify(re_presledki->PreveriUjemanje(vrstica.Vrni(), zacetek, konec));
                                    izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_presledki->interval.konec - zacetek, kodna_stran);
                                    zacetek = re_presledki->interval.konec;

                                    if (vrstica.VrniZnak(zacetek) == L')') {
                                        zacetek++;
                                        break;
                                    } else if (vrstica.VrniZnak(zacetek) == L'"') {
                                        // V oznaki je niz.
                                        izhod.ZapisiNiz(L"\"", T_UINT_MAX, kodna_stran);
                                        verify(re_zapis_niz->PreveriUjemanje(vrstica.Vrni(), ++zacetek, konec));
                                        if (vrstica.VrniZnak(re_zapis_niz->interval.konec) != L'"') {
                                            _ftprintf(stderr, _T("%s(%Iu) : error X0006: Sintaktina napaka. Niz ni pravilno zakljuen.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice2 + 1);
                                            return AOSN_NAPAKA_ISKANJE;
                                        }

                                        // Nali smo niz za prevajanje.
                                        vrstica2.Sprazni();
                                        if (re_zapis_niz->interval.zacetek < re_zapis_niz->interval.konec) {
                                            besedilo.Sprazni();
                                            rcunescape(besedilo, vrstica.Vrni() + re_zapis_niz->interval.zacetek, re_zapis_niz->interval.VrniDolzino());
                                            rocica2 = prevodi.NajdiPrevod(besedilo.Vrni(), besedilo.VrniDolzino());
                                            if (rocica2) {
                                                // Nali smo prevod.
                                                T_PREVOD &prevod = prevodi.VrniElement(rocica2);
                                                rcescape(vrstica2, prevod.prevod.Vrni(), prevod.prevod.VrniDolzino());
                                            } else {
                                                // Prevoda nismo nali, zato bomo namesto prevoda zapisali kar izvorno besedilo.
                                                _ftprintf(stderr, _T("%s(%Iu) : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice2 + 1, prevodi.ime_datoteke.Vrni());
                                                rcescape(vrstica2, besedilo.Vrni(), besedilo.VrniDolzino());
                                            }
                                        }

                                        vrstica2 += L'\"';
                                        izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
                                        zacetek = re_zapis_niz->interval.konec + 1;
                                    } else if (re_zapis_simbol->PreveriUjemanje(vrstica.Vrni(), zacetek, konec)) {
                                        // V oznaki je simbol za prevajanje.
                                        vrstica2.Sprazni();
                                        if (re_zapis_simbol->interval.zacetek < re_zapis_simbol->interval.konec) {
                                            besedilo.Sprazni();
                                            rcunescape(besedilo, vrstica.Vrni() + re_zapis_simbol->interval.zacetek, re_zapis_simbol->interval.VrniDolzino());
                                            rocica2 = prevodi.NajdiPrevod(besedilo.Vrni(), besedilo.VrniDolzino());
                                            if (rocica2) {
                                                // Nali smo prevod.
                                                T_PREVOD &prevod = prevodi.VrniElement(rocica2);
                                                vrstica2 = prevod.prevod;
                                            } else {
                                                // Prevoda nismo nali, zato bomo namesto prevoda zapisali kar izvorno besedilo.
                                                _ftprintf(stderr, _T("%s(%Iu) : warning X0001: V katalogu prevodov %s manjka prevod besedila.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice2 + 1, prevodi.ime_datoteke.Vrni());
                                                vrstica2 = besedilo;
                                            }

                                            izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
                                            zacetek = re_zapis_simbol->interval.konec;
                                        } else {
                                            _ftprintf(stderr, _T("%s(%Iu) : error X0007: Sintaktina napaka. Neznan simbol.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice2 + 1);
                                            return AOSN_NAPAKA_ISKANJE;
                                        }
                                    } else {
                                        izhod.ZapisiNiz(vrstica.Vrni() + zacetek, 1, kodna_stran);
                                        zacetek++;
                                    }
                                }
                            } else {
                                izhod.ZapisiNiz(vrstica.Vrni() + zacetek, 1, kodna_stran);
                                zacetek++;
                            }
                        }
                        izhod.ZapisiNiz(L"\n", T_UINT_MAX, kodna_stran);
                    }
                }

                st_vrstice = st_vrstice2;
            }

            je_blok = FALSE;
        } else if (je_blok) {
            for (zacetek = 0, konec = vrstica.VrniDolzino(); zacetek < konec;) {
                if (re_zacetek_oznake->PreveriUjemanje(vrstica.Vrni(), zacetek, konec)) {
                    // Prepiemo vsebino vrstice do zaetka oznake.
                    if (prevedi) izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_zacetek_oznake->interval.zacetek - zacetek, kodna_stran);
                    zacetek = re_zacetek_oznake->interval.konec;

                    while (zacetek < konec) {
                        // Preskoimo presledke.
                        if (re_presledki->PreveriUjemanje(vrstica.Vrni(), zacetek, konec)) {
                            if (prevedi) izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_presledki->interval.konec - zacetek, kodna_stran);
                            zacetek = re_presledki->interval.konec;
                        }

                        if (vrstica.VrniZnak(zacetek) == L')') {
                            zacetek++;
                            break;
                        } else if (vrstica.VrniZnak(zacetek) == L'"') {
                            // V oznaki je niz.
                            if (prevedi) izhod.ZapisiNiz(L"\"", T_UINT_MAX, kodna_stran);
                            verify(re_zapis_niz->PreveriUjemanje(vrstica.Vrni(), ++zacetek, konec));
                            if (vrstica.VrniZnak(re_zapis_niz->interval.konec) != L'"') {
                                _ftprintf(stderr, _T("%s(%Iu) : error X0006: Sintaktina napaka. Niz ni pravilno zakljuen.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                                return AOSN_NAPAKA_BRANJE;
                            }

                            // Nali smo niz za prevajanje.
                            if (prevedi) {
                                izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_zapis_niz->interval.konec + 1 - zacetek, kodna_stran);
                            } else {
                                besedilo.Sprazni();
                                rcunescape(besedilo, vrstica.Vrni() + re_zapis_niz->interval.zacetek, re_zapis_niz->interval.VrniDolzino());
                                datoteka.Oblikuj(
#ifdef _UNICODE
                                    L"%ls:%Iu",
#else
                                    L"%hs:%Iu",
#endif
                                    AOsn::locale_ACP, ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                                prevodi.Dodaj(
                                    besedilo.Vrni(), besedilo.VrniDolzino(),
                                    datoteka.Vrni(), datoteka.VrniDolzino());
                            }
                            zacetek = re_zapis_niz->interval.konec + 1;
                        } else if (re_zapis_simbol->PreveriUjemanje(vrstica.Vrni(), zacetek, konec)) {
                            // V oznaki je simbol za prevajanje.
                            if (zacetek < re_zapis_simbol->interval.konec) {
                                if (prevedi) {
                                    izhod.ZapisiNiz(vrstica.Vrni() + zacetek, re_zapis_simbol->interval.konec - zacetek, kodna_stran);
                                } else {
                                    datoteka.Oblikuj(
#ifdef _UNICODE
                                        L"%ls:%Iu",
#else
                                        L"%hs:%Iu",
#endif
                                        AOsn::locale_ACP, ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                                    prevodi.Dodaj(
                                        vrstica.Vrni() + re_zapis_simbol->interval.zacetek, re_zapis_simbol->interval.VrniDolzino(),
                                        datoteka.Vrni(), datoteka.VrniDolzino());
                                }
                                zacetek = re_zapis_simbol->interval.konec;
                            } else {
                                _ftprintf(stderr, _T("%s(%Iu) : error X0007: Sintaktina napaka. Neznan simbol.\n"), ime_vhodne_datoteke.Vrni(), st_vrstice + 1);
                                return AOSN_NAPAKA_BRANJE;
                            }
                        } else {
                            if (prevedi) izhod.ZapisiNiz(vrstica.Vrni() + zacetek, 1, kodna_stran);
                            zacetek++;
                        }
                    }
                } else {
                    if (prevedi) izhod.ZapisiNiz(vrstica.Vrni() + zacetek, 1, kodna_stran);
                    zacetek++;
                }
            }
            if (prevedi) izhod.ZapisiNiz(L"\n", T_UINT_MAX, kodna_stran);
        } else if (prevedi) {
            vrstica2  = vrstica;
            vrstica2 += L'\n';
            izhod.ZapisiNiz(vrstica2.Vrni(), vrstica2.VrniDolzino(), kodna_stran);
        }
    }

    return A_OK;
}


T_REZULTAT rc(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorRC analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext rc <izhodna datoteka.po> <vhodna datoteka.rc> ...\n"));
        rezultat = 1;
        goto konec1;
    }

    for (i = 2; i < argc; i++) {
        stanje = analizator.Analiziraj(argv[i], T_UINT_MAX);
        if (NEUSPESNO(stanje)) {
            rezultat = 2;
            goto konec1;
        }
    }

    // Shranimo katalog.
    stanje = analizator.ShraniKatalog(argv[1]);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}


T_REZULTAT rcp(T_UINT argc, TK_NIZ argv[])
{
    T_REZULTAT rezultat = 0;
    T_UINT i;
    T_STANJE stanje;
    RAnalizatorRC analizator;

    if (argc < 3) {
        _ftprintf(stderr, _T("Uporaba: rcxgettext rcp <izhodna datoteka.rc> <vhodna datoteka.rc> [<katalog prevodov.po> ...]\n"));
        rezultat = 1;
        goto konec1;
    }

    // Naloimo kataloge.
    for (i = 3; i < argc; i++) {
        stanje = analizator.NaloziKatalog(argv[i]);
        if (NEUSPESNO(stanje)) {
            rezultat = 2;
            goto konec1;
        }
    }

    stanje = analizator.Prevedi(argv[1], T_UINT_MAX, argv[2], T_UINT_MAX);
    if (NEUSPESNO(stanje)) {
        rezultat = 2;
        goto konec1;
    }

konec1:
    return rezultat;
}
