/*
    Copyright 1991-2015 Amebis

    This file is part of rcxgettext.

    rcxgettext is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rcxgettext is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rcxgettext. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StdAfx.h"


RAnalizator::RAnalizator()
{
}


T_STANJE RAnalizator::Analiziraj(TK_NIZ vhodna_datoteka, T_UINT dolzina_vh)
{
    assert(vhodna_datoteka || !dolzina_vh);

    T_STANJE stanje;

    // Odpremo vhodno datoteko.
    ime_vhodne_datoteke.Nastavi(vhodna_datoteka, dolzina_vh);
    stanje = vhod.Odpri(ime_vhodne_datoteke, RDAT_BRANJE | RDAT_SH_BERI | RDAT_BINARNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0002: Napaka pri odpiranju vhodne datoteke (koda %#010I32x).\n"), ime_vhodne_datoteke.Vrni(), stanje);
        goto napaka1;
    }

    stanje = Obdelaj(FALSE);

    verify(USPESNO(vhod.Zapri()));
napaka1:
    return stanje;
}


T_STANJE RAnalizator::Prevedi(TK_NIZ izhodna_datoteka, T_UINT dolzina_iz, TK_NIZ vhodna_datoteka, T_UINT dolzina_vh)
{
    assert(izhodna_datoteka || !dolzina_iz);
    assert(vhodna_datoteka  || !dolzina_vh);

    T_STANJE stanje;

    // Odpremo izhodno datoteko.
    ime_izhodne_datoteke.Nastavi(izhodna_datoteka, dolzina_iz);
    stanje = izhod.Odpri(ime_izhodne_datoteke, RDAT_PISANJE | RDAT_SH_BERI | RDAT_USTVARI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0009: Napaka pri odpiranju izhodne datoteke (koda %#010I32x).\n"), ime_izhodne_datoteke.Vrni(), stanje);
        goto napaka1;
    }

    // Odpremo vhodno datoteko.
    ime_vhodne_datoteke.Nastavi(vhodna_datoteka, dolzina_vh);
    stanje = vhod.Odpri(ime_vhodne_datoteke, RDAT_BRANJE | RDAT_SH_BERI | RDAT_TEKSTOVNO | RDAT_DOSTOP_ZAPOREDEN);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0002: Napaka pri odpiranju vhodne datoteke (koda %#010I32x).\n"), ime_vhodne_datoteke.Vrni(), stanje);
        goto napaka2;
    }

    stanje = Obdelaj(TRUE);

    verify(USPESNO(vhod.Zapri()));
napaka2:
    verify(USPESNO(izhod.Zapri()));
    if (NEUSPESNO(stanje)) verify(USPESNO(AOsn::ZbrisiDatoteko(ime_izhodne_datoteke)));
napaka1:
    return stanje;
}


T_STANJE RAnalizator::NaloziKatalog(TK_NIZ ime_datoteke)
{
    T_STANJE stanje;
    T_ROCICA rocica = katalogi.UstvariVozel();
    RSeznamPrevodov &prevodi = katalogi.VrniElement(rocica);

    stanje = prevodi.NaloziIzDatoteke(ime_datoteke);
    if (NEUSPESNO(stanje)) {
        katalogi.UniciVozel(rocica);
        _ftprintf(stderr, _T("%s : error X0011: Napaka pri branju datoteke kataloga (koda %#010I32x).\n"), ime_datoteke, stanje);
        return stanje;
    }
    katalogi.DodajVozel(rocica);

    return A_OK;
}


T_STANJE RAnalizator::ShraniKatalog(TK_NIZ ime_datoteke)
{
    T_STANJE stanje;

    stanje = prevodi.ShraniVDatoteko(ime_datoteke);
    if (NEUSPESNO(stanje)) {
        _ftprintf(stderr, _T("%s : error X0005: Napaka pri shranjevanju v datoteko (koda %#010I32x).\n"), ime_datoteke, stanje);
        return stanje;
    }

    return A_OK;
}


T_REZULTAT amain(T_UINT argc, TK_NIZ argv[])
{
#ifndef AOSN_DLLIMP
    verify(USPESNO(AOsn::Vzpostavi()));
#endif

    // Nastavimo locale modulu.
    asetlocale(LC_ALL, _T(".OCP"));

    T_REZULTAT rezultat = 0;

    if (argc < 2) {
        _ftprintf(stderr,
            _T("rcxgettext v") _T(RCXGETTEXT_VERZIJA_STR) _T(", Vse pravice pridrane (C) Amebis, d. o. o., 2008-") _T(BUILD_YEAR_STR) _T("\n")
            _T("Uporaba: rcxgettext <ukaz> ...\n\n")
            _T("Ukazi:\n")
            _T("rc    Poie nize za prevajanje in izdela katalog .po iz datoteke .rc\n")
            _T("rcp   Prevede datoteko .rc s prevodi iz katalogov .po\n")
            _T("mc    Poie nize za prevajanje in izdela katalog .po iz datoteke .mc\n")
            _T("mcp   Prevede datoteko .mc s prevodi iz katalogov .po\n")
            _T("idt   Poie nize za prevajanje in izdela katalog .po iz datoteke .idt\n")
            _T("idtp  Prevede datoteko .idt s prevodi iz katalogov .po\n")
            _T("idta  Analizira datoteko .idt in njen prevod ter izdela katalog .po\n")
            _T("html  Poie nize za prevajanje in izdela katalog .po iz datoteke .html\n")
            _T("htmlp Prevede datoteko .html s prevodi iz katalogov .po\n"));

        rezultat = 1;
        goto konec;
    }

    if (!astrcmp(argv[1], _T("rc")))
        rezultat = rc(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("rcp")))
        rezultat = rcp(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("mc")))
        rezultat = mc(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("mcp")))
        rezultat = mcp(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("idt")))
        rezultat = idt(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("idtp")))
        rezultat = idtp(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("idta")))
        rezultat = idta(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("html")))
        rezultat = html(argc - 1, argv + 1);
    else if (!astrcmp(argv[1], _T("htmlp")))
        rezultat = htmlp(argc - 1, argv + 1);
    else  {
        _ftprintf(stderr, _T("error X0013: Napaen ukaz!\n"));
        rezultat = 1;
    }

konec:
    verify(USPESNO(AOsn::PocakajVzpostavitev()));
#ifndef AOSN_DLLIMP
    verify(USPESNO(AOsn::Zakljuci()));
#endif
    adumpheap();
    return rezultat;
}
